
EventManager
	subclass: #SimpleHTTPServer
	instanceVariableNames: ' host port socket httpHandlers callback crlf'
	category: #SimpleHTTPServer !
SimpleHTTPServer comment: '
	Simple, pluggable non-blocking HTTP server.
	@2019-today Alejandro Reimondo aleReimondo@smalltalking.net
	Published under U8 license terms see http://u8.smalltalking.net
	See rfc2616 for more information about this HTTP server implementation
'!
SimpleHTTPServer class instanceVariableNames: 'current socketSupport'!
SimpleHTTPServer class buildAccessors: #(
	#(current initialize 'self new')
	#(socketSupport initialize 'self detectSocketSupport')
)!

! SimpleHTTPServer class methodsFor: #socketSupport !
detectSocketSupport
	" Private - Returns a valid socket implementation (or nil) "

	| api |
	(Smalltalk includesKey: #LuaStudioLibrary) ifTrue: [ ^#luaStudio ].
	(Smalltalk includesKey: #LuaSocket) ifTrue: [
		[ api := LuaSocket nativeClass ] on: Error do: [].
		api notNil ifTrue: [ ^#luaSocket ].
	].
	^nil! !

! SimpleHTTPServer methodsFor: #implementation !
support
	" Private - Returns the sockets support to run the server. "

	| api |
	api := self class socketSupport.
	api isNil ifTrue: [ self error: 'Missing socket support' ].
	^api! !

! SimpleHTTPServer methodsFor: #implementation !
hostAt: address
	" Private - Returns the host at address."

	self support = #luaStudio ifTrue: [
		address isEmpty ifTrue: [
			System systemName = #windows ifFalse: [
				self note: 'LuaStudio API is buggy in this version (under Android os), so we need to use localhost'.
				^Host @ #{Lib}[#Sys][#Net][#Host] #_localhost
			].
		].
		^Host @ address
	].
	^address = '' ifTrue: [ $* ] ifFalse: [ address ]! !

! SimpleHTTPServer methodsFor: #implementation !
bindSocket
	" Private - Instantiate the socket and bind to host and port. "

	(self support = #luaSocket) ifTrue: [
		socket := LuaSocket bind: self host port: self port.
		^self
	].
	socket := Socket new.
	socket	bind: self host port: self port;
		listen: 30; setBlocking: false! !

! SimpleHTTPServer methodsFor: #implementation !
read: client lineOr: result
	" Private - Try to read a line on client input.
	Returns the line read or result in case of no input read.
	"

	(self support = #luaSocket) ifTrue: [ | read |
		read := client receive first.
		^read isNil ifTrue: [ result ] ifFalse: [ read ]
	].
	^client input readLineOr: result! !

! SimpleHTTPServer methodsFor: #implementation !
write: data on: client
	" Private - Send data to client output. "

	(self support = #luaSocket) ifTrue: [
		^client send: data
	].
	^client write: data! !

! SimpleHTTPServer methodsFor: #implementation !
pumpRegistry
	" Private - Returns a collection where the callback is added to listen for client requests.
	The collection must be evaluated by the mail loop to maintain the receiver listening for clients.
	In case of an empty collection, the application must send #listen frequently to maintain clients happy.
	"

	Smalltalk at: #View3D ifAbsent: [ ^#() ].
	^View3D preRenderActions! !

SimpleHTTPServer buildAccessors: #(
	#(host initialize '(self hostAt: String new)')
	#(port initialize '8088')
	#(callback initialize '[:ignored| self listen. nil ]')
	#(httpHandlers initialize 'self needsHTTPHandlers')
)!
SimpleHTTPServer buildGetters: #( crlf )!

! SimpleHTTPServer class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^#(	#postProcess:with:
		#send:response:with:
		#needsRequestFor:with:
		#needsHandlerFor:with:
		#acceptVersion:
		#getHttpHandlers #getHttpHandlers:
		#openResponseLines:at:with:
		#needsResource:with:
		#closeResponseLines:at:with:
		#needsContents:with:
		#needsAccessControlAllow:at:with:
		#needsContentType:at:with:
		#needsContentLength:at:with:
	) ,super constructEventsTriggered! !

! SimpleHTTPServer class methodsFor: #sample !
example
	" server - Implements a simple (non-blocking) HTTP server "

	| server |
	server := self current.
	server when: #needsContents:with: do: [:uri :headers|
		'<html>
<head><title>' ,self class name,' example</title></head>
<body><h1>Hello from S8</h1>
<hr>Headers:<br><pre>' ,(headers keys sorted inject: '' into: [:total :key|
		total, key ,' : ' ,(headers at: key) asLiteral ,'<br>'
]),'</pre>
</body></html>'
	].
	server when: #postProcess:with: do: [:start :headers|
		(start notNil
			and: [ start first = #GET
			and: [ #('/quit' '/exit' '/stop') includes: start second ]]) ifTrue: [
			server stop.
			self note: 'server stop'
		]
	].
	server start.
	self note: 'server started at ' ,server host toString ,':' ,server port.
	^server! !

! SimpleHTTPServer methodsFor: #private !
startListening
	" Private - Start listening for clients. "

	self pumpRegistry add: self callback.! !

! SimpleHTTPServer methodsFor: #private !
stopListening
	" Private - Stop listening for clients. "

	self pumpRegistry remove: self callback ifAbsent: [].! !

! SimpleHTTPServer methodsFor: #configuration !
ip: ipAddress
	" Set the receiver's host at IP adress. "

	self host: (self hostAt: ipAddress)! !

! SimpleHTTPServer methodsFor: #query !
isRunning
	" Returns true if the receiver is runnning. "

	^socket notNil! !

! SimpleHTTPServer methodsFor: #running !
run
	" Run the receiver in a loop until stopped. "

	[ self isRunning ] whileTrue: [ self listen ]! !

! SimpleHTTPServer methodsFor: #running !
stop
	" Stop the receiver. "

	socket isNil ifTrue: [ ^self ].
	self stopListening.
	socket close.
	socket := nil.! !

! SimpleHTTPServer methodsFor: #private !
peek
	" Private - Peek a client. "

	^socket notNil ifTrue: [
		socket settimeout: 1/1000.
		socket acceptOrNil
	]! !

! SimpleHTTPServer methodsFor: #running !
start
	" Start the receiver. "

	socket notNil ifTrue: [ self close ].
	crlf := 13 asCharacter ,10 asCharacter.
	self	bindSocket;
		startListening.! !

! SimpleHTTPServer methodsFor: #private !
step
	" Private - Listen a client request. "

	| client |
	self isRunning ifFalse: [ ^false ].
	client := self peek.
	client isNil ifTrue: [ ^false ].
	client setBlocking: false.
	^self processMessageOn: client! !

! SimpleHTTPServer methodsFor: #private !
listen
	" Private - Listen client requests. "

	[ self step ] whileTrue! !

! SimpleHTTPServer methodsFor: #processing !
processMessageOn: client
	" Private - Process a client message.
	https://tools.ietf.org/html/rfc2616#section-4
	HTTP messages consist of requests from client to server and responses from server to client.
	Both types of message consist of a start-line, zero or more header fields (also known as headers), an empty line (i.e., a line with nothing preceding the CRLF) indicating the end of the header fields, and possibly a message-body.
	"

	| start line headers |
	[	line := self read: client lineOr: ''.
		line isEmpty ] whileTrue: [
		self note: self class name ,'>>#processMessageOn: start - readLine failed'.
		self write: self badRequest on: client.
		client close. ^false
	].
	start := self startLine: line.
	headers := PoolDictionary new.
	[	line := self read: client lineOr: #failed.
		line notEmpty ] whileTrue: [
		line = #failed
		ifTrue: [ self note: self class name ,'>>#processMessageOn: header - readLine failed' ]
		ifFalse: [ self addHeader: line to: headers ]
	].
	self send: client responseFor: start with: headers.
	self postProcess: start with: headers.
	^true! !

! SimpleHTTPServer methodsFor: #processing !
send: client responseFor: start with: headers
	" Private - Send the response to client for request start. "

	self	triggerEvent: #send:response:with:
		withArguments: (Array with: client with: start with: headers)
		ifNotHandled: [
			self write: (self responseFor: start with: headers) on: client.
			client close
		]! !

! SimpleHTTPServer methodsFor: #processing !
addHeader: line to: headers
	" Private - Add a request header to headers.
	https://tools.ietf.org/html/rfc2616#section-5.3
	The request-header fields allow the client to pass additional information about the request, and about the client itself, to the server.
	"

	| stream key value |
	stream := line stream.
	key := (stream upTo: $:) trimBlanks.
	value := stream upToEnd trimBlanks.
	(headers includesKey: key) ifFalse: [
		^headers at: key put: value
	].
	(headers at: key) isArray ifFalse: [
		headers at: key put: (Array with: (headers at: key))
	].
	^headers at: key put: value! !

! SimpleHTTPServer methodsFor: #processing !
startLine: line
	" Private - Returns the result of parsing a request start line.
	An array containing:
		Method		https://tools.ietf.org/html/rfc2616#section-5.1.1
		Request-URI	https://tools.ietf.org/html/rfc2616#section-5.1.2
		version
	"

	| stream method uri |
	stream := line stream.
	method := stream upTo: $ .
	uri := stream upTo: $ .
	(stream upTo: $/) = #HTTP ifFalse: [ ^nil ]. "not a start request line"
	^Array with: method with: uri with: stream upToEnd! !

! SimpleHTTPServer methodsFor: #processing !
postProcess: start with: headers
	" Private - Post process a request. "

"DEBUG- true ifTrue: [
# #print: #DEBUG with: #postProcessing.
^self stop
]."

	self	triggerEvent: #postProcess:with:
		withArguments: (Array with: start with: headers)! !

! SimpleHTTPServer methodsFor: #processing !
responseFor: start with: headers
	" Private - Returns the response string to be sent to a client requesting message with start and headers. "

	start isNil ifTrue: [ ^self badRequest ].
	start last = '1.1' ifFalse: [
		self	triggerEvent: #acceptVersion:
			withArguments: (Array with: start)
			ifNotHandled: [ ^self versionNotSupported: start last ]
	].
	^self	triggerEvent: #needsRequestFor:with:
		withArguments: (Array with: start with: headers)
		ifNotHandled: [ self defaultResponseFor: start with: headers ]! !

! SimpleHTTPServer methodsFor: #processing !
defaultResponseFor: start with: headers
	" Private - Returns the response string to be sent to a client requesting message with start and headers. "

	| block |
	block := self
		httpHandlerFor: start and: headers
		ifAbsent: [ [:s :h| self methodNotImplemented ] ].
	^block evaluateWithArguments: (Array with: start with: headers)! !

! SimpleHTTPServer methodsFor: #responses !
badRequest
	" Private - Returns the response string for a bad request.
	https://tools.ietf.org/html/rfc2616#section-10.4.1
	The request could not be understood by the server due to malformed syntax.
	The client SHOULD NOT repeat the request without modifications.
	"

	^'HTTP/1.1 400 Bad Request'! !

! SimpleHTTPServer methodsFor: #responses !
methodNotImplemented
	" Private - Returns the response string for an unimplemented request.
	https://tools.ietf.org/html/rfc2616#section-10.5.2
	The server does not support the functionality required to fulfill the request.
	This is the appropriate response when the server does not recognize the request method and is not capable of supporting it for any resource.
	"

	^'HTTP/1.1 501 Not Implemented'! !

! SimpleHTTPServer methodsFor: #responses !
versionNotSupported:  version
	" Private - Returns the response string for an unsupported HTTP version.
	https://tools.ietf.org/html/rfc2616#section-10.5.6
	"

	^'HTTP/1.1 505 HTTP Version Not Supported : ' ,version! !

! SimpleHTTPServer methodsFor: #configuration !
httpHandlerFor: start and: headers ifAbsent: aBlock
	" Private - Returns the handler of request messages with start and headers (or the result of evaluating aBlock). "

	^self	triggerEvent: #needsHandlerFor:with:
		withArguments: (Array with: start with: headers)
		ifNotHandled: [ self httpHandlers at: start first ifAbsent: aBlock ]! !

! SimpleHTTPServer methodsFor: #configuration !
needsHTTPHandlers
	" Private - Returns the default HTTP handlers of the receiver. "

	httpHandlers := PoolDictionary new.
	self	triggerEvent: #getHttpHandlers ifNotHandled: [ self installDefaultHttpHandlers ];
		triggerEvent: #getHttpHandlers: with: httpHandlers.
	^httpHandlers! !

! SimpleHTTPServer methodsFor: #defaults !
installDefaultHttpHandlers
	" Private - Install the default handlers of the receiver. "

	httpHandlers
		at: #GET put: self getHandler;
		at: #HEAD put: self headHandler;
		yourself! !

! SimpleHTTPServer methodsFor: #handlers !
headHandler
	" Private - Returns the handler of HEAD request.
	https://tools.ietf.org/html/rfc2616#section-9.4
	"

	^[:start :headers| | stream |
		stream := '' stream.
		self get: start second into: stream headers: headers body: false.
		stream contents
	]! !

! SimpleHTTPServer methodsFor: #handlers !
getHandler
	" Private - Returns the handler of GET request.
	https://tools.ietf.org/html/rfc2616#section-9.3
	"

	^[:start :headers| | stream |
		stream := '' stream.
		self get: start second into: stream headers: headers body: true.
		stream contents
	]! !

! SimpleHTTPServer methodsFor: #get !
get: uri into: stream headers: headers body: includesBody
	" Private - Emmit the response of GET operation onto stream. "

	| lines statusLine body |
	lines := self requestLinesFor: uri headers: headers withBody: includesBody.
	statusLine := lines removeFirst.
	(includesBody and: [
		statusLine second asString first = $2 and: [
		lines notEmpty ]]) ifTrue: [ "status-code with shape 2xx: Success"
		body := lines removeLast
	].

	statusLine "Status-Line = HTTP-Version SP Status-Code SP Reason-Phrase CRLF"
		do: [:each| stream nextPutAll: each asString ]
		separatedBy: [ stream space ].
	stream nextPutAll: crlf.

	lines do: [:assoc| "entity-headers https://tools.ietf.org/html/rfc2616#section-7.1"
		stream	nextPutAll: assoc key;
			space; nextPutAll: ':'; space;
			nextPutAll: assoc value toString;
			nextPutAll: crlf
	].

	stream nextPutAll: crlf. "end of header"
	body notNil ifTrue: [ stream nextPutAll: body ]! !

! SimpleHTTPServer methodsFor: #get !
requestLinesFor: uri headers: headers withBody: includesBody
	" Private - Returns the request lines of resource at uri. "

	| result resource |
	result := Array new.
	self	triggerEvent: #openResponseLines:at:with:
		withArguments: (Array with: result with: uri with: headers).
	resource := self
		triggerEvent: #needsResource:with:
		withArguments: (Array with: uri with: headers)
		ifNotHandled: [
			self dynamicResource: uri headers: headers withBody: includesBody
		].
	resource isNil ifTrue: [ ^#( #('HTTP/1.1' 404 'Not Found') ) ].
	self	addAccessControlAllow: resource at: uri with: headers to: result;
		addContentTypeAndLength: resource at: uri with: headers to: result.
	self	triggerEvent: #closeResponseLines:at:with:
		withArguments: (Array with: result with: uri with: headers).
	includesBody ifTrue: [
		result add: (self bodyOf: resource at: uri with: headers)
	].
	result addFirst: #('HTTP/1.1' 200 #OK). "The request has succeeded."
	^result! !

! SimpleHTTPServer methodsFor: #get !
dynamicResource: uri headers: headers withBody: includesBody
	" Private - Returns the resource computed for uri (or nil).
	The result MUST be aString, nil or an Array with shape
		#( mime-type bodyString [Array of Associations] )
	"

	^self	triggerEvent: #needsContents:with:
		withArguments: (Array with: uri with: headers)
		ifNotHandled: [ nil ]! !

! SimpleHTTPServer methodsFor: #get !
addAccessControlAllow: resource at: uri with: headers to: result
	" Private - Add this entity header fields of resource to result. "

	| tuples |
	tuples := self
		triggerEvent: #needsAccessControlAllow:at:with:
		withArguments: (Array with: resource with: uri with: headers)
		ifNotHandled: [ Array with: #Access-Control-Allow-Origin -> $* ].
	result addAll: tuples! !

! SimpleHTTPServer methodsFor: #get !
addContentTypeAndLength: resource at: uri with: headers to: result
	" Private - Add this entity header fields of resource to result. "

	| tuples |
	tuples := self
		triggerEvent: #needsContentType:at:with:
		withArguments: (Array with: resource with: uri with: headers)
		ifNotHandled: [ Array with: ('Content-Type' -> (self contentTypeOf: resource at: uri with: headers)) ].
	result addAll: tuples.
	tuples := self
		triggerEvent: #needsContentLength:at:with:
		withArguments: (Array with: resource with: uri with: headers)
		ifNotHandled: [ Array with: ('Content-Length' -> (self contentLengthOf: resource at: uri with: headers)) ].
	result addAll: tuples! !

! SimpleHTTPServer methodsFor: #get !
bodyOf: resource at: uri with: headers
	" Private - Returns the body of resource. "

	resource isString ifTrue: [ ^resource ].
	^resource second! !
	
! SimpleHTTPServer methodsFor: #get !
contentLengthOf: resource at: uri with: headers
	" Private - Returns this header field of resource. "

	| bytes |
	bytes := resource isString ifTrue: [ resource ] ifFalse: [ resource second ].
	^bytes size ! !
	
! SimpleHTTPServer methodsFor: #get !
contentTypeOf: resource at: uri with: headers
	" Private - Returns this header field of resource. "

	resource isString ifTrue: [
		^(self guessTypeOf: resource at: uri with: headers)
	].
	^resource first! !
	
! SimpleHTTPServer methodsFor: #get !
guessTypeOf: resource at: uri with: headers
	" Private - Return the type of resource (a String). "

	(resource startsWith: $<) ifTrue: [ ^'text/html' ].
	^'text/plain'! !
