" Wrappers to Network support for the Lua language.
Copyright (C) 2019-today Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
See also http://w3.impa.br/~diego/software/luasocket/home.html
"!

NativeObject subclass: #LuaSocketObject category: #luasocket!
#(	DNS FTP HTTP
	LTN12 LTN12Filter LTN12Pump LTN12Sink LTN12Source
	SMTP URL
) do: [:each| LuaSocketObject subclass: each ]!

LuaSocketObject subclass: #LuaSocket!
LuaSocket subclass: #TCP!
LuaSocket subclass: #UDP!
! Lua class methodsFor: #accessing !
socket
	" Returns the LuaSocket library. "

	^LuaSocket! !

! Lua class methodsFor: #accessing !
http
	" Returns the http library wrapper. "

	^HTTP new! !

! Lua class methodsFor: #accessing !
ftp
	" Returns the ftp library wrapper. "

	^FTP new! !

! Lua class methodsFor: #accessing !
smtp
	" Returns the smtp library wrapper. "

	^SMTP new! !

! LuaSocket class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^Lua @ (# #require: #socket)! !

! DNS class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^Lua @ (LuaSocket nativeClass @ #dns)! !

! FTP methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	self handle: (# #require: 'socket.ftp').
	handle isNil ifTrue: [ self error: 'Missing FTP support' ] ! !

! HTTP methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	self handle: (# #require: 'socket.http').
	handle isNil ifTrue: [ self error: 'Missing HTTP support' ] ! !

! SMTP methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	self handle: (# #require: 'socket.smtp').
	handle isNil ifTrue: [ self error: 'Missing SMTP support' ] ! !

! LuaSocket class methodsFor: #select !
select: recSocks send: sendSocks timeout: timeout
	" Waits for a number of sockets to change status. "

	| reads writes err |
{'	reads, writes, err = self:nativeClass():handle():select( recSocks:handles(), sendSocks:handles(), timeout) '}.
	^Array
		with: (Array@reads collect: [:each| LuaSocket @ each ])
		with: (Array@writes collect: [:each| LuaSocket @ each ])
		with: err! !

! LuaSocket class methodsFor: #select !
select: recSocks send: sendSocks
	" Waits for a number of sockets to change status. "

	| reads writes err |
{'	reads, writes, err = self:nativeClass():handle():select( recSocks:handles(), sendSocks:handles()) '}.
	^Array
		with: (Array@reads collect: [:each| LuaSocket @ each ])
		with: (Array@writes collect: [:each| LuaSocket @ each ])
		with: err! !

! LuaSocket class methodsFor: #select !
select: recSocks timeout: timeout
	" Waits for a number of sockets to change status. "

	| reads err |
{'	reads, _, err = self:nativeClass():handle():select( recSocks:handles(), nil, timeout) '}.
	^Array
		with: (Array@reads collect: [:each| LuaSocket @ each ])
		with: err! !

! LuaSocket class methodsFor: #select !
select: recSocks
	" Waits for a number of sockets to change status. "

	| reads err |
{'	reads, _, err = self:nativeClass():handle():select( recSocks:handles(), nil) '}.
	^Array
		with: (Array@reads collect: [:each| LuaSocket @ each ])
		with: err! !

! LuaSocket methodsFor: #ignored !
setBlocking: blocking
	" Ignore blocking setup. "! !

LuaSocket buildLuaStaticFunctions: #(
	#(tcp #() TCP)
	#(bind #(address port) TCP)
	#(bind #(address port backlog) TCP)
	#(connect #(address port) TCP)
	#(connect #(address port localaddr locport) TCP)
	#(select #(#(recvt handles) #(sendt handles)))
	#(select #(#(recvt handles) #(sendt handles) timeout))
	#(sink #(mode _socket) LTN12Sink)
	#(source #(mode _socket) LTN12Source)
	#(source #(mode _socket length) LTN12Source)
	#(udp #() UDP)
	#(newtry #(finalBlock))
	#(protect #(aBlock))
	#(sleep #(time))
	gettime
)!
LuaSocket buildFunctions: #(
	close
	#(setoption #(option))
	#(setoption #(option value))
	#(setoption #(option to))
	#(settimeout #(value))
)!

! LuaSocket methodsFor: #receiving !
receive
	| s status partial |
{'	s, status, partial = self:handle():receive() '}.
	^Array with: s with: status with: partial! !

! LuaSocket methodsFor: #accessing !
sockname
	| ip port |
{'	ip, port = self:handle():getsockname()
	if ip==nil then return nil end
'}.	^Array with: ip with: port! !

! LuaSocket methodsFor: #accessing !
peername
	| ip port |
{'	ip, port = self:handle():getpeername()
	if ip==nil then return nil end
'}.	^Array with: ip with: port! !

! LuaSocket class methodsFor: #constants !
version

	^self nativeClass @ #_VERSION! !

DNS buildLuaStaticFunctions: #(
	gethostname
	#(tohostname #(address))
	#(toip #(address))
)!

HTTP buildLuaFunctions: #(
	#(request #(url body))
	#(request #(_urlOrOptions))
)!

! HTTP class methodsFor: #access !
request: args
	^self new request: args! !

! HTTP class methodsFor: #access !
get: args
	^self new request: args! !

! HTTP methodsFor: #request !
requestAll: options
	| result count headers |
{'	result, count, headers = self:handle().request(options:json())'}.
	^Array with: result with: count with: PoolDictionary @ headers ! !

! HTTP class methodsFor: #access !
requestAll: options
	^self new requestAll: options! !

! FTP methodsFor: #get !
get: urlOrOptions onError: aBlock
	| result error |
{'	result, error = self:handle().get(urlOrOptions:json())
	if error then return aBlock:evaluateWith_(error) end'}.
	^result! !

! FTP methodsFor: #get !
get: urlOrOptions
	^self get: urlOrOptions onError: [:err| self error: err ]! !

! FTP methodsFor: #put !
put: options onError: aBlock
	| result error |
{'	result, error = self:handle().put(options:json())
	if error then return aBlock:evaluateWith_(error) end'}.
	^result! !

! FTP methodsFor: #put !
put: url contents: contents
	^self put: url contents: contents onError: [:err| self error: err ]! !

! FTP methodsFor: #put !
put: options
	^self put: options onError: [:err| self error: err ]! !

! FTP methodsFor: #put !
put: url contents: contents onError: aBlock
	| result error |
{'	result, error = self:handle().put(url,contents)
	if error then return aBlock:evaluateWith_(error) end'}.
	^result! !

! FTP class methodsFor: #get !
get: args
	^self new get: args! !

! FTP class methodsFor: #put !
put: options
	^self new put: options! !

! FTP class methodsFor: #put !
put: url contents: contents
	^self new put: url contents: contents! !

! FTP methodsFor: #get !
get: host
	sink: aLTN12Sink
	with: argumentOrPathString
	user: user password: password
	command: command
	port: port type: type
	step: step create: aBlock

	| options |
	options := #host -> host
		,(#sink -> aLTN12Sink)
		,(#path -> argumentOrPathString).
	user notNil ifTrue: [ options := options ,(#user -> user) ].
	password notNil ifTrue: [ options := options ,(#password -> password) ].
	command notNil ifTrue: [ options := options ,(#command -> command) ].
	port notNil ifTrue: [ options := options ,(#port -> port) ].
	type notNil ifTrue: [ options := options ,(#type -> type) ].
	step notNil ifTrue: [ options := options ,(#step -> step) ].
	aBlock notNil ifTrue: [ options := options ,(#create -> aBlock) ].
	^self get: options! !

! FTP methodsFor: #get !
get: host
	sink: aLTN12Sink
	with: argumentOrPathString
	user: user password: password
	command: command
	port: port type: type
	step: step

	^self get: host
		sink: aLTN12Sink
		with: argumentOrPathString
		user: user password: password
		command: command
		port: port type: type
		step: step create: nil! !

! FTP methodsFor: #get !
get: host
	sink: aLTN12Sink
	with: argumentOrPathString
	user: user password: password
	command: command
	port: port type: type

	^self get: host
		sink: aLTN12Sink
		with: argumentOrPathString
		user: user password: password
		command: command
		port: port type: type
		step: nil create: nil! !

! FTP methodsFor: #get !
get: host
	sink: aLTN12Sink
	with: argumentOrPathString
	user: user password: password
	command: command
	port: port

	^self get: host
		sink: aLTN12Sink
		with: argumentOrPathString
		user: user password: password
		command: command
		port: port type: nil step: nil create: nil! !

! FTP methodsFor: #get !
get: host
	sink: aLTN12Sink
	with: argumentOrPathString
	user: user password: password
	command: command

	^self get: host
		sink: aLTN12Sink
		with: argumentOrPathString
		user: user password: password
		command: command
		port: nil type: nil step: nil create: nil! !

! FTP methodsFor: #get !
get: host
	sink: aLTN12Sink
	with: argumentOrPathString
	user: user password: password 

	^self get: host
		sink: aLTN12Sink
		with: argumentOrPathString
		user: user password: password
		command: nil port: nil type: nil step: nil create: nil! !

! FTP methodsFor: #get !
get: host
	sink: aLTN12Sink
	with: argumentOrPathString

	^self get: host
		sink: aLTN12Sink
		with: argumentOrPathString
		user: nil password: nil! !

! FTP methodsFor: #put !
put: host
	source: aLTN12Sink
	with: argumentOrPathString
	user: user password: password
	command: command
	port: port type: type
	step: step create: aBlock

	| options |
	options := #host -> host
		,(#source -> aLTN12Sink)
		,(#path -> argumentOrPathString).
	user notNil ifTrue: [ options := options ,(#user -> user) ].
	password notNil ifTrue: [ options := options ,(#password -> password) ].
	command notNil ifTrue: [ options := options ,(#command -> command) ].
	port notNil ifTrue: [ options := options ,(#port -> port) ].
	type notNil ifTrue: [ options := options ,(#type -> type) ].
	step notNil ifTrue: [ options := options ,(#step -> step) ].
	aBlock notNil ifTrue: [ options := options ,(#create -> aBlock) ].
	^self put: options! !

! FTP methodsFor: #put !
put: host
	source: aLTN12Sink
	with: argumentOrPathString
	user: user password: password
	command: command
	port: port type: type
	step: step

	^self put: host
		source: aLTN12Sink
		with: argumentOrPathString
		user: user password: password
		command: command
		port: port type: type
		step: step create: nil! !

! FTP methodsFor: #put !
put: host
	source: aLTN12Sink
	with: argumentOrPathString
	user: user password: password
	command: command
	port: port type: type

	^self put: host
		source: aLTN12Sink
		with: argumentOrPathString
		user: user password: password
		command: command
		port: port type: type
		step: nil create: nil! !

! FTP methodsFor: #put !
put: host
	source: aLTN12Sink
	with: argumentOrPathString
	user: user password: password
	command: command
	port: port

	^self put: host
		source: aLTN12Sink
		with: argumentOrPathString
		user: user password: password
		command: command
		port: port type: nil
		step: nil create: nil! !

! FTP methodsFor: #put !
put: host
	source: aLTN12Sink
	with: argumentOrPathString
	user: user password: password
	command: command

	^self put: host
		source: aLTN12Sink
		with: argumentOrPathString
		user: user password: password
		command: command
		port: nil type: nil
		step: nil create: nil! !

! FTP methodsFor: #put !
put: host
	source: aLTN12Sink
	with: argumentOrPathString
	user: user password: password

	^self put: host
		source: aLTN12Sink
		with: argumentOrPathString
		user: user password: password
		command: nil
		port: nil type: nil step: nil create: nil! !

! FTP methodsFor: #put !
put: host
	source: aLTN12Sink
	with: argumentOrPathString

	^self put: host
		source: aLTN12Sink
		with: argumentOrPathString
		user: nil password: nil command: nil
		port: nil type: nil step: nil create: nil! !

! HTTP methodsFor: #request !
request: url
	method: method headers: headers
	sink: aLTN12Sink source: source step: step
	proxy: proxy redirect: redirect
	create: aBlock

	| options |
	options := #url -> url.
	method notNil ifTrue: [ options := options ,(#method -> method) ].
	headers notNil ifTrue: [ options := options ,(#headers -> headers) ].
	step notNil ifTrue: [ options := options ,(#step -> step) ].
	aLTN12Sink  notNil ifTrue: [ options := options ,(#sink -> aLTN12Sink ) ].
	source notNil ifTrue: [ options := options ,(#source ->source) ].
	step notNil ifTrue: [ options := options ,(#step -> step) ].
	proxy notNil ifTrue: [ options := options ,(#proxy -> proxy) ].
	redirect notNil ifTrue: [ options := options ,(#redirect -> redirect) ].
	aBlock notNil ifTrue: [ options := options ,(#create -> aBlock) ].
	^self request: options! !

! HTTP methodsFor: #request !
request: url
	method: method headers: headers
	sink: aLTN12Sink source: source step: step
	proxy: proxy redirect: redirect

	^self request: url
		method: method headers: headers
		sink: aLTN12Sink source: source step: step
		proxy: proxy redirect: redirect
		create: nil! !

! HTTP methodsFor: #request !
request: url
	method: method headers: headers
	sink: aLTN12Sink source: source step: step

	^self request: url
		method: method headers: headers
		sink: aLTN12Sink source: source step: step
		proxy: nil redirect: nil create: nil! !

! HTTP methodsFor: #request !
request: url method: method headers: headers

	^self request: url
		method: method headers: headers
		sink: nil source: nil step: nil! !

! HTTP methodsFor: #request !
request: url method: method

	^self request: url
		method: method headers: nil
		sink: nil source: nil step: nil! !

! Lua class methodsFor: #accessing !
ltn12
	" Returns the wrapper to ltn12. "

	^LTN12 new! !

! LTN12 methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	self handle: (# #require: #ltn12).
	handle isNil ifTrue: [ self error: 'Missing ltn12 support' ] ! !

! LTN12 methodsFor: #constants !
blocksize

	^self handle basicAt: #BLOSKSIZE! !

! LTN12 methodsFor: #constants !
version

	^self handle basicAt: #_VERSION! !

LTN12 buildTypedGetters: #(
	#(filter LTN12Filter)
	#(pump LTN12Pump)
	#(sink LTN12Sink)
	#(source LTN12Source)
)!

LTN12Filter buildLuaFunctions: #(
	#(chain #(_filter) LTN12Filter)
	#(chain #(_filter _filter2) LTN12Filter)
	#(chain #(_filter _filter2 _filter3) LTN12Filter)
	#(chain #(_filter _filter2 _filter3 _filter4) LTN12Filter)
	#(cycle #(_low) LTN12Filter)
	#(cycle #(_low ctx extra) LTN12Filter)
)!

! LTN12Filter methodsFor: #chain !
chainAll: anArray

	^self class @ ((self handle basicAt: #chain) evaluateWithArguments: anArray handles)! !

LTN12Pump buildLuaFunctions: #(
	#(all #(_source sink))
	#(step #(_source sink))
)!
LTN12Sink buildLuaFunctions: #(
	#(chain #(_filter sink) LTN12Sink)
	#(file #(fileHandle message) LTN12Sink)
	#(null #() LTN12Sink)
	#(table #(_table) LTN12Sink)
)!

! LTN12Sink methodsFor: #accessing !
forError: message

	^self class @ ((self handle basicAt: #error) evaluateWith: message)! !

LTN12Source buildLuaFunctions: #(
	#(cat #(_source1) LTN12Source)
	#(cat #(_source1 _source2) LTN12Source)
	#(cat #(_source1 _source2 _source3) LTN12Source)
	#(cat #(_source1 _source2 _source3 _source4) LTN12Source)
	#(chain #(_source _filter) LTN12Source)
	#(empty #() LTN12Source)
	#(file #(fileHandle message) LTN12Source)
	#(simplify #(_source) LTN12Source)
	#(string #(string) LTN12Source)
)!

! LTN12Source methodsFor: #accessing !
forError: message

	^self class @ ((self handle basicAt: #error) evaluateWith: message)! !

! LTN12Source methodsFor: #chain !
catAll: anArray

	^self class @ ((self handle basicAt: #cat) evaluateWithArguments: anArray handles)! !

LuaSocketObject subclass: #Mime!
! Lua class methodsFor: #accessing !
mime
	" Returns the wrapper to mime. "

	^Mime new! !

! Mime methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	self handle: (# #require: #mime).
	handle isNil ifTrue: [ self error: 'Missing mime support' ] ! !

Mime buildLuaFunctions: #(
	normalize
	#(normalize #(marker))
	#(decode #(type) Mime)
	#(encode #(type) Mime)
	#(encode #(type mode) Mime)
	#(stuff #() Mime)
	#(wrap #(type) Mime)
	#(wrap #(type length) Mime)
	#(b64 #(c))
	#(b64 #(c d))
	#(dot #(m))
	#(dot #(m b))
	#(eol #(c))
	#(eol #(c d marker))
	#(qp #(c))
	#(qp #(c d marker))
	#(qpwrp #(n))
	#(qpwrp #(n b length))
	#(unb64 #(c))
	#(unb64 #(c d))
	#(unqp #(c))
	#(unqp #(c d))
	#(wrp #(n))
	#(wrp #(n b length))
)!

! Mime methodsFor: #accessing !
decodeBase64

	^self decode: #base64! !

! Mime methodsFor: #accessing !
encodeBase64

	^self encode: #base64! !

! Mime methodsFor: #accessing !
decodeQuotedPrintable

	^self decode: 'quoted-printable'! !

! Mime methodsFor: #accessing !
encodeQuotedPrintable

	^self encode: 'quoted-printable'! !

! Mime methodsFor: #accessing !
wrapBase64

	^self wrap: #base64! !

! Mime methodsFor: #accessing !
wrapQuotedPrintable

	^self wrap: 'quoted-printable'! !

! Mime methodsFor: #accessing !
wrapText

	^self wrap: #text! !

! Mime methodsFor: #accessing !
wrapText: length

	^self wrap: #text length: length! !

SMTP buildLuaFunctions: #(
	#(send #(_options))
	#(message #(_mesgt) LTN12Source)
)!

! SMTP methodsFor: #sending !
send: options
	^self	send: options
		onError: [:err| self error: err ]! !

! SMTP methodsFor: #sending !
send: options onError: aBlock
	| result error |
{'	result, error = self:handle().send(options:json())
	if error then return aBlock:evaluateWith_(error) end'}.
	^true! !

! SMTP methodsFor: #send !
send: from to: rcpt source: source
	user: user password: password
	server: server port: port domain: domain
	step: step create: aBlock

	| options |
	options := #from -> from
		,(#rcpt -> rcpt)
		,(#source -> source).
	user notNil ifTrue: [ options := options ,(#user -> user) ].
	password notNil ifTrue: [ options := options ,(#password -> password) ].
	server notNil ifTrue: [ options := options ,(#server -> server) ].
	port notNil ifTrue: [ options := options ,(#port -> port) ].
	domain notNil ifTrue: [ options := options ,(#domain -> domain) ].
	step notNil ifTrue: [ options := options ,(#step -> step) ].
	aBlock notNil ifTrue: [ options := options ,(#create -> aBlock) ].
	^self send: options! !

! SMTP methodsFor: #send !
send: from to: rcpt source: source
	user: user password: password
	server: server port: port domain: domain

	^self
		send: from to: rcpt source: source
		user: user password: password
		server: server port: port domain: domain
		step: nil create: nil! !

! SMTP methodsFor: #send !
send: from to: rcpt source: source
	user: user password: password

	^self
		send: from to: rcpt source: source
		user: user password: password
		server: nil port: nil domain: nil
		step: nil create: nil! !

! SMTP methodsFor: #send !
send: from to: rcpt source: source

	^self
		send: from to: rcpt source: source
		user: nil password: nil! !

! SMTP methodsFor: #send !
message: headers body: body

	^self message: #headers -> headers ,(#body -> body)! !

! SMTP methodsFor: #send !
message: headers preamble: preamble parts: anArray epoligue: epilogue

	| parts i |
	i := 0.
	parts := anArray inject: nil into: [:total :each|
		i := i + 1.
		total isNil ifTrue: [ i -> each ]
		ifFalse: [ total ,(i -> each )]
	].
	^self message: #headers -> headers
		,(#body -> ((#preamble -> preamble)
			,parts
			,(#epilogue -> epilogue)))! !

TCP buildFunctions: #(
	#(accept TCP)
	#(bind #(address port) Boolean)
	#(connect #(address port) Boolean)
	#(getstats #() PoolDictionary)
	#(listen #(backlog) Boolean)
	#(send #(_data))
	#(send #(_data i))
	#(send #(_data i j))
	#(setstats #(received sent age))
	#(settimeout #(value mode))
	shutdown
	#(shutdown #(mode))
)!

! TCP methodsFor: #accept !
acceptOrNil
	| result err |
{'	result, err = self:handle():accept() '}.
	^TCP @ result! !

! TCP methodsFor: #receiving !
receive: pattern
	| s status partial |
{'	s, status, partial = self:handle():receive(pattern) '}.
	^Array with: s with: status with: partial! !

UDP buildFunctions: #(
	#(receive #(size))
	#(send #(_datagram))
	#(sendto #(_datagram ip port))
	#(setpeername #(address))
	#(setpeername #(address port))
	#(setsockname #(address port))
)!

! UDP methodsFor: #receiving !
receive: size
	| s status |
{'	s, status = self:handle():receive(size) '}.
	^Array with: s with: status! !

! UDP methodsFor: #receiving !
receivefrom: size
	| s status |
{'	s, status = self:handle():receivefrom(size) '}.
	^Array with: s with: status! !

! UDP methodsFor: #receiving !
receivefrom
	| s status |
{'	s, status = self:handle():receivefrom() '}.
	^Array with: s with: status! !

! UDP methodsFor: #sending !
send: datagram onError: aBlock
	| result error |
{'	result, error = self:send_(datagram,ip,port)
	if error then return aBlock:evaluateWith_(error) end'}.
	^true! !

! UDP methodsFor: #sending !
sendto: datagram ip: ip port: port onError: aBlock
	| result error |
{'	result, error = self:sendto_ip_port_(datagram,ip,port)
	if error then return aBlock:evaluateWith_(error) end'}.
	^true! !

! URL class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^Lua @ (# #require: 'socket.url')! !

URL buildLuaStaticFunctions: #(
	#(absolute #(base relative))
	#(build #(_parsed))
	#(escape #(content))
	#(parse #(url) Lua)
	#(parse #(url default) Lua)
	#(unescape #(content))
)!