" Instantiation from JSON for exchange between Smalltalk systems.
  @2018 Alejandro F. Reimondo aleReimondo@smalltalking.net
 See license.txt file
"

Object
	subclass: #Objectifier
	instanceVariableNames: 'origin objects names transforms natives useJSON'
	classVariableNames: '' poolDictionaries: ''
	category: #Objectifier !

! Objectifier class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^super constructEventsTriggered
	,#(	#needsParser
		#getNames:from:
		#getTransforms:
		#adapt:shape:
	)! !

! Objectifier class methodsFor: #converting !
@ aLiteral
	" Returns a copy of an object sent from a remote system stringified into aLiteral. "

	^self new objectFrom: aLiteral! !

! Objectifier class methodsFor: #instantiation !
from: source
	" Returns an instance of the receiver. "

	^self new
		origin: source;
		yourself! !

! Objectifier methodsFor: #accessing !
natives
	" Private - Returns the natives of the receiver. "

	natives isNil ifTrue: [ self initializeNatives ].
	^natives! !

! Objectifier methodsFor: #constants !
platform
	" Private - Returns the platform selector. "

	(Smalltalk includesKey: #OrderedCollection) ifTrue: [ ^#VSE ].
	^#S8! !

! Objectifier methodsFor: #constants !
tagStart
	" Private - Returns the tag start mask. "

	^' '! !

! Objectifier methodsFor: #reading !
objectFrom: aString
	" Returns the next object read from aString literal. "

	^self unMap: (self mapFrom: aString)! !

! Objectifier methodsFor: #parsing !
mapFrom: aString
	" Private - Returns the map read from aString literal. "

	| parser |
	useJSON := false.
	parser := self triggerEvent: #needsParser ifNotHandled: [
		self tryJSON: aString onSuccess: [:result| ^result ].
		self
	].
	^parser parse: aString! !

! Objectifier methodsFor: #initialize !
initializeNatives
	" Private - Initialize the native accessors. "

	natives := PoolDictionary new.
	self platform == #VSE
		ifTrue: [ self initializeVSEAccess ]
		ifFalse: [ self initializeS8Access ]! !

! Objectifier methodsFor: #natives !
poolIsObject
	" Private - Returns the native accessor for isObject when using PoolDictionaries. "

	^[:obj|
		(obj isMemberOf: PoolDictionary) not or: [
			(obj at: #objectifier_isObject ifAbsent: []) ~= self
		]
	]! !

! Objectifier methodsFor: #natives !
s8IsObject
	" Private - Returns the native accessor for isObject. "

	^[:obj|
		(super isObject: obj) not ifTrue: [
			natives removeKey: #isObject.
			true
		] ifFalse: [
			natives at: #isObject put: self poolIsObject.
			self poolIsObject value: obj
		]
	]! !

! Objectifier methodsFor: #initialize !
initializeVSEAccess
	" Private - Initialize the native accessors. "

	natives
		at: #get put: [:obj :key|
			useJSON
			ifTrue:  [ obj basicAt: key ]
			ifFalse: [ obj at: key ifAbsent: [nil] ]
		];
		at: #isObject put: self poolIsObject;
		at: #startsWith put: [:obj :key|
			obj size >= key size and: [
				(obj copyFrom: 1 to: key size) = key
			]
		];
		yourself! !

! Objectifier methodsFor: #initialize !
initializeS8Access
	" Private - Initialize the native accessors. "

	natives
		at: #get put: [:obj :key|
			useJSON
			ifTrue:  [ obj basicAt: key ]
			ifFalse: [ obj at: key ifAbsent: [nil] ]
		];
		at: #isObject put: self s8IsObject;
		yourself! !

! Objectifier methodsFor: #natives !
value: access for: object with: argument
	" Private - Returns the value of access the (native)object with argument. "

	^(self natives at: access) value: object value: argument! !

! Objectifier methodsFor: #natives !
readStream: source
	" Private - Returns a read stream on source. "

	| support |
	support := Smalltalk at: #ReadStream ifAbsent: [ ^source stream ].
	^support on: source! !

! Objectifier methodsFor: #natives !
writeStreamOn: container
	" Private - Returns a write stream on container. "

	| support |
	support := Smalltalk at: #WriteStream ifAbsent: [ ^container writeStream ].
	^support on: container! !

! Objectifier methodsFor: #natives !
isObject: object
	" Private - Returns true if object is a (non-native)object. "

	| block |
	block := self natives at: #isObject ifAbsent: [ ^super isObject: object ].
	^block value: object! !

! Objectifier methodsFor: #natives !
starts: string with: key
	" Private - Returns true if string starts with key. "

	| block |
	block := self natives at: #startsWith ifAbsent: [ ^string startsWith: key ].
	^block value: string value: key! !

! Objectifier methodsFor: #natives !
basic: object at: key
	" Private - Returns the result of indexing the (native)object with a key. "

	^self value: #get for: object with: key! !

! Objectifier methodsFor: #parsing !
tryJSON: aString onSuccess: aBlock
	" Private - Try to read a map from aString literal using JSON support.
	Evaluates aBlock on success.
	"

	| json result |
	(Smalltalk respondsTo: #jsObjectAt:ifAbsent:) ifTrue: [
		json := Smalltalk jsObjectAt: #JSON ifAbsent: []
	].
	json notNil ifTrue: [
		useJSON := true.
		result := (self basic: json at: #parse) value: aString.
		^aBlock value: result
	].
	^nil! !

! Objectifier methodsFor: #names !
names
	" Private - Returns the well known names map. "

	names notNil ifTrue: [ ^names ].
	names := PoolDictionary new.
	self platform == #VSE
		ifTrue: [ self setDefaultVSENames ]
		ifFalse: [ self setDefaultS8Names ].
	self	triggerEvent: #getNames:from:
		withArguments: (Array with: names with: origin).
	^names! !

! Objectifier methodsFor: #initialize !
setDefaultS8Names
	" Private - Register well known names used for instantiation mappings. "

	names	at: #OrderedCollection put: #Array;
		at: #Set put: #HashedSet;
		at: #Character put: [:json| self instantiateCharacter: json ];
		at: #TimeStamp put: #DateTime;
		yourself! !

! Objectifier methodsFor: #initialize !
setDefaultVSENames
	" Private - Register well known names used for instantiation mappings. "

	names	at: #DateTime put: #TimeStamp;
		yourself! !

! Objectifier methodsFor: #transforms !
transforms
	" Private - Returns the transforms used to fill concrete instances from structure. "

	transforms notNil ifTrue: [ ^transforms ].
	transforms := PoolDictionary new.
	self platform == #VSE
		ifTrue: [ self setDefaultVSETransforms ]
		ifFalse: [ self setDefaultS8Transforms ].
	self triggerEvent: #getTransforms.
	^transforms! !

! Objectifier methodsFor: #initialize !
setDefaultS8Transforms
	" Private - Register default transforms to fill instances from structures. "

	| dateTimeSupport |
	dateTimeSupport := Smalltalk at: #DateTime.
	transforms
		at: #Array put: [:oid :object :json :varNames | | contents |
			contents := self basic: json at: #contents.
			contents isNil ifTrue: [ self error: 'Missing Array contents' ].
			1 to: contents size do: [:i|
				object at: i put: (self object: (contents at: i))
			].
		];
		at: #String put: [:oid :object :json :varNames | "nothing to do" ];
		at: #HashedSet put: [:oid :object :json :varNames |
			object initialize: (self object: (self basic: json at: #contents))
		];
		at: #DateTime put: [:oid :object :json :varNames |
			self id: oid is: (dateTimeSupport fromString: (self basic: json at: #contents))
		];
		yourself! !

! Objectifier methodsFor: #initialize !
setDefaultVSETransforms
	" Private - Register default transforms to fill instances from structures. "

	transforms
		at: #Array put: [:oid :proto :json :varNames | | object contents |
			contents := self basic: json at: #contents.
			contents isNil ifTrue: [ self error: 'Missing Array contents' ].
			object := Array new: contents size.
			self id: oid is: object.
			1 to: contents size do: [:i|
				object at: i put: (self object: (contents at: i))
			].
		];
		at: #Set put: [:oid :object :json :varNames |
			object addAll: (self object: (self basic: json at: #contents))
		];
		at: #PoolDictionary put: [:oid :object :json :varNames |
			(self object: (self basic: json at: #associations)) do: [:assoc|
				object at: assoc key put: assoc value
			]
		];
		at: #TimeStamp put: [:oid :object :json :varNames |
			self timeStamp: object from: (self basic: json at: #contents)
		];
		yourself! !

! Objectifier methodsFor: #converting !
timeStamp: ts from: contents
	" Private - Read a TimeStamp instance from contents. "

	| aStream yyyy mm dd hh ss |
	aStream := contents asStream.
	yyyy := aStream upTo: $/.
	mm := aStream upTo: $/.
	dd := aStream upTo: $ .
	ts date: (Date fromString: dd ,'-' ,mm ,'-' ,yyyy).
	hh := (aStream upTo: $:) asInteger.
	mm := (aStream upTo: $:) asInteger.
	ss := (aStream upTo: nil) asInteger.
	ts time: (Time hours: hh minutes: mm seconds: ss)! !

! Objectifier methodsFor: #transforms !
shapeValidatingTransform
	" Private - Returns the transform to apply to fill instances from structures, validating class shape. "

	^[:oid :object :json :varNames | | transform |
		self validateClass: object class shape: varNames.
		transform := transforms at: object class name ifAbsent: [
			self error: 'Invalid shape'
		].
		transform valueWithArguments: (Array
			with: oid with: object
			with: json with: varNames)
	]! !

! Objectifier methodsFor: #transforms !
copyTransform
	" Private - Returns the transform to apply to fill instances from structures. "

	self platform == #VSE ifTrue: [ ^self copyTransformVSE ].
	^self copyTransformS8! !

! Objectifier methodsFor: #transforms !
copyTransformS8
	" Private - Returns the transform to apply to fill instances from structures. "

	^[:oid :object :json :varNames |
		1 to: varNames size do: [:i| | key value |
			key := varNames at: i.
			value := self object: (self basic: json at: i toString).
			object instVarAt: key put: value
		]
	]! !

! Objectifier methodsFor: #transforms !
copyTransformVSE
	" Private - Returns the transform to apply to fill instances from structures. "

	^[:oid :object :json :varNames | | all |
		all := object class allInstVarNames.
		1 to: all size do: [:i| | key index value |
			key := all at: i.
			index := varNames indexOf: key.
			index > 0 ifTrue: [
				value := self object: (self basic: json at: index).
				object instVarAt: i put: value
			].
		].
	]! !

! Objectifier methodsFor: #validation !
validateClass: aClass shape: varNames
	" Private - Validate shape of aClass to be mapped to varNames. "

	| all missing |
	varNames notNil ifTrue: [
		all := aClass allInstVarNames.
		(all notEmpty or: [varNames size > 0]) ifTrue: [
			missing := all reject: [:each| varNames includes: each ].
			missing notEmpty ifTrue: [
				self	triggerEvent: #adapt:shape:
					withArguments: (Array with: aClass with: varNames)
					ifNotHandled: [ self error: 'Missing class variables ' ,missing asString ]
			].
			missing := varNames reject: [:each| all includes: each ].
			missing notEmpty ifTrue: [
				self	triggerEvent: #adapt:shape:
					withArguments: (Array with: aClass with: varNames)
					ifNotHandled: [ self error: 'Missing structure variables ' ,missing asString ]
			].
		].
	].
	transforms at: aClass name put: self copyTransform! !

! Objectifier methodsFor: #converting !
unMap: aMap
	" Private - Returns the object instantiated from aMap. "

	| sender |
	sender := self basic: aMap at: #origin.
	sender notNil ifTrue: [ sender := sender asSymbol ].
	origin notNil ifTrue: [
		origin = sender ifFalse: [ self error: 'Invalid origin' ]
	] ifFalse: [ origin := sender ].
	objects isNil ifTrue: [ | count |
		count := self basic: aMap at: #id.
		objects := Array new: count.
		objects atAllPut: #missing.
	].
	^self object: (self basic: aMap at: #contents)! !

! Objectifier methodsFor: #references !
withRef: json do: aBlock
	" Private - Returns the result of evaluating aBlock with reference to known object (or json). "

	((self isObject: json) and: [json isString]) ifTrue: [
		(self starts: json with: self tagStart) ifTrue: [ | refID |
			refID := json copyFrom: self tagStart size + 1 to: json size.
			^aBlock value: (self ref: refID)
		].
	].
	^json! !

! Objectifier methodsFor: #converting !
object: json
	" Private - Returns the object instantiated from json. "

	json isNil ifTrue: [ ^nil ].
	(self isObject: json) ifFalse: [ ^self newObject: json ].
	json isArray ifTrue: [ ^json collect: [:each| self object: each ] ].
	^self withRef: json do: [:anObject| anObject ]! !

! Objectifier methodsFor: #instantiating !
instantiate: instantiator with: json allInstVarNames: varNames
	" Private - Returns a new instance from json. "

	(instantiator isKindOf: Behavior) ifFalse: [ ^instantiator value: json ].
	(instantiator inheritsFrom: Collection) ifTrue: [ ^instantiator new ].
	instantiator isClass ifFalse: [ ^self error: 'Must be a class' ].
	^instantiator basicNew ! !

! Objectifier methodsFor: #instantiating !
fill: oid object: anObject with: contents allInstVarNames: varNames
	" Private - Fill contents of anObject with contents. "

	(self transforms
		at: anObject class name
		ifAbsent: [ self shapeValidatingTransform ])
		valueWithArguments: (Array
			with: oid with: anObject
			with: contents with: varNames)! !

! Objectifier methodsFor: #instantiating !
instantiateCharacter: json
	" Private - Instantiate a character(String) literal from json structure. "

	| ascii |
	ascii := self basic: json at: 1.
	ascii isNil ifTrue: [ ascii := self basic: json at: '1' ].
	(ascii isNil or: [ ascii isNumber not ]) ifTrue: [
		^self error: 'Invalid ascii code'
	].
	^String fromCharCode: ascii! !

! Objectifier methodsFor: #converting !
newObject: json
	" Private - Returns a new instance from json. "

	| jsClass cls object varNames oid |
	jsClass := self basic: json at: #class.
	jsClass isNil ifTrue: [ ^self global: json ].

	cls := self global: jsClass.
	cls isNil ifTrue: [ self error: 'Missing class' ].
	varNames := (self isObject: jsClass) ifFalse: [ self basic: jsClass at: #allInstVarNames ].
	oid := self basic: json at: #id.

	object := self instantiate: cls with: json allInstVarNames: varNames.
	self id: oid is: object.
	self fill: oid object: object with: json allInstVarNames: varNames.
	^objects at: oid! !

! Objectifier methodsFor: #lookup !
name: aName
	" Private - Returns the (global) name to use for reference to aName. "

	^self names at: aName ifAbsent: [ aName ]! !

! Objectifier methodsFor: #lookup !
globalAt: key
	" Private - Returns the global object references by key. "

	| binding |
	key isNil ifTrue: [ ^self error: 'Missing global name' ].
	key isString ifFalse: [ "path from global reference"
		^key inject: nil into: [:result :selector|
			result isNil ifTrue: [ self globalAt: selector ]
			ifFalse: [ result perform: selector asSymbol ]
		]
	].
	binding := self name: key.
	binding isString ifFalse: [ ^binding ].
	^Smalltalk at: binding asSymbol ifAbsent: [
		self error: 'Missing global #' ,key
	]! !

! Objectifier methodsFor: #registration !
id: id is: anObject
	" Private - Register anObject as object at id. "

	id isNil ifTrue: [ ^anObject ].
	^objects at: id put: anObject! !

! Objectifier methodsFor: #references !
global: json
	" Private - Returns a global from json. "

	| result |
	self withRef: json do: [:anObject| ^anObject ].
	result := self globalAt: (self basic: json at: #global).
	^self id: (self basic: json at: #id) is: result! !

! Objectifier methodsFor: #references !
ref: ref
	" Private - Returns the object pointed at reference. "

	| value |
	(self starts: ref with: self tagStart) ifTrue: [ ^ref ].
	value := objects at: ref asNumber.
	value = #missing ifTrue: [ ^self error: 'Missing value at #' ,ref ].
	^value! !

! Objectifier methodsFor: #parsing !
parse: aString
	" Private - Returns the map read from aString literal.
	See https://www.json.org/
	"

	^self jsonObject: (self readStream: aString)! !

! Objectifier methodsFor: #parsing !
skipSeparators: stream
	" Advance the stream upto a non-separator (or end).
	Return true  if a non-separator contents is found,
	return false if run upto end of stream.
	"

	[ stream atEnd ifTrue: [ ^false ].
	  stream peek isSeparator
	] whileTrue: [ stream next ].
	^true! !

! Objectifier methodsFor: #parsing !
mustRead: char in: stream
	" Private - Ensure char is the next character to read from stream. "

	stream peek = char ifTrue: [ ^stream next ].
	(self skipSeparators: stream) ifTrue: [
		stream peek = char ifTrue: [ ^stream next ].
	].
	^self error: 'Invalid contents at ' ,(stream position + 1) asString! !

! Objectifier methodsFor: #parsing !
mustReadToken: token in: stream
	" Private - Ensure token is the next contents to read from stream. "

	| start |
	start := stream position + 1.
	token = (stream next: token size) ifFalse: [
		^self error: token ,' expected at ' ,start asString
	]! !

! Objectifier methodsFor: #parsing !
jsonObject: stream
	" Private - Returns the next json object from stream. "

	| result |
	result := PoolDictionary new.
	result at: #objectifier_isObject put: self.
	self mustRead: ${ in: stream.
	[stream peek = $}] whileFalse: [ | assoc |
		assoc := self jsonBinding: stream.
		result at: assoc key put: assoc value.
		stream peek = $, ifTrue: [ stream next ].
	].
	self mustRead: $} in: stream.
	^result! !

! Objectifier methodsFor: #parsing !
jsonBinding: stream
	" Private - Returns the next json binding from stream. "

	| key value |
	key := self jsonString: stream.
	self mustRead: $: in: stream.
	value := self jsonValue: stream.
	^Association key: key value: value! !

! Objectifier methodsFor: #instantiating !
newArray
	" Private - Returns a new extensible arrayed collection. "

	^(Smalltalk at: #OrderedCollection ifAbsent: [ Array ]) new! !

! Objectifier methodsFor: #instantiating !
numberFromString: string
	" Private - Returns the number encoded in string. "

	| support |
	support := Smalltalk at: #Float ifAbsent: [ ^string asNumber ].
	(string includes: $.) ifFalse: [ ^string asNumber ].
	^support fromString: string decimalSeparator: '.'! !

! Objectifier methodsFor: #parsing !
jsonValue: stream
	" Private - Returns the next json value from stream. "

	| char |
	char := stream peek.
	char isNil ifTrue: [ ^self error: 'Unexpected end at ' ,(stream position + 1) asString ].
	char = $" ifTrue: [ ^self jsonString: stream ].
	char isDigit ifTrue: [ ^self jsonNumber: stream ].
	char = $- ifTrue: [ stream next. ^(self jsonNumber: stream) negated ].
	char = ${ ifTrue: [ ^self jsonObject: stream ].
	char = $[ ifTrue: [ ^self jsonArray: stream ].
	char = $t ifTrue: [ self mustReadToken: 'true' in: stream. ^true ].
	char = $f ifTrue: [ self mustReadToken: 'false' in: stream. ^false ].
	char = $n ifTrue: [ self mustReadToken: 'null' in: stream. ^nil ].
	^self error: 'Corrupt contents at ' ,(stream position + 1) asString ! !

! Objectifier methodsFor: #parsing !
jsonArray: stream
	" Private - Returns the json array from stream. "

	| result |
	result := self newArray.
	self mustRead: $[ in: stream.
	[ stream peek = $] ] whileFalse: [
		result add: (self jsonValue: stream).
		stream peek = $, ifTrue: [ stream next ].
	].
	self mustRead: $] in: stream.
	^result! !

! Objectifier methodsFor: #parsing !
jsonString: stream
	" Private - Returns the json string from stream. "

	self mustRead: $" in: stream.
	^self unEscape: (self jsonStringChars: stream)! !

! Objectifier methodsFor: #parsing !
jsonStringChars: stream
	" Private - Returns the next sequence of json characters from stream. "

	| start result bar part |
	bar := $\.
	start := stream position + 1.
	[	part := stream upTo: $".
		stream atEnd ifTrue: [ ^self error: 'Missing end of string at ', start asString ].
		part notEmpty and: [ part last = bar ]
	] whileTrue: [
		part := part copyWith: $".
		result := result isNil ifTrue: [part] ifFalse: [result , part]
	].
	^result isNil ifTrue: [part] ifFalse: [result]! !

! Objectifier methodsFor: #parsing !
jsonUCharacter: fourDigits
	" Private - Returns the character escaped in fourDigits. "

	| ascii |
	ascii := (Smalltalk at: #Integer ifAbsent: [Number])
		readFrom: (self readStream: fourDigits)
		radix: 16.
	^ascii asCharacter! !

! Objectifier methodsFor: #parsing !
unEscape: string
	" Private - Returns the result of unescaping a string. "

	| input output bar |
	bar := $\.
	(string includes: bar) ifFalse: [ ^string ].
	input := self readStream: string.
	output := self writeStreamOn: ''.
	[input atEnd] whileFalse: [ | c |
		output nextPutAll: (input upTo: bar).
		input atEnd ifFalse: [
			c := input next.
			c = bar ifTrue: [ output nextPut: bar ].
			c = $/ ifTrue: [ output nextPut: $/ ].
			c = $b ifTrue: [ output nextPut: 8 asCharacter ].
			c = $f ifTrue: [ output nextPut: 12 asCharacter ].
			c = $n ifTrue: [ output nextPut: 10 asCharacter ].
			c = $r ifTrue: [ output nextPut: 13 asCharacter ].
			c = $t ifTrue: [ output nextPut: 9 asCharacter ].
			c = $u ifTrue: [ output nextPut: (self jsonUCharacter: (input next: 4)) ].
		]
	].
	^output contents! !

! Objectifier methodsFor: #parsing !
jsonNumber: stream
	" Private - Returns the json number from stream. "

	| output valid |
	output := self writeStreamOn: ''.
	['0123456789.+-eE' includes: stream peek] whileTrue: [
		output nextPut: stream next
	].
	^self numberFromString: output contents! !
