" Serialization to JSON for Object Sharing with S8 systems.
  @2018 Alejandro F. Reimondo aleReimondo@smalltalking.net
 See license.txt file
"

Object
	subclass: #Stringifier
	instanceVariableNames: 'origin rememberedSet id markedObjects objectHandlers'
	classVariableNames: '' poolDictionaries: ''
	category: #Stringifier!

! Stringifier class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^super constructEventsTriggered
	,#(	#needsOrigin
		#prepare: #needsStringifier
		#aboutToReset #reset #aboutToClose
		#needsObjectHandlers #objectHandlersContents
		#needsHandlerFor:
	)! !

! Stringifier class methodsFor: #converting !
@ anObject
	" Returns the literal to send to a remote system to obtain a copy of anObject. "

	^self new literalForOne: anObject! !

! Stringifier class methodsFor: #instantiation !
to: target
	" Returns an instance of the receiver. "

	^self new
		"target: target; - ignored"
		yourself! !

! Stringifier methodsFor: #accessing !
origin
	" Returns the origin of the receiver. "

	origin isNil ifTrue: [ self initializeOrigin ].
	^origin! !

! Stringifier methodsFor: #accessing !
id
	" Returns the id of the receiver (can be nil). "

	^id! !

! Stringifier methodsFor: #accessing !
origin: aSymbol
	" Set the origin of the receiver. "

	origin := aSymbol! !

! Stringifier methodsFor: #constants !
tagStart
	" Private - Returns the tag start mask. "

	^' '! !

! Stringifier methodsFor: #constants !
markID
	" Private - Returns the id mark token. "

	^self tagStart ,#id! !

! Stringifier methodsFor: #initialize !
defaultOrigin
	" Private - Initialize the default origin. "

	(Smalltalk includesKey: #OrderedCollection) ifTrue: [ ^#VSE ].
	^#S8! !

! Stringifier methodsFor: #initialize !
initializeOrigin
	" Private - Initialize the origin (hosting platform). "

	origin := self
		triggerEvent: #needsOrigin
		ifNotHandled: [ self defaultOrigin ]! !

! Stringifier methodsFor: #initialize !
initializeRememberedSet
	" Private - Initialize the walked objects cache. "

	rememberedSet := (Smalltalk includesKey: #IdentityDictionary)
		ifTrue: [ IdentityDictionary new ]
		ifFalse: [ self ]! !

! Stringifier methodsFor: #configuration !
reset
	" Private - Reset the receiver to initial state. "

	self triggerEvent: #aboutToReset.
	self forgetObjects.
	rememberedSet := id := nil.
	self triggerEvent: #reset! !

! Stringifier methodsFor: #configuration !
close
	" Private - Close the receiver.
	REQUIRED in platforms where there is no support for IdentityDictionary (e.g. S8 systems). "

	self triggerEvent: #aboutToClose.
	self reset.! !

! Stringifier methodsFor: #configuration !
initializeObjectHandlers
	" Private - Initialize the handlers to map objects. "

	objectHandlers := self
		triggerEvent: #needsObjectHandlers
		ifNotHandled: [ PoolDictionary new ].
	self triggerEvent: #objectHandlersContents ifNotHandled: [
		self origin == #VSE ifTrue: [ self initializeVSEObjectHandlers ].
		origin == #S8 ifTrue: [ self initializeS8ObjectHandlers ].
	]! !

! Stringifier methodsFor: #handlers !
cantHandleBlock
	" Private - Returns the block to use when the receiver do NOT handle an object. "

	^[:object :my| my error: 'Can''t handle instances of ' ,object class name ]! !

! Stringifier methodsFor: #handlers !
defaultHandleBlock
	" Private - Returns the block to handle objects. "

	^[:anObject :my| my basicMapForObject: anObject ]! !

! Stringifier methodsFor: #handlers !
dictionaryHandleBlock
	" Private - Returns the block to handle dictionaries. "

	^[:anObject :my| | result ref |
		result := my newMapForObject: anObject.
		ref := self mapForArray: (self associationsIn: anObject).
		result at: #associations put: ref.
		result
	]! !

! Stringifier methodsFor: #handlers !
asArrayHandleBlock
	" Private - Returns the block to handle arrayable collections. "

	^[:anObject :my| | result ref |
		result := my newMapForObject: anObject.
		ref := self mapForArray: anObject asArray.
		result at: #contents put: ref.
		result
	]! !

! Stringifier methodsFor: #handlers !
bytesHandleBlock
	" Private - Returns the block to handle objects. "

	^[:anObject :my| | result ref |
		result := my newMapForObject: anObject.
		ref := self mapForArray: (anObject notImplementedYet: #bytesHandleBlock).
		result at: #contents put: ref.
		result
	]! !

! Stringifier methodsFor: #handlers !
timeStampHandleBlock
	" Private - Returns the block to handle instances of TimeStamp. "

	^[:timeStamp :my| | result ref time | "we save literal representation"
		result := my newMapForObject: timeStamp.
		time := timeStamp time.
		ref := (timeStamp date printInFormat: 'YYYY/mm/dd' twoDigitYear: false)
			,' ' ,time hours asString
			,':' ,time minutes asString
			,':' ,time seconds asString.
		result at: #contents put: ref.
		result
	]! !

! Stringifier methodsFor: #handlers !
dateTimeHandleBlock
	" Private - Returns the block to handle instances of DateTime. "

	^[:dateTime :my| | result ref date | "we save literal representation"
		result := my newMapForObject: dateTime.
		date := '' ,dateTime year ,$/ ,dateTime month ,$/ ,dateTime dayOfMonth.
		ref := date ,' ' ,(dateTime asTimeString upTo: $ ).
		result at: #contents put: ref.
		result
	]! !

! Stringifier methodsFor: #private !
classNamed: aName
	" Private - Returns the class named aName (or nil). "

	^Smalltalk at: aName ifAbsent: [ ]! !

! Stringifier methodsFor: #configuration !
handleMembersOf: aClass with: aBlock
	" Private - Set the receiver to handle instances of aClass. "

	aClass isString ifTrue: [ ^self handleMembersOf: (self classNamed: aClass) with: aBlock ].
	^aClass notNil ifTrue: [ objectHandlers at: aClass name put: aBlock ]! !

! Stringifier methodsFor: #configuration !
handleInstancesOf: aClass with: aBlock
	" Private - Set the receiver to handle instances of aClass and subclasses. "

	aClass isString ifTrue: [ ^self handleInstancesOf: (self classNamed: aClass) with: aBlock ].
	^aClass notNil ifTrue: [
		aClass withAllSubclasses do: [:cls| objectHandlers at: cls name put: aBlock ]
	]! !

! Stringifier methodsFor: #configuration !
cantHandleInstancesOf: aClass
	" Private - Set the receiver to do NOT handle instances of aClass and subclasses.
	Note that aClass can be a class, a name of a class or an array of classes/classNames "

	aClass isArray ifTrue: [ ^aClass do: [:each| self cantHandleInstancesOf: each ] ].
	aClass isString ifTrue: [ ^self cantHandleInstancesOf: (self classNamed: aClass) ].
	^aClass notNil ifTrue: [ self handleInstancesOf: aClass with: self cantHandleBlock ]! !

! Stringifier methodsFor: #configuration !
initializeVSEObjectHandlers
	" Private - Initialize the handlers to map VSE objects. "

	self	cantHandleInstancesOf: #(
			#BlockClosure #CallBack
			#NotificationManager
			#Context #ExternalBuffer
			#ExternalHandle #FileSystemEntity
			#Font #Menu #FileStream
			#ViewManager #Window
			#WinLogicalObject
		);
		cantHandleInstancesOf: #Collection; "abstract"
		handleMembersOf: #Dictionary with: self dictionaryHandleBlock;
		handleMembersOf: #PoolDictionary with: self dictionaryHandleBlock;
		handleInstancesOf: #Set with: self asArrayHandleBlock;
		handleInstancesOf: #ActionSequence with: self defaultHandleBlock;
		cantHandleInstancesOf: #ByteArray; "handleInstancesOf: #ByteArray with: self bytesHandleBlock;"
		handleInstancesOf: #Interval with: self defaultHandleBlock;
		handleInstancesOf: #OrderedCollection with: self asArrayHandleBlock;
		handleInstancesOf: #Array with: self asArrayHandleBlock;
		handleInstancesOf: #TimeStamp with: self timeStampHandleBlock;
		yourself! !

! Stringifier methodsFor: #configuration !
initializeS8ObjectHandlers
	" Private - Initialize the handlers to map S8 objects. "

	self	cantHandleInstancesOf: #(
			#BlockClosure #Context
			#CompiledMethod #CompilerNode
			#NativeObject #PPParser
		);
		cantHandleInstancesOf: #Collection; "abstract"
		handleMembersOf: #Dictionary with: self dictionaryHandleBlock;
		handleMembersOf: #PoolDictionary with: self dictionaryHandleBlock;
		handleInstancesOf: #Set with: self asArrayHandleBlock;
		handleInstancesOf: #ActionSequence with: self defaultHandleBlock;
		handleInstancesOf: #Array with: self asArrayHandleBlock;
		handleInstancesOf: #DateTime with: self dateTimeHandleBlock;
		yourself! !

! Stringifier methodsFor: #configuration !
prepareFor: anObject
	" Private - Prepare the receiver to work on anObject. "

	rememberedSet isNil ifTrue: [ self initializeRememberedSet ].
	id isNil ifTrue: [ id := rememberedSet size + 1 ].
	objectHandlers isNil ifTrue: [ self initializeObjectHandlers ].
	self triggerEvent: #prepare: with: anObject! !

! Stringifier methodsFor: #accessing !
@ anObject
	" Returns the literal to send to target to obtain a copy of anObject.
	Close the receiver after encoding anObject. "

	^self literalForOne: anObject! !

! Stringifier methodsFor: #accessing !
literalForOne: anObject
	" Returns the literal to send to target to obtain a copy of anObject.
	Close the receiver after encoding anObject. "

	| result |
	result := self literalFor: anObject.
	self close.
	^result! !

! Stringifier methodsFor: #accessing !
literalFor: anObject
	" Returns the literal to send to target to obtain a copy of anObject. "

	self prepareFor: anObject.
	^self stringify: (self mapFor: anObject) object: anObject! !

! Stringifier methodsFor: #constants !
stringifierFor: structure
	" Returns the default stringifier to emmit structure.
	Note that on S8 we will use native JSON if present.
	If structure do not respondsTo #asJSON the receiver will format the output
	"

	(structure respondsTo: #asJSON) ifTrue: [
		Smalltalk jsObjectAt: #JSON ifAbsent: [ ^self ].
		^#asJSON
	].
	^self! !

! Stringifier methodsFor: #stringify !
stringify: aMap object: anObject
	" Private - Returns a string literal from aMap. "

	| package stringifier |
	package := self packageWith: aMap.
	stringifier := self
		triggerEvent: #needsStringifier
		ifNotHandled: [ self stringifierFor: package ].
	stringifier isString ifTrue: [ ^package perform: stringifier ].
	^stringifier stringify: package! !

! Stringifier methodsFor: #mapping !
mapFor: anObject
	" Private - Returns the map to export anObject. "

	| selector |
	selector := self selectorFor: anObject.
	selector == #literal ifTrue: [ ^self mapForLiteral: anObject ].
	selector == #array ifTrue: [ ^self mapForArray: anObject ].
	selector == #class ifTrue: [ ^self referenceToClass: anObject ].
	selector == #object ifTrue: [ ^self mapForObject: anObject ].
	^self error: 'Invalid selector'! !

! Stringifier methodsFor: #mapping !
selectorFor: anObject
	" Private - Returns the keyword to represent anObject as a literal or well known object. "

	anObject isNil ifTrue: [ ^#literal ].
	(anObject isString
		or: [ anObject isNumber
		or: [ anObject == true
		or: [anObject == false]]]) ifTrue: [ ^#literal ].
	(anObject isKindOf: Behavior) ifTrue: [ ^#class ].
	^#object! !

! Stringifier methodsFor: #mapping !
associationsIn: aDictionary
	" Private - Returns the associations contained in aDictionary. "

	| keys |
	keys :=  aDictionary keys asArray.
	^keys collect: [:each| aDictionary associationAt: each ]! !

! Stringifier methodsFor: #mapping !
newMap
	" Private - Returns a new map to host current object id. "

	^PoolDictionary new
		at: #id put: id;
		yourself! !

! Stringifier methodsFor: #mapping !
mapForLiteral: literal
	" Private - Returns the map to export a literal. "

	literal isNumber ifTrue: [
		literal class == Number ifTrue: [ ^literal ].
		literal isInteger ifTrue: [ ^literal ].
		literal isFloat ifTrue: [ ^self mapForFloat: literal ].
		^self mapForObject: literal
	].
	literal isString ifTrue: [
		^self mapForString: literal
	].
	^literal! !

! Stringifier methodsFor: #mapping !
mapForString: literal
	" Private - Returns the map to export string literal. "

	^(self hasTag: literal) ifTrue: [self tagStart ,literal] ifFalse: [literal]! !

! Stringifier methodsFor: #mapping !
mapForFloat: aFloat
	" Private - Returns the map to export aFloat. "

	^aFloat! !

! Stringifier methodsFor: #mapping !
mapForArray: anArray
	" Private - Returns the map to export anArray. "

	^anArray collect: [:each| self mapFor: each ]! !

! Stringifier methodsFor: #mapping !
referenceToClass: aClass
	" Private - Returns the map to reference aClass in target system. "

	^self referenceTo: aClass ifAbsent: [ | ref ivn |
		ref := self newMap.
		ref at: #global put: (self nameOfClass: aClass).
		ivn := aClass allInstVarNames.
		ivn notEmpty ifTrue: [ ref at: #allInstVarNames put: ivn ].
		ref
	]! !

! Stringifier methodsFor: #mapping !
nameOfClass: aClass
	" Private - Returns a literal to reference aClass by name. "

	aClass isClass ifTrue: [ ^aClass name ].
	^Array with: aClass instanceClass name with: #class! !

! Stringifier methodsFor: #mapping !
newMapForObject: anObject
	" Private - Returns a new/clean map literal to anObject structure. "

	^self newMap
		at: #class put: (self referenceToClass: anObject class);
		yourself! !

! Stringifier methodsFor: #mapping !
basicMapForObject: anObject
	" Private - Returns a new map literal to anObject structure. "

	| map |
	map := self newMapForObject: anObject.
	self storeFieldsOf: anObject into: map.
	^map! !

! Stringifier methodsFor: #mapping !
mapForObject: anObject
	" Private - Returns the map to export anObject. "

	^self referenceTo: anObject ifAbsent: [ | none mapper |
		none := #none.
		mapper := objectHandlers at: anObject class name ifAbsentPut: [
			self	triggerEvent: #needsHandlerFor:
				withArguments: (Array with: anObject class)
				ifNotHandled: [none]
		].
		mapper == none
			ifTrue: [ self basicMapForObject: anObject ]
			ifFalse: [ mapper value: anObject value: self ]
	]! !

! Stringifier methodsFor: #mapping !
storeFieldsOf: anObject into: map
	" Private - Store the contents of anObject into map. "

	self origin == #VSE ifTrue: [ ^self storeVSEFieldsOf: anObject into: map ].
	origin == #S8 ifTrue: [ ^self storeS8FieldsOf: anObject into: map ].
	^self error: 'Unsupported origin' ! !

! Stringifier methodsFor: #mapping !
storeVSEFieldsOf: anObject into: map
	" Private - Store the contents of anObject into map. "

	1 to: anObject class instSize do: [:i|
		map at: i put: (self mapFor: (anObject instVarAt: i))
	]! !

! Stringifier methodsFor: #mapping !
storeS8FieldsOf: anObject into: map
	" Private - Store the contents of anObject into map. "

	anObject class allInstVarNames withIndexDo: [:each :i|
		map at: i put: (self mapFor: (anObject instVarAt: each))
	]! !

! Stringifier methodsFor: #cache !
forgetObjects
	" Private - Forget/unmark the rememebered objects. "

	markedObjects notNil ifTrue: [ | markID |
		markID := self markID.
		markedObjects do: [:each| each basicDelete: markID ]
	].
	markedObjects := nil! !

! Stringifier methodsFor: #cache !
referenceTo: anObject ifAbsent: aBlock
	" Private - Save a new reference ID for anObject. "

	^rememberedSet at: anObject ifAbsent: [
		rememberedSet at: anObject put: self newReference.
		aBlock value
	]! !

! Stringifier methodsFor: #cache !
newReference
	" Private - Returns a new reference literal. "

	id := id + 1.
	^self tagStart ,id asString! !

! Stringifier methodsFor: #cache !
at: anObject ifAbsent: aBlock
	" Private - Returns the value catched for anObject or the result of evaluating aBlock.
	This implementation is a default implementation for marking objects while walking when there is no support for IdentityDictionary.
	"

	^anObject basicAt: self markID ifAbsent: aBlock! !

! Stringifier methodsFor: #cache !
at: anObject put: aValue
	" Private - Set the value catched for anObject.
	This implementation is a default implementation for marking objects while walking when there is no support for IdentityDictionary.
	"

	markedObjects isNil ifTrue: [ markedObjects := #() ].
	markedObjects add: anObject.
	anObject basicAt: self markID put: aValue.
	^aValue! !

! Stringifier methodsFor: #private !
writeStream
	" Private - Returns new write stream. "

	origin = #VSE ifTrue: [ ^'' asStream ].
	^'' stream! !

! Stringifier methodsFor: #stringify !
stringify: aMap
	" Private - Returns a string literal from map. "

	| stream |
	stream := self writeStream.
	self stringify: aMap into: stream.
	^stream contents! !

! Stringifier methodsFor: #stringify !
packageWith: aMap
	" Private - Returns aMap wrapped with information of the receiver. "

	^PoolDictionary new
		at: #origin put: self origin;
		at: #id put: id;
		at: #contents put: aMap;
		yourself! !

! Stringifier methodsFor: #formating !
printFloat: literal into: stream
	" Private - Dump literal formated as JSON. "

	literal printOn: stream decimalSeparator: '.'! !

! Stringifier methodsFor: #formating !
hasTag: literal
	" Private - Returns true if the literal starts with tagStart. "

	| tag |
	tag := self tagStart.
	self origin == #S8 ifTrue: [ ^literal startsWith: tag ].
	tag size > literal size ifTrue: [ ^false ].
	1 to: tag size do: [:i|
		(tag at: i) = (literal at: i) ifFalse: [ ^false ]
	].
	^true! !

! Stringifier methodsFor: #formating !
jsonEscapeUnicode: ascii
	" Private - Returns the escape sequence for unicode value. "

	^'u' ,(ascii printPaddedWith: $0 to: 4 base: 16)! !

! Stringifier methodsFor: #formating !
jsonEscape: ascii char: c
	" Private - Returns the escape sequence for ascii code. "

	#( #(10 n) #(13 r) #(9 t) #(12 f) #(8 b) ) do: [:tuple|
		tuple first = ascii ifTrue: [ ^tuple last ]
	].
	(ascii between: 32 and: 126) ifFalse: [
		^self jsonEscapeUnicode: ascii
	].
	^String with: c! !

! Stringifier methodsFor: #formating !
jsonEncodeChar: c into: stream
	" Private - Encode character c into stream.
	See https://www.json.org/
	"

	| ascii |
	ascii := c asciiValue.
	(ascii < 32 or: [ascii > 126 or: [ #(34 92 47) includes: ascii ]])
	ifTrue: [ stream nextPut: $\; nextPutAll: (self jsonEscape: ascii char: c) ]
	ifFalse: [ stream nextPut: c ]! !

! Stringifier methodsFor: #formating !
printString: literal into: stream
	" Private - Dump literal formated as JSON. "

	stream nextPut: $".
	literal asString do: [:c| self jsonEncodeChar: c into: stream ].
	stream nextPut: $"! !

! Stringifier methodsFor: #stringify !
stringify: literal into: stream
	" Private - Dump literal onto stream. "

	literal isNil ifTrue: [ ^stream nextPutAll: #null ].
	literal isArray ifTrue: [
		stream nextPut: $[.
		literal	do: [:each| self stringify: each into: stream ]
			andBetweenDo: [ stream nextPut: $, ].
		stream nextPut: $].
		^self
	].
	(literal isMemberOf: PoolDictionary) ifTrue: [
		stream nextPut: ${.
		literal keys do: [:key|
			self printString: key into: stream.
			stream nextPut: $:.
			self stringify: (literal at: key) into: stream ]
			andBetweenDo: [ stream nextPut: $, ].
		stream nextPut: $}.
		^self
	].
	literal isString ifTrue: [
		^self printString: literal into: stream
	].
	(literal isNumber and: [literal class name = 'Float']) ifTrue: [
		^self printFloat: literal into: stream
	].
	stream nextPutAll: literal printString! !