
EventManager
	subclass: #SystemChangesManager
	instanceVariableNames: #systemChanges
	category: #SystemServer !

! SystemChangesManager methodsFor: 'tracking-classes' !
classChanged: aBehavior
	" Private - aBehavior has been changed. "

	^self systemHasChanged: (Array with: #classChanged: with: aBehavior)! !

! SystemChangesManager methodsFor: 'tracking-methods' !
methodBound: aMethod to: aClass
	" Private - The method has been bound to aClass. "

	aMethod selector = String doItSelector ifTrue: [ ^self "ignored" ].
	^self systemHasChanged: (Array with: #methodBound:to: with: aMethod with: aClass)! !

! SystemChangesManager methodsFor: 'tracking-methods' !
methodRemoved: aMethod from: aClass
	" Private - The method has been removed from aClass. "

	aMethod selector = String doItSelector ifTrue: [ ^self "ignored" ].
	^self systemHasChanged: (Array with: #methodRemoved:from: with: aMethod with: aClass)! !

! SystemChangesManager methodsFor: 'tracking-methods' !
memoryWarning: anArgument
	" Private - A memory warning event has been thrown. "

	^self systemHasChanged: (Array with: #memoryWarning: with: anArgument)! !

! SystemChangesManager methodsFor: 'system changes' !
processSystemChanges: changes
	" Private - Update reflecting system changes. "

	self print: #processSystemChanges:
		,' ', changes asString
		,' MUST be implemented at '
		,self class name! !

! SystemChangesManager methodsFor: #changes !
changesDo: aOneArgumentBlock
	" Private - Process the changes with aOneArgumentBlock. "
	
	| changes |
	changes := systemChanges.
	(changes notNil and: [changes notEmpty]) ifTrue: [
		systemChanges := nil.
		aOneArgumentBlock value: changes
	].! !

! SystemChangesManager methodsFor: #changes !
processTheChanges
	" Private - Process the changes. "
	
	^self changesDo: [:changes| self processSystemChanges: changes ]! !

! SystemChangesManager methodsFor: 'system changes' !
processSystemChanges
	" Private - The receiver has changes to be processed.
	The default action is to defer proccessing of the changes.
	This method can be overwritten/refined by subclasses to implement alternatives of change management.
	"
	
	[ self processTheChanges ] valueDeferred: 750.! !

! SystemChangesManager methodsFor: 'system changes' !
systemHasChanged: aSystemChange
	" Private - Signal a event of system change to be resolved in the future. "

	systemChanges notNil ifTrue: [
		^systemChanges add: aSystemChange
	].
	systemChanges := Array with: aSystemChange.
	self processSystemChanges.! !

! SystemChangesManager methodsFor: 'system changes' !
bindSystemManager
	" Private - Set the receiver to start listening current system changes.
	Note: Only the events implemented by the receiver will be hooked.
	"

	(SystemManager eventsTriggered copyWithout: #changed) do: [:each|
		(self respondsTo: each) ifTrue: [
			SystemManager current when: each send: each to: self
		]
	]! !

! SystemChangesManager methodsFor: 'system changes' !
unbindSystemManager
	" Private - rest the link from the receiver to system changes. "

	| aManager |
	aManager := SystemManager currentOrNil.
	aManager notNil ifTrue: [ aManager removeAllActionsWithReceiver: self ]! !

SystemChangesManager
	subclass: #SystemServer
	instanceVariableNames: 'server port methodReferences'
	category: #SystemServer !

SystemServer class instanceVariableNames: 'current silent platform'!
SystemServer class buildAccessors: #(
	#( current initialize 'self new' )
	#( silent initialize #false )
	#( platform initialize 'self detectedPlatform' )
)!
SystemServer buildAccessors: #(
	#( port initialize #8088 )
	#( methodReferences initialize 'Smalltalk current methodReferences' )
)!

! SystemServer class methodsFor: #platform!
detectedPlatform
	" Private - Returns the platform we are running (or nil).
	Note that the platform will be used to solve transport functionality and the associated messages MUST be implemented by the receiver.
	"

	#(	#(NodeJS node)
		#(GCDWebServer gcd)
		#(Turbo turbo)
		#(SimpleHTTPServer luaSimple)
	) do: [:tuple| (Smalltalk includesKey: tuple first) ifTrue: [ ^tuple last ] ].
	^nil! !

! SystemServer methodsFor: #platform!
platformSend: message withArguments: anArray
	" Private - Returns the result of sending message (platform dependent) to receiver.
	Trigger an error if the message is not implemented.
	"

	| token msg |
	token := self class platform.
	token isNil ifTrue: [ token := #default ].
	msg := token , message.
"	(self respondsTo: msg) ifFalse: [ ^self error: 'Missing server support for ' ,msg ]. "
	^self perform: msg withArguments: anArray! !

! SystemServer class methodsFor: #singleton!
currentOrNil
	" Returns the current instance (or nil). "

	^current! !

! SystemServer class methodsFor: #server!
start
	" Start the server. "

	^self current start! !

! SystemServer class methodsFor: #server!
close
	" Close the current server. "

	^current notNil ifTrue: [ current close ]! !

! SystemServer class methodsFor: #server!
stop
	" Stop the current server. "

	^self close! !

! SystemServer class methodsFor: #html !
htmlSelectorsFor: classname
	" Return an html fragment containing the selectors implemented by classname. "

	| stream class classdef |
	class := Smalltalk at: classname ifAbsent:[ ^self ].
	stream := '' writeStream.	
	stream nextPutAll: '<BR><BR><a href=',$'
		,'/SystemServer/htmlHierarchy',$'
		,'><< S8 Hierarchy</a><BR>'.
	classdef := Exporter new fileOutDefinitionOf: class.
	classdef := classdef replace:'\r\n' with:'<BR>'.
	classdef := classdef replace:'\t' with:'&nbsp;&nbsp;&nbsp;'.
	stream nextPutAll: '<BR>',classdef, '<H4>', classname,' class methods</H4>'.
	class class selectors do: [:each| 
		stream nextPutAll: '<a href=',$'
			,'/SystemServer/htmlFileOut:selector:side:/', class name
			,'/',each,'/class', $', '>', each, '</a><BR>'
	].	
	stream nextPutAll: '<H4>', classname,' instance methods</H4>'.
	class selectors do:[:each| 
		stream nextPutAll: '<a href=',$'
			,'/SystemServer/htmlFileOut:selector:/', class name
			,'/',each,$', '>', each, '</a><BR>'
	].
	^stream contents! !

! SystemServer class methodsFor: #html !
htmlFileOut: classname selector: selector side: side
	" Return an html fragment of selector source code. "

	| stream class mth source |
	class := Smalltalk at: classname ifAbsent:[ ^self ].
	mth := (side = #class) ifTrue:[ class class >> selector 
	] ifFalse: [ class >> selector ].
	source := Exporter new fileOutMethod: mth of: class.
	source := source replace:'\r\n' with:'<BR>'.
	source := source replace:'\t' with:'&nbsp;&nbsp;&nbsp;'.
	source := source replace: $\,$" with: '&quot;'.
	source := '<BR><BR><a href='
		,'/SystemServer/htmlSelectorsFor:/'
		,classname ,'><< ' ,classname
		,' methods</a><BR>',source
		,'<BR>'.
	^source! 

! SystemServer class methodsFor: #html !
htmlFileOut: classname selector: selector
	" Return an html fragment of selector source code. "

	^self htmlFileOut: classname 
		selector: selector 
		side: #instance! !

! SystemServer class methodsFor: #html !
htmlHierarchy
	" Returns the system hierarchy in html fasion. "

	| stream |
	stream := '' writeStream.
	stream nextPutAll:'<H2>S8 Hierarchy</H2>'.
	Smalltalk rootClasses do: [:cls|
		self htmlWalk: cls blanks: 0 into: stream
	].
	^stream contents! !
	
! SystemServer class methodsFor: #html !
htmlWalk: aClass blanks: blanks into: stream
	" Private - Returns the description for hierarchy level @ blanks. "

	stream nextPutAll:'<BR>'. 
	blanks timesRepeat: [ stream nextPutAll: '&nbsp;&nbsp;' ].
	stream nextPutAll: (self htmlDescriptionForClass: aClass).
	aClass subclasses do: [:cls|
		self htmlWalk: cls blanks: blanks + 1 into: stream
	].! !

! SystemServer class methodsFor: #html !
htmlDescriptionForClass: aClass
	" Private - Returns the contents to show for aClass. "

	^'<a href=', $', '/SystemServer/htmlSelectorsFor:/', aClass name, $' ,$>, aClass name,'</a> [',aClass category, ']'! !

! SystemServer methodsFor: #query!
isRunning
	" Returns true if the receiver is online. "

	^server notNil! !

! SystemServer methodsFor: #runnning!
step
	" Run a server step. "

	^self isRunning ifTrue: [ server step ]! !

! SystemServer methodsFor: #runnning!
listen
	" Run a server step. "

	^server listen! !

! SystemServer methodsFor: #runnning!
run
	" Run the receiver in a loop until stopped. "

	[self isRunning] whileTrue: [ self listen ]! !

! SystemServer methodsFor: #server!
online
	" Returns true if the receiver is online. "

	^server notNil! !

! SystemServer methodsFor: #server!
start
	" Start the receiver. "

	^self startServer! !

! SystemServer methodsFor: #server!
close
	" Close the receiver. "

	self server: nil! !

! SystemServer methodsFor: #private!
server: aServer
	" Private - Set the server of the receiver. "

	(server respondsTo: #close) ifTrue: [ server close ].
	server := aServer.
	server isNil
		ifTrue: [ self unbindSystemManager ]
		ifFalse: [ self bindSystemManager ]! !

! SystemServer methodsFor: #server!
startServer
	" Start the server. "

	^self startServerOn: self port! !
	
! SystemServer methodsFor: #server!
startServerOn: aPort
	" Start the server. "

	| query |
	query := self systemRequestBlock.
	query isNil ifTrue: [ ^false ].
	^self startWith: [:req :resp|
		'<html><body><pre>'
		,(query value: req value: resp)
		,'</pre></body></html>'.		
	] on: aPort! !
	
! SystemServer methodsFor: #server!
startWith: processBlock on: aPort
	
	^self	startBlock: processBlock 
		port: aPort! !
	
! SystemServer methodsFor: #server!
startBlock: aBlock port: aPort
	" Private - Start the server. "

	self server: (self
		start: self
		processBlock: aBlock
		port: aPort).
	^server notNil! !

! SystemServer methodsFor: #server!
start: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	^self	platformSend: #Start:processBlock:port:
		withArguments: (Array with: aServer with: aBlock with: aPort)! !

! SystemServer methodsFor: #server !
systemRequestBlock
	" Private - Return the request processing block used for remote system tools.
	Note that the block is evaluated for each request and must return the code of the body of HTML formated result.
	"

	^self platformSend: #RequestBlock withArguments: #()! !

! SystemServer methodsFor: #server !
jsonParse: string
	" Private - Return the result of parsing JSON string. "

	^self platformSend: #Parse: withArguments: (Array with: string)! !

! SystemServer methodsFor: #server !
jsonStringify: literal
	" Private - Return the result of JSON string. "

	^self platformSend: #Stringify: withArguments: (Array with: literal)! !

! SystemServer methodsFor: #defaults !
defaultRequestBlock
	" Private - Return the request processing block used for remote system tools. "

	^self error: 'Missing support for server request'! !

! SystemServer methodsFor: #defaults !
defaultStart: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	^self error: 'Missing support for server start'! !

! SystemServer methodsFor: #defaults !
defaultParse: string
	" Private - Return the result of parsing JSON string. "

	^NativeHandle support @ (#{JSON} #parse: string)! !

! SystemServer methodsFor: #defaults !
defaultStringify: literal
	" Private - Return the result of JSON string. "

	^#{JSON} #stringify: literal! !

! SystemServer methodsFor: #json !
checkArrays: aHandle
	" Private,Lua - Scan native handle values for arrayed contents recursivelly.
	Note that JSON data is hierarchical non-recursive structures we walk the structure descending recursivelly in the structure mutating array candidates.
	"

	| toDo |
	toDo := Array with: aHandle.
	[toDo notEmpty] whileTrue: [ | literal |
		(literal := toDo removeFirst) notNil ifTrue: [
			(literal basicAt: 1 ifAbsent: []) notNil "mutate to Array"
			ifTrue: [ {'literal' lua'smalltalk:newArray(literal)'}. toDo addAll: literal ]
			ifFalse: [ "walk normal object contents"
				(self isObject: literal) ifFalse: [
					NativeHandle support @ literal propertiesDo: [:key :value|
						(value basicAt: 1 ifAbsent: []) notNil ifTrue: [
							{'value' lua'smalltalk:newArray(value)'}. toDo addAll: value
						] ifFalse: [ toDo add: value ]
					]
				]
			]
		]
	].! !

! SystemServer methodsFor: #notification !
say: text
	" Show text to the user. "

	self class silent ifTrue: [ ^self ].
	self print: '//' ,text! !

! SystemServer methodsFor: #private !
getCommand: command
	" Private - Return the result of evaluating the command. "

	^(command isEmpty or: [command first isEmpty])
	ifTrue:  [ self systemHomePage ]
	ifFalse: [ (command size = 1
		and: [ self respondsTo: command first ])
		ifTrue: [ self perform: command first ]
		ifFalse: [ self resultForCommand: command stream ]
	]! !

! SystemServer methodsFor: #private !
postCommand: command stream: stream
	" Private - Return the result of evaluating the command. "

	stream upTo: $=.
	^(command isEmpty or: [command first isEmpty])
	ifTrue:  [ self print: stream upToEnd ]
	ifFalse: [
		command add: stream upToEnd.
		self resultForCommand: command stream.
	]! !

! SystemServer methodsFor: #private !
requestBlockFor: urlBlock method: methodBlock post: postBlock
	" Private - Return the request processing block used for remote system tools. "

	^[:arg1 :arg2| | command method | self jsonValue: [
		command := urlBlock value: arg1 value: arg2.
		method := methodBlock value: arg1 value: arg2.
		self say: '' ,method ,' ' ,command.
		command := command asArrayOfSubstringsSeparatedBy: $/.
		command removeFirst.
		method = #GET ifTrue: [ self getCommand: command ]
		ifFalse: [
			postBlock value: arg1 value: [:stream|
				self postCommand: command stream: stream
			]
		]
	]]! !
	
! SystemServer methodsFor: '#commands' !
systemHomePage
	" Private - Returns the contents for system home page. "

	^'<h1>S8 System server</h1>'
	,'Try commands like <a href=',$' ,'/SystemServer/htmlHierarchy', $', '>hierarchy</a>&nbsp;'
	,'or more complex expressions: <a href=', $', '/Object/htmlSelectors', $', '>Object selectors</a>'! !

! SystemServer methodsFor: #private !
objectsFromArguments: array
	" Private - Return the objects to use as arguments received from client. "

	^array! !

! SystemServer methodsFor: #private !
perform: stream on: receiver
	" Private - Returns the result of performing messages in stream on receiver. "

	| result |
	result := receiver.
	[stream atEnd] whileFalse: [ | selector |
		selector := stream next.
		selector isEmpty ifTrue: [ selector := $/ ].
		selector first = ${ ifTrue: [ | json |
			[selector last = $}] whileFalse: [
				stream atEnd ifTrue: [ ^self error: 'Corrupted message ' ,selector ].
				selector := selector ,$/ ,stream next.
			].
			json := self jsonParse: selector.
			selector := json @ #selector.
			selector isNil ifTrue: [ self error: 'Missing selector' ].
			result := (Message
				receiver: result selector: selector toString
				arguments: (self objectsFromArguments: json @ #arguments)
				) perform
		] ifFalse: [ | args argCount |
			argCount := selector argumentCount.
			args := (1 to: argCount) collect: [:i|
				stream atEnd ifFalse: [ stream next ]
			].
			result := result perform: selector withArguments: args
		]
	].
	^result! !

! SystemServer methodsFor: #commands !
resultForCommand: stream
	" Private - Return the result of processing the command in stream. "

	| command cls |
	command := stream next.
	command = #doIt    ifTrue: [ ^stream next doIt ].
	command = #fileIn  ifTrue: [ ^stream next fileIn ].
	command = #eval    ifTrue: [ ^Smalltalk current eval: stream next ].
	command = #fileOut ifTrue: [ ^stream next fileOut ].
	(#( #fileOut #category ) includes: command) ifTrue: [ ^stream next fileOut ].
	(command first isUppercase and: [
		cls := Smalltalk at: command ifAbsent: [].
		cls notNil
	]) ifTrue: [ ^self perform: stream on: cls ].
	^'Unknown command ',command! !
 
! SystemServer methodsFor: #commands !
categories
	" Private - Returns the categories in the system. "

	^Class categories sorted printString! !

! SystemServer methodsFor: #commands !
classes
	" Private - Returns the contents for this command. "

	^Smalltalk classes inject: '' into: [:total :cls|
		total,'
',(self shortDescriptionForClass: cls)
	]! !

! SystemServer methodsFor: #commands !
hierarchy
	" Private - Returns the contents for this command. "

	| stream |
	stream := '' writeStream.
	Smalltalk rootClasses do: [:cls|
		self walk: cls blanks: 0 into: stream
	].
	^stream contents! !

! SystemServer methodsFor: #commands !
walk: aClass blanks: blanks into: stream
	" Private - Returns the description for hierarchy level @ blanks. "

	stream cr. blanks timesRepeat: [ stream nextPut: $ ].
	stream nextPutAll: (self shortDescriptionForClass: aClass);cr.
	aClass subclasses do: [:cls|
		self walk: cls blanks: blanks + 1 into: stream
	].! !
	
! SystemServer methodsFor: #private !
shortDescriptionForClass: aClass
	" Private - Returns the contents to show for aClass. "

	^aClass name,' ',aClass category! !

! SystemServer methodsFor: #output !
jsonValue: aBlock
	" Private - Returns a JSON package string with result of evaluating aBlock. "

	| response json |
	response := [ | result |
		result := self literalFor: aBlock evaluate.
		self changesWith: #result -> result
		] on: Error do: [:ex| #error -> ex ].
	^[	json := response json.
		self jsonStringify: response json
	] on: Error do: [ json toString ]! !

! SystemServer methodsFor: #output !
literalFor: anObject
	" Private - Returns the literal for anObject. "

	anObject isNil ifTrue: [ ^'!!nil' ].
	(self isObject: anObject) ifTrue: [
		anObject isString ifTrue: [
			(anObject startsWith: $!!) ifTrue: [ ^$!! ,anObject ].
			^anObject
		].
		anObject isArray ifTrue: [ ^anObject collect: [:each| self literalFor: each ] ].
		(anObject isKindOf: Behavior) ifTrue: [ ^'!!',anObject printString ].
		(anObject isKindOf: CompiledMethod) ifTrue: [ ^'!!',anObject printString ].
		^anObject printString
	].
	^anObject! !

! SystemServer methodsFor: #output !
changesWith: results
	" Private - Returns the results with logged changes. "

	| changes |
	changes := systemChanges.
	systemChanges := nil.
	(changes isNil or: [changes isEmpty]) ifTrue: [ ^results ].
	changes := self literalFor: changes.
	^results ,(#changes -> changes)! !

! SystemServer methodsFor: 'system changes' !
processSystemChanges
	" Private - The receiver has changes to be processed.
	We do nothing, the changes will be sent to client(s?) on demmand.
	"! !

"--NodeJS customization----------------------------------------------"!
! SystemServer methodsFor: #node !
nodeStart: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	| localIP http httpServer |
	[ http := NodeJS http ] on: Error do: [
		self print: self class name ,' missing HTTP interface.'.
		^nil
	].
	localIP := NodeJS os localIP.
	localIP isNil ifTrue: [
		self print: self class name ,' missing network interface.'.
		^nil
	].
	httpServer := http createServer:[:req :resp| | request response |
		request  := NodeIncomingMessage @ req.
		response := NodeHTTPServerResponse @ resp.
		response 
			setHeader:'Access-Control-Allow-Origin' value:'*';
			writeHead: 200;
			end: (aBlock value: request value: response).
	].
	httpServer listen: aPort hostname: nil json.
	self print: self class name ,' running at http://' ,localIP ,':' ,aPort.
	^httpServer! !

! SystemServer methodsFor: #node !
nodeRequestBlock
	" Private - Return the request processing block used for remote system tools. "

	^self
		requestBlockFor: [:req| req url asDecodedURI ]
		method: [:req| req method ]
		post: [:req :action| | body |
			body := ''.
			req
				on: #data do: [:data| body := body ,data ];
				on: #end do: [ action value: body asDecodedURI stream ].
			#ok
		] ! !

! SystemServer methodsFor: #node !
nodeParse: string
	" Private - Return the result of parsing JSON string. "

	^self defaultParse: string! !

! SystemServer methodsFor: #node !
nodeStringify: literal
	" Private - Return the result of JSON string. "

	^self defaultStringify: literal! !

"--GCD customization----------------------------------------------"!
! SystemServer methodsFor: #gcd !
gcdStart: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	^GCDWebServer startHTML: aBlock withPort: aPort ! !

! SystemServer methodsFor: #gcd !
gcdRequestBlock
	" Private - Return the request processing block used for remote system tools. "

	^self
		requestBlockFor: [:req| req path ]
		method: [:req| #GET ]
		post: [:req :action| | body |
			body := ''.
			req	notImplementedYet: #gcdRequestBlock ,' (POST)';
				on: #data do: [:data| body := body ,data ];
				on: #end do: [ action value: body asDecodedURI stream ].
			#ok
		]! !

! SystemServer methodsFor: #gcd !
gcdParse: string
	" Private - Return the result of parsing JSON string. "

	^self defaultParse: string! !

! SystemServer methodsFor: #gcd !
gcdStringify: literal
	" Private - Return the result of JSON string. "

	^self defaultStringify: literal! !

"--Turbo(Lua) customization----------------------------------------------"!
! SystemServer methodsFor: #turbo !
turboStart: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	| http |
	http := HTTPServer new: [:hRequest| | hdr request msg buf crlf |
		crlf := String crlf.
		request := HTTPRequest @ hRequest.
		request basicAt: #path put: request path.
		msg := aBlock value: request.
		hdr := HTTPHeaders new.
		hdr
			add: 'Content-Length' value: '',msg size;
			add: 'Connection' value: 'Keep-Alive';
			add: 'Access-Control-Allow-Origin' value: $*;
			setStatusCode: 200;
			setVersion: 'HTTP/1.1';
			yourself.
		buf := hdr stringifyAsResponse ,crlf ,msg.
		request write: buf; finish
	].
	http listen: aPort.
	self print: 'SystemServer setup for port ',aPort.
	"Ioloop instance start"
	^http! !

! SystemServer methodsFor: #turbo !
turboRequestBlock
	" Private - Return the request processing block used for remote system tools. "

	^self
		requestBlockFor: [:req| Escape new unescape: req#path ]
		method: [:req| #GET ]
		post: [:req :action|
			req notImplementedYet: #gcdRequestBlock ,' (POST)'.
			#ok
		]! !

! SystemServer methodsFor: #turbo !
turboParse: string
	" Private - Return the result of parsing JSON string. "

	| result |
	result := NativeHandle support @ (Escape new jsonDecode: string).
	result notNil ifTrue: [ self checkArrays: result handle ].
	^result! !

! SystemServer methodsFor: #turbo !
turboStringify: literal
	" Private - Return the result of JSON string. "

	^Escape new jsonEncode: NativeHandle support @ literal! !

"--luaSimple customization----------------------------------------------"!
! SystemServer methodsFor: #luaSimple !
luaSimpleStart: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	| http |
	(Smalltalk includesKey: #Json) ifFalse: [
		^self error: 'Missing Json support (SimpleJson.st required).'
	].
	http := SimpleHTTPServer new.
	http	port: aPort;
		when: #needsContents:with: do: [:uri :headers|
			aBlock value: uri value: headers
		]; start.
	self note: ' Server started at ' ,http host toString ,':' ,http port.
	^http! !

! SystemServer methodsFor: #luaSimple !
luaSimpleRequestBlock
	" Private - Return the request processing block used for remote system tools. "

	^self
		requestBlockFor: [:uri :headers| uri unescaped ]
		method: [:uri :headers| #GET ]
		post: [:req :action|
			req notImplementedYet: #luaSimpleRequestBlock ,' (POST)'.
			#ok
		]! !

! SystemServer methodsFor: #luaSimple !
luaSimpleParse: string
	" Private - Return the result of parsing JSON string. "

	| json js args |
	js := Json parse: string.
	js isNil ifTrue: [ ^self error: 'Broken Json :',string ].
	args := js @ #arguments.
	{'args' lua'smalltalk:newArray(args)'}.
	json := ((#selector -> (js @ #selector) toString)
		,(#arguments -> args)) json.
	^NativeHandle support @ json! !

! SystemServer methodsFor: #luaSimple !
luaSimpleStringify: literal
	" Private - Return the result of JSON string. "

	^Json stringify: NativeHandle support @ literal! !