--[[
S8 - A modern Smalltalk environment running on the "invisible machine".
Copyright (C) 2018-today Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
For questions and bug reports join us at S8 Facebook group
	https://www.facebook.com/groups/s8smalltalk/
--]]

local smalltalk = {}
smalltalk["$nativeLanguage"] = "lua"
smalltalk.useManager = false

function smalltalk.do_onError_(self, aBlock, errorBlock)
	local status, result = pcall(aBlock)
	if status then return result end
	return errorBlock(result) 
end

function smalltalk.isNil_(self, aValue)
	if not aValue then return aValue==nil end
	if type(aValue)=="table" then
		if type(aValue.isNil)=="function" then
			return aValue:isNil()
		end
	end
	return false
end

function smalltalk.notNil_(self, aValue)
	if not aValue then return aValue~=nil end
	if type(aValue)=="table" then
		if type(aValue.notNil)=="function" then
			return aValue:notNil()
		end
	end
	return true
end

function smalltalk.basic_at_ifAbsent_(self, receiver, key, aBlock)
	local result = receiver[key]
	if self:isNil_(result) then
		return aBlock:value()
	end
	return result
end

function smalltalk.basic_at_put_(self, receiver, key, aValue)
	receiver[key] = aValue
	return aValue
end

function smalltalk.ifUndefined_do_(self, receiver, aBlock)
	local result = receiver
	if self:isNil_(result) then
		return aBlock:value()
	end
	return result
end

function smalltalk.orNil_(self, receiver)
	local result = receiver
	if self:isNil_(result) then return nil end
	return result
end

function smalltalk.newConstructor (self, super)
	-- return a new constructor of instances inheriting from super (constructor)
	-- super can be nil to return a constructor for a root class
	local constructor = {}
	local prototype = {}
	prototype.__index = prototype
	if super then setmetatable(prototype,super.prototype)
	else
		local dnu = {}
		dnu.__index = function (rootPrototype, key)
			--print("--keyNotFound: "..key)
			smalltalk.dnu=key
			return nil
		end
		setmetatable(prototype,dnu)
	end
	setmetatable(constructor,prototype)
	constructor.prototype = prototype
	prototype.__call = function (f,rcv,...)
		if f==nil and smalltalk.dnu then
			if string.find(smalltalk.dnu,"[%a_%+%-%/%*%|%&%=%<%>%,%~%\\]")~=1 then return nil end
--DEBUG
print("--DNU "..tostring(smalltalk.dnu))
if tostring(smalltalk.dnu):first():isUppercase() then return nil end
--DEBUG DNU detected dump stack trace
if debug then
	print(debug.traceback("Stack trace"))
	print(debug.getinfo(1))
	print("Stack trace end")
end
--if smalltalk.dnu=="arguments" then error("DNU exiting...") end
error("DNU")
			-- Syntetize a doesNotUnderstand message send
			local args = (pack or table.pack)(...)
			local msg = smalltalk.Message:receiver_selector_arguments_(
				rcv, smalltalk.dnu,
				smalltalk:newArray(args)
			)
			smalltalk.dnu = nil
			local result = nil
			if smalltalk:isObject_(rcv) then
				result = rcv:doesNotUnderstand_( msg )
			else -- DNU goes to smalltalk for native objects
				result = smalltalk:doesNotUnderstand_( msg )
			end
			return result
		end
		local result = {}
		setmetatable(result,prototype)
		return result
	end
	return constructor
end

local StObject = smalltalk:newConstructor()
local StBehavior = smalltalk:newConstructor(StObject)
local StClass = smalltalk:newConstructor(StBehavior)
local StMetaclass = smalltalk:newConstructor(StBehavior)
StMetaclass.prototype["$meta"] = true
local StMethod = smalltalk:newConstructor(StObject)
local StSystemDictionary = smalltalk:newConstructor(StObject)

function smalltalk.will_be_(self, anObject, className)
	-- return an object that will become an instance of class (named className) when accessed
	local result = anObject
	local mt = {}
	mt.__index = function(table, key)
		if self[className] then
			self:become_a_(table,self[className])
		end
		return table[key]
	end
	setmetatable(result,mt)
	return result
end

function smalltalk.linkNative_to_ (self, instance, aClass)
	-- link native species of instance to delegate to aClass instance
	local mt = getmetatable(instance)
	if mt	then
		local index=mt["__index"]
		setmetatable(index,aClass["$fn"].prototype)
	else
		if debug then
			self:do_onError_(function()
				debug.setmetatable(instance,aClass["$fn"].prototype)
			end ,function()
				print("-- ERROR cant set metatable for "..tostring(instance))
			end)
		else
			print("-- WARNING: debug support missing")
		end
	end
end

function smalltalk.newArray(self, object)
	-- return a new Array instance mutating the provided object
	return self:will_be_(object or {},"Array")
end

function smalltalk.become_a_(self, object, aClass)
	-- mutate object to an instance of aClass
	setmetatable(object,aClass["$fn"].prototype)
	return object
end

function smalltalk.setProtoOf_to_ (self, nativeObject, proto)
	local mt = {}
	mt.__index = proto
	setmetatable(nativeObject,mt)
end

function smalltalk.bindRoots(self, metaclass)
	-- Private - Link the system roots to ensure proper delegation chains
	local objectClassPrototype = self.Object["$klass"]["$fn"].prototype
	self:setProtoOf_to_(objectClassPrototype,self.Class["$fn"].prototype)
	self:setProtoOf_to_(self.Smalltalk["$klass"]["$fn"].prototype,objectClassPrototype)
	self:setProtoOf_to_(self.Behavior["$klass"]["$fn"].prototype,objectClassPrototype)
	self.Object["$klass"]["$superclass"] = self.Class
	self:become_a_(smalltalk,self.Smalltalk)
end

function smalltalk.klass(self, spec)
	local aClass
	if spec.meta then
		aClass = StMetaclass()
	else
		local metaSpec = {meta=true}
		metaSpec.className = spec.className .." class"
		if spec.superclass then
			metaSpec.superclass = spec.superclass["$klass"]
		end
		local metaclass = self:klass(metaSpec)
		aClass = metaclass["$fn"]()
		if aClass["$klass"]~=metaclass then -- this should NOT happen
			print("--! "..(spec.className).." BROKEN metaclass="..tostring(aClass["$klass"]))
			aClass["$klass"] = metaclass
		end
	end
	self:setClass(aClass,spec)
	aClass["$fn"].prototype["$klass"] = aClass
	aClass["$fn"].prototype["$methods"] = {}
	if aClass["$className"]=="Metaclass" then
		self:bindRoots(aClass)
	end
	return aClass
end

function smalltalk.setVarNames(self, aClass, iVarNames, aValue)
	if iVarNames==nil then iVarNames = self:newArray() end
	aClass["$iVarNames"] = iVarNames
	local i = #iVarNames
	while (i > 0) do
		aClass["$fn"].prototype["@"..iVarNames[i]] = aValue
		i = i - 1
	end
end

function smalltalk.setClass (self, aClass, spec)
	aClass["$className"] = spec.className
	if spec.meta==nil then
		aClass["$klass"]["@instanceClass"] = aClass
	end
	aClass["$category"] = spec.category or ""
	aClass["$superclass"] = spec.superclass --or nil
	if aClass["$superclass"] then
		aClass["$superclass"]["@subclasses"] = nil
		if aClass["$klass"] then
			if spec.meta and (aClass["$klass"]~=self.Metaclass) then
				aClass["$klass"]["$superclass"] = aClass["$superclass"]["$klass"]
			end

			if self.Metaclass~=nil then -- OBSOLETE - Metaclass validation & repair
				if self.Metaclass["$superclass"]~=self.Behavior then
					print("-- Warning broken Metaclass spec.meta="..tostring(spec.meta).." aClass="..aClass["$className"])
					self.Metaclass["$superclass"]=self.Behavior
				end
			end

			if aClass["$klass"]["$superclass"] then
				aClass["$klass"]["$superclass"]["@subclasses"] = nil
			end
		end
	end
	aClass["$fn"] = self:fnInSpec(spec)
	self:setVarNames(aClass,spec.iVarNames,nil)
	return aClass
end

function smalltalk.fnInSpec (self, spec)
	if spec.meta and (spec.superclass == nil) then
		spec.superclass = self.Class
	end
	if spec.fn then
		if spec.superclass then
			self:setProtoOf_to_(spec.fn.prototype,spec.superclass["$fn"].prototype)
		end
		return spec.fn
	end
	local superC = nil
	if spec.superclass then
		superC = spec.superclass["$fn"]
	end
	return self:newConstructor(superC)
end

function smalltalk.classes (self)
	local result = self:newArray()
	for key, aClass in pairs(smalltalk) do
		if type(aClass)=="table" then
			if aClass.isClass and aClass:isClass() then
				result[#result+1] = aClass
			end
		end
	end
	return result
end

function smalltalk.computeSubclasses (self)
	if self.Metaclass==nil then print("-- Warning system broken/incomplete") end
	if self.Metaclass["$superclass"]~=self.Behavior then -- OBSOLETE - Metaclass validation & repair
		print("-- Warning broken Metaclass")
		self.Metaclass["$superclass"]=self.Behavior
	end
	local classes = self:classes()
	for _,cls in pairs(classes) do
		cls["@subclasses"] = self:newArray()
		cls["$klass"]["@subclasses"] = self:newArray()
	end
	for _,cls in pairs(classes) do
		if cls["$superclass"]~=nil then
			local array = cls["$superclass"]["@subclasses"]
			array[#array+1] = cls
			array = cls["$superclass"]["$klass"]["@subclasses"]
			array[#array+1] = cls["$klass"]
		end
	end
	if self.Class then
		self.Class["@subclasses"] = self:newArray()
		self.Class["$klass"]["@subclasses"] = self:newArray()
	end
end

function smalltalk.subclasses (self, aClass)
	if aClass["@subclasses"]==nil then
		self:computeSubclasses()
	end
	return aClass["@subclasses"]
end

function smalltalk.allSubclassesDo (self, aClass, aBlock)
	for _,cls in pairs(self:subclasses(aClass)) do
		aBlock(cls)
		self:allSubclassesDo(cls,aBlock)
	end
end

function smalltalk.mapClassName (self, className, category, fn, superclass)
	local spec = {
		className = className, 
		category = category, 
		superclass = superclass,
		fn = fn }
	if self[className] then
		return self:setClass(self[className],spec)
	end
	self[className] = self:klass(spec)
	return self[className]
end

function smalltalk.isArray_equalTo_(self, table1, table2)
	if table1==table2 then return true end
	if #table1 ~= #table2 then return false end
	local i = #table1
	while (i > 0) do
		if table1[i] ~= table2[i] then return false end
		i = i - 1
	end
	return true
end

function smalltalk.isObject_equalTo_(self, receiver, anObject)
	if receiver==anObject then return true end
	if type(receiver)=="table" and type(anObject)=="table" then
		if receiver["_eq"] then
			return receiver:_eq(anObject)
		end
	end
	return false
end

function smalltalk.object_comma_(self, receiver, anObject)
	if type(receiver)=="string" then
		if type(anObject)=="string" then return receiver..anObject end
		if type(anObject)=="number" then return receiver..tostring(anObject) end
		return self:object_comma_(receiver,(receiver:toString_(anObject)))
	end
	if type(receiver)=="number" then
		return self:object_comma_(tostring(receiver),anObject)
	end
	if not (type(receiver)=="table") then
		return self:object_comma_(self:toString_(receiver),anObject)
	end
	return receiver:_comma(anObject)
end

function smalltalk.boolean_and_(self, receiver ,option)
	if receiver~=true then return false end
	if type(option)=="function" then return option() end
	if type(option)=="boolean" then return option end
	return option:evaluate()==true;
end

function smalltalk.boolean_or_(self, receiver, option)
	if receiver==true then return true end
	if type(option)=="function" then return option() end
	if type(option)=="boolean" then return option end
	return option:evaluate()==true;
end

function smalltalk.booleanNot_(self, receiver)
	if receiver==true then return false end
	if type(receiver)~="table" then return not receiver end
	return receiver:__not()
end

function smalltalk.times_repeat_(self, count, aBlock)
	for i=1,count do aBlock(i) end
	return count
end

function smalltalk.if_nil_notNil_(self, anObject, block1, block2)
	if self:isNil_(anObject) then
		return block1:evaluate()
	end
	return block2:evaluate()
end

function smalltalk.if_true_false_(self, aBoolean, blk1, blk2)
	if aBoolean==true then
		if blk1 then return blk1() end
	else
		if blk2 then return blk2() end
	end
	return nil
end

function smalltalk.while_do_(self, condition, aBlock)
	while condition()==true do
		if aBlock then aBlock() end
	end
	return nil
end

function smalltalk.whileNot_do_(self, condition, aBlock)
	while condition()==false do
		if aBlock then aBlock() end
	end
	return nil
end

function smalltalk.asCharacter_(self, receiver)
	if type(receiver)=="number" then
		return string.char(receiver)
	end
	return receiver:asCharacter()
end

function smalltalk.toString_(self, receiver)
	local aType = type(receiver)
	if aType=="string" then return receiver end
	if aType=="number" then return tostring(receiver) end
	if aType=="boolean" then
		if receiver then return "true" end
		return "false"
	end
	if type(receiver.toString)=="function" then return receiver:toString() end
	return tostring(receiver)
end

function smalltalk.printString_(self, receiver)
	if type(receiver)=="string" then return receiver:printString() end
	if type(receiver)=="table" and receiver.printString then return receiver:printString() end
	return tostring(receiver)
end

function smalltalk.asString_(self, receiver)
	if type(receiver)=="table" and receiver.asString then return receiver:asString() end
	return tostring(receiver)
end

function smalltalk.mustBeNumber_(self, receiver)
	if type(receiver)=="number" then return receiver end
	if type(receiver)=="table" then return receiver:mustBeNumber() end
	return self:error_("Must be a Number")
end

function smalltalk.math_max_(self, receiver, aNumber)
	if type(receiver)~="number" then return receiver:max_(aNumber) end
	if type(aNumber)~="number" then return aNumber:max_(receiver) end
	if receiver>aNumber then return receiver end
	return aNumber
end

function smalltalk.math_min_(self, receiver, aNumber)
	if type(receiver)~="number" then return receiver:min_(aNumber) end
	if type(aNumber)~="number" then return aNumber:min_(receiver) end
	if receiver<aNumber then return receiver end
	return aNumber
end

function smalltalk.addClass (self, className, superclass, iVarNames, category)
	self:debugAddClass(className, superclass, iVarNames, category)
	if(self[className]) then
		local aClass = self[className]
		local sameShape = (aClass["$superclass"] == superclass)
			and (self:isArray_equalTo_(aClass["$iVarNames"],iVarNames))
		if sameShape and (aClass["$category"] == category) then
			return aClass
		end
		self:aboutToChangeClass(aClass,superclass,iVarNames,category)
		if(aClass["$superclass"]) then
			aClass["$superclass"]["@subclasses"] = nil
			local aMeta = aClass["$klass"]
			if aMeta and superclass then
				aMeta["$superclass"]["@subclasses"] = nil
				aMeta["$superclass"] = superclass["$klass"]
			end
		end
		aClass["$superclass"] = superclass
		if(superclass and (superclass["@subclasses"])) then
			if superclass["@subclasses"] == nil then
				superclass["@subclasses"] = self:newArray()
			end
			superclass["@subclasses"].push(aClass) 
		end
		self:setVarNames(aClass,iVarNames,nil)
		aClass["$category"] = category or aClass["$category"]
		if sameShape == nil then self:classShapeChanged(aClass) end
	else
		self:aboutToCreateClass(className,superclass,iVarNames,category)
		self[className] = self:klass({
			className = className, 
			iVarNames = iVarNames,
			superclass= superclass
		})
		self[className]["$category"] = category or ""
	end
	self:classChanged(self[className])
	return self[className]
end

function smalltalk.renameClass (self, aClass, aString)
	self[aClass["$className"]] = nil
	aClass["$className"] = aString
	self[aString] = aClass
	self:classChanged(aClass)
end

function smalltalk.removeClass (self, aClass)
	local deleted = "Deleted-"
	if aClass["$superclass"] then aClass["$superclass"]["@subclasses"] = nil end
	self[aClass["$className"]] = nil
	aClass["$className"] = deleted .. aClass["$className"]
	aClass["$klass"]["$className"] = deleted .. aClass["$klass"]["$className"]
	self:classChanged(aClass)
end

function smalltalk.metaVarNames (self, aMetaClass, iVarNames)
	if self:isArray_equalTo_(aMetaClass["$iVarNames"],iVarNames) then return false end
	self:setVarNames(aMetaClass,iVarNames,nil)
	self:classShapeChanged(aMetaClass)
	self:classChanged(aMetaClass)
	return true
end

function smalltalk.smDo (self, block)
	if (self.useManager and self.SystemManager
		and self.SystemManager.currentOrNil) then
		local sm = self.SystemManager:currentOrNil()
		if sm then return block(sm) end
	end
end

function smalltalk.classChanged (self, aBehavior)
	self:smDo(function(sm) sm:classChanged_(aBehavior) end)
end
function smalltalk.classShapeChanged (self, aBehavior)
	self:smDo(function(sm) sm:classShapeChanged_(aBehavior) end)
end
function smalltalk.aboutToCreateClass (self, className, superclass, iVarNames, category)
	self:smDo(function(sm)
		sm:aboutToCreateClass_superclass_iVarNames_category_(className,superclass,iVarNames,category)
	end)
end
function smalltalk.aboutToChangeClass (self, aClass, superclass, iVarNames, category)
	self:smDo(function(sm)
		sm:aboutToChange_superclass_iVarNames_category_(aClass,superclass,iVarNames,category)
	end)
end
function smalltalk.methodBoundTo (self, method, klass)
	self:smDo(function(sm) sm:methodBound_to_(method,klass) end)
	return method
end
function smalltalk.aboutToBindTo (self, method,klass)
	self:smDo(function(sm) sm:aboutToBind_to_(method,klass) end)
	return method
end
function smalltalk.aboutToRemoveFrom (self, method, klass)
	self:smDo(function(sm) sm:aboutToRemove_from_(method,klass) end)
	return method
end
function smalltalk.methodRemovedFrom (self, method, klass)
	self:smDo(function(sm) sm:methodRemoved_from_(method,klass) end)
	return method
end
function smalltalk.objectNotFound (self, object, indexOrKey, collection)
	self:smDo(function(sm) sm:object_at_notFoundIn_(object,indexOrKey,collection) end)
	return collection
end

function smalltalk.whoImplements (self, klass, jsSelector)
	local cls = klass
	repeat
		if(rawget(cls["$fn"].prototype,jsSelector)) then return cls end
		cls = cls["$superclass"]
	until (cls==nil)
	return nil
end

function smalltalk.superImplementor (self, klass, jsSelector)
	if (self:isNil_(klass["$superclass"])
		or self:isNil_(klass["$superclass"]["$fn"])) then
		return klass:missingImplementorFor_(jsSelector) end
	local cls = self:whoImplements(klass["$superclass"],jsSelector)
	if self:isNil_(cls) then
		return klass["$superclass"]:missingImplementorFor_(jsSelector)
	end
	return cls["$fn"].prototype[jsSelector]
end

function smalltalk.addMethod (self, jsSelector, method, klass)
	self:aboutToBindTo(method,klass)
	klass["$fn"].prototype[jsSelector] = method["$fn"]
	klass["$fn"].prototype["$methods"][method["$selector"]] = method
	method["@classField"] = klass
	return self:methodBoundTo(method,klass)
end

function smalltalk.bind (self, klass, stSelector, jsSelector, jsFunction, category, stSource, nativeSource ,stamp)
	self:debugBind(klass,stSelector,jsSelector,jsFunction,category,stSource,nativeSource,stamp)
	local method = self.CompiledMethod["$fn"]()
	method["$selector"] = stSelector
	method["$fn"] = jsFunction
	method["$category"] = category
	method["$source"] = stSource
	method["$nativeSource"] = nativeSource
	if self:notNil_(stamp) then method["$stamp"] = stamp end
	method["$generation"] = self["$state"]
	if self:notNil_(klass) then
		if(jsSelector == 0) then jsSelector = string.gsub(stSelector,":","_") end
		self:addMethod(jsSelector,method,klass)
	end
	return method
end

function smalltalk.removeMethod (self, aClass, aMethod)
	self:aboutToRemoveFrom(aMethod,aClass)
	aClass["$fn"].prototype[aMethod["$selector"]:asSelector()] = nil
	aClass["$fn"].prototype["$methods"][aMethod["$selector"]] = nil
	return self:methodRemovedFrom(aMethod,aClass)
end

function smalltalk.newMethod (self, currentClass, selector, fn, source, nativeSource, stamp)
	return self:bind(nil,selector,nil,fn,nil,source,nativeSource,stamp)
end
function smalltalk.method (self, spec) -- OBSOLETE
	return self:bind(nil,spec.selector,nil,spec.fn,spec.category,spec.source,spec.nativeSource,spec.stamp)
end

function smalltalk.readJSObject (self, js)
	print("--[NIY] smalltalk:readJSObject()")
	return nil
end

function smalltalk.basen(self, n, b)
	local floor,insert = math.floor, table.insert
	n = floor(n)
	if not b or b == 10 then return tostring(n) end
	local digits = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ"
	local t = {}
	local sign = ""
	if n < 0 then
		sign = "-"
		n = -n
	end
	repeat
		local d = (n % b) + 1
		n = floor(n / b)
		insert(t, 1, digits:sub(d,d))
	until n == 0
	return sign .. table.concat(t,"")
end

function smalltalk.globalsDo (self, action)
	for key, aValue in pairs(smalltalk) do
		if type(aValue)=="table" then
			if aValue.isClass and not aValue:isClass()
				then action(key) end
		end
	end
	return smalltalk
end

function smalltalk.launchImage (self, imageName)
	self.useManager = true
	if(self:notNil_(imageName)) then
		self["$imageFileName"] = imageName
		dofile(imageName) --load(imageName)
	end
	self["$state"] = "launching"
	if(self.launch) then self:launch() end
	self["$state"] = "running"
	return smalltalk
end

function smalltalk.is_kindOf_ (self, anObject, aClass)
	if(self:isNil_(anObject) or (self:isNil_(anObject.isKindOf_))) then return false end
	return anObject:isKindOf_(aClass)
end

function smalltalk.responds_to_ (self, anObject, aSelector)
	if(self:isNil_(anObject) or (self:isNil_(anObject.respondsTo_))) then return false end
	return anObject:respondsTo_(aSelector)
end

function smalltalk.debugBind (self,klass,stSelector,jsSelector,jsFunction,category,stSource,nativeSource,stamp) return self end
function smalltalk.debugAddClass (self,className, superclass, iVarNames, category) return self end

function smalltalk.toString (self)
	return "smalltalk"
end

function smalltalk.bindCore (self) --bind the minimal functions (e.g i/o)
	local global = _G
	if global==nil then return false end

	function global.quit ()
		os.exit()
		return false
	end
	function global.version ()
		return _VERSION
	end
	function global.fileContents (fileName)
		local f,err = io.open(fileName, "rb")
		if f==nil then return nil end
		local content = f:read("*all")
		f:close()
		return content
	end
	global.read=global.fileContents
	function global.fileWrite (fileName,contents,mode)
		local f = assert(io.open(fileName, mode))
		f:write(contents)
		f:close()
		return contents:size()
	end

	function global.fileIn (fileName)
		local contents = read(fileName)
		if contents==nil then 
			smalltalk:error_("Missing file "..tostring(fileName)) 
		end
		return contents:fileIn()
	end
	function global.emmit (fileName)
		return smalltalk:emmit_(fileName)
	end
	function global.fileInJS (fileName)
		return smalltalk:eval_(read(fileName))
	end

	function global.exportCategory (aString)
		return smalltalk.Exporter:exportCategory_(aString)
	end
	function global.saveImage ()
		return smalltalk:saveImage()
	end
	function global.saveImageAs (fileName)
		return smalltalk:saveImageAs_(fileName)
	end

	function global.isFileOlderThan (fileName,referenceFileName)
		local f,err = io.open(fileName, "r")
		if f==nil then return false end
		f:close()
		f,err = io.open(referenceFileName, "r")
		if f==nil then return false end
		f:close()
		-- there is NO way to get the datetime of a file in Lua
		return true
	end

	self["$imageFileName"] = self["$imageFileName"] or "s8.image.lua"
	return true
end

function smalltalk.startup (self)
	self["$state"] = "booting"
	self.OverwriteMethods = self.OverwriteMethods or true
	self.DumpOverwritenMethods = self.DumpOverwritenMethods or false

	self["$state"] = "singularity"
	self:mapClassName("Object","Kernel",StObject)
	self:mapClassName("Smalltalk","Kernel",StSystemDictionary,self.Object)
	self:mapClassName("Behavior","Kernel",StBehavior,self.Object)
	self:mapClassName("Class","Kernel",StClass,self.Behavior)
	self:mapClassName("Metaclass","Kernel",StMetaclass,self.Behavior)
	self:mapClassName("CompiledMethod","Kernel",StMethod,self.Object)
	self:mapClassName("Number","Kernel",nil,self.Object)
	self:mapClassName("BlockClosure","Kernel",nil,self.Object)
	self:mapClassName("Boolean","Kernel",nil,self.Object)
--	self:mapClassName("DateTime","Kernel",Date,self.Object)
	self:mapClassName("UndefinedObject","Kernel",nil,self.Object)
	self:mapClassName("Collection","Kernel",nil,self.Object)
	self:mapClassName("SequenceableCollection","Kernel",nil,self.Collection)
	self:mapClassName("String","Kernel",nil,self.SequenceableCollection)
	self:mapClassName("Array","Kernel",nil,self.SequenceableCollection)
--	self:mapClassName("RegularExpression","Kernel",RegExp,self.Object)
	self["$state"] = "whiteHole"
	self:linkNative_to_("",self.String)
	self:linkNative_to_(function()end,self.BlockClosure)
	self:linkNative_to_(true,self.Boolean)
	self:linkNative_to_(nil,self.UndefinedObject)
	self:linkNative_to_(1,self.Number)
	self["$state"] = "wormHole"
	self:bindCore()
	self["$state"] = "startup"
	return self
end

return smalltalk:startup()