" LuaStudio wrappers
Disclaimer: we are NOT recomending/suggesting to use Lua Studio for application development,
 this API implementation has been written with the sole goal to test our native access features for Lua.
Copyright (C) 2018 Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
See also http://scormpool.com/LuaStudio/doc/html/LuaStudio%20API%20documentation.html
"!

NativeObject
	subclass: #LuaStudioLibrary instanceVariableNames: ''
	classVariableNames: #HasBeenInstalled poolDictionaries: ''
	category: #LuaStudio!
LuaStudioLibrary subclass: #Lib!
LuaStudioLibrary subclass: #EventDispatcher!

! LuaStudioLibrary class methodsFor: #api !
withString: data do: aBlock
	" Private - If data is an instance of String, returns the result of evaluating aBlock (or nil). "

	^(self typeOf: data) = #string ifTrue: aBlock! !

! LuaStudioLibrary class methodsFor: #u8 !
pathToSystemFolder
	" Private - Returns the path for system root (or nil). "

	(#(android ios) includes: System systemName) ifTrue: [
		^MediaFile documentsDirectory nativePath ,$/
	].
	^$/! !

! LuaStudioLibrary class methodsFor: #u8 !
pathToM8n
	" Private - Returns the local path of m8n folder (or nil). "

	^self pathToSystemFolder ,'m8n/'! !

! LuaStudioLibrary class methodsFor: #u8 !
pathToU8Cache
	" Private - Returns the local path of U8 cache folder (or nil). "

	| contributionURL |
	contributionURL := Smalltalk at: #ContributionURL ifAbsent: [ ^nil ].
	^self pathToM8n ,(contributionURL fromLast: '/profile/')! !

! LuaStudioLibrary class methodsFor: #u8 !
isLocalPath: pathName
	" Private - Returns true if the pathName is a local file name. "

	^(pathName includes: $:)
	or: [(pathName startsWith: $/)
	or: [(pathName startsWith: $\)]]! !

! LuaStudioLibrary class methodsFor: #u8 !
isAssetPath: pathName
	" Private - Returns true if the file at pathName is stored in assets. "

	^(self isLocalPath: pathName) not! !

! LuaStudioLibrary class methodsFor: #u8 !
pathToFile: pathName
	" Private - Returns the local path of file at pathName. "

	(self isAssetPath: pathName) ifTrue: [ ^'/assets/' ,pathName ].
	^pathName! !

! LuaStudioLibrary class methodsFor: #u8 !
withU8Path: aPath do: aBlock
	" Private - Ensure a local file at aPath exists.
	If aPath is only a fileName will search for file in well known folders; if a file exists that local path is used.
	Path for files stored in contribution starts with u8:/
	If file does not exist try to download the file from contribution path.
	Returns the result of evaluating aBlock with local(downloaded) file path or nil.
	"

	| u8Root path contributionURL filePath url |
	HasBeenInstalled == true ifFalse: [
		HasBeenInstalled := true.
		self withAllSubclasses do: [:cls| cls launch ]
	].

	u8Root := 'u8:/'.
	(self isAssetPath: aPath) ifTrue: [ "fileName in default folder"
		(Smalltalk includesKey: #ContributionURL) ifTrue: [
			^self withU8Path: u8Root ,aPath do: aBlock
		].
		filePath := self pathToM8n ,path.
		(Sys fileSystem exists: filePath) ifTrue: [ ^aBlock value: filePath ].
		filePath := self pathToSystemFolder ,path.
		(Sys fileSystem exists: filePath) ifTrue: [ ^aBlock value: filePath ].
		^nil
	].

	(aPath startsWith: u8Root) not ifTrue: [
		((self isLocalPath: aPath) and: [Sys fileSystem exists: aPath]) ifTrue: [
			^aBlock value: aPath
		].
		^nil ].
	path := aPath copyFrom: u8Root size + 1 to: aPath size.
	contributionURL := Smalltalk at: #ContributionURL ifAbsent: [
		^self error: 'Contribution URL missing'
	].
	filePath := self pathToU8Cache ,path.
	(Sys fileSystem exists: filePath) ifTrue: [ ^aBlock value: filePath ].
	Sys fileSystem createDirectory: (filePath upToLast: $/) ,$/.

	url := contributionURL ,path.
	Http get: url onBytes: [:hBytes|
		self note: 'Downloading ' ,filePath.
		IOFile new
			saveBytes: filePath
			bytes: (ByteArray fromBytes: Bytes @ hBytes).
	] onError: [:err|
		# #print: err with: url.
		^self error: err
	].
	^aBlock value: filePath! !

! LuaStudioLibrary class methodsFor: #u8 !
deleteU8Cache
	" Delete all the files cached for current contribution. "

	| folder |
	folder := self pathToU8Cache.
	^folder notNil and: [ Sys fileSystem prune: folder ]! !

! LuaStudioLibrary class methodsFor: #u8 !
deleteM8n
	" Delete all the files cached for m8n platform. "

	^Sys fileSystem prune: self pathToM8n! !

! LuaStudioLibrary class methodsFor: #u8 !
withPath: aPath do: aBlock
	" Private - Returns the result of evaluating aBlock with file path. "

	| cached done |
	done := self withU8Path: aPath do: [:path|
		cached := aBlock value: path.
		#done
	].
	^done isNil ifTrue: [ aBlock value: aPath ] ifFalse: [ cached ] ! !

! LuaStudioLibrary methodsFor: #reflect !
send: funcName withArguments: args ifFailed: aBlock
	" Private - Call the method funcName on the receiver in safe mode.
	Return the result of evaluation or the result of evaluating aBlock if failed.
	"

	| result err |
	result := Lib reflect
		callMethodSafe: self funcName: funcName
		args: args errFunc: [ err := true ].
	err == true ifTrue: [ ^aBlock value ].
	^result! !

! LuaStudioLibrary methodsFor: #reflect !
send: funcName ifFailed: aBlock
	" Private - Call the method funcName on the receiver in safe mode.
	Return the result of evaluation or the result of evaluating aBlock if failed.
	"

	^self	send: funcName
		withArguments: #()
		ifFailed: aBlock! !

! String methodsFor: #converting !
asLocalPath
	" Return the path to a local file pathName, read from U8 service if required (not cached). "

	| result |
	LuaStudioLibrary withPath: self do: [:filePath| result := filePath ].
	^result! !

! LuaStudioLibrary class methodsFor: #builders !
buildConstants: constants
	" Build the methods for constants. "

	^self build: #Constants -> constants doing: [
		constants isString ifTrue: [ | stream |
			stream := constants stream.
			[stream atEnd] whileFalse: [
				self buildConstant: stream nextLine trimBlanks
			]
		] ifFalse: [ constants do: [:each| self buildConstant: each ] ]
	]! !

! LuaStudioLibrary class methodsFor: #builders !
buildConstant: spec
	" Private - Build the accessor for constant spec. "

	| selector |
	spec isEmpty ifTrue: [ ^self ].
	selector := (spec fromLast: $.) asSmalltalkName: false.
	self class
		implement: selector
		comment: 'Return the constant value for ' ,spec
		as: '^{' ,$' ,spec ,$' ,'}' category: #constants! !

! LuaStudioLibrary class methodsFor: #builders !
modules: modules
	" Build support implementing link to module. "
 
	modules do: [:each|
		EventDispatcher subclass: each.
		self class
			implement: each asLocalName
			comment: 'Returns a wrapper for library'
			as: '^' ,each category: #accessing
	]! !

#(	Away3D Media OAuth Project
	Reflect SQLight Sys WebServer
	WebView Web
) do: [:each|
	LuaStudioLibrary subclass: each.
	Lib class
		implement: each asLocalName
		comment: 'Returns a wrapper for library'
		as: '^' ,each ,' new' category: #accessing.
	(Smalltalk at: each)
		implement: #initializeHandle
		comment: 'Private - Initialize the handle of the receiver.'
		as: 'handle := #{Lib}[#' ,each ,']'
		category: #initialize.
]!

Project buildLuaFunctions: #(
	#(getText #(pathInProject))
	#(getBytes #(pathInProject) ByteArray)
)!

Reflect buildLuaFunctions: #(
	#(hasField #(_o field))
	#(field #(_o field))
	#(setField #(_o field value))
	#(getProperty #(_o field))
	#(setProperty #(_o field value))
	#(callMethod #(_o func args))
	#(callMethodSafe #(_o funcName args errFunc))
	#(fields #(_o))
	#(isFunction #(_f))
	#(isObject #(_v))
	#(isEnumValue #(_v))
	#(deleteField #(_o field))
	#(copy #(_o))
)!

WebView buildSetters: #(
	onClose onURLChanging
)!
WebView buildLuaFunctions: #(
	isSupported
	#(open #(url userAgent floating urlWhiteList urlBlackList))
)!

WebServer buildLuaFunctions: #(
	#(start #(host port path oncomplete))
	#(isStarted #(host port))
	#(stop #(host port))
)!

#(	Crypto IO Json EReg
	FileSystem Net Utf8
	VM Xml Zip
) do: [:each|
	LuaStudioLibrary subclass: each.
	Sys class
		implement: each asLocalName
		comment: 'Returns a wrapper for library'
		as: '^' ,each ,' new' category: #accessing.
	(Smalltalk at: each)
		implement: #initializeHandle
		comment: 'Private - Initialize the handle of the receiver.'
		as: 'handle := #{Lib}[#Sys][#' ,each ,']'
		category: #initialize.
]!
Sys buildGetters: #( version )!
Sys buildLuaFunctions: #(
	getStamp getTime
	#(trace #(message))
	#(throw #(errorString))
	#(sleep #(seconds))
)!

#(	Base64 Md5 Sha1 Sha224 Sha256
) do: [:each|
	Crypto subclass: each.
	Crypto class
		implement: each asLocalName
		comment: 'Returns a wrapper for library'
		as: '^' ,each ,' new' category: #accessing.
	(Smalltalk at: each)
		implement: #initializeHandle
		comment: 'Private - Initialize the handle of the receiver.'
		as: 'handle := #{Lib}[#Sys][#Crypto][#' ,each ,']'
		category: #initialize.
]!
Crypto buildLuaFunctions: #(
	#(encode #(string))
	#(decode #(string))
)!
Base64 buildLuaFunctions: #(
	#(encode #(_bytes complement))
	#(decode #(string complement) Bytes)
)!

IO subclass: #Bytes!
! IO class methodsFor: #accessing !
bytes
	" Returns a wrapper for library "

	^Bytes! !

! Bytes class methodsFor: #instantiation !
alloc: length
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.IO.Bytes.alloc(length))'}! !

! Bytes class methodsFor: #instantiation !
new: length
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.IO.Bytes.alloc(length))'}! !

! Bytes class methodsFor: #instantiation !
ofString: aString
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.IO.Bytes.ofString(aString))'}! !

Bytes buildGetters: #(
	length
)!
Bytes buildLuaFunctions: #(
	toString toHex
	#(getString #(pos len))
	#(fill #(pos len value))
	#(sub #(pos len) Bytes)
	#(blit #(pos _src srcPos len))
	#(get #(pos))
	#(set #(pos value))
)!

! Bytes methodsFor: #accessing !
size
	" Returns the size of the receiver. "

	^self length! !

LuaStudioLibrary subclass: #Input!
Input buildAccessors: #(
	bigEndian
)!
Input buildLuaFunctions: #(
	readByte close
	#(readBytes #(_bytes pos len))
	#(readAll #(bufSize))
	#(readFullBytes #(_bytes pos len))
	#(read #(nbytes))
	#(readUntil #(end))
	readLine
	#(readString #(len))
	readFloat readDouble
	readInt8 readInt16 readUInt16
	readInt24 readUInt24
	readInt32
)!

! Input methodsFor: #nonBlocking !
readLineOr: anObject
	" Returns the next line read from the receiver (or anObject). "

	^self send: #readLine ifFailed: [ anObject ]! !

! Input methodsFor: #nonBlocking !
readLineOrNil
	" Returns the next line read from the receiver (or nil). "

	^self readLineOr: nil! !

LuaStudioLibrary subclass: #Output!
Output buildAccessors: #(
	bigEndian
)!
Output buildLuaFunctions: #(
	#(writeByte #(c))
	#(writeBytes #(_bytes pos len))
	flush close
	#(writeFullBytes #(_bytes pos len))
	#(writeFloat #(x))
	#(writeDouble #(x))
	#(writeInt8 #(x))
	#(writeInt16 #(x))
	#(writeUInt16 #(x))
	#(writeInt24 #(x))
	#(writeUInt24 #(x))
	#(writeInt32 #(x))
	#(prepare #(nbytes))
	#(writeInput #(input bufsize))
	#(writeString #(string))
)!

Input subclass: #BytesInput!
! BytesInput class methodsFor: #instantiation !
bytes: bytes pos: pos len: len
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.IO.BytesInput.new(bytes:handle(),pos,len))'}! !

! BytesInput class methodsFor: #instantiation !
bytes: bytes
	" Returns an instance of the receiver. "

	^self bytes: bytes pos: 0 len: bytes length! !

BytesInput buildAccessors: #( position )!
BytesInput buildGetters: #( length )!

Output subclass: #BytesOutput!
! BytesOutput methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Sys.IO.BytesOutput.new()'}! !

BytesOutput buildGetters: #( length )!
BytesOutput buildLuaFunctions: #(
	#(getBytes Bytes)
)!

IO subclass: #IOFile!
! IO class methodsFor: #accessing !
file
	" Returns a wrapper for library "

	^IOFile!

! IOFile class methodsFor: #reading !
bytesAt: fileName
	" Returns the bytes read at fileName. "

	| result data |
	result  := self withU8Path: fileName do: [:path|
		self new getLocalBytes: path
	].
	result notNil ifTrue: [ ^result ].
	data := Lib project getBytes: (self pathToFile: fileName).
	^data notNil ifTrue: [ data ]! !

! IOFile class methodsFor: #reading !
textAt: fileName
	" Returns the text content at fileName. "

	| result data |
	result  := self withU8Path: fileName do: [:path|
		self new getLocalContent: path
	].
	result notNil ifTrue: [ ^result ].
	data := Lib project getText: (self pathToFile: fileName).
	^data notNil ifTrue: [ data ]! !

! IOFile class methodsFor: #reading !
byteArrayAt: fileName
	" Returns the ByteArray read at fileName. "

	^ByteArray fromBytes: (self bytesAt: fileName)! !

IOFile buildLuaFunctions: #(
	#(getContent #(#(path asLocalPath)))
	#(saveContent #(path content))
	#(getBytes #(#(path asLocalPath)) Bytes)
	#(saveBytes #(path _bytes))
	#(read #(#(path asLocalPath) binary) FileInput)
	#(write #(path binary) FileOutput)
	#(append #(path binary))
	#(copy #(#(path asLocalPath) dstPath))
)!

! IOFile methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Sys.IO.File'}! !

! IOFile methodsFor: #accessing !
getLocalBytes: path

	^Bytes @ (self handle #_getBytes: path)! !

! IOFile methodsFor: #accessing !
getLocalContent: path

	^self handle #_getContent: path! !

IOFile class instanceVariableNames: #nativeRead!
! IOFile class methodsFor: #startup !
launch
	" Private - The system has been launched and start execution.
	Add support for reading native contents.
	"

	super launch.
	nativeRead isNil ifFalse: [ ^self "already launched" ].
	nativeRead := #{read}.
	{' read = self:nativeReadBlock()'}! !

! IOFile class methodsFor: #private !
nativeReadBlock
	" Private - Returns a block to be evaluated when reading native data. "

	^[:fileName|
		(#(http https) includes: ((fileName upTo: $:) asLowercase))
		ifTrue: [ Http get: fileName ]
		ifFalse: [ IOFile textAt: fileName ]
	]! !
! IOFile class methodsFor: #private !
fileContentsAt: aString
	" Private - Returns the contents at aString (or nil). "

	nativeRead isNil ifTrue: [ self launch ].
	^# #read: aString! !
! String methodsFor: #reading !
fileContents
	" Return the contents of file stored at pathName given by the receiver.
	Return nil if the file do not exist.
	"

	^IOFile fileContentsAt: self! !


Input subclass: #FileInput!
FileInput buildLuaFunctions: #(
	#(seek #(position type))
	tell eof
)!

! IOFile class methodsFor: #constants !
seekBegin
	^{'Lib.Sys.IO.FileSeek.SeekBegin'}!
seekCur
	^{'Lib.Sys.IO.FileSeek.SeekCur'}!
seekEnd
	^{'Lib.Sys.IO.FileSeek.SeekEnd'}!
begin
	^self seekBegin!
current
	^self seekCur!
end
	^self seekEnd!
seekType: selector
	selector == #begin ifTrue: [ ^self seekBegin ].
	selector == #current ifTrue: [ ^self seekCur ].
	selector == #end ifTrue: [ ^self seekEnd ].
	selector == #cur ifTrue: [ ^self seekCur ].
	^self error: 'Unknown seek constant'! !

! FileInput methodsFor: #query !
atEnd
	^self eof! !
! FileInput methodsFor: #accessing !
position
	^self tell! !
! FileInput methodsFor: #seek !
seekTo: position type: selector
	^self seek: position type: (IOFile seekType: selector)! !

Output subclass: #FileOutput!
FileOutput buildLuaFunctions: #(
	#(seek #(position type))
	tell
)!

! FileOutput methodsFor: #accessing !
position
	^self tell! !
! FileOutput methodsFor: #seek !
seekTo: position type: selector
	^self seek: position type: (IOFile seekType: selector)! !

Sys subclass: #Date!
! Sys class methodsFor: #accessing !
date
	" Returns a wrapper for library "

	^Date! !

! Date class methodsFor: #instantiation !
year: year month: month day: day hour: hour min: min sec: sec
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.Date.new(year, month, day, hour, min, sec))'}! !

! Date class methodsFor: #instantiation !
now
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.Date.now())'}! !

! Date class methodsFor: #instantiation !
fromTime: aFloat
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.Date.fromTime(aFloat))'}! !

! Date class methodsFor: #instantiation !
fromString: aString
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.Date.fromString(aString))'}! !

Date buildLuaFunctions: #(
	getTime
	getHours
	getMinutes
	getSeconds
	getFullYear
	getMonth
	getDate
	getDay
	toString
)!

! Date class methodsFor: #accessing !
tools
	" Returns the tools library. "

	^Sys dateTools! !

! Date methodsFor: #tools !
formatTo: format
	" Returns the receiver converted to String with specified format. "

	^self class tools
		format: self
		to: format! !

! Date methodsFor: #tools !
monthDays
	" Returns the number of days in the month of the receiver. "

	^self class tools getMonthDays: self handle! !

Sys subclass: #DateTools!
! Sys class methodsFor: #accessing !
dateTools
	" Returns a wrapper for library "

	^DateTools! !

! DateTools methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Sys.DateTools'}! !

DateTools buildLuaFunctions: #(
	#(format #(_aDate to))
	#(getMothDays #(_aDate))
)!

Sys subclass: #List!
! Sys class methodsFor: #accessing !
list
	" Returns a wrapper for library "

	^List! !

! List methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Sys.List.new()'}! !

List buildGetters: #( length )!
List buildLuaFunctions: #(
	#(add #(item))
	clear
	#(filter #(aBlock))
	first isEmpty
	#(iterator Iterator)
	#(join #(separator))
	last
	#(map #(aBlock) List)
	pop
	#(push #(item))
	#(remove #(v))
	toString
)!

! List methodsFor: #iterating !
do: aBlock
	" Evaluate aBlock with the contents of the receiver. "

	^self iterator do: aBlock!

! List methodsFor: #converting !
asArray
	" Return an array with the contents of the receiver. "

	^self iterator asArray! !

! List methodsFor: #accessing !
size
	^self length! !

! List methodsFor: #removing !
remove: anElement ifAbsent: aBlock
	| found |
	found := self remove: anElement.
	found = false ifTrue: [ ^aBlock value ].
	^anElement! !

! Sys class methodsFor: #accessing !
json
	" Returns a wrapper for library "

	^Json! !

! Json methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Sys.Json'}! !

Json buildLuaFunctions: #(
	#(parse #(aString) Lua)
	#(stringify #(_anObject replacer space))
)!

! Json class methodsFor: #converting !
parse: aString

	^self new parse: aString! !

! Json class methodsFor: #converting !
stringify: anObject

	^self new stringify: anObject replacer: nil space: ' '! !

! Json class methodsFor: #converting !
stringify: anObject replacer: replacer

	^self new stringify: anObject replacer: replacer space: ' '! !

! Json class methodsFor: #converting !
stringify: anObject replacer: replacer space: space

	^self new stringify: anObject replacer: replacer space: space! !


! Sys class methodsFor: #accessing !
eReg
	" Returns a wrapper for library "

	^EReg! !

! EReg class methodsFor: #instantiation !
with: pattern option: option 
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.EReg.new(pattern,option))'}! !

! EReg class methodsFor: #instantiation !
with: pattern
	" Returns an instance of the receiver. "

	^self with: pattern option: ''! !

EReg buildLuaFunctions: #(
	#(match #(s))
	#(matched #(n))
	matchedLeft
	matchedRight
	matchedPos
	#(matchSub #(s pos len))
	#(split #(s))
	#(replace #(s by))
	#(map #(s f))
)!

FileSystem buildLuaFunctions: #(
	#(exists #(path))
	#(rename #(path newPath))
	#(stat #(path))
	#(fullPath #(path))
	#(absolutePath #(path))
	#(isDirectory #(path))
	#(createDirectory #(path))
	#(deleteFile #(path))
	#(deleteDirectory #(path))
	#(readDirectory #(path) Array)
)!

! FileSystem methodsFor: #files !
prune: pathName
	" Delete a file at pathName or all the files and folders in directory at pathName. "

	| toDo |
	pathName isNil ifTrue: [ ^false ].
	pathName = $/ ifTrue: [ ^self error: 'We do not remove root folder' ].
	pathName last = $/ ifTrue: [ ^self prune: (pathName upToLast: $/) ].
	(self exists: pathName) ifFalse: [ ^false ].
	toDo := Array with: pathName.
	[toDo notEmpty] whileTrue: [ | current |
		current := toDo removeLast.
		(self isDirectory: current) ifFalse: [ self deleteFile: current ]
		ifTrue: [ | files path folders |
			files := self readDirectory: current.
			folders := #().
			files do: [:each|
				path := current ,$/ ,each.
				(self isDirectory: path)
					ifTrue: [ folders add: path ]
					ifFalse: [ self deleteFile: path ]
			].
			folders isEmpty
				ifTrue: [ self deleteDirectory: current ]
				ifFalse: [ toDo add: current; addAll: folders ]
		]
	].
	^true! !

| modules |
modules := #( Host Http Socket ).
Net modules: modules.
modules remove: #Host.
modules do: [:each|
	(Smalltalk at: each)
		implement: #initializeHandle
		comment: 'Private - Initialize the handle of the receiver.'
		as: 'handle := #{Lib}[#Sys][#Net][#' ,each ,'] #_new'
		category: #initialize.
]!

Socket subclass: #SSL !
! SSL class methodsFor: #select !
select: read write: write others: others timeout: timeout
	" Wait until one of the sockets groups is ready for the given operation. "

	^{'Lib.Sys.SSL.Socket.select(read:handles(), write:handles(), others:handles(), timeout)'}! !

! SSL methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Sys.SSL.Socket.new()'}! !

! Host class methodsFor: #instantiation !
@ ipHostName
	" Returns an instance of the receiver "

	#string = (self typeOf: ipHostName) ifTrue: [
		^self new: ipHostName
	].
	^super @ ipHostName! !

! Host class methodsFor: #instantiation !
new: ipHostName
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Sys.Net.Host.new(ipHostName))'}! !

! Host class methodsFor: #private !
dataToString: data
	" Returns the socket data as a string. "

	| array |
	array := ((data copyFrom: 2 to: data size - 1)
		asArrayOfSubstringsSeparatedBy: $,)
		collect: [:c| c asNumber ].
	^String fromCharCode: array! ! 

Host buildAccessors: #( host ip )!
Host buildLuaFunctions: #( toString reverse "localhost" )!
Host buildLuaStaticFunctions: #(
	reverse
	#(localhost #() Host)
)!

! Http class methodsFor: #instantiation !
@ url
	" Returns an instance of the receiver "

	#string = (self typeOf: url) ifTrue: [ ^self new: url ].
	^super @ url! !

! Http class methodsFor: #instantiation !
new: url
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Sys.Net.Http.new(url))'}! !

! Http class methodsFor: #instantiation !
requestUrl: url
	" Make http request with url and return result as string. "

	^{'Lib.Sys.Net.Http.requestUrl(url)'}! !

! Http class methodsFor: #instantiation !
get: url
	" Make http request with url and return result as string. "

	^self requestUrl: url! !

! Http class methodsFor: #request !
get: url onData: dataBlock onBytes: bytesBlock onError: errorBlock
	" Request a Get operation at url evaluating blocks. "

	| instance |
	instance := self @ url.
	dataBlock notNil ifTrue: [ instance onData: dataBlock ].
	errorBlock notNil ifTrue: [ instance onError: errorBlock ].
	bytesBlock notNil ifTrue: [ instance onBytesData: bytesBlock ].
	^instance request: false! !

! Http class methodsFor: #request !
get: url onBytes: bytesBlock onError: errorBlock
	" Request a Get operation at url evaluating blocks. "

	^self	get: url
		onData: nil onBytes: bytesBlock
		onError: errorBlock! !

! Http class methodsFor: #request !
get: url onData: dataBlock onError: errorBlock
	" Request a Get operation at url evaluating blocks. "

	^self	get: url
		onData: dataBlock onBytes: nil
		onError: errorBlock! !

! Http methodsFor: #configuration !
proxy: settings
	" Set the proxy settings. "

	^{'Lib.Sys.Net.Http.PROXY(settings)'}! !

Http buildAccessors: #(
	cnxTimeout
	noShutdown
	responseData
	url
)!
Http buildSetters: #(
	onData
	onBytesData
	onError
	onStatus
)!
Http buildTypedAccessors: #(
	#(responseHeaders Map)
)!
Http buildLuaFunctions: #(
	#(addHeader #(header value))
	#(addParameter #(param value))
	#(customRequest #(post _api _sock method))
	#(fileTransfer #(argname filename _file size mimeType))
	#(request #(isPost))
	#(setHeader #(header value))
	#(setParameter #(param value))
	#(setPostData #(data))
)!

! Socket class methodsFor: #select !
select: read write: write others: others timeout: timeout
	" Wait until one of the sockets groups is ready for the given operation. "

	^{'Lib.Sys.Net.Socket.select(read:handles(), write:handles(), others:handles(), timeout)'}! !

! Socket methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Sys.Net.Socket.new()'}! !

Socket buildAccessors: #(
	custom
)!
Socket buildTypedGetters: #(
	#(input Input)
	#(output Output)
)!
Socket buildLuaFunctions: #(
	#(accept Socket)
	#(bind #(_host port))
	close
	#(connect #(_host port))
	#(host Host)
	#(listen #(connections))
	#(peer Host)
	read
	#(setBlocking #(aBoolean))
	#(setFastSend #(aBoolean))
	#(setTimeout #(aFloat))
	#(shutdown #(read write))
	waitForRead
	#(write #(content))
)!

! Socket methodsFor: #nonBlocking !
settimeout: aFloat

	^self setTimeout: aFloat! !

! Socket methodsFor: #nonBlocking !
acceptOrNil
	" Returns an accepted client socket (or nil). "

	^self class @ (self send: #accept ifFailed: [])! !

! Utf8 class methodsFor: #instantiation !
new: size
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.Utf8.new(size))'}! !

! Utf8 class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Sys.Utf8)'} ! !

Utf8 buildLuaFunctions: #(
	#(addChar #(c))
	toString
)!
Utf8 buildLuaStaticFunctions: #(
	#(encode #(s))
	#(decode #(s))
	#(iter #(s doing))
	#(charCodeAt #(s index))
	#(validate #(s))
	#(length #(s))
	#(compare #(a with))
	#(sub #(s pos len))
)!

VM subclass: #Thread!
! VM class methodsFor: #accessing !
thread
	" Returns a wrapper for library "

	^Thread! !

! Thread class methodsFor: #instantiation !
new: scriptOrPathName
	" Returns an instance of the receiver. "

	^self create: scriptOrPathName! !

! Thread class methodsFor: #instantiation !
@ data
	" Returns an instance of the receiver. "

	self withString: data do: [ ^self create: data ].
	^super @ data! !

! Thread class methodsFor: #instantiation !
create: scriptOrPathName
	" Returns an instance of the receiver.
	The instance will execute the script as independent instance, then exit.
	"

	^{'self:_at(Lib.Sys.VM.Thread.create(scriptOrPathName))'}! !

! Thread class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Sys.VM.Thread)'}! !

Thread buildLuaFunctions: #(
	#(sendMessage #(_msg))
)!
Thread buildLuaStaticFunctions: #(
	#(current Thread)
	#(readMessage #(aBoolean))
)!

! Xml class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Sys.Xml)'}! !

! Xml class methodsFor: #constants !
type: selector
	" Returns the value of the constant. "

	^(self nativeClass handle basicAt: #XmlType)
		basicAt: selector
		ifAbsent: [ self error: 'Invalid Xml type name' ]! !

! Xml class methodsFor: #instantiation !
pathName: fileName
	" Returns an instance of the receiver. "

	| result data |
	result  := self withU8Path: fileName do: [:path|
		IOFile new getContent: path
	].
	result notNil ifTrue: [ ^self parse: result ].
	data := Lib project getText: (self pathToFile: fileName).
	^data notNil ifTrue: [ self parse: data ]! !

Xml buildLuaStaticFunctions: #(
	#(createCData #(data) Xml)
	#(createComment #(data) Xml)
	#(createDocType #(data) Xml)
	#(createDocument Xml)
	#(createElement #(name) Xml)
	#(createPCData #(data) Xml)
	#(createProcessingInstruction #(data) Xml)
	#(parse #(string) Xml)
)!
Xml	buildAccessors: #(nodeName nodeValue);
	buildGetters: #(nodeType);
	buildTypedGetters: #( #(parent Xml) )!
Xml buildLuaFunctions: #(
	#(addChild #(_x))
	#(attributes Iterator)
	#(elements Iterator)
	#(elementsNamed #(name) Iterator)
	#(exists #(att))
	#(firstChild Xml)
	#(firstElement Xml)
	#(get #(att) Xml)
	#(insertChild #(_x pos))
	#(iterator Iterator)
	#(remove #(att))
	#(removeChild #(_x))
	#(set #(att value))
	toString
)!

! Xml methodsFor: #accessing !
xmlElements
	" Returns anArray with the elements of the receiver as Xml instances. "

	^self elements asArrayOf: Xml! !

! Xml methodsFor: #iterating !
do: aBlock
	" Evaluates aBlock with the elements of the receiver. "

	self elements do: [:each| aBlock value: Xml @ each handle ]! !

Zip subclass: #Compress!
Zip subclass: #Uncompress!
Zip subclass: #ZipReader!
Zip subclass: #ZipWriter!
Zip subclass: #ZipEntry!
! Zip class methodsFor: #accessing !
compress
	" Returns a wrapper for library "

	^Compress! !

! Zip class methodsFor: #util !
compress: bytes level: level
	" Compress bytes with level of compression.
	Return compresed Bytes object. "

	^self compress run: bytes level: level! !

! Zip class methodsFor: #accessing !
uncompress
	" Returns a wrapper for library "

	^Uncompress! !

! Zip class methodsFor: #util !
uncompress: bytes size: bufsize
	" Uncompress bytes with bufsize.
	Return uncompresed Bytes object. "

	^self uncompress run: bytes bufsize: bufsize! !

! Zip class methodsFor: #util !
uncompress: bytes
	" Uncompress bytes.
	Return uncompresed Bytes object. "

	^self uncompress run: bytes bufsize: nil! !

! Zip class methodsFor: #accessing !
entriesIn: zipPath
	" Returns the entries in zip file zipPath "

	| bytes input result |
	bytes := IOFile new getBytes: zipPath.
	input := BytesInput bytes: bytes.
	result := ZipReader readZip: input.
	input close.
	^result! !

! Zip class methodsFor: #util !
unzip: zipPath to: destination
	" Decompress the contents in zip file at zipPath to destination folder. "

	(Sys fileSystem exists: destination) ifFalse: [
		^self error: 'Destination folder missing ' ,destination
	].

	(Zip entriesIn: zipPath) do: [:each| | fileName file path |
		fileName := each fileName replace: $\ with: $/.
		(fileName startsWith: '..') ifFalse: [ | target dataBytes |
			(fileName includes: $/) ifFalse: [
				file := fileName.
				path := ''.
			] ifTrue: [
				file := fileName fromLast: $/.
				path := fileName upToLast: $/.
			].
			target := destination ,$/ ,path.
			target last = $/ ifFalse: [ target := target ,$/ ].
			(Sys fileSystem exists: target) ifFalse: [ Sys fileSystem createDirectory: target ].
			dataBytes := ZipReader unzip: each.
"# #print: target,file with: dataBytes length with: #bytes."
			dataBytes length > 0 ifTrue: [
				IOFile new
					saveBytes: target ,file
					bytes: (ByteArray fromBytes: dataBytes)
			].
		].
	]! !

! Compress class methodsFor: #instantiation !
run: bytes level: level
	" Returns an instance of the receiver. "

	^{'smalltalk.Bytes:_at(Lib.Sys.Zip.Compress.run(bytes:handle(), level))'}! !

! Uncompress class methodsFor: #instantiation !
run: bytes bufsize: bufsize
	" Returns an instance of the receiver. "

	^{'smalltalk.Bytes:_at(Lib.Sys.Zip.Uncompress.run(bytes:handle(), bufsize))'}! !

! Zip class methodsFor: #accessing !
reader
	" Returns a wrapper for library "

	^ZipReader! !

! Zip class methodsFor: #accessing !
writer
	" Returns a wrapper for library "

	^ZipWriter! !

! ZipReader class methodsFor: #instantiation !
new: input
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.Zip.Reader.new(input:handle()))'}! !

! ZipReader class methodsFor: #instantiation !
on: input
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.Zip.Reader.new(input:handle()))'}! !

! ZipReader class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Sys.Zip.Reader)'}! !

ZipReader buildLuaStaticFunctions: #(
	#(readZip #(_input) 'Array[ZipEntry]')
	#(unzip #(_entry) Bytes)
)!
ZipReader buildLuaFunctions: #(
	#(read 'List[ZipEntry]')
	#(readEntryHeader ZipEntry)
)!

! ZipWriter class methodsFor: #instantiation !
new: output
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.Zip.Writer.new(output:handle()))'}! !

! ZipWriter class methodsFor: #instantiation !
on: output
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Sys.Zip.Writer.new(output:handle()))'}! !

ZipWriter buildLuaFunctions: #(
	#(write #(#(files handles)))
	writeCDR
	#(writeEntryHeader #(_f))
)!

ZipEntry buildAccessors: #(
	fileName fileSize
	compressed dataSize
	crc32
);	buildTypedAccessors: #(
	#(fileTime Date)
	#(data Bytes)
	#(extraFields List)
)!

LuaStudioLibrary subclass: #Request!
LuaStudioLibrary subclass: #Response!
! Web methodsFor: #accessing !
request
	" Returns a wrapper for library "

	^Request! !

! Web methodsFor: #accessing !
response
	" Returns a wrapper for library "

	^Response! !

! Request methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Web.Request'}! !

Request buildAccessors: #(
	queryString form headers
	requestType url path
)!

! Response methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Web.Response'}! !

Response buildLuaFunctions: #(
	#(addHeader #(name value))
	#(writeString #(aString))
	#(redirect #(url))
)!

LuaStudioLibrary subclass: #Map!
Map buildLuaFunctions: #(
	#(set #(key value))
	#(exists #(key))
	#(remove #(key))
	#(keys Iterator)
	#(iterator Iterator)
	toString
)!

LuaStudioLibrary subclass: #Iterator!
Iterator buildLuaFunctions: #(
	hasNext
	#(next #() Lua)
)!

! Iterator methodsFor: #iterating !
do: aBlock
	" Evaluate aBlock with the contents of the receiver. "

	[self hasNext] whileTrue: [ aBlock value: self next ]!

! Iterator methodsFor: #iterating !
with: aNativeClass do: aBlock
	" Evaluate aBlock with the contents of the receiver. "

	[self hasNext] whileTrue: [
		aBlock value: (aNativeClass new
			handle: self next handle)
	]!

! Iterator methodsFor: #accessing !
next: aNativeClass
	" Returns the next element of the receiver as an instance of NativeClass. "

	| result |
	result := self next.
	^result notNil ifTrue: [ aNativeClass new handle: result handle ]! !

! Iterator methodsFor: #converting !
asArray
	" Return an array with the contents of the receiver. "

	| result |
	result := Array new.
	self do: [:each| result add: each ].
	^result! !

! Iterator methodsFor: #converting !
asArrayOf: aNativeClass
	" Return an array with the contents of the receiver. "

	| result |
	result := Array new.
	self with: aNativeClass do: [:each| result add: each ].
	^result! !

LuaStudioLibrary subclass: #SqliteConnection!
LuaStudioLibrary subclass: #SqliteBlob!
Iterator subclass: #SqliteResultSet!

! SQLight class methodsFor: #accessing!
connection
	" Returns the connection wrapper. "

	^SqliteConnection! !

! SqliteConnection class methodsFor: #instantiation !
new: fileLocation
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.SQLight.Connection.new(fileLocation))'}! !

SqliteConnection buildLuaFunctions: #(
	#(blobOpen #(tableName fieldName row) SqliteBlob)
	#(request #(query parameters) SqliteResultSet)
	#(request #(query with) SqliteResultSet)
	close
)!

! SqliteConnection methodsFor: #request !
request: query
	" Returns the result of request without parameters. "

	^self request: query parameters: nil! !

SqliteBlob buildLuaFunctions: #(
	size close
	#(write #(_data size dataPos blobPos))
	#(read #(_data size dataPos blobPos))
)!

LuaStudioLibrary subclass: #OAuthToken!
OAuth buildConstants: '
	Lib.OAuth.OAuthVersion.V1
	Lib.OAuth.OAuthVersion.V2
'!

LuaStudioLibrary subclass: #Client!
! Client class methodsFor: #instantiation !
version: version consumer: consumer
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.OAuth.Client.new(version,consumer:handle()))'}! !

Client buildAccessors: #(
	version
)!
Client buildTypedAccessors: #(
	#(consumer Consumer)
	#(accessToken OAuthToken)
	#(refreshToken RefreshToken)
)!
Client buildLuaFunctions: #(
	#(getRequestToken #(uri callback post) RefreshToken)
	#(getAccessToken1 #(uri verifier post) Client)
	#(getAccessToken2 #(uri code redirectUrl post) Client)
	#(refreshAccessToken #(uri post) Client)
	#(requestUrlEncoded #(uri post postData))
	#(requestJSON #(uri post postData) Json)
	#(request #(uri post postData))
)!

LuaStudioLibrary subclass: #Consumer!
! Consumer class methodsFor: #instantiation !
key: key secret: secret
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.OAuth.Consumer.new(key,secret))'}! !

Consumer buildGetters: #(
	key secret
)!

LuaStudioLibrary subclass: #OAuth1!
! OAuth1 class methodsFor: #instantiation !
consumer: consumer accessToken: accessToken
	" Returns a Client. "

	^{'smalltalk.Client:_at(Lib.OAuth.OAuth1.connect(consumer:handle(),accessToken:handle()))'}! !

! OAuth1 class methodsFor: #accessing !
authUrl: baseUri requestToken: requestToken
	" Returns authentication url from baseUri and requestToken strings. "

	^{'Lib.OAuth.OAuth1.buildAuthUrl(baseUri, requestToken)'}! !

LuaStudioLibrary subclass: #OAuth2!
! OAuth2 class methodsFor: #constants !
nonce
	" Returns the nonce string. "

	^{'Lib.OAuth.OAuth2.nonce()'}! !

! OAuth2 class methodsFor: #instantiation !
consumer: consumer accessToken: accessToken refreshToken: refreshToken
	" Returns a Client. "

	^{'smalltalk.Client:_at(Lib.OAuth.OAuth2.connect(consumer:handle(),accessToken:handle(),refreshToken:handle()))'}! !

! OAuth2 class methodsFor: #accessing !
authUrl: baseUri clientId: clientId opts: opts additionalParamsObject: additionalParamsObject
	" Returns authentication url from baseUri, clientId, opts object and additionalParamsObject parameters. "

	^{'Lib.OAuth.OAuth2.buildAuthUrl(baseUri, clientId, opts:handle(), additionalParamsObject:handle())'}! !

OAuthToken subclass: #OAuth1AccessToken!
! OAuth1AccessToken class methodsFor: #instantiation !
token: token secret: secret
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.OAuth.Tokens.OAuth1AccessToken.new(token, secret))'}! !

OAuthToken subclass: #OAuth2AccessToken!
! OAuth2AccessToken class methodsFor: #instantiation !
token: token expires: expires
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.OAuth.Tokens.OAuth2AccessToken.new(token, expires))'}! !

OAuthToken subclass: #RefreshToken!
! RefreshToken class methodsFor: #instantiation !
token: token
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.OAuth.Tokens.RefreshToken.new(token))'}! !

OAuthToken subclass: #RequestToken!
! RequestToken class methodsFor: #instantiation !
token: token secret: secret
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.OAuth.Tokens.RequestToken.new(token, secret))'}! !

RequestToken buildGetters: #( token secret )!

Media modules: #( Audio Events Graphics )!
Audio modules: #( SoundTransform )!
Events modules: #( Event )!
EventDispatcher subclass: #SoundChannel!
EventDispatcher subclass: #Camera!
EventDispatcher subclass: #Sound!

! Sound class methodsFor: #instantiation !
stream: stream context: context forcePlayAsMusic: forcePlayAsMusic inEngine: inEngine
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Audio.Sound.new(stream:handle(), context:handle(), forcePlayAsMusic, inEngine))'}! !

! Sound class methodsFor: #instantiation !
music
	" Returns an instance of the receiver. "

	^self stream: nil context: nil forcePlayAsMusic: true inEngine: nil! !

! Sound class methodsFor: #instantiation !
@ data
	" Returns an instance of the receiver. "

	self withString: data do: [ ^self pathName: data ].
	^super @ data! !

! Sound class methodsFor: #instantiation !
pathName: fileName
	" Returns an instance of the receiver. "

	^self pathName: fileName forcePlayAsMusic: true inEngine: nil! !

! Sound class methodsFor: #instantiation !
pathName: fileName forcePlayAsMusic: forcePlayAsMusic inEngine: inEngine
	" Returns an instance of the receiver. "

	^self new
		load: (IOFile byteArrayAt: fileName)
		forcePlayAsMusic: forcePlayAsMusic
		inEngine: inEngine! !

! Sound methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Audio.Sound.new(nil,nil,false,nil)'}! !

Sound buildAccessors: #(
	bytesLoaded
	bytesTotal
	id3
	isBuffering
	length
	url
)!
Sound buildLuaFunctions: #(
	getEngine
	close
	#(load #(_stream _context forcePlayAsMusic inEngine))
	#(loadCompressedDataFromByteArray #(_bytes length forcePlayAsMusic inEngine))
	#(loadPCMFromByteArray #(_bytes samples format stereo sampleRate inEngine))
	#(play #(startTime loops _sndTransform) SoundChannel)
)!

! Sound methodsFor: #loading !
load: data forcePlayAsMusic: forcePlayAsMusic inEngine: inEngine
	" Load the data onto the receiver. "

	self	loadCompressedDataFromByteArray: data
		length: data size forcePlayAsMusic: forcePlayAsMusic
		inEngine: inEngine! !

! SoundTransform class methodsFor: #instantiation !
volume: volume pan: pan
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Audio.SoundTransform.new(volume, pan))'}! !

SoundTransform buildAccessors: #( pan volume )!

SoundChannel buildGetters: #(
	leftPeak rightPeak position
)!
SoundChannel buildTypedGetters: #(
	soundTransform
)!
SoundChannel buildLuaFunctions: #( stop )!

! Camera class methodsFor: #instantiation !
named: aName
	" Returns an instance of the receiver. "

	^self getCamera: name! !

! Camera class methodsFor: #instantiation !
getCamera: name
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Camera.getCamera(name))'}! !

! Camera class methodsFor: #instantiation !
@ data
	" Returns an instance of the receiver. "

	self withString: data do: [ ^self getCamera: data ].
	^super @ data! !

! Camera class methodsFor: #instantiation !
default
	" Returns an instance of the receiver. "

	^self getCamera: nil! !

! Camera class methodsFor: #accessing !
names
	" Returns an Array of strings containing the names of all available cameras. "

	| names |
	{' names = Lib.Media.Camera.names '}.
	^Array @ names! !

! Camera methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Camera.getCamera(nil)'}! !

! Camera methodsFor: #query !
isValid
	" Returns true if the receiver is valid. "

	^handle notNil and: [ self width > 0 ]! !

Camera buildTypedAccessors: #(
	#(bitmapData BitmapData)
)!
Camera buildGetters: #(
	name position
	width height
)!
Camera buildLuaFunctions: #(
	activate
	destroy
)!

! Camera methodsFor: #accessing !
extent
	" Returns the extent of the receiver. "

	^self width @ self height! !

Graphics buildLuaFunctions: #(
	#(lineStyle #(thickness color alpha pixelHinting scaleMode caps joints miterLimit))
	#(beginFill #(color alpha))
	#(drawRect #(x y width height))
	endFill
	#(beginGradientFill #(type colors alphas ratios _matrix spreadMethod interpolationMethod focalPointRatio))
	#(drawCircle #(x y radius))
	#(moveTo #(x y))
	#(lineTo #(x y))
	clear
	#(beginBitmapFill #(_bitmap _matrix repeat smooth))
	#(drawTiles #(_sheet xyid smooth flags count))
	#(curveTo #(controlX controlY anchorX anchorY))
	#(drawEllipse #(x y width height))
	#(drawGraphicsData #(_graphicsData))
	#(drawPath #(commands data winding))
	#(drawRoundRect #(x y width height ellipseWidth ellipseHeight))
	#(drawTriangles #(vertices indices uvtData culling colours blendMode))
	#(lineBitmapStyle #(_bitmap _matrix repeat smooth))
	#(lineGradientStyle #(type colors alphas ratios _matrix spreadMethod interpolationMethod focalPointRatio))
	#(lineStyle #(thickness color alpha pixelHinting scaleMode caps joints miterLimit))
)!

Media modules: #( CameraPosition Capabilities )!
CameraPosition buildConstants: '
	Lib.Media.CameraPosition.BACK
	Lib.Media.CameraPosition.FRONT
	Lib.Media.CameraPosition.UNKNOWN
'!
Capabilities buildConstants: '
	Lib.Media.Capabilities.pixelAspectRatio
	Lib.Media.Capabilities.screenDPI
	Lib.Media.Capabilities.screenResolutionX
	Lib.Media.Capabilities.screenResolutionY
	Lib.Media.Capabilities.language
'!	"Lib.Media.Capabilities.screenResolutions"

! Capabilities class methodsFor: #accessing !
screenResolution
	" Returns the maximum screen coordinate (in pixels). "

	^self screenResolutionX @ self screenResolutionY! !

! Capabilities class methodsFor: #accessing !
screenResolutions
	" Returns the supported resolutions on device. "

	^Array @ ({'Lib.Media.Capabilities.screenResolutions'})
		collect: [:tuple| Array @ tuple ]! !

Media modules: #( Clipboard )!

! Clipboard methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Clipboard.new()'}! !

! Clipboard class methodsFor: #accessing !
generalClipboard
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Clipboard.generalClipboard)'}! !

Clipboard buildTypedGetters: #(
	#(formats Array)
)!
Clipboard buildLuaFunctions: #(
	clear
	#(clearData #(format))
	#(getData #(format transferMode))
	#(hasFormat #(format))
	#(setData #(format _data serializable))
	#(setDataHandler #(format handler serializable))
)!

Media modules: #( ClipboardFormats )!
ClipboardFormats buildConstants: '
	Lib.Media.ClipboardFormats.AIR_PREFIX
	Lib.Media.ClipboardFormats.BITMAP_FORMAT
	Lib.Media.ClipboardFormats.FILE_LIST_FORMAT
	Lib.Media.ClipboardFormats.FILE_PROMISE_LIST_FORMAT
	Lib.Media.ClipboardFormats.FLASH_PREFIX
	Lib.Media.ClipboardFormats.HTML_FORMAT
	Lib.Media.ClipboardFormats.REFERENCE_PREFIX
	Lib.Media.ClipboardFormats.RICH_TEXT_FORMAT
	Lib.Media.ClipboardFormats.SERIALIZATION_PREFIX
	Lib.Media.ClipboardFormats.TEXT_FORMAT
	Lib.Media.ClipboardFormats.URL_FORMAT
'!

EventDispatcher buildLuaFunctions: #(
	#(addEventListener #(type doing useCapture priority useWeakReference))
	#(dispatchEvent #(_event))
	#(hasEventListener #(type))
	#(removeEventListener #(type doing useCapture))
	#(removeEventListener #(type doing capture)) "OBSOLETE"
	toString
	#(willTrigger #(type))
)!

! EventDispatcher methodsFor: #events !
addEventListener: type doing: aBlock useCapture: capture priority: priority useWeak: useWeak
	self addEventListener: type doing: aBlock useCapture: capture priority: priority useWeakReference: useWeak!
addEventListener: type doing: aBlock useCapture: capture priority: priority
	self addEventListener: type doing: aBlock useCapture: capture priority: priority useWeakReference: false!
addEventListener: type doing: aBlock useCapture: capture
	self addEventListener: type doing: aBlock useCapture: capture priority: 0 useWeakReference: false!
addEventListener: type doing: aBlock
	self addEventListener: type doing: aBlock useCapture: false priority: 0 useWeakReference: false!
addEventListener: type with: aBlock
	self addEventListener: type doing: aBlock useCapture: false priority: 0 useWeakReference: false!
removeEventListener: type doing: aBlock
	self removeEventListener: type doing: aBlock useCapture: false! !

EventDispatcher subclass: #DisplayObject!
DisplayObject buildAccessors: #(
	alpha blendMode cacheAsBitmap
	pedanticBitmapCaching pixelSnapping
	height name opaqueBackground
	rotation scaleX scaleY visible width
	x y
)!
DisplayObject buildGetters: #(
	mouseX mouseY
)!
DisplayObject buildTypedAccessors: #(
	#(graphics Graphics)
	#(loaderInfo LoaderInfo)
	#(transform Transform)
	#(parent DisplayObjectContainer)
	#(mask DisplayObject)
	#(scrollRect ARectangle)
	#(scale9Grid Lua)
)!
DisplayObject buildSetters: #(
	rotationX rotationY
)!
DisplayObject buildLuaFunctions: #(
	#(getBounds #(_targetCoordinateSpace) ARectangle)
	#(getRect #(_targetCoordinateSpace) ARectangle)
	#(globalToLocal #(_inGlobal) APoint)
	#(hitTestObject #(_object))
	#(hitTestPoint #(x y shapeFlag))
	#(localToGlobal #(_inLocal) APoint)
)!

! DisplayObject methodsFor: #accessors !
position
	^self x @ self y! !

! DisplayObject methodsFor: #accessors !
position: aPoint
	| p |
	p := aPoint asPoint.
	self x: p x; y: p y.! !

! DisplayObject methodsFor: #accessors !
extent
	^self width @ self height! !

! DisplayObject methodsFor: #accessors !
extent: aPoint
	| p |
	p := aPoint asPoint.
	self width: p x; height: p y.! !

! DisplayObject methodsFor: #accessors !
scale
	^self scaleX @ self scaleY! !

! DisplayObject methodsFor: #accessors !
scale: aPoint
	| p |
	p := aPoint asPoint.
	self scaleX: p x; scaleY: p y.! !

! DisplayObject methodsFor: #accessors !
mouse
	^self mouseX @ self mouseY! !

! DisplayObject methodsFor: #accessors !
filters
	^self handle basicAt: #filters! !

! DisplayObject methodsFor: #accessors !
filters: anArray
	^self handle basicAt: #filters put: anArray handles! !

DisplayObject subclass: #Display!
DisplayObject subclass: #Bitmap!
Display modules: #( BitmapData )!
! Media class methodsFor: #accessing !
display
	" Returns a wrapper for library "

	^Display! !
! Display class methodsFor: #accessing !
bitmap
	" Returns a wrapper for library "

	^Bitmap! !

! Bitmap class methodsFor: #instantiation !
bitmapData: bitmapData pixelSnapping: pixelSnapping smoothing: smoothing
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Display.Bitmap.new(bitmapData:handle(), pixelSnapping, smoothing))'}! !

! Bitmap class methodsFor: #instantiation !
@ data
	" Returns an instance of the receiver. "

	self withString: data do: [ ^self pathName: data ].
	^super @ data! !

! Bitmap class methodsFor: #instantiation !
pathName: pathName
	" Returns an instance of the receiver. "

	^self	bitmapData: BitmapData @ pathName
		pixelSnapping: PixelSnapping auto
		smoothing: true! !

Bitmap buildAccessors: #( smoothing )!
Bitmap buildTypedAccessors: #(
	#(bitmapData BitmapData)
)!

! Bitmap methodsFor: #accessing !
extent
	" Returns the extent of the receiver. "

	^self bitmapData extent! !

! BitmapData class methodsFor: #instantiation !
inWidth: inWidth inHeight: inHeight inTransparent: inTransparent fillColor: fillColor inGPUMode: inGPUMode
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Display.BitmapData.new(inWidth, inHeight, inTransparent, fillColor, inGPUMode))'}! !

! BitmapData class methodsFor: #instantiation !
bytes: bytes rawAlpha: rawAlpha
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Display.BitmapData.loadFromBytes(bytes:handle(), rawAlpha:handle()))'}! !

! BitmapData class methodsFor: #instantiation !
@ data
	" Returns an instance of the receiver. "

	| bytes |
	(self isObject: data) ifFalse: [ ^super @ data ].
	self withString: data do: [ ^self pathName: data ].
	data size = 1 ifTrue: [ ^self @ (data copyWith: nil) ].
	data size = 2 ifFalse: [ ^self error: 'Invalid size' ].
	bytes := data collect: [:each|
		(self typeOf: each) = #string
		ifTrue: [ Lib project getBytes: each ]
		ifFalse: [ each ]
	].
	^self bytes: bytes first rawAlpha: bytes second! !

! BitmapData class methodsFor: #instantiation !
from: data
	" Returns an instance of the receiver "

	self withString: data do: [ ^self pathName: data ].
	^{'self:_at(Lib.Away3D.Utils.Cast.bitmapData(data:handle()))'}! !

! BitmapData class methodsFor: #converting !
pathName: fileName
	" Returns an instance of the receiver "

	| result pathName |
	result  := self withU8Path: fileName do: [:path|
		self bytes: (ByteArray fromBytes: (IOFile new getLocalBytes: path)) rawAlpha: nil
	].
	result notNil ifTrue: [ ^result ].
	pathName := self pathToFile: fileName.
	^{'self:_at(Lib.Away3D.Utils.Cast.bitmapData(pathName))'}! !

BitmapData buildGetters: #(
	height width transparent
)!
BitmapData buildTypedGetters: #(
	#(rect ARectangle)
)!
BitmapData buildLuaFunctions: #(
	#(applyFilter #(_sourceBitmapData _sourceRect _destPoint filter))
	clone
	#(colorTransform #(_rect _colorTransform))
	#(copyChannel #(_sourceBitmapData _sourceRect _destPoint inSourceChannel inDestChannel))
	#(copyPixels #(_sourceBitmapData _sourceRect _destPoint _alphaBitmapData _alphaPoint mergeAlpha))
	dispose
	#(draw #(_source _matrix _colorTransform _blendMode _clipRect smoothing))
	#(encode #(format quality))
	#(fillRect #(_rect colour))
	#(fillRectEx #(_rect colour alpha))
	#(getColorBoundsRect #(mask color findColor))
	#(getPixel #(x y))
	#(getPixel32 #(x y))
	#(getPixels #(_rect) ByteArray)
	#(getVector #(_rect) Array)
	#(noise #(randomSeed low high channelOptions grayScale))
	#(scroll #(x y))
	#(setPixel #(x y color))
	#(setPixel32 #(x y color))
	#(setPixels #(_rect _inputByteArray))
	#(setVector #(_rect inputVector))
	#(threshold #(_sourceBitmapData _sourceRect _destPoint operation threshold color mask copySource))
)!

! BitmapData methodsFor: #accessing !
extent
	" Returns the extent of the receiver. "

	^self width @ self height! !

Display modules: #( BitmapDataChannel BlendMode CapsStyle FPS )!
BitmapDataChannel buildConstants: '
	Lib.Media.Display.BitmapDataChannel.ALPHA
	Lib.Media.Display.BitmapDataChannel.BLUE
	Lib.Media.Display.BitmapDataChannel.GREEN
	Lib.Media.Display.BitmapDataChannel.RED
'!
BlendMode buildConstants: '
	Lib.Media.Display.BlendMode.ADD
	Lib.Media.Display.BlendMode.ALPHA
	Lib.Media.Display.BlendMode.DARKEN
	Lib.Media.Display.BlendMode.DIFFERENCE
	Lib.Media.Display.BlendMode.ERASE
	Lib.Media.Display.BlendMode.HARDLIGHT
	Lib.Media.Display.BlendMode.INVERT
	Lib.Media.Display.BlendMode.LAYER
	Lib.Media.Display.BlendMode.LIGHTEN
	Lib.Media.Display.BlendMode.MULTIPLY
	Lib.Media.Display.BlendMode.NORMAL
	Lib.Media.Display.BlendMode.OVERLAY
	Lib.Media.Display.BlendMode.SCREEN
	Lib.Media.Display.BlendMode.SHADER
	Lib.Media.Display.BlendMode.SUBTRACT
'!
CapsStyle buildConstants: '
	Lib.Media.Display.CapsStyle.NONE
	Lib.Media.Display.CapsStyle.ROUND
	Lib.Media.Display.CapsStyle.SQUARE
'!

! FPS methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Display.FPS.new()'}! !

FPS buildGetters: #( currentFPS )!

Display modules: #(
	GradientType GraphicsPathCommand
	GraphicsPathWinding JointStyle
	InterpolationMethod LineScaleMode
)!
GradientType buildConstants: '
	Lib.Media.Display.GradientType.LINEAR
	Lib.Media.Display.GradientType.RADIAL
'!
GraphicsPathCommand buildConstants: '
	Lib.Media.Display.GraphicsPathCommand.CUBIC_CURVE_TO
	Lib.Media.Display.GraphicsPathCommand.CURVE_TO
	Lib.Media.Display.GraphicsPathCommand.LINE_TO
	Lib.Media.Display.GraphicsPathCommand.MOVE_TO
	Lib.Media.Display.GraphicsPathCommand.NO_OP
	Lib.Media.Display.GraphicsPathCommand.WIDE_LINE_TO
	Lib.Media.Display.GraphicsPathCommand.WIDE_MOVE_TO
'!
GraphicsPathWinding buildConstants: '
	Lib.Media.Display.GraphicsPathWinding.EVEN_ODD
	Lib.Media.Display.GraphicsPathWinding.NON_ZERO
'!
JointStyle buildConstants: '
	Lib.Media.Display.JointStyle.BEVEL
	Lib.Media.Display.JointStyle.MITER
	Lib.Media.Display.JointStyle.ROUND
'!
InterpolationMethod buildConstants: '
	Lib.Media.Display.InterpolationMethod.LINEAR_RGB
	Lib.Media.Display.InterpolationMethod.RGB
'!
LineScaleMode buildConstants: '
	Lib.Media.Display.LineScaleMode.HORIZONTAL
	Lib.Media.Display.LineScaleMode.NONE
	Lib.Media.Display.LineScaleMode.NORMAL
	Lib.Media.Display.LineScaleMode.VERTICAL
'!

DisplayObject subclass: #InteractiveObject!
InteractiveObject buildAccessors: #(
	doubleClickEnabled
	mouseEnabled
	needsSoftKeyboard
	softKeyboard
)!
InteractiveObject buildLuaFunctions: #(
	requestSoftKeyboard
)!

InteractiveObject subclass: #DisplayObjectContainer!
DisplayObjectContainer buildAccessors: #(
	mouseChildren tabChildren
)!
DisplayObjectContainer buildGetters: #( numChildren )!
DisplayObjectContainer buildLuaFunctions: #(
	#(addChild #(_child))
	#(addChildAt #(_child index))
	#(areInaccessibleObjectsUnderPoint #(_point))
	#(contains #(_child))
	#(getChildAt #(index))
	#(getChildByName #(name))
	#(getChildIndex #(_child))
	#(getObjectsUnderPoint #(_point))
	#(removeChild #(_child))
	#(removeChildAt #(index))
	#(removeChildren #(beginIndex endIndex))
	#(setChildIndex #(_child index))
	#(swapChildren #(_child1 _child2))
	#(swapChildrenAt #(index1 index2))
)!

DisplayObjectContainer subclass: #Sprite!
! Display class methodsFor: #accessing !
sprite
	" Returns a wrapper for library "

	^Sprite! !
! Sprite methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Display.Sprite.new()'}! !

Sprite buildAccessors: #(
	buttonMode useHandCursor
)!
Sprite buildLuaFunctions: #(
	#(startDrag #(lockCenter _bounds))
	stopDrag
)!

Sprite subclass: #Loader!
! Display class methodsFor: #accessing !
loader
	" Returns a wrapper for library "

	^Loader! !

! Loader methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Display.Loader.new()'}! !

Loader buildTypedGetters: #(
	data
	#(content DisplayObject)
	#(contentLoaderInfo LoaderInfo)
)!
Loader buildLuaFunctions: #(
	#(load #(_request context))
	#(loadBytes #(_bytes context))
	unload
)!
EventDispatcher subclass: #LoaderInfo!
LoaderInfo buildGetters: #(
	childAllowsParent
	frameRate height width
	loaderURL parentAllowsChild
	sameDomain url
)!
LoaderInfo buildTypedGetters: #(
	#(bytes ByteArray)
	#(content DisplayObject)
	#(loader Loader)
	#(parameters PoolDictionary)
	#(sharedEvents EventDispatcher)
)!
LoaderInfo buildAccessors: #(
	contentType
)!
LoaderInfo buildLuaFunctions: #(
	#(load #(_request))
)!

Display modules: #(
	PixelSnapping TriangleCulling SpreadMethod
)!
PixelSnapping buildConstants: '
	Lib.Media.Display.PixelSnapping.ALWAYS
	Lib.Media.Display.PixelSnapping.AUTO
	Lib.Media.Display.PixelSnapping.NEVER
'!
Display buildConstants: '
	Lib.Media.Display.PAD
	Lib.Media.Display.REFLECT
	Lib.Media.Display.REPEAT
'!
TriangleCulling buildConstants: '
	Lib.Media.Display.TriangleCulling.NEGATIVE
	Lib.Media.Display.TriangleCulling.NONE
	Lib.Media.Display.TriangleCulling.POSITIVE
'!
SpreadMethod buildConstants: '
	Lib.Media.Display.SpreadMethod.PAD
	Lib.Media.Display.SpreadMethod.REFLECT
	Lib.Media.Display.SpreadMethod.REPEAT
'!

DisplayObject subclass: #Shape!
! Shape methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Display.Shape.new()'}! !

DisplayObjectContainer subclass: #Stage!
! Display class methodsFor: #accessing!
stage
	" Returns the object that represents the main drawing area."

	| stage |
	{'stage = Lib.Media.Display.stage'}.
	^Stage @ stage! !

Stage buildAccessors: #( frameRate )!
Stage buildGetters: #( stageWidth stageHeight color softKeyboardRect )!
Stage buildTypedGetters: #(
	#(stage3Ds 'Array[Stage3D]')
	#(stageVideos 'Array[StageVideo]')
)!
Stage buildLuaFunctions: #( getOrientation )!

! Stage methodsFor: #accessing !
stageExtent
	^self stageWidth @ self stageHeight! !

! Stage methodsFor: #accessing !
stageExtent: aPoint
	| p |
	p := aPoint asPoint.
	self stageWidth: p x; stageHeight: p y! !

EventDispatcher subclass: #Stage3D !
Stage3D buildLuaFunctions: #(
	#(requestContext3D #(context3DRenderMode))
)!
Stage3D buildTypedGetters: #(
	#(context3D Context3D)
)!
EventDispatcher subclass: #StageVideo !
StageVideo buildAccessors: #(
	depth
)!
StageVideo buildGetters: #(
	colorSpaces
	videoHeight videoWidth
	duration
)!
StageVideo buildTypedGetters: #(
	#(viewPort ARectangle)
	#(pan APoint)
	#(zoom APoint)
)!
StageVideo buildLuaFunctions: #(
	#(attachNetStream #(_netstream))
)!

LuaStudioLibrary subclass: #Tilesheet !
! Tilesheet class methodsFor: #instantiation !
new: bitmapdata
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Display.Tilesheet.new(bitmapdata:handle()))'}! !
Tilesheet buildLuaFunctions: #(
	#(addTileRect #(_rectangle _centerPoint))
	#(getTileRect #(index _into) ARectangle)
	#(drawTiles #(_graphics  #(tileData handles) smooth flags count))
)!
Tilesheet buildConstants: '
	Lib.Media.Display.Tilesheet.TILE_SCALE
	Lib.Media.Display.Tilesheet.TILE_ROTATION
	Lib.Media.Display.Tilesheet.TILE_RGB
	Lib.Media.Display.Tilesheet.TILE_ALPHA
	Lib.Media.Display.Tilesheet.TILE_RECT
	Lib.Media.Display.Tilesheet.TILE_ORIGIN
	Lib.Media.Display.Tilesheet.TILE_NO_ID
	Lib.Media.Display.Tilesheet.TILE_BLEND_NORMAL
'!

! Tilesheet class methodsFor: #constants !
tileTrans2x2
	" Return the constant value for Lib.Media.Display.Tilesheet.TILE_TRANS_2x2 "

	^{'Lib.Media.Display.Tilesheet.TILE_TRANS_2x2'}! !

DisplayObject subclass: #Display3D!
! Media class methodsFor: #accessing !
display3D
	" Returns a wrapper for library "

	^Display3D! !

Display3D modules: #(
	Context3DBlendFactor Context3DClearMask
	Context3DCompareMode Context3DProgramType
	Context3DTextureFilter Context3DTextureFormat
	Context3DTriangleFace Context3DVertexBufferFormat
	Context3DWrapMode Context3DMipFilter
)!
Context3DBlendFactor buildConstants: '
	Lib.Media.Display3D.Context3DBlendFactor.DESTINATION_ALPHA
	Lib.Media.Display3D.Context3DBlendFactor.DESTINATION_COLOR
	Lib.Media.Display3D.Context3DBlendFactor.ONE
	Lib.Media.Display3D.Context3DBlendFactor.ONE_MINUS_DESTINATION_ALPHA
	Lib.Media.Display3D.Context3DBlendFactor.ONE_MINUS_DESTINATION_COLOR
	Lib.Media.Display3D.Context3DBlendFactor.ONE_MINUS_SOURCE_ALPHA
	Lib.Media.Display3D.Context3DBlendFactor.ONE_MINUS_SOURCE_COLOR
	Lib.Media.Display3D.Context3DBlendFactor.SOURCE_ALPHA
	Lib.Media.Display3D.Context3DBlendFactor.SOURCE_COLOR
	Lib.Media.Display3D.Context3DBlendFactor.ZERO
'!
Context3DClearMask buildConstants: '
	Lib.Media.Display3D.Context3DClearMask.ALL
	Lib.Media.Display3D.Context3DClearMask.COLOR
	Lib.Media.Display3D.Context3DClearMask.DEPTH
	Lib.Media.Display3D.Context3DClearMask.STENCIL
'!
Context3DCompareMode buildConstants: '
	Lib.Media.Display3D.Context3DCompareMode.ALWAYS
	Lib.Media.Display3D.Context3DCompareMode.EQUAL
	Lib.Media.Display3D.Context3DCompareMode.GREATER
	Lib.Media.Display3D.Context3DCompareMode.GREATER_EQUAL
	Lib.Media.Display3D.Context3DCompareMode.LESS
	Lib.Media.Display3D.Context3DCompareMode.LESS_EQUAL
	Lib.Media.Display3D.Context3DCompareMode.NEVER
	Lib.Media.Display3D.Context3DCompareMode.NOT_EQUAL
'!
Context3DMipFilter buildConstants: '
	Lib.Media.Display3D.Context3DMipFilter.MIPLINEAR
	Lib.Media.Display3D.Context3DMipFilter.MIPNEAREST
	Lib.Media.Display3D.Context3DMipFilter.MIPNONE
'!
Context3DProgramType buildConstants: '
	Lib.Media.Display3D.Context3DProgramType.VERTEX
	Lib.Media.Display3D.Context3DProgramType.FRAGMENT
'!
Context3DTextureFilter buildConstants: '
	Lib.Media.Display3D.Context3DTextureFilter.ANISOTROPIC16X
	Lib.Media.Display3D.Context3DTextureFilter.ANISOTROPIC2X
	Lib.Media.Display3D.Context3DTextureFilter.ANISOTROPIC4X
	Lib.Media.Display3D.Context3DTextureFilter.ANISOTROPIC8X
	Lib.Media.Display3D.Context3DTextureFilter.LINEAR
	Lib.Media.Display3D.Context3DTextureFilter.NEAREST
'!
Context3DTextureFormat buildConstants: '
	Lib.Media.Display3D.Context3DTextureFormat.BGRA
	Lib.Media.Display3D.Context3DTextureFormat.COMPRESSED
	Lib.Media.Display3D.Context3DTextureFormat.COMPRESSED_ALPHA
'!
Context3DTriangleFace buildConstants: '
	Lib.Media.Display3D.Context3DTriangleFace.BACK
	Lib.Media.Display3D.Context3DTriangleFace.FRONT
	Lib.Media.Display3D.Context3DTriangleFace.FRONT_AND_BACK
	Lib.Media.Display3D.Context3DTriangleFace.NONE
'!
Context3DVertexBufferFormat buildConstants: '
	Lib.Media.Display3D.Context3DVertexBufferFormat.BYTES_4
	Lib.Media.Display3D.Context3DVertexBufferFormat.FLOAT_1
	Lib.Media.Display3D.Context3DVertexBufferFormat.FLOAT_2
	Lib.Media.Display3D.Context3DVertexBufferFormat.FLOAT_3
	Lib.Media.Display3D.Context3DVertexBufferFormat.FLOAT_4
'!
Context3DWrapMode buildConstants: '
	Lib.Media.Display3D.Context3DWrapMode.CLAMP
	Lib.Media.Display3D.Context3DWrapMode.CLAMP_U_REPEAT_V
	Lib.Media.Display3D.Context3DWrapMode.REPEAT
	Lib.Media.Display3D.Context3DWrapMode.REPEAT_U_CLAMP_V
'!

LuaStudioLibrary subclass: #AGALMiniAssembler !
! AGALMiniAssembler methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Display3D.Shaders.AGLSL.AGALMiniAssembler.new()'}! !

AGALMiniAssembler buildLuaFunctions: #(
	#(assemble #(type source) Shader)
)!

LuaStudioLibrary subclass: #Program3D !
Program3D subclass: #GLSLProgram !
Program3D buildLuaFunctions: #(
	#(upload #(_vertexShader _fragmentShader))
	dispose
)!
! GLSLProgram class methodsFor: #instantiation !
new: context3D
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Display3D.Shaders.GLSL.GLSLProgram.new(context3D:handle()))'}! !

GLSLProgram buildLuaFunctions: #(
	attach
	#(setVertexUniformFromMatrix #(name _with transposedMatrix))
	#(setVertexBufferAt #(name _with ofsset format))
)!

LuaStudioLibrary subclass: #Shader !
Shader subclass: #GLSLShader !
GLSLShader subclass: #GLSLFragmentShader !
GLSLShader subclass: #GLSLVertexShader !
GLSLShader buildLuaFunctions: #(
	#(nativeShader Shader)
)!
! GLSLFragmentShader class methodsFor: #instantiation !
source: source opt: option
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Display3D.Shaders.GLSL.GLSLFragmentShader.new(source, option))'}! !
! GLSLVertexShader class methodsFor: #instantiation !
source: source opt: option
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Display3D.Shaders.GLSL.GLSLVertexShader.new(source, option))'}! !

! GLSLShader class methodsFor: #instantiation !
source: source
	" Returns an instance of the receiver. "

	^self source: source opt: nil! !

! GLSLShader class methodsFor: #instantiation !
new: source
	" Returns an instance of the receiver. "

	^self source: source opt: nil! !

LuaStudioLibrary subclass: #Texture !
LuaStudioLibrary subclass: #CubeTexture !
LuaStudioLibrary subclass: #RectangleTexture !
Texture buildLuaFunctions: #(
	#(uploadFromBitmapData #(_source miplevel))
	#(uploadFromByteArray #(_data byteArrayOffset miplevel))
	#(uploadCompressedTextureFromByteArray #(_data byteArrayOffset async))
)!
CubeTexture buildLuaFunctions: #(
	#(uploadCompressedTextureFromByteArray #(_data byteArrayOffset async))
	#(uploadFromBitmapData #(_source side miplevel))
	#(uploadFromByteArray #(_data byteArrayOffset side miplevel))
)!
RectangleTexture buildLuaFunctions: #(
	#(uploadFromBitmapData #(_source))
	#(uploadFromByteArray #(_data byteArrayOffset))
)!

LuaStudioLibrary subclass: #IndexBuffer3D !
IndexBuffer3D buildLuaFunctions: #(
	dispose
	#(uploadFromByteArray #(_data byteArrayOffset startOffset count))
	#(uploadFromVector #(_data startOffset count))
	#(uploadFromByteArray #(_data byteArrayOffset startVertex numVertices))
	#(uploadFromVector #(_data startVertex numVertices))
)!
LuaStudioLibrary subclass: #VertexBuffer3D !
VertexBuffer3D buildLuaFunctions: #(
	dispose
	#(uploadFromByteArray #(_data byteArrayOffset startVertex numVertices))
	#(uploadFromVector #(_data startVertex numVertices))
	#(uploadFromByteArray #(_data byteArrayOffset startOffset count))
	#(uploadFromVector #(_data startOffset count))
)!

LuaStudioLibrary subclass: #Context3D!
Context3D buildGetters: #( driverInfo )!
Context3D buildAccessors: #( enableErrorChecking )!
Context3D buildLuaFunctions: #(
	#(clear #(red green blue alpha depth stencil mask))
	#(configureBackBuffer #(width height antiAlias enableDepthAndStencil))
	#(createCubeTexture #(size format optimize streamingLevels) CubeTexture)
	#(createIndexBuffer #(numIndices) IndexBuffer3D)
	#(createProgram Program3D)
	#(createRectangleTexture #(width height format optimize) RectangleTexture)
	#(createTexture #(width height format optimize levels) Texture)
	#(createVertexBuffer #(numVertices data32PerVertex) VertexBuffer3D)
	dispose
	#(drawToBitmapData #(_destination))
	#(drawTriangles #(_indexBuffer firstIndex numTriangles))
	present
	#(setBlendFactors #(sourceFactor destinationFactor))
	#(setColorMask #(red green blue alpha))
	#(setCulling #(triangleFaceToCull))
	#(setDepthTest #(depthMask passCompareMode))
	#(setProgram #(_program))
	#(setProgramConstantsFromByteArray #(programType firstRegister numRegisters _data byteArrayOffset))
	#(setProgramConstantsFromMatrix #(programType firstRegister _matrix transposedMatrix))
	#(setProgramConstantsFromVector #(programType firstRegister _data numRegisters))
	#(setGLSLProgramConstantsFromByteArray #(locationName _data byteArrayOffset))
	#(setGLSLProgramConstantsFromMatrix #(locationName _matrix transposedMatrix))
	#(setGLSLProgramConstantsFromVector4 #(locationName _data startIndex))
	setRenderToBackBuffer
	#(setRenderToTexture #(_texture enableDepthAndStencil antiAlias surfaceSelector))
	#(setSamplerStateAt #(_sampler wrap _filter _mipfilter))
	#(setScissorRectangle #(_rectangle))
	#(setStencilActions #(_triangleFace compareMode actionOnBothPass actionOnDepthFail actionOnDepthPassStencilFail))
	#(setStencilReferenceValue #(referenceValue readMask writeMask))
	#(setTextureAt #(_sampler _texture))
	#(setGLSLTextureAt #(locationName_ texture textureIndex))
	#(setVertexBufferAt #(index _buffer bufferOffset format))
	#(setGLSLVertexBufferAt #(locationName _buffer bufferOffset format))
	#(setRenderMethod #(aBlock))
)!

! Event class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.Event.new(type, bubbles, cancelable))'}! !

Event buildAccessors: #(
	bubbles cancelable currentTarget
	eventPhase target type
)!
Event buildLuaFunctions: #(
	clone stopImmediatePropagation
	stopPropagation toString
)!
Event buildConstants: '
	Lib.Media.Events.Event.ACTIVATE
	Lib.Media.Events.Event.ADDED
	Lib.Media.Events.Event.ADDED_TO_STAGE
	Lib.Media.Events.Event.CANCEL
	Lib.Media.Events.Event.CHANGE
	Lib.Media.Events.Event.CLOSE
	Lib.Media.Events.Event.COMPLETE
	Lib.Media.Events.Event.CONNECT
	Lib.Media.Events.Event.CONTEXT3D_CREATE
	Lib.Media.Events.Event.CONTEXT3D_LOST
	Lib.Media.Events.Event.DEACTIVATE
	Lib.Media.Events.Event.ENTER_FRAME
	Lib.Media.Events.Event.ID3
	Lib.Media.Events.Event.INIT
	Lib.Media.Events.Event.MOUSE_LEAVE
	Lib.Media.Events.Event.OPEN
	Lib.Media.Events.Event.REMOVED
	Lib.Media.Events.Event.REMOVED_FROM_STAGE
	Lib.Media.Events.Event.RENDER
	Lib.Media.Events.Event.RESIZE
	Lib.Media.Events.Event.SCROLL
	Lib.Media.Events.Event.SELECT
	Lib.Media.Events.Event.SOUND_COMPLETE
	Lib.Media.Events.Event.TAB_CHILDREN_CHANGE
	Lib.Media.Events.Event.TAB_ENABLED_CHANGE
	Lib.Media.Events.Event.TAB_INDEX_CHANGE
	Lib.Media.Events.Event.UNLOAD
	Lib.Media.Events.Event.VIDEO_FRAME
'!

Event subclass: #AccelerometerEvent !
! AccelerometerEvent class methodsFor: #instantiation !
type: type bubbles: bubbles
	cancelable: cancelable timestamp: timestamp
	x: x y: y z: z
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.AccelerometerEvent.new(type, bubbles, cancelable, timestamp, accelerationX, accelerationY, accelerationZ))'}! !
AccelerometerEvent buildAccessors: #(
	accelerationX accelerationY accelerationZ
	timestamp
)!
AccelerometerEvent buildConstants: '
	Lib.Media.Events.AccelerometerEvent.UPDATE
'!

Event subclass: #TextEvent!
! TextEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable text: text
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.TextEvent.new(type, bubbles, cancelable, text))'}! !
TextEvent buildConstants: '
	Lib.Media.Events.TextEvent.LINK
	Lib.Media.Events.TextEvent.TEXT_INPUT
'!

TextEvent subclass: #ErrorEvent!
! ErrorEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable text: text id: id
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.ErrorEvent.new(type, bubbles, cancelable, text, id))'}! !
ErrorEvent buildGetters: #( errorID )!
ErrorEvent buildConstants: '
	Lib.Media.Events.ErrorEvent.ERROR
'!

ErrorEvent subclass: #AsyncErrorEvent !
! AsyncErrorEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable text: text error: error
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.AsyncErrorEvent.new(type, bubbles, cancelable, text, error))'}! !
AsyncErrorEvent buildGetters: #( error )!
AsyncErrorEvent buildConstants: '
	Lib.Media.Events.AsyncErrorEvent.ASYNC_ERROR
'!

ErrorEvent subclass: #IOErrorEvent !
! IOErrorEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable text: text id: id
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.IOErrorEvent.new(type, bubbles, cancelable, text, id))'}! !
IOErrorEvent buildConstants: '
	Lib.Media.Events.IOErrorEvent.IO_ERROR
'!

Event subclass: #HTTPStatusEvent !
! HTTPStatusEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable status: status
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.HTTPStatusEvent.new(type, bubbles, cancelable, status))'}! !
HTTPStatusEvent buildGetters: #( status )!
HTTPStatusEvent buildTypedGetters: #( #(responseHeaders 'Array[URLRequestHeader]') )!
HTTPStatusEvent buildConstants: '
	Lib.Media.Events.HTTPStatusEvent.HTTP_STATUS
'!

Event subclass: #KeyboardEvent !
! KeyboardEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable
	charCodeValue: charCodeValue keyCodeValue: keyCodeValue keyLocationValue: keyLocationValue
	ctrlKeyValue: ctrlKeyValue altKeyValue: altKeyValue shiftKeyValue: shiftKeyValue
	controlKeyValue: controlKeyValue commandKeyValue: commandKeyValue
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.KeyboardEvent.new(type, bubbles, cancelable, charCodeValue, keyCodeValue, keyLocationValue, ctrlKeyValue, altKeyValue, shiftKeyValue, controlKeyValue, commandKeyValue))'}! !
KeyboardEvent buildGetters: #( altKey charCode commandKey controlKey ctrlKey keyCode keyLocation shiftKey)!
KeyboardEvent buildConstants: '
	Lib.Media.Events.KeyboardEvent.KEY_DOWN
	Lib.Media.Events.KeyboardEvent.KEY_UP
'!

Event subclass: #MouseEvent !
! MouseEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable
	localX: localX localY: localY relatedObject: relatedObject
	ctrlKey: ctrlKey altKey: altKey shiftKey: shiftKey
	buttonDown: buttonDown delta: delta
	commandKey: commandKey clickCount: clickCount
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.MouseEvent.new(type, bubbles, cancelable, localX, localY, relatedObject, ctrlKey, altKey, shiftKey, buttonDown, delta, commandKey, clickCount))'}! !
MouseEvent buildGetters: #( altKey buttonDown clickCount commandKey ctrlKey delta localX localY relatedObject shiftKey stageX stageY)!
MouseEvent buildConstants: '
	Lib.Media.Events.MouseEvent.CLICK
	Lib.Media.Events.MouseEvent.DOUBLE_CLICK
	Lib.Media.Events.MouseEvent.MIDDLE_CLICK
	Lib.Media.Events.MouseEvent.MIDDLE_MOUSE_DOWN
	Lib.Media.Events.MouseEvent.MIDDLE_MOUSE_UP
	Lib.Media.Events.MouseEvent.MOUSE_DOWN
	Lib.Media.Events.MouseEvent.MOUSE_MOVE
	Lib.Media.Events.MouseEvent.MOUSE_OUT
	Lib.Media.Events.MouseEvent.MOUSE_OVER
	Lib.Media.Events.MouseEvent.MOUSE_UP
	Lib.Media.Events.MouseEvent.MOUSE_WHEEL
	Lib.Media.Events.MouseEvent.RIGHT_CLICK
	Lib.Media.Events.MouseEvent.RIGHT_MOUSE_DOWN
	Lib.Media.Events.MouseEvent.RIGHT_MOUSE_UP
	Lib.Media.Events.MouseEvent.ROLL_OUT
	Lib.Media.Events.MouseEvent.ROLL_OVER
'!

Event subclass: #NetStatusEvent !
! NetStatusEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable info: info
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.NetStatusEvent.new(type, bubbles, cancelable, info))'}! !
NetStatusEvent buildGetters: #( info )!
NetStatusEvent buildConstants: '
	Lib.Media.Events.NetStatusEvent.NET_STATUS 
'!

Event subclass: #FocusEvent !
! FocusEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable
	relatedObject: relatedObject shiftKey: shiftKey
	keyCode: keyCode direction: direction
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.FocusEvent.new(type, bubbles, cancelable, relatedObject, shiftKey, keyCode, direction))'}! !
FocusEvent buildGetters: #( direction keyCode relatedObject shiftKey )!
FocusEvent buildConstants: '
	Lib.Media.Events.FocusEvent.FOCUS_IN
	Lib.Media.Events.FocusEvent.FOCUS_OUT
	Lib.Media.Events.FocusEvent.KEY_FOCUS_CHANGE
	Lib.Media.Events.FocusEvent.MOUSE_FOCUS_CHANGE
'!

Event subclass: #ProgressEvent !
! ProgressEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable
	bytesLoaded: bytesLoaded bytesTotal: bytesTotal
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.ProgressEvent.new(type, bubbles, cancelable, bytesLoaded, bytesTotal))'}! !
ProgressEvent buildGetters: #( bytesLoaded bytesTotal )!
ProgressEvent buildConstants: '
	Lib.Media.Events.ProgressEvent.PROGRESS
	Lib.Media.Events.ProgressEvent.SOCKET_DATA
'!

Event subclass: #SoftKeyboardEvent !
! SoftKeyboardEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable
	relatedObject: relatedObject triggerType: triggerType
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.SoftKeyboardEvent.new(type, bubbles, cancelable, relatedObject, triggerType))'}! !
SoftKeyboardEvent buildGetters: #( relatedObject triggerType )!
SoftKeyboardEvent buildConstants: '
	Lib.Media.Events.SoftKeyboardEvent.SOFT_KEYBOARD_ACTIVATE
	Lib.Media.Events.SoftKeyboardEvent.SOFT_KEYBOARD_ACTIVATING
	Lib.Media.Events.SoftKeyboardEvent.SOFT_KEYBOARD_DEACTIVATE
'!

Event subclass: #StageVideoEvent !
StageVideoEvent buildConstants: '
	Lib.Media.Event.StageVideoEvent.RENDER_STATE
	Lib.Media.Event.StageVideoEvent.RENDER_STATUS_ACCELERATED
	Lib.Media.Event.StageVideoEvent.RENDER_STATUS_SOFTWARE
	Lib.Media.Event.StageVideoEvent.RENDER_STATUS_UNAVAILABLE
'!

Event subclass: #TimerEvent !
! TimerEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.TimerEvent.new(type, bubbles, cancelable))'}! !
TimerEvent buildConstants: '
	Lib.Media.Events.TimerEvent.TIMER
	Lib.Media.Events.TimerEvent.TIMER_COMPLETE
'!

MouseEvent subclass: #TouchEvent !
! TouchEvent class methodsFor: #instantiation !
type: type bubbles: bubbles cancelable: cancelable
	touchPointID: touchPointID isPrimaryTouchPoint: isPrimaryTouchPoint
	localX: localX localY: localY sizeX: sizeX sizeY: sizeY
	relatedObject: relatedObject ctrlKey: ctrlKey altKey: altKey shiftKey: shiftKey
	buttonDown: buttonDown delta: delta commandKey: commandKey clickCount: clickCount
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Events.TouchEvent.new(type, bubbles, cancelable, touchPointID, isPrimaryTouchPoint, localX, localY, sizeX, sizeY, relatedObject, ctrlKey, altKey, shiftKey, buttonDown, delta, commandKey, clickCount))'}! !
TouchEvent buildGetters: #( altKey buttonDown clickCount commandKey ctrlKey delta isPrimaryTouchPoint localX localY relatedObject shiftKey sizeX sizeY stageX stageY touchPointID )!
TouchEvent buildConstants: '
	Lib.Media.Events.TouchEvent.TOUCH_BEGIN
	Lib.Media.Events.TouchEvent.TOUCH_END
	Lib.Media.Events.TouchEvent.TOUCH_MOVE
	Lib.Media.Events.TouchEvent.TOUCH_OUT
	Lib.Media.Events.TouchEvent.TOUCH_OVER
	Lib.Media.Events.TouchEvent.TOUCH_ROLL_OUT
	Lib.Media.Events.TouchEvent.TOUCH_ROLL_OVER
	Lib.Media.Events.TouchEvent.TOUCH_TAP
'!

LuaStudioLibrary subclass: #GlowFilter !
! GlowFilter class methodsFor: #instantiation !
color: color alpha: alpha
	blurX: blurX blurY: blurY
	strength: strength quality: quality
	inner: inner knockout: knockout
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Filters.GlowFilter.new(color, alpha, blurX, blurY, strength, quality, inner, knockout))'}! !
GlowFilter buildAccessors: #( alpha blurX blurY color inner knockout quality strength )!
GlowFilter buildLuaFunctions: #(
	#(clone self)
)!

LuaStudioLibrary subclass: #DropShadowFilter !
! DropShadowFilter class methodsFor: #instantiation !
distance: distance angle: angle
	color: color alpha: alpha
	blurX: blurX blurY: blurY
	strength: strength quality: quality
	inner: inner knockout: knockout
	hideObject: hideObject
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Filters.DropShadowFilter.new(distance, angle, color, alpha, blurX, blurY, strength, quality, inner, knockout, hideObject))'}! !

DropShadowFilter buildAccessors: #( distance color angle alpha blurX blurY color distance hideObject inner knockout quality strength )!
DropShadowFilter buildLuaFunctions: #(
	#(clone self)
)!

LuaStudioLibrary subclass: #ColorMatrixFilter !
! ColorMatrixFilter class methodsFor: #instantiation !
matrix: matrix
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Filters.ColorMatrixFilter.new(matrix))'}! !
! ColorMatrixFilter class methodsFor: #instantiation !
new: matrix
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Filters.ColorMatrixFilter.new(matrix))'}! !
ColorMatrixFilter buildAccessors: #( matrix )!
ColorMatrixFilter buildLuaFunctions: #(
	#(clone self)
)!

LuaStudioLibrary subclass: #BlurFilter !
! BlurFilter class methodsFor: #instantiation !
blurX: blurX blurY: blurY quality: quality
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Filters.BlurFilter.new(blurX, blurY, quality))'}! !
BlurFilter buildAccessors: #( blurX blurY quality )!
BlurFilter buildLuaFunctions: #(
	#(clone self)
)!

Media modules: #( Geom )!
Geom modules: #(
	ColorTransform
	Matrix Matrix3D
	Orientation3D
	APoint ARectangle
	Transform Vector3D
)!

! ColorTransform class methodsFor: #instantiation !
@ array
	" Returns an instance of the receiver. "

	(self isObject: array) ifFalse: [ ^super @ array ].
	array size = 8 ifFalse: [ ^self error: 'Invalid array size' ].
	^self	redMultiplier: array[1]
		greenMultiplier: array[2]
		blueMultiplier: array[3]
		alphaMultiplier: array[4]
		redOffset: array[5]
		greenOffset: array[6]
		blueOffset: array[7]
		alphaOffset: array[8]! !

! ColorTransform class methodsFor: #instantiation !
redMultiplier: redMultiplier greenMultiplier: greenMultiplier
	blueMultiplier: blueMultiplier alphaMultiplier: alphaMultiplier
	redOffset: redOffset greenOffset: greenOffset
	blueOffset: blueOffset alphaOffset: alphaOffset
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Geom.ColorTransform.new(redMultiplier, greenMultiplier, blueMultiplier, alphaMultiplier, redOffset, greenOffset, blueOffset, alphaOffset))'}! !
ColorTransform buildAccessors: #(
	color
	alphaMultiplier alphaOffset blueMultiplier blueOffset
	greenMultiplier greenOffset redMultiplier redOffset )!
ColorTransform buildLuaFunctions: #(
	#(concat #(_second))
	toString
)!

! Matrix class methodsFor: #instantiation !
a: a b: b c: c d: d tx: tx ty: ty
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Geom.Matrix.new(a, b, c, d, tx, ty))'}! !
Matrix buildAccessors: #( a b c d tx ty )!
Matrix buildLuaFunctions: #(
	#(clone self)
	#(concat #(_m))
	#(copyColumnFrom #(_column _vector3D))
	#(copyColumnTo #(_column _vector3D))
	#(copyFrom #(_sourceMatrix))
	#(copyRowFrom #(_row _vector3D))
	#(copyRowTo #(_row _vector3D))
	#(createBox #(scaleX scaleY rotation tx ty))
	#(createGradientBox #(width height rotation tx ty))
	#(deltaTransformPoint #(_point))
	identity invert
	#(rotate #(angle))
	#(scale #(sx sy))
	#(setTo #(aa ba ca da txa tya))
	toString
	#(transformPoint #(_point))
	#(translate #(dx dy))
)!

! Matrix3D class methodsFor: #instantiation !
v: v
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Geom.Matrix3D.new(v))'}! !

! Matrix3D class methodsFor: #instantiation !
new: v
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Geom.Matrix3D.new(v))'}! !

! Matrix3D class methodsFor: #instantiation !
createOrtho: array
	" Returns an instance of the receiver. "

	array size = 6 ifFalse: [ ^self error: 'Invalid array size' ].
	^{'self:_at(Lib.Media.Geom.Matrix3D.createOrtho(array[1],array[2],array[3],array[4],array[5],array[6]))'}! !

! Matrix3D class methodsFor: #instantiation !
create2D: array
	" Returns an instance of the receiver. "

	array size = 4 ifFalse: [ ^self error: 'Invalid array size' ].
	^{'self:_at(Lib.Media.Geom.Matrix3D.create2D(array[1],array[2],array[3],array[4]))'}! !

! Matrix3D methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Geom.Matrix3D.new(nil)'}! !

Matrix3D buildGetters: #( determinant )!
Matrix3D buildAccessors: #( rawData )!
Matrix3D buildTypedAccessors: #(
	#(position Vector3D)
)!
Matrix3D buildLuaFunctions: #(
	#(append #(_lhs))
	#(appendRotation #(degrees _axis _pivotPoint))
	#(appendScale #(xScale yScale zScale))
	#(appendTranslation #(x y z))
	#(clone Matrix3D)
	#(copyColumnFrom #(_column _vector3D))
	#(copyColumnTo #(_column _vector3D))
	#(copyFrom #(_sourceMatrix3D))
	#(copyRawDataFrom #(_vector index transpose))
	#(copyRawDataTo #(_vector index transpose))
	#(copyRowFrom #(_row _vector3D))
	#(copyRowTo #(_row _vector3D))
	#(copyToMatrix3D #(_dest))
	#(decompose #(orientationStyle))
	#(deltaTransformVector #(_v))
	identity
	#(interpolate #(_thisMat _toMat percent))
	#(interpolateTo #(_toMat percent))
	invert
	#(pointAt #(pos at up))
	#(prepend #(_rhs))
	#(prependRotation #(degrees axis _pivotPoint))
	#(prependScale #(xScale yScale zScale))
	#(prependTranslation #(x y z))
	#(recompose #(#(components handles) orientationStyle))
	#(transformVector #(_v))
	#(transformVectors #(vin vout))
	transpose
)!

Orientation3D buildConstants: '
	Lib.Media.Geom.Orientation3D.AXIS_ANGLE
	Lib.Media.Geom.Orientation3D.EULER_ANGLES
	Lib.Media.Geom.Orientation3D.QUATERNION
'!

! APoint class methodsFor: #instantiation !
x: x y: y
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Geom.Point.new(x, y))'}! !

! APoint class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Media.Geom.Point)'}! !

APoint buildGetters: #( length )!
APoint buildAccessors: #( x y )!
APoint buildLuaStaticFunctions: #(
	#(distance #(_pt1 _to))
	#(interpolate #(_pt1 _with level))
	#(polar #(len angle))
)!
APoint buildLuaFunctions: #(
	#(add #(_v))
	#(clone self)
	#(copyFrom #(_sourcePoint))
	#(equals #(_to))
	#(normalize #(thickness))
	#(offset #(dx dy))
	#(setTo #(xa ya))
	#(subtract #(_v))
	toString
)!

! ARectangle class methodsFor: #instantiation !
x: x y: y width: width height: height
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Geom.Rectangle.new(x, y, width, height))'}! !

! ARectangle class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Media.Geom.Rectangle)'}! !

ARectangle buildAccessors: #( bottom bottomRight height left right size top topLeft width x y )!
ARectangle buildLuaFunctions: #(
	#(clone self)
	#(contains #(x y))
	#(containsPoint #(_point))
	#(containsRect #(_rect))
	#(copyFrom #(_sourceRect))
	#(equals #(_toCompare))
	#(inflate #(dx dy))
	#(inflatePoint #(_point))
	#(intersection #(_toIntersect))
	#(intersects #(_toIntersect))
	isEmpty
	#(offset #(dx dy))
	#(offsetPoint #(_point))
	setEmpty
	#(setTo #(xa ya widtha heighta))
	toString
	#(union #(_toUnion))
)!

! Transform class methodsFor: #instantiation !
inParent: inParent
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Geom.Transform.new(inParent:handle()))'}! !
! Transform class methodsFor: #instantiation !
new: inParent
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Geom.Transform.new(inParent:handle()))'}! !
! Transform class methodsFor: #instantiation !
in: inParent
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Geom.Transform.new(inParent:handle()))'}! !
! Transform class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Media.Geom.Transform)'}! !
Transform buildTypedAccessors: #(
	#(colorTransform ColorTransform)
	#(matrix Matrix)
)!
Transform buildTypedGetters: #(
	#(concatenatedColorTransform ColorTransform)
	#(concatenatedMatrix Matrix)
	#(pixelBounds ARectangle)
)!

! Vector3D class methodsFor: #instantiation !
@ array
	" Returns an instance of the receiver. "

	(self isObject: array) ifFalse: [ ^super @ array ].
	array size = 3 ifTrue: [
		^self	x: array[1] y: array[2]
			z: array[3] w: 0
	].
	array size = 4 ifTrue: [
		^self	x: array[1] y: array[2]
			z: array[3] w: array[4]
	].
	^self error: 'Invalid array size' ! !

! Vector3D class methodsFor: #instantiation !
x: x y: y z: z w: w
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.Geom.Vector3D.new(x, y, z, w))'}! !

! Vector3D class methodsFor: #instantiation !
x: x y: y z: z
	" Returns an instance of the receiver. "

	^self x: x y: y z: z w: 0! !

! Vector3D class methodsFor: #constants !
zero
	" Returns an instance of the receiver. "

	^self x: 0 y: 0 z: 0 w: 0! !

! Vector3D class methodsFor: #constants !
xAxis
	" Returns an instance of the receiver. "

	^self x: 1 y: 0 z: 0 w: 0! !

! Vector3D class methodsFor: #constants !
yAxis
	" Returns an instance of the receiver. "

	^self x: 0 y: 1 z: 0 w: 0! !

! Vector3D class methodsFor: #constants !
zAxis
	" Returns an instance of the receiver. "

	^self x: 0 y: 0 z: 1 w: 0! !

! Vector3D class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Media.Geom.Vector3D)'}! !
Vector3D buildAccessors: #( w x y z )!
Vector3D buildGetters: #( length lengthSquared )!
Vector3D buildLuaStaticFunctions: #(
	#(angleBetween #(_a _b))
	#(distance #(_pt1 _pt2))
)!
Vector3D buildLuaFunctions: #(
	#(add #(_a))
	#(clone self)
	#(copyFrom #(_sourceVector3D))
	#(crossProduct #(_a))
	#(decrementBy #(_a))
	#(dotProduct #(_a))
	#(equals #(_toCompare allFour))
	#(incrementBy #(_a))
	#(nearEquals #(_toCompare tolerance allFour))
	negate
	normalize
	project
	#(scaleBy #(_s))
	#(setTo #(xa ya za))
	#(subtract #(_a))
	toString
)!

! Vector3D methodsFor: #converting !
xyzw
	^Array with: self x with: self y  with: self z  with: self w! !

! Vector3D methodsFor: #converting !
asArray
	^Array with: self x with: self y  with: self z  with: self w! !

! Vector3D methodsFor: #converting !
xyz
	^Array with: self x with: self y  with: self z! !

LuaStudioLibrary subclass: #MediaFileSystem !
! Media class methodsFor: #accessing !
fileSystem
	" Returns the file module. "

	^MediaFileSystem! !
MediaFileSystem modules: #( MediaFile )!
! MediaFileSystem class methodsFor: #accessing !
file
	" Returns the file module. "

	^MediaFile! !

! MediaFile class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Media.FileSystem.File)'}! !

! MediaFile class methodsFor: #instantiation !
new: path
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Media.FileSystem.File.new(path))'}! !

! MediaFile methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.FileSystem.File.new(nil)'}! !

MediaFile buildAccessors: #(
	url nativePath preventBackup
)!
MediaFile buildLuaFunctions: #(
	#(browseForDirectory #(title))
	#(browseForOpen #(title _typeFilter))
	#(browseForSave #(title))
)!

#(	applicationDirectory applicationStorageDirectory
	desktopDirectory documentsDirectory userDirectory
) do: [:selector|
	MediaFile class
		implement: selector
		comment: 'Returns the ' ,selector
		as: '^self @ (self nativeClass handle basicAt: #' ,selector ,$)
		category: #folders
]!

! Media class methodsFor: #accessing !
net
	" Private - Returns the handle to the native module. "

	^Net! !

Net modules: #(
	FileFilter NetConnection NetStream
	URLLoader URLLoaderDataFormat
	URLRequest URLRequestMethod
	URLVariables
)!

! FileFilter class methodsFor: #instantiation !
description: description extension: extension macType: macType
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Net.FileFilter.new(description, extension, macType))'}! !
FileFilter buildAccessors: #( description extension macType )!

! NetConnection methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Net.NetConnection.new()'}! !
! NetConnection class methodsFor: #constants !
defaultObjectEncoding
	" Returns the default encoding for NetConnection objects. "

	^{'Lib.Media.Net.NetConnection.defaultObjectEncoding'}! !
NetConnection buildAccessors: #(
	client objectEncoding
)!
NetConnection buildGetters: #(
	connected uri
	usingTLS proxyType
	connectedProxyType
)!
NetConnection buildLuaFunctions: #(
	#(connect #(command p1 p2 p3 p4 p5))
	close
)!

! NetStream class methodsFor: #instantiation !
connection: connection peerID: peerID
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Net.NetStream.new(connection, peerID))'}! !
NetStream buildAccessors: #(
	client soundTransform
)!
NetStream buildGetters: #(
	bytesTotal bytesLoaded
	objectEncoding peerStreams
	time
)!
NetStream buildLuaFunctions: #(
	#(attach #(_connection))
	#(seek #(offset))
	close dispose pause
	#(play #(inFilename startSeconds lenSeconds p4 p5))
	togglePause resume
	#(receiveAudio #(flag))
	#(receiveVideo #(flag))
	#(onPeerConnect #(_subscriber))
)!

! URLLoader class methodsFor: #instantiation !
request: request
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Net.URLLoader.new(request:handle()))'}! !
URLLoader buildAccessors: #(
	bytesLoaded bytesTotal
	data dataFormat
)!
URLLoader buildLuaFunctions: #(
	#(load #(_request))
	close
)!
URLLoaderDataFormat buildConstants: '
	Lib.Media.Net.URLLoaderDataFormat.BINARY
	Lib.Media.Net.URLLoaderDataFormat.TEXT
	Lib.Media.Net.URLLoaderDataFormat.VARIABLES
'!
URLRequestMethod buildConstants: '
	Lib.Media.Net.URLRequestMethod.DELETE
	Lib.Media.Net.URLRequestMethod.GET
	Lib.Media.Net.URLRequestMethod.HEAD
	Lib.Media.Net.URLRequestMethod.OPTIONS
	Lib.Media.Net.URLRequestMethod.POST
	Lib.Media.Net.URLRequestMethod.PUT
'!

! URLRequest class methodsFor: #instantiation !
new: url
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Net.URLRequest.new(url))'}! !

! URLRequest class methodsFor: #instantiation !
url: url
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Net.URLRequest.new(url))'}! !

! URLRequest class methodsFor: #instantiation !
@ url
	" Returns an instance of the receiver "

	#string = (self typeOf: url) ifTrue: [ ^self new: url ].
	^super @ url! !

URLRequest buildTypedAccessors: #(
	url userAgent requestHeaders
	method contentType
	#(data Lua)
	followRedirects
	cookieString
)!
URLRequest buildLuaFunctions: #(
	launchBrowser
	#(basicAuth #(key value))
)!

! URLVariables class methodsFor: #instantiation !
new: source
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Net.URLVariables.new(source))'}! !

! URLVariables class methodsFor: #instantiation !
@ url
	" Returns an instance of the receiver "

	#string = (self typeOf: url) ifTrue: [ ^self new: url ].
	^super @ url! !

URLVariables buildTypedAccessors: #(
	name email submit method
	#(data Lua)
)!
URLVariables buildLuaFunctions: #(
	#(decode #(source))
	toString
)!

Media modules: #( Sensors System Text UI Utils )!
EventDispatcher subclass: #Accelerometer!
! Sensors class methodsFor: #accessing !
accelerometer
	" Returns a wrapper for library "

	^Accelerometer! !

! Accelerometer methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Sensors.Accelerometer.new()'}! !
! Accelerometer class methodsFor: #constants !
isSupported
	^{'Lib.Media.Sensors.Accelerometer.isSupported'}! !
Accelerometer buildGetters: #( muted )!
Accelerometer buildLuaFunctions: #(
	#(setRequestedUpdateInterval #(interval))
)!

! System class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Media.System)'}! !
#( totalMemory totalMemoryNumber deviceID exeName ) do: [:selector|
	System class
		implement: selector
		comment: 'Returns the ' ,selector
		as: '^self nativeClass handle basicAt: #' ,selector
		category: #constants
]!
System buildLuaStaticFunctions: #(
	#(exit #(code))
	gc #(pauseForGCIfCollectionImminent #(imminence))
	systemName getLocalIpAddress
	#(restart #(code))
	#(getUrl #(_request target))
)!

Text modules: #(
	AntiAliasType Font
	FontStyle FontType
	TextFormat
	SoftKeyboardType
)!
AntiAliasType buildConstants: '
	Lib.Media.Text.AntiAliasType.ADVANCED
	Lib.Media.Text.AntiAliasType.NORMAL
'!

! Font class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Media.Text.Font)'}! !
! Font class methodsFor: #instantiation !
name: fontName style: fontStyle type: fontType
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Text.Font.new(fontName, fontStyle, fontType))'}! !
Font buildGetters: #( fontName fontStyle fontType )!
Font buildLuaFunctions: #( toString )!
Font buildLuaStaticFunctions: #(
	#(loadBytes #(_bytes))
	#(registerFontData #(_instance _bytes))
	#(enumerateFonts #(enumerateDeviceFonts) 'Array[Font]')
	#(isRegistered #(_instance))
	#(unregisterFontData #(_instance))
)!
FontStyle buildConstants: '
	Lib.Media.Text.FontStyle.BOLD
	Lib.Media.Text.FontStyle.BOLD_ITALIC
	Lib.Media.Text.FontStyle.ITALIC
	Lib.Media.Text.FontStyle.REGULAR
'!
FontType buildConstants: '
	Lib.Media.Text.FontType.DEVICE
	Lib.Media.Text.FontType.EMBEDDED
	Lib.Media.Text.FontType.EMBEDDED_CFF
'!
SoftKeyboardType buildConstants: '
	Lib.Media.Text.SoftKeyboardType.CONTACT
	Lib.Media.Text.SoftKeyboardType.DEFAULT
	Lib.Media.Text.SoftKeyboardType.EMAIL
	Lib.Media.Text.SoftKeyboardType.NUMBER
	Lib.Media.Text.SoftKeyboardType.PUNCTUATION
	Lib.Media.Text.SoftKeyboardType.URL
'!
InteractiveObject subclass: #TextField!

! Text class methodsFor: #accessing !
textField
	" Returns a wrapper for library "

	^TextField! !
! TextField class methodsFor: #instantiating !
label
	" Returns an instance of the receiver.
	A multi line text box set with default values.
	"

	^self new
		wordWrap: true; multiline: true; border: false;
		background: true; backgroundColor: 16rFFFFFF;
		yourself! !
! TextField class methodsFor: #instantiating !
box
	" Returns an instance of the receiver.
	A multi line text box set with default values.
	"

	^self label
		border: true; borderColor: 16r000000;
		yourself! !
! TextField class methodsFor: #instantiating !
input
	" Returns an instance of the receiver.
	A multi line input field set with default values.
	"

	^self box
		type: TextFieldType input;
		yourself! !

! TextField methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Text.TextField.new()'}! !

TextField buildGetters: #(
	bottomScrollV
	length
	caretIndex
	maxScrollH
	maxScrollV
	numLines
	selectionBeginIndex
	selectionEndIndex
	textHeight
	textWidth
)!
TextField buildTypedAccessors: #(
	antiAliasType
	autoSize
	background
	backgroundColor
	border
	borderColor
	#(defaultTextFormat TextFormat)
	displayAsPassword
	embedFonts
	gridFitType
	htmlText
	maxChars
	mouseWheelEnabled
	multiline
	scrollH
	scrollV
	selectable
	sharpness
	text
	textColor
	type
	wordWrap
)!
TextField buildLuaFunctions: #(
	#(appendText #(text))
	#(getCharNoundaries #(charIndex) ARectangle)
	#(getCharIndexAtPoint #(x y))
	#(getFirstCharInParagraph #(charIndex))
	#(getLineIndexAtPoint #(x y))
	#(getLineIndexOfChar #(charIndex))
	#(getLineLength #(lineIndex))
	#(getLineMetrics #(lineIndex))
	#(getLineOffset #(lineIndex))
	#(getLineText #(lineIndex))
	#(getTextFormat #(beginIndex endIndex))
	#(getTextRuns #(beginIndex endIndex) 'Array[TextRun]')
	#(replaceText #(beginIndex endIndex text))
	#(setSelection #(beginIndex endIndex))
	#(setTextFormat #(_format beginIndex endIndex))
)!

! TextField methodsFor: #accessors !
text: aString position: position extent: extent
	self
		text: aString;
		position: position;
		extent: extent! !

! TextField methodsFor: #accessors !
text: aString boundingBox: aRectangle
	self
		text: aString;
		position: aRectangle origin;
		extent: aRectangle extent! !

! TextField methodsFor: #accessors !
text: aString extent: aPoint
	self text: aString; extent: aPoint! !

! TextField methodsFor: #accessors !
textExtent: aPoint
	| p |
	p := aPoint asPoint.
	self textWidth: p x; textHeight: p y.! !

! TextField methodsFor: #accessors !
textExtent
	^self textWidth @ self textHeight! !

LuaStudioLibrary subclass: #TextRun!
TextRun buildTypedAccessors: #(
	beginIndex endIndex
	#(textFormat TextFormat)
)!

! TextFormat class methodsFor: #instantiation !
font: font size: size color: color bold: bold italic: italic underline: underline
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Text.TextFormat.new(font, size, color, bold, italic, underline))'}! !

TextFormat buildAccessors: #(
	align
	blockIndent
	bold
	bullet
	color
	font
	indent
	italic
	kerning
	leading
	leftMargin
	letterSpacing
	rightMargin
	size
	tabStops
	target
	underline
	url
	lineThrough
	baselineShift
)!

Text modules: #(
	TextFieldAutoSize TextFieldType
	TextFormatAlign
)!
TextFieldAutoSize buildConstants: '
	Lib.Media.Text.TextFieldAutoSize.CENTER
	Lib.Media.Text.TextFieldAutoSize.LEFT
	Lib.Media.Text.TextFieldAutoSize.NONE
	Lib.Media.Text.TextFieldAutoSize.RIGHT
'!
TextFieldType buildConstants: '
	Lib.Media.Text.TextFieldType.DYNAMIC
	Lib.Media.Text.TextFieldType.INPUT
'!
TextFormatAlign buildConstants: '
	Lib.Media.Text.TextFormatAlign.CENTER
	Lib.Media.Text.TextFormatAlign.JUSTIFY
	Lib.Media.Text.TextFormatAlign.LEFT
	Lib.Media.Text.TextFormatAlign.RIGHT
'!

UI modules: #( Multitouch MultitouchInputMode )!
! Multitouch class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^{'smalltalk.Lua:_at(Lib.Media.UI.Multitouch)'}! !

#( inputMode maxTouchPoints supportedGestures supportsGestureEvents supportsTouchEvents ) do: [:selector|
	Multitouch class
		implement: selector
		comment: 'Returns the ' ,selector
		as: '^self nativeClass handle basicAt: #' ,selector
		category: #accessing
]!

! Multitouch class methodsFor: #accessing!
inputMode: mode
	^self nativeClass handle basicAt: #inputMode put: mode! !

MultitouchInputMode buildConstants: '
	Lib.Media.UI.MultitouchInputMode.GESTURE
	Lib.Media.UI.MultitouchInputMode.NONE
	Lib.Media.UI.MultitouchInputMode.TOUCH_POINT
'!

Utils modules: #( CompressionAlgorithm Endian )!

Bytes subclass: #ByteArray !
! Utils class methodsFor: #accessing !
byteArray
	" Returns a wrapper for library "

	^ByteArray! !
! ByteArray class methodsFor: #instantiation !
new: size
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Utils.ByteArray.new(size))'}! !
! ByteArray class methodsFor: #instantiation !
fromBytes: bytes
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Utils.ByteArray.fromBytes(bytes:handle()))'}! !
ByteArray buildAccessors: #( endian length position )!
ByteArray buildGetters: #( bytesAvailable )!
ByteArray buildLuaFunctions: #(
	clear
	#(compress #(algorithm))
	deflate inflate
	readBoolean readByte
	#(readBytes #(_bytes offset length))
	readDouble readFloat readInt
	#(readMultiByte #(length charSet))
	readShort readUnsignedByte
	readUnsignedInt readUnsignedShort
	readUTF #(readUTFBytes #(length))
	toString
	#(uncompress #(algorithm))
	#(writeBoolean #(value))
	#(writeByte #(value))
	#(writeBytes #(_bytes offset length))
	#(writeDouble #(value))
	#(writeFloat #(value))
	#(writeInt #(value))
	#(writeMultiByte #(value charSet))
	#(writeShort #(value))
	#(writeUnsignedInt #(value))
	#(writeUTF #(value))
	#(writeUTFBytes #(value))
	asString
)!
CompressionAlgorithm buildConstants: '
	Lib.Media.Utils.CompressionAlgorithm.DEFLATE
	Lib.Media.Utils.CompressionAlgorithm.LZMA
	Lib.Media.Utils.CompressionAlgorithm.ZLIB
	Lib.Media.Utils.CompressionAlgorithm.GZIP
'!
Endian buildConstants: '
	Lib.Media.Utils.Endian.BIG_ENDIAN
	Lib.Media.Utils.Endian.LITTLE_ENDIAN
'!

EventDispatcher subclass: #Timer!
! Utils class methodsFor: #accessing !
timer
	" Returns a wrapper for library "

	^Timer! !
! Timer class methodsFor: #instantiation !
delay: delay repeatCount: repeatCount
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Utils.Timer.new(delay, repeatCount))'}! !
Timer  buildAccessors: #( delay repeatCount )!
Timer  buildGetters: #( currentCount running )!
Timer buildLuaFunctions: #( reset start stop )!

Display modules: #(
	GraphicsBitmapFill GraphicsEndFill
	GraphicsGradientFill GraphicsPath
	GraphicsSolidFill GraphicsStroke
)!

! GraphicsBitmapFill class methodsFor: #instantiation !
bitmapData: bitmapData matrix: matrix repeat: count smooth: smooth
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Display.GraphicsBitmapFill.new(bitmapData, matrix, count, smooth))'}! !

! GraphicsEndFill methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Media.Display.GraphicsEndFill.new()'}! !

! GraphicsGradientFill class methodsFor: #instantiation !
type: type colors: colors
	alphas: alphas ratios: ratios
	matrix: matrix spreadMethod: spreadMethod
	interpolationMethod: interpolationMethod focalPointRatio: focalPointRatio
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Display.GraphicsGradientFill.new(type, colors, alphas, ratios, matrix, spreadMethod, interpolationMethod, focalPointRatio))'}! !

! GraphicsPath class methodsFor: #instantiation !
commands: commands data: data winding: winding
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Display.GraphicsPath.new(commands, data, winding))'}! !
GraphicsPath buildAccessors: #( commands data )!
GraphicsPath buildLuaFunctions: #(
	#(curveTo #(controlX controlY anchorX anchorY))
	#(lineTo #(x y))
	#(moveTo #(x y))
	#(wideLineTo #(x y))
	#(wideMoveTo #(x y))
)!

! GraphicsSolidFill class methodsFor: #instantiation !
color: color alpha: alpha
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Display.GraphicsSolidFill.new(color, alpha))'}! !

! GraphicsStroke class methodsFor: #instantiation !
thickness: thickness pixelHinting: pixelHinting scaleMode: scaleMode caps: caps joints: joints miterLimit: miterLimit fill: fill
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Media.Display.GraphicsStroke.new(thickness, pixelHinting, scaleMode, caps, joints, miterLimit, fill))'}! !

Away3D modules: #(
	Animators Cameras Containers
	Controllers Core
	Debug Entities Lights
	Materials Primitives
	Text Textures
)!
Animators modules: #(
	ParticleAnimationData
	ParticlePropertiesMode
	Nodes
)!
ParticlePropertiesMode buildConstants: '
	Lib.Away3D.Animators.Data.ParticlePropertiesMode.GLOBAL
	Lib.Away3D.Animators.Data.ParticlePropertiesMode.LOCAL_DYNAMIC
	Lib.Away3D.Animators.Data.ParticlePropertiesMode.LOCAL_STATIC
'!
ParticleAnimationData buildAccessors: #(
	delay duration index
	numVertices startTime startVertexIndex
	totalTime
)!

EventDispatcher subclass: #NamedAssetBase!
NamedAssetBase buildAccessors: #( id name )!
NamedAssetBase buildGetters: #( assetFullPath assetNamespace originalName )!
NamedAssetBase buildLuaFunctions: #(
	#(assetPathEquals #(name ns))
	#(resetAssetPath #(name ns overrideOriginal))
)!
! NamedAssetBase class methodsFor: #constants !
defaultNamespace

	^{'Lib.Away3D.Animators.NamedAssetBase.DEFAULT_NAMESPACE'}! !

NamedAssetBase subclass: #AnimationNodeBase!
AnimationNodeBase buildGetters: #( assetType )!
AnimationNodeBase buildLuaFunctions: #( dispose )!
AnimationNodeBase subclass: #ParticleNodeBase!

ParticleNodeBase buildGetters: #( dataLength mode oneData priority )!
ParticleNodeBase buildLuaFunctions: #(
	#(getAGALFragmentCode #(_pass _animationRegisterCache))
	#(getAGALUVCode #(_pass _animationRegisterCache))
	#(getAGALVertexCode #(_pass _animationRegisterCache))
	#(getAnimationState #(_animator) ParticleNodeState)
)!

ParticleNodeBase subclass: #ParticleColorNode!
!Nodes class methodsFor: #accessing !
particleColorNode
	" Returns a wrapper for library "

	^ParticleColorNode! !
! ParticleColorNode class methodsFor: #instantiation !
mode:mode usesMultiplier: usesMultiplier
	usesOffset: usesOffset usesCycle: usesCycle usesPhase: usesPhase
	startColor: startColor endColor: endColor
	cycleDuration: cycleDuration cyclePhase: cyclePhase
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Animators.Nodes.ParticleColorNode.new(mode, usesMultiplier,
		 usesOffset, usesCycle, usesPhase,
		startColor:handle(), endColor:handle(),
		cycleDuration, cyclePhase))'}! !
ParticleColorNode buildLuaFunctions: #(
	#(getAnimationState #(_animator) ParticleColorState)
)!

ParticleNodeBase subclass: #ParticleBillboardNode !
!Nodes class methodsFor: #accessing !
particleBillboardNode
	" Returns a wrapper for library "

	^ParticleBillboardNode! !
! ParticleBillboardNode class methodsFor: #instantiation !
new: billboardAxis
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Animators.Nodes.ParticleBillboardNode.new(billboardAxis))'}! !
! ParticleBillboardNode methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Animators.Nodes.ParticleBillboardNode.new(nil)'}! !
ParticleBillboardNode buildLuaFunctions: #(
	#(getAnimationState #(_animator) ParticleBillboardState)
)!

ParticleNodeBase subclass: #ParticleScaleNode !
!Nodes class methodsFor: #accessing !
particleScaleNode
	" Returns a wrapper for library "

	^ParticleScaleNode! !
! ParticleScaleNode class methodsFor: #instantiation !
mode: mode usesCycle: usesCycle usesPhase: usesPhase
	minScale: minScale maxScale: maxScale
	cycleDuration: cycleDuration cyclePhase: cyclePhase
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Animators.Nodes.ParticleScaleNode.new(mode, usesCycle, usesPhase, minScale, maxScale, cycleDuration, cyclePhase))'}! !
ParticleScaleNode buildLuaFunctions: #(
	#(getAnimationState #(_animator) ParticleScaleState)
)!

ParticleNodeBase subclass: #ParticleVelocityNode !
!Nodes class methodsFor: #accessing !
particleVelocityNode
	" Returns a wrapper for library "

	^ParticleVelocityNode! !
! ParticleVelocityNode class methodsFor: #instantiation !
mode: mode velocity: velocity
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Animators.Nodes.ParticleVelocityNode.new(mode, velocity:handle()))'}! !
ParticleVelocityNode buildLuaFunctions: #(
	#(getAnimationState #(_animator) ParticleVelocityState)
)!
! ParticleVelocityNode class methodsFor: #constants !
velocityVector3D

	^{'Lib.Away3D.Animators.Nodes.ParticleVelocityNode.VELOCITY_VECTOR3D'}! !

NamedAssetBase subclass: #AnimationSetBase !
AnimationSetBase buildGetters: #( animationNames animations assetType usesCPU )!
AnimationSetBase buildTypedGetters: #( #(animations 'Array[AnimationNodeBase]') )!
AnimationSetBase buildLuaFunctions: #(
	#(addAnimation #(_node))
	cancelGPUCompatibility
	dispose
	#(getAnimation #(name))
	#(hasAnimation #(name))
)!

AnimationSetBase subclass: #ParticleAnimationSet !
! Animators class methodsFor: #accessing !
particleAnimationSet
	" Returns a wrapper for library "

	^ParticleAnimationSet! !
! ParticleAnimationSet class methodsFor: #instantiation !
usesDuration: usesDuration usesLooping: usesLooping usesDelay: usesDelay
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Animators.ParticleAnimationSet.new(usesDuration, usesLooping, usesDelay))'}! !
ParticleAnimationSet buildAccessors: #(
	hasBillboard
	hasColorAddNode
	hasColorMulNode
	hasUVNode
	initParticleFunc
	needVelocity
)!
ParticleAnimationSet buildGetters: #( particleNodes )!
ParticleAnimationSet buildLuaFunctions: #(
	#(activate #(_stage3DProxy _pass))
	#(deactivate #(_stage3DProxy _pass))
	#(doneAGALCode #(_pass))
	#(getAGALFragmentCode #(_pass shadedTarget profile))
	#(getAGALUVCode #(_pass uvSource uvTarget))
	#(getAGALVertexCode #(_pass sourceRegisters targetRegisters profile))
)!

NamedAssetBase subclass: #AnimatorBase !
AnimatorBase buildGetters: #( absoluteTime activeAnimation activeAnimationName assetType )!
AnimatorBase buildAccessors: #( autoUpdate playbackSpeed time updatePosition )!
AnimatorBase buildTypedGetters: #( #(activeState AnimationNodeBase) )!
AnimatorBase buildLuaFunctions: #(
	dispose
	#(getAnimationState #(_node))
	#(getAnimationStateByName #(name))
	#(phase #(value))
	#(reset #(name offset))
	start stop
	#(update #(time))
)!

NamedAssetBase subclass: #Object3D!
Object3D buildTypedGetters: #(
	#(backVector Vector3D)
	#(downVector Vector3D)
	#(eulers Vector3D)
	#(forwardVector Vector3D)
	#(leftVector Vector3D)
	#(rightVector Vector3D)
	#(upVector Vector3D)
)!
Object3D buildTypedAccessors: #(
	extra
	#(pivotPoint Vector3D)
	#(position Vector3D)
	rotationX rotationY	rotationZ
	scaleX scaleY scaleZ
	#(transform Matrix3D)
	x y z zOffset
)!
Object3D buildLuaFunctions: #(
	#(clone self)
	dispose disposeAsset
	#(lookAt #(_target _upAxis))	
	#(moveBackward #(distance))
	#(moveDown #(distance))
	#(moveForward #(distance))
	#(moveLeft #(distance))
	#(movePivot #(dx dy dz))
	#(moveRight #(distance))
	#(moveTo #(dx dy dz))
	#(moveUp #(distance))
	#(pitch #(angle))
	#(roll #(angle))
	#(rotate #(axis angle))
	#(rotateTo #(ax ay az))
	#(scale #(value))
	#(translate #(_axis distance))
	#(translateLocal #(_axis distance))
	#(yaw #(angle))
)!

! Object3D methodsFor: #rotation !
rotateX: delta

	^self rotationX: self rotationX + delta! !

! Object3D methodsFor: #rotation !
rotateY: delta

	^self rotationY: self rotationY + delta! !

! Object3D methodsFor: #rotation !
rotateZ: delta

	^self rotationZ: self rotationZ + delta! !

! Object3D methodsFor: #accessing !
lookAt: target

	^self lookAt: target upAxis: nil! !

! Object3D methodsFor: #accessing !
x: x y: y z: z
	" Set the receiver's position "

	self x: x; y: y; z: z! !

Object3D subclass: #ObjectContainer3D !
! Containers class methodsFor: #accessing !
objectContainer3D
	" Returns a wrapper for library "

	^ObjectContainer3D! !
! ObjectContainer3D methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Containers.ObjectContainer3D.new()'}! !
ObjectContainer3D buildAccessors: #(
	ignoreTransform mouseChildren mouseEnabled
)!
ObjectContainer3D buildTypedGetters: #(
	assetType
	inverseSceneTransform
	maxX maxY maxZ
	minX minY minZ
	numChildren
	#(parent ObjectContainer3D)
	#(partition Partition3D)
	#(scene Scene3D)
	#(scenePosition Vector3D)
	#(sceneTransform Matrix3D)
	visible
)!
ObjectContainer3D buildLuaFunctions: #(
	#(addChild #(_child))
	#(addChildren #(#(array handles)))
	#(contains #(_child))
	disposeWithChildren
	#(getChildAt #(index))
	#(removeChild #(_child))
	#(removeChildAt #(index))
	updateImplicitVisibility
)!

ObjectContainer3D subclass: #Entity !
Entity buildTypedAccessors: #(
	#(bounds BoundingVolumeBase)
	pickingCollider shaderPickingDetails
	showBounds staticNode
)!
Entity buildTypedGetters: #(
	#(pickingCollisionVO PickingCollisionVO)
	#(worldBounds BoundingVolumeBase)
)!
Entity buildLuaFunctions: #(
	#(getEntityPartitionNode EntityNode)
	#(isIntersectingRay #(_rayPosition _rayDirection))
)!

Object3D subclass: #Camera3D !
! Cameras class methodsFor: #accessing !
camera3D
	" Returns a wrapper for library "

	^Camera3D! !
! Camera3D class methodsFor: #instantiation !
new: lens
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Cameras.Camera3D.new(lens))'}! !
! Camera3D methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Cameras.Camera3D.new(nil)'}! !
Camera3D buildTypedGetters: #(
	#(frustumPlanes 'Array[Plane3D]')
	#(viewProjection Matrix3D)
)!
Camera3D buildTypedAccessors: #( #(lens LensBase) )!
Camera3D buildLuaFunctions: #(
	#(getRay #(x y z) Vector3D)
	#(project #(_point3d) Vector3D)
	#(unproject #(x y z) Vector3D)
)!

Cameras modules: #( Lenses )!
EventDispatcher subclass: #LensBase !
LensBase buildTypedAccessors: #(
	#(frustumCorners Array)
	#(matrix Matrix3D)
	far near
)!
LensBase buildTypedGetters: #(
	#(unprojectionMatrix Matrix3D)
)!
LensBase buildLuaFunctions: #(
	#(clone self)
	#(project #(_point3d) Vector3D)
	#(unproject #(x y z) Vector3D)
)!

LensBase subclass: #PerspectiveLens !
! Lenses class methodsFor: #accessing !
perspectiveLens
	" Returns a wrapper for library "

	^PerspectiveLens! !
! PerspectiveLens class methodsFor: #instantiation !
new: fieldOfView
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Cameras.Lenses.PerspectiveLens.new(fieldOfView))'}! !
PerspectiveLens buildAccessors: #( fieldOfView focalLength )!

Entity subclass: #LightBase !
LightBase buildAccessors: #(
	ambient ambientColor
	castsShadows
	color diffuse
	shadowMapper
	specular
)!

EventDispatcher subclass: #Scene3D!
! Containers class methodsFor: #accessing !
scene3D
	" Returns a wrapper for library "

	^Scene3D! !
! Scene3D methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Containers.Scene3D.new()'}! !
Scene3D buildTypedAccessors: #(
	#(partition Partition3D)
)!
Scene3D buildGetters: #( numChildren )!
Scene3D buildLuaFunctions: #(
	#(addChild #(_child) ObjectContainer3D)
	#(contains #(_child))
	#(getChildAt #(index) ObjectContainer3D)
	#(removeChild #(_child))
	#(removeChildAt #(index))
	#(traversePartitions #(_traverser))
)!

Sprite
	subclass: #View3D
	instanceVariableNames: ''
	classVariableNames: #PreRenderActions
	poolDictionaries: ''!
! View3D class methodsFor: #rendering !
preRenderActions
	" Returns the actions to evaluate just before rendering instances of the receiver. "

	PreRenderActions isNil ifTrue: [ PreRenderActions := #() ].
	^PreRenderActions! !

! View3D class methodsFor: #rendering !
aboutToRender: aViewOrNil
	" Private - About to render aView. "

	self preRenderActions do: [:action| action evaluateWith: aViewOrNil ].! !

! Containers class methodsFor: #accessing !
view3D
	" Returns a wrapper for library "

	^View3D! !

! View3D methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Containers.View3D.new()'}! !

View3D buildTypedAccessors: #(
	antiAlias
	#(background Texture2DBase)
	backgroundAlpha
	backgroundColor
	#(camera Camera3D)
	deltaTime
	depthPrepass
	#(filters3d Array)
	forceMouseMove
	layeredView
	#(mousePicker IPicker)
	renderedFacesCount
	#(renderer RendererBase)
	rightClickMenuEnabled
	#(scene Scene3D)
	#(shareContext Context3D)
	#(stage3DProxy Stage3DProxy)
	#(touchPicker IPicker)
	#(transform Transform)
)!
View3D buildSetters: #(
	rotationX rotationY rotationZ
	scaleX scaleY scaleZ
	visible width height
	x y z
)!
View3D buildLuaFunctions: #(
	"render See custom implementation below"
	#(addSourceURL #(url))
	dispose
	#(getRay #(x y z) Vector3D)
	#(project #(_point3d) Vector3D)
	#(unproject #(x y z) Vector3D)
	#(setRenderCallback #(aBlock))
)!

! View3D methodsFor: #rendering !
basicRender
	" Private - Render the receiver. "

	^self handle #_render! !

! View3D methodsFor: #rendering !
render
	" Private - Render the receiver. "

	self class aboutToRender: self.
	^self basicRender! !

LuaStudioLibrary subclass: #ControllerBase!
ControllerBase buildTypedAccessors: #(
	autoUpdate #(targetObject Entity)
)!
ControllerBase buildLuaFunctions: #(
	#(update #(interpolate))
)!

ControllerBase subclass: #HoverController!
! Controllers class methodsFor: #accessing !
hoverController
	" Returns a wrapper for library "

	^HoverController! !
! HoverController class methodsFor: #instantiation !
target: targetObject lookAt: lookAtObject
	panAngle: panAngle tiltAngle: tiltAngle distance: distance
	minTiltAngle: minTiltAngle maxTiltAngle: maxTiltAngle
	minPanAngle: minPanAngle maxPanAngle: maxPanAngle
	steps: steps yFactor: yFactor
	wrapPanAngle: wrapPanAngle
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Controllers.HoverController.new(targetObject:handle(), lookAtObject:handle(),
		panAngle, tiltAngle, distance,
		minTiltAngle, maxTiltAngle, minPanAngle, maxPanAngle,
		steps, yFactor, wrapPanAngle))'}! !
HoverController buildAccessors: #(
	distance
	maxPanAngle maxTiltAngle
	minPanAngle minTiltAngle
	panAngle steps
	tiltAngle wrapPanAngle
	yFactor
)!

ControllerBase subclass: #LookAtController!
! Controllers class methodsFor: #accessing !
lookAtController
	" Returns a wrapper for library "

	^LookAtController! !
! LookAtController class methodsFor: #instantiation !
target: targetObject lookAt: lookAtObject
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Controllers.LookAtController.new(targetObject, lookAtObject))'}! !
LookAtController buildTypedAccessors: #(
	#(targetObject ObjectContainer3D)
	#(lookAtObject Vector3D)
)!

Core modules: #( Data Managers Pick Render Sort Traverse )!
Data modules: #( EntityListItem RenderableListItem )!
Managers modules: #( Mouse3DManager Touch3DManager )!

! EntityListItem methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Core.Data.EntityListItem.new()'}! !
EntityListItem buildTypedAccessors: #(
	#(entity Entity)
	#(next EntityListItem)
)!

! RenderableListItem methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Core.Data.RenderableListItem.new()'}! !
RenderableListItem buildTypedAccessors: #(
	cascaded materialId
	#(next RenderableListItem)
	#(renderable IRenderable)
	renderOrderId
	#(renderSceneTransform Matrix3D)
	zIndex
)!

! Mouse3DManager methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Core.Managers.Mouse3DManager.new()'}! !
Mouse3DManager buildTypedAccessors: #(
	forceMouseMove
	#(mousePicker IPicker)
)!
Mouse3DManager buildLuaFunctions: #(
	#(addViewLayer #(_view))
	#(disableMouseListeners #(_view))
	dispose
	#(enableMouseListeners #(_view))
	fireMouseEvents
	#(updateCollider #(_view))
)!

! Touch3DManager methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Core.Managers.Touch3DManager.new()'}! !
Touch3DManager buildTypedAccessors: #(
	forceTouchMove
	#(touchPicker IPicker)
)!
Touch3DManager buildLuaFunctions: #(
	#(disableTouchListeners #(_view))
	dispose
	#(enableTouchListeners #(_view))
	fireTouchEvents
	#(updateCollider #(_view))
)!

Pick modules: #( PickingCollisionVO RaycastPicker ShaderPicker )!

! PickingCollisionVO class methodsFor: #instantiation !
new: entity
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Core.Pick.PickingCollisionVO.new(entity:handle()))'}! !

PickingCollisionVO buildTypedAccessors: #(
	#(entity Entity) index
	#(localNormal Vector3D)
	#(localPosition Vector3D)
	#(localRayDirection Vector3D)
	#(localRayPosition Vector3D)
	#(rayDirection Vector3D)
	rayEntryDistance
	rayOriginIsInsideBounds
	#(rayPosition Vector3D)
	#(renderable IRenderable)
	subGeometryIndex
	#(uv APoint)
)!

! RaycastPicker class methodsFor: #instantiation !
new: findClosestCollision
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Core.Pick.RaycastPicker.new(findClosestCollision))'}! !

RaycastPicker buildAccessors: #( onlyMouseEnabled )!
RaycastPicker buildLuaFunctions: #(
	dispose
	#(getEntityCollision #(_position _direction #(entities handles)) PickingCollisionVO)
	#(getSceneCollision #(_position _direction _scene) PickingCollisionVO)
	#(getViewCollision #(x y _view) PickingCollisionVO)
	#(setIgnoreList #(#(entities handles)))
)!

! ShaderPicker methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Core.Pick.ShaderPicker.new()'}! !

ShaderPicker buildAccessors: #( onlyMouseEnabled )!
ShaderPicker buildLuaFunctions: #(
	dispose
	#(getSceneCollision #(_position _direction _scene) PickingCollisionVO)
	#(getViewCollision #(x y _view) PickingCollisionVO)
)!

Render modules: #( BackgroundImageRenderer )!

! BackgroundImageRenderer class methodsFor: #instantiation !
new: stage3DProxy
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Core.Render.BackgroundImageRenderer.new(stage3DProxy:handle()))'}! !

BackgroundImageRenderer buildTypedAccessors: #(
	#(stage3DProxy Stage3DProxy)
	#(texture Texture2DBase)
)!
BackgroundImageRenderer buildLuaFunctions: #(
	dispose render
)!

Sort modules: #( RenderableMergeSort )!

! RenderableMergeSort methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Core.Sort.RenderableSorter.new()'}! !

RenderableMergeSort buildLuaFunctions: #(
	#(sort #(_collector))
)!

LuaStudioLibrary subclass: #PartitionTraverser!
PartitionTraverser buildTypedAccessors: #( #(scene Scene3D) )!
PartitionTraverser buildTypedGetters: #( #(entryPoint Vector3D) )!
PartitionTraverser buildLuaFunctions: #(
	#(applyDirectionalLight #(_light))
	#(applyEntity #(_entity))
	#(applyLightProbe #(_light))
	#(applyPointLight #(_light))
	#(applyRenderable #(_renderable))
	#(applySkyBox #(_renderable))
	#(applyUnknownLight #(_light))
	#(enterNode #(_node))
)!

PartitionTraverser subclass: #EntityCollector!

!Traverse class methodsFor: #accesing!
entityCollector
	" Returns a wrapper for library "

	^EntityCollector ! !

! EntityCollector methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Core.Traverse.EntityCollector.new()'}! !

EntityCollector buildTypedAccessors: #(
	#(blendedRenderableHead RenderableListItem)
	#(camera Camera3D)
	#(cullPlanes 'Array[Plane3D]')
)!
EntityCollector buildTypedGetters: #(
	#(directionalLights 'Array[DirectionalLight]')
	#(entityHead EntityListItem)
	#(lightProbes 'Array[LightProbe]')
	#(lights 'Array[LightBase]')
	numMouseEnableds
	numTriangles
	#(opaqueRenderableHead RenderableListItem)
	#(pointLights 'Array[PointLight]')
	#(skyBox IRenderable)
)!
EntityCollector buildLuaFunctions: #( cleanUp clear )!

Sprite subclass: #AwayFPS!
!Debug class methodsFor: #accesing!
awayFPS
	" Returns a wrapper for library "

	^AwayFPS ! !
! AwayFPS class methodsFor: #instantiation !
view:view xOff: xOff yOff: yOff color: color scale: scale
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Debug.AwayFPS.new(view:handle(), xOff, yOff, color, scale))'}! !
! AwayFPS class methodsFor: #instantiation !
xOff: xOff yOff: yOff color: color scale: scale
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Debug.AwayFPS.new(nil, xOff, yOff, color, scale))'}! !


Entity subclass: #Mesh!
!Entities class methodsFor: #accesing!
mesh
	" Returns a wrapper for library "

	^Mesh ! !
! Mesh class methodsFor: #instantiation !
geometry: geometry material: material
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Entities.Mesh.new(geometry:handle(), material:handle()))'}! !
Mesh buildTypedAccessors: #(
	#(animator AnimatorBase)
	castsShadows
	#(geometry Geometry)
	#(material MaterialBase)
	shareAnimationGeometry
)!
Mesh buildTypedGetters: #(
	#(subMeshes 'Array[SubMesh]')
)!
Mesh buildLuaFunctions: #(
	bakeTransformations
	clearAnimationGeometry
	disposeWithAnimatorAndChildren
	#(getSubMeshForSubGeometry #(_subGeometry) SubMesh)
)!

Event subclass: #AnimatorEvent!
! AnimatorEvent class methodsFor: #instantiation !
type: type animator: animator
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Events.AnimatorEvent.new(type, animator:handle()))'}! !
AnimatorEvent buildTypedGetters: #(
	#(animator AnimatorBase)
)!
AnimatorEvent buildConstants: '
	Lib.Away3D.Events.AnimatorEvent.CYCLE_COMPLETE
	Lib.Away3D.Events.AnimatorEvent.START
	Lib.Away3D.Events.AnimatorEvent.STOP
'!

Event subclass: #Asset3DEvent!
! Asset3DEvent class methodsFor: #instantiation !
type: type asset: asset prevName: prevName
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Events.Asset3DEvent.new(type, asset:handle(), prevName))'}! !
Asset3DEvent buildTypedGetters: #(
	#(asset IAsset)
	assetPrevName
)!
Asset3DEvent buildConstants: '
	Lib.Away3D.Events.Asset3DEvent.ANIMATION_NODE_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.ANIMATION_SET_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.ANIMATION_STATE_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.ANIMATOR_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.ASSET_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.ASSET_CONFLICT_RESOLVED
	Lib.Away3D.Events.Asset3DEvent.ASSET_RENAME
	Lib.Away3D.Events.Asset3DEvent.CAMERA_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.CONTAINER_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.EFFECTMETHOD_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.ENTITY_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.GEOMETRY_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.LIGHT_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.LIGHTPICKER_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.MATERIAL_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.MESH_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.SEGMENT_SET_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.SHADOWMAPMETHOD_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.SKELETON_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.SKELETON_POSE_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.SKYBOX_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.STATE_TRANSITION_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.TEXTURE_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.TEXTURE_PROJECTOR_COMPLETE
	Lib.Away3D.Events.Asset3DEvent.TEXTURE_SIZE_ERROR
'!

Event subclass: #MouseEvent3D!
! MouseEvent3D class methodsFor: #instantiation !
type: type
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Events.Mouse3DEvent.new(type))'}! !
MouseEvent3D buildTypedGetters: #(
	altKey ctrlKey
	delta index
	#(localNormal Vector3D)
	#(localPosition Vector3D)
	#(material MaterialBase)
	#(object ObjectContainer3D)
	#(renderable IRenderable)
	#(sceneNormal Vector3D)
	#(scenePosition Vector3D)
	screenX screenY
	shiftKey
	subGeometryIndex
	#(uv APoint)
	#(view View3D)
)!
MouseEvent3D buildConstants: '
	Lib.Away3D.Events.Mouse3DEvent.CLICK
	Lib.Away3D.Events.Mouse3DEvent.DOUBLE_CLICK
	Lib.Away3D.Events.Mouse3DEvent.MOUSE_DOWN
	Lib.Away3D.Events.Mouse3DEvent.MOUSE_MOVE
	Lib.Away3D.Events.Mouse3DEvent.MOUSE_OUT
	Lib.Away3D.Events.Mouse3DEvent.MOUSE_OVER
	Lib.Away3D.Events.Mouse3DEvent.MOUSE_UP
	Lib.Away3D.Events.Mouse3DEvent.MOUSE_WHEEL
'!

Event subclass: #Object3DEvent!
! Object3DEvent class methodsFor: #instantiation !
type: type object: object
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Events.Object3DEvent.new(type, object:handle()))'}! !
Object3DEvent buildTypedAccessors: #(
	#(object Object3D)
)!
Object3DEvent buildConstants: '
	Lib.Away3D.Events.Object3DEvent.POSITION_CHANGED
	Lib.Away3D.Events.Object3DEvent.ROTATION_CHANGED
	Lib.Away3D.Events.Object3DEvent.SCALE_CHANGED
	Lib.Away3D.Events.Object3DEvent.SCENE_CHANGED
	Lib.Away3D.Events.Object3DEvent.SCENETRANSFORM_CHANGED
	Lib.Away3D.Events.Object3DEvent.VISIBLITY_UPDATED
'!

LuaStudioLibrary subclass: #Asset3DType!
Asset3DType buildConstants: '
	Lib.Away3D.Library.Assets.Asset3DType.ANIMATION_NODE
	Lib.Away3D.Library.Assets.Asset3DType.ANIMATION_SET
	Lib.Away3D.Library.Assets.Asset3DType.ANIMATION_STATE
	Lib.Away3D.Library.Assets.Asset3DType.ANIMATOR
	Lib.Away3D.Library.Assets.Asset3DType.CAMERA
	Lib.Away3D.Library.Assets.Asset3DType.CONTAINER
	Lib.Away3D.Library.Assets.Asset3DType.EFFECTS_METHOD
	Lib.Away3D.Library.Assets.Asset3DType.ENTITY
	Lib.Away3D.Library.Assets.Asset3DType.GEOMETRY
	Lib.Away3D.Library.Assets.Asset3DType.LIGHT
	Lib.Away3D.Library.Assets.Asset3DType.LIGHT_PICKER
	Lib.Away3D.Library.Assets.Asset3DType.MATERIAL
	Lib.Away3D.Library.Assets.Asset3DType.MESH
	Lib.Away3D.Library.Assets.Asset3DType.SEGMENT_SET
	Lib.Away3D.Library.Assets.Asset3DType.SHADOW_MAP_METHOD
	Lib.Away3D.Library.Assets.Asset3DType.SKELETON
	Lib.Away3D.Library.Assets.Asset3DType.SKELETON_POSE
	Lib.Away3D.Library.Assets.Asset3DType.SKYBOX
	Lib.Away3D.Library.Assets.Asset3DType.STATE_TRANSITION
	Lib.Away3D.Library.Assets.Asset3DType.TEXTURE
	Lib.Away3D.Library.Assets.Asset3DType.TEXTURE_PROJECTOR
'!

LightBase subclass: #DirectionalLight !
DirectionalLight buildTypedAccessors: #(
	#(direction Vector3D)
)!
DirectionalLight buildTypedGetters: #(
	#(sceneDirection Vector3D)
)!
! DirectionalLight class methodsFor: #instantiation !
xDir: xDir yDir: yDir zDir: zDir
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Lights.DirectionalLight.new(xDir, yDir, zDir))'}! !

LightBase subclass: #LightProbe !
LightProbe buildTypedAccessors: #(
	#(diffuseMap CubeTextureBase)
	#(specularMap CubeTextureBase)
)!
! LightProbe class methodsFor: #instantiation !
diffuseMap: diffuseMap specularMap: specularMap
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Lights.LightProbe.new(diffuseMap:handle(), specularMap:handle()))'}! !

LightBase subclass: #PointLight !
PointLight buildAccessors: #( fallOff radius )!
! PointLight methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Lights.PointLight.new()'}! !

NamedAssetBase subclass: #MaterialBase !
MaterialBase buildTypedAccessors: #(
	alphaPremultiplied blendMode
	bothSides depthCompareMode
	extra
	#(lightPicker LightPickerBase)
	mipmap repeat requiresBlending smooth
)!
MaterialBase buildGetters: #( assetType uniqueId )!
MaterialBase buildLuaFunctions: #( dispose )!

MaterialBase subclass: #SinglePassMaterialBase !
SinglePassMaterialBase buildGetters: #( numMethods )!
SinglePassMaterialBase buildTypedAccessors: #(
	alphaBlending alphaThreshold
	ambient ambientColor
	#(ambientMethod BasicAmbientMethod)
	#(colorTransform ColorTransform)
	diffuseLightSources
	#(diffuseMethod BasicDiffuseMethod)
	enableLightFallOff gloss
	#(normalMap Texture2DBase)
	#(normalMethod BasicNormalMethod)
	shadowMethod
	specular specularColor specularLightSources
	#(specularMap Texture2DBase)
	#(specularMethod BasicSpecularMethod)
)!
SinglePassMaterialBase buildLuaFunctions: #(
	#(addMethod #(_method))
	#(addMethodAt #(_method index))
	#(getMethodAt #(index))
	#(hasMethod #(_method))
	#(removeMethod #(_method))
)!

SinglePassMaterialBase subclass: #ColorMaterial !
! Materials class methodsFor: #accesing !
colorMaterial
	" Returns a wrapper for library "

	^ColorMaterial ! !
! ColorMaterial class methodsFor: #instantiation !
color: color alpha: alpha
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Materials.ColorMaterial.new(color, alpha))'}! !
ColorMaterial buildAccessors: #( alpha color )!

Materials modules: #( LightPickers )!

NamedAssetBase subclass: #LightPickerBase !
LightPickerBase buildTypedGetters: #(
	#(allPickedLights 'Array[LightBase]')
	assetType
	#(castingDirectionalLights 'Array[DirectionalLight]')
	#(castingPointLights 'Array[PointLight]')
	#(directionalLights 'Array[DirectionalLight]')
	#(lightProbes 'Array[LightProbe]')
	#(lightProbeWeights Array)
	numCastingDirectionalLights numCastingPointLights
	numDirectionalLights numLightProbes numPointLights
	#(pointLights 'Array[PointLight]')
)!
LightPickerBase buildLuaFunctions: #(
	#(collectLights #(_renderable _entityCollector))
	dispose
)!

LightPickerBase subclass: #StaticLightPicker !
! StaticLightPicker class methodsFor: #instantiation !
lights: lights
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Materials.LightPickers.StaticLightPicker.new(lights:handles()))'}! !
StaticLightPicker buildTypedGetters: #(
	#(lights 'Array[LightBase]')
)!
! StaticLightPicker methodsFor: #accessing !
lights: array
	" Set the lights of the receiver "

	self handle basicAt: #lights put: array handles! !

NamedAssetBase subclass: #ShadingMethodBase !
ShadingMethodBase buildTypedGetters: #(
	#(passes MaterialPassBase)
)!
ShadingMethodBase buildLuaFunctions: #(
	#(copyFrom #(_method))
	dispose
)!

ShadingMethodBase subclass: #BasicAmbientMethod !
BasicAmbientMethod buildTypedAccessors: #(
	ambient ambientColor
	#(texture Texture2DBase)
)!

ShadingMethodBase subclass: #BasicNormalMethod !
BasicNormalMethod buildTypedAccessors: #(
	#(normalMap Texture2DBase)
)!

ShadingMethodBase subclass: #EffectMethodBase !
EffectMethodBase buildGetters: #( assetType )!

EffectMethodBase subclass: #EnvMapMethod !
! EnvMapMethod class methodsFor: #instantiation !
envMap: envMap alpha: alpha
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Materials.Methods.EnvMapMethod.new(envMap:handle(), alpha))'}! !
EnvMapMethod buildTypedAccessors: #(
	alpha
	#(envMap CubeTextureBase)
	#(mask Texture2DBase)
)!

SinglePassMaterialBase subclass: #TextureMaterial !
! TextureMaterial class methodsFor: #instantiation !
texture: texture smooth: smooth repeat: count mipmap: mipmap anisotropy: anisotropy
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Materials.TextureMaterial.new(texture:handle(), smooth, count, mipmap, anisotropy))'}! !
! TextureMaterial class methodsFor: #instantiation !
texture: texture smooth: smooth repeat: count mipmap: mipmap
	" Returns an instance of the receiver "

	^self texture: texture smooth: smooth repeat: count mipmap: mipmap anisotropy: nil! !
! TextureMaterial class methodsFor: #instantiation !
texture: texture
	" Returns an instance of the receiver "

	^self texture: texture smooth: true repeat: false mipmap: true anisotropy: nil! !

TextureMaterial buildTypedAccessors: #(
	anisotropy alpha animateUVs
	#(ambientTexture Texture2DBase)
	#(texture Texture2DBase)
)!

MaterialBase subclass: #MultiPassMaterialBase !
MultiPassMaterialBase buildGetters: #( numMethods )!
MultiPassMaterialBase buildTypedAccessors: #(
	alphaThreshold ambient ambientColor
	#(ambientMethod BasicAmbientMethod)
	diffuseLightSources
	#(diffuseMethod BasicDiffuseMethod)
	enableLightFallOff gloss
	#(normalMap Texture2DBase)
	#(normalMethod BasicNormalMethod)
	#(shadowMethod ShadowMapMethodBase)
	specular specularColor specularLightSources
	#(specularMap Texture2DBase)
	#(specularMethod BasicSpecularMethod)
)!
MultiPassMaterialBase buildLuaFunctions: #(
	#(addMethod #(_method))
	#(addMethodAt #(_method index))
	#(getMethodAt #(index))
	#(hasMethod #(_method))
	#(removeMethod #(_method))
)!

MultiPassMaterialBase subclass: #TextureMultiPassMaterial !
! TextureMultiPassMaterial class methodsFor: #instantiation !
texture: texture smooth: smooth repeat: count mipmap: mipmap anisotropy: anisotropy
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Materials.TextureMultiPassMaterial.new(texture:handle(), smooth, count, mipmap, anisotropy))'}! !
TextureMultiPassMaterial buildTypedAccessors: #(
	#(ambientTexture Texture2DBase)
	animateUVs
	#(texture Texture2DBase)
)!

NamedAssetBase subclass: #Geometry !
Geometry buildTypedGetters: #(
	assetType
	#(subGeometries 'Array[SubGeometry]')
)!
Geometry buildLuaFunctions: #(
	#(addSubGeometry #(_subGeometry))
	#(applyTransformation #(_transform))
	#(clone self) dispose
	convertToSeparateBuffers
	#(removeSubGeometry #(_subGeometry))
	#(scale #(scale))
	#(scaleUV #(scaleU scaleV))
)!

Geometry subclass: #PrimitiveBase !
PrimitiveBase buildLuaFunctions: #(
	#(buildGeometry #(_target))
	#(buildUVs #(_target))
	invalidateGeometry
	invalidateUVs
)!

PrimitiveBase subclass: #CubeGeometry !
! Primitives class methodsFor: #accesing !
cubeGeometry
	" Returns a wrapper for library "

	^CubeGeometry ! !
! CubeGeometry class methodsFor: #instantiation !
width: width height: height depth: depth
	segmentsW: segmentsW segmentsH: segmentsH
	segmentsD: segmentsD tile6: tile6
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Primitives.CubeGeometry.new(width, height, depth, segmentsW, segmentsH, segmentsD, tile6))'}! !
CubeGeometry buildAccessors: #(
	depth height segmentsD segmentsH segmentsW tile6 width
)!

PrimitiveBase subclass: #PlaneGeometry !
! Primitives class methodsFor: #accesing !
planeGeometry
	" Returns a wrapper for library "

	^PlaneGeometry ! !
! PlaneGeometry class methodsFor: #instantiation !
width: width height: height
	segmentsW: segmentsW segmentsH: segmentsH
	yUp: yUp doubleSided: doubleSided
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Primitives.PlaneGeometry.new(
		width, height,
		segmentsW, segmentsH,
		yUp, doubleSided))'}! !
PlaneGeometry buildAccessors: #(
	depth height segmentsH segmentsW width doubleSided yUp
)!

Entity subclass: #SkyBox !
! Primitives class methodsFor: #accesing !
skyBox
	" Returns a wrapper for library "

	^SkyBox ! !
! SkyBox class methodsFor: #instantiation !
cubeMap: cubeMap
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Primitives.SkyBox.new(cubeMap:handle()))'}! !
SkyBox buildTypedGetters: #(
	#(animator AnimatorBase)
	assetType castsShadows
	#(indexData Array)
	#(material MaterialBase)
	numTriangles
	#(sourceEntity Entity)
	#(UVData Array)
	#(uvTransform Matrix3D)
	#(vertexData Array)
	#(vertexNormalData Array)
	vertexNormalOffset
	vertexOffset vertexStride
	#(vertexTangentData Array)
	vertexTangentOffset
)!
SkyBox buildLuaFunctions: #(
	#(activateSecondaryUVBuffer #(index _stage3DProxy))
	#(activateUVBuffer #(index _stage3DProxy))
	#(activateVertexBuffer #(index _stage3DProxy))
	#(activateVertexNormalBuffer #(index _stage3DProxy))
	#(activateVertexTangentBuffer #(index _stage3DProxy))
	#(getIndexBuffer #(_stage3DProxy) IndexBuffer3D)
	#(getRenderSceneTransform #(_camera) Matrix3D)
)!

PrimitiveBase subclass: #SphereGeometry !
! Primitives class methodsFor: #accesing !
sphereGeometry
	" Returns a wrapper for library "

	^SphereGeometry ! !
! SphereGeometry class methodsFor: #instantiation !
radius: radius segmentsW: segmentsW segmentsH: segmentsH yUp: yUp
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Primitives.SphereGeometry.new(radius, segmentsW, segmentsH, yUp))'}! !
SphereGeometry buildAccessors: #(
	radius segmentsH segmentsW yUp
)!

PrimitiveBase subclass: #TorusGeometry !
! Primitives class methodsFor: #accesing !
torusGeometry
	" Returns a wrapper for library "

	^TorusGeometry ! !
! TorusGeometry class methodsFor: #instantiation !
radius: radius tubeRadius: tubeRadius segmentsR: segmentsR segmentsT: segmentsT yUp: yUp
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Primitives.TorusGeometry.new(radius, tubeRadius, segmentsR, segmentsT, yUp))'}! !
TorusGeometry buildAccessors: #(
	radius segmentsR segmentsT tubeRadius yUp
)!

LuaStudioLibrary subclass: #Text!
Text modules: #( HAlign VAlign )!
HAlign buildConstants: '
	Lib.Away3D.Text.HAlign.LEFT
	Lib.Away3D.Text.HAlign.CENTER
	Lib.Away3D.Text.HAlign.RIGHT
'!
VAlign buildConstants: '
	Lib.Away3D.Text.VAlign.TOP
	Lib.Away3D.Text.VAlign.CENTER
	Lib.Away3D.Text.VAlign.BOTTOM
'!

LuaStudioLibrary subclass: #BitmapChar !
BitmapChar buildGetters: #(
	charID
	xOffset yOffset xAdvance
	width height x y
)!
BitmapChar buildLuaFunctions: #(
	#(addKerning #(charID amount))
	#(getKerning #(charID))
)!

LuaStudioLibrary subclass: #BitmapFont !
! BitmapFont class methodsFor: #instantiation !
texture: texture fontXml: fontXml
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Text.BitmapFont.new(texture:handle(), fontXml:handle()))'}! !
BitmapFont buildConstants: '
	Lib.Away3D.Text.BitmapFont.NATIVE_SIZE
	Lib.Away3D.Text.BitmapFont.MINI
'!
BitmapFont buildAccessors: #( name size )!
BitmapFont buildGetters: #(
	lineHeight baseline
	offsetX offsetY
	fontMaterial
)!
BitmapFont buildLuaFunctions: #(
	dispose
	#(getChar #(charID))
	#(addChar #(charID _bitmapChar))
	#(getCharIDs #(result))
	#(hasChars #(text))
	#(fillBatched #(data indices width height text fontSize hAlign vAlign autoScale kerning letterSpacing))
	#(getMaterialClone #() TextureMaterial)
)!

Mesh subclass: #TextField !
! TextField class methodsFor: #instantiation !
width: width height: height text: text
	bitmapFont: bitmapFont fontSize: fontSize
	color: color bold: bold hAlign: hAlign
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Text.TextField.new(width, height,
		text, bitmapFont:handle(),
		fontSize, color, bold,
		hAlign))'}! !
TextField buildAccessors: #(
	alpha bold italic underline kerning
	letterSpacing autoScale autoSize
)!
TextField buildTypedGetters: #(
	textHeight textWidth
	#(boundsRect ARectangle)
)!

Textures modules: #( Anisotropy )!
Anisotropy buildConstants: '
	Lib.Away3D.Textures.Anisotropy.NONE
	Lib.Away3D.Textures.Anisotropy.ANISOTROPIC2X
	Lib.Away3D.Textures.Anisotropy.ANISOTROPIC4X
	Lib.Away3D.Textures.Anisotropy.ANISOTROPIC8X
	Lib.Away3D.Textures.Anisotropy.ANISOTROPIC16X
'!

NamedAssetBase subclass: #TextureProxyBase !
TextureProxyBase buildAccessors: #( assetType format hasMipMaps height width )!
TextureProxyBase buildLuaFunctions: #(
	#(getTextureForStage3D #(_stage3DProxy) TextureBase)
	dispose invalidateContent
)!

TextureProxyBase subclass: #Texture2DBase !

Texture2DBase subclass: #BitmapTexture !
! Textures class methodsFor: #accesing !
bitmapTexture
	" Returns a wrapper for library "

	^BitmapTexture ! !
! BitmapTexture class methodsFor: #instantiation !
bitmapData: bitmapData generateMipmaps: generateMipmaps
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Textures.BitmapTexture.new(bitmapData:handle(), generateMipmaps))'}! !
BitmapTexture buildTypedAccessors: #( #(bitmapData BitmapData) )!

! BitmapTexture class methodsFor: #converting !
from: data
	" Returns an instance of the receiver "

	self withString: data do: [ ^self pathName: data ].
	^{'self:_at(Lib.Away3D.Utils.Cast.bitmapTexture(data:handle()))'}! !

! BitmapTexture class methodsFor: #instantiating !
@ data
	" Returns an instance of the receiver "

	self withString: data do: [ ^self pathName: data ].
	^super @ data! !

! BitmapTexture class methodsFor: #converting !
pathName: fileName
	" Returns an instance of the receiver "

	| result pathName |
	result := self withU8Path: fileName do: [
		self bitmapData: BitmapData @ fileName generateMipmaps: true
	].
	result notNil ifTrue: [ ^result ].
	pathName := self pathToFile: fileName.
	^{'self:_at(Lib.Away3D.Utils.Cast.bitmapTexture(pathName))'}! !

NamedAssetBase subclass: #TextureProxyBase !
TextureProxyBase buildGetters: #( assetType format hasMipMaps height width )!
TextureProxyBase buildLuaFunctions: #(
	#(getTextureForStage3D #(_stage3DProxy) TextureBase)
	invalidateContent dispose
)!

TextureProxyBase subclass: #CubeTextureBase !
CubeTextureBase buildGetters: #( size )!

Texture2DBase subclass: #BitmapCubeTexture !
! Textures class methodsFor: #accesing !
bitmapCubeTexture
	" Returns a wrapper for library "

	^BitmapCubeTexture ! !
! BitmapCubeTexture class methodsFor: #instantiation !
posX: posX negX: negX posY: posY negY: negY posZ: posZ negZ: negZ
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Textures.BitmapCubeTexture.new(posX:handle(),
		negX:handle(), posY:handle(), negY:handle(),
		posZ:handle(), negZ:handle()))'}! !

BitmapCubeTexture buildTypedAccessors: #(
	#(negativeX BitmapData)
	#(negativeY BitmapData)
	#(negativeZ BitmapData)
	#(positiveX BitmapData)
	#(positiveY BitmapData)
	#(positiveZ BitmapData)
)!

! BitmapCubeTexture class methodsFor: #instantiation !
@ anArray
	" Returns an instance of the receiver "

	| array |
	(self isObject: anArray) ifFalse: [ ^super @ anArray ].
	anArray size = 6 ifFalse: [ ^self error: 'Invalid size' ].
	array := anArray collect: [:each|
		each isString
		ifTrue: [ BitmapData @ each ]
		ifFalse: [ each ]
	].
	^self	posX: array[1] negX: array[2]
		posY: array[3] negY: array[4]
		posZ: array[5] negZ: array[6]! !

LuaStudioLibrary subclass: #ParticleGeometryTransform !
! ParticleGeometryTransform methodsFor: #initialize !
initializeHandle
	" Private - Initialize the handle of the receiver. "

	handle := {'Lib.Away3D.Tools.Helpers.Data.ParticleGeometryTransform.new()'}! !

ParticleGeometryTransform buildTypedAccessors: #(
	#(invVertexTransform Matrix3D)
)!
ParticleGeometryTransform buildTypedGetters: #(
	#(UVTransform Matrix)
	#(vertexTransform Matrix3D)
)!

LuaStudioLibrary subclass: #ParticleGeometryHelper !
! ParticleGeometryHelper class methodsFor: #instantiation !
geometries: geometries transforms: transforms
	" Returns an instance of the receiver "

	^{'smalltalk.ParticleGeometry:_at(Lib.Away3D.Tools.Helpers.ParticleGeometryHelper(geometries:handles(), transforms:handles()))'}! !
! ParticleGeometryHelper class methodsFor: #instantiation !
geometries: geometries
	" Returns an instance of the receiver "

	^{'smalltalk.ParticleGeometry:_at(Lib.Away3D.Tools.Helpers.ParticleGeometryHelper(geometries:handles(),nil))'}! !
ParticleGeometryHelper buildConstants: '
	Lib.Away3D.Tools.Helpers.ParticleGeometryHelper.MAX_VERTEX
'!

LuaStudioLibrary subclass: #AnimationStateBase !
AnimationStateBase buildTypedGetters: #(
	#(positionDelta Vector3D)
)!
AnimationStateBase buildLuaFunctions: #(
	#(offset #(startTime))
	#(phase #(value))
	#(update #(time))
)!
AnimationStateBase subclass: #ParticleStateBase !
ParticleStateBase buildGetters: #( needUpdateTime )!
ParticleStateBase buildLuaFunctions: #(
	#(setRenderState #(_stage3DProxy _renderable _animationSubGeometry _animationRegisterCache _camera))
)!
ParticleStateBase subclass: #ParticleBillboardState !
ParticleBillboardState buildTypedAccessors: #(
	#(billboardAxis Vector3D)
)!
ParticleStateBase subclass: #ParticleColorState !
ParticleColorState buildTypedAccessors: #(
	cycleDuration cyclePhase
	#(endColor ColorTransform)
	#(startColor ColorTransform)
)!
ParticleStateBase subclass: #ParticleScaleState !
ParticleScaleState buildAccessors: #(
	cycleDuration cyclePhase
	maxScale minScale
)!
ParticleStateBase subclass: #ParticleVelocityState !
ParticleVelocityState buildTypedAccessors: #(
	#(billboardAxis Vector3D)
)!
ParticleVelocityState buildLuaFunctions: #(
	#(getVelocities #() 'Array[Vector3D]')
	#(setVelocities #(#(value handles)))
)!

LuaStudioLibrary subclass: #AnimationSubGeometry !
AnimationSubGeometry buildTypedGetters: #(
	numProcessedVertices
	numVertices
	totalLenOfOneVertex
	#(vertexData Array)
)!
AnimationSubGeometry buildTypedAccessors: #(
	#(animationParticles 'Array[ParticleAnimationData]')
	previousTime
)!
AnimationSubGeometry buildLuaFunctions: #(
	#(activateVertexBuffer #(index bufferOffset _stage3DProxy format))
	#(createVertexData #(numVertices totalLenOfOneVertex))
	dispose
	invalidateBuffer
)!

LuaStudioLibrary subclass: #BoundingVolumeBase !
BoundingVolumeBase buildTypedGetters: #(
	#(aabbPoints Array)
	#(boundingRenderable WireframePrimitiveBase)
	#(max Vector3D)
	#(min Vector3D)
)!
BoundingVolumeBase buildLuaFunctions: #(
	#(classifyToPlane #(_plane))
	clone
	#(containsPoint #(_position))
	disposeRenderable
	#(fromExtremes #(minX minY minZ maxX maxY maxZ))
	#(fromGeometry #(_geometry))
	#(fromSphere #(_center radius))
	#(fromVertices #(#(vertices handles)))
	#(isInFrustum #(#(planes handles) numPlanes))
	nullify
	#(overlaps #(_bounds))
	#(rayIntersection #(_position _direction _targetNormal))
	#(transformFrom #(_bounds _matrix))
)!

EventDispatcher subclass: #MaterialPassBase !
MaterialPassBase buildTypedGetters: #(
	needFragmentAnimation
	needUVAnimation
	numUsedFragmentConstants
	numUsedStreams
	numUsedVaryings
	numUsedVertexConstants
	renderToTexture
)!
MaterialPassBase buildTypedAccessors: #(
	alphaPremultiplied
	#(animationRegisterCache AnimationRegisterCache)
	animationSet bothSides
	depthCompareMode
	#(material MaterialBase)
	mipmap repeat
	smooth writeDepth
)!
MaterialPassBase buildLuaFunctions: #(
	dispose
	#(setBlendMode #(value))
)!

Geometry subclass: #ParticleGeometry !
ParticleGeometry buildTypedAccessors: #(
	numParticles
	#(particles 'Array[ParticleData]')
)!

! ParticleGeometry class methodsFor: #instantiation !
geometries: geometries transforms: transforms
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Tools.Helpers.ParticleGeometryHelper.generateGeometry(geometries:handles(), transforms:handles()))'}! !
! ParticleGeometry class methodsFor: #instantiation !
geometries: geometries
	" Returns an instance of the receiver "

	^{'self:_at(Lib.Away3D.Tools.Helpers.ParticleGeometryHelper.generateGeometry(geometries:handles(),nil))'}! !

ShadingMethodBase subclass: #LightingMethodBase !
LightingMethodBase subclass: #BasicDiffuseMethod !
BasicDiffuseMethod buildTypedAccessors: #(
	alphaThreshold diffuseAlpha diffuseColor
	#(texture Texture2DBase)
)!
BasicDiffuseMethod buildLuaFunctions: #(
	#(generateMip #(_stage3DProxy))
)!

LightingMethodBase subclass: #BasicSpecularMethod !
BasicSpecularMethod buildTypedAccessors: #(
	gloss specular specularColor
	#(texture Texture2DBase)
)!

LuaStudioLibrary subclass: #NodeBase !
NodeBase buildTypedGetters: #(
	#(parent NodeBase)
)!
NodeBase buildAccessors: #( showDebugBounds )!
NodeBase buildLuaFunctions: #(
	#(acceptTraverser #(_traverser))
	#(findPartitionForEntity #(_entity))
	#(isInFrustum #(#(planes handles) numPlanes))
	#(isIntersectingRay #(_rayPosition _rayDirection))
)!

NodeBase subclass: #EntityNode !
EntityNode buildTypedGetters: #(
	#(entity Entity)
)!
EntityNode buildLuaFunctions: #( removeFromParent )!

LuaStudioLibrary subclass: #ParticleData !
ParticleData buildTypedAccessors: #(
	numVertices particleIndex startVertexIndex
	#(subGeometry CompactSubGeometry)
)!

LuaStudioLibrary subclass: #Partition3D !
Partition3D buildAccessors: #( showDebugBounds )!
Partition3D buildLuaFunctions: #(
	#(traverse #(_traverser))
)!

LuaStudioLibrary subclass: #PartitionTraverser !
PartitionTraverser buildTypedGetters: #(
	#(entryPoint Vector3D)
)!
PartitionTraverser buildTypedAccessors: #(
	#(scene Scene3D)
)!
PartitionTraverser buildLuaFunctions: #(
	#(applyDirectionalLight #(_light))
	#(applyEntity #(_light))
	#(applyLightProbe #(_light))
	#(applyPointLight #(_light))
	#(applyRenderable #(_renderable))
	#(applySkyBox #(_renderable))
	#(applyUnknownLight #(_light))
	#(enterNode #(_node))
)!

LuaStudioLibrary subclass: #Plane3D !
Plane3D buildAccessors: #( a b c d )!
Plane3D buildLuaFunctions: #(
	#(classifyPoint #(_p epsilon))
	#(distance #(_p))
	#(fromNormalAndPoint #(_normal _point))
	#(fromPoints #(_p0 _p1 _p2))
	normalize toString
)!

LuaStudioLibrary subclass: #RendererBase !
RendererBase buildTypedGetters: #(
	#(backgroundImageRenderer BackgroundImageRenderer)
)!
RendererBase buildTypedAccessors: #(
	antiAlias
	#(renderableSorter RenderableMergeSort)
)!
RendererBase buildLuaFunctions: #(
	#(queueSnapshot #(_bitmapData))
)!

LuaStudioLibrary subclass: #ShaderRegisterCache !
ShaderRegisterCache buildTypedGetters: #(
	#(fragmentOutputRegister ShaderRegisterElement)
	numUsedFragmentConstants numUsedStreams
	numUsedTextures numUsedVaryings
	numUsedVertexConstants
)!
ShaderRegisterCache buildAccessors: #(
	fragmentConstantOffset
	varyingsOffset
	vertexAttributesOffset
	vertexConstantOffset
)!
ShaderRegisterCache buildLuaFunctions: #(
	#(addFragmentTempUsages #(_register usageCount))
	#(addVertexTempUsages #(_register usageCount))
	dispose
	getFreeFragmentConstant
	getFreeFragmentSingleTemp
	getFreeFragmentVectorTemp
	getFreeTextureReg getFreeVarying
	getFreeVertexAttribute getFreeVertexConstant
	getFreeVertexSingleTemp getFreeVertexVectorTemp
	#(removeFragmentTempUsage #(_register))
	#(removeVertexTempUsage #(_register))
	reset
)!

ShaderRegisterCache subclass: #AnimationRegisterCache !
AnimationRegisterCache buildGetters: #(
	numFragmentConstant
	numVertexConstant
)!
AnimationRegisterCache buildTypedAccessors: #(
	#(colorAddTarget ShaderRegisterElement)
	#(colorAddVary ShaderRegisterElement)
	#(colorMulTarget ShaderRegisterElement)
	#(colorMulVary ShaderRegisterElement)
	#(fragmentConstantData Array)
	hasBillboard hasColorAddNode hasColorMulNode
	hasUVNode needFragmentAnimation
	needUVAnimation needVelocity
	#(positionAttribute ShaderRegisterElement)
	#(positionTarget ShaderRegisterElement)
	#(rotationRegisters ShaderRegisterElement)
	#(scaleAndRotateTarget ShaderRegisterElement)
	#(sourceRegisters ShaderRegisterElement)
	#(targetRegisters ShaderRegisterElement)
	#(uvAttribute ShaderRegisterElement)
	#(uvTarget ShaderRegisterElement)
	#(uvVar ShaderRegisterElement)
	#(velocityTarget ShaderRegisterElement)
	#(vertexConstantData Array)
	#(vertexLife ShaderRegisterElement)
	#(vertexOneConst ShaderRegisterElement)
	#(vertexTime ShaderRegisterElement)
	#(vertexTwoConst ShaderRegisterElement)
	#(vertexZeroConst ShaderRegisterElement)
)!
AnimationRegisterCache buildLuaFunctions: #(
	#(getColorCombinationCode #(shadedTarget))
	getColorPassCode getCombinationCode
	getInitCode setDataLength
	#(getRegisterIndex #(_node parameterIndex))
	#(setFragmentConst #(index x y z w))
	#(setRegisterIndex #(_node parameterIndex registerIndex))
	#(setUVSourceAndTarget #(uvAttribute uvVaring))
	#(setVertexConst #(index x y z w))
	#(setVertexConstFromMatrix #(index _matrix))
	#(setVertexConstFromVector #(index #(data handles)))
)!

LuaStudioLibrary subclass: #ShaderRegisterElement !
ShaderRegisterElement buildGetters: #( index regName )!
ShaderRegisterElement buildFunctions: #( toString )!

EventDispatcher subclass: #Stage3DProxy !
Stage3DProxy buildTypedGetters: #(
	usesSoftwareRendering
	#(viewPort ARectangle)
	#(context3D Context3D)
	driverInfo profile
	renderSurfaceSelector
	#(renderTarget TextureBase)
	#(stage3D Stage3D)
	stage3DIndex
)!
Stage3DProxy buildTypedAccessors: #(
	antiAlias bufferClear
	color height
	enableDepthAndStencil
	#(mouse3DManager Mouse3DManager)
	#(scissorRect ARectangle)
	#(touch3DManager Touch3DManager)
	visible width x y
)!
Stage3DProxy buildLuaFunctions: #(
	clear clearDepthBuffer
	#(configureBackBuffer #(backBufferWidth backBufferHeight antiAlias enableDepthAndStencil))
	dispose present recoverFromDisposal
	#(setRenderTarget #(_target enableDepthAndStencil surfaceSelector))
)!
Event buildConstants: '
	Lib.Media.Events.Event.ENTER_FRAME
'!

LuaStudioLibrary subclass: #SubGeometryBase !
SubGeometryBase buildTypedGetters: #(
	numTriangles scaleU scaleV
	#(UVData Array)
	UVOffset UVStride
	vertexData
	vertexNormalData
	vertexNormalOffset
	vertexNormalStride
	vertexOffset
	vertexPositionData
	vertexStride
	vertexTangentData
	vertexTangentOffset
	vertexTangentStride
)!
SubGeometryBase buildTypedAccessors: #(
	autoDeriveVertexNormals
	autoDeriveVertexTangents
	autoGenerateDummyUVs
	#(parentGeometry Geometry)
	useFaceWeights
)!
SubGeometryBase buildLuaFunctions: #(
	#(applyTransformation #(_transform))
	dispose
	#(getIndexBuffer #(_stage3DProxy))
	#(scale #(scale))
	#(scaleUV #(scaleU scaleV))
	#(updateIndexData #(indices))
)!

SubGeometryBase subclass: #CompactSubGeometry !
CompactSubGeometry buildTypedGetters: #(
	numVertices
	secondaryUVOffset
	secondaryUVStride
)!
CompactSubGeometry buildLuaFunctions: #(
	#(activateSecondaryUVBuffer #(index _stage3DProxy))
	#(activateUVBuffer #(index stage3DProxy))
	#(activateVertexBuffer #(index _stage3DProxy))
	#(activateVertexNormalBuffer #(index _stage3DProxy))
	#(activateVertexTangentBuffer #(index _stage3DProxy))
	#(clone self)
	#(cloneWithSeperateBuffers SubGeometry)
	#(fromVectors #(verts uvs normals tangents))
	#(stripBuffer #(offset numEntries))
	#(updateData #(#(data handles)))
)!

SubGeometryBase subclass: #SubGeometry !
SubGeometry buildTypedGetters: #(
	numVertices
	secondaryUVData
	secondaryUVOffset
	secondaryUVStride
)!
SubGeometry buildLuaFunctions: #(
	#(activateSecondaryUVBuffer #(index _stage3DProxy))
	#(activateUVBuffer #(index stage3DProxy))
	#(activateVertexBuffer #(index _stage3DProxy))
	#(activateVertexNormalBuffer #(index _stage3DProxy))
	#(activateVertexTangentBuffer #(index _stage3DProxy))
	#(clone self)
	#(cloneWithSeperateBuffers SubGeometry)
	#(fromVectors #(verts uvs normals tangents))
	#(updateSecondaryUVData #(uvs))
	#(updateUVData #(uvs))
	#(updateVertexData #(vertices))
	#(updateVertexNormalData #(vertexNormals))
	#(updateVertexTangentData #(vertexTangents))
)!

LuaStudioLibrary subclass: #SubMesh !
SubMesh buildTypedGetters: #(
	#(animator AnimatorBase)
	#(bounds BoundingVolumeBase)
	castsShadows
	#(indexData Array)
	#(inverseSceneTransform Matrix3D)
	mouseEnabled numTriangles numVertices
	#(sceneTransform Matrix3D)
	shaderPickingDetails
	#(sourceEntity Entity)
	UVData UVOffset UVStride
	#(uvTransform Matrix)
	#(vertexData Array)
	#(vertexNormalData Array)
	vertexNormalOffset vertexOffset vertexStride
	vertexTangentData vertexTangentOffset
	visible
)!
SubMesh buildTypedAccessors: #(
	#(animationSubGeometry AnimationSubGeometry)
	#(material MaterialBase)
	offsetU offsetV scaleU scaleV uvRotation
	#(subGeometry SubGeometry)
)!
SubMesh buildLuaFunctions: #(
	#(activateSecondaryUVBuffer #(index _stage3DProxy))
	#(activateUVBuffer #(index stage3DProxy))
	#(activateVertexBuffer #(index _stage3DProxy))
	#(activateVertexNormalBuffer #(index _stage3DProxy))
	#(activateVertexTangentBuffer #(index _stage3DProxy))
	#(getIndexBuffer #(_stage3DProxy) IndexBuffer3D)
	#(getRenderSceneTransform #(_camera) Matrix3D)
	dispose
)!

AnimatorBase  subclass: #ParticleAnimator !
!Animators class methodsFor: #accessing !
particleAnimator
	" Returns a wrapper for library "

	^ParticleAnimator! !
! ParticleAnimator class methodsFor: #instantiation !
new: particleAnimationSet
	" Returns an instance of the receiver. "

	^{'self:_at(Lib.Away3D.Animators.ParticleAnimator.new(particleAnimationSet:handle()))'}! !
ParticleAnimator buildLuaFunctions: #(
	#(clone #() ParticleAnimator)
	#(resetTime #(offset))
	#(setRenderState #(_stage3DProxy _renderable vertexConstantOffset vertexStreamOffset _camera))
	#(testGPUCompatibility #(_pass))
)!

"-HTTP Server (non-blocking)-------------------------------------------------------------------"!
EventManager
	subclass: #NonBlockingHTTPServer
	instanceVariableNames: ' host port socket httpHandlers callback crlf'
	category: #LuaStudio !
NonBlockingHTTPServer comment: '
	Pluggable non-blocking HTTP server.
	@2019-today Alejandro Reimondo aleReimondo@smalltalking.net
	Published under U8 license terms see http://u8.smalltalking.net
	See rfc2616 for more information about this HTTP server implementation
'!
NonBlockingHTTPServer class instanceVariableNames: 'current'!
NonBlockingHTTPServer class buildAccessors: #(
	#(current initialize 'self new')
)!
NonBlockingHTTPServer buildAccessors: #(
	#(host initialize 'Host @ String new')
	#(port initialize '8088')
	#(callback initialize '[:ignored| self listen ]')
	#(httpHandlers initialize 'self needsHTTPHandlers')
)!
NonBlockingHTTPServer buildGetters: #( crlf )!

! NonBlockingHTTPServer class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^#(	#postProcess:with:
		#send:response:with:
		#needsRequestFor:with:
		#needsHandlerFor:with:
		#acceptVersion:
		#getHttpHandlers #getHttpHandlers:
		#openResponseLines:at:with:
		#needsResource:with:
		#closeResponseLines:at:with:
		#needsContents:with:
		#needsAccessControlAllow:at:with:
		#needsContentType:at:with:
		#needsContentLength:at:with:
	) ,super constructEventsTriggered! !

! NonBlockingHTTPServer class methodsFor: #sample !
example
	" server - Implements a simple (non-blocking) HTTP server "

	| server |
	server := self current.
	server when: #needsContents:with: do: [:uri :headers|
		'<html>
<head><title>' ,self class name,' example</title></head>
<body><h1>Hello from S8</h1>
<hr>Headers:<br><pre>' ,(headers keys sorted inject: '' into: [:total :key|
		total, key ,' : ' ,(headers at: key) asLiteral ,'<br>'
]),'</pre>
</body></html>'
	].
	server when: #postProcess:with: do: [:start :headers|
		(start notNil
			and: [ start first = #GET
			and: [ #('/quit' '/exit' '/stop') includes: start second ]]) ifTrue: [
			server stop.
			self note: 'server stop'
		]
	].
	server start.
	self note: ' server started at ' ,server host toString ,':' ,server port.
	^server! !

! NonBlockingHTTPServer methodsFor: #configuration !
ip: ipAddress
	" Set the receiver's host at IP adress. "

	self host: Host @ ipAddress! !

! NonBlockingHTTPServer methodsFor: #private !
startListening
	" Private - Start listening for clients. "

	View3D preRenderActions add: self callback.! !

! NonBlockingHTTPServer methodsFor: #private !
stopListening
	" Private - Stop listening for clients. "

"	Display stage
		removeEventListener: Event enterFrame
		doing: self callback capture: false."
	View3D preRenderActions remove: self callback ifAbsent: [].! !

! NonBlockingHTTPServer methodsFor: #running !
stop
	" Stop the receiver. "

	socket isNil ifTrue: [ ^self ].
	self stopListening.
	socket close.
	socket := nil.! !

! NonBlockingHTTPServer methodsFor: #private !
peek
	" Private - Peek a client. "

	^socket notNil ifTrue: [ socket acceptOrNil ]! !

! NonBlockingHTTPServer methodsFor: #running !
start
	" Start the receiver. "

	socket notNil ifTrue: [ self close ].
	crlf := 13 asCharacter ,10 asCharacter.
	socket := Socket new.
	socket
		bind: self host port: self port;
		listen: 30;
		setBlocking: false.
	self startListening.! !

! NonBlockingHTTPServer methodsFor: #private !
listen
	" Private - Listen client requests. "

	| client |
	[ client := self peek.
	  client notNil ] whileTrue: [
	  client setBlocking: false.
	  self processMessageOn: client
	].
	^nil! !

! NonBlockingHTTPServer methodsFor: #processing !
processMessageOn: client
	" Private - Process a client message.
	https://tools.ietf.org/html/rfc2616#section-4
	HTTP messages consist of requests from client to server and responses from server to client.
	Both types of message consist of a start-line, zero or more header fields (also known as headers), an empty line (i.e., a line with nothing preceding the CRLF) indicating the end of the header fields, and possibly a message-body.
	"

	| start line headers |
	[	line := client input readLineOr: ''.
		line isEmpty ] whileTrue: [
		self note: 'start - readLine failed'
	].
	start := self startLine: line.
	headers := PoolDictionary new.
	[	line := client input readLineOr: #failed.
		line notEmpty ] whileTrue: [
		line = #failed
		ifTrue: [ self note: 'header - readLine failed' ]
		ifFalse: [ self addHeader: line to: headers ]
	].
	self send: client responseFor: start with: headers.
	self postProcess: start with: headers! !

! NonBlockingHTTPServer methodsFor: #processing !
send: client responseFor: start with: headers
	" Private - Send the response to client for request start. "

	self	triggerEvent: #send:response:with:
		withArguments: (Array with: client with: start with: headers)
		ifNotHandled: [ client write: (self responseFor: start with: headers); close ]! !

! NonBlockingHTTPServer methodsFor: #processing !
addHeader: line to: headers
	" Private - Add a request header to headers.
	https://tools.ietf.org/html/rfc2616#section-5.3
	The request-header fields allow the client to pass additional information about the request, and about the client itself, to the server.
	"

	| stream key value |
	stream := line stream.
	key := (stream upTo: $:) trimBlanks.
	value := stream upToEnd trimBlanks.
	(headers includesKey: key) ifFalse: [
		^headers at: key put: value
	].
	(headers at: key) isArray ifFalse: [
		headers at: key put: (Array with: (headers at: key))
	].
	^headers at: key put: value! !

! NonBlockingHTTPServer methodsFor: #processing !
startLine: line
	" Private - Returns the result of parsing a request start line.
	An array containing:
		Method		https://tools.ietf.org/html/rfc2616#section-5.1.1
		Request-URI	https://tools.ietf.org/html/rfc2616#section-5.1.2
		version
	"

	| stream method uri |
	stream := line stream.
	method := stream upTo: $ .
	uri := stream upTo: $ .
	(stream upTo: $/) = #HTTP ifFalse: [ ^nil ]. "not a start request line"
	^Array with: method with: uri with: stream upToEnd! !

! NonBlockingHTTPServer methodsFor: #processing !
postProcess: start with: headers
	" Private - Post process a request. "

"DEBUG- true ifTrue: [
# #print: #DEBUG with: #postProcessing.
^self stop
]."

	self	triggerEvent: #postProcess:with:
		withArguments: (Array with: start with: headers)! !

! NonBlockingHTTPServer methodsFor: #processing !
responseFor: start with: headers
	" Private - Returns the response string to be sent to a client requesting message with start and headers. "

	start isNil ifTrue: [ ^self badRequest ].
	start last = '1.1' ifFalse: [
		self	triggerEvent: #acceptVersion:
			withArguments: (Array with: start)
			ifNotHandled: [ ^self versionNotSupported: start last ]
	].
	^self	triggerEvent: #needsRequestFor:with:
		withArguments: (Array with: start with: headers)
		ifNotHandled: [ self defaultResponseFor: start with: headers ]! !

! NonBlockingHTTPServer methodsFor: #processing !
defaultResponseFor: start with: headers
	" Private - Returns the response string to be sent to a client requesting message with start and headers. "

	| block |
	block := self
		httpHandlerFor: start and: headers
		ifAbsent: [ [:s :h| self methodNotImplemented ] ].
	^block evaluateWithArguments: (Array with: start with: headers)! !

! NonBlockingHTTPServer methodsFor: #responses !
badRequest
	" Private - Returns the response string for a bad request.
	https://tools.ietf.org/html/rfc2616#section-10.4.1
	The request could not be understood by the server due to malformed syntax.
	The client SHOULD NOT repeat the request without modifications.
	"

	^'HTTP/1.1 400 Bad Request'! !

! NonBlockingHTTPServer methodsFor: #responses !
methodNotImplemented
	" Private - Returns the response string for an unimplemented request.
	https://tools.ietf.org/html/rfc2616#section-10.5.2
	The server does not support the functionality required to fulfill the request.
	This is the appropriate response when the server does not recognize the request method and is not capable of supporting it for any resource.
	"

	^'HTTP/1.1 501 Not Implemented'! !

! NonBlockingHTTPServer methodsFor: #responses !
versionNotSupported:  version
	" Private - Returns the response string for an unsupported HTTP version.
	https://tools.ietf.org/html/rfc2616#section-10.5.6
	"

	^'HTTP/1.1 505 HTTP Version Not Supported : ' ,version! !

! NonBlockingHTTPServer methodsFor: #configuration !
httpHandlerFor: start and: headers ifAbsent: aBlock
	" Private - Returns the handler of request messages with start and headers (or the result of evaluating aBlock). "

	^self	triggerEvent: #needsHandlerFor:with:
		withArguments: (Array with: start with: headers)
		ifNotHandled: [ self httpHandlers at: start first ifAbsent: aBlock ]! !

! NonBlockingHTTPServer methodsFor: #configuration !
needsHTTPHandlers
	" Private - Returns the default HTTP handlers of the receiver. "

	httpHandlers := PoolDictionary new.
	self	triggerEvent: #getHttpHandlers ifNotHandled: [ self installDefaultHttpHandlers ];
		triggerEvent: #getHttpHandlers: with: httpHandlers.
	^httpHandlers! !

! NonBlockingHTTPServer methodsFor: #defaults !
installDefaultHttpHandlers
	" Private - Install the default handlers of the receiver. "

	httpHandlers
		at: #GET put: self getHandler;
		at: #HEAD put: self headHandler;
		yourself! !

! NonBlockingHTTPServer methodsFor: #handlers !
headHandler
	" Private - Returns the handler of HEAD request.
	https://tools.ietf.org/html/rfc2616#section-9.4
	"

	^[:start :headers| | stream |
		stream := '' stream.
		self get: start second into: stream headers: headers body: false.
		stream contents
	]! !

! NonBlockingHTTPServer methodsFor: #handlers !
getHandler
	" Private - Returns the handler of GET request.
	https://tools.ietf.org/html/rfc2616#section-9.3
	"

	^[:start :headers| | stream |
		stream := '' stream.
		self get: start second into: stream headers: headers body: true.
		stream contents
	]! !

! NonBlockingHTTPServer methodsFor: #get !
get: uri into: stream headers: headers body: includesBody
	" Private - Emmit the response of GET operation onto stream. "

	| lines statusLine body |
	lines := self requestLinesFor: uri headers: headers withBody: includesBody.
	statusLine := lines removeFirst.
	(includesBody and: [
		statusLine second asString first = $2 and: [
		lines notEmpty ]]) ifTrue: [ "status-code with shape 2xx: Success"
		body := lines removeLast
	].

	statusLine "Status-Line = HTTP-Version SP Status-Code SP Reason-Phrase CRLF"
		do: [:each| stream nextPutAll: each ]
		separatedBy: [ stream space ].
	stream nextPutAll: crlf.

	lines do: [:assoc| "entity-headers https://tools.ietf.org/html/rfc2616#section-7.1"
		stream	nextPutAll: assoc key;
			space; nextPutAll: ':'; space;
			nextPutAll: assoc value toString;
			nextPutAll: crlf
	].

	stream nextPutAll: crlf. "end of header"
	body notNil ifTrue: [ stream nextPutAll: body ]! !

! NonBlockingHTTPServer methodsFor: #get !
requestLinesFor: uri headers: headers withBody: includesBody
	" Private - Returns the request lines of resource at uri. "

	| result resource |
	result := Array new.
	self	triggerEvent: #openResponseLines:at:with:
		withArguments: (Array with: result with: uri with: headers).
	resource := self
		triggerEvent: #needsResource:with:
		withArguments: (Array with: uri with: headers)
		ifNotHandled: [
			self dynamicResource: uri headers: headers withBody: includesBody
		].
	resource isNil ifTrue: [ ^#( #('HTTP/1.1' 404 'Not Found') ) ].
	self	addAccessControlAllow: resource at: uri with: headers to: result;
		addContentTypeAndLength: resource at: uri with: headers to: result.
	self	triggerEvent: #closeResponseLines:at:with:
		withArguments: (Array with: result with: uri with: headers).
	includesBody ifTrue: [
		result add: (self bodyOf: resource at: uri with: headers)
	].
	result addFirst: #('HTTP/1.1' 200 #OK). "The request has succeeded."
	^result! !

! NonBlockingHTTPServer methodsFor: #get !
dynamicResource: uri headers: headers withBody: includesBody
	" Private - Returns the resource computed for uri (or nil).
	The result MUST be aString, nil or an Array with shape
		#( mime-type bodyString [Array of Associations] )
	"

	^self	triggerEvent: #needsContents:with:
		withArguments: (Array with: uri with: headers)
		ifNotHandled: [ nil ]! !

! NonBlockingHTTPServer methodsFor: #get !
addAccessControlAllow: resource at: uri with: headers to: result
	" Private - Add this entity header fields of resource to result. "

	| tuples |
	tuples := self
		triggerEvent: #needsAccessControlAllow:at:with:
		withArguments: (Array with: resource with: uri with: headers)
		ifNotHandled: [ Array with: #Access-Control-Allow-Origin -> $* ].
	result addAll: tuples! !

! NonBlockingHTTPServer methodsFor: #get !
addContentTypeAndLength: resource at: uri with: headers to: result
	" Private - Add this entity header fields of resource to result. "

	| tuples |
	tuples := self
		triggerEvent: #needsContentType:at:with:
		withArguments: (Array with: resource with: uri with: headers)
		ifNotHandled: [ Array with: ('Content-Type' -> (self contentTypeOf: resource at: uri with: headers)) ].
	result addAll: tuples.
	tuples := self
		triggerEvent: #needsContentLength:at:with:
		withArguments: (Array with: resource with: uri with: headers)
		ifNotHandled: [ Array with: ('Content-Length' -> (self contentLengthOf: resource at: uri with: headers)) ].
	result addAll: tuples! !

! NonBlockingHTTPServer methodsFor: #get !
bodyOf: resource at: uri with: headers
	" Private - Returns the body of resource. "

	resource isString ifTrue: [ ^resource ].
	^resource second! !
	
! NonBlockingHTTPServer methodsFor: #get !
contentLengthOf: resource at: uri with: headers
	" Private - Returns this header field of resource. "

	| bytes |
	bytes := resource isString ifTrue: [ resource ] ifFalse: [ resource second ].
	^bytes size ! !
	
! NonBlockingHTTPServer methodsFor: #get !
contentTypeOf: resource at: uri with: headers
	" Private - Returns this header field of resource. "

	resource isString ifTrue: [
		^(self guessTypeOf: resource at: uri with: headers)
	].
	^resource first! !
	
! NonBlockingHTTPServer methodsFor: #get !
guessTypeOf: resource at: uri with: headers
	" Private - Return the type of resource (a String). "

	(resource startsWith: $<) ifTrue: [ ^'text/html' ].
	^'text/plain'! !
