" LuaStudio wrappers
Disclaimer: we are NOT recomending/suggesting to use Love for application development,
 this API implementation has been written with the sole goal to test our native access features for Lua.
Copyright (C) 2019-today Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
See also https://love2d.org
"!

NativeObject subclass: #LoveObject category: #love!

! LoveObject class methodsFor: #builders !
reservedWords
	" Private - Returns the reserved words for the receiver. "

	^#( class name )
"	^String reservedWords
"! !

! LoveObject class methodsFor: #builders !
stMessageHeader: msgHeader
	" Private - Return a message header string with nice smalltalk looking. "

	| default mode stream word |
	default := [ super stMessageHeader: msgHeader ].
	(msgHeader upTo: $:) size < 4 ifTrue: [ ^default value ].
	(msgHeader at: 4) isUppercase not ifTrue: [ ^default value ].
	mode := msgHeader copyFrom: 1 to: 3.
	(#(get set) includes: mode) ifFalse: [ ^default value ].
	stream := ((msgHeader at: 4) asLowercase
		,(msgHeader copyFrom: 5 to: msgHeader size)) stream.
	word := stream upTo: $:.
	(self reservedWords includes: word) ifTrue: [ ^default value ].
	stream atEnd ifFalse: [ word := word ,$: ,stream upToEnd ].
	^super stMessageHeader: word! !



! LoveObject class methodsFor: #api !
withString: data do: aBlock
	" Private - If data is an instance of String, returns the result of evaluating aBlock (or nil). "

	^(self typeOf: data) = #string ifTrue: aBlock! !

! LoveObject class methodsFor: #u8 !
hasFileAccess
	" Private - Returns true if we can access/save files. "

	^(self typeOf: Filesystem nativeClass @ #getInfo) == #function! !

! LoveObject class methodsFor: #u8 !
pathToSystemFolder
	" Private - Returns the path for system root (or nil). "

	^Filesystem saveDirectory ,$/! !

! LoveObject class methodsFor: #u8 !
pathToM8n
	" Private - Returns the local path of m8n folder (or nil). "

	^self pathToSystemFolder ,'m8n/'! !

! LoveObject class methodsFor: #u8 !
pathToU8Cache
	" Private - Returns the local path of U8 cache folder (or nil). "

	| contributionURL |
	self hasFileAccess ifFalse: [ ^nil ].
	contributionURL := Smalltalk at: #ContributionURL ifAbsent: [ ^nil ].
	^self pathToM8n ,(contributionURL fromLast: '/profile/')! !

! LoveObject class methodsFor: #u8 !
isLocalPath: pathName
	" Private - Returns true if the pathName is a local file name. "

	^(pathName includes: $:)
	or: [(pathName startsWith: $/)
	or: [(pathName startsWith: $\)]]! !

! LoveObject class methodsFor: #u8 !
isAssetPath: pathName
	" Private - Returns true if the file at pathName is stored in assets. "

	^(self isLocalPath: pathName) not! !

! LoveObject class methodsFor: #u8 !
pathToFile: pathName
	" Private - Returns the local path of file at pathName. "

	(self isAssetPath: pathName) ifTrue: [ ^'/assets/' ,pathName ].
	^pathName! !

! LoveObject class methodsFor: #u8 !
withU8Path: aPath do: aBlock
	" Private - Ensure a local file at aPath exists.
	If aPath is only a fileName will search for file in well known folders; if a file exists that local path is used.
	Path for files stored in contribution starts with u8:/
	If file does not exist try to download the file from contribution path.
	Returns the result of evaluating aBlock with local(downloaded) file path or nil.
	"

	| u8Root path contributionURL filePath url |
	HasBeenInstalled == true ifFalse: [
		HasBeenInstalled := true.
		self withAllSubclasses do: [:cls| cls launch ]
	].

	u8Root := 'u8:/'.
	(self isAssetPath: aPath) ifTrue: [ "fileName in default folder"
		(Smalltalk includesKey: #ContributionURL) ifTrue: [
			^self withU8Path: u8Root ,aPath do: aBlock
		].
		filePath := self pathToM8n ,path.
		(Filesystem info: filePath) notNil ifTrue: [
			 ^aBlock value: filePath ].
		filePath := self pathToSystemFolder ,path.
		(Filesystem info: filePath) notNil ifTrue: [
			^aBlock value: filePath ].
		^nil
	].

	(aPath startsWith: u8Root) not ifTrue: [
		((self isLocalPath: aPath) and: [(Filesystem info: aPath) notNil ]) ifTrue: [
			^aBlock value: aPath
		].
		^nil ].
	path := aPath copyFrom: u8Root size + 1 to: aPath size.
	contributionURL := Smalltalk at: #ContributionURL ifAbsent: [
		^self error: 'Contribution URL missing'
	].
	filePath := self pathToU8Cache ,path.
	(Filesystem info: filePath) notNil ifTrue: [ ^aBlock value: filePath ].
	Filesystem createDirectory: (filePath upToLast: $/) ,$/.

	url := contributionURL ,path.
	[|data|
		data := (# #_require: 'socket.http') #_request: url.
		Filesystem write: filePath data: data size: data size
	] on: Error do: [:err|
		# #print: err with: url.
		^self error: err
	].
	^aBlock value: filePath! !

! LoveObject class methodsFor: #u8 !
deleteU8Cache
	" Delete all the files cached for current contribution. "

	| folder |
	folder := self pathToU8Cache.
	^folder notNil and: [ Filesystem prune: folder ]! !

! LoveObject class methodsFor: #u8 !
deleteM8n
	" Delete all the files cached for m8n platform. "

	^Filesystem prune: self pathToM8n! !

! LoveObject class methodsFor: #u8 !
withPath: aPath do: aBlock
	" Private - Returns the result of evaluating aBlock with file path. "

	| cached done |
	done := self withU8Path: aPath do: [:path|
		cached := aBlock value: path.
		#done
	].
	^done isNil ifTrue: [ aBlock value: aPath ] ifFalse: [ cached ] ! !

LoveObject subclass: #Love!

! Love class methodsFor: #accessing !
nativeClass
	" Private - Returns the native module. "

	^Lua @ #{love} ! !

! Love class methodsFor: #callbacks !
callbackSelector: aName
	" Private - Returns the selector for callback hook at aName. "

	| result |
	result := aName asLowercase copyWithout: $:.
	(self callbackNames includes: result) ifFalse: [
		self error: 'Invalid callback name'
	].
	^result ! !

! Love class methodsFor: #callbacks !
callbackNames
	" Returns the names of callback hooks. "

	^#(
	directorydropped draw
	errhand errorhandler
	filedropped focus
	keypressed keyreleased	
	load lowmemory
	mousefocus
	mousemoved
	mousepressed
	mousereleased
	quit resize run
	textedited textinput
	threaderror
	touchmoved
	touchpressed
	touchreleased
	update visible
	wheelmoved
	gamepadaxis
	gamepadpressed
	gamepadreleased
	joystickadded joystickremoved
	joystickaxis joystickhat
	joystickpressed joystickreleased
)! !

Love buildLuaStaticFunctions: #(
	getVersion
	hasDeprecationOutput
	#(setDeprecationOutput #(enable))
)!

! Love class methodsFor: #version !
version
	" Returns version information (anArray). "

	| major minor revision codename |
{'	major, minor, revision, codename = love.getVersion() '}.
	^Array with: major with: minor with: revision with: codename! !

! Love class methodsFor: #accessing !
module: aName
	" Returns the module (handle or nil). "

	^Lua @ (self nativeClass handleAt: aName)! !

! Love class methodsFor: #callbacks !
callbackFor: aName
	" Returns the current callback value (aBlock). "

	^self nativeClass @ (self callbackSelector: aName)! !

! Love class methodsFor: #callbacks !
on: aName do: aBlock
	" Set the callback for selector.
	Returns the old callback handler.
	"

	| old selector |
	selector := self callbackSelector: aName.
	old := self nativeClass handleAt: selector.
	self nativeClass handleAt: selector put: aBlock.
	^old! !

#(	audio data event filesystem font graphics
	joystick keyboard math mouse
	physics sound system thread
	timer touch video window
) do: [:module| | cls |
	cls := module asGlobalName.
	LoveObject subclass: cls.
	cls := Smalltalk at: cls.
	cls class
		implement: #nativeClass
		comment: 'Private - Returns the wrapper for native binding of the receiver.'
		as: '^Love module: #' ,module
		category: #private
]!

! Love class methodsFor: #u8 !
deleteM8n
	" Delete all the files cached for m8n platform. "

	^Filesystem prune: #m8n! !

! Love class methodsFor: #u8 !
deleteU8Cache
	" Delete all the files cached for current contribution. "

	| folder |
	folder := self pathToU8Cache.
	^folder notNil and: [ Filesystem prune: folder ]! !

! Love class methodsFor: #u8 !
pathToU8Cache
	" Private - Returns the local path of U8 cache folder (or nil). "

	| contributionURL |
	contributionURL := Smalltalk at: #ContributionURL ifAbsent: [ ^nil ].
	^'m8n/', (contributionURL fromLast: '/profile/')! !

Audio buildLuaStaticFunctions: #(
	#(getActiveEffects #() Array)
	getActiveSourceCount
	getDistanceModel
	getDopplerScale
	#(getEffect #(name) PoolDictionary)
	getMaxSceneEffects
	getMaxSourceEffects
	getNumSources
	#(getRecordingDevices #() PoolDictionary)
	getSourceCount
	getVolume
	isEffectsSupported
	#(newQueueableSource #(samplerate bitdepth channels buffercount) Source)
	#(newSource #(filename type) Source)
	#(pause #() 'Array[Source]')
	#(pause #(#(sources handles)))
	#(play #(#(sources handles)))
	resume #(resume #(_source))
	rewind #(rewind #(_source))
	#(setDistanceModel #(model))
	#(setDopplerScale #(scale))
	#(setEffect #(name _settings) Boolean)
	#(setMixWithSystem #(mix))
	#(setOrientation #(fx fy fz ux uy uz))
	#(setPosition #(x y z))
	#(setVelocity #(x y z))
	#(setVolume #(volume))
	stop #(stop #(#(sources handles)))
)!

! Audio class methodsFor: #accessing !
orientation
	| fx fy fz ux uy uz |
{'	fx, fy, fz, ux, uy, uz = love.audio.getOrientation( ) '}.
	^Array	with: (Array with: fx with: fy with: fz)
		with: (Array with: ux with: uy with: uz)! !

! Audio class methodsFor: #accessing !
position
	| x y z |
{'	x, y, z = love.audio.getPosition( ) '}.
	^Array with: x with: y with: z! !

! Audio class methodsFor: #accessing !
velocity
	| x y z |
{'	x, y, z = love.audio.getVelocity( ) '}.
	^Array with: x with: y with: z! !

#(	Byte Compressed CompressedImage
	File Font Glyph Image Sound
) do: [:type| Data subclass: type ,#Data ]!

Data buildLuaStaticFunctions: #(
	#(compress #(container format _data level) CompressedData)
	#(decode #(container format _data))
	#(decompress #(container _data))
	#(encode #(container format _data linelength))
	#(getPackedSize #(format))
	#(hash #(hashFunction string))
	#(newByteData #(dataOrSize) ByteData)
	#(newByteData #(_data offset size) ByteData)
	#(newDataView #(_data offset size) Data)
)!

Data buildFunctions: #(
	#(clone #() Data)
	getPointer getSize
	getString
)!
CompressedData buildFunctions: #(
	getFormat
)!
CompressedImageData buildFunctions: #(
	getFormat getHeight getWidth getMipmapCount
)!

! CompressedImageData methodsFor: #accessing !
extent
	^self width @ self height! !

! CompressedImageData methodsFor: #accessing !
dimensions
	| width height |
{'	width, height = self:handle():getDimensions( ) '}.
	^Array with: width with: height! !

! CompressedImageData methodsFor: #accessing !
dimensionsAt: level
	| width height |
{'	width, height = self:handle():getDimensions(level) '}.
	^Array with: width with: height! !

FileData buildFunctions: #(
	getExtension getFilename
)!

ImageData buildFunctions: #(
	#(encode #(format filename) FileData)
	#(encode #(format) FileData)
	getFormat getHeight getWidth
	getString
	#(mapPixel #(mapFunction))
	#(paste #(_source dx dy sx sy sw sh))
	#(setPixel #(x y r g b a))
)!

! ImageData methodsFor: #accessing !
extent
	^self width @ self height! !

! ImageData methodsFor: #accessing !
dimensions
	| width height |
{'	width, height = self:handle():getDimensions( ) '}.
	^Array with: width with: height! !

! ImageData methodsFor: #accessing !
pixelAt: aPoint
	| r g b a |
{'	r, g, b, a = self:handle():getPixel(aPoint:x(),aPoint:y()) '}.
	^Array with: r with: g with: b with: a! !

SoundData buildFunctions: #(
	getBitDepth getBits getDuration
	getChannelCount getChannels
	#(getSample #(i))
	#(getSample #(i channel))
	getSampleCount getSampleRate
	#(setSample #(i channel sample))
)!

Event buildLuaStaticFunctions: #(
	clear
	pool
	pump
	#(push #(event a b c d e f))
	#(push #(event a b c d e))
	#(push #(event a b c d))
	#(push #(event a b c))
	#(push #(event a b))
	#(push #(event a))
	quit #(quit #(exitstatus))
)!

! Event class methodsFor: #wait !
wait
	| n a b c d e f |
{'	n, a, b, c, d, e, f = love.event.wait( ) '}.
	^(Array with: n)
		,(Array with: a with: b with: c)
		,(Array with: d with: e with: f)! !

! Event class methodsFor: #pool !
poolDo: aBlock
	| n a b c d e f |
{'	for n, a, b, c, d, e, f in love.event.pool() do
		aBlock:evaluateWithArguments_(smalltalk.Array:_at({n,a,b,c,d,e,f}))
	end'}.
	^self! !

Filesystem buildLuaStaticFunctions: #(
	#(append #(fileName string size))
	#(append #(fileName string))
	areSymlinksEnabled
	#(createDirectory #(dirname))
	#(enumerate #(dir) Array)
	#(exists #(filename))
	getAppdataDirectory
	getCRequirePath
	#(getDirectoryItems #(dir callback) Array)
	#(getDirectoryItems #(dir) Array)
	getIdentity
	#(getInfo #(path) PoolDictionary)
	#(getInfo #(path filtertype) PoolDictionary)
	#(getLastModified #(filename))
	#(getRealDirectory #(filepath))
	getRequirePath
	getSaveDirectory
	#(getSize #(filename))
	getSource getSourceBaseDirectory
	getUserDirectory getWorkingDirectory
	#(isDirectory #(filename))
	#(isFile #(filename))
	#(isSymlink #(filename))
	isFused
	#(load #(filename))
	#(mkdir #(dirname))
	#(mount #(_archive mountpoint))
	#(mount #(_archive mountpoint appendToPath))
	#(mount #(_data archiveName mountpoint))
	#(mount #(_data archiveName mountpoint appendToPath))
	#(newFile #(filename) File)
	#(newFileData #(filepath) FileData)
	#(newFileData #(contents filename) FileData)
	#(newFileData #(contents filename _decoder) FileData)
	#(read #(filename))
	#(read #(filename size))
	#(remove #(filename))
	#(setCRequirePath #(paths))
	#(setIdentity #(filename))
	#(setIdentity #(filename appendToPath))
	#(setRequirePath #(paths))
	#(setSource #(path))
	#(setSymlinksEnabled #(enable))
	#(unmount #(archive))
	#(write #(filename _data))
	#(write #(filename _data size))
)!

! Filesystem class methodsFor: #reading !
lines: fileName do: aBlock
	" Evaluates aBlock on each line of file at fileName "

{'	for line in love.filesystem.lines(fileName) do
		aBlock:evaluateWith_(line)
	end'}.
	^self! !

! Filesystem class methodsFor: #files !
prune: pathName
	" Delete a file at pathName or all the files and folders in directory at pathName. "

	pathName isNil ifTrue: [ ^false ].
	(pathName = $/ or: [ pathName = (self saveDirectory, $/) ])
		ifTrue: [ ^self error: 'We do not remove root folder' ].
	^self recursivelyDelete: pathName! !

! Filesystem class methodsFor: #files !
recursivelyDelete: pathName
	" Private - Delete a file at pathName or all the files and folders in directory at pathName. "

	(self info: pathName filtertype: #directory) notNil ifTrue: [
		(self directoryItems: pathName) do: [:child | | path |
			path := pathName ,$/ ,child.
			self recursivelyDelete: path.
		]
	].
	(self info: pathName) isNil ifTrue: [ ^false ].
	self remove: pathName.
	^true! !

LoveObject subclass: #File!
File subclass: #DroppedFile!

File buildFunctions: #(
	close eof flush
	getFilename
	getMode getSize tell
	isEOF isOpen
	#(open #(mode))
	#(read #(bytes))
	#(seek #(pos))
	#(setBuffer #(mode))
	#(setBuffer #(mode size))
	#(write #(data size))
	#(write #(data))
)!

! File methodsFor: #testing !
atEnd
	^self eof! !

! File methodsFor: #accessing !
position
	^self tell! !

! File methodsFor: #accesing !
buffer
	| mode size |
{'	mode, size = self:handle():getBuffer() '}.
	^Array with: mode with: size! !

! File methodsFor: #reading !
linesDo: aBlock
{'	for line in self:handle():lines() do
		aBlock:evaluateWith_(line)
	end'}.
	^self! !

! File methodsFor: #reading !
read: container bytes: bytes
	| contents size |
{'	contents, size = self:handle():read(container:json(),bytes) '}.
	^Array with: contents with: size! !

LoveObject subclass: #Rasterizer!
Rasterizer buildFunctions: #(
	getAdvance
	getAscent
	getDescent
	getGlyphCount
	getGlyphData
	getHeight	Get
	getLineHeight
	hasGlyphs
)!

Font buildLuaStaticFunctions: #(
	#(newBMFontRasterizer #(_imageData glyphs) Rasterizer)
	#(newFontData #(_rasterizer) FontData)
	#(newGlyphData #(_rasterizer glyph) GlyphData)
	#(newRasterizer #(_data) Rasterizer)
	#(newRasterizer #(size mode) Rasterizer)
	#(newRasterizer #(_imageData glyphs) Rasterizer)
	#(newTrueTypeRasterizer #(size mode) Rasterizer)
)!
Font buildFunctions: #(
	getAscent getBaseline
	getDPIScale 
	getDescent
	getHeight getLineHeight
	getWidth
	#(hasGlyphs #(text)) #(hasGlyphs #(c1 c2))
	#(setFallbacks #(font))
	#(setFallbacks #(font font2))
	#(setFallbacks #(font font2 font3))
	#(setFilter #(min mag))
	#(setFilter #(min mag anisotropy))
	#(setLineHeight #(height))
)!

! Font methodsFor: #accessing !
extent
	^self width @ self height! !

! Font methodsFor: #accessing !
wrap: text limit: wraplimit
	| width wrappedtext |
{'	width, wrappedtext = self:handle():getWrap( text, wraplimit )'}.
	^Array with: width with: wrappedtext! !

! Font methodsFor: #accessing !
filter
	| min mag anisotropy |
{'	min, mag, anisotropy = self:handle():getFilter( )'}.
	^Array with: min with: mag with: anisotropy! !

LoveObject subclass: #Drawable!
#(	Framebuffer Texture
	ParticleSystem SpriteBatch
	Mesh Text Video
) do: [:each|
	(Smalltalk includesKey: each) ifTrue: [
		self error: 'Class ' ,each ,' already defined.'.
		^self cancelFileIn
	].
	Drawable subclass: each
]!
#(	Canvas Image
) do: [:each|
	(Smalltalk includesKey: each) ifTrue: [
		self error: 'Class ' ,each ,' already defined.'.
		^self cancelFileIn
	].
	Texture subclass: each
]!

Framebuffer buildFunctions: #(
	#(getImageData ImageData)
	#(renderTo #(aBlock))
	#(setWrap #(horiz vert))
)!

! Framebuffer methodsFor: #accessing !
wrap
	| horiz vert |
{'	horiz, vert = self:handle():getWrap()'}.
	^Array with: horiz with: vert! !

Texture buildFunctions: #(
	getDPIScale
	getDepth getHeight getWidth
	getDepthSampleMode
	getFormat getLayerCount
	getMipmapCount
	getPixelHeight getPixelWidth
	getTextureType
	isReadable
	setDepthSampleMode
	#(setDepthSampleMode #(compare))
	#(setFilter #(min mag))
	#(setFilter #(min mag anisotropy))
	#(setMipmapFilter #(filtermode))
	#(setMipmapFilter #(filtermode sharpness))
	#(setWrap #(horiz vert))
)!

! Texture methodsFor: #accessing !
extent
	^self width @ self height! !

! Texture methodsFor: #accessing !
wrap
	| horiz vert |
{'	horiz, vert = self:handle():getWrap()'}.
	^Array with: horiz with: vert! !

! Texture methodsFor: #accessing !
mipmapFilter
	| mode sharpness |
{'	mode, sharpness = self:handle():getMipmapFilter()'}.
	^Array with: mode with: sharpness! !

! Texture methodsFor: #accessing !
filter
	| min mag anisotropy |
{'	min, mag, anisotropy = self:handle():getFilter()'}.
	^Array with: min with: mag with: anisotropy! !

! Texture methodsFor: #accessing !
pixelDimensions
	| width height |
{'	width, height = self:handle():getPixelDimensions()'}.
	^Array with: width with: height! !

! Texture methodsFor: #accessing !
dimensions
	| width height |
{'	width, height = self:handle():getDimensions()'}.
	^Array with: width with: height! !

Canvas buildFunctions: #(
	#(clear #(red green blue))
	#(clear #(red green blue alpha))
	#(clear #(_rgba))
	clear
	getFSAA getMSAA getFormat
	#(getImageData ImageData)
	#(newImageData ImageData)
	#(newImageData #(slice mipmap x y width height) ImageData)
	#(newImageData #(x y width height) ImageData)
	#(renderTo #(aBlock))
)!

! Canvas methodsFor: #accessing !
getPixel: aPoint
	| r g b a |
{'	r, g, b, a = self:handle():getPixel(aPoint:x(),aPoint:y())'}.
	^Array with: r with: g with: b with: a! !

Image buildFunctions: #(
	#(getData ImageData)
	#(getFlags PoolDictionary)
	isCompressed
	refresh #(refresh #(x y width height))
	#(replacePixels #(_data slice mipmap x y reloadmipmaps))
	#(replacePixels #(_data slice mipmap x y))
	#(replacePixels #(_data slice mipmap))
	#(replacePixels #(_data slice))
)!

Mesh buildFunctions: #(
	#(attachAttribute:from: #(name _mesh))
	#(detachAttribute #(name))
	flush
	getDrawMode
	getDrawRange
	#(getImage Image)
	#(getTexture Texture)
	#(getVertex #(index))
	#(getVertexAttribute #(vertex index))
	getVertexCount
	#(getVertexFormat 'Array[Array]')
	#(getVertexMap Lua)
	#(getVertices 'Array[Array]')
	hasVertexColors
	#(isAttributeEnabled #(name))
	#(setAttributeEnabled #(name enable))
	#(setDrawMode #(mode))
	setDrawRange #(setDrawRange #(start count))
	#(setImage #(_image))
	#(setTexture #(_texture))
	#(setVertex #(index component x y u v r g b a))
	#(setVertex #(index _vertex))
	#(setVertex #(index x y u v r g b a))
	#(setVertexAttribute #(vertex index value1))
	#(setVertexAttribute #(vertex index value1 value2))
	#(setVertexAttribute #(vertex index value1 value2 value3))
	#(setVertexAttribute #(vertex index value1 value2 value3 value4))
	#(setVertexColors #(aBoolean))
	#(setVertexMap #(_map))
	#(setVertexMap #(vi1 vi2 vi3 vi4 vi5 vi6 vi7 vi8))
	#(setVertexMap #(_data datatype))
	#(setVertices #(_vertices startvertex))
	#(setVertices #(vertices))
)!

! Mesh methodsFor: #access !
vertex: index component: attributecomponent with: array

	^(self handle basicAt: #setVertex)
		evaluateWithArguments: (Array
			with: self with: index
			with: attributecomponent) ,array! !

! Mesh methodsFor: #access !
vertexAttribute: vertexindex index: attributeindex with: array

	^(self handle basicAt: #setVertexAttribute)
		evaluateWithArguments: (Array
			with: self with: vertexindex
			with: attributeindex) ,array! !

! Mesh methodsFor: #access !
vertexMap: array

	^(self handle basicAt: #setVertexMap)
		evaluateWithArguments: (Array with: self) ,array! !

! Mesh methodsFor: #access !
drawRange
	| min max |
{'	min, max = self:handle():getDrawRange() '}.
	^Array with: min with: max! !

! Mesh methodsFor: #access !
vertex: index
	| x y u v r g b a |
{'	x, y, u, v, r, g, b, a = self:handle():getVertex(index) '}.
	^(Array with: x with: y with: u with: v)
		,(Array with: r with: g with: b with: a)! !

! Mesh methodsFor: #access !
vertex: index attribute: attIndex
	| v1 v2 v3 v4 v5 v6 v7 v8 |
{'	v1, v2, v3, v4, v5, v6, v7, v8 = self:handle():getVertexAttribute(index,attIndex) '}.
	^(Array with: v1 with: v2 with: v3 with: v4)
		,(Array with: v5 with: v6 with: v7 with: v8)! !


ParticleSystem buildFunctions: #(
	#(clone ParticleSystem)
	count
	#(emit #(numparticles))
	getBufferSize
	getColors getCount
	getDirection
	getEmissionRate
	getEmitterLifetime
	#(getImage Image)
	getInsertMode
	getOffsetX getOffsetY
	#(getQuads Array)
	getSizeVariation
	getSpinVariation
	getSpread
	#(getTexture Texture)
	getX getY
	hasRelativeRotation
	isActive isEmpty isFull
	isPaused isStopped
	#(moveTo #(x y))
	pause reset
	#(setAreaSpread #(distribution dx dy))
	#(setBufferSize #(size))
	#(setColor #(r1 g1 b1 a1 r2 g2 b2 a2))
	#(setDirection #(direction))
	#(setEmissionArea #(distribution dx dy))
	#(setEmissionArea #(distribution dx dy angle))
	#(setEmissionArea #(distribution dx dy angle relative))
	#(setEmissionRate #(rate))
	#(setEmitterLifetime #(seconds))
	#(setGravity #(min))
	#(setGravity #(min max))
	#(setImage #(_image))
	#(setInsertMode #(mode))
	#(setLifetime #(seconds))
	#(setLinearAcceleration #(xmin ymin))
	#(setLinearAcceleration #(xmin ymin xmax ymax))
	#(setLinearDamping #(min))
	#(setLinearDamping #(min max))
	#(setOffset #(x y))
	#(setParticleLife #(min))
	#(setParticleLife #(min max))
	#(setParticleLifetime #(min))
	#(setParticleLifetime #(min max))
	#(setPosition #(x y))
	#(setQuads #(quads))
	#(setRadialAcceleration #(min))
	#(setRadialAcceleration #(min max))
	#(setRelativeRotation #(enable))
	#(setRotation #(min))
	#(setRotation #(min max))
	#(setSize #(min))
	#(setSize #(min max))
	#(setSize #(min max variation))
	#(setSizeVariation #(variation))
	#(setSpeed #(min))
	#(setSpeed #(min max))
	#(setSpin #(min))
	#(setSpin #(min max))
	#(setSpinVariation #(variation))
	#(setSpread #(spread))
	#(setSprite #(_anImage))
	#(setTangentialAcceleration #(min))
	#(setTangentialAcceleration #(min max))
	#(setTexture #(_texture))
	start stop
	#(update #(dt))
)!

! ParticleSystem methodsFor: #access !
setSizes: array

	^(self handle basicAt: #setSizes)
		evaluateWithArguments: (Array with: self) ,array! !

! ParticleSystem methodsFor: #access !
setColors: array

	^(self handle basicAt: #setColors)
		evaluateWithArguments: (Array with: self) ,array! !

! ParticleSystem methodsFor: #access !
tangentialAcceleration
	| min max |
{'	min, max = self:handle():getTangentialAcceleration() '}.
	^Array with: min with: max! !

! ParticleSystem methodsFor: #access !
spin
	| min max variation |
{'	min, max, variation = self:handle():getSpin() '}.
	^Array with: min with: max with: variation! !

! ParticleSystem methodsFor: #access !
sizes
	| v1 v2 v3 v4 v5 v6 v7 v8 |
{'	v1, v2, v3, v4, v5, v6, v7, v8 = self:handle():getSizes() '}.
	^(Array with: v1 with: v2 with: v3 with: v4)
		,(Array with: v5 with: v6 with: v7 with: v8)! !

! ParticleSystem methodsFor: #access !
speed
	| min max |
{'	min, max = self:handle():getSpeed() '}.
	^Array with: min with: max! !

! ParticleSystem methodsFor: #access !
rotation
	| min max |
{'	min, max = self:handle():getRotation() '}.
	^Array with: min with: max! !

! ParticleSystem methodsFor: #access !
radialAcceleration
	| min max |
{'	min, max = self:handle():getRadialAcceleration() '}.
	^Array with: min with: max! !

! ParticleSystem methodsFor: #access !
position
	| x y |
{'	x, y = self:handle():getPosition() '}.
	^x@y! !

! ParticleSystem methodsFor: #access !
particleLifetime
	| min max |
{'	min, max = self:handle():getParticleLifetime() '}.
	^Array with: min with: max! !

! ParticleSystem methodsFor: #access !
offset
	| x y |
{'	x, y = self:handle():getOffset() '}.
	^x@y! !

! ParticleSystem methodsFor: #access !
linearDamping
	| min max |
{'	min, max = self:handle():getLinearDamping() '}.
	^Array with: min with: max! !

! ParticleSystem methodsFor: #access !
linearAcceleration
	| xmin ymin xmax ymax |
{'	xmin, ymin, xmax, ymax = self:handle():getLinearAcceleration() '}.
	^Array
		with: xmin@ymin
		with: xmax@ymax! !

! ParticleSystem methodsFor: #access !
emissionArea
	| distribution dx dy angle relative |
{'	distribution, dx, dy, angle, relative = self:handle():getEmissionArea() '}.
	^Array
		with: distribution with: dx@dy
		with: angle with: relative! !

! ParticleSystem methodsFor: #access !
areaSpread
	| distribution dx dy |
{'	distribution, dx, dy = self:handle():getAreaSpread() '}.
	^Array with: distribution with: dx@dy! !

LoveObject subclass: #LoveTarget !
LoveTarget subclass: #PixelEffect !
LoveTarget buildFunctions: #(
	getWarnings
	#(send #(name _value))
	#(send #(name _value1 _value2))
	#(send #(name _value1 _value2 _value3))
	#(send #(name _value1 _value2 _value3 _value4))
	#(send #(name array))
	#(send #(name matrix))
)!
LoveObject subclass: #Quad !
Quad buildFunctions: #(
	#(flip #(x y))
	#(setViewport #(x y w h))
)!

! Quad methodsFor: #access !
textureDimensions
	| x y |
{'	x, y = self:handle():getTextureDimensions() '}.
	^x@y! !

! Quad methodsFor: #access !
viewport: aRectangle
	^self viewport: aRectangle left y: aRectangle top
		w: aRectangle width h: aRectangle height! !

! Quad methodsFor: #access !
viewport
	| x y w h |
{'	x, y, w, h = self:handle():getViewport() '}.
	^x@y extent: w@h! !

LoveTarget subclass: #Shader !
Shader buildFunctions: #(
	#(hasUniform #(name))
	#(sendColor #(rgbaArray))
)!

! Shader methodsFor: #access !
externVariable: name
	| t components count |
{'	t, components, count = self:handle():getExternVariable(name) '}.
	^Array with: t with: components with: count! !

LoveObject subclass: #Transform!
Transform buildFunctions: #(
	#(apply #(_other) Transform)
	#(close Transform)
	#(inverse Transform)
	isAffine2DTransform
	#(reset Transform)
	#(rotate #(angle) Transform)
	#(scale #(x y) Transform)
	#(setMatrix #(array) Transform)
	#(setTransformation #(x y angle sx sy ox oy kx ky) Transform)
	#(shear #(x y) Transform)
	#(translate #(x y) Transform)
)!

! Transform methodsFor: #accessing !
matrix: array
	(self handle basicAt: #setMatrix) evaluateWithArguments: (Array with: self) ,array! !

! Transform methodsFor: #accessing !
matrix: layout withAll: array
	(self handle basicAt: #setMatrix) evaluateWithArguments: (Array with: self with: layout) ,array! !

! Transform methodsFor: #accessing !
matrix
	| e11 e12 e13 e14  e21 e22 e23 e24
	  e31 e32 e33 e34  e41 e42 e43 e44 |
{'	e11, e12, e13, e14, e21, e22, e23, e24, e31, e32, e33, e34, e41, e42, e43, e44 = self:handle():getMatrix() '}.
	^(Array with: e11 with: e12 with: e13 with: e14)
	 ,(Array with: e21 with: e22 with: e23 with: e24)
	 ,(Array with: e31 with: e32 with: e33 with: e34)
	 ,(Array with: e41 with: e42 with: e43 with: e44)! !

! Transform methodsFor: #accessing !
inverseTransformPoint: aPoint
	| x y |
{'	x, y = self:handle():inverseTransformPoint(aPoint:x(),aPoint:y()) '}.
	^x@y! !

! Transform methodsFor: #accessing !
transformPoint: aPoint
	| x y |
{'	x, y = self:handle():transformPoint(aPoint:x(),aPoint:y()) '}.
	^x@y! !

SpriteBatch buildFunctions: #(
	#(add #(x y r sx sy ox oy kx ky))
	#(add #(quad x y r sx sy ox oy kx ky))
	#(addLayer #(layerindex x y r sx sy ox oy kx ky))
	#(addLayer #(layerindex _transform))
	#(addLayer #(layerindex quad _transform))
	#(addq #(_quad x y r sx sy ox oy kx ky))
	#(attachAttribute #(name _mesh))
	bind unbind clear flush
	getBufferSize getCount
	#(getImage Image)
	#(getTexture Texture)
	#(set #(spriteindex x y r sx sy ox oy kx ky))
	#(set #(spriteindex _quad x y r sx sy ox oy kx ky))
	#(setBufferSize #(size))
	#(setColor #(r g b a)) setColor
	#(setDrawRange #(start count)) setDrawRange
	#(setImage #(_image))
	#(setLayer #(spriteindex layerindex x y r sx sy ox oy kx ky))
	#(setLayer #(spriteindex layerindex _quad x y r sx sy ox oy kx ky))
	#(setLayer #(spriteindex layerindex _quad _transform))
	#(setTexture #(_texture))
	#(setq #(id _quad x y r sx sy ox oy kx ky))
)!

! SpriteBatch methodsFor: #accessing !
color
	| r g b a |
{'	r, g, b, a = self:handle():getColor() '}.
	^Array with: r with: g with: b with: a! !

! SpriteBatch methodsFor: #accessing !
color: array
	(self handle basicAt: #setColor) evaluateWithArguments: (Array with: self) ,array! !

Text buildFunctions: #(
	#(add #(aStringOrColoredText x y angle sx sy ox oy kx ky))
	#(addf #(aStringOrColoredText wraplimit alignmode x y angle sx sy ox oy kx ky))
	clear
	#(getFont Font)
	getHeight #(getHeight #(index))
	getWidth #(getWidth #(index))
	set #(set #(_aStringOrColoredText))
	#(setFont #(_font))
	#(setf #(aStringOrColoredText wraplimit alignmode))
)!

! Text methodsFor: #accessing !
extent
	^self width @ self height! !

! Text methodsFor: #accessing !
dimensions
	| x y |
{'	x, y = self:handle():getDimensions() '}.
	^x@y! !

! Text methodsFor: #accessing !
dimensions: index
	| x y |
{'	x, y = self:handle():getDimensions(index) '}.
	^x@y! !

LoveObject subclass: #Source!
Source buildFunctions: #(
	getAirAbsorption
	getChannelCount
	getChannels
	#(getDuration #(unit))
	#(getEffect #(name _filtersettings) Lua)
	#(getFilter Lua)
	getFreeBufferCount getPitch
	getRolloff getType
	getVolume
	isLooping isPaused isPlaying
	isRelative isStatic isStopped
	pause play
	#(queue #(_sounddata))
	resume reqind
	#(seek #(offset))
	#(seek #(offset unit))
	#(setAirAbsorption #(amount))
	#(setAttenuationDistances #(ref max))
	#(setCone #(innerAngle outerAngle outerVolume))
	#(setDirection #(x y z))
	#(setDistance #(ref max))
	#(setEffect #(name enable))
	#(setFilter #(_settings))
	#(setLooping #(aBoolean))
	#(setPitch #(pitch))
	#(setPosition #(x y z))
	#(setRelative #(enable))
	#(setRolloff #(rolloff))
	#(setVelocity #(x y z))
	#(setVolume #(volume))
	#(setVolumeLimits #(min max))
	stop
	tell #(tell #(unit))
)!

! Source methodsFor: #accessing !
volumeLimits
	| min max |
{'	min, max = self:handle():getVolumeLimits() '}.
	^Array with: min with: max! !

! Source methodsFor: #accessing !
distance
	| ref max |
{'	ref, max = self:handle():getDistance() '}.
	^Array with: ref with: max! !

! Source methodsFor: #accessing !
attenuationDistances
	| ref max |
{'	ref, max = self:handle():getAttenuationDistances() '}.
	^Array with: ref with: max! !

! Source methodsFor: #accessing !
cone
	| innerAngle outerAngle outerVolume |
{'	innerAngle, outerAngle, outerVolume = self:handle():getCone() '}.
	^Array with: innerAngle with: outerAngle with: outerVolume! !

! Source methodsFor: #accessing !
direction
	| x y z |
{'	x, y, z = self:handle():getDirection() '}.
	^Array with: x with: y with: z! !

! Source methodsFor: #accessing !
position
	| x y z |
{'	x, y, z = self:handle():getPosition() '}.
	^Array with: x with: y with: z! !

! Source methodsFor: #accessing !
velocity
	| x y z |
{'	x, y, z = self:handle():getVelocity() '}.
	^Array with: x with: y with: z! !

LoveObject subclass: #VideoStream!

Video buildFunctions: #(
	getHeight
	#(getSource Source)
	#(getStream VideoStream)
	getWidth play rewind
	isPlaying pause
	#(seek #(offset))
	#(setFilter #(min mag))
	#(setFilter #(min mag anisotropy))
	setSource #(setSource #(_source))
	tell
)!

! Video methodsFor: #accessing !
extent
	^self width @ self height! !

! Video methodsFor: #accessing !
dimensions
	| min mag anisotropy |
{'	min, mag, anisotropy = self:handle():getfilter() '}.
	^Array with: min with: mag with: anisotropy! !

! Video methodsFor: #accessing !
dimensions
	| x y |
{'	x, y = self:handle():getDimensions() '}.
	^x@y! !

"Drawing"
Graphics buildLuaStaticFunctions: #(
	#(arc #(drawmode x y radius angle1 angle2))
	#(arc #(drawmode x y radius angle1 angle2 segments))
	#(arc #(drawmode arctype x y radius angle1 angle2))
	#(arc #(drawmode arctype x y radius angle1 angle2 segments))
	#(circle #(mode x y radius))
	#(circle #(mode x y radius segments))
	clear #(clear #(r g b))
	clear #(clear #(r g b a))
	#(clear #(r g b a clearstencil))
	#(clear #(r g b a clearstencil cleardepth))
	#(clear #(clearcolor clearstencil cleardepth))
	discard #(discard #(discardcolor discardstencil))
	#(draw #(_drawable))
	#(draw #(_drawable x))
	#(draw #(_drawable x y))
	#(draw #(_drawable x y r))
	#(draw #(_drawable x y r sx sy))
	#(draw #(_drawable x y r sx sy ox oy))
	#(draw #(_drawable x y r sx sy ox oy kx ky))
	#(draw #(_texture _quad x y))
	#(draw #(_texture _quad x y r))
	#(draw #(_texture _quad x y r sx sy))
	#(draw #(_texture _quad x y r sx sy ox oy))
	#(draw #(_texture _quad x y r sx sy ox oy kx ky))
	#(draw #(_texture _quad _transform))
	#(draw #(_drawable  _transform))
	#(drawInstanced #(_mesh instancecount x y))
	#(drawInstanced #(_mesh instancecount x y r))
	#(drawInstanced #(_mesh instancecount x y r sx sy))
	#(drawInstanced #(_mesh instancecount x y r sx sy ox oy))
	#(drawInstanced #(_mesh instancecount x y r sx sy ox oy kx ky))
	#(drawInstanced #(_mesh instancecount _transform))
	#(drawLayer #(_texture layerindex))
	#(drawLayer #(_texture layerindex x y))
	#(drawLayer #(_texture layerindex x y r))
	#(drawLayer #(_texture layerindex x y r sx sy))
	#(drawLayer #(_texture layerindex x y r sx sy ox oy))
	#(drawLayer #(_texture layerindex x y r sx sy ox oy kx ky))
	#(drawLayer #(_texture layerindex _quad))
	#(drawLayer #(_texture layerindex _quad x y))
	#(drawLayer #(_texture layerindex _quad x y r))
	#(drawLayer #(_texture layerindex _quad x y r sx sy))
	#(drawLayer #(_texture layerindex _quad x y r sx sy ox oy))
	#(drawLayer #(_texture layerindex _quad x y r sx sy ox oy kx ky))
	#(drawq #(_image _quad x y))
	#(drawq #(_image _quad x y r))
	#(drawq #(_image _quad x y r sx sy))
	#(drawq #(_image _quad x y r sx sy ox oy))
	#(drawq #(_image _quad x y r sx sy ox oy kx ky))
	#(ellipse #(mode x y radiusx radiusy))
	#(ellipse #(mode x y radiusx radiusy segments))
	flushBatch
	#(line #(x1 y1 x2 y2))
	#(line #(points))
	#(point #(x y))
	#(points #(points))
	#(polygon #(mode vertices))
	present
	#(print #(_text))
	#(print #(_text x y))
	#(print #(_text x y r))
	#(print #(_text x y r sx sy))
	#(print #(_text x y r sx sy ox oy))
	#(print #(_text x y r sx sy ox oy kx ky))
	#(print #(_text _transform))
	#(printf #(text x y limit))
	#(printf #(text x y limit align))
	#(printf #(text x y limit align r))
	#(printf #(text x y limit align r sx sy))
	#(printf #(text x y limit align r sx sy ox oy))
	#(printf #(text x y limit align r sx sy ox oy kx ky))
	#(printf #(text _font x y limit))
	#(printf #(text _font x y limit align))
	#(printf #(text _font x y limit align r))
	#(printf #(text _font x y limit align r sx sy))
	#(printf #(text _font x y limit align r sx sy ox oy))
	#(printf #(text _font x y limit align r sx sy ox oy kx ky))
	#(quad #(mode x1 y1 x2 y2 x3 y3 x4 y4))
	#(rectangle #(mode x y width height))
	#(rectangle #(mode x y width height rx))
	#(rectangle #(mode x y width height rx ry))
	#(rectangle #(mode x y width height rx ry segments))
	#(stencil #(stencilfunction action))
	#(stencil #(stencilfunction action value))
	#(stencil #(stencilfunction action value keepvalues))
	#(triangle #(mode x1 y1 x2 y2 x3 y3))
)!

"Object Creation"
Graphics buildLuaStaticFunctions: #(
	#(captureScreenshot #(arg))
	#(newArrayImage #(slices) Image)
	#(newArrayImage #(slices _settings) Image)
	#(newCanvas Canvas)
	#(newCanvas #(width height) Canvas)
	#(newCanvas #(width height layers) Canvas)
	#(newCanvas #(width height layers _settings) Canvas)
	#(newCubeImage #(arg) Image)
	#(newCubeImage #(arg _settings) Image)
	#(newFont #(filename size) Font)
	#(newFont #(filename size hinting) Font)
	#(newFont #(size hinting) Font)
	#(newFramebuffer Framebuffer)
	#(newFramebuffer #(width height) Framebuffer)
	#(newImage #(arg) Image)
	#(newImage #(arg _flags) Image)
	#(newImage #(arg _format) Image)
	#(newImageFont #(_arg glyphs) Font)
	#(newMesh #(vertices) Mesh)
	#(newMesh #(vertices mode) Mesh)
	#(newMesh #(vertices mode usage) Mesh)
	#(newParticleSystem #(_arg buffer) ParticleSystem)
	#(newPixelEffect #(code) PixelEffect)
	#(newQuad #(x y width height sw sh) Quad)
	#(newScreenshot ImageData)
	#(newScreenshot #(copyAlpha) ImageData)
	#(newShader #(code) Shader)
	#(newShader #(pixelcode vertexcode) Shader)
	#(newStencil #(stencilFunction))
	#(newSpriteBatch #(_image maxsprites) SpriteBatch)
	#(newSpriteBatch #(_image maxsprites usage) SpriteBatch)
	#(newText #(_font) Text)
	#(newText #(_font textstring) Text)
	#(newVideo #(_arg) Video)
	#(newVideo #(_arg _settings) Video)
	#(newVolumeImage #(#(layers handles)) Image)
	#(newVolumeImage #(#(layers handles) _settings) Image)
	#(setNewFont #(size) Font)
	#(setNewFont #(_arg size) Font)
)!

! Graphics class methodsFor: #shader !
validateShader: gles code: code
	| status message |
{'	status, message = love.graphics.validateShader(gles, code) '}.
	^Array with: status with: message! !

! Graphics class methodsFor: #shader !
validateShader: gles pixelcode: pixelcode vertexcode: vertexcode
	| status message |
{'	status, message = love.graphics.validateShader(gles, pixelcode, vertexcode) '}.
	^Array with: status with: message! !


! Graphics class methodsFor: #accessing !
backgroundColor
	| r g b a |
{'	r, g, b, a = love.graphics.getBackgroundColor( ) '}.
	^Array with: r with: g with: b with: a! !

! Graphics class methodsFor: #accessing !
blendMode
	| mode alphamode |
{'	mode, alphamode = love.graphics.getBlendMode( ) '}.
	^Array with: mode with: alphamode! !

! Graphics class methodsFor: #accessing !
color
	| r g b a |
{'	r, g, b, a = love.graphics.getColor( ) '}.
	^Array with: r with: g with: b with: a! !

! Graphics class methodsFor: #accessing !
colorMask
	| r g b a |
{'	r, g, b, a = love.graphics.getColorMask( ) '}.
	^Array with: r with: g with: b with: a! !

! Graphics class methodsFor: #accessing !
defaultFilter
	| min mag anisotropy |
{'	min, mag, anisotropy = love.graphics.getDefaultFilter( ) '}.
	^Array with: min with: mag with: anisotropy! !

! Graphics class methodsFor: #accessing !
defaultImageFilter
	| min mag |
{'	min, mag = love.graphics.getDefaultImageFilter( ) '}.
	^Array with: min with: mag! !

! Graphics class methodsFor: #accessing !
depthMode
	| comparemode write |
{'	comparemode, write = love.graphics.getDepthMode( ) '}.
	^Array with: comparemode with: write! !

! Graphics class methodsFor: #accessing !
lineStipple
	| pattern rpt |
{'	pattern, rpt = love.graphics.getLineStipple( ) '}.
	^Array with: pattern with: rpt! !

! Graphics class methodsFor: #accessing !
scissor
	| x y width height |
{'	x, y, width, height = love.graphics.getScissor( ) '}.
	^x@y extent: width@height! !

! Graphics class methodsFor: #accessing !
stencilTest
	| mode value |
{'	mode, value = love.graphics.getStencilTest( ) '}.
	^Array with: mode with: value! !

! Graphics class methodsFor: #accessing !
intersectScissor: aRectangle
	^self intersectScissor: aRectangle left
		y: aRectangle top
		width: aRectangle width
		height: aRectangle height! !

! Graphics class methodsFor: #accessing !
scissor: aRectangle
	^self scissor: aRectangle left
		y: aRectangle top
		width: aRectangle width
		height: aRectangle height! !

"Graphics State"
Graphics buildLuaStaticFunctions: #(
	#(getCanvas Canvas)
	getColorMode
	#(getFont Font)
	getFrontFaceWinding
	getLineJoin getLineStyle getLineWidth
	#(getMeshCullMode #(mode))
	getPixelEffect getPointSize getPointStyle
	#(getShader Shader)
	getStackDepth
	#(intersectScissor #(x y width height))
	isActive isGammaCorrect
	#(isSupported #(feature))
	isWireframe reset
	#(setBackgroundColor #(red green blue))
	#(setBackgroundColor #(red green blue alpha))
	#(setBackgroundColor #(rgba))
	#(setBlendMode #(mode))
	#(setBlendMode #(mode alphamode))
	setCanvas
	#(setCanvas #(_canvas))
	#(setCanvas #(_canvas mipmap))
	#(setColor #(rgba))
	#(setColor #(red green blue))
	setColorMask #(setColorMask #(red green blue alpha))
	#(setColorMode #(mode))
	#(setDefaultFilter #(min mag))
	#(setDefaultFilter #(min mag anisotropy))
	#(setDefaultImageFilter #(min mag))
	#(setDepthMode #(comparemode write))
	#(setFont #(filename size))
	#(setFont #(_arg))
	setFrontFaceWinding
	#(setInvertedStencil #(aBlock))
	#(setLine #(width))
	#(setLine #(width style))
	#(setLineJoin #(_join))
	#(setLineStipple #(_pattern))
	#(setLineStipple #(_pattern count))
	#(setLineStyle #(style))
	#(setLineWidth #(width))
	#(setMeshCullMode #(_mode))
	#(setPixelEffect #(_pixeleffect))
	#(setPoint #(size style))
	#(setPointSize #(size))
	#(setPointStyle #(style))
	setRenderTarget #(setRenderTarget #(_framebuffer))
	#(setScissor #(x y width height))
	setShader #(setShader #(_shader))
	setStencil #(setStencil #(aBlock))
	setStencilTest #(setStencilTest #(comparemode comparevalue))
	#(setWireframe #(enable))
)!

! Graphics class methodsFor: #transform !
inverseTransformPoint: aPoint
	| x y |
{'	x, y = love.graphics.inverseTransformPoint( aPoint:x(), aPoint:y() ) '}.
	^x@y! !

! Graphics class methodsFor: #transform !
transformPoint: aPoint
	| x y |
{'	x, y = love.graphics.transformPoint( aPoint:x(), aPoint:y() ) '}.
	^x@y! !

"Coordinate System"
Graphics buildLuaStaticFunctions: #(
	#(applyTransform #(_transform))
	origin pop push
	#(replaceTransform #(_transform))
	#(rotate  #(angle))
	#(scale #(scale))
	#(scale #(x y))
	#(shear #(k))
	#(shear #(x y))
	#(translate #(x y))
)!

! Graphics class methodsFor: #accessing !
dimensions
	| x y |
{'	x, y = love.graphics.getDimensions() '}.
	^x@y! !

! Graphics class methodsFor: #accessing !
pixelDimensions
	| x y |
{'	x, y = love.graphics.getPixelDimensions() '}.
	^x@y! !

! Graphics class methodsFor: #accessing !
mode
	| width height fullscreen vsync fsaa |
{'	width, height, fullscreen, vsync, fsaa = love.graphics.getMode( ) '}.
	^Array with: width @ height with: fullscreen with: vsync with: fsaa! !

! Graphics class methodsFor: #accessing !
modes
	| modes |
{'	modes = love.graphics.getModes( ) '}.
	^(Array @ modes) collect: [:tuple| PoolDictionary @ tuple ]! !

"Window"
Graphics buildLuaStaticFunctions: #(
	#(checkMode #(width height))
	#(checkMode #(width height fullscreen))
	getCaption getDPIScale
	getHeight getWidth
	getPixelHeight getPixelWidth
	hasFocus isCreated
	#(setCaption #(caption))
	#(setIcon #(_image))
	#(setMode #(width height))
	#(setMode #(width height fullscreen))
	#(setMode #(width height fullscreen vsync))
	#(setMode #(width height fullscreen vsync fsaa))
	toggleFullscreen
)!

! Graphics class methodsFor: #accessing !
extent
	^self width @ self height! !

! Graphics class methodsFor: #accessing !
rendererInfo
	| aName version vendor device |
{'	aName, version, vendor, device = love.graphics.getRendererInfo( ) '}.
	^Array	with: aName with: version
		with: vendor with: device! !

! Graphics class methodsFor: #accessing !
imageFormats
	| formats |
{'	formats = love.graphics.getImageFormats( ) '}.
	^PoolDictionary @ formats! !

! Graphics class methodsFor: #accessing !
compressedImageFormats
	| formats |
{'	formats = love.graphics.getCompressedImageFormats( ) '}.
	^PoolDictionary @ formats! !

! Graphics class methodsFor: #accessing !
canvasFormats
	| formats |
{'	formats = love.graphics.getCanvasFormats( ) '}.
	^PoolDictionary @ formats! !

"System Information"
Graphics buildLuaStaticFunctions: #(
	getMaxImageSize
	getMaxPointSize
	#(getStats PoolDictionary)
	#(getSupported PoolDictionary)
	#(getSystemLimit #(limittype))
	#(getSystemLimits PoolDictionary)
	#(getTextureTypes PoolDictionary)
)!

Image buildLuaStaticFunctions: #(
	#(isCompressed #(_file))
	#(newCompressedData #(_file) CompressedImageData)
	#(newImageData #(width height) ImageData)
	#(newImageData #(width height format) ImageData)
	#(newImageData #(width height format _data) ImageData)
)!


! Joystick methodsFor: #accessing !
axes
	| dir1 dir2 dir3 dir4 |
{'	dir1, dir2, dir3, dir4 = self:handle():getAxes( ) '}.
	^Array with: dir1 with: dir2 with: dir3 with: dir4! !

! Joystick methodsFor: #accessing !
id
	| id instanceid |
{'	id, instanceid = self:handle():getID( ) '}.
	^Array with: id with: instanceid! !

! Joystick methodsFor: #accessing !
gamepadMapping: axis
	| inputtype inputindex hatdirection |
{'	inputtype, inputindex, hatdirection = self:handle():getGamepadMapping( axis ) '}.
	^Array with: inputtype with: inputindex with: hatdirection! !

! Joystick methodsFor: #accessing !
vibration
	| left right |
{'	left, right = self:handle():getVibration( ) '}.
	^Array with: left with: right! !

Joystick buildFunctions: #(
	#(getAxis #(axis))
	getAxisCount
	getButtonCount
	getGUID
	#(getGamepadAxis #(axis))
	#(getHat #(hat))
	getHatCount
	getName
	isConnected
	#(isDown #(button))
	isGamepad
	#(isGamepadDown #(button))
	isVibrationSupported
	#(setVibration #(left right))
	#(setVibration #(left right duration))
)!
Joystick buildLuaStaticFunctions: #(
	#(close #(joystick))
	#(getAxis #(joystick axis))
	getJoystickCount getNumJoysticks
	#(getJoysticks 'Array[Joystick]')
	#(loadGamepadMappings #(aString))
	#(saveGamepadMappings #(filename))
	saveGamepadMappings
	#(setGamepadMapping #(guid button inputtype inputindex hatdir))
	#(setGamepadMapping #(guid axis inputtype inputindex hatdir))
)!


! Keyboard class methodsFor: #accessing !
keyRepeat
	| delay interval |
{'	delay, interval = love.keyboard.getKeyRepeat( ) '}.
	^Array with: delay with: interval! !

Keyboard buildLuaStaticFunctions: #(
	#(getKeyFromScancode #(scancode))
	#(getScancodeFromKey #(key))
	hasKeyRepeat
	hasTextInput
	#(isDown #(key))
	#(isScancodeDown #(scancode))
	#(setKeyRepeat #(enable))
	#(setTextInput #(enable))
	#(setTextInput #(enable x y w h))
)!

! Math class methodsFor: #accessing !
gammaToLinear: r g: g b: b
	| lr lg lb |
{'	lr, lg, lb = love.math.gammaToLinear( r, g, b ) '}.
	^Array with: lr with: lg with: lb! !

! Math class methodsFor: #accessing !
linearToGama: lr g: lg b: lb
	| r g b |
{'	r, g, b = love.math.linearToGama( lr, lg, lb ) '}.
	^Array with: r with: g with: b! !

! Math class methodsFor: #accessing !
randomSeed
	| low high |
{'	low, high = love.math.getRandomSeed( ) '}.
	^Array with: low with: high! !

! Math class methodsFor: #accessing !
gammaToLinear: color
	| lr lg lb |
{'	lr, lg, lb = love.math.gammaToLinear( color:json() ) '}.
	^Array with: lr with: lg with: lb! !

LoveObject subclass: #BezierCurve!

! BezierCurve methodsFor: #accessing !
evaluate: t
	| x y |
{'	x, y = self:handle():evaluate(t) '}.
	^Array with: x with: y! !

! BezierCurve methodsFor: #accessing !
controlPoint: i
	| x y |
{'	x, y = self:handle():getControlPoint(i) '}.
	^Array with: x with: y! !

BezierCurve buildFunctions: #(
	getControlPointCount
	getDegree
	#(getDerivative BezierCurve)
	#(getSegment #(startpoint endpoint) BezierCurve)
	#(insertControlPoint #(x y))
	#(insertControlPoint #(x y i))
	#(removeControlPoint #(index))
	#(render Array) #(render #(depth) Array)
	#(renderSegment #(startpoint endpoint) Array)
	#(renderSegment #(startpoint endpoint depth) Array)
	#(rotate #(angle))
	#(rotate #(angle ox oy))
	#(scale #(s))
	#(scale #(s ox oy))
	#(setControlPoint #(i x y))
	#(translate #(dx dy))
)!

LoveObject subclass: #RandomGenerator!

! RandomGenerator methodsFor: #accessing !
seed
	| x y |
{'	x, y = self:handle():getSeed( ) '}.
	^Array with: x with: y! !

RandomGenerator buildFunctions: #(
	getState random
	#(randomNormal #(stddev mean))
	#(setSeed #(seed)) #(setSeed #(low high))
	#(setState #(state))
)!

Math buildLuaStaticFunctions: #(
	#(compress #(_arg) CompressedData)
	#(compress #(_arg format) CompressedData)
	#(compress #(_arg format level) CompressedData)
	#(decompress #(_compressedData))
	#(decompress #(_data format))
	#(isConvex #(vertices))
	#(newBezierCurve #(vertices) BezierCurve)
	#(newRandomGenerator RandomGenerator)
	#(newRandomGenerator #(seed) RandomGenerator)
	#(newRandomGenerator #(low high) RandomGenerator)
	#(newTransform Transform)
	#(newTransform #(x y angle sx sy ox oy kx ky) Transform)
	#(noise #(x))
	#(noise #(x y))
	#(noise #(x y z))
	#(noise #(x y z w))
	random #(random #(max)) #(random #(min max))
	#(randomNormal #(stddev mean))
	#(setRandomSeed #(x))
	#(setRandomSeed #(low high))
	#(triangulate #(polygon) 'Array[Array]')
)!

LoveObject subclass: #Cursor!
Cursor buildFunctions: #(
	getType
)!

! Mouse class methodsFor: #accessing !
position
	| x y |
{'	x, y = love.mouse.getPosition( ) '}.
	^x@y! !

! Mouse class methodsFor: #accessing !
position: aPoint
	^self position: aPoint x y: aPoint y! !

Mouse buildLuaStaticFunctions: #(
	#(getCursor Cursor)
	getRelativeMode
	#(getSystemCursor #(ctype) Cursor)
	getX getY
	hasCursor isCursorSupported
	#(isDown #(button))
	isGrabbed
	isVisible
	#(newCursor #(_imageData hotx hoty) Cursor)
	#(newCursor #(_imageData) Cursor)
	#(setGrab #(grab)) #(setGrabbed #(grab))
	#(setPosition #(x y))
	#(setRelativeMode #(enable))
	#(setVisible #(visible))
	#(setX #(x)) #(setY #(y))
)!

LoveObject subclass: #Body!

! Body methodsFor: #accesing !
linearVelocity
	| x y |
{'	x, y = self:handle():getLinearVelocity( ) '}.
	^x@y! !

! Body methodsFor: #accesing !
linearVelocityFromLocalPoint: aPoint
	| x y |
{'	x, y = self:handle():getLinearVelocityFromLocalPoint( aPoint:x(), aPoint:y() ) '}.
	^x@y! !

! Body methodsFor: #accesing !
linearVelocityFromWorldPoint: aPoint
	| x y |
{'	x, y = self:handle():getLinearVelocityFromWorldPoint( aPoint:x(), aPoint:y() ) '}.
	^x@y! !

! Body methodsFor: #accesing !
localCenter
	| x y |
{'	x, y = self:handle():getLocalCenter() '}.
	^x@y! !

! Body methodsFor: #accesing !
localPoint: aPoint
	| x y |
{'	x, y = self:handle():getLocalPoint( aPoint:x(), aPoint:y() ) '}.
	^x@y! !

! Body methodsFor: #accesing !
localVector: aPoint
	| x y |
{'	x, y = self:handle():getLocalVector( aPoint:x(), aPoint:y() ) '}.
	^x@y! !

! Body methodsFor: #accesing !
massdata
	| x y mass inertia |
{'	x, y, mass, inertia = Body:getMassData( ) '}.
	^Array with: x@y with: mass with: inertia! !

! Body methodsFor: #accesing !
position
	| x y |
{'	x, y = self:handle():getPosition() '}.
	^x@y! !

! Body methodsFor: #accesing !
worldCenter
	| x y |
{'	x, y = self:handle():getWorldCenter() '}.
	^x@y! !

! Body methodsFor: #accesing !
worldPoint: aPoint
	| x y |
{'	x, y = self:handle():getWorldPoint( aPoint:x(), aPoint:y() ) '}.
	^x@y! !

! Body methodsFor: #accesing !
worldVector: aPoint
	| x y |
{'	x, y = self:handle():getWorldVector( aPoint:x(), aPoint:y() ) '}.
	^x@y! !

Body buildFunctions: #(
	#(applyAngularImpulse #(impulse))
	#(applyForce #(fx fy))
	#(applyForce #(fx fy x y))
	#(applyImpulse #(ix iy))
	#(applyImpulse #(ix iy x y))
	#(applyLinearImpulse #(ix iy))
	#(applyLinearImpulse #(ix iy x y))
	#(applyTorque #(torque))
	destroy
	getAllowSleeping
	getAngle
	getAngularDamping
	getAngularVelocity
	#(getContactList 'Array[Contact]')
	#(getFixtureList 'Array[Fixture]')
	getGravityScale
	getInertia
	#(getJointList 'Array[Joint]')
	getLinearDamping
	getMass getType getUserData
	#(getWorld World)
	getX getY
	isActive isAwake isBullet isDestroyed isDynamic
	isFixedRotation isFrozen isSleeping isSleepingAllowed
	isStatic #(isTouching #(_body))
	putToSleep
	resetMassData
	#(setActive #(aBoolean))
	#(setAllowSleeping #(aBoolean))
	#(setAngle #(angle))
	#(setAngularDamping #(aBoolean))
	#(setAngularVelocity #(w))
	#(setAwake #(aBoolean))
	#(setBullet #(aBoolean))
	#(setFixedRotation #(aBoolean))
	#(setGravityScale #(aBoolean))
	#(setInertia #(value))
	#(setLinearDamping #(value))
	#(setActive #(aBoolean))
	#(setLinearVelocity #(x y))
	#(setMass #(mass))
	#(setMassData #(x y mass inertia))
	setMassFromShapes
	#(setPosition #(x y))
	#(setSleepingAllowed #(allowed))
	#(setType #(type))
	#(setUserData #(value))
	#(setX #(x)) #(setY #(y))
	wakeUp
)!

LoveObject subclass: #Contact !

! Contact methodsFor: #accesing !
children
	| a b |
{'	a, b = self:handle():getChildren( ) '}.
	^Array with: a with: b! !

! Contact methodsFor: #accesing !
fixtures
	| a b |
{'	a, b = self:handle():getFixtures( ) '}.
	^Array with: a with: b! !

! Contact methodsFor: #accesing !
normal
	| x y |
{'	x, y = self:handle():getNormal( ) '}.
	^x@y! !

! Contact methodsFor: #accesing !
position
	| x y |
{'	x, y = self:handle():getPosition( ) '}.
	^x@y! !

! Contact methodsFor: #accesing !
positions
	| x1 y1 x2 y2 |
{'	x1, y1, x2, y2 = self:handle():getPositions( ) '}.
	^Array with: x1@y1 with: x2@y2! !

! Contact methodsFor: #accesing !
velocity
	| x y |
{'	x, y = self:handle():getVelocity( ) '}.
	^x@y! !

Contact buildFunctions: #(
	getFriction
	getRestitution
	getSeparation
	isEnabled
	isTouching
	resetFriction
	resetRestitution
	#(setEnabled #(enabled))
	#(setFriction #(friction))
	#(setRestitution #(restitution))
)!
LoveObject subclass: #LoveForm !
LoveForm subclass: #Shape !
LoveForm subclass: #Fixture !

! LoveForm methodsFor: #accesing !
boundingBox
	| left top right bottom |
{'	left, top, right, bottom = self:handle():getBoundingBox( ) '}.
	^left@top rightBottom: right@bottom! !

! LoveForm methodsFor: #accesing !
category
	| v1 v2 v3 v4 |
{'	v1, v2, v3, v4 = self:handle():getCategory( ) '}.
	^Array with: v1 with: v2 with: v3 with: v4! !

! LoveForm methodsFor: #accesing !
computeMass: density
	| x y mass inertia |
{'	x, y, mass, inertia = self:handle():computeMass(density) '}.
	^Array with: x@y with: mass with: inertia! !

! LoveForm methodsFor: #accesing !
filterData
	| categories mask group |
{'	categories, mask, group = self:handle():getFilterData( ) '}.
	^Array with: categories with: mask with: group! !

! LoveForm methodsFor: #accesing !
mask
	| v1 v2 v3 v4 |
{'	v1, v2, v3, v4 = self:handle():getMask( ) '}.
	^Array with: v1 with: v2 with: v3 with: v4! !

! LoveForm methodsFor: #accesing !
rayCast: x1 y1: y1 x2: x2 y2: y2
	maxFraction: maxFraction
	tx: tx ty: ty tr: tr
	childIndex: childIndex
	| x y fraction |
{'	x, y, fraction = self:handle():rayCast( x1, y1, x2, y2, maxFraction, tx, ty, tr, childIndex ) '}.
	^Array with: x@y with: fraction! !

LoveForm buildFunctions: #(
	destroy isDestroyed
	#(getBody Body)
	getDensity
	getFriction
	getRestitution
	isSensor
	#(setCategory: #(c1))
	#(setCategory:with: #(c1 c2))
	#(setCategory:with:with: #(c1 c2 c3))
	#(setCategory:with:with:with: #(c1 c2 c3 c4))
	#(setDensity #(x))
	#(setFilterData #(categories mask group))
	#(setFriction #(friction))
	#(setMask: #(v1))
	#(setMask:with: #(v1 v2))
	#(setMask:with:with: #(v1 v2 v3))
	#(setMask:with:with:with: #(v1 v2 v3 v4))
	#(setRestitution #(restitution))
	#(setSensor #(sensor))
	#(testPoint #(tx ty tr x y))
)!

! Shape methodsFor: #accesing !
computeAABB: tx ty: ty tr: tr childIndex: childIndex
	| left top right bottom |
{'	left, top, right, bottom = self:handle():computeAABB( tx, ty, tr, childIndex) '}.
	^left@top rightBottom: right@bottom! !

! Shape methodsFor: #accesing !
testSegment: x1 y1: y1 x2: x2 y2: y2
	| t x y |
{'	t, x, y = self:handle():testSegment( x1, y1, x2, y2 ) '}.
	^Array with: t with: x@y! !

Shape buildFunctions: #(
	getCategoryBits
	getChildCount
	getData
	getMask
	getRadius
	getType
	#(setData #(x))
)!

! Fixture methodsFor: #accesing !
massData
	| x y mass inertia |
{'	x, y, mass, inertia = self:handle():getMassData( ) '}.
	^Array with: x@y with: mass with: inertia! !

Fixture buildFunctions: #(
	getGroupIndex
	#(getShape Shape)
	getUserData
	#(setUserData #(x))
)!

LoveObject subclass: #Joint !

! Joint methodsFor: #accesing !
anchors
	| x1 y1 x2 y2 |
{'	x1, y1, x2, y2 = self:handle():getAnchors( ) '}.
	^Array with: x1@y1 with: x2@y2! !

! Joint methodsFor: #accesing !
bodies
	| a b |
{'	a, b = self:handle():getBodies( ) '}.
	^Array with: Body @ a with: Body @ b! !

! Joint methodsFor: #accesing !
reactionForce: invdt
	| x y |
{'	x, y = self:handle():getReactionForce( invdt ) '}.
	^x@y! !

Joint buildFunctions: #(
	destroy isDestroyed
	getCollideConnected
	#(getReactionTorque #(invdt))
	getType
	getUserData
	#(setUserData #(x))
	#(setCollideConnected #(collide))
)!

LoveObject subclass: #World!

! World methodsFor: #accessing !
callbacks
	| f1 f2 f3 f4 |
{'	f1, f2, f3, f4 = self:handle():getCallbacks() '}.
	^Array with: f1 with: f2 with: f3 with: f4! !

! World methodsFor: #accessing !
gravity
	| x y |
{'	x, y = self:handle():getGravity() '}.
	^x@y! !

! World methodsFor: #collision !
queryBoundingBox: aRectangle doing: aBlock
	^self queryBoundingBox: aRectangle top
		left: aRectangle left
		right: aRectangle right
		bottom: aRectangle bottom
		callback: aBlock! !

! World methodsFor: #collision !
rayCast: fromPoint to: toPoint doing: aBlock
	^self	rayCast: fromPoint x y1: fromPoint y
		x2: toPoint x y2: toPoint y
		callback: aBlock! !

World buildFunctions: #(
	destroy isDestroyed
	getAllowSleeping
	#(getBodyList 'Array[Body]')
	getBodyCount getContactCount
	getContactFilter
	#(getContactList 'Array[Contact]')
	getJointCount
	#(getJointList 'Array[Joint]')
	isAllowSleep
	isLocked
	isSleepingAllowed
	#(queryBoundingBox #(top left right bottom callback))
	#(rayCast #(x1 y1 x2 y2 callback ))
	#(setAllowSleep #(permission))
	#(setAllowSleeping #(allowSleep))
	#(setCallbacks #(beginContact endContact preSolve postSolve))
	#(setContactFilter #(aBlock))
	#(setGravity #(x y))
	#(setMeter #(scale))
	#(setSleepingAllowed #(allow))
	#(translateOrigin #(x y))
	#(update #(dt))
	#(update #(dt velocityiterations))
	#(update #(dt velocityiterations positioniterations))
)!

#(	Chain Circle Edge Polygon
) do: [:each| Shape subclass: each ,#Shape ].
#(	Distance Friction Gear Motor Mouse Prismatic
	Pulley Revolute Rope Weld Wheel
) do: [:each| Joint subclass: each ,#Joint ].

DistanceJoint buildFunctions: #(
	getDampingRatio #(setDampingRatio #(x))
	getFrequency #(setFrequency #(x))
	getLength #(setLength #(x))
)!
FrictionJoint buildFunctions: #(
	getMaxForce #(setMaxForce #(x))
	getMaxTorque #(setMaxTorque #(x))
)!

! GearJoint  methodsFor: #accessing !
joints
	| joint1 joint2 |
{'	joint1, joint2 = self:handle():getJoints( ) '}.
	^Array with: joint1 with: joint2! !

GearJoint buildFunctions: #(
	getRatio #(setRatio #(x))
)!

MotorJoint buildFunctions: #(
	getAngularOffset #(setAngularOffset #(x))
	getLinearOffset #(setLinearOffset #(x))
)!

! MouseJoint methodsFor: #accessing !
target
	| x y |
{'	x, y = self:handle():getTarget( ) '}.
	^x@y! !

! MouseJoint methodsFor: #accessing !
target: aPoint
	^self target: aPoint x y: aPoint y! !

MouseJoint buildFunctions: #(
	getDampingRatio #(setDampingRatio #(x))
	getFrequency #(setFrequency #(x))
	getMaxForce #(setMaxForce #(x))
	getTarget #(setTarget #(x y))
)!

! PrismaticJoint methodsFor: #accessing !
axis
	| x y |
{'	x, y = self:handle():getAxis( ) '}.
	^x@y! !

! PrismaticJoint methodsFor: #accessing !
limits
	| lower upper |
{'	lower, upper = self:handle():getLimits( ) '}.
	^Array with: lower with: upper! !

PrismaticJoint buildFunctions: #(
	areLimitsEnabled #(enableLimit #( enable)) hasLimitsEnabled
	#(enableMotor #(enable))
	getJointSpeed getJointTranslation
	getLowerLimit getUpperLimit
	getMaxMotorForce #(setMaxMotorForce #(x))
	#(getMotorForce #(invdt))
	getMotorSpeed #(setMotorSpeed #(s))
	isMotorEnabled
	#(setLimits #(lower upper))
	#(setLowerLimit #(x))
	#(setUpperLimit #(x))
)!

! Physics class methodsFor: #accessing !
distance: fixture1 to: fixture2
	| distance x1 y1 x2 y2 |
{'	distance, x1, y1, x2, y2 = love.physics.getDistance( fixture1:json() , fixture2:json() ) '}.
	^Array with: distance with: x1@y1 with: x2@y2! !

Physics buildLuaStaticFunctions: #(
	getMeter
	#(newBody #(_world x y) Body)
	#(newBody #(_world x y type) Body)
	#(newChainShape #(loop points ) ChainShape)
	#(newCircleShape #(radius) CircleShape)
	#(newCircleShape #(x y radius) CircleShape)
	#(newDistanceJoint #(_body1 _body2 x1 y1 x2 y2) Joint)
	#(newDistanceJoint #(_body1 _body2 x1 y1 x2 y2 collideConnected) Joint)
	#(newEdgeShape #(x1 y1 x2 y2) EdgeShape)
	#(newFixture #(_body _shape) Fixture)
	#(newFixture #(_body _shape density) Fixture)
	#(newFrictionJoint #(_body1 _body2 x y) FrictionJoint)
	#(newFrictionJoint #(_body1 _body2 x y collideConnected) FrictionJoint)
	#(newFrictionJoint #(_body1 _body2 x1 y1) FrictionJoint)
	#(newFrictionJoint #(_body1 _body2 x1 y1 x2 y2 collideConnected) FrictionJoint)
	#(newGearJoint #(_joint1 _joint2 ratio) GearJoint)
	#(newGearJoint #(_joint1 _joint2 ratio collideConnected) GearJoint)
	#(newMotorJoint #(_body1 _body2) MotorJoint)
	#(newMotorJoint #(_body1 _body2 correctionFactor) MotorJoint)
	#(newMotorJoint #(_body1 _body2 correctionFactor collideConnected ) MotorJoint)
	#(newMouseJoint #(_body x y ) MouseJoint)
	#(newPolygonShape #(vertices) PolygonShape)
	#(newPrismaticJoint #(_body1 _body2 x1 y1 x2 y2 ax ay) PrismaticJoint)
	#(newPrismaticJoint #(_body1 _body2 x1 y1 x2 y2 ax ay collideConnected) PrismaticJoint)
	#(newPrismaticJoint #(_body1 _body2 x1 y1 x2 y2 ax ay collideConnected referenceAngle) PrismaticJoint)
	#(newPulleyJoint #(_body1 _body2 gx1 gy1 gx2 gy2 x1 y1 x2 y2) PulleyJoint)
	#(newPulleyJoint #(_body1 _body2 gx1 gy1 gx2 gy2 x1 y1 x2 y2 ratio) PulleyJoint)
	#(newPulleyJoint #(_body1 _body2 gx1 gy1 gx2 gy2 x1 y1 x2 y2 ratio collideConnected) PulleyJoint)
	#(newRectangleShape #(x y) PolygonShape)
	#(newRectangleShape #(x y width height) PolygonShape)
	#(newRectangleShape #(x y width height angle) PolygonShape)
	#(newRevoluteJoint #(_body1 _body2 x1 y1 x2 y2) RevoluteJoint)
	#(newRevoluteJoint #(_body1 _body2 x1 y1 x2 y2 collideConnected) RevoluteJoint)
	#(newRevoluteJoint #(_body1 _body2 x1 y1 x2 y2 collideConnected referenceAngle) RevoluteJoint)
	#(newRopeJoint #(_body1 _body2 x1 y1 x2 y2 maxLength) RopeJoint)
	#(newRopeJoint #(_body1 _body2 x1 y1 x2 y2 maxLength collideConnected) RopeJoint)
	#(newWeldJoint #(_body1 _body2 x1 y1) WeldJoint)
	#(newWeldJoint #(_body1 _body2 x1 y1 collideConnected) WeldJoint)
	#(newWeldJoint #(_body1 _body2 x1 y1 x2 y2) WeldJoint)
	#(newWeldJoint #(_body1 _body2 x1 y1 x2 y2 collideConnected ) WeldJoint)
	#(newWheelJoint #(_body1 _body2 x y ax ay) WheelJoint)
	#(newWheelJoint #(_body1 _body2 x y ax ay collideConnected) WheelJoint)
	#(newWorld World)
	#(newWorld #(sleep) World)
	#(newWorld #(xg yg) World)
	#(newWorld #(xg yg sleep) World)
	#(setMeter #(scale))
)!

LoveObject subclass: #Decoder!
Decoder buildFunctions: #(
	#(decode SoundData)
	getBitDepth getBits
	getChannelCount getChannels
	getDuration getSampleRate
)!

Sound buildLuaStaticFunctions: #(
	#(newDecoder #(_file buffer) Decoder)
	#(newSoundData #(_arg) SoundData)
)!

! System class methodsFor: #accessing !
powerInfo
	| state percent seconds |
{'	state, percent, seconds = love.system.getPowerInfo( ) '}.
	^Array with: state with: percent with: seconds! !

System buildLuaStaticFunctions: #(
	getClipboardText
	getOS
	getProcessorCount
	hasBackgroundMusic
	#(openURL #(url))
	#(setClipboardText #(text))
	vibrate #(vibrate #(seconds))
)!

! PulleyJoint methodsFor: #accessing !
groundAnchors
	| a1x a1y a2x a2y |
{'	a1x, a1y, a2x, a2y = self:handle():getGroundAnchors( ) '}.
	^Array with: a1x@a1y with: a2x@a2y! !

! PulleyJoint methodsFor: #accessing !
maxLengths
	| len1 len2 |
{'	len1, len2 = self:handle():getMaxLengths( ) '}.
	^Array with: len1 with: len2! !

PulleyJoint buildFunctions: #(
	getConstant #(setConstant #(x))
	getLengthA getLengthB
	getRatio #(setRatio #(x))
	#(setMaxLengths #(max1 max2))
)!

! RevoluteJoint methodsFor: #accessing !
limits
	| lower upper |
{'	lower, upper = self:handle():getLimits( ) '}.
	^Array with: lower with: upper! !

RevoluteJoint buildFunctions: #(
	areLimitsEnabled #(enableLimit #(enable)) #(setLimitsEnabled #(enable))
	getJointAngle getJointSpeed
	getLowerLimit getUpperLimit
	#(setLowerLimit #(x)) #(setUpperLimit #(x))
	getMaxMotorTorque #(setMaxMotorTorque #(x))
	getMotorSpeed #(setMotorSpeed #(x))
	hasLimitsEnabled isMotorEnabled
	#(setLimits #(lower upper))
)!

RopeJoint buildFunctions: #(
	getMaxLength
	#(setMaxLength #(x))
)!

WeldJoint buildFunctions: #(
	getDampingRatio #(setDampingRatio #(x))
	getFrequency #(setFrequency #(x))
)!

! WheelJoint methodsFor: #accessing !
axis
	| x y |
{'	x, y = self:handle():getAxis( ) '}.
	^x@y! !

WheelJoint buildFunctions: #(
	#(enableMotor #(enable)) isMotorEnabled
	getJointSpeed getJointTranslation
	getMaxMotorTorque #(setMaxMotorTorque #(x))
	getMaxMotorTorque #(setMaxMotorTorque #(x))
	#(getMotorTorque #(invdt))
	getSpringDampingRatio #(setSpringDampingRatio #(x))
	getSpringFrequency #(setSpringFrequency #(x))
)!


LoveObject subclass: #Channel!
Channel buildFunctions: #(
	clear
	demand #(demand #(timeout))
	getCount peek
	#(hasRead #(id))
	pop #(push #(value))
	#(supply #(value))
)!

Thread buildFunctions: #(
	getError
	isRunning
	start
	#(start #(arg1))
	#(start #(arg1 arg2))
	#(start #(arg1 arg2 arg3))
	#(start #(arg1 arg2 arg3 arg4))
	wait
)!
Thread buildLuaStaticFunctions: #(
	#(getChannel #(aName) Channel)
	#(getThread #(aName) Thread)
	#(getThreads 'Array[Thread]')
	#(newChannel Channel)
	#(newThread #(_arg) Thread)
)!

Timer buildLuaStaticFunctions: #(
	getAverageDelta
	getDelta
	getFPS
	getMicroTime getTime
	#(sleep #(s))
	step
)!

! Touch class methodsFor: #accessing !
position: id
	| x y |
{'	x, y = love.touch.getPosition( id )'}.
	^x@y! !

Touch buildLuaStaticFunctions: #(
	#(getPressure #(id))
	#(getTouches Array)
)!

Video buildLuaStaticFunctions: #(
	#(newVideoStream #(_arg) VideoStream)
)!

! Window class methodsFor: #accessing !
fromPixels: px py: py
	| x y |
{'	x, y = love.window.fromPixels( px, py )'}.
	^x@y! !

! Window class methodsFor: #accessing !
dimensions
	| x y |
{'	x, y = love.window.getDimensions()'}.
	^x@y! !

! Window class methodsFor: #accessing !
desktopDimensions: display
	| x y |
{'	x, y = love.window.getDesktopDimensions( display )'}.
	^x@y! !

! Window class methodsFor: #accessing !
desktopDimensions
	^self desktopDimensions: 1! !

! Window class methodsFor: #accessing !
fullScreen
	| fullscreen fstype |
{'	fullscreen, fstype = love.window.getFullscreen( )'}.
	^Array with: fullscreen with: fstype! !

! Window class methodsFor: #accessing !
mode
	| width height flags |
{'	width, height, flags = love.window.getMode( )'}.
	^Array with: width with: height with: Lua@flags! !

! Window class methodsFor: #accessing !
positionAndDisplay
	| x y display |
{'	x, y, display = love.window.getPosition()'}.
	^Array with: x@y with: display! !

! Window class methodsFor: #accessing !
position
	| x y display |
{'	x, y, display = love.window.getPosition()'}.
	^x@y! !

! Window class methodsFor: #accessing !
pixels: x y: y
	| px py |
{'	px, py = love.window.toPixels(x,y)'}.
	^px@py! !

Window buildLuaStaticFunctions: #(
	close
	#(fromPixels #(value))
	getDPIScale
	getDisplayCount
	getDisplayName #(getDisplayName #(index))
	#(getFullscreenModes #(display) 'Array[Lua]')
	getHeight getWidth
	#(getIcon ImageData)
	getPixelScale
	getTitle
	hasFocus hasMouseFocus
	isDisplaySleepEnabled
	isMaximized isMinimized
	isOpen isVisible
	maximize minimize restore
	requestAttention #(requestAttention #(continuous))
	#(setDisplaySleepEnabled #(enable))
	#(setFullscreen #(fullscreen))
	#(setFullscreen #(fullscreen fstype))
	#(setIcon #(_imagedata))
	#(setMode #(width height))
	#(setMode #(width height flags))
	#(setPosition #(x y))
	#(setPosition #(x y display))
	#(setTitle #(title))
	#(showMessageBox #(title message))
	#(showMessageBox #(title message type))
	#(showMessageBox #(title message type attachtowindow))
	#(updateMode #(width height _settings))
)!

! Window methodsFor: #accessing !
extent
	^self width @ self height! !

! Window methodsFor: #accessing !
position: aPoint
	^self position: aPoint x y: aPoint y! !

ChainShape buildFunctions: #(
	#(getChildEdge #(index) EdgeShape)
	getVertexCount
	#(setNextVertex #(x y))
	#(setPreviousVertex #(x y))
)!

! ChainShape methodsFor: #accessing !
nextVertex
	" Returns the next vertex of the receiver. "

	| x y |
{'	x, y = self:handle():getNextVertex() '}.
	^x notNil ifTrue: [ x@y ]! !

! ChainShape methodsFor: #accessing !
pointAt: index
	" Returns the point at index of the receiver. "

	| x y |
{'	x, y = self:handle():getPoint(index) '}.
	^x@y! !

! ChainShape methodsFor: #accessing !
previousVertex
	" Returns the next vertex of the receiver. "

	| x y |
{'	x, y = self:handle():getPreviousVertex() '}.
	^x notNil ifTrue: [ x@y ]! !


CircleShape buildFunctions: #(
	getRadius
	#(setPoint #(x y))
	#(setRadius #(radius))
)!

! CircleShape methodsFor: #accessing !
localCenter
	" Returns the next vertex of the receiver. "

	| x y |
{'	x, y = self:handle():getLocalCenter() '}.
	^x@y! !

! CircleShape methodsFor: #accessing !
point
	" Returns the point of the receiver. "

	| x y |
{'	x, y = self:handle():getPoint() '}.
	^x@y! !

! CircleShape methodsFor: #accessing !
worldCenter
	" Returns the center of the receiver in world coordinates. "

	| x y |
{'	x, y = self:handle():getWorldCenter() '}.
	^x@y! !


EdgeShape buildFunctions: #(
	#(setNextVertex #(x y))
	#(setPreviousVertex #(x y))
)!

! EdgeShape methodsFor: #accessing !
nextVertex
	" Returns the next vertex of the receiver. "

	| x y |
{'	x, y = self:handle():getNextVertex() '}.
	^x notNil ifTrue: [ x@y ]! !

! EdgeShape methodsFor: #accessing !
points
	" Returns the points of the receiver. "

	| x y u v |
{'	x, y, u, v = self:handle():getPoints() '}.
	^Array with: x@y with: u@v! !

! EdgeShape methodsFor: #accessing !
previousVertex
	" Returns the next vertex of the receiver. "

	| x y |
{'	x, y = self:handle():getPreviousVertex() '}.
	^x notNil ifTrue: [ x@y ]! !

PolygonShape buildFunctions: #(
	validate
)!

! PolygonShape methodsFor: #accessing !
points
	" Returns the points of the receiver. "

	| x1 y1 x2 y2 x3 y3 x4 y4 x5 y5 x6 y6 x7 y7 x8 y8 result |
	result := Array new.
{'	x1, y1, x2, y2, x3, y3, x4, y4, x5, y5, x6, y6, x7, y7, x8, y8 = self:handle():getPoints() '}.
	x1 notNil ifTrue: [ result add: x1@y1 ] ifFalse: [ ^result ].
	x2 notNil ifTrue: [ result add: x2@y2 ] ifFalse: [ ^result ].
	x3 notNil ifTrue: [ result add: x3@y3 ] ifFalse: [ ^result ].
	x4 notNil ifTrue: [ result add: x4@y4 ] ifFalse: [ ^result ].
	x5 notNil ifTrue: [ result add: x5@y5 ] ifFalse: [ ^result ].
	x6 notNil ifTrue: [ result add: x6@y6 ] ifFalse: [ ^result ].
	x7 notNil ifTrue: [ result add: x7@y7 ] ifFalse: [ ^result ].
	x8 notNil ifTrue: [ result add: x8@y8 ] ifFalse: [ ^result ].
	^result! !