EventManager
	subclass: #MobileDevice
	instanceVariableNames: ' name statusBarHeight units imageCache screenExtent typeWidgetMapping'
	classVariableNames: #Current
	category: #MobileDevice !
MobileDevice comment: '
	Abstract model of a Mobile device.
	MobileDevice define minimal functionality that a device must provide to MobileApplication.
	Subclasses must refine methods providing implementation.
	@2013 Alejandro Reimondo - aleReimondo@smalltalking.net'!
MobileDevice buildAccessors: #(
	#( #name #initialize 'self class name')
	#( #typeWidgetMapping #initialize #PoolDictionary)
	#( #units #initialize '#pt' )
	#( #statusBarHeight #initialize #20 )
	#( #imageCache #initialize )
	#( #screenExtent #initialize '320@480' )
).!

! MobileDevice class methodsFor: #accessing !
current
	" Return the current instance of the receiver. "

	Current isNil ifTrue: [ self initializeCurrent ].
	^Current! !

! MobileDevice class methodsFor: #accessing !
currentOrNil
	" Return the current instance of the receiver (or nil). "

	^Current! !

! MobileDevice class methodsFor: #accessing !
current: anObject
	" Set the current instance of the receiver. "

	Current := anObject! !

! MobileDevice class methodsFor: #initialize !
initializeCurrent
	" Private - Initialize the current of the receiver. "

	Current := self installed new! !

! MobileDevice class methodsFor: #defaults !
isDefault
	" Private - Return true if the receiver can act as default implementation.
	This method must be implemented by at least one subclass of MobileDevice.
	"

	^false! !

! MobileDevice class methodsFor: #defaults !
installed
	" Private - Return the installed support for the receiver. "

	| default |
	self allSubclasses reverseDo: [:each|
		each isInstalled ifTrue: [ ^each ].
		(default isNil and: [ each isDefault ]) ifTrue: [ default := each ].
	].
	default isNil ifTrue: [
		self error: 'Missing default implementation for ',self name
	].
	^default! !

! MobileDevice class methodsFor: #defaults !
isInstalled
	" Private - Return true if the receiver implements installed support.
	Concrete subclasses must implement detection of real device and return true if detected.
	"

	^false! !

! MobileDevice class methodsFor: #defaults !
mainViewSupport
	" Private - Return the support for mainView(or nil). "

	^nil! !

! MobileDevice class methodsFor: #accessing !
imageWidgetSupport
	" Return the support for ImageWidgets. "

	^self subclassResponsibility: #imageWidgetSupport! !

! MobileDevice class methodsFor: #accessing !
userAgent
	" Private - Return the user agent. "

	^self subclassResponsibility: #userAgent! !

! MobileDevice methodsFor: #accessing !
imageWidgetSupport
	" Return the support for ImageWidgets. "

	^self class imageWidgetSupport! !

! MobileDevice methodsFor: #query !
isInstalled
	" Private - Return true if the receiver is installed. "

	^(self isMemberOf: MobileDevice) not and: [
		self class isInstalled
	]! !

! MobileDevice methodsFor: #ui !
fullScreen
	" Private - Return true if the receiver display app in fullScreen mode. "

	^self isInstalled! !

! MobileDevice methodsFor: #accessing !
id
	" Private - Return the device ID of the receiver. "

	^self subclassResponsibility: #id! !

! MobileDevice methodsFor: #accessing !
platform
	" Private - Return the platform of the receiver (or #unknown ). "

	^self subclassResponsibility: #platform! !

! MobileDevice methodsFor: #accessing !
details
	" Private - Return the details array of the receiver. "

	^self subclassResponsibility: #details! !

! MobileDevice methodsFor: #accessing !
model
	" Private - Return the model of the receiver. "

	^self details first! !

! MobileDevice methodsFor: #accessing !
basicConnectionType
	" Private - Return the connection type. "

	^self subclassResponsibility: #basicConnectionType! !

! MobileDevice methodsFor: #connection !
connectionType
	" Return the type of active conection.
	It is a fast way to determine the device's network connection state, and type of connection.
	"

	| result |
	self isInstalled ifFalse: [ "DEBUG, Development - connection granted while testing. "
		^#wifi
	].
	[	result := self basicConnectionType.
		result = 0 ifTrue: [ ^#wifi ]."library support missing"
	] on: Error do: [ ^#none ].
	^result! !

! MobileDevice methodsFor: #query !
isOnline
	" Return true if the receiver is online. "

	| connectionType |
	connectionType := self connectionType.
	^(#( unknown none ) includes: connectionType asLowercase) not! !

! MobileDevice methodsFor: #geo !
instructionsToTarget: aLocation
	latitude: latitude longitude: longitude
	on: aTaskView
	" Assist the user with directions to target. "

	^self subclassResponsibility: #instructionsToTarget:latitude:longitude:on:! !

! MobileDevice methodsFor: #accessing !
mainViewFor: anApplication
	" Return a mainView (GUI container) for the receiver(or nil). "

	| support result |
	support := self mainViewSupport.
	support isNil ifTrue: [ ^nil ].
	result := support new owner: anApplication; yourself.
	self customizeMainView: result application: anApplication.
	^result! !

! MobileDevice methodsFor: #defaults !
mainViewSupport
	" Private - Return the support for mainView of the receiver(or nil). "

	^self class mainViewSupport! !

! MobileDevice methodsFor: #defaults !
customizeMainView: aMainView application: anApplication
	" Private - Customize the mainView. "

	^self subclassResponsibility: #customizeMainView:application: ! !

! MobileDevice methodsFor: #gui !
selectorToOpen: anApplication
	" Return the selector to use to open anApplication. "

	^#basicOpen! !

! MobileDevice methodsFor: #gui !
openTabsFor: anApplication
	" Open tabs to host application UI. "

	^self subclassResponsibility: #openTabsFor: ! !

! MobileDevice methodsFor: #gui !
renderApplicationContents: aMobileApplication
	" Render the contents of anApplication. "

	| active |
	active := aMobileApplication active.
	active notNil ifTrue: [ active buildContentsIn: aMobileApplication ]! !

! MobileDevice methodsFor: #app !
newThreadFor: aMobileApplication

	^MobileThreadAction new
		application: aMobileApplication;
		yourself! !

! MobileDevice methodsFor: #gui !
newWidget: aName for: aMobileCoordinator

	| support |
	support := Smalltalk at: aName ifAbsent: [
		self error: '// Missing widget: ',aName printString
	].
	^support new
		coordinator: aMobileCoordinator;
		yourself! !

! MobileDevice methodsFor: #camera !
openCameraDoing: action onError: errorBlock with: options
	" Evaluate action with snapshot captured from camera. "

	self subclassResponsibility: #openCameraDoing:onError:with:.
	^errorBlock value: 'Camera - Not implemented. Yet'! !

! MobileDevice methodsFor: #multimedia !
openFileAt: pathName
	" Open the (local) file at pathName. "

	^self subclassResponsibility: #openFileAt:! !

! MobileDevice methodsFor: #multimedia !
openFileAtUrl: url
	" Download and open the file at url. "

	^self subclassResponsibility: #openFileAtUrl:! !

! MobileDevice methodsFor: #multimedia !
openFileContents: fileContents type: filetype version: version
	" Save the contents in local storage and open as a document of specified type. "

	^self subclassResponsibility: #openFileContents:type:version: ! !

! MobileDevice methodsFor: #accessing !
peek: aMobileApplication title: aTitle
	selecting: initialSelection in: options
	doing: anAction onClose: closeBlock
	" Open an instance of the receiver prompting aValue from list of options. "

	^self subclassResponsibility: #peek:title: ! !

! MobileDevice methodsFor: #gui !
scaleMainViewFor: anApplication
	" Private - Scale the main view of anApplication. "

	^self subclassResponsibility: #scaleMainViewFor:! !

! MobileDevice methodsFor: #reading !
jsObjectInFile: aPathName do: aBlock onError: errorBlock
	" Evaluate aBlock when reading a JSON object from file at aPathName.
	In case of error reading the object, evaluates errorBlock.
	"

" DEBUG - TODO - needs implementation. See
	IMLT8App>>#withSettingsDo:then:
	IMLT8App>>#settingsPathName (pathName can have folder references between [])
"
	errorBlock evaluateWith: 'jsObjectInFile:... notImplemented. Yet'! !

! MobileDevice methodsFor: #barcode !
userScanCodeFor: aMobileApplication doing: aTwoArgumentsBlock onClose: onClose onError: errorBlock
	" Let the user enter scan barcode from keyboard. "

"OBSOLETE -
| value |
	value := self prompt: 'Enter scan code'.
	value ~= nil ifTrue: [
		aTwoArgumentsBlock
			value: (value notNil ifTrue: [ #user ])
			value: value
	].
	onClose evaluate.
"	^self subclassResponsibility: #userScanCodeFor:doing:onClose:onError:! !

! MobileDevice methodsFor: #barcode !
scanCodeFor: aMobileApplication doing: aTwoArgumentsBlock onClose: onClose onError: errorBlock
	" Let the user scan barcode and evaluates aTwoArgumentBlock on success.
	Evaluates errorBlock in case of error during scanning.
	Optimistic implementation - We will try to use native interface for scanning (if available)."

	^self subclassResponsibility: #scanCodeFor:doing:onClose:onError:! !

! MobileDevice methodsFor: #barcode !
scanCode: native for: aMobileApplication doing: aTwoArgumentsBlock onClose: onClose onError: errorBlock
	" Let the user scan barcode and evaluates aTwoArgumentBlock on success.
	Evaluates errorBlock in case of error during scanning.
	Optimistic implementation - We will try to use native interface for scanning (if available).
	"

	| checkResult noResult done |
	native isNil ifTrue: [
		"	self error: 'Native barcode API missing.'	"
		^self
			userScanCodeFor: aMobileApplication
			doing: aTwoArgumentsBlock
			onClose: onClose onError: errorBlock
	].
	done := false.
	[ native scan ] on: Error do: [:ex| ^errorBlock value: ex ].
	noResult := [
		done := true.
		aTwoArgumentsBlock value: nil value: nil
	].
	checkResult := [ | result format data |
		done ifFalse: [
			done := true.
			result := native scanResult.
			result isNil ifTrue: [ noResult value ]
			ifFalse: [
				format := result getFormatName.
				data := result getContents.
				aTwoArgumentsBlock value: format value: data
			]
		]
	].
	self
		confirm: aMobileApplication
		text: 'Is scan code valid ?'
		action: (Array
			with: 'Accept' -> checkResult
			with: 'Enter code' -> [
				self	userScanCodeFor: aMobileApplication
					doing: aTwoArgumentsBlock
					onClose: [ #ignore ]
					onError: errorBlock
			])
		cancel: noResult
		close: onClose! !

! MobileDevice methodsFor: #barcode !
scanCodeFor: aMobileApplication doing: aTwoArgumentsBlock onClose: onClose onError: errorBlock
	" Let the user scan barcode and evaluates aTwoArgumentBlock on success.
	Evaluates errorBlock in case of error during scanning.
	Optimistic implementation - We will try to use native interface for scanning (if available).
	"

	| native checkResult noResult done |
	native := aMobileApplication native.
	native isNil ifTrue: [
		"	self error: 'Native barcode API missing.'	"
		^self
			userScanCodeFor: aMobileApplication
			doing: aTwoArgumentsBlock
			onClose: onClose onError: errorBlock
	].
	done := false.
	[ native scan ] on: Error do: [:ex| ^errorBlock value: ex ].
	noResult := [
		done := true.
		aTwoArgumentsBlock value: nil value: nil
	].
	checkResult := [ | result format data |
		done ifFalse: [
			done := true.
			result := native scanResult.
			result isNil ifTrue: [ noResult value ]
			ifFalse: [
				format := result getFormatName.
				data := result getContents.
				aTwoArgumentsBlock value: format value: data
			]
		]
	].
	self
		confirm: aMobileApplication
		text: 'Is scan code valid ?'
		action: (Array
			with: 'Accept' -> checkResult
			with: 'Enter code' -> [
				self	userScanCodeFor: aMobileApplication
					doing: aTwoArgumentsBlock
					onClose: [ #ignore ]
					onError: errorBlock
			])
		cancel: noResult
		close: onClose! !

! MobileDevice methodsFor: #gui !
screenFor: anApplication

	^self subclassResponsibility: #screenFor: ! !

! MobileDevice methodsFor: #defaults !
sourceForImage: imageName
	"  Return the URL for source of resource image at imageName. "

	^self subclassResponsibility: #sourceForImage: ! !

! MobileDevice methodsFor: #widgets !
text: text with: tags
	" Private - Return a widget for text styled with tags. "

	^self subclassResponsibility: #text:with: ! !

! MobileDevice methodsFor: #widgets !
text: text
	" Private - Return a widget for text. "

	^self text: text with: Array new! !

! MobileDevice methodsFor: #gui !
button: title for: imageProvider doing: aBlock
	" Private - Return a button with title evaluating anAction when clicked. "

	^self subclassResponsibility: #button:for:doing:! !

! MobileDevice methodsFor: #list !
configure: cell title: aTitle description: aDescription detail: aDetail
	" Private - Default configuration for cells in groups. "

	^self subclassResponsibility: #configure:title:description:detail: ! !

! MobileDevice methodsFor: #gui !
confirm: aMobileApplication text: question action: actionAssoc
	" Prompt the user for confirmation and evaluates actionAssoc when confirmed. "

	^self confirm: aMobileApplication text: question action: actionAssoc cancel: [ "ignored" ] close: [ "ignored" ]! !

! MobileDevice methodsFor: #gui !
confirm: aMobileApplication text: question action: actionAssoc cancel: cancelAction
	" Prompt the user for confirmation and evaluates actionAssoc when confirmed. "

	^self confirm: aMobileApplication text: question action: actionAssoc cancel: cancelAction close: [ "ignored" ]! !

! MobileDevice methodsFor: #gui !
confirm: aMobileApplication text: question action: actionAssoc cancel: cancelAction close: closeAction
	" Prompt the user for confirmation and evaluates anAction when confirmed. "

	^self subclassResponsibility: #confirm:text:action:cancel:close:! !


! MobileDevice methodsFor: #gui !
addButton: actionAssoc in: pane onClose: forceClose for: imageProvider
	" Private - Arrange an action button on pane. "

	^self subclassResponsibility: #addButton:in:onClose:for:! !

! MobileDevice methodsFor: #gui !
centerButton: aButton background: background color: color
	" Private - Center the button. "

	^self subclassResponsibility: #centerButton:background:color:! !

! MobileDevice methodsFor: #gui !
dialog: aMobileApplication
	text: explanation
	custom: customBlock
	doing: anAction
	" Open a modal dialog customized by customBlock.
	Evaluate anAction when the user close the dialog.
	Return the action to be evaluated to force to close the dialog.
	"

	^self
		dialog: aMobileApplication
		text: explanation
		custom: customBlock
		doing: anAction
		closeOnClick: true! !

! MobileDevice methodsFor: #gui !
dialog: aMobileApplication
	text: explanation
	custom: customBlock
	doing: anAction
	closeOnClick: closeOnClick
	" Open a modal dialog customized by customBlock. "

	^self subclassResponsibility: #dialog:text:custom:doing:closeOnClick:! !

! MobileDevice methodsFor: #gui !
tool: title for: imageProvider doing: aBlock
	" Private - Return an action tool. "

	^self button: title for: imageProvider doing: aBlock! !	

! MobileDevice methodsFor: #gui !
prompt: text
	" Private - Return the result of prompting the user about text. "

	self -- ('// Needs ',self class name,' implementor for #prompt: ',text printString).
	^super prompt: text! !

! MobileDevice methodsFor: #gui !
warning: aMobileApplication text: explanation doing: anAction
	" Open a modal dialog with explanation shown to the user.
	Evaluate anAction when the user close the dialog.
	"

	^self subclassResponsibility: #warning:text:doing:! !

! MobileDevice methodsFor: #gui !
toolbarWith: left and: right for: imageProvider
	" Return a simple toolbar with two elements. "

	^self subclassResponsibility: #toolbarWith:and:for: ! !

! MobileDevice methodsFor: #private !
initializeImageCache
	" Private - Initialize the image cache. "

	imageCache := PoolDictionary new.
	"populate image cache with well known application resources"
	self iconNames do: [:imageName|
		imageCache
			at: imageName
			put: (self sourceForImage: imageName)
	]! !

! MobileDevice methodsFor: #defaults !
iconNames
	" Private - Return the names of icons in resources storage.
	This method must be refined for application icons.
	"

	^#()! !

! MobileDevice methodsFor: #gui !
imageNamed: aSymbol
	" Return the default image named aSymbol (or nil). "

	^self imageCache
		at: aSymbol
		ifAbsent: []! !

! MobileDevice methodsFor: #gui !
navigationBarFor: anApplication
	" Private - Return the/a navigationBar installed on the receiver. "

	^self subclassResponsibility: #navigationBarFor:! !

! MobileDevice methodsFor: #gui !
thread: aMobileThread actionsBarDo: aBlock
	" Private - Evaluate aBlock with actionsBar (if present). "

	| bar |
	bar := aMobileThread application screen actionsBar.
	^bar notNil ifTrue: [ aBlock evaluateWith: bar ]! !
	
! MobileDevice methodsFor: #gui !
newWidgetFromJSon: aJSon for: aMobileCoordinator
	^self	newWidget: (self widgetClassNameForJS: JS @ aJSon)
		for: aMobileCoordinator. ! !

! MobileDevice methodsFor: #gui !
widgetClassNameForJS: aJS
	" Return a class name for widget specified in aJS (or nil). "

	| typeName className|
	className := aJS @ #class.
	className notNil ifTrue: [ ^className ].
	typeName := aJS @ #type.
	typeName notNil ifTrue: [
		^self widgetClassNameForType: typeName
	].
	^nil! !
	
! MobileDevice methodsFor: #gui !
widgetClassNameForType: typeName
	" Return a class name for widget specified in aJS. "
	
	^self typeWidgetMapping
		at: typeName
		ifAbsent: [ #Mobile ,(typeName asSmalltalkName: true) ]! !

! MobileDevice methodsFor: #gui !
newFieldEditor: aString  field: aField

	^MobileWebFieldEditor for: aString field: aField! !