" Wrappers to Lua Standard Libraries.
Copyright (C) 2018 Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
See also Lua manual, section:
	5 - Standard Libraries
	http://www.lua.org/manual/5.1/manual.html#5
"!

NativeObject subclass: #LuaBasicFunctions category: #StandardLibrary!
LuaBasicFunctions comment: '
	Wrapper to Lua manual section 6.1 - Basic Functions
'!

! Lua class methodsFor: #accessing !
basic
	" Returns the wrapper to Lua basic functions. "

	^LuaBasicFunctions new! !

! LuaBasicFunctions methodsFor: #initialization !
initializeHandle
	" Private - Initialize the native handle. "

	handle := Smalltalk global.
	handle isNil ifTrue: [ ^self error: 'Missing access to Global scope' ].! !

! LuaBasicFunctions methodsFor: #assertions !
assert: value message: message
	" Trigger error if the value is false (i.e., nil or false); otherwise, returns all its arguments.
	In case of error, message is the error object; when nil, it defaults to assertion failed
	"

	^{lua'assert(value,message)'}! !

! LuaBasicFunctions methodsFor: #gc !
collectgarbage: option with: arg
	" Interface to the garbage collector.
	It performs different functions according to option.
	"

	^{lua'collectgarbage(option,arg)'}! !

! LuaBasicFunctions methodsFor: #gc !
collectgarbage: option
	" Interface to the garbage collector.
	It performs different functions according to option.
	"

	^{lua'collectgarbage(option)'}! !

! LuaBasicFunctions methodsFor: #gc !
fullGC
	" Performs a full garbage-collection cycle. "

	^self collectgarbage: #collect! !

! LuaBasicFunctions methodsFor: #gc !
gc
	" Performs a full garbage-collection cycle. "

	^self collectgarbage: #collect! !

! LuaBasicFunctions methodsFor: #gc !
stopGC
	" Stops automatic execution of the garbage collector.
	The collector will run only when explicitly invoked, until a call to restart it.
	"

	^self collectgarbage: #stop! !

! LuaBasicFunctions methodsFor: #gc !
restartGC
	" Restarts automatic execution of the garbage collector. "

	^self collectgarbage: #restart! !

! LuaBasicFunctions methodsFor: #gc !
countGC
	" Returns the total memory in use by Lua in Kbytes.
	The value has a fractional part, so that it multiplied by 1024 gives the exact number of bytes in use by Lua (except for overflows).
	"

	^self collectgarbage: #count! !

! LuaBasicFunctions methodsFor: #gc !
kBytesInUse
	" Returns the total memory in use by Lua in Kbytes. "

	^self collectgarbage: #count! !

! LuaBasicFunctions methodsFor: #gc !
stepGC: size
	" Performs a garbage-collection step.
	The step size is controlled by arg.
	With a zero value, the collector will perform one basic (indivisible) step.
	For non-zero values, the collector will perform as if that amount of memory (in KBytes) had been allocated by Lua.
	Returns true if the step finished a collection cycle.
	"

	^self collectgarbage: #step with: size! !

! LuaBasicFunctions methodsFor: #gc !
pauseGC: value
	" Sets the new value for the pause of the collector
	Returns the previous value for pause.
	"

	^self collectgarbage: #setpause with: value! !

! LuaBasicFunctions methodsFor: #gc !
stepGC: value
	" Sets the new value for the step multiplier of the collector.
	Returns the previous value for step.
	"

	^self collectgarbage: #setstepmul with: value! !

! LuaBasicFunctions methodsFor: #gc !
isRunningGC
	" Returns a boolean that tells whether the collector is running (i.e., not stopped). "

	^self collectgarbage: #isrunning! !

! LuaBasicFunctions methodsFor: #gc !
isCollecting
	" Returns a boolean that tells whether the collector is running (i.e., not stopped). "

	^self collectgarbage: #isrunning! !

! LuaBasicFunctions methodsFor: #evaluating !
doFile: fileName
	" Opens the named file and executes its contents as a Lua chunk.
	When fileName isNil, dofile executes the contents of the standard input (stdin).
	Returns all values returned by the chunk.
	In case of errors, dofile propagates the error (that is, dofile does not run in protected mode).
	"

	^{lua'dofile(fileName)'}! !

! LuaBasicFunctions methodsFor: #error !
error: message level: level
	" Terminates the last protected function called and returns message as the error object.
	This method never returns.
	Usually, error adds some information about the error position at the beginning of the message, if the message is a string.
	The level specifies how to get the error position.
	With level=1 (the default), the error position is where the error function was called.
	Level=2 points the error to where the function that called error was called; and so on.
	Passing a level=0 avoids the addition of error position information to the message.
	"

	^{lua'error(message,level)'}! !

! LuaBasicFunctions methodsFor: #utilities !
getmetatable: object
	" If object does not have a metatable, returns nil.
	Otherwise, if the object's metatable has a __metatable field, returns the associated value.
	Otherwise, returns the metatable of the given object.
	"

	^{lua'getmetatable(object)'}! !

! LuaBasicFunctions methodsFor: #utilities !
ipairs: t
	" Returns three values (an iterator function, the table t, and 0) so that the construction
		for i,v in ipairs(t) do body end
	will iterate over the key-value pairs (1,t[1]), (2,t[2]), ..., up to the first nil value.
	"

	^{lua'ipairs(t)'}! !

! LuaBasicFunctions methodsFor: #utilities !
pairs: t
	" Returns three values (the next function, the table t, and nil) so that the construction
		for i,v in pairs(t) do body end
	will iterate over the key-value pairs of t
	"

	^{lua'pairs(t)'}! !

! LuaBasicFunctions methodsFor: #utilities !
load: chunk chunkname: chunkname mode: mode env: env
	" Loads a chunk. "

	^{lua'load(chunk,chunkname,mode,env)'}! !

! LuaBasicFunctions methodsFor: #utilities !
load: chunk chunkname: chunkname mode: mode
	" Loads a chunk. "

	^{lua'load(chunk,chunkname,mode)'}! !

! LuaBasicFunctions methodsFor: #utilities !
load: chunk chunkname: chunkname
	" Loads a chunk. "

	^{lua'load(chunk,chunkname)'}! !

! LuaBasicFunctions methodsFor: #utilities !
load: chunk
	" Loads a chunk. "

	^{lua'load(chunk)'}! !

! LuaBasicFunctions methodsFor: #utilities !
loadfile: filename mode: mode env: env
	" Similar to load, but gets the chunk from file filename or from the standard input, if no filename is given. "

	^{lua'loadfile(filename,mode,env)'}! !

! LuaBasicFunctions methodsFor: #utilities !
loadfile: filename mode: mode
	" Similar to load, but gets the chunk from file filename or from the standard input, if no filename is given. "

	^{lua'loadfile(filename,mode)'}! !

! LuaBasicFunctions methodsFor: #utilities !
loadfile: filename
	" Similar to load, but gets the chunk from file filename or from the standard input, if no filename is given. "

	^{lua'loadfile(filename)'}! !

! LuaBasicFunctions methodsFor: #utilities !
next: table index: index
	" Traverse a field of a table. "

	^{lua'next(table,index)'}! !

! LuaBasicFunctions methodsFor: #utilities !
next: table
	" Returns an index to initial field of a table and its associated value. "

	^{lua'next(table)'}! !

! LuaBasicFunctions methodsFor: #utilities !
rawequal: v1 to: v2
	" Checks whether v1 is equal to v2, without activating the __eq metamethod.
	Returns a boolean.
	"

	^{lua'rawequal(v1,v2)'}! !

! LuaBasicFunctions methodsFor: #utilities !
rawget: table index: index
	" Returns the value of table[index], without invoking the __index metamethod.
	table must be a table; index may be any value. "

	^{lua'rawget(table,index)'}! !

! LuaBasicFunctions methodsFor: #utilities !
rawlen: v
	" Returns the length of the object v, which must be a table or a string, without invoking the __len metamethod.
	Returns an integer. "

	^{lua'rawlen(v)'}! !

! LuaBasicFunctions methodsFor: #utilities !
rawset: table index: index value: value
	" Sets the value of table[index] to value, without invoking the __newindex metamethod.
	table must be a table, index any value different from nil and NaN, and value any Lua value.
	Returns table. "

	^{lua'rawset(table,index,value)'}! !

! LuaBasicFunctions methodsFor: #utilities !
setmetatable: table to: metatable
	" Sets the metatable for the given table.
	(To change the metatable of other types from Lua code, you must use the debug library)
	If metatable is nil, removes the metatable of the given table.
	If the original metatable has a __metatable field, raises an error.
	Returns table.
	"

	^{lua'setmetatable(table,metatable)'}! !

! LuaBasicFunctions methodsFor: #utilities !
tonumber: e base: base

	^{lua'tonumber(e,base)'}! !

! LuaBasicFunctions methodsFor: #utilities !
tonumber: e

	^{lua'tonumber(e)'}! !

! LuaBasicFunctions methodsFor: #utilities !
tostring: v
	" Receives a value of any type and returns the result of converting to a string in a human-readable format. "

	^{lua'tostring(v)'}! !

! LuaBasicFunctions methodsFor: #utilities !
type: v
	" Returns the type of v. "

	^{lua'type(v)'}! !

! LuaBasicFunctions methodsFor: #utilities !
version
	" Returns a string containing the running Lua version. "

	^self handle basicAt: #_VERSION! !

NativeObject subclass: #Coroutine category: #StandardLibrary!
Coroutine comment: '
	Wrapper to Lua manual section 6.2 - Coroutine Manipulation
'!

! Lua class methodsFor: #accessing !
coroutine
	" Returns the wrapper to library support (aClass). "

	^Coroutine! !

! Coroutine class methodsFor: #accessing !
nativeLibrary
	" Private - returns the handle to the native module. "

	^#{coroutine}! !

! Coroutine class methodsFor: #instantiation !
new: aBlock
	" Returns an instance with body aBlock. "

	| result |
	{lua'	result = coroutine.create(aBlock) '}.
	^self @ result! !

! Coroutine class methodsFor: #instantiation !
wrap: aBlock
	" Returns a Block that resumes a created coroutine each time it is called. "

	^{lua'coroutine.wrap(aBlock)'}! !

! Coroutine class methodsFor: #query !
isyieldable
	" Returns true when the running coroutine can yield.
	A running coroutine is yieldable if it is not the main thread and it is not inside a non-yieldable C function.
	"

	^{lua'coroutine.isyieldable()'}! !

! Coroutine class methodsFor: #query !
yieldable
	" Returns true when the running coroutine can yield.
	A running coroutine is yieldable if it is not the main thread and it is not inside a non-yieldable C function.
	"

	^{lua'coroutine.isyieldable()'}! !

! Coroutine class methodsFor: #accessing !
running
	" Returns the running instance. "

	| result |
	{lua' result = coroutine.running()'}.
	^self @ result ! !

! Coroutine methodsFor: #executing !
resume
	" Starts or continues the execution of the receiver. "

	^{lua'coroutine.resume(self:handle())'}! !

! Coroutine methodsFor: #executing !
resume: arg1
	" Starts or continues the execution of the receiver. "

	^{lua'coroutine.resume(self:handle(),arg1)'}! !

! Coroutine methodsFor: #executing !
resume: arg1 with: arg2
	" Starts or continues the execution of the receiver. "

	^{lua'coroutine.resume(self:handle(),arg1,arg2)'}! !

! Coroutine methodsFor: #executing !
resume: arg1 with: arg2 with: arg3
	" Starts or continues the execution of the receiver. "

	^{lua'coroutine.resume(self:handle(),arg1,arg2,arg3)'}! !

! Coroutine methodsFor: #executing !
resume: arg1 with: arg2 with: arg3 with: arg4
	" Starts or continues the execution of the receiver. "

	^{lua'coroutine.resume(self:handle(),arg1,arg2,arg3,arg4)'}! !

! Coroutine methodsFor: #status !
status
	" returns the status of the receiver. "

	^{lua'coroutine.status(self:handle())'}! !

! Coroutine class methodsFor: #executing !
yield
	" Suspends the execution of the calling coroutine. "

	^{lua'coroutine.yield()'}! !

! Coroutine methodsFor: #executing !
yield: arg1
	" Suspends the execution of the calling coroutine.
	Any arguments to yield are passed as extra results to resume.
	"

	^{lua'coroutine.yield(arg1)'}! !

! Coroutine methodsFor: #executing !
yield: arg1 with: arg2
	" Suspends the execution of the calling coroutine.
	Any arguments to yield are passed as extra results to resume.
	"

	^{lua'coroutine.yield(arg1,arg2)'}! !

! Coroutine methodsFor: #executing !
yield: arg1 with: arg2 with: arg3
	" Suspends the execution of the calling coroutine.
	Any arguments to yield are passed as extra results to resume.
	"

	^{lua'coroutine.yield(arg1,arg2,arg3)'}! !

! Coroutine methodsFor: #executing !
yield: arg1 with: arg2 with: arg3 with: arg4
	" Suspends the execution of the calling coroutine.
	Any arguments to yield are passed as extra results to resume.
	"

	^{lua'coroutine.yield(arg1,arg2,arg3,arg4)'}! !

NativeObject subclass: #Package category: #StandardLibrary!
Package comment: '
	Wrapper to Lua manual section 6.3 - Modules
'!

! Lua class methodsFor: #accessing !
package
	" Returns the wrapper to library support (aClass). "

	^Package! !

! Lua class methodsFor: #modules !
require: moduleName
	" Returns the package wrapper to a loaded module. "

	^Package require: moduleName! !

! Package class methodsFor: #loading !
require: moduleName
	" Returns the package wrapper to a loaded module. "

	| module |
	{lua'	module = require(moduleName) '}.
	^self @ module! !

! Package class methodsFor: #accessing !
config
	" Returns a string describing some compile-time configurations for packages. "

	^{lua'package.config'}! !

! Package class methodsFor: #accessing !
cpath
	" Returns the path used by require to search for a C loader. "

	^{lua'package.cpath'}! !

! Package class methodsFor: #accessing !
loaded
	" Returns a PoolDictionary with the modules already loaded. "

	| copy |
	{lua' copy = package.loaded '}.
	^PoolDictionary @@ copy! !

! Package class methodsFor: #accessing !
loadlib: libname funcname: funcname
	" Dynamically links the host program with the C library libname. "

	^{lua'package.loadlib(libname,funcname)'}! !

! Package class methodsFor: #accessing !
path
	" Returns the path used by require to search for a Lua loader. "

	^{lua'package.path'}! !

! Package class methodsFor: #accessing !
preload
	" Returns a PoolDictionary with a copy of the table to store loaders for specific modules. "

	| copy |
	{lua' copy = package.preload '}.
	^PoolDictionary @@ copy! !

! Package class methodsFor: #accessing !
searchers
	" Returns an Array with functions used by require to control how to load modules. "

	| array |
	{lua' array = package.searchers '}.
	^array notNil ifTrue: [ Array @ array ]! !

! Package class methodsFor: #accessing !
searchers
	" Returns an Array with functions used by require to control how to load modules. "

	| array |
	{lua' array = package.searchers '}.
	^array notNil ifTrue: [ Array @ array ]! !

! Package class methodsFor: #accessing !
searchpath: name in: path with: sep with: rep
	" Searches for the given name in the given path.
	Returns the resulting name of the first file that it can open in read mode (after closing the file), or nil plus an error message if none succeeds. 
	"

	^{lua' package.searchpath (name,path,sep,rep) '}! !

! Package class methodsFor: #accessing !
searchpath: name in: path with: sep
	" Searches for the given name in the given path.
	Returns the resulting name of the first file that it can open in read mode (after closing the file), or nil plus an error message if none succeeds. 
	"

	^{lua' package.searchpath (name,path,sep) '}! !

! Package class methodsFor: #accessing !
searchpath: name in: path
	" Searches for the given name in the given path.
	Returns the resulting name of the first file that it can open in read mode (after closing the file), or nil plus an error message if none succeeds. 
	"

	^{lua' package.searchpath (name,path) '}! !

"Implemented by S8 String
6.4 - String Manipulation
6.4.1 - Patterns
6.4.2 - Format Strings for Pack and Unpack
"

NativeObject subclass: #UTF8 category: #StandardLibrary!
UTF8 comment: '
	Wrapper to Lua manual section 6.5 - UTF-8 Support
'!

! Lua class methodsFor: #accessing !
utf8
	" Returns the wrapper to library support (aClass). "

	^UTF8! !

! UTF8 class methodsFor: #converting !
char: arg
	" Receives an integer or Array (of integers), converts each one to its corresponding UTF-8 byte sequence and returns a string with the concatenation of all these sequences. "

	^arg isNumber
	ifTrue:  [ {lua'utf8.char(arg) '} ]
	ifFalse: [ {lua'utf8.char(unpack(arg)) '} ]! !

! UTF8 class methodsFor: #constants !
charpattern
	" Returns the pattern [\0-\x7F\xC2-\xF4][\x80-\xBF]* which matches exactly one UTF-8 byte sequence, assuming that the subject is a valid UTF-8 string. "

	^{lua'utf8.charpattern'}! !

! UTF8 class methodsFor: #converting !
codes: s
	^{lua'utf8.codes(s)'}! !

! UTF8 class methodsFor: #converting !
codepoint: s from: i to: j
	" Returns the codepoints (as integers) from all characters in s. "

	^{lua'utf8.codepoint(s,i,j)'}! !

! UTF8 class methodsFor: #converting !
	" Returns the codepoints (as integers) from all characters in s. "

codepoint: s from: i
	^{lua'utf8.codepoint(s,i)'}! !

! UTF8 class methodsFor: #converting !
codepoint: s
	" Returns the codepoints (as integers) from all characters in s. "

	^{lua'utf8.codepoint(s)'}! !

! UTF8 class methodsFor: #converting !
len: s from: i to: j
	" Returns the number of UTF-8 characters in string s. "

	^{lua'utf8.len(s,i,j)'}! !

! UTF8 class methodsFor: #converting !
len: s from: i
	" Returns the number of UTF-8 characters in string s. "

	^{lua'utf8.len(s,i)'}! !

! UTF8 class methodsFor: #converting !
len: s
	" Returns the number of UTF-8 characters in string s. "

	^{lua'utf8.len(s)'}! !

! UTF8 class methodsFor: #converting !
offset: s n: n i: i
	" Returns the position (in bytes) where the encoding of the n-th character of s (counting from position i) starts. "

	^{lua'utf8.offset(s,n,i)'}! !

! UTF8 class methodsFor: #converting !
offset: s n: n
	" Returns the position (in bytes) where the encoding of the n-th character of s starts. "

	^{lua'utf8.offset(s,n)'}! !

"Implemented by S8 collections & numbers
6.6 - Table Manipulation
6.7 - Mathematical Functions
"!

NativeObject subclass: #LuaFile category: #StandardLibrary!
LuaFile comment: '
	Wrapper to Lua manual section 6.8 - Input and Output Facilities
'!

! Lua class methodsFor: #accessing !
io
	" Returns the wrapper to library support (aClass). "

	^LuaFile! !

! Lua class methodsFor: #accessing !
file
	" Returns the wrapper to library support (aClass). "

	^LuaFile! !

! LuaFile class methodsFor: #io !
close
	" Close the default output file. "

	^{lua'io.close()'}! !

! LuaFile class methodsFor: #io !
close: hFile
	" Close the file. "

	^{lua'io.close(hFile)'}! !

! LuaFile class methodsFor: #io !
flush
	" Flush the default output file. "

	^{lua'io.flush()'}! !

! LuaFile class methodsFor: #io !
input: hFile
	" When called with a file name, it opens the named file (in text mode), and sets its handle as the default input file.
	When called with a file handle, it simply sets this file handle as the default input file.
	When called without arguments, it returns the current default input file.
	In case of errors this function raises the error, instead of returning an error code. "

	| result |
	{lua' result = io.input(hFile)'}.
	^self @ result! !

! LuaFile class methodsFor: #io !
input
	" Returns the current default input file. "

	| result |
	{lua' result = io.input()'}.
	^self @ result! !

! LuaFile class methodsFor: #io !
output: hFile
	" When called with a file name, it opens the named file (in text mode), and sets its handle as the default output file.
	When called with a file handle, it simply sets this file handle as the default output file.
	When called without arguments, it returns the current default output file.
	In case of errors this function raises the error, instead of returning an error code. "

	| result |
	{lua' result = io.output(hFile)'}.
	^self @ result! !

! LuaFile class methodsFor: #io !
output
	" Returns the current default output file. "

	| result |
	{lua' result = io.output()'}.
	^self @ result! !

! LuaFile class methodsFor: #io !
lines: filename withAll: array
	" Opens the given file name in read mode and returns an iterator function that works like file:lines(...) over the opened file. "

	| result |
	{lua' result = io.lines(filename,unpack(array))'}.
	^self @ result! !

! LuaFile class methodsFor: #io !
lines
	" Returns an iterator over the lines of the default input file. "

	| result |
	{lua' result = io.lines()'}.
	^self @ result! !

! LuaFile class methodsFor: #io !
open: filename mode: mode
	" Returns an instance. Opens a file, in the mode specified in the string mode. "

	| result |
	{lua' result = io.open(filename,mode)'}.
	^self @ result! !

! LuaFile class methodsFor: #io !
open: filename
	" Returns an instance. Opens a file, in default(read) mode. "

	| result |
	{lua' result = io.open(filename)'}.
	^self @ result! !

! LuaFile class methodsFor: #io !
popen: program mode: mode
	" Starts program in a separated process and returns a file that you can use to read data from this program (if mode is $r) or to write data to this program (if mode is $w).
	Note: System dependent and not available on all platforms.
	"

	| result |
	{lua' result = io.popen(program,mode)'}.
	^self @ result! !

! LuaFile class methodsFor: #io !
popen: program
	" Starts program in a separated process and returns a file that you can use to read data from this program.
	Note: System dependent and not available on all platforms.
	"

	| result |
	{lua' result = io.popen(program)'}.
	^self @ result! !

! LuaFile class methodsFor: #io !
tmpfile
	" Returns a temporary file.
	This file is opened in update mode and it is automatically removed when the program ends.
	"

	| result |
	{lua' result = io.tmpfile()'}.
	^self @ result! !

! LuaFile class methodsFor: #io !
type: obj
	" Checks whether obj is a valid file handle.
	Returns #file if obj is an open file handle, 'closed file' if obj is a closed file handle, or nil if obj is not a file handle.
	"

	^{lua'io.type(obj)'}! !

! LuaFile methodsFor: #state !
state
	" Returns the receiver's handle state.
	Returns #file if obj is an open file handle, 'closed file' if obj is a closed file handle, or nil if obj is not a file handle.
	"

	| state |
	state := self class type: self handle.
	#file = state ifTrue: [ ^#open ].
	#'closed file' = state ifTrue: [ ^#closed ].
	^state! !

! LuaFile methodsFor: #lines !
lines: format
	" Returns an iterator function that, each time it is called, reads the file according to the given format. "

	^self handle #lines: format! !

! LuaFile methodsFor: #lines !
linesWithAll: formats
	" Returns an iterator function that, each time it is called, reads the file according to the given formats. "

	^{lua'self:handle():lines(unpack(formats))'}! !

! LuaFile methodsFor: #reading !
readWithAll: formats
	" Reads the file file, according to the given formats, which specify what to read. "

	^{lua'self:handle():read(unpack(formats))'}! !

! LuaFile methodsFor: #reading !
writeWithAll: values
	" Writes the value of each of its arguments to file.
	The values must be strings or numbers.
	In case of success, this function returns the handle.
	Otherwise it returns nil plus a string describing the error.
	"

	^{lua'self:handle():write(unpack(values))'}! !

LuaFile buildMethods: #(
	close flush seek read write
	#(lines #(format))
	#(read #(format))
	#(write #(value))
	#(seek #(whence))
	#(seek #(whence offset))
	#(setvbuf #(mode))
	#(setvbuf #(mode size))
)!

NativeObject subclass: #OS category: #StandardLibrary!
LuaFile comment: '
	Wrapper to Lua manual section 6.9 - Operating System Facilities
'!

! Lua class methodsFor: #accessing !
os
	" Returns the wrapper to library support (aClass). "

	^OS! !

! OS class methodsFor: #accessing !
clock
	" Returns an approximation of the amount in seconds of CPU time used by the system. "

	^{lua'os.clock()'}! !

! OS class methodsFor: #datetime !
date: format time: time
	" Returns a string containing date and time. "

	^{lua'os.date(format,time)'}! !

! OS class methodsFor: #datetime !
date: format
	" Returns a string or a table containing date and time, formatted according to the given string format.
		Lua@>(OS date: '*t')
		OS date: 'today is %A, in %B'
	"

	^{lua'os.date(format)'}! !

! OS class methodsFor: #datetime !
date
	" Returns a string containing date and time. "

	^{lua'os.date()'}! !

! OS class methodsFor: #datetime !
difftime: t2 from: t1
	" Returns the difference, in seconds, from time t1 to time t2. "

	^{lua'os.difftime(t2,t1)'}! !

! OS class methodsFor: #util !
execute: command
	" It passes command to be executed by an operating system shell.
	Its first result is true if the command terminated successfully, or nil otherwise.
	"

	| result string number |
{lua'	result,string,number = os.execute(command)
'}.	^result == true ifTrue: [ Array with: string with: number ]! !

! OS class methodsFor: #util !
exit: code close: close
	" Terminate the host program. "

	{lua'os.exit(code,close)'}! !

! OS class methodsFor: #util !
exit: code
	" Terminate the host program. "

	{lua'os.exit(code)'}! !

! OS class methodsFor: #util !
exit
	" Terminate the host program. "

	{lua'os.exit()'}! !

! OS class methodsFor: #util !
getenv: varname
	" Returns the value of the process environment variable varname, or nil if the variable is not defined. "

	^{lua'os.getenv(varname)'}! !

! OS class methodsFor: #files !
remove: filename ifFailed: aBlock
	" Deletes the file (or empty directory, on POSIX systems) with the given name. "

	| success description code |
{lua'	success,description,code = os.remove(filename)'}.
	success ~= true ifTrue: [ ^aBlock value: description value: code ].
	^success! !

! OS class methodsFor: #files !
rename: filename to: newname ifFailed: aBlock
	" Renames the file to newname. "

	| success description code |
{lua'	success,description,code = os.rename(filename,newname)'}.
	success ~= true ifTrue: [ ^aBlock value: description value: code ].
	^success! !

! OS class methodsFor: #util !
setlocale: locale category: category
	" Sets the current locale. "

	^{lua'os.setlocale(locale,category)'}! !

! OS class methodsFor: #util !
setlocale: locale
	" Sets the current locale. "

	^{lua'os.setlocale(locale)'}! !

! OS class methodsFor: #util !
getlocale: category
	" Gets the current locale. "

	^{lua'os.setlocale(nil,category)'}! !

! OS class methodsFor: #datetime !
time: table
	" Returns a time from table containing values for year,month,day and optionally hour,min,sec. "

	^{lua'os.time(table)'}! !

! OS class methodsFor: #files !
tmpname
	" Returns a string with a file name that can be used for a temporary file.
	The file must be explicitly opened before its use and explicitly removed when no longer needed.
	"

	^{lua'os.tmpname()'}! !

NativeObject subclass: #Debug category: #StandardLibrary!
NativeObject
	subclass: #DebugInfo
	instanceVariableNames: ' index '
	category: #StandardLibrary!
Debug comment: '
	Wrapper to Lua manual section 6.10 - The Debug Library
'!

! Lua class methodsFor: #accessing !
debug
	" Returns the wrapper to library support (aClass). "

	^Debug! !

! Debug class methodsFor: #debug !
debug
	" Enters an interactive mode with the user, running each string that the user enters.
	Using simple commands and other debug facilities, the user can inspect global and local variables, change their values, evaluate expressions, and so on.
	A line containing only the word cont finishes this function, so that the caller continues its execution.
	"

	^{lua'debug.debug()'}! !

! Debug class methodsFor: #accessing !
gethook
	" Returns the current hook settings of the current thread. "

	| hook mask count |
{lua'	hook,mask,count  = debug.gethook()'}.
	^Array with: hook with: mask with: count! !

! Debug class methodsFor: #accessing !
gethook: thread
	" Returns the current hook settings of the thread. "

	| hook mask count |
{lua'	hook,mask,count  = debug.gethook(thread)'}.
	^Array with: hook with: mask with: count! !

! Debug class methodsFor: #accessing !
getinfo: thread f: f what: what
	" Returns a table with information about a function. "

	^DebugInfo index: f at: ({lua'debug.getinfo(thread,f,what)'})! !

! Debug class methodsFor: #accessing !
getinfo: thread f: f
	" Returns a table with information about a function. "

	^DebugInfo index: f at: ({lua'debug.getinfo(thread,f)'})! !

! Debug class methodsFor: #accessing !
getinfo: f what: what
	" Returns a table with information about a function. "

	^DebugInfo index: f at: ({lua'debug.getinfo(f,what)'})! !

! Debug class methodsFor: #accessing !
getinfo: f
	" Returns a table with information about a function. "

	^DebugInfo index: f at: ({lua'debug.getinfo(f)'})! !

! Debug class methodsFor: #accessing !
getlocal: thread f: f local: index
	" Returns the name and the value of the local variable at index of the function at level f (aNumber) of the stack.
	Accesses not only explicit local variables, but also parameters, temporaries, etc.
	The first parameter or local variable has index 1, and so on, following the order that they are declared in the code, counting only the variables that are active in the current scope of the function.
	Negative indices refer to vararg arguments; -1 is the first vararg argument.
	Returns nil if there is no variable with the given index, and raises an error when called with a level out of range.
	Variable names starting with $( (open parenthesis) represent variables with no known names
	 (internal variables such as loop control variables, and variables from chunks saved without debug information).
	f may also be a function. In that case, getlocal returns only the name of function parameters.
	"

	| aName value |
{lua'	aName,value  = debug.getlocal(thread,f,index)'}.
	aName isNil ifTrue: [ ^nil ].
	^aName -> value ! !

! Debug class methodsFor: #accessing !
getlocal: f local: index
	" Returns the name and the value of the local variable at index of the function at level f (aNumber) of the stack. "

	| aName value |
{lua'	aName,value  = debug.getlocal(f,index)'}.
	aName isNil ifTrue: [ ^nil ].
	^aName -> value ! !

! Debug class methodsFor: #accessing !
getlocals: f
	" Returns the name and the value of the local variables of the function at level f (aNumber) of the stack. "

	| result tuple |
	result := #().
	[	tuple := self getlocal: f local: result size + 1.
		tuple notNil ] whileTrue: [ result add: tuple ].
	^result ! !

! Debug class methodsFor: #accessing !
getvarargs: f
	" Returns the name and the value of the variable arguments of the function at level f (aNumber) of the stack. "

	| result tuple |
	result := #().
	[	tuple := self getlocal: f local: result size + 1 * -1.
		tuple notNil ] whileTrue: [ result add: tuple ].
	^result ! !

! Debug class methodsFor: #accessing !
getmetatable: value
	" Returns the metatable of the given value or nil if it does not have a metatable. "

	^{lua'debug.getmetatable(value)'}! !

! Debug class methodsFor: #accessing !
setmetatable: value to: table
	" Set the metatable of the given value to a metatable (or nil). "

	^{lua'debug.setmetatable(value,table)'}! !

! Debug class methodsFor: #accessing !
getregistry
	" Returns the registry table. "

	^{lua'debug.getregistry()'}! !

! Debug class methodsFor: #accessing !
registry
	" Returns the registry table. "

	^{lua'debug.getregistry()'}! !

! Debug class methodsFor: #accessing !
getupvalue: aFunction at: index
	" Returns  the name and the value of the upvalue at index of aFunction (a closure).
	Returns nil if there is no upvalue with the given index.
	Variable names starting with $( (open parenthesis) represent variables with no known names (variables from chunks saved without debug information).
	"

	| aName value |
{lua'	aName,value  = debug.getupvalue(aFunction,index)'}.
	aName isNil ifTrue: [ ^nil ].
	^aName -> value ! !

! Debug class methodsFor: #accessing !
getupvalues: aFunction
	" Returns the name and the value of the upvalues of aFunction. "

	| result tuple |
	result := #().
	[	tuple := self getupvalue: aFunction at: result size + 1.
		tuple notNil ] whileTrue: [ result add: tuple ].
	^result ! !

! Debug class methodsFor: #accessing !
getuservalue: u
	" Returns the Lua value associated to u.
	If u is not a full userdata, returns nil.
	"

	^{lua'debug.getuservalue(u)'}! !

! Debug class methodsFor: #accessing !
sethook: thread hook: hook mask: mask count: count
	" Sets the given function as a hook. "

	^{lua'debug.sethook(thread,hook,mask,count)'}! !

! Debug class methodsFor: #accessing !
sethook: hook mask: mask count: count
	" Sets the given function as a hook. "

	^{lua'debug.sethook(hook,mask,count)'}! !

! Debug class methodsFor: #accessing !
sethook: thread hook: hook mask: mask
	" Sets the given function as a hook. "

	^{lua'debug.sethook(thread,hook,mask)'}! !

! Debug class methodsFor: #accessing !
sethook: hook mask: mask
	" Sets the given function as a hook. "

	^{lua'debug.sethook(hook,mask)'}! !

! Debug class methodsFor: #accessing !
setlocal: thread level: level local: index value: value
	" Assigns the value value to the local variable at index of the function at level level of the stack.
	Returns nil if there is no local variable with the given index, and raises an error when called with a level out of range.
	"

	^{lua'debug.setlocal(thread,level,index,value)'}! !

! Debug class methodsFor: #accessing !
setlocal: level local: index value: value
	" Assigns the value value to the local variable at index of the function at level level of the stack.
	Returns nil if there is no local variable with the given index, and raises an error when called with a level out of range.
	"

	^{lua'debug.setlocal(level,index,value)'}! !

! Debug class methodsFor: #accessing !
setupvalue: f up: up value: value
	" Assigns the value value to the upvalue with index up of the function f.
	Returns nil if there is no upvalue with the given index.
	Otherwise, it returns the name of the upvalue.
	"

	^{lua'debug.setupvalue(f,up,value)'}! !

! Debug class methodsFor: #accessing !
setuservalue: udata to: value
	" Sets the given value as the Lua value associated to the given udata.
	udata must be a full userdata.
	Returns udata. "

	^{lua'debug.setuservalue(udata,value)'}! !

! Debug class methodsFor: #accessing !
traceback: thread message: message level: level
	" If message is present but is neither a string nor nil, this function returns message without further processing.
	Otherwise, it returns a string with a traceback of the call stack.
	The optional message string is appended at the beginning of the traceback.
	An optional level number tells at which level to start the traceback
	"

	^{lua'debug.traceback(thread,message,level)'}! !

! Debug class methodsFor: #accessing !
traceback: thread message: message

	^{lua'debug.traceback(thread,message)'}! !

! Debug class methodsFor: #accessing !
traceback: thread

	^{lua'debug.traceback(thread)'}! !

! Debug class methodsFor: #accessing !
traceback: message level: level

	^{lua'debug.traceback(message,level)'}! !

! Debug class methodsFor: #accessing !
upvalueid: f at: n
	" Returns a unique identifier (as a light userdata) for the upvalue numbered n from the given function.
	These unique identifiers allow to check whether different closures share upvalues.
	Lua closures that share an upvalue (that is, that access a same external local variable) will return identical ids for those upvalue indices.
	"

	^{lua'debug.upvalueid(f,n)'}! !

! Debug class methodsFor: #util !
upvaluejoin: f1 n1: n1 f2: f2 n2: n2
	" Make the n1-th upvalue of the Lua closure f1 refer to the n2-th upvalue of the Lua closure f2. "

	^{lua'debug.upvaluejoin(f1,n1,f2,n2)'}! !

! Debug class methodsFor: #util !
getfenv: obj
	" Returns the environment of object obj.
	Returns the environment of the argument, which is only meaningful for objects of type function, userdata or thread.
	"

	^{lua'debug.getfenv(obj)'}! !

! Debug class methodsFor: #stack !
stackFrom: level do: aBlock
	" Evaluates aBlock with each info extracted from stack at level and up. "

	| info n |
	n := level max: 0.
	[	info := self getinfo: n.
		info notNil
	] whileTrue: [
		aBlock evaluateWith: info.
		n := n + 1.
	]! !

! Debug class methodsFor: #stack !
stackDo: aBlock
	" Evaluates aBlock with each info extracted from stack. "

	^self stackFrom: 0 do: aBlock! !

! Debug class methodsFor: #stack !
stackFrom: level
	" Returns an array with information extracted from stack. "

	| result |
	result := #().
	self stackFrom: level do: [:info| result add: info ].
	^result! !

! Debug class methodsFor: #stack !
stack
	" Returns an array with information extracted from stack. "

	^self stackFrom: 0! !

! DebugInfo class methodsFor: #instantiation !
index: index at: aHandle
	" Return an instance of the receiver (or nil). "

	^aHandle notNil ifTrue: [
		(self @ aHandle) index: index
	]! !

! DebugInfo methodsFor: #accessing !
index
	^index ! !

! DebugInfo methodsFor: #accessing !
index: anIndex
	index := anIndex! !

! DebugInfo methodsFor: #accessing !
currentLine
	" Return the line that is currently active (only applies to active functions). "

	^self handle basicAt: #currentline! !

! DebugInfo methodsFor: #accessing !
source
	" Returns where the function was defined.
	If in a file, it is the file name prefixed by $@.
	If the function was defined in a string returns the source string (or '=(load)').
	If the function was defined interactively returns #stdin
	If the function was defined in a C program, then source will be '[C]'.
	"

	^self handle basicAt: #source! !
 
! DebugInfo methodsFor: #accessing !
lineNumber
	" Returns the first line number, in the source, where this function was defined (for Lua functions). "

	^self handle basicAt: #linedefined ! !

! DebugInfo methodsFor: #accessing !
lastLine
	" Returns the last line number, in the source, where this function was defined (for Lua functions). "

	^self handle basicAt: #lastlinedefined ! !

! DebugInfo methodsFor: #accessing !
lineInterval
	" Returns the interval for line source. "

	^self lineNumber to: self lastLine! !

! DebugInfo methodsFor: #accessing !
what
	" Returns what this function is.
	Can be #Lua for a regular Lua function, #C if it is a C function, or #main if it is part of the main Lua chunk (ie. outside any function)."

	^self handle basicAt: #what ! !

! DebugInfo methodsFor: #accessing !
name
	" Returns the name (or nil)."

	^self handle basicAt: #name ! !

! DebugInfo methodsFor: #accessing !
nameWhat
	" Returns the what the #name means.
	It can be #global, #local, #method, #field, or an empty string.
	The empty string means Lua did not find the name, and thus the name field will not be present."

	^self handle basicAt: #namewhat ! !

! DebugInfo methodsFor: #accessing !
nups
	" Returns the number of upvalues that this function has. "

	^self handle basicAt: #nups ! !

! DebugInfo methodsFor: #accessing !
activeLines
	" Returns a table of the active lines of the function. "

	^self handle basicAt: #activelines ! !

! DebugInfo methodsFor: #accessing !
currentLine
	" Returns the line that is currently active (or -1) "

	^self handle basicAt: #currentline ! !

! DebugInfo methodsFor: #accessing !
function
	" Returns the function that is active at that stack level "

	^self handle basicAt: #func ! !

! DebugInfo methodsFor: #accessing !
locals
	" Returns the locals defined in the receiver. "

	^Debug getlocals: index ! !

! DebugInfo methodsFor: #accessing !
varargs
	" Returns the variable arguments defined in the receiver. "

	^Debug getvarargs: index ! !

! DebugInfo methodsFor: #accessing !
upvalues
	" Returns the upvalues defined in the functions of receiver. "

	^Debug getupvalues: self function ! !

"-----DateTime implementation------------------------------------------------------------"!

! DateTime methodsFor: #internal !
table
	" Private - Return the internal table of the receiver. "

	^self basicAt: #s8Table ifAbsentPut: [
		Lua @ (OS date: '*t')
	]! !

! DateTime methodsFor: #accessing !
day
	" Return the day of the receiver. "

	^self table @ #day! !

! DateTime methodsFor: #accessing !
day: aNumber
	" Set the day of the receiver. "

	self table handleAt: #day put: aNumber! !

! DateTime methodsFor: #accessing !
dayOfMonth
	" Return the day of month of the receiver. "

	^self table @ #day! !

! DateTime methodsFor: #accessing !
dayOfMonth: aNumber
	" Set the day of month of the receiver. "

	self table handleAt: #day put: aNumber! !

! DateTime methodsFor: #accessing !
dayOfWeek
	" Return the day of week of the receiver. "

	^self table @ #wday! !

! DateTime methodsFor: #accessing !
dayOfWeek: aNumber
	" Set the day of week of the receiver. "

	self table handleAt: #wday put: aNumber! !

! DateTime methodsFor: #accessing !
hours
	" Return the hours of the receiver. "

	^self table @ #hour! !

! DateTime methodsFor: #accessing !
hours: aNumber
	" Set the hours of the receiver. "

	self table handleAt: #hour put: aNumber! !

! DateTime methodsFor: #accessing !
milliseconds
	" Return the milliseconds of the receiver. "

	^self notImplementedYet: #milliseconds! !

! DateTime methodsFor: #accessing !
milliseconds: aNumber
	" Set the milliseconds of the receiver. "

	^self notImplementedYet: #milliseconds:! !

! DateTime methodsFor: #accessing !
minutes
	" Return the minutes of the receiver. "

	^self table @ #min! !

! DateTime methodsFor: #accessing !
minutes: aNumber
	" Set the minutes of the receiver. "

	self table handleAt: #min put: aNumber! !

! DateTime methodsFor: #accessing !
month
	" Return the month of the receiver. "

	^self table @ #month! !

! DateTime methodsFor: #accessing !
month: aNumber
	" Set the month of the receiver. "

	self table handleAt: #month put: aNumber! !

! DateTime methodsFor: #accessing !
seconds
	" Return the seconds of the receiver. "

	^self table @ #sec! !

! DateTime methodsFor: #accessing !
seconds: aNumber
	" Set the seconds of the receiver. "

	self table handleAt: #sec put: aNumber! !

! DateTime methodsFor: #accessing !
time
	" Return the time of the receiver. "

	^self notImplementedYet: #time! !

! DateTime methodsFor: #accessing !
time: aNumber
	" Set the time of the receiver. "

	^self notImplementedYet: #time:! !

! DateTime methodsFor: #accessing !
year
	" Return the year of the receiver. "

	^self table @ #year! !

! DateTime methodsFor: #accessing !
year: aNumber
	" Set the year of the receiver. "

	self table handleAt: #year put: aNumber! !

! DateTime methodsFor: #converting !
asDateString
	" Return the date string of the receiver. "

	^self notImplementedYet: #asDateString! !

! DateTime methodsFor: #converting !
asLocaleString
	" Return the locale string of the receiver. "

	^self notImplementedYet: #asLocaleString! !

! DateTime methodsFor: #converting !
asMilliseconds
	" Return the receiver as milliseconds. "

	^self notImplementedYet: #asMilliseconds! !

! DateTime methodsFor: #converting !
asNumber
	" Return the receiver as a Number. "

	^OS time: self table handle! !

! DateTime methodsFor: #converting !
asTimeString
	" Return the time string of the receiver. "

	^self notImplementedYet: #asTimeString! !
