
EventManager
	subclass: #BootSector
	instanceVariableNames: ' lang stream indentation '
	category: #Eden!
BootSector comment: '
	Native representation of S8 boot section of an image (as a String).
	The subclasses refine behavior to jump to other runtimes.
	@2023-today Alejandro Reimondo aleReimondo@smalltalking.net

	(BootSector for: #js) contents

'!
BootSector buildAccessors: #( lang )!

! BootSector class methodsFor: #testing !
encode: aSelector
	" Private, testing - Retruns the encoding for implementation of aSelector. "

	| tester |
	tester := self new.
	^tester encode: [ tester perform: aSelector ] ! !

! BootSector class methodsFor: #instantiation !
for: nativeLanguage
	" Returns an instance of the receiver encoding to nativeLanguage. "

	| implementor |
	implementor := self
		supportFor: nativeLanguage
		ifAbsent: [ self error: 'Missing implementation for ' ,nativeLanguage asLiteral ].
	^implementor new
		lang: nativeLanguage;
		yourself! !

! BootSector class methodsFor: #private !
supportFor: nativeLanguage ifAbsent: aBlock
	" Private - Returns implementor for nativeLanguage. "

	^self allSubclasses
		detect: [:one| one supportsLanguage: nativeLanguage ]
		ifNone: aBlock! !

! BootSector class methodsFor: #instantiation !
supportsLanguage: aLanguage
	" Private - Return true if instances of the receiver can encode to aLanguage. "

	^false! !

! BootSector class methodsFor: #reading !
storages
	" Private - Return an Array of pathNames to external folders containing literal contents of instances.
	The default action returns an empty collection.
	This method can be refined by subclasses to support external repositories.
	See also uses of #u8Storages "

	^#()! !

! BootSector methodsFor: #reading !
u8Storages
	" Private - Return an Array of pathNames to external folders containing literal contents of instances. "

	^#(	's8/start/'
		'u8:aleReimondo/s8/start/'
	)! !

! BootSector methodsFor: #contents !
contentsFromStorages: storages
	" Private - Return the contents of the receiver read from external storage (or nil). "

	storages do: [:folder| | result pathName |
		pathName := folder ,'boot.' ,self lang.
		[ result := pathName fileContents ] on: Exception do: ["ignore error reading contents"].
		result isString ifTrue: [ ^result ].
	].
	^nil! !

! BootSector methodsFor: #contents !
contentsFromStorages
	" Private - Return the contents of the receiver read from external storage (or nil). "

	^self contentsFromStorages: self u8Storages! !

! BootSector methodsFor: #contents !
contents
	" Return the contents (aString) to be dump as source of the receiver. "

	^self header
		,self definitions
		,self startupExpression! !

! BootSector methodsFor: #contents !
header
	" Private - Return the header of the receiver. "

	^self multilineComment: self headerLines! !

! BootSector methodsFor: #contents !
definitions
	" Private - Return the encoding for global definitions of the receiver. "

	^self initialDefinition
	,self nilDefinition
	,self smalltalkDefinition! !

! BootSector methodsFor: #contents !
startupExpression
	" Private - Return the encoding for startup expression of the receiver. "

	^self encode: [ self send: #startup to: #smalltalk;cr ]! !

! BootSector methodsFor: #private !
encode: aBlock
	" Private - Return the contents encoded in aBlock. "

	| back result |
	back := stream.
	stream := '' writeStream.
	[	aBlock evaluate.
		result := stream contents
	] ensure: [ stream := back ].
	^result ! !

! BootSector methodsFor: #indentation !
indentation
	" Private - Return the indentation of the receiver. "

	indentation isNil ifTrue: [ indentation := 0 ].
	^indentation ! !

! BootSector methodsFor: #indentation !
indent
	" Private - Enter the next level of indentation. "

	indentation := self indentation + 1! !

! BootSector methodsFor: #indentation !
dedent
	" Private - Exit the current level of indentation. "

	indentation := self indentation - 1 max: 0! !

! BootSector methodsFor: #dump !
tab
	" Private - Encode a tab character. "

	stream tab ! !

! BootSector methodsFor: #dump !
space
	" Private - Encode a space character. "

	stream space ! !

! BootSector methodsFor: #dump !
cr
	" Private - Encode a new line. "

	stream cr.
	self indentation timesRepeat: [ self tab ]! !

! BootSector methodsFor: #dump !
nextPutAll: aString
	" Private - Encode aString to receiver's stream. "

	aString isNil ifTrue: [ ^self ].
	aString isArray ifTrue: [
		aString
			do: [:line| self nextPutAll: line ]
			separatedBy: [ self cr ].
		^self
	].
	aString isString ifTrue: [
		stream nextPutAll: aString.
		^self
	].
	aString isValuable ifTrue: [ ^aString evaluate ].
	^self error: 'Invalid argument to encode - ' ,aString ! !

! BootSector methodsFor: #dump !
nextPutLines: lines
	" Private - Encode lines to receiver's stream. "

	| src |
	lines isArray ifTrue: [ ^lines do: [:each| self nextPutAll: each ] separatedBy: [self cr] ].
	src := lines stream.
	[src atEnd] whileFalse: [
		self nextPutAll: src nextLine.
		src atEnd ifFalse: [ self cr ]
	] ! !

! BootSector methodsFor: #encoding !
lineComment: aString
	" Private - Encode aString as line comment. "

	^self implementedBySubclass: #lineComment: ! !

! BootSector methodsFor: #encoding !
multilineComment: array
	" Private - Encode an array as a multiline comment. "

	^self encode: [ array do: [:line| self lineComment: line ] ]! !

! BootSector methodsFor: #encoding !
comment: aString
	" Private - Encode aString as comment. "

	aString isNil ifTrue: [ ^nil ].
	aString isString ifFalse: [ ^self lineComment: aString ].
	(aString includes: String cr) ifTrue: [ ^self multilineComment: aString ].
	(aString includes: String lf) ifTrue: [ ^self multilineComment: aString ].
	^self lineComment: aString ! !

! BootSector methodsFor: #header !
headerLines
	" Private - Return an array of lines for header. "

	^(Array	with: self product ,' - ' ,self description
		with: self copyright with: self license with: '')
	,(Array	with: 'Version ' ,self version
		with: self versionDetail with: '')
	,#(
	'For questions and bug reports join us at S8 Facebook group'
	'https://www.facebook.com/groups/s8smalltalk/'
	) ! !

! BootSector methodsFor: #constants !
product
	" Private - Returns this value. "

	^Smalltalk current product! !

! BootSector methodsFor: #constants !
description
	" Private - Returns this value. "

	^'A modern Smalltalk environment running on the "invisible machine".'! !

! BootSector methodsFor: #constants !
copyright
	" Private - Returns this value. "

	^'Copyright (C) '
		,Smalltalk current licenseYear
		,'-today Alejandro F. Reimondo <aleReimondo@smalltalking.net>'! !

! BootSector methodsFor: #constants !
license
	" Private - Returns this value. "

	^'MIT License - See license.txt file'! !

! BootSector methodsFor: #constants !
version
	" Private - Returns the version string (dotted number and date).
	e.g.   ^'0.7 - Feb 2010'
	"

	^self implementedBySubclass: #version! !

! BootSector methodsFor: #constants !
versionDetail
	" Private - Returns this value.
	This method can be refined to give details of status of development.
	e.g.	^'Note that this is a pre-release of XX execution support under social development(20xx-today).'
	"

	^''! !

! BootSector methodsFor: #definitions !
initialDefinition
	" Private - Return the encoding for initializing execution environment.
	The default result is an empty String.
	This method can be refined to prepare global execution context for creation "

	^'' ! !

! BootSector methodsFor: #definitions !
nilDefinition
	" Private - Return the encoding for nil. "

	^(self encode: [
		self defineClass: #UndefinedObject doing: self undefinedObjectInternals.
		self define: #nil asInstanceOf: #UndefinedObject
	]) ,self nilImplementation ! !

! BootSector methodsFor: #definitions !
smalltalkDefinition
	" Private - Return the encoding for smalltalk. "

	^(self encode: [
		self defineClass: #SystemDictionary doing: [ self systemDictionaryInternals ].
		self define: #smalltalk asInstanceOf: #SystemDictionary
	]) ,self smalltalkImplementation ! !

! BootSector methodsFor: #implementations !
undefinedObjectInternals
	" Private - Return the internal encoding for UndefinedObject. "

	^String new "nothing to encode for UndefinedObject" ! !

! BootSector methodsFor: #implementations !
systemDictionaryInternals
	" Private - Return the internal encoding for SystemDictionary. "

	^self implementedBySubclass: #systemDictionaryInternals ! !

! BootSector methodsFor: #implementations !
smalltalkImplementation
	" Private - Return the encoding for smalltalk implementation. "

	^self implementedBySubclass: #smalltalkImplementation ! !

! BootSector methodsFor: #implementations !
smalltalkImplementation
	" Private - Return the encoding for smalltalk implementation. "

	^self implementedBySubclass: #smalltalkImplementation ! !

! BootSector methodsFor: #asm-constants !
selfOP
	" Private - Encoding for self. "

	^#self ! !

! BootSector methodsFor: #encoding !
send: msg withArguments: args
	" Private - Return the encoding for self message sending. "

	^self send: msg to: self selfOP withArguments: args! !

! BootSector methodsFor: #encoding !
send: msg to: aName withArguments: args
	" Private - Return the encoding for sending msg to aName. "

	msg isString ifTrue: [
		^self sendMessage: (Message
			receiver: aName selector: msg
			arguments: args)
	].
	^self sendMessage: (Message
		receiver: aName selector: msg selector
		arguments: args)! !

! BootSector methodsFor: #encoding !
send: msg
	" Private - Return the encoding for self message sending. "

	^self send: msg to: self selfOP withArguments: #()! !

! BootSector methodsFor: #encoding !
send: msg with: arg1
	" Private - Return the encoding for self message sending. "

	^self send: msg to: self selfOP with: arg1! !

! BootSector methodsFor: #encoding !
send: msg with: arg1 with: arg2
	" Private - Return the encoding for self message sending. "

	^self send: msg to: self selfOP withArguments: (Array with: arg1 with: arg2)! !

! BootSector methodsFor: #encoding !
send: msg to: aName with: arg1
	" Private - Return the encoding for sending msg to aName. "

	^self send: msg to: aName withArguments: (Array with: arg1)! !

! BootSector methodsFor: #encoding !
send: msg to: aName
	" Private - Return the encoding for sending msg to aName. "

	^self send: msg to: aName withArguments: #()! !

! BootSector methodsFor: #asm-constants !
sendMessageOP
	" Private - Encoding for send message OP "

	^$. ! !

! BootSector methodsFor: #asm-constants !
expressionPars
	" Private - Encoding for expression (or nil) "

	^'()' ! !

! BootSector methodsFor: #asm-constants !
argsPars
	" Private - Encoding for arguments (or nil) "

	^'()' ! !

! BootSector methodsFor: #asm !
pars: expression
	" Private - Encoding for expression with parentesis. "

	^self between: self argsPars do: expression indent: false ! !

! BootSector methodsFor: #encoding !
sendMessage: aMessage
	" Private - Return the encoding for sending aMessage. "

	| rcv |
	rcv := aMessage receiver.
	(rcv isString or: [rcv isValuable])
	ifTrue: [ self between: (rcv isString ifFalse: [self expressionPars]) do: rcv indent: false ]
	ifFalse: [
		rcv isMessage ifFalse: [ self error: 'Receiver must be aMessage or aString' ].
		self between: self expressionPars do: [ self sendMessage: rcv ] indent: false
	].
	self nextPutAll: self sendMessageOP.
	self nextPutAll: aMessage selector asSelector.
	self pars: [ self sendMessageArguments: aMessage arguments ] ! !

! BootSector methodsFor: #encoding !
sendMessageArguments: args
	" Private - Encoding for aMessage arguments. "

	args size = 0 ifTrue: [ ^self ].
	self nextPutArguments: args ! !

! BootSector methodsFor: #encoding !
call: aFunction
	" Private - Return the encoding for calling aFunction. "

	^self call: aFunction withArguments: #() ! !

! BootSector methodsFor: #encoding !
call: aFunction with: arg1
	" Private - Return the encoding for calling aFunction. "

	^self call: aFunction withArguments: (Array with: arg1) ! !

! BootSector methodsFor: #encoding !
call: aFunction withArguments: array
	" Private - Return the encoding for calling aFunction. "

	aFunction isNil ifTrue: [ self error: 'Missing function to call' ].
	aFunction isString ifTrue: [
		self nextPutAll: aFunction
	] ifFalse: [
		self between: self expressionPars do: aFunction indent: false
	].
	self pars: [ self nextPutArguments: array ] ! !

! BootSector methodsFor: #asm-constants !
blockOPs
	" Private - Encoding for expression block OPs "

	^self implementedBySubclass: #blockOPs ! !

! BootSector methodsFor: #asm-constants !
expressionsOPs
	" Private - Encoding for expression list OPs "

	^self blockOPs ! !

! BootSector methodsFor: #encoding !
implementFunction: aName arguments: args doing: expression comment: aComment
	" Private - Define the function named aName doing expression. "

	self	nextPutAll: #function; space; nextPutAll: aName;
		pars: [ self nextPutArguments: args ];
		between: self expressionsOPs do: expression indent: true comment: aComment! !

! BootSector methodsFor: #encoding !
implementFunction: aName arguments: args doing: expression
	" Private - Define the function named aName doing expression. "

	^self	implementFunction: aName arguments: args
		doing: expression comment: nil ! !

! BootSector methodsFor: #encoding !
implementFunction: aName doing: expression
	" Private - Define the function named aName doing expression. "

	^self implementFunction: aName arguments: #() doing: expression ! !

! BootSector methodsFor: #encoding !
implementFunction: aName
	" Private - Define the function named aName doing expression. "

	^self implementFunction: aName arguments: #() doing: '' ! !

! BootSector methodsFor: #encoding !
defineClass: aName doing: expression
	" Private - Return the encoding for defining a new S8 class named aName. "

	^self define: #Object subclass: aName doing: expression ! !

! BootSector methodsFor: #encoding !
define: parentClass subclass: aName doing: expression
	" Private - Return the encoding for defining a new S8 class named aName. "

	^self implementedBySubclass: #define:subclass:doing: ! !

! BootSector methodsFor: #encoding !
define: aName asInstanceOf: clsName
	" Private - Return the encoding for defining a variable named aName. "

	^self implementedBySubclass: #define:asInstanceOf: ! !

! BootSector methodsFor: #encoding !
between: tags do: aBlock indent: aBoolean comment: aComment
	" Private - Encode aBlock expression between tags. "

	tags size = 0 ifTrue: [
		^self	between: #('' '') do: aBlock
			indent: aBoolean comment: aComment
	].
	self nextPutAll: tags first.
	self comment: aComment.
	(aBlock isString and: [aBlock isEmpty]) ifFalse: [
		aBoolean ifTrue: [ self indent; cr ].
		self nextPutAll: aBlock.
		aBoolean ifTrue: [ self dedent; cr ].
	].
	self nextPutAll: tags last.! !

! BootSector methodsFor: #encoding !
between: tags do: aBlock indent: aBoolean
	" Private - Encode aBlock expression between tags. "

	^self between: tags do: aBlock indent: aBoolean comment: nil ! !

! BootSector methodsFor: #asm-constants !
conditionPars: expression
	" Private - Encoding for condition expression(or nil) "

	^self expressionPars ! !

! BootSector methodsFor: #asm-constants !
opPars: expression
	" Private - Encoding for expression in operation(or nil) "

	^'()' ! !

! BootSector methodsFor: #asm !
expr: expr1 op: op with: expr2 pars: pars
	" Private - Encoding for OP expression. "

	pars isNil ifTrue: [
		^self expr: expr1 op: op with: expr2 pars: (Array
			with: (self opPars: expr1)
			with: (self opPars: expr2))
	].
	pars isEmpty ifTrue: [ ^self expr: expr1 op: op with: expr2 pars: #('') ].
	pars size = 1 ifTrue: [ ^self expr: expr1 op: op with: expr2 pars: (Array with: pars with: pars) ].
	pars size = 2 ifFalse: [ ^self error: 'Invalid paren sizes. Must be 2' ].
	self between: pars first do: [ self nextPutAll: expr1 ] indent: false.
	self space; nextPutAll: op; space.
	self between: pars last do: [ self nextPutAll: expr2 ] indent: false.! !

! BootSector methodsFor: #asm !
expr: expr1 op: op with: expr2
	" Private - Encoding for OP expression. "

	self expr: expr1 op: op with: expr2 pars: nil ! !

! BootSector methodsFor: #asm-constants !
orOP
	" Private - Encoding for operand. "

	^#|| ! !

! BootSector methodsFor: #asm-constants !
andOP
	" Private - Encoding for operand. "

	^#&& ! !

! BootSector methodsFor: #asm-constants !
eqeqOP
	" Private - Encoding for operand. "

	^#== ! !

! BootSector methodsFor: #asm-constants !
noteqOP
	" Private - Encoding for operand. "

	^#!!= ! !

! BootSector methodsFor: #asm-constants !
eqeqeqOP
	" Private - Encoding for operand. "

	^#=== ! !

! BootSector methodsFor: #asm-constants !
noteqeqOP
	" Private - Encoding for operand. "

	^self noteqOP ,#= ! !

! BootSector methodsFor: #asm !
expr: expr1 or: expr2
	" Private - Encoding for or expression. "

	self expr: expr1 op: self orOP with: expr2! !

! BootSector methodsFor: #asm !
expr: expr1 and: expr2
	" Private - Encoding for and expression. "

	self expr: expr1 op: self andOP with: expr2! !

! BootSector methodsFor: #asm !
expr: expr1 eqeq: expr2
	" Private - Encoding for == expression. "

	self expr: expr1 op: self eqeqOP with: expr2! !

! BootSector methodsFor: #asm !
expr: expr1 noteq: expr2
	" Private - Encoding for ~= expression. "

	self expr: expr1 op: self noteqOP with: expr2! !

! BootSector methodsFor: #asm !
expr: expr1 eqeqeq: expr2
	" Private - Encoding for === expression. "

	self expr: expr1 op: self eqeqeqOP with: expr2! !

! BootSector methodsFor: #asm !
expr: expr1 noteqeq: expr2
	" Private - Encoding for ~== expression. "

	self expr: expr1 op: self noteqeqOP with: expr2! !

! BootSector methodsFor: #asm-constants !
notOP
	" Private - Return the OP for negation. "

	^$!! ! !

! BootSector methodsFor: #asm !
not: condition
	" Private - Encoding for negation of condition. "

	self nextPutAll: self notOP.
	self between: (self conditionPars: condition) do: condition indent: false ! !

! BootSector methodsFor: #asm !
assign: assignee to: expression
	" Private - Assignment expression. "

	self	expr: assignee op: #= with: expression
		pars: #(#('' '') #('' ''))! !

! BootSector methodsFor: #asm !
return: expression
	" Private - Encoding for expression return. "

	self nextPutAll: #return; space.
	self nextPutAll: expression! !

! BootSector methodsFor: #asm-constants !
whileDoOPs
	" Private - Return the OPs for while-do expression. "

	^Array with: #while with: #do ! !

! BootSector methodsFor: #asm-constants !
doWhileOPs
	" Private - Return the OPs for do-while expression. "

	^Array with: #do with: #while! !

! BootSector methodsFor: #asm !
while: test do: expression
	" Private - Encoding for while loop. "

	self	nextPutAll: self whileDoOPs first;space;
		between: (self conditionPars: test) do: test indent: false;
		space;nextPutAll: self whileDoOPs last;
		between: self expressionsOPs do: expression indent: true! !

! BootSector methodsFor: #asm !
do: expression while: test
	" Private - Encoding for while loop. "

	self	nextPutAll: self doWhileOPs first; space;
		between: self expressionsOPs do: expression indent: true;
		space; nextPutAll: self doWhileOPs last; space;
		between: (self conditionPars: test) do: test indent: false! !

! BootSector methodsFor: #asm-constants !
repeatOPs
	" Private - Return the OPs for repeat-until expression. "

	^Array with: #repeat with: #until! !

! BootSector methodsFor: #asm !
repeat: expression until: test
	" Private - Encoding for repeat loop. "

	self	between: self repeatOPs do: expression indent: true;
		space; between: (self conditionPars: test) do: test indent: false! !

! BootSector methodsFor: #asm-constants !
forInDoOPs
	" Private - Return the OPs for for-in-do expression. "

	^Array with: #for with: #in with: #do! !

! BootSector methodsFor: #asm-constants !
forInDoPars
	" Private - Return the parentesis for for-in-do expression. "

	^self expressionPars! !

! BootSector methodsFor: #asm !
for: aName in: array do: expression
	" Private - Encoding for for(var...) loop. "

	self	nextPutAll: self forInDoOPs first; space;
		between: self forInDoPars do: [
			self nextPutAll: aName;space.
			self nextPutAll: self forInDoOPs second;space.
			self nextPutAll: array
		] indent: false;
		space;nextPutAll: self forInDoOPs last;
		between: self expressionsOPs do: expression indent: true! !

! BootSector methodsFor: #asm-constants !
forDoOPs
	" Private - Return the OPs for for-do expression. "

	^Array with: #for with: #do! !

! BootSector methodsFor: #asm-constants !
forDoPars
	" Private - Return the parentesis for for-do expression. "

	^self expressionPars! !

! BootSector methodsFor: #asm !
for: counter do: expression
	" Private - Encoding for for(var...) loop. "

	self	nextPutAll: self forDoOPs first; space;
		between: self forDoPars do: [ self nextPutAll: counter ] indent: false;
		space;nextPutAll: self forDoOPs last;
		between: self expressionsOPs do: expression indent: true! !

! BootSector methodsFor: #asm-constants !
arrayOPs
	" Private - Return the OPs for literal array definition. "

	^'[]' ! !

! BootSector methodsFor: #asm !
array: array indent: indented
	" Private - Encoding for array literal. "

	| withCrs |
	withCrs := indented and: [array size > 3].
	^self between: self arrayOPs do: [|first|
		first := true.
		 self nextPutArguments:
			(array collect: [:each| [
				withCrs ifTrue: [
					first ifTrue: [ first := false ] ifFalse: [ self cr ]
				].
				self nextPutAll: each
			]])
	] indent: withCrs ! !

! BootSector methodsFor: #asm !
array: array
	" Private - Encoding for array literal. "

	^self array: array indent: true! !

! BootSector methodsFor: #asm !
emptyArray
	" Private - Encoding for empty array literal. "

	^self array: #() indent: false! !

! BootSector methodsFor: #asm-constants !
fieldOP
	" Private - Return the OP for tuple separator. "

	^self implementedBySubclass: #fieldOP ! !

! BootSector methodsFor: #asm-constants !
objectOPs
	" Private - Return the OPs for literal object definition. "

	^self blockOPs ! !

! BootSector methodsFor: #asm !
object: associations indent: indented
	" Private - Encoding for object literal. "

	| array withCrs |
	(associations isKindOf: Association) ifTrue: [
		^self object: (Array with: associations) indent: indented
	].
	array := associations asArray.
	withCrs := indented and: [array size > 3].
	^self between: self objectOPs do: [|first|
		first := true.
		 self nextPutArguments:
			(array collect: [:assoc| [
				withCrs ifTrue: [
					first ifTrue: [ first := false ] ifFalse: [ self cr ]
				].
				self nextPutAll: assoc key.
				self nextPutAll: self fieldOP; space.
				self nextPutAll: assoc value
			]])
	] indent: withCrs ! !

! BootSector methodsFor: #asm !
object: associations
	" Private - Encoding for object literal. "

	^self object: associations indent: true! !

! BootSector methodsFor: #asm !
emptyObject
	" Private - Encoding for empty object literal. "

	^self object: #() indent: false! !

! BootSector methodsFor: #asm-constants !
ifOP
	" Private - Encoding for if OP (or nil) "

	^'if ' ! !

! BootSector methodsFor: #asm-constants !
elseOP
	" Private - Encoding for else OP "

	^#else ! !

! BootSector methodsFor: #asm !
if: condition
	" Private - Encoding for conditional. "

	self ifOP notNil ifTrue: [ self nextPutAll: self ifOP ].
	self between: (self conditionPars: condition) do: condition indent: false.! !

! BootSector methodsFor: #asm !
if: condition do: trueBlock
	" Private - Encoding for conditional. "

	self if: condition do: trueBlock else: nil! !

! BootSector methodsFor: #asm !
if: condition do: trueBlock else: falseBlock
	" Private - Encoding for conditional. "

	self if: condition; space.
	self between: self blockOPs do: trueBlock indent: true.
	falseBlock notNil ifTrue: [
		self nextPutAll: self elseOP.
		self between: self blockOPs do: falseBlock indent: true.
	]! !

! BootSector methodsFor: #asm !
if: condition return: expression
	" Private - Encoding for conditional expression return. "

	self if: condition; space; return: expression! !

! BootSector methodsFor: #asm !
if: expr1 or: expr2 return: expression
	" Private - Encoding for conditional expression return. "

	self if: [ self expr: expr1 or: expr2 ] return: expression ! !

! BootSector methodsFor: #asm !
if: expr1 or: expr2 or: expr3 return: expression
	" Private - Encoding for conditional expression return. "

	self if: [ self expr: expr1 or: [self expr: expr2 or: expr3]] return: expression ! !

! BootSector methodsFor: #asm !
if: expr1 and: expr2 return: expression
	" Private - Encoding for conditional expression return. "

	self if: [ self expr: expr1 and: expr2 ] return: expression ! !

! BootSector methodsFor: #asm !
if: expr1 or: expr2 do: expression
	" Private - Encoding for conditional expression. "

	self if: [ self expr: expr1 or: expr2 ] do: expression ! !

! BootSector methodsFor: #asm !
if: expr1 or: expr2 or: expr3 do: expression
	" Private - Encoding for conditional expression. "

	self if: [ self expr: expr1 or: [self expr: expr2 or: expr3]] do: expression ! !

! BootSector methodsFor: #asm !
if: expr1 and: expr2 do: expression
	" Private - Encoding for conditional expression. "

	self if: [ self expr: expr1 and: expr2 ] do: expression ! !

! BootSector methodsFor: #asm-constants !
argumentsSeparatorOp
	" Private - Return the OP for separating arguments. "

	^$, ! !

! BootSector methodsFor: #asm !
nextPutArgument: arg
	" Private - Encode argument. "

	(arg isString or: [arg isValuable]) ifTrue: [ ^self nextPutAll: arg ].
	^self error: 'Invalid argument ' ,arg asLiteral ! !

! BootSector methodsFor: #asm !
nextPutArguments: args
	" Private - Encode arguments. "

	args
	    do: [:each | self nextPutArgument: each ]
	    separatedBy: [ self nextPutAll: self argumentsSeparatorOp ]! !

! BootSector methodsFor: #asm !
bind: selector arguments: args doing: implementation comment: aComment
	" Private - Encoding for internal method implementation. "

	^self implementedBySubclass: #bind:arguments:doing: ! !

! BootSector methodsFor: #asm !
bind: selector arguments: args doing: implementation
	" Private - Encoding for internal method implementation. "

	^self	bind: selector arguments: args
		doing: implementation comment: nil ! !

! BootSector methodsFor: #asm !
bind: selector arguments: args private: implementation
	" Private - Encoding for internal method implementation. "

	^self	bind: selector arguments: args
		doing: implementation comment: #Private ! !

! BootSector methodsFor: #asm-constants !
localOP
	" Private - Return the OP for defining new local variable (or nil). "

	^nil ! !

! BootSector methodsFor: #asm !
local: localName with: expression
	" Private - Encoding for new local variable. "

	self localOP notNil ifTrue: [
		self nextPutAll: self localOP; space
	].
	expression isNil
	ifTrue: [ self nextPutAll: localName ]
	ifFalse: [ self assign: localName to: expression ]! !

! BootSector methodsFor: #asm !
local: localName
	" Private - Encoding for new local variable. "

	self local: localName with: nil ! !

! BootSector methodsFor: #asm-constants !
stringOPs
	" Private - Return the OP for raw string. "

	^$" ! !

! BootSector methodsFor: #asm !
string: token
	" Private - Encoding for raw string literal. "

	self between: self stringOPs do: token indent: false! !

! BootSector methodsFor: #asm-constants !
typeofOPs
	" Private - Return the OP for typeOf expression. "

	^#('typeof ' '') ! !

! BootSector methodsFor: #asm-constants !
typeofArg: token
	" Private - Return the OP for typeOf argument. "

	^[self string: token] ! !

! BootSector methodsFor: #asm !
typeOf: expression is: token
	" Private - Encoding for testing typeof expression. "

	self	expr: [self between: self typeofOPs do: expression indent: false]
		eqeq: (self typeofArg: token) ! !

! BootSector methodsFor: #asm !
typeOf: expression isNot: token
	" Private - Encoding for testing typeof expression. "

	self	expr: [self between: self typeofOPs do: expression indent: false]
		noteq: (self typeofArg: token) ! !