" Tests and examples of Network support for the Lua language.
Copyright (C) 2019-today Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
See also http://w3.impa.br/~diego/software/luasocket/home.html
"!

! LuaSocket class methodsFor: #samples !
echoServer
	" Sample - Simple Echo server.
	Binds to an ephemeral port (one that is chosen by the operating system) on the local host and awaits client connections on that port. When a connection is established, reads a line from the remote end and sends it back, closing the connection immediately.
	You can test it using telnet. "

	| server host newLine |
	newLine := '
'.	server := LuaSocket bind: '*' port: 0.
	host := server sockname.
	self note: 'Please telnet to localhost(' ,host first ,') on port ' ,host last.
	self note: 'After connecting you have 10s to enter a line to be echoed'.
	[	| client line |
		client := server accept.
		client settimeout: 10.
		line := client receive first.
		line notNil ifTrue: [
			client send: line ,newLine.
			self print: line.
		].
		client close.
	true] whileTrue! !

! LuaSocket class methodsFor: #samples !
stServer
	" Sample - Simple smalltalk fileIn server (at port 8088). "

	^self stServer: 8087! !

! LuaSocket class methodsFor: #samples !
stServer: port
	" Sample - Simple smalltalk fileIn server.
	Capture all lines sent by clients until a line with a command is found.
	When a command is received it is executed with on all captured lines, and the capture lines are cleared.
	Read to learn about accepted commands
	"

	| server host newLine lines |
	newLine := '
'.	server := LuaSocket bind: '*' port: port.
	host := server sockname.
	self note: 'Please telnet to localhost(' ,host first ,') on port ' ,host last ,' enter ##close to exit'.
	self note: 'After connecting you have 10s to enter a line to be captured'.
	lines := #().
	[	| client line |
		client := server accept.
		client settimeout: 10.
		line := client receive first.
		line notNil ifTrue: [
			(line startsWith: '##') ifFalse: [ lines add: line ] ifTrue: [
				(#('##clear' '##reset' '##cls') includes: line) ifTrue: [ lines := #() ].
				(#('##fileIn' '##showIt') includes: line) ifTrue: [
					client send: (# #tostring: (lines join: newLine) fileIn).
					lines := #()
				].
				line = '##close' ifTrue: [
					client close.
					server close.
					^self note: #done
				].
			].
		].
		client close.
	true] whileTrue! !

! LuaSocket class methodsFor: #samples !
stClient
	" Sample - Simple smalltalk client reading from default input and sending to localhost at port 8088. "

	^self stClient: #localhost! !

! LuaSocket class methodsFor: #samples !
stClient: host
	" Sample - Simple smalltalk client reading from default input and sending to host at port 8088. "

	^self stClient: (Array with: host with: 8087) file: LuaFile input! !

! LuaSocket class methodsFor: #samples !
stClient: array file: aLuaFile
	" Sample - Simple smalltalk client.
	Send all lines in aFile to stServer at hostAndPort in array.
	"

	| client line newLine |
	newLine := '
'.	client := LuaSocket tcp.
	client connect: array first port: array last.
	[ line := aLuaFile read: '*l'.
	  line notNil ] whileTrue: [
	  client send: line ,newLine.
	  client receive second == #closed ifTrue: [ ^self ].
	].
	client close! !

! LuaSocket class methodsFor: #samples !
daytime
	" Sample - A simple daytime client.
	Connects to a remote server and tries to retrieve the daytime, printing the answer it got or an error message. "

	| ip port server |
	 port := 13.
	ip := DNS toip: "#localhost" #raspberrypi.
	server := LuaSocket udp.
	self note: 'Sending UDP datagram to ' ,ip ,':' ,port.
	server sendto: #anything ip: ip port: port onError: [:err|
		^self error: err
	].
	self note: 'Waiting for daytime data'.
	self note: server receive asLiteral.
	self note: #done! !

! FTP class methodsFor: #samples !
getSpeedtest
	" Sample - Get file from public FTP server.
	The Tele2 Speedtest Service helps you test your Internet connection speed through various methods and is available not only to customers of Tele2 but anyone with an Internet connection. Test your connection using speedtest.net's tool, downloading a file via your web browser (HTTP) or downloading and uploading via FTP.	http://speedtest.tele2.net/"

	| url |
	url := 'ftp://speedtest.tele2.net'."/1MB.zip'. "
	self note: 'Reading contents at ' ,url.
	self note: ' ',(FTP get: url) size ,' bytes read'.
	self note: #done! !

! HTTP class methodsFor: #example !
headerTest: url
	" Sample - Get header of file at url. "

	| array headers |
	array := HTTP requestAll: 
		#method -> #Head
		,(#url -> url).
	array first = 1 ifFalse: [ ^self error: 'Invalid result - ' ,array asLiteral ].
	headers := array last.
	headers keys do: [:key|
		self note: key ,'=' ,(headers at: key)
	].
	self note: #done.
	^headers! !

! HTTP class methodsFor: #example !
headerTest
	" Sample - Get header of index file of U8 service. "

	^self headerTest: 'http://u8.smalltalking.net/index.html'! !

! HTTP class methodsFor: #example !
getTest: url
	" Sample - Returns contents of file at url (or error). "

	| data array headers |
	array := HTTP requestAll: 
		#method -> #GET ,(#url -> url)
		,(#sink -> [:chunk| chunk notNil ifTrue: [
			data := data isNil ifTrue: [ chunk ] ifFalse: [ data ,chunk ]
		]. true ]).
	array first = 1 ifFalse: [ ^self error: array asLiteral ].
"	headers := array last.
	headers keys do: [:key|
		self note: key ,'=' ,(headers at: key)
	]."
	self note: #done.
	^data! !

! HTTP class methodsFor: #example !
getTest
	" Sample - Get of index file of U8 service. "

	^self getTest: 'http://u8.smalltalking.net/index.html'! !

! SMTP class methodsFor: #example !
example
	" Sample - Send an email to example users. "

	SMTP send:
		  #from -> '<luasocket@example.com>'
		,(#rcpt -> #(
			'<fulano@example.com>'
			'<beltrano@example.com>'
			'<sicrano@example.com>'
			))
		,(#source -> (Lua smtp message: (
			#body -> 'I hope this works. If it does, I can send you another 1000 copies.'
			,(#headers -> (
				  #to -> 'Fulano da Silva <fulano@example.com>'
				,(#cc -> '"Beltrano F. Nunes" <beltrano@example.com>')
				,(#subject -> 'My first message')
			))
		))).
	self note: #done! !

