" Runtime bindings for S8 Smalltalk.
  Author: Alejandro F. Reimondo [aleReimondo@smalltalking.net]
  Authored @2020 S8 see http://u8.smalltalking.net
"
Object
	subclass: #S8BindingObject
	instanceVariableNames: ' '
	category: #Bindings !
S8BindingObject
	subclass: #S8Binding
	instanceVariableNames: ' key source '
	category: #Bindings !
S8Binding subclass: #S8LiteralBinding !
S8LiteralBinding subclass: #S8CopyBinding !
S8BindingObject
	subclass: #S8Context
	instanceVariableNames: ' parent locals home '
	category: #Bindings !
S8BindingObject
	subclass: #S8Block
	instanceVariableNames: ' node context '
	classVariableNames: ' Optimizations '
	category: #Bindings !

! S8Context class methodsFor: #instantiation !
in: aContext
	" Returns an instance of the receiver. "

	^self basicNew
		initialize: aContext
		with: PoolDictionary new! !

! S8Context class methodsFor: #activation !
in: aContext clear: array do: aBlock
	" Activates a new context in aContext and evaluate aBlock clearing temporaries array. "

	| ctx |
	ctx := self in: aContext.
	array do: [:each| each clear: ctx ].
	^aBlock value: ctx ! !

! S8Context class methodsFor: #instantiation !
for: aReceiver
	" Returns an instance of the receiver. "

	^(self in: nil)
		self: aReceiver;
		yourself! !

! S8Context methodsFor: #initialize !
initialize: aContext with: aPoolDictionary
	" Private - Initialize the receiver. "

	parent := aContext.
	locals := aPoolDictionary.! !

! S8Context methodsFor: #setup !
clear: aBinding
	" Clear the local value for aBinding. "

	^locals at: aBinding key put: nil ! !

! S8Context methodsFor: #accessing !
home
	" Returns the home context of the receiver (or nil). "

	home notNil ifTrue: [ ^home ].
	^parent notNil ifTrue: [ parent  home ] ! !

! S8Context methodsFor: #accessing !
home: anObject
	" Set the home context of the receiver. "

	home := anObject ! !

! S8Context methodsFor: #accessing !
at: aKey
	" Returns the value for aKey in the receiver (or nil). "

	^locals at: aKey ifAbsent: [
		parent notNil ifTrue: [
			parent at: aKey
		]
	]! !

! S8Context methodsFor: #private !
set: aKey to: aValue
	" Private - Returns true if the receiver change the value for aKey to aValue. "

	(locals includesKey: aKey) ifTrue: [
		locals at: aKey put: aValue.
		^true
	].
	^parent notNil and: [ parent set: aKey to: aValue ] ! !

! S8Context methodsFor: #accessing !
at: aKey put: aValue
	" Set the value for aKey in the receiver to aValue. "

	(locals includesKey: aKey) ifFalse: [
		(parent notNil and: [ parent set: aKey to: aValue ])
		ifTrue: [ ^aValue ].
	].
	^locals at: aKey put: aValue ! !

! S8Context methodsFor: #accessing !
self: anObject
	" Set the object bound to self in the receiver. "

	^self at: #self put: anObject ! !

! S8Context methodsFor: #accessing !
self
	" Returns the object bound to self in the receiver. "

	^self at: #self ! !

! S8Context methodsFor: #accessing !
associationsDo: aBlock
	" Evaluates aBlock on each association in the receiver. "

	parent notNil ifTrue: [
		parent associationsDo: aBlock
	].
	locals notNil ifTrue: [
		locals associationsDo: aBlock
	].! !

! S8Binding class methodsFor: #instantiating !
key: aKey
	" Return an instance of the receiver. "

	^self basicNew initialize: aKey source: nil ! !

! S8Binding class methodsFor: #instantiating !
key: aKey in: aDictionary
	" Return an instance of the receiver. "

	^self basicNew initialize: aKey source: aDictionary ! !

! S8Binding class methodsFor: #instantiating !
key: aKey evaluate: oneArgumentBlock assign: threeArgumentBlock
	" Return an instance of the receiver. "

	| aBlock pool |
	pool := PoolDictionary new.
	pool	at: #key put: [ aKey ];
		at: #evaluateIn: put: oneArgumentBlock;
		at: #assign:to:in: put: threeArgumentBlock;
		yourself.
	aBlock := [:what :arg1 :arg2 :arg3 |
		(pool at: what) evaluateWithArguments: (Array with: arg1 with: arg2 with: arg3)
	].
	^self basicNew initialize: aBlock source: nil ! !

! S8Binding class methodsFor: #instantiating !
toInstanceVariable: key
	" Return an instance of the receiver. "

	| getter |
	getter := [:ctx| ctx self instVarAt: key ].
	^self	key: key evaluate: getter
		assign: [:ndx :value :ctx|
			ndx isNil
			ifTrue: [ ctx self instVarAt: key put: value ]
			ifFalse: [ | target |
				target := ndx value: (getter value: ctx).
				target first basicAt: target last put: value
			]
		] ! !

! S8Binding methodsFor: #initialize !
initialize: aKeyOrBlock source: aSource
	" Private - Initialize the receiver. "

	key := aKeyOrBlock.
	source := aSource. ! !

! S8Binding methodsFor: #accessing !
key
	" Return the key of the receiver. "

	key isString ifFalse: [ ^key evaluateWith: #key ].
	^key ! !

! S8Binding methodsFor: #accessing !
source
	" Private - Return the source of the receiver. "

	^source ! !

! S8Binding methodsFor: #scope !
local
	" Returns the receiver acting on runtime context scope. "

	source := nil ! !

! S8Binding methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	key isString ifFalse: [
		^key evaluateWithArguments: (Array with: #evaluateIn: with: aContext)
	].
	source notNil ifTrue: [ ^source at: key ifAbsent: [] ].
	^aContext at: key ! !

! S8Binding methodsFor: #evaluating !
assign: indexBlock to: aValue in: aContext
	" Set the value for the receiver in aContext. "

	| target |
	key isString ifFalse: [
		^key evaluateWithArguments: (Array
			with: #assign:to:in:
			with: indexBlock with: aValue
			with: aContext)
	].
	target := source isNil ifTrue: [aContext] ifFalse: [source].
	indexBlock isNil ifTrue: [ ^target at: key put: aValue ].
	target := indexBlock value: (target at: key).
	^target first basicAt: target last put: aValue ! !

! S8LiteralBinding class methodsFor: #instantiating !
value: aValue
	" Return an instance of the receiver. "

	^self basicNew initialize: #constant source: aValue ! !

! S8LiteralBinding methodsFor: #evaluating !
assign: indexBlock to: aValue in: aContext
	" Set the value for the receiver in aContext. "

	^self invalidMessage: #assign:to:in: ! !

! S8LiteralBinding methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^source ! !

! S8CopyBinding methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^source shallowCopy ! !

! S8Block class methodsFor: #actions !
defaultOptimizations
	" Private - Returns the default pool of optimized actions. "

	^PoolDictionary new
		at: #class put: [:receiver :msg :values :context| receiver class ];
		at: #isNil put: [:receiver :msg :values :context| receiver isNil ];
		at: #notNil put: [:receiver :msg :values :context| receiver notNil ];
		at: #== put: [:receiver :msg :values :context| receiver == values first ];
		at: #=== put: [:receiver :msg :values :context| receiver === values first ];
		at: #basicAt: put: [:receiver :msg :values :context| receiver basicAt: values first ];
		at: #basicAt:ifAbsent: put: [:receiver :msg :values :context|
			receiver basicAt: values first ifAbsent: [ values second value ] ];
		at: #basicAt:put: put: [:receiver :msg :values :context|
			receiver basicAt: values first put: values second ];
		at: #orNil put: [:receiver :msg :values :context| receiver orNil ];
		at: #respondsTo: put: [:receiver :msg :values :context|
			receiver respondsTo: values first ];
		yourself ! !

! S8Block class methodsFor: #actions !
optimizations
	" Private - Returns the pool of optimized actions. "

	Optimizations isNil ifTrue: [ Optimizations := self defaultOptimizations ].
	^Optimizations! !

! S8Block class methodsFor: #instantiation !
for: aNode in: aContext
	" Returns an instance of the receiver. "

	^self new
		node: aNode;
		context: aContext;
		yourself ! !

! S8Block methodsFor: #accessing !
node: aNode
	" Sets the receiver's parent node. "

	node := aNode ! !

! S8Block methodsFor: #evaluating !
statements
	" Returns the receiver's statements. "

	^node statements ! !

! S8Block methodsFor: #accessing !
args
	" Returns the receiver's arguments. "

	^node arguments ! !

! S8Block methodsFor: #accessing !
arguments
	" Returns the receiver's arguments. "

	^node arguments ! !

! S8Block methodsFor: #accessing !
temporaries
	" Returns the receiver's temporaries. "

	^node temporaries ! !

! S8Block methodsFor: #accessing !
argumentCount
	" Returns the number of arguments of the receiver. "

	^self arguments size ! !

! S8Block methodsFor: #accessing !
context: aContext
	" Sets the receiver's context. "

	context := aContext ! !

! S8Block methodsFor: #binding !
bindTo: anArray in: aContext
	" Private - Bind the arguments to values in anArray. "

	| args stream |
	args := self arguments.
	args size = 0 ifTrue: [ ^self ].
	args size = anArray size ifFalse: [
"		^self error: 'Invalid number of arguments.'	
DEBUG-		self note: 'Invalid number of arguments (' ,anArray size ,') must be ' ,args size
"	].

	stream := anArray stream.
	args do: [:each| each assignValue: (stream atEnd ifFalse: [ stream next ]) in: aContext ].
	self temporaries do: [:each| each assignValue: nil in: aContext ]. ! !

! S8Block methodsFor: #converting !
asActionSequence
	" Return an ActionSequence from the receiver. "

	^ActionSequence with: self ! !

! S8Block methodsFor: #converting !
valueBlock
	" Private - Return the receiver as a native block to compute value. "

	^[ self value ] ! !

! S8Block methodsFor: #evaluating !
ensure: aBlock
	" Return the result of evaluating the receiver.
	Execute aBlock after the receiver is executed.
	"

	^self valueBlock ensure: aBlock ! !

! S8Block methodsFor: #evaluating !
evaluate
	" Return the result of evaluating the receiver. "

	^self value ! !

! S8Block methodsFor: #evaluating !
evaluateFor: anObject
	" Return the result of evaluating the receiver with anObject. "

	^self evaluateWithArguments: (Array with: anObject) ! !

! S8Block methodsFor: #evaluating !
evaluateWithArguments: anArray
	" Return the result of evaluating the receiver. "

	^self argumentCount = 0
		ifTrue: [ self value ]
		ifFalse: [ self valueWithArguments: anArray ] ! !

! S8Block methodsFor: #evaluating !
evaluateWithoutInterrupts
	" OBSOLETE - Disable interrupts, evaluate the receiver, enable interrupts, and answer the result of the evaluation. "

	^self valueBlock evaluateWithoutInterrupts ! !

! S8Block methodsFor: #evaluating !
ifCurtailed: exceptionBlock
	" Return the result of evaluating the receiver.
	If evaluation is limited by exiting before completion, the exception block is evaluated.
	"

	^self valueBlock ifCurtailed: exceptionBlock ! !

! S8Block methodsFor: #evaluating !
on: anExceptionType do: aBlock
	" Return the result of evaluating the receiver catching exceptions. "

	^self valueBlock on: anExceptionType do: aBlock ! !

! S8Block methodsFor: #evaluating !
repeat
	" OBSOLETE - Evaluate the receiver repeatedly, ending only if the block forces some stopping condition. "

	^self valueBlock repeat ! !

! S8Block methodsFor: #evaluating !
value
	" Return the result of evaluating the receiver. "

	^self valueWithArguments: #() ! !

! S8Block methodsFor: #evaluating !
value: argument
	" Return the result of evaluating the receiver with anObject. "

	^self valueWithArguments: (Array with: argument) ! !

! S8Block methodsFor: #evaluating !
value: anObject value: anotherObject
	" Return the result of evaluating the receiver with arguments. "

	^self valueWithArguments: (Array with: anObject with: anotherObject) ! !

! S8Block methodsFor: #evaluating !
value: anObject value: anotherObject value: thirdObject
	" Return the result of evaluating the receiver with arguments. "

	^self valueWithArguments: (Array with: anObject with: anotherObject with: thirdObject) ! !

! S8Block methodsFor: #evaluating !
value: firstArgument value: secondArgument value: thirdArgument value: fourthArgument
	" Return the result of evaluating the receiver with arguments. "

	^self valueWithArguments: (Array
		with: firstArgument with: secondArgument
		with: thirdArgument with: fourthArgument) ! !

! S8Block methodsFor: #evaluating !
valueWithArguments: anArray
	" Return the result of evaluating the receiver with arguments. "

	^S8Context
		in: context clear: self arguments ,self temporaries
		do: [:ctx|
			self bindTo: anArray in: ctx.
			node statements inject: nil
				into: [:total :each| each evaluateIn: ctx ]
	] ! !

! S8Block methodsFor: #evaluating !
valueDeferred
	" Evaluate the receiver deferred in time.
	Return the receiver.
	"

	self valueDeferred: 0! !

! S8Block methodsFor: #evaluating !
valueDeferred: milliseconds
	" Evaluate the receiver deferred in time.
	Return the receiver.
	"

	self valueWithTimeout: milliseconds! !

! S8Block methodsFor: #timeout !
valueWithTimeout: aNumber
	" Evaluate the receiver with time out. "

	| fn | fn := self callableOf: self.
	^{''js'setTimeout(fn,aNumber)'
	lua'self:notImplementedYet_("valueWithTimeout:")'
	py'self.notImplementedYet_(smalltalk.String("valueWithTimeout:"))'}! !

! S8Block methodsFor: #evaluating !
whileFalse
	" Evaluate the receiver until the result is true. "

	^self whileFalse: [] ! !

! S8Block methodsFor: #evaluating !
whileFalse: aBlock
	" Evaluate aBlock while the result of evaluating the receiver is false. "

	self valueBlock whileFalse: [ aBlock value ].
	^nil ! !

! S8Block methodsFor: #evaluating !
whileTrue
	" Evaluate the receiver until the result is false. "

	^self whileTrue: [] ! !

! S8Block methodsFor: #evaluating !
whileTrue: aBlock
	" Evaluate aBlock while the result of evaluating the receiver is true. "

	self valueBlock whileTrue: [ aBlock value ].
	^nil ! !



! KeyedCollection methodsFor: #bindings !
bindingAt: aKey
	" Private - Return the binding at aKey, or generates an error if key is not present. "

	^self bindingAt: aKey ifAbsent: [ self error: 'Missing key ' ,aKey asLiteral ]! !

! KeyedCollection methodsFor: #bindings !
bindingAt: aKey ifAbsent: aBlock
	" Private - Return the binding at aKey, or the result of evaluating aBlock. "

	(self includesKey: aKey) ifFalse: [ ^aBlock value ].
	^S8Binding key: aKey in: self! !

! Smalltalk methodsFor: #bindings !
bindingAt: aKey ifAbsent: aBlock
	" Private - Return the binding at aKey, or the result of evaluating aBlock. "

	(self includesKey: aKey) ifFalse: [ ^aBlock value ].
	^S8Binding key: aKey in: self! !

! Smalltalk methodsFor: #bindings !
bindingAt: aKey
	" Private - Return the binding at aKey, or generates an error if key is not present. "

	^self bindingAt: aKey ifAbsent: [ self error: 'Missing key ' ,aKey asLiteral ]! !