" Scripting support for S8 Smalltalk using S8 Compiler framework.
  Author: Alejandro F. Reimondo [aleReimondo@smalltalking.net]
  Authored @2020 S8 see http://u8.smalltalking.net

Simple test expressions:
(Compiler new parse: '#hello') evaluate = #hello
(Compiler new parse: '|tres| tres := 3. #hello. tres') evaluate = 3
(Compiler new parse: '| i | i := 12. #(1 2 3 4) do: [:i| self print: i * i]. i') evaluate = 12
"
(Smalltalk includesKey: #S8Context) ifFalse: [
	^self error: 'Missing Bindings framework'
]!

! Compiler methodsFor: #parsing !
parse: smalltalkExpression
	" Return the result of parsing a method to evaluate smalltalkExpression. "

	^self parseMethod: String doItSelector
		,' ^[', smalltalkExpression
		,'] value' ! !

! Compiler methodsFor: #parsing !
parse: smalltalkExpression onError: aBlock
	" Return the result of parsing a method to evaluate smalltalkExpression. "

	| syntax |
	[ syntax := self parse: smalltalkExpression.
	  syntax isParseFailure ifTrue: [ self error: syntax toString ].
	] on: Error do: [:ex| ^aBlock evaluateWith: ex ].
	^syntax ! !

! CompilerNode methodsFor: #evaluating !
evaluate
	" Evaluate the receiver in an empty context. "

	^self evaluate: nil ! !

! CompilerNode methodsFor: #evaluating !
evaluate: aReceiver
	" Returns the result of evaluating the receiver in a context bound to aReceiver. "

	^self evaluateIn: (S8Context for: aReceiver) ! !

! CompilerNode methodsFor: #evaluating !
evaluateContentsIn: aContext
	" Private - Returns the result of evaluating the nodes of receiver in aContext. "

	| result |
	self nodes do: [:each| result := each evaluateIn: aContext ].
	^result ! !

! CompilerNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext.
	This message must be implemented by the subclasses. "

	^self subclassResponsibility: #evaluateIn: ! !

! CompilerNode methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver.
	This block will be used to abort execution of the receiver.
	Propagate to nodes by default.
	This method can be refined by subclasses. "

	nodes notNil ifTrue: [
		nodes do: [:each| each exitBlock: aBlock ]
	] ! !

! CompilerNode methodsFor: #query !
isOO
	" Returns true if the receiver refers to an object. "

	^true ! !

! AssignmentNode methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver. "

	super exitBlock: aBlock.
	self left exitBlock: aBlock.
	self right exitBlock: aBlock ! !

! AssignmentNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^self left bind: (self right evaluateIn: aContext) in: aContext ! !

! BlockNode methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver. "

	super exitBlock: aBlock.
	self parameters do: [:each| each exitBlock: aBlock ] ! !

! BlockNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^S8Block for: self in: aContext ! !

! BlockNode methodsFor: #accessing !
arguments
	" Return the arguments of the receiver. "

	^self parameters ! !

! BlockNode methodsFor: #accessing !
temporaries
	" Return the temporaries of the receiver. "

	^#() ! !

! BlockNode methodsFor: #accessing !
statements
	" Return the statements of the receiver. "

	^self nodes ! !

! CascadeNode methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver. "

	super exitBlock: aBlock.
	receiver notNil ifTrue: [ receiver exitBlock: aBlock ]. ! !

! MethodNode methodsFor: #evaluating !
clearExitBlock
	" Clear the exit block. "

	self exitBlock: nil ! !

! MethodNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	self exitBlock: [:result| self clearExitBlock. ^result ].
	^nodes size > 0
		ifTrue: [ self evaluateContentsIn: aContext ]
		ifFalse: [ aContext self ] ! !

! ReturnNode methodsFor: #accessing !
exitBlock: aBlock
	" Set the receiver's exitBlock. "

	super exitBlock: aBlock.
	exitBlock := aBlock. ! !

! ReturnNode methodsFor: #evaluating !
exitWith: anObject
	" Exit the receiver with anObject as returned value. "

	exitBlock isNil ifTrue: [ ^self error: 'Invalid context.' ].
	^exitBlock value: anObject ! !

! ReturnNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^self exitWith: (self evaluateContentsIn: aContext) ! !


! SendNode class methodsFor: #actions !
actionToSend: selector ifAbsent: aBlock
	" Private - Returns the block to evaluate for sending message selector. "

	^S8Block optimizations at: selector ifAbsent: aBlock ! !

! SendNode methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver. "

	super exitBlock: aBlock.
	receiver notNil ifTrue: [ receiver exitBlock: aBlock ].
	self arguments do: [:each| each exitBlock: aBlock ]. ! !

! SendNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	| value |
	value := self receiver evaluateIn: aContext.
	^self evaluate: self receiver value: value in: aContext ! !

! SendNode methodsFor: #evaluating !
evaluate: aReceiver value: anObject in: aContext
	" Private - Returns the result of sending message to anObject in aContext. "

	| values |
	values := self arguments collect: [:each| each evaluateIn: aContext ].
	(aReceiver isSuper or: [ aReceiver isOO not ]) ifTrue: [
		^aReceiver send: self selector to: anObject withArguments: values in: aContext
	].
	^self send: self selector to: anObject withArguments: values in: aContext ! !

! SendNode class methodsFor: #private !
selectorToSend: msg
	" Private - Returns the selector to use to lookup method for sending message msg. "

	^msg first = $# ifTrue: [
		(msg copyFrom: 2 to: msg size) upTo: $:
	] ifFalse: [ msg ]! !

! SendNode methodsFor: #evaluating !
send: msg to: anObject withArguments: values in: aContext
	" Private - Returns the result of sending message to anObject in aContext. "

	| sel |
	sel := self class selectorToSend: msg.
	action isNil ifTrue: [ action := self actionToSend: sel ].
	^action	value: anObject value: sel
		value: values value: aContext ! !

! SendNode methodsFor: #actions !
actionToSend: aSymbol
	" Private - Returns the block to evaluate for sending message aSymbol. "

	^self class actionToSend: aSymbol ifAbsent: [
		[:receiver :msg :values :context|
			receiver perform: msg withArguments: values
		]
	] ! !

! CascadeNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	| rec result |
	rec := self receiver evaluateIn: aContext.
	nodes do: [:each| result := each evaluate: self receiver value: rec in: aContext ].
	^result ! !

! InlineNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	| src lang |
	lang := aContext nativeLanguage.
	lang isNil ifTrue:  [ lang := Smalltalk nativeLanguage ].
	src := self sourceFor: lang.
	^Smalltalk
		nativeObjectAt: src
		ifAbsent: [ self error: 'Invalid/missing native code - ' ,src ] ! !

! SequenceNode methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver. "

	super exitBlock: aBlock.
	self temps do: [:each| each exitBlock: aBlock ]. ! !

! SequenceNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^S8Context
		in: aContext clear: self temps
		do: [:ctx| self evaluateContentsIn: ctx ] ! !

! ExtendedSequenceNode methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver. "

	super exitBlock: aBlock.
	self directives do: [:each| each exitBlock: aBlock ] ! !

! ExtendedSequenceNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	self directives do: [:each| each evaluateIn: aContext ].
	^super evaluateIn: aContext ! !

! ValueNode class methodsFor: #instantiation !
literal: aName value: aValue
	" Return an instance of the receiver. "

	^self new literal: aName value: aValue! !

! ValueNode methodsFor: #initialize !
literal: aName value: aValue
	" Private - Initialize the receiver as a literal constant with aValue. "

	value := aName.
	binding := S8LiteralBinding value: aValue ! !

! ValueNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^self index: (self valueIn: aContext) in: aContext ! !

! ValueNode methodsFor: #indexing !
index: anObject in: aContext
	" Private - Returns the result of indexing anObject. "

	^anObject ! !

! ValueNode methodsFor: #initialize !
initializeBinding: aContext
	" Private - Initialize the binding of the receiver. "

	binding := #constant ! !

! ValueNode methodsFor: #bindings !
enter: aContext
	" Private - Returns true if the receiver is bound to aContext. "

	binding isNil ifTrue: [ self initializeBinding: aContext ].
	binding = #constant ifTrue: [ ^false ].
	^binding notNil ! !

! ValueNode methodsFor: #value !
valueIn: aContext
	" Private - Returns the value of the receiver in aContext. "

	(self enter: aContext) ifFalse: [ ^value shallowCopy ].
	^binding evaluateIn: aContext ! !

! ValueNode methodsFor: #evaluating !
clear: aContext
	" Clear the receiver's contents. "

	(self enter: aContext) ifFalse: [
		^aContext clear: value -> nil
	].
	^aContext clear: binding ! !

! ValueNode methodsFor: #binding !
bind: aValue in: aContext
	" Set the receiver's value to aValue in aContext.
	WARNING: The receiver MUST return the value set. "

	^self assignValue: aValue in: aContext ! !

! ValueNode methodsFor: #evaluating !
followBlockIn: aContext
	" Private - Returns the block to follow indices. "

	^nil ! !

! ValueNode methodsFor: #evaluating !
assignValue: aValue in: aContext
	" Private - Set the receiver's value to aValue.
	WARNING: The receiver MUST return the value set. "

	(self enter: aContext) ifFalse: [
		^aContext at: value put: aValue
	].
	^binding
		assign: (self followBlockIn: aContext)
		to: aValue in: aContext ! !

! VariableNode methodsFor: #indexing !
index: anObject in: aContext
	" Private - Returns the result of indexing anObject. "

	indices isNil ifTrue: [ ^anObject ].
	^indices inject: anObject into: [:result :index|
		result	basicAt: (self key: index in: aContext)
			ifAbsent: [ ]
	]! !

! VariableNode methodsFor: #indexing !
key: index in: aContext
	" Private - Returns the key to index an object in aContext. "

	^(index isKindOf: CompilerNode)
		ifTrue: [ index evaluateIn: aContext ]
		ifFalse: [ index ]! !

! VariableNode methodsFor: #initialize !
initializeBinding: aContext
	" Private - Initialize the binding of the receiver. "

	(#( #self #super ) includes: value) ifTrue: [
		^binding := S8Binding key: #self
	].
	value = #nil ifTrue: [
		value := nil.
		^binding := #constant
	].
	(#( #true #false ) includes: value) ifTrue: [
		value := value = #true.
		^binding := #constant
	].
	(aContext self class allInstVarNames includes: self value) ifTrue: [
		^binding := S8Binding toInstanceVariable: self value
	].

	binding := S8Binding key: value ! !

! VariableNode methodsFor: #evaluating !
clear: aContext
	" Clear the receiver's contents. "

	self enter: aContext.
	binding = #constant ifTrue: [
		^self error: 'Should not be constant'
	].
	^super clear: aContext ! !

! VariableNode methodsFor: #evaluating !
followBlockIn: aContext
	" Private - Returns the block to follow indices. "

	^indices notNil ifTrue: [ [:root| self follow: root in: aContext ] ] ! !

! VariableNode methodsFor: #indexing !
follow: anObject in: aContext
	" Private - Returns the result of following the indices from anObject and last index to evaluate (or nil). "

	| target index |
	indices isNil ifTrue: [ ^nil ].
	target := anObject.
	1 to: indices size - 1 do: [:i|
		index := self key: (indices at: i) in: aContext.
		target := target basicAt: index ifAbsent: []
	].
	index := self key: indices last in: aContext.
	^Array with: target with: index ! !

! VariableNode methodsFor: #evaluating !
send: selector to: anObject withArguments: anArray in: aContext
	" Private - Returns the result of sending message to anObject in aContext. "

	| anImplementor |
	anImplementor := aContext self class.
	anImplementor isNil ifTrue: [ anImplementor := anObject class ].
	^self	apply: (self functionToCall: selector in: anImplementor)
		to: anObject withArguments: anArray
		in: aContext ! !

! VariableNode methodsFor: #evaluating !
functionToCall: selector in: anImplementor
	" Private - Returns the native function to call for selector. "

	^Smalltalk current
		#superImplementor: anImplementor
		of: selector ! !

! VariableNode methodsFor: #evaluating !
apply: fn to: anObject withArguments: anArray in: aContext
	" Private - Returns the result of evaluating the native function in anObject @ aContext. "

	^{''
	js'fn.apply(anObject,anArray)'
	lua'fn(anObject,(table.unpack or unpack)(anArray))'
	py'fn(anObject,*smalltalk.objects_toCall_(anArray,fn))'}. ! !

! AlienNode methodsFor: #initialize !
initializeBinding: aContext
	" Private - Initialize the binding of the receiver. "

	binding := S8Binding key: [ Smalltalk nativeObjectAt: value ] ! !

! GlobalReferenceNode methodsFor: #visiting !
bindingIn: aContext
	" Private - Returns the binding of the receiver. "

	| st cls aName |
	aName := self value.
	cls := aContext self class.
	cls := cls notNil
		ifTrue: [ cls instanceClass ]
		ifFalse: [ nil class ].
	st := aContext self smalltalk.
	cls withAllSuperclassesDo: [:each| | vars |
		vars := each classVariables.
		(vars notNil and: [ vars includesKey: aName ]) ifTrue: [
			^S8Binding key: aName in: vars
		]
	].
	cls sharedPoolsAndNamesDo: [:poolName :pool|
		(pool includesKey: aName) ifTrue: [
			^S8Binding
				key: aName
				in: (st at: poolName)
		]
	].
	^S8Binding key: aName in: st ! !

! GlobalReferenceNode methodsFor: #visiting !
initializeBinding: aContext
	" Private - Initialize the binding of the receiver. "

	binding := self bindingIn: aContext ! !

! NoReceiverNode methodsFor: #query !
isOO
	" Returns true if the receiver refers to an object. "

	^false ! !

! NoReceiverNode methodsFor: #evaluating !
send: msg to: anObject withArguments: anArray in: aContext
	" Private - Returns the result of sending message to anObject in aContext. "

	| fn selector |
	selector := SendNode selectorToSend: msg.
	selector = #self ifTrue: [
		self note: self class name,'>>#send:to:... need testing'.
		^aContext self evaluateWithArguments: anArray
	].
	fn := Smalltalk nativeObjectAt: selector ifAbsent: [
		^self error: 'Missing native function ' ,selector asLiteral
	].
	^{''
	js'fn.apply(anObject,anArray)'
	lua'fn(anObject,(table.unpack or unpack)(anArray))'
	py'fn(anObject,*smalltalk.objects_toCall_(anArray,fn))'}. ! !

! JSDirectiveNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^self error: 'Directive support missing' ! !