" (Compiler based) Step by step execution support for S8 Smalltalk.
  Author: Alejandro F. Reimondo [aleReimondo@smalltalking.net]
  Authored @2020 S8 see http://u8.smalltalking.net
"

MethodNode
	subclass: #StepOverNode
	instanceVariableNames: ' action '
	category: #S8Steps !

! Compiler class methodsFor: #S8Steps-private !
evaluationScript
	" Return the support for skipping method execution. "

	^StepOverNode ! !

! Compiler methodsFor: #S8Steps-private !
isSourcecode: string
	" Private - Returns true if a string is formated as sourcecode. "

	| input |
	input := string stream.
	[input atEnd] whileFalse: [
		((input upTo: String crlf) includes: String lf) ifTrue: [ ^false ]
	].
	^true! !

! Compiler methodsFor: #S8Steps-private !
sourcecode: string
	" Private - Returns a string formated as sourcecode. "

	| input output |
	(self isSourcecode: string) ifTrue: [ ^string ].
	input := string stream.
	output := '' writeStream.
	[input atEnd] whileFalse: [
		output nextPutAll: input nextLine.
		input atEnd ifFalse: [ output cr ]
	].
	^output contents! !

! Compiler methodsFor: #S8Steps-scripting !
scriptForMethod: source onError: aBlock
	" Return the compiled script for method with source code or the result of evaluating aBlock with error. "

	| node |
	^[	node := self parseMethod: (self sourcecode: source).
		node isParseFailure ifTrue: [ self error: node toString ].
		node
	] on: Error do: [:ex| aBlock value: ex ] ! !

! Compiler methodsFor: #S8Steps-scripting !
scriptForMethod: source
	" Return the compiled script for method with source code (or nil). "

	^self scriptForMethod: source onError: [ nil ] ! !

! CompilerNode methodsFor: #S8Steps-testing !
isMessageNode
	" Returns true if the receiver is a message send. "

	^false ! !

! CompilerNode methodsFor: #S8Steps-testing !
isValid
	" Returns true if the receiver is valid (for simulation). "

	self isEmpty ifFalse: [
		nodes do: [:each| each isValid ifFalse: [ ^false ] ]
	].
	^true ! !

! CompilerNode methodsFor: #S8Steps-simulating !
emptyActions: actions
	" The actions taken to evaluate the receiver are empty. "

	^self note: 'WARNING: Empty actions for ' ,self ! !

! CompilerNode methodsFor: #S8Steps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext.
	The default implementation returns the actions to evaluate the contents of receiver.
	This method can be refined by the subclasses.
	"

	| result |
	result := self nodes inject: #() into: [:total :each|
		total addAll: (each actionsIn: aContext); yourself
	].
	result isEmpty ifTrue: [ self emptyActions: result ].
	^result! !

! AssignmentNode methodsFor: #S8Steps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	^Array
		with: [:simulator| simulator add: self right in: aContext ]
		with: [:simulator| self left bind: simulator value in: aContext ] ! !

! AssignmentNode methodsFor: #S8Steps-test !
isValid
	" Returns true if the receiver is valid (for simulation). "

	^self right isValid and: [ self left isValid ] ! !

! DirectiveNode methodsFor: #S8Steps-test !
isValid
	" Returns true if the receiver is valid (for simulation). "

	^false ! !

! InlineNode methodsFor: #S8Steps-test !
isValid
	" Returns true if the receiver is valid (for simulation).
	Actually we do NOT simulate methods with inlined code if not encoded for #s8.
	"

	^nodes notNil ! !

! MethodNode methodsFor: #S8Steps-converting !
script
	" Returns the receiver capable to be evaluated as a script. "

	^self isValid ifTrue: [ self ]! !

! MethodNode methodsFor: #S8Steps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	| result |
	result := super actionsIn: aContext.
	result addFirst: [:simulator|
		self exitBlock: [:value| simulator exit: value ].
	].
	^result ! !

! MethodNode methodsFor: #S8Steps-native !
argsIn: mth
	" Private - Return the arguments bound to mth. "

	^mth arguments ! !

! MethodNode methodsFor: #S8Steps-evaluating !
contextFor: aReceiver binding: args to: array
	" Private - Returns the context to evaluate the receiver on aReceiver with parameters. "

	| result count |
	result := S8Context for: aReceiver.
	count := args size min: array size.
	1 to: count do: [:i| | key |
		key := (args at: i) value.
		result at: key put: (array at: i)
	].
	^result ! !

! MethodNode methodsFor: #S8Steps-simulating !
stepsFor: aReceiver entering: aMethod withArguments: array
	" Returns the steps to evaluate execution of the receiver activating aMethod. "

	| ctx |
	ctx := self contextFor: aReceiver binding: (self argsIn: aMethod) to: array.
	^(self actionsIn: ctx) collect: [:each|
		SimulationStep for: self doing: each in: ctx
	]! !

! StepOverNode class methodsFor: #instantiation !
doing: aBlock
	" Returns an instance of the receiver. "

	^self new initialize: aBlock ! !

! StepOverNode methodsFor: #initialize !
initialize: anAction
	" Private - Initialize the receiver. "

	action := anAction ! !

! StepOverNode methodsFor: #simulating !
stepsFor: aReceiver entering: aMethod withArguments: array
	" Returns the steps to evaluate execution of the receiver activating aMethod. "

	self sourceRange: (0 to: aMethod source size - 1).
	^super stepsFor: aReceiver entering: aMethod withArguments: array ! !

! StepOverNode methodsFor: #evaluating !
contextFor: aReceiver binding: args to: array
	" Returns the context to evaluate the receiver on aReceiver with parameters. "

	| result count |
	result := S8Context for: aReceiver.
	count := args size min: array size.
	1 to: count do: [:i| result at: (args at: i) put: (array at: i) ].
	^result ! !

! StepOverNode methodsFor: #native !
argsIn: aMethod
	" Private - Return the arguments bound to aMethod. "

	| i | i := 0.
	^(aMethod selector asArrayOfSubstringsSeparatedBy: $:) collect: [:each|
		i := i + 1. $( ,i ,$) ,each
	] ! !

! StepOverNode methodsFor: #evaluating !
evaluateContentsIn: localContext
	" Private - Returns the result of evaluating the contents of the receiver in local context. "

	self isEmpty ifTrue: [ ^action evaluateWith: nil ].
	^self invalidMessage: #evaluateContentsIn: ! !

! StepOverNode methodsFor: #internal !
inject: simulator in: aContext
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	self isEmpty ifTrue: [
		^simulator add: self
			step: [:simulator| simulator value: (action evaluateWith: simulator) ]
			in: aContext
	].
	^self invalidMessage: #inject:in: ! !

! StepOverNode methodsFor: #S8Steps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	^Array with: [:simulator| simulator value: (action value: simulator) ] ! !

! ReturnNode methodsFor: #S8Steps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	^(super actionsIn: aContext)
		add: [:simulator|
			simulator return: self in: aContext then: [
				exitBlock value: simulator value
			]
		];
		yourself ! !

! ScopedNode methodsFor: #S8Steps-internal !
inject: simulator in: aContext
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	^self	inject: simulator in: aContext
		catch: nil curtail: nil finally: nil ! !

! ScopedNode methodsFor: #S8Steps-internal !
inject: simulator in: aContext catch: catcher curtail: curtail finally: final
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	simulator
		addBlock: self contents: self nodes in: aContext
		catch: catcher curtail: curtail finally: final.
	self isEmpty ifTrue: [
		^simulator add: self step: [ simulator value: nil ] in: aContext
	].! !

! ScopedNode methodsFor: #S8Steps-activation !
contextVariables
	" Private - Returns the variables involved in activation of the receiver. "

	^self implementedBySubclass: #contextVariables ! !

! ScopedNode methodsFor: #S8Steps-internal !
content: simulator actionsIn: aContext
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	^self implementedBySubclass: #content:actionsIn: ! !

! ScopedNode methodsFor: #S8Steps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	self isEmpty ifTrue: [
		^Array with: [:simulator| self content: simulator actionsIn: aContext ]
	].
	^Array with: [:simulator|
		S8Context
			in: aContext clear: self contextVariables
			do: [:ctx| self content: simulator actionsIn: ctx ]
	] ! !

! BlockNode methodsFor: #bindings !
contextVariables
	" Private - Returns the variables involved in activation of the receiver. "

	^self parameters ! !

! BlockNode methodsFor: #S8Steps-internal !
content: simulator actionsIn: aContext
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	| block |
	block := SimulationBlock for: self in: aContext.
	^simulator add: self step: [
		block simulator: simulator.		
		simulator value: block
	] in: aContext ! !

! CascadeNode methodsFor: #S8Steps-query !
isMessageNode
	" Returns true if the receiver is a message send. "

	^true ! !

! CascadeNode methodsFor: #S8Steps-test !
isValid
	" Returns true if the receiver is valid (for simulation). "

	^super isValid and: [ self receiver isValid ] ! !

! CascadeNode methodsFor: #bindings !
contextVariables
	" Private - Returns the variables involved in activation of the receiver. "

	^#() ! !

! CascadeNode methodsFor: #S8Steps-internal !
content: simulator actionsIn: aContext
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	| head tail |
	tail := self nodes collect: [:each| each shallowCopy ].
	tail isEmpty ifTrue: [
		^self error: 'Empty cascade - Should not happen'
	].
	head := self receiver.
	^simulator add: self steps: (Array
		with: [ simulator add: head in: aContext ]
		with: [ | value literal |
			value := simulator value.
			literal := ValueNode literal: '$rec' value: value.
			literal sourceRange: head sourceRange.
			tail do: [:each| each receiver: literal ].
			" aContext at: '$rec' put: value. "
			simulator addAll: tail in: aContext
		]) in: aContext ! !

! SequenceNode methodsFor: #bindings !
contextVariables
	" Private - Returns the variables involved in activation of the receiver. "

	^self temps ! !

! SequenceNode methodsFor: #S8Steps-internal !
content: simulator actionsIn: aContext
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	^self inject: simulator in: aContext ! !

! ExtendedSequenceNode methodsFor: #S8Steps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	| result |
	result := self directives inject: #() into: [:total :each|
		total addAll: (each actionsIn: aContext); yourself
	].
	result addAll: (super actionsIn: aContext).
	^result! !

! SendNode methodsFor: #S8Steps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	^Array
		with: [:simulator| simulator add: self receiver in: aContext ]
		with: [:simulator| simulator add: self steps: (self steps: simulator value in: aContext) in: aContext ] ! !

! SendNode methodsFor: #S8Steps-simulating !
steps: anObject in: aContext
	" Private - Returns the steps to simulate message sent to anObject in aContext. "

	| type |
	self receiver isOO ifTrue: [ type := #message ].
	self receiver isSuper ifTrue: [ type := #super ].
	^Array	with: [:simulator| simulator collect: self all: self arguments in: aContext ]
		with: [:simulator| | send |
			send := self
				steps: type toSend: self selector
				to: anObject withArguments: simulator value
				in: aContext.
			simulator add: self steps: send in: aContext 
		] ! !

! SendNode methodsFor: #S8Steps-sending !
valuesToSend: msg to: anObject withArguments: values in: aContext
	" Private - Returns the arguments to send msg to anObject. "

	(msg startsWith: $#) ifFalse: [ ^values ].
	^values collect: [:each|
		(each isKindOf: SimulationBlock)
		ifTrue: [ self callableOf: each ]
		ifFalse: [ each ]
	] ! !

! SendNode methodsFor: #S8Steps-simulating !
steps: type toSend: msg to: anObject withArguments: objects in: aContext
	" Private - Returns the steps to simulate message sent to anObject in aContext. "

	(self class actionToSend: msg ifAbsent: []) notNil ifTrue: [ "optimized/native message sent"
		^Array with: [:simulator|
			simulator value: (self send: msg to: anObject withArguments: objects in: aContext)
		]
	].
	^Array with: [:simulator| | cls |
		cls := type = #super
			ifTrue: [ simulator topFrame method classField superclass ]
			ifFalse: [ anObject class ].
		simulator
			activate: cls message: (self class selectorToSend: msg)
			for: anObject withArguments: (self
				valuesToSend: msg to: anObject
				withArguments: objects in: aContext)
	] ! !

! SendNode methodsFor: #S8Steps-query !
isMessageNode
	" Returns true if the receiver is a message send. "

	^true ! !

! SendNode methodsFor: #S8Steps-test !
isValid
	" Returns true if the receiver is valid (for simulation). "

	self receiver notNil ifTrue: [ receiver isValid ifFalse: [ ^false ] ].
	self arguments do: [:each| each isValid ifFalse: [ ^false ] ].
	^super isValid ! !

! ValueNode methodsFor: #S8Steps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	^Array with: [:simulator|
		simulator value: (self evaluateIn: aContext)
	] ! !