" Script Compiler for S8 Smalltalk.
  Author: Alejandro F. Reimondo [aleReimondo@smalltalking.net]
  Authored @2020 S8 see http://u8.smalltalking.net
"

Compiler
	subclass: #ScriptCompiler
	instanceVariableNames: #cache
	category: #ScriptCompiler !

CompiledMethod
	subclass: #CompiledScript
	instanceVariableNames: ' args script source '
	category: #ScriptCompiler !
CompiledScript buildAccessors!

! ScriptCompiler class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^#(	#needsScriptFor:
		#codeGenerated:
		#eval:
	) ,super constructEventsTriggered! !

! ScriptCompiler class methodsFor: #installation !
install
	" Install the receiver as default Compiler support.
	Note that effect of changing the compiler is inmediate.
	DO NOT install if you don't know what you are doing.
	"

	self install: self >> #scriptCompilerSupport ! !

! ScriptCompiler class methodsFor: #testing !
isInstalled
	" Returns true if the receiver is installed as default Compiler support. "

	^(self >> #scriptCompilerSupport) fn = (Smalltalk >> #compilerSupport) fn ! !

! ScriptCompiler class methodsFor: #installation !
uninstall
	" Install the default Compiler support. "

	self install: self >> #s8CompilerSupport ! !

! ScriptCompiler class methodsFor: #private !
install: support
	" Private - Install the given compiler support as default Compiler. "

	| selector cm code src |
	selector := #compilerSupport.
	code := '' writeStream.
	src := support source stream.
	src nextLine.
	src atEnd ifFalse: [
		code nextPutAll: selector; cr.
		code nextPutAll: src upToEnd
	].

	cm := Smalltalk >> selector.
	cm source: code contents; fn: support fn.
	cm classField addCompiledMethod: cm ! !

! ScriptCompiler class methodsFor: #private !
evaluationScript
	" Return the support for skipping method execution. "

	^StepOverScript ! !

! ScriptCompiler methodsFor: #installation !
s8CompilerSupport
	" Return the compiler support for the receiver. "

	^Compiler ! !

! ScriptCompiler methodsFor: #installation !
scriptCompilerSupport
	" Return the compiler support for the receiver. "

	^ScriptCompiler ! !

! ScriptCompiler methodsFor: #accessing !
cache
	" Private - Returns the internal cache. "

	cache isNil ifTrue: [ cache := PoolDictionary new ].
	^cache! !

! ScriptCompiler methodsFor: #private !
scriptFor: smalltalkExpression
	" Private - Return the script result of parsing expression to be evaluated. "

	| result |
	result := (ScriptParser for: self currentClass)
		parse: smalltalkExpression onError: [:ex|
			^self error: ex description
		].
	^result! !

! ScriptCompiler methodsFor: #private !
isSourcecode: string
	" Private - Returns true if a string is formated as sourcecode. "

	| input |
	input := string stream.
	[input atEnd] whileFalse: [
		((input upTo: String crlf) includes: String lf) ifTrue: [ ^false ]
	].
	^true! !

! ScriptCompiler methodsFor: #private !
sourcecode: string
	" Private - Returns a string formated as sourcecode. "

	| input output |
	(self isSourcecode: string) ifTrue: [ ^string ].
	input := string stream.
	output := '' writeStream.
	[input atEnd] whileFalse: [
		output nextPutAll: input nextLine.
		input atEnd ifFalse: [ output cr ]
	].
	^output contents! !

! ScriptCompiler methodsFor: #scripting !
scriptForMethod: source onError: aBlock
	" Return the compiled script for method with source code or the result of evaluating aBlock with error. "

	| method |
	method := CompiledScript new.
	method classField: self currentClass.
	(ScriptParser for: self currentClass)
		parse: (self sourcecode: source) method: method
		onError: [:ex| ^aBlock value: ex ].
	^method! !

! ScriptCompiler methodsFor: #scripting !
scriptForMethod: source
	" Return the compiled script for method with source code (or nil). "

	^self scriptForMethod: source onError: [ nil ]! !

! ScriptCompiler methodsFor: #evaluation !
evaluate: smalltalkExpression in: aClass to: aReceiver
	" Return the result of evaluating the smalltalk expression with aReceiver. "

	self currentClass: aClass.
	^(self scriptFor: smalltalkExpression)
		evaluate: aReceiver! !

! ScriptCompiler methodsFor: #private !
compile: aString
	" Return the native code result of compiling aString for currentClass. "

	| code |
	code := super compile: aString.
	self cache at: #code put: code.
	self triggerEvent: #codeGenerated: with: code.
	^code! !

! ScriptCompiler methodsFor: #evaluation !
basicEval: expression
	" Private - Return the result of evaluating a native expression in global context.
	Note that we do NOT evaluate the code in case we parsed to a compiled script.
	"

	| default script code |
	default := [ super eval: expression ].
	script := self cache at: #compiledScript ifAbsent: [ ^default value ].
	code := self cache at: #code ifAbsent: [ ^default value ].
	code = expression ifTrue: [ ^script ].
	^default value! !

! ScriptCompiler methodsFor: #evaluation !
eval: nativeExpression
	" Return the result of evaluating a native expression in global context. "

	^self	triggerEvent: #eval: with: nativeExpression
		ifNotHandled: [ self basicEval: nativeExpression ]! !

! ScriptCompiler methodsFor: #parsing !
parseMethod: aString
	" Return the result of parsing aString as a method source code. "

	| script |
	script := self
		triggerEvent: #needsScriptFor: with: aString
		ifNotHandled: [ self scriptForMethod: aString ].
	script notNil ifTrue: [ self cache at: #compiledScript put: script ].
	^super parseMethod: aString! !

! CompiledScript methodsFor: #accessing !
fn
	" Return the native implementation of the receiver. "

	| fn |
	fn := super fn.
	fn isNil ifTrue: [
		fn := script nativeImplementationFor: self.
		self fn: fn.
	].
	^fn ! !

! CompiledScript methodsFor: #export !
exportFor: anExporter on: aStream
	" Export the native representation of the receiver on to aStream. "

	| encoder cls nClassName |
self compiledSource isNil ifTrue: ["NIY" ^false ].
	cls := self classField.
	encoder := anExporter encoderFor: cls.
"ToDo - encode to recompile on first run, with native source equivalent to:
	[ self recompile. #{this} basicPerform: self selector withArguments: #{arguments} ]
"	^encoder export: cls method: self on: aStream ! !
