" (script based) Step by step execution support for S8 Smalltalk.
  Author: Alejandro F. Reimondo [aleReimondo@smalltalking.net]
  Authored @2020 S8 see http://u8.smalltalking.net
"
(Smalltalk includesKey: #CompiledScript) ifFalse: [
	^self error: 'ScriptCompiler framework missing'
]!
(Smalltalk includesKey: #S8Block) ifFalse: [
	^self error: 'Bindings framework missing'
]!
(Smalltalk includesKey: #S8Block) ifFalse: [
	^self error: 'Scripting framework missing'
]!

! CompiledScript methodsFor: #ScriptSteps-simulating !
compiledScript
	" Returns the (compiled)script to execute when simulating the receiver. "

	^self ! !

! Script methodsFor: #ScriptSteps-simulating !
stepsFor: aReceiver entering: aMethod withArguments: array
	" Returns the steps to evaluate execution of the receiver activating aMethod. "

	| ctx |
	ctx := self contextFor: aReceiver binding: (self argsIn: aMethod) to: array.
	^(self actionsIn: ctx) collect: [:each|
		SimulationStep for: self doing: each in: ctx
	]! !

! ParseNode methodsFor: #ScriptSteps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext.
	The default implementation will evaluate the receiver in a single step.
	This method can be refined by the subclasses.
	"

	^Array with: [ self evaluateIn: aContext ] ! !

! Script methodsFor: #ScriptSteps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	| result |
	result := super actionsIn: aContext.
	result addFirst: [:simulator|
		self exitBlock: [:value| simulator exit: value ]
	].
	^result ! !

! ParseStatements methodsFor: #ScriptSteps-internal !
inject: simulator in: aContext
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	^self	inject: simulator in: aContext
		catch: nil curtail: nil finally: nil ! !

! ParseStatements methodsFor: #ScriptSteps-internal !
inject: simulator in: aContext catch: catcher curtail: curtail finally: final
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	simulator
		addBlock: self contents: self statements in: aContext
		catch: catcher curtail: curtail finally: final.
	self isEmpty ifTrue: [
		^simulator add: self step: [ simulator value: nil ] in: aContext
	].! !

! ParseStatements methodsFor: #ScriptSteps-internal !
content: simulator actionsIn: aContext
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	^self inject: simulator in: aContext ! !

! ParseStatements methodsFor: #ScriptSteps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	self isEmpty ifTrue: [
		^Array with: [:simulator| self content: simulator actionsIn: aContext ]
	].
	^Array	with: [:simulator|
			S8Context
				in: aContext clear: self contextVariables
				do: [:ctx| self content: simulator actionsIn: ctx ]
		] ! !

! BlockClosureNode methodsFor: #ScriptSteps-internal !
content: simulator actionsIn: aContext
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	| block |
	block := SimulationBlock for: self in: aContext.
	^simulator add: self step: [
		block simulator: simulator.		
		simulator value: block
	] in: aContext ! !

! CascadedMessageNode methodsFor: #ScriptSteps-internal !
content: simulator actionsIn: aContext
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	| head tail |
	tail := self statements collect: [:each| each shallowCopy ].
	tail isEmpty ifTrue: [
		^self error: 'Empty cascade - Should not happen'
	].
	head := tail first receiver.
	^simulator add: self steps: (Array
		with: [ simulator add: head in: aContext ]
		with: [ | value literal |
			value := simulator value.
			literal := ParseLiteral
				sourceRange: head sourceRange
				value: value.
			tail do: [:each| each receiver: literal ].
			" aContext at: '$rec' put: value. "
			simulator addAll: tail in: aContext
		]) in: aContext ! !


! ParseVariable methodsFor: #ScriptSteps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	| result |
	result := Array with: [:simulator| simulator value: (self valueIn: aContext) ].
	indices notNil ifTrue: [
		result add: [:simulator| simulator value: (self index: simulator value in: aContext) ]
	].
	^result ! !

! MethodReturnNode methodsFor: #ScriptSteps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	^Array
		with: [:simulator| simulator add: self expression in: aContext ]
		with: [:simulator|
			simulator return: self in: aContext then: [
				exitBlock value: simulator value
			]
		] ! !

! AssignNode methodsFor: #ScriptSteps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	^Array
		with: [:simulator| simulator add: self expression in: aContext ]
		with: [:simulator| self variable bind: simulator value in: aContext ] ! !

! MessageNode methodsFor: #ScriptSteps-simulating !
actionsIn: aContext
	" Returns the steps to evaluate while simulating execution of the receiver in aContext. "

	^Array
		with: [:simulator| simulator add: self receiver in: aContext ]
		with: [:simulator| simulator add: self steps: (self steps: simulator value in: aContext) in: aContext ] ! !

! MessageNode methodsFor: #ScriptSteps-simulating !
steps: anObject in: aContext
	" Private - Returns the steps to simulate message sent to anObject in aContext. "

	| sender |
	sender := (self receiver isSuper or: [ self receiver isOO not ])
		ifTrue: [ self receiver ] ifFalse: [ self ].
	^Array	with: [:simulator| simulator collect: self all: self args in: aContext ]
		with: [:simulator| | send |
			send := sender
				stepsToSend: self selector
				to: anObject withArguments: simulator value
				in: aContext.
			simulator add: self steps: send in: aContext 
		] ! !

! MessageNode methodsFor: #ScriptSteps-simulating !
stepsToSend: msg to: anObject withArguments: objects in: aContext
	" Private - Returns the steps to simulate message sent to anObject in aContext. "

	(self class actionToSend: msg ifAbsent: []) notNil ifTrue: [ "optimized/native message sent"
		^Array with: [:simulator|
			simulator value: (self send: msg to: anObject withArguments: objects in: aContext)
		]
	].
	^Array with: [:simulator|
		simulator
			activate: anObject class message: msg
			for: anObject withArguments: objects
	] ! !

! NativeMessageNode methodsFor: #ScriptSteps-simulating !
stepsToSend: msg to: anObject withArguments: values in: aContext
	" Private - Returns the steps to simulate message sent to anObject in aContext. "

	^Array with: [:simulator|
		simulator value: (self
			send: msg to: anObject
			withArguments: values in: aContext)
	] ! !

! ParseSuper methodsFor: #ScriptSteps-simulating !
stepsToSend: msg to: anObject withArguments: values in: aContext
	" Private - Returns the steps to simulate message sent to anObject in aContext. "

	^Array with: [:simulator| | mth cls |
		cls := implementor superclass.
		cls notNil ifTrue: [ mth := cls >>> msg ].
		mth isNil ifTrue: [
			simulator
				activate: implementor class message: #missingImplementorFor:
				for: implementor withArguments: (Array with: msg)
		] ifFalse: [
			simulator
				activate: cls message: msg
				for: anObject withArguments: values
		]
	] ! !

! ParseNoReceiver methodsFor: #ScriptSteps-simulating !
stepsToSend: msg to: anObject withArguments: values in: aContext
	" Private - Returns the steps to simulate message sent to anObject in aContext. "

	^Array with: [:simulator|
		simulator value: (self send: msg to: anObject withArguments: values in: aContext)
	] ! !