" Scripting support for S8 Smalltalk.
  Author: Alejandro F. Reimondo [aleReimondo@smalltalking.net]
  Authored @1999 SUGAR http://www.sugarWeb.com
  Updated  @2014 Smalltalking http://www.smalltalking.net
  Updated  @2020 S8 see http://u8.smalltalking.net

TestSet:
	swiki:s8-media/TestScriptingIndex
	SwikiCodeRobot @> #(test scripting)

Simple test expressions:
(ScriptParser new parse: '#hello') evaluate = #hello
(ScriptParser new parse: '|tres| tres := 3. #hello. tres') evaluate = 3
(ScriptParser new parse: '| i | i := 12. #(1 2 3 4) do: [:i| self print: i * i]. i') evaluate = 12
"
(Smalltalk includesKey: #S8Context) ifFalse: [
	^self error: 'Missing Bindings framework'
]!

Object
	subclass: #ScriptingObject
	instanceVariableNames: ''
	category: #Scripting !
ScriptingObject
	subclass: #ParseContext
	instanceVariableNames: ' map parent '
	category: #Scripting !
ScriptingObject
	subclass: #ParseNode
	instanceVariableNames: ' sourceRange '
	category: #Scripting !
ParseNode
	subclass: #AssignNode
	instanceVariableNames: ' variable expression '
	category: #Scripting !
ParseNode
	subclass: #MessageNode
	instanceVariableNames: ' receiver selector args action '
	category: #Scripting !
MessageNode
	subclass: #NativeMessageNode
	instanceVariableNames: ''
	category: #Scripting !
ParseNode
	subclass: #ParseStatements
	instanceVariableNames: ' temporaries statements '
	category: #Scripting !
ParseStatements
	subclass: #BlockClosureNode
	instanceVariableNames: ' args '
	category: #Scripting !
ParseStatements
	subclass: #CascadedMessageNode
	instanceVariableNames: ' receiver '
	category: #Scripting !
ParseStatements
	subclass: #Script
	instanceVariableNames: ''
	category: #Scripting !
Script
	subclass: #StepOverScript
	instanceVariableNames: ' action '
	category: #Scripting !
ParseNode
	subclass: #ParseVariable
	instanceVariableNames: ' binding indices '
	category: #Scripting !
ParseVariable
	subclass: #ParseArgument
	instanceVariableNames: ' canBeAssigned '
	category: #Scripting !
ParseVariable
	subclass: #ParseGlobalVariable
	instanceVariableNames: ''
	category: #Scripting !
ParseGlobalVariable
	subclass: #ParseGlobalConstant
	instanceVariableNames: ''
	category: #Scripting !
ParseVariable
	subclass: #ParseSpecial
	instanceVariableNames: ''
	category: #Scripting !
ParseSpecial
	subclass: #ParseLiteral
	instanceVariableNames: ''
	category: #Scripting !
ParseSpecial
	subclass: #ParseInstanceVariable
	instanceVariableNames: ' name index '
	category: #Scripting !
ParseSpecial
	subclass: #ParseUniversal
	instanceVariableNames: ''
	category: #Scripting !
ParseUniversal
	subclass: #ParseSelf
	instanceVariableNames: ''
	category: #Scripting !
ParseSelf
	subclass: #ParseSuper
	instanceVariableNames: 'implementor'
	category: #Scripting !
ParseSpecial
	subclass: #ParseNoReceiver
	instanceVariableNames: ''
	category: #Scripting !
ParseSpecial
	subclass: #ParseNativeVariable
	instanceVariableNames: ''
	category: #Scripting !
ParseNode
	subclass: #MethodExitNode
	instanceVariableNames: ' exitBlock '
	category: #Scripting !
MethodExitNode
	subclass: #MethodReturnNode
	instanceVariableNames: ' expression '
	category: #Scripting !
MethodExitNode
	subclass: #PrimitiveNode
	instanceVariableNames: ' action '
	category: #Scripting !
ScriptingObject
	subclass: #ScriptParser
	instanceVariableNames: ' currentClass source scope instanceVariables errorBlock '
	category: #Scripting !
ScriptParser class instanceVariableNames: ' primitiveHandlers defaultPrimitiveHandler ' !
ScriptingObject
	subclass: #ScriptingError
	instanceVariableNames: ' description '
	category: #Scripting !
ScriptingError
	subclass: #ParseError
	instanceVariableNames: ' source position '
	category: #Scripting !
ParseError
	subclass: #SyntaxError
	instanceVariableNames: ''
	category: #Scripting !


! ParseContext class methodsFor: #defaults !
default
	" Returns an instance of the receiver set to the default scope. "

	^self new
		beDefault;
		yourself ! !

! ParseContext class methodsFor: #instantiation !
in: aParentScope
	" Returns the a new scope included in aParentScope. "

	^self new
		parent: aParentScope;
		yourself ! !

! ParseContext methodsFor: #accessing !
at: aKey put: anObject
	"Creates a new variable binding for aKey in the receiver's local context (the map). "

	^self map at: aKey put: anObject! !

! ParseContext methodsFor: #accessing !
map
	" Private - Returns the receiver's map. "

	map isNil ifTrue: [ self initializeMap ].
	^map ! !

! ParseContext methodsFor: #accessing !
parent: aScope
	" Private - Sets the receiver's parent scope. "

	parent := aScope ! !

! ParseContext methodsFor: #binding !
bindingAt: aName
	" LookUp for variable named aName in the receiver.
	Returns the binding found for aName or produces an error if failed. "

	^self bindingAt: aName ifAbsent: [ self error: 'Absent element.' ] ! !

! ParseContext methodsFor: #binding !
bindingAt: aName ifAbsent: errorBlock
	" LookUp for variable named aName in the receiver.
	Returns the binding found for aName or evaluates errorBlock if failed. "

	| assoc key |
	aName size = 0 ifTrue: [ ^nil ].

	assoc := self bindingIn: self map at: aName ifAbsent: [].
	assoc notNil ifTrue: [ ^assoc local ].

	(parent isKindOf: KeyedCollection) ifTrue: [
		assoc := self bindingIn: parent at: aName ifAbsent: [].
		assoc notNil ifTrue: [ ^assoc ].
	].

	^parent notNil
	ifTrue: [ parent bindingAt: aName ifAbsent: errorBlock ]
	ifFalse: [ errorBlock value ] ! !

! ParseContext methodsFor: #binding !
bindingIn: aDictionary at: aName ifAbsent: errorBlock
	" Private - LookUp for variable named aName in the dictionary aDictionary. "

	| assoc key |
	^aDictionary bindingAt: aName ifAbsent: [
		key := aDictionary keys detect: [:one| one asString = aName ] ifNone: [].
		key notNil ifTrue: [ aDictionary bindingAt: key ifAbsent: [] ]
	] ! !

! ParseContext methodsFor: #initialize !
initializeMap
	" Private - Initializes the receiver's map. "

	map := PoolDictionary new ! !

! ParseContext methodsFor: #private !
beDefault
	" Private - Set the receiver as child of default context. "

	parent := self smalltalk ! !

! ParseNode class methodsFor: #instantiation !
sourceRange: interval
	" Returns an instance of the receiver. "

	^self new sourceRange: interval ! !

! ParseNode class methodsFor: #instantiation !
from: start to: stop
	" Returns an instance of the receiver. "

	^self sourceRange: (start to: stop) ! !


! ParseNode methodsFor: #accessing !
sourcePosition: anInterval
	" Set the receiver's sourceRange. "

	^self sourceRange: anInterval ! !

! ParseNode methodsFor: #accessing !
sourceRange
	" Returns the receiver's sourceRange. "

	^sourceRange! !

! ParseNode methodsFor: #accessing !
sourceRange: anInterval
	" Set the receiver's sourceRange. "

	sourceRange := anInterval ! !

! ParseNode methodsFor: #composing !
, aParseNode
	" Returns a composite parse node with the receiver and aParseNode appended at end. "

	^ParseStatements
		with: self
		with: aParseNode ! !

! ParseNode methodsFor: #evaluating !
evaluate
	" Evaluate the receiver in an empty context. "

	^self evaluate: nil ! !

! ParseNode methodsFor: #evaluating !
evaluate: aReceiver
	" Returns the result of evaluating the receiver in a context bound to aReceiver. "

	^self evaluateIn: (S8Context for: aReceiver) ! !

! ParseNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext.
	This message must be implemented by the subclasses. "

	^self subclassResponsibility: #evaluateIn: ! !

! ParseNode methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver.
	This block will be used to abort execution of the receiver.
	Do nothing by default.
	This method can be refined by subclasses. " ! !

! ParseNode methodsFor: #private !
clear: aContext
	" Private - Clear the receiver's cached state.
	This method must be implemented by subclasses. "

	^self subclassResponsibility ! !

! ParseNode methodsFor: #query !
isComposite
	" Returns true if the receiver is a composite parse node. "

	^false ! !

! ParseNode methodsFor: #query !
isReturn
	" Returns true if the receiver is a return parse node. "

	^false ! !

! ParseNode methodsFor: #query !
isSuper
	" Returns true if the receiver is a reference to super. "

	^false ! !

! ParseNode methodsFor: #query !
isOO
	" Returns true if the receiver refers to an object. "

	^true ! !

! ParseNode methodsFor: #query !
isMessageNode
	" Returns true if the receiver is a message send. "

	^false ! !

! ParseNode methodsFor: #query !
isPrimitiveNode
	" Returns true if the receiver is a primitive. "

	^false ! !

! ParseNode methodsFor: #query !
isVariable
	" Returns true if the receiver is a variable parse node. "

	^false ! !


! AssignNode methodsFor: #accessing !
expression
	" Returns the receiver's expression. "

	^expression ! !

! AssignNode methodsFor: #accessing !
expression: anExpression
	" Set the receiver's expression. "

	expression := anExpression! !

! AssignNode methodsFor: #accessing !
variable
	" Returns the receiver's variable. "

	^variable ! !

! AssignNode methodsFor: #accessing !
variable: aVariable
	" Set the receiver's variable. "

	variable := aVariable ! !

! AssignNode methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver. "

	super exitBlock: aBlock.
	self variable exitBlock: aBlock.
	self expression exitBlock: aBlock ! !

! AssignNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^self variable bind: (self expression evaluateIn: aContext) in: aContext ! !

! MessageNode class methodsFor: #actions !
actionToSend: selector ifAbsent: aBlock
	" Private - Returns the block to evaluate for sending message selector. "

	^S8Block optimizations at: selector ifAbsent: aBlock ! !

! MessageNode class methodsFor: #instantiation !
native: isNative from: start to: stop
	" Returns an instance of the receiver. "

	^(isNative
		ifTrue: [ NativeMessageNode ]
		ifFalse: [ MessageNode ])
		from: start to: stop ! !

! MessageNode class methodsFor: #instantiation !
receiver: receiver selector: selector arguments: args
	" Returns a new instance of the receiver. "

	^self new receiver: receiver selector: selector arguments: args ! !

! MessageNode methodsFor: #accessing !
receiver: aParseNode selector: aSelector arguments: aCollection
	" Set the receiver contents. "

	self	receiver: aParseNode;
		selector: aSelector;
		arguments: aCollection ! !

! MessageNode methodsFor: #accessing !
args
	^args ! !

! MessageNode methodsFor: #accessing !
args: aCollection
	args := aCollection ! !

! MessageNode methodsFor: #accessing !
arguments
	^self args ! !

! MessageNode methodsFor: #accessing !
arguments: aCollection
	self args: aCollection! !

! MessageNode methodsFor: #accessing !
receiver
	^receiver! !

! MessageNode methodsFor: #accessing !
receiver: aParseNode
	receiver := aParseNode ! !

! MessageNode methodsFor: #accessing !
selector
	^selector! !

! MessageNode methodsFor: #accessing !
selector: aSymbol
	selector := aSymbol ! !

! MessageNode methodsFor: #query !
isMessageNode
	" Returns true if the receiver is a message send. "

	^true ! !

! MessageNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	| value |
	value := self receiver evaluateIn: aContext.
	^self evaluate: self receiver value: value in: aContext ! !

! MessageNode methodsFor: #actions !
actionToSend: aSymbol
	" Private - Returns the block to evaluate for sending message aSymbol. "

	^self class actionToSend: aSymbol ifAbsent: [
		[:receiver :msg :values :context|
			receiver perform: msg withArguments: values
		]
	] ! !

! MessageNode methodsFor: #sending !
valuesToSend: msg to: anObject withArguments: values in: aContext
	" Private - Returns the arguments to send msg to anObject. "

	^values ! !

! MessageNode methodsFor: #sending !
send: msg to: anObject withArguments: values in: aContext
	" Private - Returns the result of sending message to anObject in aContext. "

	action isNil ifTrue: [ action := self actionToSend: msg ].
	^action	value: anObject value: msg
		value: (self
			valuesToSend: msg to: anObject
			withArguments: values in: aContext)
		value: aContext ! !

! MessageNode methodsFor: #evaluating !
evaluate: aReceiver value: anObject in: aContext
	" Private - Returns the result of sending message to anObject in aContext. "

	| values |
	values := self args collect: [:each| each evaluateIn: aContext ].
	(aReceiver isSuper or: [ aReceiver isOO not ]) ifTrue: [
		^aReceiver send: self selector to: anObject withArguments: values in: aContext
	].
	^self send: self selector to: anObject withArguments: values in: aContext! !

! MessageNode methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver. "

	super exitBlock: aBlock.
	self receiver exitBlock: aBlock.
	self args do: [:each| each exitBlock: aBlock ] ! !

! NativeMessageNode methodsFor: #accessing !
selector: aSymbol
	" Set the selector converting aSymbol to native selector. "

	super selector: (aSymbol upTo: $:) ! !

! NativeMessageNode methodsFor: #sending !
valuesToSend: msg to: anObject withArguments: values in: aContext
	" Private - Returns the arguments to send msg to anObject. "

	^values collect: [:each|
		(each isKindOf: SimulationBlock)
		ifTrue: [ self callableOf: each ]
		ifFalse: [ each ]
	] ! !

! NativeMessageNode methodsFor: #sending !
send: msg to: anObject withArguments: values in: aContext
	" Private - Returns the result of sending message to anObject in aContext. "

	| fn anArray |
	fn := anObject basicAt: msg ifAbsent: [
		^self error: 'Missing function ' ,msg asLiteral
	].
	anArray := self
		valuesToSend: msg to: anObject
		withArguments: values in: aContext.
	^{''
	js'fn.apply(anObject,anArray)'
	lua'fn(anObject,(table.unpack or unpack)(anArray))'
	py'__0__.fn(anObject,*smalltalk.objects_toCall_(anArray,__0__.fn))'}! !

! ParseStatements class methodsFor: #instantiation !
with: aStatement
	" Returns an instance of the receiver containing aStatement. "

	^self new
		add: aStatement;
		yourself ! !

! ParseStatements class methodsFor: #instantiation !
with: statement1 with: statement2
	" Returns an instance of the receiver containing the statements. "

	^self new
		add: statement1;
		add: statement2;
		yourself ! !

! ParseStatements class methodsFor: #instantiation !
with: statement1 with: statement2 with: statement3
	" Returns an instance of the receiver containing the statements. "

	^self new
		add: statement1;
		add: statement2;
		add: statement3;
		yourself ! !

! ParseStatements class methodsFor: #instantiation !
with: statement1 with: statement2 with: statement3 with: statement4
	" Returns an instance of the receiver containing the statements. "

	^self new
		add: statement1;
		add: statement2;
		add: statement3;
		add: statement4;
		yourself ! !

! ParseStatements class methodsFor: #instantiation !
withAll: aCollection
	" Returns an instance of the receiver containing all the statements in aCollection. "

	^self new
		addAll: aCollection;
		yourself ! !

! ParseStatements methodsFor: #accessing !
statements
	" Returns the receiver's statements. "

	statements isNil ifTrue: [ self initializeStatements ].
	^statements ! !

! ParseStatements methodsFor: #accessing !
temporaries
	" Returns the receiver's temporaries. "

	temporaries isNil ifTrue: [ self initializeTemporaries ].
	^temporaries ! !

! ParseStatements methodsFor: #accessing !
temporaries: aCollection
	" Set the receiver's temporaries. "

	temporaries := aCollection ! !

! ParseStatements methodsFor: #accessing !
sourceRange
	" Returns the receiver's sourceRange. "

	| start end |
	sourceRange notNil ifTrue: [ ^sourceRange ].
	statements size = 0 ifTrue: [ ^nil ].

	start := 999. end := 0.
	statements do: [:each| | range |
		range := each sourceRange.
		range notNil ifTrue: [
			start := start min: range start.
			end := end max: range end.
		].
	].
	end = 0 ifTrue: [ ^nil ].

	^start to: end ! !

! ParseStatements methodsFor: #collection !
add: aStatement
	"Add aStatement to the receiver. "

	^self addLast: aStatement ! !

! ParseStatements methodsFor: #collection !
addAll: aCollection
	"Add aCollection of statements to the receiver. "

	^self statements addAll: aCollection ! !

! ParseStatements methodsFor: #collection !
addFirst: aStatement
	"Add aStatement to the receiver. "

	^self statements addFirst: aStatement ! !

! ParseStatements methodsFor: #collection !
addFirstAll: aCollection
	"Add aCollection of statements to the receiver. "

	aCollection asArray reverseDo: [:each|
		self statements addFirst: each
	] ! !

! ParseStatements methodsFor: #collection !
addLast: aStatement
	"Add aStatement to the receiver. "

	^self statements addLast: aStatement ! !

! ParseStatements methodsFor: #collection !
remove: aStatement
	"Remove aStatement from the receiver. "

	^self statements remove: aStatement ! !

! ParseStatements methodsFor: #collection !
removeAll: aCollection
	"Remove aCollection of statements from the receiver. "

	^self statements removeAll: aCollection ! !

! ParseStatements methodsFor: #activation !
contextVariables
	" Private - Returns the variables involved in activation of the receiver. "

	^self temporaries ! !

! ParseStatements methodsFor: #evaluating !
evaluateContentsIn: localContext
	" Private - Returns the result of evaluating the contents of the receiver in local context. "

	^self statements inject: nil into: [:result :each|
		each evaluateIn: localContext
	]! !

! ParseStatements methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^S8Context
		in: aContext clear: self contextVariables
		do: [:ctx| self evaluateContentsIn: ctx ] ! !

! ParseStatements methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver. "

	super exitBlock: aBlock.
	self temporaries do: [:each| each exitBlock: aBlock ].
	self statements do: [:each| each exitBlock: aBlock ] ! !

! ParseStatements methodsFor: #initialize !
initializeStatements
	" Private - Initialize the receiver's statements. "

	statements := Array new ! !

! ParseStatements methodsFor: #initialize !
initializeTemporaries
	" Private - Initialize the receiver's temporaries. "

	temporaries := Array new ! !

! ParseStatements methodsFor: #query !
isComposite
	" Returns true if the receiver is a composite parse node. "

	^true ! !

! ParseStatements methodsFor: #query !
isEmpty
	" Returns true if the receiver is empty. "

	^self statements isEmpty ! !


! BlockClosureNode methodsFor: #accessing !
args
	" Returns the receiver's arguments. "

	args isNil ifTrue: [ self initializeArgs ].
	^args ! !

! BlockClosureNode methodsFor: #accessing !
args: aCollection
	" Set the receiver's arguments. "

	args := aCollection! !

! BlockClosureNode methodsFor: #accessing !
arguments
	" Returns the receiver's arguments. "

	^self args ! !

! BlockClosureNode methodsFor: #accessing !
arguments: aCollection
	" Set the receiver's arguments. "

	self args: aCollection ! !

! BlockClosureNode methodsFor: #activation !
contextVariables
	" Private - Returns the variables involved in activation of the receiver. "

	^self args ,super contextVariables ! !

! BlockClosureNode methodsFor: #evaluating !
exitBlock: aBlock
	" Set the exit block of the receiver. "

	super exitBlock: aBlock.
	self args do: [:each| each exitBlock: aBlock ] ! !

! BlockClosureNode methodsFor: #initialize !
initializeArgs
	" Private - Initializes the receiver's arguments. "

	args := Array new ! !

! BlockClosureNode methodsFor: #evaluating !
evaluateContentsIn: localContext
	" Private - Returns the result of evaluating the contents of the receiver in local context.
	Block defer evaluation until message #value:... is sent
	 (re-using a local context for execution) "

	^S8Block for: self in: localContext ! !


! CascadedMessageNode class methodsFor: #instantiation !
for: aMessageNode
	" Return an instance of the receiver. "

	^self new initialize: aMessageNode ! !

! CascadedMessageNode methodsFor: #accessing !
receiver
	" Returns the receiver's receiver. "

	^receiver ! !

! CascadedMessageNode methodsFor: #accessing !
receiver: aReceiver
	" Set the receiver's receiver. "

	receiver := aReceiver ! !

! CascadedMessageNode methodsFor: #initialize !
initialize: expression
	" Private - Initialize the receiver to start execution with expression. "

	self add: (MessageNode new
		receiver: expression receiver
		selector: expression selector
		arguments: expression arguments) ! !

! CascadedMessageNode methodsFor: #evaluating !
evaluateContentsIn: localContext
	" Private - Returns the result of evaluating the contents of the receiver in local context. "

	| result rec assoc |
	self statements do: [:each|
		assoc isNil ifTrue: [
			rec := each receiver.
			assoc := '$rec' -> (rec evaluateIn: localContext).
		].
		result := each evaluate: rec value: assoc value in: localContext.
	].
	^result ! !

! Script methodsFor: #evaluating !
clearExitBlock
	" Clear the exit block. "

	self exitBlock: nil ! !

! Script methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	self exitBlock: [:result| self clearExitBlock. ^result ].
	^super evaluateIn: aContext ! !

! Script methodsFor: #evaluating !
contextFor: aReceiver binding: args to: array
	" Private - Returns the context to evaluate the receiver on aReceiver with parameters. "

	| result count |
	result := S8Context for: aReceiver.
	count := args size min: array size.
	1 to: count do: [:i| result at: (args at: i) binding key put: (array at: i) ].
	^result ! !

! Script methodsFor: #evaluating !
evaluate: aReceiver binding: args to: array
	" Returns the result of evaluating the receiver with activation bindings. "

	^self evaluateIn: (self contextFor: aReceiver binding: args to: array)! !

! Script methodsFor: #native !
nativeImplementationFor: mth
	" Return the native implementation to run a compiled script (mth). "

	| msg |
	msg := Smalltalk nativeLanguage ,#ImplementationFor:args:.
	(self respondsTo: msg) ifTrue: [
		^self perform: msg withArguments: (Array with: mth with: (self argsIn: mth))
	].
	^self error: 'Should not happen ' ,#nativeImplementationFor: ! !

! Script methodsFor: #native !
argsIn: mth
	" Private - Return the arguments bound to mth. "

	^mth args ! !

! Script methodsFor: #native !
jsImplementationFor: mth args: args
	" Private - Return the function to run mth. "

	^[ self evaluate: #{this} binding: args to: (Array slice: #{arguments}) ]! !

! Script methodsFor: #native !
luaImplementationFor: mth args: args
	" Private - Return the function to run mth. "

	| script fn hash |
	hash := $#.
	script := self.
	{''lua'fn = function (...)
		local array = smalltalk:newArray()
		for i=1,select(hash,...),1 do array:add_(select(i,...)) end
		return script:evaluate_binding_to_(self,args,array)
	end'}.
	^fn! !

! Script methodsFor: #native !
pyImplementationFor: mth args: args
	" Private - Return the function to run mth. "

	| script |
	script := self.
	{''py'def result(rcv,*params):%>%return __0__.script.evaluate_binding_to_(rcv,args,%Array%(params))%<%if True: return result'}! !

! StepOverScript class methodsFor: #instantiation !
doing: aBlock
	" Returns an instance of the receiver. "

	^self new initialize: aBlock ! !

! StepOverScript methodsFor: #initialize !
initialize: anAction
	" Private - Initialize the receiver. "

	action := anAction ! !

! StepOverScript methodsFor: #native !
argsIn: aMethod
	" Private - Return the arguments bound to aMethod. "

	| i | i := 0.
	^(aMethod selector asArrayOfSubstringsSeparatedBy: $:) collect: [:each|
		i := i + 1. $( ,i ,$) ,each
	] ! !

! StepOverScript methodsFor: #evaluating !
contextFor: aReceiver binding: args to: array
	" Returns the context to evaluate the receiver on aReceiver with parameters. "

	| result count |
	result := S8Context for: aReceiver.
	count := args size min: array size.
	1 to: count do: [:i| result at: (args at: i) put: (array at: i) ].
	^result ! !

! StepOverScript methodsFor: #simulating !
stepsFor: aReceiver entering: aMethod withArguments: array
	" Returns the steps to evaluate execution of the receiver activating aMethod. "

	self sourceRange: (0 to: aMethod source size - 1).
	^super stepsFor: aReceiver entering: aMethod withArguments: array ! !

! StepOverScript methodsFor: #internal !
inject: simulator in: aContext
	" Private - Add the steps for evaluation of contents of the receiver in aContext. "

	self isEmpty ifTrue: [
		^simulator add: self
			step: [:simulator| simulator value: (action evaluateWith: simulator) ]
			in: aContext
	].
	^self invalidMessage: #inject:in: ! !

! StepOverScript methodsFor: #evaluating !
evaluateContentsIn: localContext
	" Private - Returns the result of evaluating the contents of the receiver in local context. "

	self isEmpty ifTrue: [ ^action evaluateWith: nil ].
	^self invalidMessage: #evaluateContentsIn: ! !

! ParseVariable class methodsFor: #instantiation !
boundTo: assoc
	" Returns an instance of the receiver. "

	^self new binding: assoc ! !

! ParseVariable methodsFor: #accessing !
name
	" Returns the receiver's name. "

	^self binding key ! !

! ParseVariable methodsFor: #accessing !
indices: anArray
	" Set the receiver's indices. "

	indices := anArray ! !

! ParseVariable methodsFor: #binding !
bind: aValue in: aContext
	" Set the receiver's value to aValue in aContext.
	Check if the receiver is assignable.
	WARNING: The receiver MUST return the value set. "

	self canBeAssigned ifFalse: [
		^self error: 'Invalid assignment'
	].
	^self assignValue: aValue in: aContext ! !

! ParseVariable methodsFor: #binding !
binding
	" Returns the receiver's binding. "

	^binding ! !

! ParseVariable methodsFor: #binding !
binding: aBinding
	" Set the receiver's binding. "

	binding := aBinding ! !

! ParseVariable methodsFor: #indexing !
key: index in: aContext
	" Private - Returns the key to index an object in aContext. "

	^(index isKindOf: ParseNode)
	ifTrue: [ index evaluateIn: aContext ]
	ifFalse: [ index ]! !

! ParseVariable methodsFor: #indexing !
index: anObject in: aContext
	" Private - Returns the result of indexing anObject. "

	indices isNil ifTrue: [ ^anObject ].
	^indices inject: anObject into: [:result :index|
		result	basicAt: (self key: index in: aContext)
			ifAbsent: [ ]
	]! !

! ParseVariable methodsFor: #indexing !
follow: anObject in: aContext
	" Private - Returns the result of following the indices from anObject and last index to evaluate (or nil). "

	| target index |
	indices isNil ifTrue: [ ^nil ].
	target := anObject.
	1 to: indices size - 1 do: [:i|
		index := self key: (indices at: i) in: aContext.
		target := target basicAt: index ifAbsent: []
	].
	index := self key: indices last in: aContext.
	^Array with: target with: index ! !

! ParseVariable methodsFor: #evaluating !
assignValue: aValue in: aContext
	" Private - Set the receiver's value to aValue.
	WARNING: The receiver MUST return the value set. "

	^self binding
		assign: (indices notNil ifTrue: [ [:root| self follow: root in: aContext ] ])
		to: aValue in: aContext ! !

! ParseVariable methodsFor: #value !
valueIn: aContext
	" Private - Returns the value of the receiver in aContext. "

	^self binding evaluateIn: aContext ! !

! ParseVariable methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^self index: (self valueIn: aContext) in: aContext ! !

! ParseVariable methodsFor: #query !
canBeAssigned
	" Returns true if the receiver is a variable that can be assigned. "

	^false ! !

! ParseVariable methodsFor: #query !
isVariable
	" Returns true if the receiver is a variable parse node. "

	^true ! !


! ParseArgument methodsFor: #evaluating !
clear: aContext
	" Clear the receiver's contents. "

	^aContext clear: self binding ! !

! ParseArgument methodsFor: #query !
canBeAssigned
	" Returns true if the receiver is a variable that can be assigned. "

	^canBeAssigned == true! !

! ParseArgument methodsFor: #query !
canBeAssigned: aBoolean
	" Set canBeAssigned to aBoolean. "

	canBeAssigned := aBoolean ! !


! ParseGlobalVariable methodsFor: #query !
canBeAssigned
	" Returns true if the receiver is a variable that can be assigned. "

	^true ! !


! ParseGlobalConstant methodsFor: #query !
canBeAssigned
	" Returns true if the receiver is a variable that can be assigned. "

	^false! !


! ParseSpecial class methodsFor: #evaluating !
value: aValue
	" Returns an instance of the receiver with value aValue. "

	^self new
		binding: (S8LiteralBinding value: aValue);
		yourself ! !

! ParseLiteral class methodsFor: #instantiation !
from: start to: stop copy: aValue
	" Returns an instance of the receiver with value aValue. "

	^(self from: start to: stop)
		binding: (S8CopyBinding value: aValue);
		yourself ! !

! ParseLiteral class methodsFor: #instantiation !
from: start to: stop value: aValue
	" Returns an instance of the receiver with value aValue. "

	^(self from: start to: stop)
		binding: (S8LiteralBinding value: aValue);
		yourself ! !

! ParseLiteral class methodsFor: #instantiation !
sourceRange: interval value: aValue
	" Returns an instance of the receiver with value aValue. "

	^(self sourceRange: interval)
		binding: (S8LiteralBinding value: aValue);
		yourself ! !


! ParseInstanceVariable class methodsFor: #instantiation !
name: aName index: index
	" Returns an instance of the receiver with name aName. "

	^self new
		name: aName;
		index: index;
		yourself ! !

! ParseInstanceVariable methodsFor: #accessing !
index
	" Returns the receiver's index. "

	^index ! !

! ParseInstanceVariable methodsFor: #accessing !
index: anIndex
	" Set the receiver's index. "

	index := anIndex ! !

! ParseInstanceVariable methodsFor: #accessing !
name
	" Returns the receiver's name. "

	^name ! !

! ParseInstanceVariable methodsFor: #accessing !
name: aName
	" Set the receiver's name. "

	name := aName ! !

! ParseInstanceVariable methodsFor: #accessing !
instVarKey
	" Private - Returns the instance access index of the receiver. "

	^self name "
		was: index
		Note: S8 has better access to named i.v.
	" ! !

! ParseInstanceVariable methodsFor: #binding !
bind: aValue in: aContext
	" Set the receiver's value to aValue in aContext.
	Check if the receiver is assignable.
	WARNING: The receiver MUST return the set value. "

	self canBeAssigned ifFalse: [
		^self error: 'Invalid assignment.'
	].
	^aContext self
		instVarAt: self instVarKey
		put: aValue ! !

! ParseInstanceVariable methodsFor: #value !
valueIn: aContext
	" Private - Returns the value of the receiver in aContext. "

	^aContext self instVarAt: self instVarKey ! !

! ParseInstanceVariable methodsFor: #query !
canBeAssigned
	" Returns true if the receiver is a variable that can be assigned. "

	^true ! !

! ParseUniversal methodsFor: #query !
canBeAssigned
	" Returns true if the receiver is a variable that can be assigned. "

	^indices size > 0 ! !

! ParseUniversal methodsFor: #evaluating !
assignValue: aValue in: aContext
	"Private - Set the receiver's value to aValue.
	WARNING: The receiver MUST return the value set. "

	| target |
	target := self valueIn: aContext.
	target := self follow: target in: aContext.
	^target first basicAt: target last put: aValue ! !

! ParseSelf methodsFor: #value !
valueIn: aContext
	" Private - Returns the value of the receiver in aContext. "

	^aContext self ! !

! ParseSuper class methodsFor: #instantiation !
in: aClass
	" Returns an instance of the receiver. "

	^self new initialize: aClass! !

! ParseSuper methodsFor: #initialize !
initialize: aClass
	" Private - Initialize the receiver. "

	implementor := aClass! !

! ParseSuper methodsFor: #query !
isSuper
	" Returns true if the receiver is a reference to super. "

	^true ! !

! ParseSuper methodsFor: #evaluating !
functionToCall: selector in: anImplementor
	" Private - Returns the native function to call for selector. "

	^Smalltalk current
		#superImplementor: anImplementor
		of: selector ! !

! ParseSuper methodsFor: #evaluating !
apply: fn to: anObject withArguments: anArray in: aContext
	" Private - Returns the result of evaluating the native function in anObject @ aContext. "

	^{''
	js'fn.apply(anObject,anArray)'
	lua'fn(anObject,(table.unpack or unpack)(anArray))'
	py'fn(anObject,*smalltalk.objects_toCall_(anArray,fn))'}. ! !

! ParseSuper methodsFor: #evaluating !
send: selector to: anObject withArguments: anArray in: aContext
	" Private - Returns the result of sending message to anObject in aContext. "

	| anImplementor |
	anImplementor := implementor isNil
		ifTrue: [ anObject class ]
		ifFalse: [ implementor ].
	^self	apply: (self functionToCall: selector in: anImplementor)
		to: anObject withArguments: anArray
		in: aContext ! !

! ParseNoReceiver methodsFor: #query !
isOO
	" Returns true if the receiver refers to an object. "

	^false ! !

! ParseNoReceiver methodsFor: #value !
valueIn: aContext
	" Private - Returns the value of the receiver in aContext. "

	^aContext ! !

! ParseNoReceiver methodsFor: #evaluating !
send: selector to: anObject withArguments: anArray in: aContext
	" Private - Returns the result of sending message to anObject in aContext. "

	| fn |
	selector = #self ifTrue: [ "note that anObject is the context of evaluation"
		^anObject self evaluateWithArguments: anArray
	].
	fn := Smalltalk nativeObjectAt: selector.
	^fn evaluateWithArguments: anArray ! !

! ParseNativeVariable class methodsFor: #instantiation !
named: aName from: start to: stop
	" Returns an instance of the receiver. "

	^(self from: start to: stop)
		binding: aName -> nil;
		yourself ! !

! ParseNativeVariable methodsFor: #value !
valueIn: aContext
	" Private - Returns the value of the receiver in aContext. "

	^Smalltalk global basicAt: self name ! !

! ParseNativeVariable methodsFor: #query !
canBeAssigned
	" Returns true if the receiver is a variable that can be assigned. "

	^true ! !

! ParseNativeVariable methodsFor: #evaluating !
assignValue: aValue in: aContext
	"Private - Set the receiver's value to aValue.
	WARNING: The receiver MUST return the value set. "

	| target |
	target := Smalltalk global.
	indices isNil ifTrue: [ ^target basicAt: self name put: aValue ].
	target := target basicAt: self name ifAbsent: [
		^self error: 'Native Global not found - ' ,self name
	].
	target := self follow: target in: aContext.
	^target first basicAt: target last put: aValue ! !


! MethodExitNode methodsFor: #accessing !
exitBlock
	" Returns the receiver's exitBlock. "

	^exitBlock ! !

! MethodExitNode methodsFor: #accessing !
exitBlock: aBlock
	" Set the receiver's exitBlock. "

	super exitBlock: aBlock.
	exitBlock := aBlock.! !

! MethodExitNode methodsFor: #evaluating !
exitWith: anObject
	" Exit the receiver with anObject as returned value. "

	exitBlock isNil ifTrue: [ ^self error: 'Invalid script context.' ].
	^exitBlock value: anObject ! !

! MethodReturnNode methodsFor: #accessing !
expression
	" Returns the receiver's expression. "

	^expression ! !

! MethodReturnNode methodsFor: #accessing !
expression: anExpressionNode
	" Set the receiver's expression. "

	expression := anExpressionNode ! !

! MethodReturnNode methodsFor: #accessing !
exitBlock: aBlock
	" Set the receiver's exitBlock. "

	super exitBlock: aBlock.
	expression exitBlock: aBlock ! !

! MethodReturnNode methodsFor: #query !
isReturn
	" Returns true if the receiver is a return parse node. "

	^true ! !

! MethodReturnNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	^self exitWith: (self expression evaluateIn: aContext) ! !

! PrimitiveNode class methodsFor: #instantiation !
doing: anAction
	" Returns an instance of the receiver. "

	^self new
		action: anAction;
		yourself ! !

! PrimitiveNode methodsFor: #query !
isPrimitiveNode
	" Returns true if the receiver is a primitive. "

	^true ! !

! PrimitiveNode methodsFor: #accessing !
action: anAction
	" Set the receiver's action. "

	action := anAction ! !

! PrimitiveNode methodsFor: #evaluating !
evaluateIn: aContext
	" Returns the result of evaluating the receiver in aContext. "

	action isNil ifTrue: [ ^self error: 'Action to evaluate missing.' ].
	^action evaluateWithArguments: (Array with: aContext with: self) ! !


! ScriptParser class methodsFor: #instantiation !
for: aReceiver on: aScope
	" OBSOLETE - Returns an instance of the receiver for evaluating on objects like aReceiver on given scope. "

	^self new
		scope: aScope;
		receiver: aReceiver;
		yourself ! !

! ScriptParser class methodsFor: #instantiation !
for: aClass
	" Returns an instance of the receiver to parse in aClass. "

	^self new
		currentClass: aClass;
		yourself ! !

! ScriptParser class methodsFor: #instantiation !
on: aScope
	" Returns an instance of the receiver to parse on given scope. "

	^self new
		scope: aScope;
		yourself ! !

ScriptParser class buildGetters: #(
	#(primitiveHandlers initialize PoolDictionary)
)!
ScriptParser class buildAccessors: #(
	defaultPrimitiveHandler
)!

! ScriptParser class methodsFor: #parsing !
handlerForPrimitive: type
	" Private - Returns a handler to evaluate for parsing a primitive of specified type. "

	^type notNil ifTrue: [
		self primitiveHandlers at: type ifAbsent: [
			self defaultPrimitiveHandler
		]
	] ! !

! ScriptParser class methodsFor: #parsing !
parsePrimitive: aParser in: aContext
	" Private - Parse a primitive in aContext.
	Returns nil if failed. "

	| handler type |
	type := aParser nextSelectorOrNil.
	handler := self handlerForPrimitive: type.
	^handler notNil ifTrue: [
		handler evaluateWithArguments:
			(Array with: aParser with: aContext with: type)
	] ! !

! ScriptParser methodsFor: #parsing !
expression: aContext
	" Private - Parse the receiver looking for an expression using variables in aContext. Returns nil if failed.
	<rule> expression = primitive | {variableName ':='} (primary | messageExpression {';' cascadeMessage})
	"

	| variable expression aReceiver start |
	start := self startOfNextToken.
	(self match: self primitiveStart) ifTrue: [
		^self nextPrimitive: aContext
	].

	variable := self nextVariable: aContext.
	(variable notNil and: [ self match: self leftArrow ])
	ifTrue: [ ^self assignment: variable in: aContext startPosition: start ].

	self position: start.
	aReceiver := self primaryExpression: aContext.
	self atEnd ifTrue: [ ^aReceiver ].
	expression := self message3: aReceiver repeat: true in: aContext.
	expression isNil ifTrue: [ ^aReceiver ].
	(self match: self semicolon) ifTrue: [
		expression := self cascade: expression to: aReceiver in: aContext
	].
	^expression! !

! ScriptParser methodsFor: #accessing !
position
	" Returns the current position. "

	| result |
	result := source position.
	^result isNil ifTrue: [0] ifFalse: [result]! !

! ScriptParser methodsFor: #accessing !
position: index
	" Set the current position. "

	^source position: index! !

! ScriptParser methodsFor: #accessing !
currentClass
	" Return the current class of the receiver. "

	currentClass isNil ifTrue: [ self currentClass: nil ].
	^currentClass! !

! ScriptParser methodsFor: #accessing !
currentClass: aClass
	" Set the current class of the receiver. "

	| index |
	currentClass := aClass isNil ifTrue: [ nil class ] ifFalse: [ aClass ].
	index := 0.
	self instanceVariables: (currentClass allInstVarNames collect: [:each|
		index := index + 1.
		ParseInstanceVariable name: each index: index.
	])! !

! ScriptParser methodsFor: #accessing !
receiver: anObject
	" Set the receiver to parse considering the receiver will be a member of anObject species when evaluated. "

	self currentClass: anObject class ! !

! ScriptParser methodsFor: #parsing !
parseTemporaries: aContext
	" Private - Read temporaries from the receiver's stream.
	<rule> temporaries = [ '|' (variable) ] (variable, ..., variable)*
	"

	| vars word start temporary |
	(self match: self verticalBar) ifTrue: [ ^#() ].
	vars := Array new.
	[	start := self startOfNextToken.
		word := self nextWord.
		word notNil
	] whileTrue: [
		self position: start.
		temporary := self nextArgument: aContext.
		temporary isNil ifTrue: [ ^self expected: 'local name' ].
		temporary canBeAssigned: true.
		vars addLast: temporary.
	].
	(self match: self verticalBar) ifTrue: [ ^vars ].
	^self expected: self verticalBar! !

! ScriptParser methodsFor: #parsing !
assignment: aVariable in: aContext
	" Private - Get the assignment from the receiver.
	<rule> expression = {variableName ':='} (primary | messageExpression {';' cascadeMessage} ).
	 "

	| expression |
	aVariable canBeAssigned ifFalse: [ ^self syntaxError: 'Cant assign to this variable.' ].
	expression := self expression: aContext.
	expression isNil ifTrue: [ ^self expected: #expression ].
	^AssignNode new
		variable: aVariable;
		expression: expression;
		yourself! !

! ScriptParser methodsFor: #parsing !
assignment: variable in: aContext startPosition: startPosition
	" Private - Get the an assignement from the receiver. "

	| assignment |
	assignment := self assignment: variable in: aContext.
	assignment sourceRange: (startPosition to: self position).
	^assignment! !

! ScriptParser methodsFor: #initialize !
initializeInstanceVariables
	" Private - Initializes the receiver's instanceVariables. "

	self instanceVariables: Array new! !

! ScriptParser methodsFor: #initialize !
initializeScope
	" Private - Initializes the receiver's scope. "

	self scope: ParseContext default! !

! ScriptParser methodsFor: #query !
atEnd
	" Returns true when the receiver is atEnd. "

	self skipSeparators.
	^source atEnd! !

! ScriptParser methodsFor: #query !
isBinarySelector: aSelector
	" Private - Returns true if aSelector is a keyword selector. "

	aSelector isEmpty ifTrue: [ ^false ].
	aSelector
		detect: [:one| (self isSelectorCharacter: one) not ]
		ifNone: [ ^true ].
	^false! !

! ScriptParser methodsFor: #query !
isKeywordSelector: aSelector
	" Private - Returns true if aSelector is a keyword selector. "

	^aSelector last = $:! !

! ScriptParser methodsFor: #query !
isAlpha: aCharacter
	" Private - Returns true if aCharacter is good in a word. "

	^aCharacter isAlphaNumeric or: [ '_' includes: aCharacter ]! !

! ScriptParser methodsFor: #constants !
binarySelectorChars
	" Return the binary selector characters mask. "

	^'|&+*/=><,@%~-?!!\' ! !

! ScriptParser methodsFor: #query !
isSelectorCharacter: aCharacter
	" Private - Returns true if aCharacter is a selector character.
	<rule> selectorCharacter = ',' | '+' | '-' | '/' | '\' | '*' | '~' | '>' | '<' | '=' | '@' | '%' | '|' | '&' | '?' | '!!' .
	"

	^self binarySelectorChars includes: aCharacter! !

! ScriptParser methodsFor: #query !
isUnarySelector: aSelector
	" Private - Returns true if aSelector is a unary selector. "

	aSelector isEmpty ifTrue: [ ^false ].
"	aSelector first isLetter ifFalse: [ ^false ]."
	aSelector
		detect: [:one| (self isAlpha: one) not ]
		ifNone: [ ^true ].
	^false! !

! ScriptParser methodsFor: #parsing !
arrayLiteralElement: aContext
	" Private - Gets the next array element from the receiver. "

	^self constant: aContext ! !

! ScriptParser methodsFor: #parsing !
basicPeek
	"Peek the next character from source stream. "

	source peek = $" ifTrue: [ "skip coments"
		source next;upTo: $"
	].
	^source peek! !

! ScriptParser methodsFor: #parsing !
cascade: expression to: aReceiver in: aContext
	" Private - Parse the receiver looking for a cascade expression using variables in aContext.
	<rule> cascadeMessage = unaryMessage | binaryMessage | keywordMessage.
	"

	| start cascade message |
	start := self position.
	cascade := CascadedMessageNode for: expression.
	[	message := self message3: aReceiver repeat: false in: aContext.
		message isNil ifTrue: [ ^self expected: #message ].
		cascade add: message.
		self match: self semicolon ] whileTrue: [].
	cascade sourceRange: (start to: self position).
	^cascade! !

! ScriptParser methodsFor: #parsing !
checkForOldSyntax: radix with: value on: aStream
	" Private - Answer aNumber as described on the stream, aStream. "

	(aStream atEnd not and: [aStream peekFor: $.])
	ifTrue: [
		(aStream atEnd not and: [aStream peek digitValue between: 0 and: radix-1])
		ifTrue: [ ^self syntaxError: 'Non-radix 10 numbers may not contain decimal points' ]
		ifFalse: [ aStream skip: -1 ]
	].
	(aStream atEnd not and: [aStream peek asLowercase = $e])
	ifTrue: [ ^self syntaxError: 'Non-radix 10 numbers may not have an exponent' ].
	^value! !

! ScriptParser methodsFor: #constants !
colon
	" Private - Returns the colon token. "

	^$:! !

! ScriptParser methodsFor: #parsing !
nextDashOrNumber
	" Private - Returns the next Number or #- "

	| result |
	(source peek = $-) ifFalse: [ ^self nextNumber ].
	result := source next.
	^(source atEnd not and: [source peek isDigit])
		ifTrue: [ self nextNumber negated ]
		ifFalse: [ result ]! !

! ScriptParser methodsFor: #parsing !
constant: aContext
	" Private - Recognize a constant (named, number, character, string, or quoted).
	Returns the constant value. "

	self skipSeparators.
	(self match: #false) ifTrue: [ ^false ].
	(self match: #true) ifTrue: [ ^true ].
	(self match: #nil) ifTrue: [ ^nil ].
	(self match: self singleQuote) ifTrue: [ ^self nextString ].
	(self peekAny: '0123456789-' advance: false) notNil ifTrue: [ ^self nextDashOrNumber ].
	(self match: self dollar) ifTrue: [ ^self nextCharacter ].
	(self match: (Array with: self literalQuote with: self leftParenthesis)) ifTrue: [ ^self nextArray: aContext ].
	(self match: self literalQuote) ifTrue: [ ^self nextSymbol ].
	(self match: self leftParenthesis) ifTrue: [ ^self nextArray: aContext ].
	(self match: self leftBracket) ifTrue: [ ^self nextByteArray: aContext ].
	(self match: self leftBrace) ifTrue: [ ^self nextBraceLiteral: aContext ].
	^self nextSymbol ! !

! ScriptParser methodsFor: #constants !
dollar
	" Private - Returns the dollar token. "

	^$$! !

! ScriptParser methodsFor: #error !
syntaxError: aString
	" A Syntax error has been occurred at current position in source. "

	| syntaxError |
	syntaxError := SyntaxError description: aString.
	syntaxError
		source: source contents;
		position: self position + 1;
		yourself.
	^errorBlock value: syntaxError! !

! ScriptParser methodsFor: #error !
expected: aString
	"An expected element can't be found at current position in source. "

	^self syntaxError: aString ,' expected'! !

! ScriptParser methodsFor: #accessing !
instanceVariables
	" Returns the receiver's instanceVariables. "

	instanceVariables isNil ifTrue: [ self initializeInstanceVariables ].
	^instanceVariables! !

! ScriptParser methodsFor: #accessing !
instanceVariables: aCollection
	" Set the receiver's instanceVariables. "

	instanceVariables := aCollection! !

! ScriptParser methodsFor: #constants !
leftArrow
	" Private - Returns the leftArrow token. "

	^':='! !

! ScriptParser methodsFor: #constants !
period
	" Private - Returns the period token. "

	^$.! !

! ScriptParser methodsFor: #constants !
leftBrace
	" Private - Returns the leftBrace token. "

	^${! !

! ScriptParser methodsFor: #constants !
leftBracket
	" Private - Returns the leftBracket token. "

	^$[! !

! ScriptParser methodsFor: #constants !
leftParenthesis
	" Private - Returns the leftParentesis token. "

	^$(! !

! ScriptParser methodsFor: #constants !
literalQuote
	" Private - Returns the literal quote token. "

	^$#! !

! ScriptParser methodsFor: #constants !
primitiveStart
	" Private - Returns the start primitive token. "

	^$<! !

! ScriptParser methodsFor: #constants !
primitiveEnd
	" Private - Returns the end primitive token. "

	^$>! !

! ScriptParser methodsFor: #parsing !
match: token
	" Try to match token.
	token can be a single token or a collection of tokens.
	Returns true if success and advance the receiver's stream.
	If failed, restore the receiver's position. "

	| position |
	self atEnd ifTrue: [^false].
	position := self position.
	token isCharacter ifTrue: [
		self peek = token
		ifTrue: [ self next.^true ]
	] ifFalse: [
		(token isMemberOf: BlockClosure) ifTrue: [
			token argumentCount = 0
			ifTrue: [ token value ifTrue: [ ^true ] ]
			ifFalse: [ (token value: self next) ifTrue: [ ^true ] ]
		] ifFalse: [
			token
				detect: [:one| (self match: one) not ]
				ifNone: [ ^true ].
		].
	].
	self position: position.
	^false! !

! ScriptParser methodsFor: #parsing !
message1: aNode repeat: repeat in: aContext
	" Private - Parse the receiver looking for a message expression using variables in aContext.
	<rule> messageExpression = unaryExpression | binaryExpression | keywordExpression.
	<rule> unaryExpression = primaryMessage {unaryMessage}.
	<rule> binaryExpression = (unaryExpression | primary) binaryMessage {binaryMessage}.
	<rule> keywordExpression = (binaryExpression | primary) keywordMessage.
	"

	| aReceiver start selector params argument part partStart |
	aReceiver := aNode.
	[
		start := self startOfNextToken.
		selector := self nextSelectorOrNil.
		selector isNil ifTrue: [ ^aReceiver ].
		(self isUnarySelector: selector) "simple message"
		ifTrue: [ params := #() ]
		ifFalse: [
			aReceiver isSuper ifTrue: [ ^self expected: #message ].
			self position: start.
			^aReceiver
		].
		aReceiver isNil ifTrue: [ ^self expected: 'Receiver or name' ].
		aReceiver := (MessageNode from: start to: self position)
			receiver: aReceiver
			selector: (self newSelectorName: selector)
			arguments: params.
	repeat ] whileTrue: [].
	^aReceiver! !

! ScriptParser methodsFor: #parsing !
message2: aNode repeat: repeat in: aContext
	" Private - Parse the receiver looking for a message expression using variables in aContext.
	<rule> messageExpression = unaryExpression | binaryExpression | keywordExpression.
	<rule> unaryExpression = primaryMessage {unaryMessage}.
	<rule> binaryExpression = (unaryExpression | primary) binaryMessage {binaryMessage}.
	<rule> keywordExpression = (binaryExpression | primary) keywordMessage.
	"

	| aReceiver start selector params argument part partStart |
	aReceiver := aNode.
	[
		start := self startOfNextToken.
		selector := self nextSelectorOrNil.
		selector isNil ifTrue: [ ^aReceiver ].
		(selector = #| or: [self isBinarySelector: selector]) "level>=2"
		ifTrue: [
			argument := self primaryExpression: aContext.
			argument isNil ifTrue: [ ^self expected: #argument ].
			argument := self message1: argument repeat: true in: aContext.
			argument isNil ifTrue: [ ^self expected: #argument ].
			params := Array with: argument.
		]
		ifFalse: [
			(self isUnarySelector: selector) "simple message"
			ifTrue: [ params := #() ]
			ifFalse: [
				aReceiver isSuper ifTrue: [ ^self expected: #message ].
				self position: start.
				^aReceiver
			]
		].
		aReceiver isNil ifTrue: [ ^self syntaxError: 'Invalid receiver or invalid name' ].
		aReceiver := (MessageNode from: start to: self position)
			receiver: aReceiver
			selector: (self newSelectorName: selector)
			arguments: params.
	repeat ] whileTrue: [].
	^aReceiver! !

! ScriptParser methodsFor: #parsing !
message3: aNode repeat: repeat in: aContext
	" Private - Parse the receiver looking for a message expression using variables in aContext.
	<rule> messageExpression = unaryExpression | binaryExpression | keywordExpression.
	<rule> unaryExpression = primaryMessage {unaryMessage}.
	<rule> binaryExpression = (unaryExpression | primary) binaryMessage {binaryMessage}.
	<rule> keywordExpression = (binaryExpression | primary) keywordMessage.
	"

	| aReceiver start selector params argument part partStart isNative |
	aReceiver := aNode.
	[	start := self startOfNextToken.
		isNative := self match: self literalQuote.
		selector := self nextSelectorOrNil.
		selector isNil ifTrue: [ ^aReceiver ].
		(self isKeywordSelector: selector) "level>=3" ifTrue: [
			argument := self primaryExpression: aContext.
			argument isNil ifTrue: [ ^self expected: #argument ].
			argument := self message2: argument repeat: true in: aContext.
			argument isNil ifTrue: [ ^self expected: #argument ].
			params := Array with: argument.
			[ partStart := self startOfNextToken.
			part := self nextSelectorOrNil.
			part notNil and: [ self isKeywordSelector: part ]
			] whileTrue: [
				selector := selector , part.
				argument := self primaryExpression: aContext.
				argument isNil ifTrue: [ ^self expected: #argument ].
				argument := self message2: argument repeat: true in: aContext.
				argument isNil ifTrue: [ ^self expected: #argument ].
				params := params copyWith: argument.
			].
			self position: partStart.
		] ifFalse: [
			(selector = #| or: [self isBinarySelector: selector]) "level>=2" ifTrue: [
				argument := self primaryExpression: aContext.
				argument isNil ifTrue: [ ^self expected: #argument ].
				argument := self message1: argument repeat: true in: aContext.
				argument isNil ifTrue: [ ^self expected: #argument ].
				params := Array with: argument.
			] ifFalse: [
				(self isUnarySelector: selector) "simple message"
				ifTrue: [ params := #() ]
				ifFalse: [
					aReceiver isSuper ifTrue: [ ^self expected: #message ].
					self position: start.
					^aReceiver
				]
			]
		].
		aReceiver isNil ifTrue: [ ^self syntaxError: 'Invalid receiver or invalid name' ].
		aReceiver := (MessageNode native: isNative from: start to: self position)
			receiver: aReceiver
			selector: (self newSelectorName: selector)
			arguments: params.
	repeat ] whileTrue: [].
	^aReceiver! !

! ScriptParser methodsFor: #parsing !
bodyScope: statements
	" Private - Returns a context for body of script.
	<rule> script = [ | temporaries ] statements.
	"

	| result |
	(self match: self verticalBar) ifFalse: [ ^self scope ].
	result := ParseContext in: self scope.
	statements temporaries: (self parseTemporaries: result).
	^result ! !

! ScriptParser methodsFor: #parsing !
mustHaveA: aCharacter
	" Private - Check that the next character is aCharacter.
	Read the character if success. Generates a syntax error in absent. "

	self peek = aCharacter ifFalse: [
		^self syntaxError: aCharacter printString,' expected.'
	].
	self atEnd ifFalse: [ self next ]! !

! ScriptParser methodsFor: #parsing !
newSelectorName: aSelectorName
	" Private - Build a new message name from aSelectorName. "

	^aSelectorName asSymbol ! !

! ScriptParser methodsFor: #parsing !
next
	"Advance to the next character onto source stream. "

	self skipSeparators.
	^source next! !

! ScriptParser methodsFor: #parsing !
nextArgument: aContext
	" Private - Parse the receiver looking for a new argument using variables in aContext.
	Returns nil if failed. "

	| start word assoc |
	start := self startOfNextToken.
	word := self nextWord.
	word isNil ifTrue: [ ^nil ].
	aContext at: word put: nil.
	assoc := aContext bindingAt: word.
	^(ParseArgument
		from: start to: self position)
		binding: assoc;
		yourself! !

! ScriptParser methodsFor: #parsing !
nextArray: aContext
	" Private - Parse the receiver looking for an Array using variables in aContext.
	Returns anArray or generates an error if failed.
	<rule> array = '(' {number | string | symbol | array | characterConstant} ')'.
	"

	| stream |
	stream := Array new writeStream.
	[ self match: self rightParenthesis ] whileFalse: [
		stream nextPut: (self arrayLiteralElement: aContext)
	].
	^stream contents! !

! ScriptParser methodsFor: #parsing !
nextArrayConstant: aContext
	" Private - Parse the receiver looking for an Array using variables in aContext.
	Returns nil if failed.
	<rule> arrayConstant = '#' array.
	<rule> array = '(' {number | string | symbol | array | characterConstant} ')'.
	"

	| start value |
	start := self startOfNextToken.
	(self match: (Array with: self literalQuote with: self leftParenthesis)) ifFalse: [ ^nil ].
	value := self nextArray: aContext.
	value isNil ifTrue: [ self position: start. ^nil ].
	^ParseLiteral
		from: start to: self position
		copy: value ! !

! ScriptParser methodsFor: #parsing !
nextBlock: aContext
	" Private - Parse the receiver looking for a block using variables in aContext.
	Returns nil if failed.
	<node> $[ ($: arg)* temporaries statements $] => BlockClosureNode
	"

	| start args block peek |
	start := self position - 1.
	args := #().
	[ self match: self colon ] whileTrue: [ | arg |
		arg := self nextArgument: aContext.
		arg isNil ifTrue: [ ^self expected: #argument ].
		args add: arg.
	].
	args size > 0 ifTrue: [
		peek := self peek.
		(peek = self rightBracket or: [ peek = self verticalBar ])
		ifFalse: [ ^self expected: self verticalBar ].
		peek = self verticalBar ifTrue: [ self skip: self verticalBar ].
	].
	block := self statementsArgs: args in: aContext.
	(self match: self rightBracket) ifFalse: [
		^self expected: 'Period or right bracket'
	].
	block sourcePosition: (start to: self position).
	^block! !

! ScriptParser methodsFor: #parsing !
nextPrimitive: aContext
	" Private - Parse the receiver looking for primitive in aContext.
	Returns nil if failed.
	<node> $< ...prim... $> => PrimitiveNode
	"

	| start result |
	start := self position - 1.
	self skipSeparators.
	result := self class parsePrimitive: self in: aContext.
	result isNil ifTrue: [ ^nil ].
	(self match: self primitiveEnd) ifFalse: [
		^self expected: 'End of primitive'
	].
	result sourcePosition: (start to: self position).
	^result! !

! ScriptParser methodsFor: #parsing !
nextBraceLiteral: aContext
	" Private - Parse the receiver looking for a brace literal in aContext.
	Returns the literal or generates an error if failed.
	"

	^self notImplementedYet: #nextBraceLiteral: ! !

! ScriptParser methodsFor: #inlines !
nextInline: aContext
	" Private - Returns the next inline in the receiver's stream (or nil). "

	| start lang code codeStart |
	start := self startOfNextToken.
	lang := self nextSelectorOrNil.
	codeStart := self startOfNextToken.
	code := self nextStringConstant: aContext.
	^code notNil ifTrue: [
		Array	with: lang with: codeStart
			with: code binding source
	] ! !

! ScriptParser methodsFor: #inlines !
inlineStatements: pool in: aContext
	" Private - Returns a statements sequence parsed from pool (or nil). "

	| inline |
	inline := pool at: #s8 ifAbsent: [
		pool at: #s8 ,$_ ,self nativeLanguage ifAbsent: [ ^nil ]
	].
	^self	inlineStatements: inline last
		startingAt: inline first + 1
		in: aContext! !

! ScriptParser methodsFor: #inlines !
inlineStatements: code startingAt: start in: aContext
	" Private - Returns a statements sequence parsed on code (or nil). "

	| src result |
	src := source.
	[	source := ((source contents copyFrom: 1 to: start), code ,$]) readStream.
		source position: start.
		result := (MessageNode from: start to: self position)
			receiver: (self statementsArgs: #() in: aContext)
			selector: #value arguments: #()
	] ensure: [ source := src ].
	^result! !

! ScriptParser methodsFor: #parsing !
nextBraceExpression: aContext
	" Private - Parse the receiver looking for a brace expression in aContext.
	Returns nil if failed.
	<node> ${ ({lang} string)* $} => statements
	"

	| start pool block |
	start := self position.
	pool := PoolDictionary new.
	[ self match: self rightBrace ] whileFalse: [ | inline |
		inline := self nextInline: aContext.
		inline notNil ifTrue: [
			pool at: inline removeFirst put: inline
		].
	].
	pool isEmpty ifTrue: [
		self position: start.
		^self expected: 'Inline code'
	].
	block := self inlineStatements: pool in: aContext.
	block isNil ifTrue: [
		self position: start.
		^self expected: 'Code for s8 language'
	].
	block sourcePosition: (start - 1 to: self position).
	^block! !

! ScriptParser methodsFor: #parsing !
nextByteArray: aContext
	" Private - Parse the receiver looking for aByteArray using variables in aContext.
	Returns aByteArray or generates an error if failed.
	"

	^self notImplementedYet: #nextByteArray: ! !

! ScriptParser methodsFor: #parsing !
nextCharacter
	" Private - Read aCharacter from the receiver. "

	^source atEnd ifTrue: [ $  ] ifFalse: [ source next ]! !

! ScriptParser methodsFor: #parsing !
nextCharacterConstant: aContext
	" Private - Parse the receiver looking for aCharacter using variables in aContext.
	Returns nil if failed. "

	| start |
	start := self startOfNextToken.
	(self match: self dollar) ifFalse: [ ^nil ].
	^ParseLiteral
		from: start to: self position
		value: self nextCharacter ! !

! ScriptParser methodsFor: #parsing !
nextLiteral: aContext
	" Private - Parse the receiver looking for a literal using variables in aContext.
	Returns nil if failed.
	<rule> literal = number | string | characterConstant | symbolConstant | arrayConstant.
	"

	| node |
	node := self nextNumberConstant: aContext.
	node notNil ifTrue: [ ^node ].
	node := self nextStringConstant: aContext.
	node notNil ifTrue: [ ^node ].
	node := self nextCharacterConstant: aContext.
	node notNil ifTrue: [ ^node ].
	node := self nextArrayConstant: aContext.
	node notNil ifTrue: [ ^node ].
	node := self nextSymbolConstant: aContext.
	node notNil ifTrue: [ ^node ].
	^nil! !

! ScriptParser methodsFor: #parsing !
nextNumber
	" Private - Parse the receiver looking for a number using variables in aContext.
	Returns aNumber or generates an error if failed.
	<rule> number = [digits 'r'] ['-'] bigDigits ['.'bigDigits] ['e' ['-'] digits].
	<rule> digits = digit {digit}.
	<rule> Digit = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'.
	<rule> bigDigits = bigDigit {bigDigit}.
	<rule> bigDigit = digit | capitalLetter.
	"

	^self readSmalltalkNumberFrom: source! !

! ScriptParser methodsFor: #parsing !
nextNumberConstant: aContext
	" Private - Parse the receiver looking for a number using variables in aContext.
	Returns nil if failed.
	<rule> number = [digits 'r'] ['-'] bigDigits ['.'bigDigits] ['e' ['-'] digits].
	<rule> digits = digit {digit}.
	<rule> Digit = '0' | '1' | '2' | '3' | '4' | '5' | '6' | '7' | '8' | '9'.
	<rule> bigDigits = bigDigit {bigDigit}.
	<rule> bigDigit = digit | capitalLetter.
	"

	| start value |
	start := self startOfNextToken.
	(self peekAny: '0123456789-' advance: false) isNil ifTrue: [ ^nil ].
	value := self nextNumber.
	value isNil ifTrue: [ self position: start. ^nil ].
	^ParseLiteral
		from: start to: self position
		value: value ! !

! ScriptParser methodsFor: #parsing !
nextSelector
	" Private - Returns the next selector in the receiver.
	Generates an error if failed.
		<rule> symbol = unarySelector | binarySelector | keyword {keyword}.
		<rule> unarySelector = identifier.
		<rule> binarySelector = selectorCharacter [selectorCharacter] | '-'.
		<rule> keyword = identifier ':'.
	"

	| result |
	result := self nextSelectorOrNil.
	result isNil ifTrue: [ ^self expected: #selector ].
	^result! !

! ScriptParser methodsFor: #parsing !
nextSelectorOrNil
	" Private - Returns the next selector in the receiver.
	Returns nil if failed.
		<rule> symbol = unarySelector | binarySelector | keyword {keyword}.
		<rule> unarySelector = identifier.
		<rule> binarySelector = selectorCharacter [selectorCharacter] | '-'.
		<rule> keyword = identifier ':'.
	"

	| contents peek |
	peek := self basicPeek.
	(peek isNil or: [ peek isSeparator ]) ifTrue: [ ^nil ].

	(peek = $- or: [ self isSelectorCharacter: peek ]) "binarySelector"
	ifTrue: [ ^self nextSelectorWith: peek ].

	(self isAlpha: peek) ifFalse: [ ^nil ].
	contents := String new.
	[	contents := contents, self nextWord.
		self basicPeek = self colon ] whileTrue: [
		contents := contents ,self colon.
		self skip: self colon.
		peek := self basicPeek.
		(peek isNil or: [ (self isAlpha: peek) not ]) ifTrue: [ ^contents ]
	].
	^contents asSymbol! !

! ScriptParser methodsFor: #parsing !
nextSelectorWith: start
	" Private - Returns the next selector starting with start character. "

	| result peek |
	result := start.
	[	self next.
		peek := self basicPeek.
		self isSelectorCharacter: peek
	] whileTrue: [
		result := result , peek asString
	].
	^result! !

! ScriptParser methodsFor: #parsing !
nextString
	" Private - Returns the next String contents from the receiver. "

	| contents |
	contents := String new.
	[	contents := contents , (source upTo: self singleQuote).
		source peek = self singleQuote
	] whileTrue: [
		contents := contents copyWith: $'.
		source next
	].
	^contents! !

! ScriptParser methodsFor: #parsing !
nextStringConstant: aContext
	" Private - Parse the receiver looking for a string using variables in aContext.
	Returns nil if failed. "

	| start contents |
	start := self startOfNextToken.
	(self match: self singleQuote) ifFalse: [ ^nil ].
	^ParseLiteral
		from: start to: self position
		value: self nextString ! !

! ScriptParser methodsFor: #parsing !
nextSymbol
	" Private - Returns the characters in next Symbol (or nil). "

	| c result |
	result := String new.
	[	c := source peek.
		c isNil ifTrue: [
			^result notEmpty ifTrue: [ result ]
		].
		(c between: $a and: $z)
		or: [ (c between: $A and: $Z)
		or: [ (c between: $0 and: $9)
		or: [ (self binarySelectorChars includes: c)
		or: [ '_:' includes: c ]]]]
	] whileTrue: [
		result := result ,source next
	].
	^result notEmpty ifTrue: [ result ] ! !

! ScriptParser methodsFor: #parsing !
nextNativeConstant: aContext
	" Private - Parse the receiver looking for a name of native constant in aContext.
	Returns nil if failed.
		<rule> nativeConstant = '{' word '}'
	"

	| start aName |
	start := self startOfNextToken.
	(self match: self leftBrace) ifFalse: [ ^nil ].
	aName := self nextWord.
	(self match: self rightBrace) ifFalse: [ ^self expected: self rightBrace ].
	^ParseNativeVariable named: aName from: start to: self position ! !

! ScriptParser methodsFor: #parsing !
nextSymbolConstant: aContext
	" Private - Parse the receiver looking for aSymbol using variables in aContext.
	Returns nil if failed.
		<rule> symbolConstant = '#' string | symbolChars | ${ nativeConstant
	"

	| start value |
	start := self startOfNextToken.
	(self match: self literalQuote) ifFalse: [ ^nil ].
	source peek = self leftBrace ifTrue: [ ^self nextNativeConstant: aContext ].
	value := source peek = self singleQuote
		ifTrue: [ source next. self nextString ]
		ifFalse: [ self nextSymbol ].
	value isNil ifTrue: [
		^ParseNoReceiver
			from: start to: self position
	].
	^ParseLiteral
		from: start to: self position
		value: value asSymbol ! !

! ScriptParser methodsFor: #parsing !
classBindingNamed: varName
	" Private - Returns a binding known by the currentClass with name varName. "

	| aClass pool binding |
	aClass := self currentClass instanceClass.
	aClass withAllSuperclassesDo: [:cls|
		pool := cls classVariables.
		pool notNil ifTrue: [
			binding := pool bindingAt: varName ifAbsent: [].
			binding notNil ifTrue: [ ^binding ].
		]
	].
	aClass withAllSuperclassesDo: [:cls|
		cls poolDictionaryNames do: [:aName|
			pool := Smalltalk poolDictionary: aName ifAbsent: [].
			pool notNil ifTrue: [
				binding := pool bindingAt: varName ifAbsent: [].
				binding notNil ifTrue: [ ^binding ].
			]
		].
	].
	^nil! !

! ScriptParser methodsFor: #parsing !
collectIndices: variable in: aContext
	" Private - Collect the indices following variable reference in aContext.
	<rule> indices = {'#' string | '[' expression ']'}
	"

	| start indices done |
	indices := #(). done := false.
	[	(source peek = self literalQuote
		or: [ source peek = self leftBracket ])
		and: [ done not ]
	] whileTrue: [ | c index |
		start := self position.
		c := source next.
		c = self literalQuote ifTrue: [
			index := source peek = self singleQuote
				ifTrue: [ source next. self nextString ]
				ifFalse: [ self nextSymbol ].
		].
		c = self leftBracket ifTrue: [
			index := self expression: aContext.
			index notNil ifTrue: [
				self startOfNextToken.
				source peek = self rightBracket
				ifTrue: [ source next ]
				ifFalse: [ index := nil ]
			]
		].
		index isNil ifTrue: [
			self position: start.
			done := true
		] ifFalse: [ indices add: index ]
	].
	indices notEmpty ifTrue: [
		variable indices: indices
	]! !

! ScriptParser methodsFor: #parsing !
nextVariable: aContext
	" Private - Parse the receiver looking for a variable name.
	Returns the variable binding or nil if failed. "

	| word assoc start variable found |
	start := self startOfNextToken.
	word := self nextWord.
	word isNil ifTrue: [ ^nil ].
	found := [
		self collectIndices: variable in: aContext.
		variable sourceRange: (start to: self position)
	].

	variable := self specialVariableNamed: word.
	variable notNil ifTrue: [ ^found value ].
	assoc := self classBindingNamed: word.
	assoc isNil ifTrue: [
		assoc := aContext bindingAt: word ifAbsent: [].
	].
	(assoc isNil and: [ word first isUppercase ]) ifTrue: [
		assoc := S8Binding key: word in: Smalltalk.
	].
	assoc isNil ifTrue: [
		self position: start.
		^nil
	].
	variable := ParseGlobalVariable boundTo: assoc.
	^found value! !

! ScriptParser methodsFor: #parsing !
nextWord
	" Returns the next word in the receiver or nil if failed.
	 Advance the receiver. "

	| start word |
	self atEnd ifTrue: [ ^nil ].
	start := self startOfNextToken.
	(self isAlpha: self peek) ifFalse: [ ^nil ].

	word := self next.
	[	source atEnd not
		and: [ self isAlpha: source peek ]
	] whileTrue: [ word := word , source next ].

	^word! !

! ScriptParser methodsFor: #error !
notImplementedYet: aString
	" Returns a syntax error because the functionality is not implemented yet. "

	^self syntaxError: aString ,' not implemented. Yet' ! !

! ScriptParser methodsFor: #parsing !
parse: code doing: action onError: aBlock
	" Private - Parse code with action and returns the root node of the parsing. "

	| result |
	errorBlock := aBlock.
	source := code isString
		ifTrue: [ code stream ]
		ifFalse: [ code ].
	result := action evaluate.
	errorBlock := source := nil.
	^result! !

! ScriptParser methodsFor: #scripting !
parse: aStringOrStream onError: aBlock
	" Parse S8 code (aStringOrStream) and returns the root node of the parsing.
	If occurs an error aBlock will be valued with the cause as an argument.
	Example:
		ScriptParser new parse: 'Smalltalk keys' onError: [:error| ^error printString ].
	"

	^self	parse: aStringOrStream doing: [ | statements |
			statements := ParseStatements new.
			self	parseStatements: statements
				in: (self bodyScope: statements)
				innerBlock: false.
			Script with: statements
		] onError: aBlock! !

! ScriptParser methodsFor: #scripting !
parse: aStringOrStream
	" Parse S8 code (aStringOrStream) and returns the root node of the parsing (or an error). "

	| result |
	result := self parse: aStringOrStream onError: [:error| ^error ].
	^result ! !

! ScriptParser methodsFor: #parsing !
peek
	" Peek the next character from source stream. "

	self skipSeparators.
	^self basicPeek! !

! ScriptParser methodsFor: #parsing !
peekAny: aCollection
	" Try to match the next token of the receiver to any of the elements in aCollection.
	Returns the first element in aCollection that matches or nil if none. "

	^self peekAny: aCollection advance: true! !

! ScriptParser methodsFor: #parsing !
peekAny: aCollection advance: advance
	" Try to match the next token of the receiver to any of the elements in aCollection.
	Returns the first element in aCollection that matches or nil if none. "

	| start |
	start := self position.
	aCollection do: [:each|
		(self match: each) ifTrue: [
			advance ifFalse: [ self position: start ].
			^each
		].
	].
	^nil! !

! ScriptParser methodsFor: #parsing !
primaryExpression: aContext
	" Private - Parse the receiver looking for a primary expression using variables in aContext.
	Returns nil if failed.
	<rule> primary = literal | variableName | brace | block | '(' expression ')' | '<' primitive '>' .
	"

	| node |
	node := self nextLiteral: aContext.
	node notNil ifTrue: [ ^node ].
	node := self nextVariable: aContext.
	node notNil ifTrue: [ ^node ].
	(self match: self leftBrace) ifTrue: [
		^self nextBraceExpression: aContext
	].
	(self match: self leftBracket) ifTrue: [
		^self nextBlock: aContext
	].

	(self match: self leftParenthesis) ifTrue: [
		node := self expression: aContext.
		(self match: self rightParenthesis) ifTrue: [ ^node ].
		^self expected: self rightParenthesis
	].
	^nil " self expected: #expression "! !

! ScriptParser methodsFor: #parsing !
pushNil
	" Returns a node returning nil. "

	^self specialVariableNamed: #nil ! !

! ScriptParser methodsFor: #parsing !
pushSelf
	" Returns a node returning self. "

	^self specialVariableNamed: #self ! !

! ScriptParser methodsFor: #parsing !
readIntegerFrom: aStream radix: radix
	" Private - Answer aNumber of the concrete subclasses if Integer.
	Initial minus sign accepted, and bases > 10 use letters A-Z.
	Imbedded radix specifiers not allowed; use Integer
	class readFrom: for that. Answer zero if there are no digits. "

	| digit value |
	value := 0.
	[aStream atEnd] whileFalse: [
		digit := aStream next.
		digit := digit isDigit ifTrue: [ digit digitValue ] ifFalse: [-1].
		(digit < 0 or: [digit >= radix]) ifTrue: [
			aStream skip: -1.
			^value
		].
		value := value * radix + digit
	].
	^value! !

! ScriptParser methodsFor: #parsing !
readSmalltalkFloat: integerPart from: aStream
	" Private -Answer aNumber as described on the stream, aStream. "

	| value coercionClass possibleCoercionClass precision digit num den exp neg endOfNumber eChar atEnd |
	precision := 0. num := 0. den := 1.
	coercionClass := #Integer.
	(aStream peekFor: $.) ifTrue: ["<integer>.<fraction>"
		(aStream atEnd not and: [aStream peek isDigit]) ifTrue: [
			coercionClass := #Float.
			[ (atEnd := aStream atEnd) or: [(digit := aStream next) isDigit not]
			] whileFalse: [
				num := num * 10 + digit digitValue.
				precision := precision + 1
			].
			den := 10 raisedTo: precision.
			atEnd ifFalse: [ aStream skip: -1 ]
		] ifFalse: ["just <integer>. un-gobble the period"
			aStream skip: -1
		]
	].
	eChar := aStream peek.
	eChar == nil
	ifTrue: [ possibleCoercionClass := nil ]
	ifFalse: [
		"true ifTrue: [ "eChar := eChar asLowercase" ]".
		possibleCoercionClass := eChar = $e ifTrue: [ #Float ].
		possibleCoercionClass == nil ifFalse: [ aStream next ]
	].
	exp := nil.
	possibleCoercionClass == nil ifFalse: [
		coercionClass := possibleCoercionClass.
		endOfNumber := aStream position.
		neg := aStream peekFor: $-.
		((digit := aStream peek) notNil and: [digit isDigit]) ifTrue: [
			exp := self readIntegerFrom: aStream radix: 10.
			neg ifTrue: [ exp := exp negated ]
		] ifFalse: [ aStream position: endOfNumber ]
	].
	value := integerPart + (num / den). "The exponent will be added in the next step. "
	" ^self
		coerce: value
		to: coercionClass
		precision: precision
		exponent: exp
		exponentChar: eChar "
	^exp isNil ifTrue: [ value ] ifFalse: [ value * (10 raisedTo: exp) ]! !

! ScriptParser methodsFor: #parsing !
readSmalltalkNumberFrom: aStream
	" Private - Answer aNumber as described on the stream, aStream. "

	| value neg radix |
	(aStream atEnd or: [aStream peek isLetter]) ifTrue: [^0].
	neg := aStream peekFor: $-.
	value := self readIntegerFrom: aStream radix: 10.
	(aStream peekFor: $r) ifTrue: [
		radix := value.
		value := self readSmalltalkRadix: value from: aStream.
		value := self checkForOldSyntax: radix with: value on: aStream
	] ifFalse: [
		value := self readSmalltalkFloat: value from: aStream
	].
	^neg
		ifTrue: [value negated]
		ifFalse: [value]! !

! ScriptParser methodsFor: #parsing !
readSmalltalkRadix: radix from: aStream
	" Private - Answer aNumber as described on the stream, aStream. "

	radix < 2 ifTrue: [ ^self syntaxError: 'Invalid radix' ].
	(aStream atEnd or: [(aStream peek digitValue between: 0 and: radix - 1) not])
	ifTrue: [ aStream skip: -1. ^radix ].
	^self readIntegerFrom: aStream radix: radix! !

! ScriptParser methodsFor: #constants !
rightBrace
	" Private - Returns the rightBrace token. "

	^$}! !

! ScriptParser methodsFor: #constants !
rightBracket
	" Private - Returns the rightBracket token. "

	^$]! !

! ScriptParser methodsFor: #constants !
rightParenthesis
	" Private - Returns the rightParenthesis token. "

	^$)! !

! ScriptParser methodsFor: #accessing !
scope
	" Returns the receiver's root scope. "

	scope isNil ifTrue: [ self initializeScope ].
	^scope! !

! ScriptParser methodsFor: #accessing !
scope: aScope
	" Set the receiver's root scope. "

	scope := aScope! !

! ScriptParser methodsFor: #constants !
semicolon
	" Private - Returns the semicolon token. "

	^$;! !

! ScriptParser methodsFor: #constants !
singleQuote
	" Private - Returns the singleQuote token. "

	^$'! !

! ScriptParser methodsFor: #parsing !
skip: aCharacter
	" Skip aCharacter from source stream if present. "

	self peek = aCharacter ifTrue: [ self next ]! !

! ScriptParser methodsFor: #parsing !
skipSeparators
	" Advance to the source stream to the next non-separator character. "

	[	self basicPeek isNil ifTrue: [ ^self ].
		self basicPeek isSeparator
	] whileTrue: [ source next ]! !

! ScriptParser methodsFor: #parsing !
specialVariableNamed: aName
	" Private - Returns a variable known by the receiver with name aName. "

	| index |
	aName = #self ifTrue: [ ^ParseSelf new ].
	aName = #nil ifTrue: [ ^ParseUniversal value: nil ].
	aName = #true ifTrue: [ ^ParseUniversal value: true ].
	aName = #false ifTrue: [ ^ParseUniversal value: false ].
	aName = #super ifTrue: [ ^ParseSuper in: self currentClass ].
	^self instanceVariables detect: [:one| one name = aName ] ifNone: [] ! !

! ScriptParser methodsFor: #parsing !
startOfNextToken
	" Advance to next token and return current position. "

	self skipSeparators.
	^self position! !

! ScriptParser methodsFor: #parsing !
parseStatements: statements in: aContext innerBlock: isInnerBlock
	" Private - Parse the statements in the receiver looking for variables in aContext. "

	| returns more expression start origin |
	origin := self startOfNextToken.
	returns := false.
	more := true.
	[more] whileTrue: [
		start := self startOfNextToken.
		returns := self match: self upArrow.
		returns ifTrue: [ "Return expression. "
			expression := self expression: aContext.
			expression isNil ifTrue: [ ^self expected: 'expression to return' ].
			statements addLast: (
				MethodReturnNode new
					expression: expression;
					sourceRange: (start to: self position);
					yourself).
		]
		ifFalse: [
			expression := self expression: aContext.
			expression isNil ifTrue: [
				expression := isInnerBlock
					ifTrue: [ self pushNil ]
					ifFalse: [ self pushSelf ]
			].
			statements addLast: expression
		].
		returns ifTrue: [ self skip: self period ].
		more := returns not and: [
			expression isPrimitiveNode or: [
				self match: self period ] ]
	].
	self skipSeparators.
	self atEnd ifFalse: [ ^self expected: 'message or period' ].
	statements sourceRange: (origin to: self position).
	^statements! !

! ScriptParser methodsFor: #parsing !
statementsArgs: argNodes in: aScope
	" Private - Parse the receiver looking for a statements in a block using variables in aScope.
	Returns nil if failed.
	<node> $[ ($: var)* temporaries statements $] => BlockClosureNode
	"

	| stmts start returnStart parseNode aContext |
	aContext := aScope.
	stmts := BlockClosureNode new.
	stmts arguments: argNodes.		
	start := self position.
	self skipSeparators.
	(self match: self verticalBar) ifTrue: [ 
		aContext := ParseContext in: aContext.
		stmts temporaries: (self parseTemporaries: aContext).
	].
	[	(self match: self upArrow) ifTrue: [
			returnStart := self position.
			parseNode := self expression: aContext.
			parseNode isNil ifTrue: [ ^self expected: 'expression to return' ].
			parseNode := MethodReturnNode new
				expression: parseNode;
				sourcePosition: (returnStart to: self position);
				yourself.
			stmts addLast: parseNode.
			self match: self period.
			false
		] ifFalse: [
			parseNode := self expression: aContext.
			parseNode notNil ifTrue: [
				stmts addLast: parseNode.
				self match: self period
			] ifFalse: [false]
		]
	] whileTrue: [].
	stmts sourcePosition: (start to: self position).
	^stmts! !

! ScriptParser methodsFor: #constants !
upArrow
	" Private - Returns the upArrow token. "

	^$^! !

! ScriptParser methodsFor: #constants !
verticalBar
	" Private - Returns the verticalBar token. "

	^$|! !

! ScriptParser methodsFor: #parsing !
nextMessage: args in: aContext
	" Private - Returns the selector from next message header.
	<rule> message = (selector | (selector $: argument)*) [comment]
	"

	| exit nextArg start selector part |
	start := self startOfNextToken.
	selector := self nextSelector.
	exit := [ self skipSeparators. selector ].
	nextArg := [ | arg |
		arg := self nextArgument: aContext.
		arg notNil ifTrue: [
			arg canBeAssigned: true.
			args addLast: arg
		]
	].
	selector last = self colon ifFalse: [
		selector argumentCount timesRepeat: [
			nextArg value isNil ifTrue: [
				^self expected: #argument
			]
		].
		^exit value
	].
	nextArg value.
	[	start := self startOfNextToken.
		part := self nextWord.
		part isNil ifTrue: [ ^exit value ].
		self basicPeek = self colon
	] whileTrue: [
		part := part ,self next.
		nextArg value isNil ifTrue: [
			self position: start.
			^selector
		].
		selector := selector ,part.
	].
	self position: start.
	^exit value! !

! ScriptParser methodsFor: #parsing !
methodScope: aCompiledScript for: statements
	" Private - Returns a context for parsing aCompiledScript.
	<rule> compiledScript = message script.
	<rule> message = (selector [[:] argument]...) [comment].
	"

	| args selector ctx |
	ctx := ParseContext new.
	args := #().
	aCompiledScript args: args.
	selector := self nextMessage: args in: ctx.
	aCompiledScript selector: selector.
	args notEmpty ifTrue: [
		ctx parent: self scope.
		scope := ctx.
	].
	^self bodyScope: statements! !

! ScriptParser methodsFor: #parsing !
methodReturn
	" Private - Returns an implisit return node. "

	^(MethodReturnNode
		from: self position to: self position)
		expression: ParseSelf new ! !

! ScriptParser methodsFor: #parsing !
parseMethod: aCompiledScript
	" Private - Parse method (aCompiledScript) reading from source stream. "

	| body methodScope statements |
	statements := ParseStatements new.
	methodScope := self methodScope: aCompiledScript for: statements.
	body := self parseStatements: statements in: methodScope innerBlock: false.
	(body respondsTo: #isEmpty) ifFalse: [ ^nil ].
	(body isEmpty or: [body statements last isReturn not]) ifTrue: [
		body statements add: self methodReturn
	].
	aCompiledScript script: (Script with: body)! !

! ScriptParser methodsFor: #scripting !
parse: aString method: aCompiledScript onError: aBlock
	" Parse aString as method source code implemented in aCompiledScript.
	Returns the receiver.
	See ScriptCompiler framework to learn more on #scripting.
	"

	self parse: aString doing: [
		self currentClass: aCompiledScript classField.
		self parseMethod: aCompiledScript.
		aCompiledScript source: aString.
		] onError: aBlock! !


! ScriptingError class methodsFor: #accessing !
description: aString
	" Returns an instance of the receiver with description aString. "

	^self new
		description: aString;
		yourself! !

! ScriptingError methodsFor: #testing !
isParseFailure
	" Return true if the receiver is a parse failure. "

	^true! !

! ScriptingError methodsFor: #accessing !
description
	" Returns the receiver's description string. "

	description isNil ifTrue: [ self initializeDescription ].
	^description! !

! ScriptingError methodsFor: #accessing !
description: aText
	" Set the receiver's description string. "

	description := aText! !

! ScriptingError methodsFor: #initialize !
initializeDescription
	" Private - Initializes the receiver's description. "

	description := '*Error*'! !

! ScriptingError methodsFor: #printing !
descriptionString
	" Returns the receiver's description as shown to the user. "

	^self description! !

! ScriptingError methodsFor: #printing !
printOn: aStream
	" Append to the argument aStream, a sequence of characters that describes the receiver. "

	aStream
		nextPutAll: self class name ,': ';
		nextPutAll: self descriptionString! !


! ParseError methodsFor: #accessing !
position
	" Returns the receiver's position. "

	^position! !

! ParseError methodsFor: #accessing !
position: index
	" Set the receiver's position. "

	position := index! !

! ParseError methodsFor: #accessing !
source
	" Returns the receiver's source. "

	^source! !

! ParseError methodsFor: #accessing !
source: aString
	" Set the receiver's source. "

	source := aString! !

! ParseError methodsFor: #printing !
descriptionString
	" Returns the receiver's description as shown to the user. "

	| stream |
	self source isNil ifTrue: [ ^super descriptionString ].
	stream := self source stream.
	^' '
	,(stream next: (self position min: stream size))
	,'<---' ,self description
	,' ' ,stream upToEnd! !

! SyntaxError methodsFor: #initialize !
initializeDescription
	" Private - Initializes the receiver's description. "

	description := 'Syntax Error'! !