" Step by step execution support for S8 Smalltalk.
  Author: Alejandro F. Reimondo [aleReimondo@smalltalking.net]
  Authored @2020 S8 see http://u8.smalltalking.net
"
(Smalltalk includesKey: #S8Context) ifFalse: [
	^self error: 'Bindings framework missing'
]!

EventManager
	subclass: #StepsObject
	instanceVariableNames: ' id '
	classVariableNames: ' LastID '
	category: #Steps !
StepsObject comment: '
	Instances are part of Steps framework, can trigger events and maintain a reference ID.
'!

! StepsObject class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^#(	needsID
	) ,super constructEventsTriggered! !

! StepsObject class methodsFor: #id !
newID
	" Private - Returns a new/unique ID to reference an instance of the receiver. "

	LastID isNil ifTrue: [ LastID := 0 ].
	LastID := LastID + 1.
	^LastID ! !

! StepsObject methodsFor: #initialize !
initializeId
	" Private - Initialize the id of the receiver. "

	id := self
		triggerEvent: #needsID
		ifNotHandled: [ self class newID ] ! !

! StepsObject methodsFor: #accessing !
id
	" Returns the id of the receiver. "

	id isNil ifTrue: [ self initializeId ].
	^id ! !

! StepsObject methodsFor: #ui !
descriptionForBinding: aBinding
	" Private - Returns a (short) description of aBinding. "

	| text max |
	max := 30.
	text := self descriptionForValue: aBinding value.
	text size > max ifTrue: [ text := (text copyFrom: 1 to: max - 3) ,'...'].
	^aBinding key ,' := ' ,text ! !

! StepsObject methodsFor: #ui !
descriptionForValue: anObject
	" Private - Returns a (short) description of anObject. "

	| text |
	anObject isNil ifTrue: [ ^#nil ].

	[ text := self toString: anObject
	] on: Exception do: [:ex|
		^'Note: ' ,ex messageText
	].

	(self isObject: anObject) ifTrue: [ ^anObject asLiteral ].
	^text ! !

! StepsObject methodsFor: #debug !
info
	" Returns the description of the receiver. "

	^NativeHandle support @> self ! !

S8Block
	subclass: #SimulationBlock
	instanceVariableNames: ' simulator method '
	category: #Steps !

! SimulationBlock methodsFor: #accessing !
simulator: aSimulator
	" Private - Set the receiver. "

	| tos |
	simulator := aSimulator.
	tos := aSimulator topFrame.
	tos notNil ifTrue: [ method := tos method ] ! !

! SimulationBlock methodsFor: #accessing !
node
	" Private - Returns the node of the receiver. "

	^node ! !

! SimulationBlock methodsFor: #accessing !
context
	" Private - Returns the activation context of the receiver. "

	^context ! !

! SimulationBlock methodsFor: #accessing !
source
	" Return the source of the receiver (or nil). "

	| interval |
	(method isNil or: [ method source size = 0 ]) ifTrue: [ ^nil ].
	interval := node sourceRange.
	^interval notNil ifTrue: [
		method source
			copyFrom: (interval start max: 1)
			to: interval end
	] ! !

! SimulationBlock methodsFor: #converting !
asLiteral
	" Return the receiver in literal notation. "

	| result |
	result := self source.
	^result notNil ifTrue: [ result ] ifFalse: [ super asLiteral ] ! !

! SimulationBlock methodsFor: #converting !
asBlock
	" Returns the receiver an an evaluable block. "

	^S8Block for: node in: context ! !

! SimulationBlock methodsFor: #evaluating !
valueWithArguments: anArray
	" Evaluate the receiver with parameters anArray.
	Result of evaluation is loaded on simulator value when finished.
	"

	S8Context
		in: context clear: self arguments ,self temporaries
		do: [:ctx|
			self bindTo: anArray in: ctx.
			node inject: simulator in: ctx
	]. ! !

! SimulationBlock methodsFor: #evaluating !
ensure: aBlock
	" Return the result of evaluating the receiver.
	Execute aBlock after the receiver is executed.
	"

	self bindTo: #() in: context.
	node	inject: simulator in: context
		catch: nil curtail: nil
		finally: aBlock ! !

! SimulationBlock methodsFor: #evaluating !
ifCurtailed: exceptionBlock
	" Return the result of evaluating the receiver.
	If evaluation is limited by exiting before completion, the exception block is evaluated.
	"

	self bindTo: #() in: context.
	node	inject: simulator in: context
		catch: nil curtail: exceptionBlock
		finally: nil ! !

! SimulationBlock methodsFor: #evaluating !
on: anExceptionType do: aBlock
	" Return the result of evaluating the receiver catching exceptions. "

	self bindTo: #() in: context.
	node	inject: simulator in: context
		catch: (Association key: anExceptionType value: aBlock)
		curtail: nil finally: nil ! !

! SimulationBlock methodsFor: #evaluating !
repeat
	" OBSOLETE - Evaluate the receiver repeatedly, ending only if the block forces some stopping condition. "

	simulator
		repeat: self while: [ true ]
		doing: nil then: [] ! !

! SimulationBlock methodsFor: #evaluating !
whileFalse: aBlock
	" Evaluate aBlock while the result of evaluating the receiver is false. "

	simulator
		repeat: self while: [ false = simulator value ]
		doing: aBlock then: [ simulator value: nil ] ! !

! SimulationBlock methodsFor: #evaluating !
whileTrue: aBlock
	" Evaluate aBlock while the result of evaluating the receiver is true. "

	simulator
		repeat: self while: [ true = simulator value ]
		doing: aBlock then: [ simulator value: nil ] ! !

! SimulationBlock methodsFor: #testing !
isValuable
	" Return true if the receiver can be evaluated. "

	^true! !

! SimulationBlock methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	^self! !

StepsObject
	subclass: #SimulationStep
	instanceVariableNames: ' source action context '
	category: #Steps !
SimulationStep buildAccessors!

! SimulationStep class methodsFor: #instantiation !
for: aNodeOrMethod doing: action in: aContext
	" Return an instance of the receiver. "

	^self new initialize: aNodeOrMethod doing: action in: aContext ! !

! SimulationStep methodsFor: #initialize !
initialize: aNodeOrMethod doing: aBlock in: aContext
	" Private - Initialize the receiver. "

	source := aNodeOrMethod.
	action := aBlock.
	context := aContext ! !

! SimulationStep methodsFor: #evaluating !
evaluateWithArguments: anArray
	" Return the result of evaluating the receiver. "

"self note: source class name ,$	,(JS@> source)."
	^self action evaluateWithArguments: anArray ! !

! SimulationStep methodsFor: #accessing !
sourceRange
	" Returns the source range of source (or nil). "

	^(source notNil and: [ source respondsTo: #sourceRange ])
		ifTrue: [ source sourceRange ] ! !

! SimulationStep methodsFor: #accessing !
sourceLimits
	" Returns the source range limits of source (or nil). "

	| range |
	range := self sourceRange.
	^range notNil ifTrue: [ Array with: range start with: range end ] ! !

! SimulationStep methodsFor: #accessing !
bindings
	" Returns the bindings of the receiver (or nil). "

	| array |
	context isNil ifTrue: [ ^nil ].
	array := #().
	context associationsDo: [:assoc| array add: assoc ].
	^array ! !

! SimulationStep methodsFor: #virtuals !
@ key
	" Returns the result of accessing the receiver at key (or nil). "

	context isNil ifTrue: [ ^nil ].
	context associationsDo: [:assoc|
		assoc key = key ifTrue: [ ^assoc ]
	].
	^nil ! !

! SimulationStep methodsFor: #accessing !
bindingAt: key put: aValue
	" Set the binding of the receiver. "

	context isNil ifTrue: [ ^nil ].
	^context at: key put: aValue ! !

! SimulationStep methodsFor: #printing !
printString
	" Return a printable representation of the receiver. "

	(source respondsTo: #sourceRange) ifFalse: [
		^super printString
	].
	^$[ ,source asLiteral ,$@ ,source sourceRange ,$] ! !

StepsObject
	subclass: #SimulationFrame
	instanceVariableNames: ' method steps receiver values'
		,' catcher curtail final '
	category: #Steps !
SimulationFrame buildAccessors!

! SimulationFrame class methodsFor: #instantiation !
for: aMethod doing: steps
	" Return an instance of the receiver. "

	^self new initialize: aMethod doing: steps ! !

! SimulationFrame class methodsFor: #instantiation !
for: aMethod catch: catcher curtail: curtail finally: final
	" Return an instance of the receiver. "

	^(self for: aMethod doing: #())
		catcher: catcher;
		curtail: curtail;
		final: final;
		yourself ! !

! SimulationFrame methodsFor: #initialize !
initialize: aMethod doing: array
	" Private - Initialize the receiver. "

	method := aMethod.
	steps := array copy ! !

! SimulationFrame methodsFor: #accessing !
receiver: anObject values: array
	" Private - Set activation data. "

	receiver := anObject.
	values := array ! !

! SimulationFrame methodsFor: #accessing !
selector
	" Returns the selector of message sent. "

	^self method selector ! !

! SimulationFrame methodsFor: #accessing !
contents
	" Returns the receiver's contents. "

	^steps isNil
		ifTrue: [ #() ]
		ifFalse: [ steps ] ! !

! SimulationFrame methodsFor: #accessing !
bindings
	" Returns the associations bound to the receiver (or nil). "

	^steps size > 0 ifTrue: [ steps first bindings ] ! !

! SimulationFrame methodsFor: #virtuals !
bindingsInfo
	" Returns the information for bindings of the receiver. "

	| array |
	array := self bindings.
	^array notNil ifTrue: [
		Array	with: steps first id
			with: (array collect: [:each|
				Array	with: each key
					with: (self descriptionForBinding: each)
			])
	] ! !

! SimulationFrame methodsFor: #query !
isEmpty
	" Returns true if the receiver is empty. "

	^steps size = 0 ! !

! SimulationFrame methodsFor: #stream !
peek
	" Returns the next action to evaluate (or nil). "

	^steps size > 0 ifTrue: [ steps first ]! !

! SimulationFrame methodsFor: #stream !
next
	" Returns and remove the next action to evaluate (or nil). "

	^steps size > 0 ifTrue: [ steps removeFirst ]! !

! SimulationFrame methodsFor: #add !
add: aStep
	" Add an inmediate step to evaluate. "

	^steps addFirst: aStep ! !

! SimulationFrame methodsFor: #stack !
restart: simulator
	" Private - Set simulator to restart the execution of the receiver.
	NOTE: the receiver will be (silently) removed from stack and a new frame activation will occur for the original message.
	"

	simulator callStack removeFirst = self ifFalse: [ ^self error: 'Broken stack' ].
	simulator
		send: self selector to: self receiver
		withArguments: self values ! !

! SimulationFrame methodsFor: #execute !
execute: step for: simulator
	" Private - Capture execution of step in simulator. "

	[ step evaluateWith: simulator
	] on: Exception do: [:ex|
		simulator	handleException: ex
	]! !

! SimulationFrame methodsFor: #printing !
printString
	" Return a printable representation of the receiver. "

	method isNil ifTrue: [ ^super printString ].
	^method classField asLiteral ,#~> ,method selector asLiteral ! !

! SimulationFrame methodsFor: #virtuals !
methodInfo
	" Private - Returns the information to catch for the method (or nil). "

	^method notNil ifTrue: [
		Array	with: method classField name 
			with: method selector
			with: method source
	] ! !

! SimulationFrame methodsFor: #virtuals !
stepIDs
	" Private - Returns an array with steps IDs. "

	^self steps collect: [:each| each id ] ! !

! SimulationFrame methodsFor: #virtuals !
@ key
	" Returns the result of accessing the receiver at key (or nil). "

	| anID |
	anID := key asNumber.
	^self steps detect: [:one| one id = anID ] ifNone: [] ! !

StepsObject
	subclass: #S8Simulator
	instanceVariableNames: ' callStack value state shortcuts scriptCompiler breakpoints'
	classVariableNames: ' InUse '
	category: #Steps !
S8Simulator buildGetters: #(
	#(callStack initialize Array)
	#(breakpoints initialize Array)
)!
S8Simulator buildAccessors: #(
	#(scriptCompiler initialize 'self defaultScriptCompiler')
	#(shortcuts initialize 'self
		triggerEvent: #needsShortcuts
		ifNotHandled: [ PoolDictionary new ]')
)!
S8Simulator class instanceVariableNames: 'current useScripting'.
S8Simulator class buildAccessors: #(
	current
	#(useScripting initialize #false)
)!

! S8Simulator class methodsFor: #example !
example
	" DEBUG - Open an instance with a simple expression. "

	(self open: 'self halt. 3 + 4 ') run ! !

! S8Simulator class methodsFor: #ui !
open: initializer
	" Returns an instance opened with initializer. "

	^self new open: initializer ! !

! S8Simulator class methodsFor: #ui !
run: initializer
	" Returns an instance opened with initializer. "

	^(self open: initializer) run ! !

! S8Simulator class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^#(	aboutToOpen: opened
		ready paused breakpoint:
		aboutToRun atEnd
		aboutToClose closed
		valueChanged:
		enter: enter:with:
		exit: exit:with: aboutToExit:
		activateMethod: activateNative:
		activateBlock: deactivateBlock:
		terminate: discard:
		aboutToExecute: executed:
		needsShortcuts getShortcuts:
	) ,super constructEventsTriggered! !

! S8Simulator class methodsFor: #accessing !
inUse
	" Returns the pool of open instances of the receiver. "

	InUse isNil ifTrue: [ InUse := PoolDictionary new ].
	^InUse! !

! S8Simulator class methodsFor: #accessing !
@ key
	" Returns the information given at key (or nil). "

	^self inUse @ key asNumber ! !

! S8Simulator class methodsFor: #instantiation !
empty
	" Returns a fresh instance of the receiver. "

	^self open: nil ! !

! S8Simulator class methodsFor: #private !
reset
	" Reset the receiver. "

	LastID := InUse := nil ! !

! S8Simulator methodsFor: #defaults !
defaultScriptCompiler
	" Returns the default script compiler for the receiver. "

	self class useScripting ifTrue: [
		^Smalltalk at: #ScriptCompiler
	].
	^Compiler ! !

! S8Simulator methodsFor: #accessing !
value
	" Returns the value of the receiver. "

	^value ! !

! S8Simulator methodsFor: #accessing !
value: aValue
	" Set the value of the receiver. "

	value := aValue.
	self triggerEvent: #valueChanged: with: aValue ! !

! S8Simulator methodsFor: #breakpoints !
breakpointFromString: aString
	" Private - Returns a breakpoint specified by aString. "

	(aString includes: #>>) ifTrue: [
		(aString upTo: #>>) trimBlanks notEmpty ifTrue: [
			^aString fileIn
		].
	].
	^aString ! !

! S8Simulator methodsFor: #accessing !
breakpoints: array
	" Set the breakpoints of the receiver. "

	breakpoints := array notNil ifTrue: [
		array collect: [:each|
			each isString
			ifTrue: [ self breakpointFromString: each ]
			ifFalse: [ each ]
		]
	]! !

! S8Simulator methodsFor: #query !
isEmpty
	" Returns true if the receiver is empty. "

	^callStack size = 0 ! !

! S8Simulator methodsFor: #accessing !
size
	" Return the size of the receiver. "

	^callStack size ! !

! S8Simulator methodsFor: #accessing !
topFrame
	" Returns the frame at top (or nil). "

	^self isEmpty not ifTrue: [ callStack first ] ! !

! S8Simulator methodsFor: #accessing !
bindings
	" Returns the bindings of the receiver (or nil).
	Note: it is nil or an Array with result (when finished).
	"

	^self atEnd ifTrue: [ Array with: #result -> self value ] ! !

! S8Simulator methodsFor: #virtuals !
bindingKeys
	" Returns the keys for bindings of the receiver. "

	| array |
	array := self bindings.
	^array notNil ifTrue: [ array collect: [:each| each key ] ] ! !

! S8Simulator methodsFor: #virtuals !
topFrameID
	" Returns the ID of the frame at top (or nil). "

	| part |
	part := self topFrame.
	^part notNil ifTrue: [ part id ]! !

! S8Simulator methodsFor: #virtuals !
callStackIDs
	" Returns the IDs of the callStack elements. "

	^self callStack collect: [:each| each id ] ! !

! S8Simulator methodsFor: #accessing !
needsWorkTo: aFrame
	" Returns true when more work must be done to reach aFrame on top of stack. "

	^(self includes: aFrame) and: [ callStack first ~= aFrame ] ! !

! S8Simulator methodsFor: #virtuals !
needsWorkToID: aFrameID
	" Returns true when more work must be done to reach aFrame on top of stack. "

	^self needsWorkTo: self @ aFrameID ! !

! S8Simulator methodsFor: #accessing !
frames
	" Returns a snapshot of the frames of the receiver. "

	^self callStack collect: [:each| each asLiteral ] ! !

! S8Simulator methodsFor: #virtuals !
@ key
	" Returns the result of accessing the receiver at key (or nil). "

	| anID |
	key = #result ifTrue: [ ^#result -> self value ].

	anID := key asNumber.
	^self callStack detect: [:one| one id = anID ] ifNone: [] ! !

! S8Simulator methodsFor: #stack !
removeTop
	" Remove the method at top of call stack. "

	| frame aMethod |
	self isEmpty ifTrue: [ ^nil ].
	frame := callStack removeFirst.
	aMethod := frame method.
	self	triggerEvent: #exit: with: aMethod;
		triggerEvent: #exit:with: with: aMethod with: value.
	self curtail: frame curtail finally: frame final ! !

! S8Simulator methodsFor: #exceptions !
handleException: exception
	" Private - Leave activated frames handling captured exception "

	| handler |
	handler := callStack detect: [:each|
		self catcher: each catcher canHandle: exception
	] ifNone: [ ^self error: 'Uncaptured exception ' ,exception description ].
	self discard: handler; removeTop.
	handler catcher value evaluateWith: exception. ! !

! S8Simulator methodsFor: #exceptions !
catcher: assoc canHandle: exception
	" Private - Returns true if catcher assoc can handle exception "

	^assoc notNil and: [
		[] exception: exception matches: assoc key
	] ! !

! S8Simulator methodsFor: #stream !
peek
	" Returns the next action to evaluate (or nil). "

	| result |
	[	self isEmpty ifTrue: [ ^nil ].
		result := self topFrame peek.
		result isNil
	] whileTrue: [ self removeTop ].
	^result! !

! S8Simulator methodsFor: #virtuals !
peekID
	" Returns the ID of next action to evaluate (or nil). "

	| top part |
	top := self topFrame.
	top isNil ifTrue: [ ^nil ].
	part := top peek.
	^part notNil ifTrue: [ Array with: top id with: part id ]! !

! S8Simulator methodsFor: #accessing !
peekSource
	" Returns the source of next step (or nil). "

	| step |
	step := self peek.
	^step notNil ifTrue: [ step source ]! !

! S8Simulator methodsFor: #query !
hasMessage
	" Returns true when the next action to evaluate is a message send. "

	| node |
	node := self peekSource.
	^(node respondsTo: #isMessageNode) and: [ node isMessageNode ] ! !

! S8Simulator methodsFor: #accessing !
peekSourceID
	" Returns the ID of current source (or nil). "

	| source hash |
	source := self peekSource.
	source isNil ifTrue: [ ^nil ].
	hash := (source respondsTo: #sourceRange)
		ifTrue: [ source sourceRange ]
		ifFalse: [ #none ].
	^Array	with: source class name
		with: hash toString ! !

! S8Simulator methodsFor: #query !
stepsOn: aFrame hasSource: node
	" Private - Returns true if the source is on sources to execute for frame. "

	aFrame steps do: [:each| each source = node ifTrue: [ ^true ] ].
	^false ! !

! S8Simulator methodsFor: #query !
includes: aFrame
	" Returns true if the callStack includes frame. "

	^self callStack includes: aFrame ! !

! S8Simulator methodsFor: #query !
includesID: aFrameID
	" Returns true if the callStack includes frame with id aFrameID. "

	^self includes: self @ aFrameID ! !

! S8Simulator methodsFor: #execute !
execute: step in: frame
	" Private - Evaluate the step in the frame (actually on top of callstack). "

	frame execute: step for: self! !

! S8Simulator methodsFor: #stream !
atEnd
	" Returns true if the receiver is at end. "

	^self peek isNil ! !

! S8Simulator methodsFor: #stream !
next
	" Evaluate the next action in the receiver; returns true when it is expected to continue execution. "

	| tos step |
	self atEnd ifTrue: [ ^false ].
	state := #running.
	tos := self topFrame.
	step := tos next.
	self triggerEvent: #aboutToExecute: with: step.
	self execute: step in: tos.
	self triggerEvent: #executed: with: step.
	^state ~~ #paused ! !

! S8Simulator methodsFor: #gui !
open: initializer
	" Open the receiver to execute initializer. "

	^self open: initializer on: nil ! !

! S8Simulator methodsFor: #gui !
open: initializer on: aReceiver
	" Open the receiver to execute initializer. "

	| rcv |
	rcv := aReceiver notNil ifTrue: [ aReceiver ].
	self triggerEvent: #aboutToOpen: with: initializer.
	self class inUse at: self id put: self.
	self class current: self.
	self prepare: initializer on: rcv.
	self	triggerEvent: #opened;
		triggerEvent: #ready ! !

! S8Simulator methodsFor: #gui !
close
	" Close the receiver. "

	self triggerEvent: #aboutToClose.
	self class inUse removeKey: self id ifAbsent: [].
	callStack := value := nil.
	self triggerEvent: #closed.
	^true ! !

! S8Simulator methodsFor: #running !
runWhile: condition
	" Execute the receiver (until paused, empty or condition value is true). "

	self triggerEvent: #aboutToRun.
	[	self atEnd ifTrue: [ ^self triggerEvent: #atEnd ].
		condition value and: [ self next ]
	] whileTrue ! !

! S8Simulator methodsFor: #running !
run
	" Start execution of the receiver (until paused or empty). "

	self runWhile: [ true ] ! !

! S8Simulator methodsFor: #actions !
stepIn
	" Step in current message sent. "

	| top node |
	node := self peekSource.
	node isNil ifTrue: [ ^nil ].
	top := self topFrame.
	self runWhile: [ self peekSource = node ] ! !

! S8Simulator methodsFor: #actions !
stepOver
	" Run until the current message is consumed. "

	| top node |
	node := self peekSource.
	node isNil ifTrue: [ ^nil ].
	top := self topFrame.
	self runWhile: [
		(self includes: top) and: [
			self stepsOn: top hasSource: node
		]
	] ! !

! S8Simulator methodsFor: #actions !
stepOut
	" Run the receiver while current frame is still on the callStack. "

	| top |
	top := self topFrame.
	top isNil ifTrue: [ ^nil ].
	self runWhile: [ self includes: top ] ! !

! S8Simulator methodsFor: #running !
pause
	" Pause execution of the receiver. "

	self add: #pause step: [
		self triggerEvent: #paused.
		state := #paused
	] in: nil ! !

! S8Simulator methodsFor: #initialize !
prepare: initializer on: rcv
	" Private - Initialize the receiver to execute initializer. "

	(initializer isMemberOf: Message) ifTrue: [
		^self	send: initializer selector
			to: initializer receiver
			withArguments: initializer arguments
	].
	((self isObject: initializer) not or: [initializer isNil]) ifTrue: [ ^self ].
	initializer isString ifTrue: [ | cls mth |
		cls := rcv class.
		mth := cls compiler
			compiledMethodToEvaluate: initializer
			in: cls.
		^self callMethod: mth for: rcv withArguments: #()
	].
	^self send: #evaluate to: initializer withArguments: #()! !

! S8Simulator methodsFor: #execute !
exit: aValue
	" Exit the execution of current method returning aValue. "

	self triggerEvent: #aboutToExit: with: aValue.
	value := aValue.
	^self removeTop ! !

! S8Simulator methodsFor: #stack !
stack: eventName remove: aFrame doing: aBlock
	" Private - Remove frames at top of the callStack until aFrame is found. "

	[  self isEmpty or: [aFrame = self topFrame] ] whileFalse: [ | frame |
		frame := self callStack removeFirst.
		self triggerEvent: eventName with: frame.
		aBlock evaluateWith: frame.
	]! !

! S8Simulator methodsFor: #stack !
removeTop: frames then: aBlock
	" Private - Remove top of the callStack and continue terminating frames. "

	self removeTop.
	self add: #terminate step: [ self terminate: frames then: aBlock ] in: nil. ! !

! S8Simulator methodsFor: #stack !
terminate: frames then: aBlock
	" Private - Terminate the execution of frames at top of the callStack. "

	| frame |
	frames isEmpty ifTrue: [ ^aBlock evaluate ].
	frame := frames removeFirst.
	self topFrame = frame ifFalse: [ self error: 'Broken call stack' ].
	self triggerEvent: #terminate: with: frame ifNotHandled: [
		self removeTop: frames then: aBlock
	]. ! !

! S8Simulator methodsFor: #stack !
discard: aFrame
	" Private - Discard the execution of frames at top of the callStack until aFrame is found. "

	self triggerEvent: #discard: with: aFrame ifNotHandled: [
		self stack: #discard: remove: aFrame doing: nil
	] ! !

! S8Simulator methodsFor: #private !
curtail: curtail finally: final
	" Private - Evaluate curtail & final blocks in current frame. "

	curtail notNil ifTrue: [ ^curtail ensure: final ].
	final notNil ifTrue: [ final ensure: nil ] ! !

! S8Simulator methodsFor: #stack !
restart: aFrame
	" Restart execution at start of aFrame. "

	self discard: aFrame.
	self isEmpty ifTrue: [ ^nil ].
	aFrame restart: self ! !

! S8Simulator methodsFor: #virtuals !
restartID: aFrameID
	" Restart execution at start of aFrame with id aFrameID. "

	^self restart: self @ aFrameID ! !

! S8Simulator methodsFor: #execute !
return: aNode in: aContext then: aBlock
	" Prepare the receiver to return from method. "

	| tos all |
	tos := aContext home.
	tos isNil ifTrue: [ ^self error: 'Method context missing' ].
	all := #().
	self callStack do: [:frame|
		tos = frame ifTrue: [ ^self terminate: all then: aBlock ].
		all add: frame.
	].
	^self error: 'Activation(' ,tos ,') missing in call stack'! !

! S8Simulator methodsFor: #execute !
addAll: statements in: aContext
	" Add steps to simulate statements in current method with aContext. "

	statements reverseDo: [:each| self add: each in: aContext ] ! !

! S8Simulator methodsFor: #execute !
activateBlock: aFrame catch: catcher curtail: curtail finally: final
	" Private - (re)Activate a method context to evaluate a block. "

	| tos |
	(aFrame = self topFrame
		and: [ catcher = aFrame catcher 
		and: [ curtail = aFrame curtail
		and: [ final = aFrame final ]]])
		ifTrue: [ ^aFrame ].
	tos := SimulationFrame
		for: aFrame method
		catch: catcher curtail: curtail finally: final.
	self callStack addFirst: tos.
	self triggerEvent: #activateBlock: with: tos.
	^tos ! !

! S8Simulator methodsFor: #execute !
deactivateBlock: aFrame finally: final
	" Private - Deactivate a block context that run up to end. "

	self triggerEvent: #deactivateBlock: with: aFrame.
	self curtail: nil finally: final ! !

! S8Simulator methodsFor: #execute !
addBlock: aNode contents: nodes in: aContext
	catch: catcher curtail: curtail finally: final
	" Add steps to activate a block of statements with aContext. "

	| home tos |
	home := aContext home.
	tos := self
		activateBlock: home
		catch: catcher curtail: curtail
		finally: final.
	self	add: aNode
		step: [ self deactivateBlock: home finally: final ]
		in: aContext.
	self addAll: nodes in: aContext. ! !

! S8Simulator methodsFor: #execute !
collect: aStatement all: nodes in: aContext
	" Add steps to simulate nodes in current method with aContext and collect results. "

	| array |
	value := #().
	nodes isEmpty ifTrue: [ ^self ].
	self add: aStatement step: [ value := array ] in: aContext.
	nodes reverseDo: [:each|
		self add: aStatement step: [ array add: value ] in: aContext.
		self add: each in: aContext
	].
	array := value.! !

! S8Simulator methodsFor: #execute !
add: aStatement step: action in: aContext
	" Add step to simulate aStatement in current method with aContext. "

	self topFrame add: (SimulationStep for: aStatement doing: action in: aContext) ! !

! S8Simulator methodsFor: #execute !
add: aStatement steps: array in: aContext
	" Add steps to simulate aStatement in current method with aContext. "

	array reverseDo: [:action| self add: aStatement step: action in: aContext ] ! !

! S8Simulator methodsFor: #execute !
add: aStatement in: aContext
	" Add steps to simulate aStatement in current method with aContext. "

	^self	add: aStatement
		steps: (aStatement actionsIn: aContext)
		in: aContext! !

! S8Simulator methodsFor: #activation !
activateMethod: aFrame in: aContext
	" Private - Activate the method context. "

	aContext notNil ifTrue: [
		self triggerEvent: #activateMethod: with: aFrame.
		aContext home: aFrame
	] ! !

! S8Simulator methodsFor: #execute !
enter: anObject method: aMethod withAll: steps
	" Private - Activate execution of Method and returns the frame pushed on stack. "

	| tos |
	tos := SimulationFrame for: aMethod doing: steps.
	self callStack addFirst: tos.
	self	triggerEvent: #enter: with: aMethod;
		triggerEvent: #enter:with: with: aMethod with: anObject.
	steps notEmpty ifTrue: [
		self activateMethod: tos in: steps first context
	].
	^tos ! !

! S8Simulator methodsFor: #private !
evaluationScript
	" Private - Return the support for skipping method execution. "

	^self scriptCompiler evaluationScript ! !

! S8Simulator methodsFor: #private !
scriptCompilerFor: aClass
	" Private - Return the script compiler to use for aClass. "

	| support |
	support := self scriptCompiler.
	support = Compiler ifTrue: [ ^aClass compiler ].
	^support for: aClass ! !

! S8Simulator methodsFor: #private !
compiledScriptFor: aMethod
	" Private - Returns the compiled script of aMethod (or nil). "

	| script |
	(aMethod respondsTo: #compiledScript)
		ifTrue: [ ^aMethod compiledScript ].

	script := aMethod basicAt: #s8_compiledScript ifAbsentPut: [
		(self scriptCompilerFor: aMethod classField)
			scriptForMethod: aMethod source
			onError: [ #none ]
	].
	^script = #none ifFalse: [ script ] ! !

! S8Simulator methodsFor: #execute !
callMethod: aMethod for: anObject withArguments: array
	" Call aMethod for anObject with given arguments. "

	| steps mth |
	aMethod isNil ifTrue: [ ^self error: 'Missing method' ].
	mth := self compiledScriptFor: aMethod.
	mth notNil ifTrue: [ | script |
		script := mth script.
		script notNil ifTrue: [
			steps := script
				stepsFor: anObject entering: mth
				withArguments: array.
		].
	].
	steps isNil ifTrue: [
		steps := self
			stepsOver: aMethod for: anObject
			withArguments: array.
	].
	(self enter: anObject method: aMethod withAll: steps)
		receiver: anObject values: array ! !

! S8Simulator methodsFor: #execute !
stepsOver: aMethod for: aReceiver withArguments: array
	" Private - The method do not have a script to evaluate.
	Return the steps to execute the method without entering script.
	"

	| script |
	array do: [:each| ((self isObject: each) and: [
		(each isMemberOf: SimulationBlock) and: [
		aMethod messages includes: #callableOf: ]]) ifTrue: [
			self note: 'Step in method calling block ' ,each asLiteral ,'
	Method ' ,aMethod ,' needs edition to run.
	Edit inline expressions to avoid block activations'
	] ].

	script := self evaluationScript doing: [:simulator|
		aReceiver
			basicPerform: aMethod selector asSelector
			withArguments: array
	].
	^script stepsFor: aReceiver entering: aMethod withArguments: array ! !

! S8Simulator methodsFor: #shortcuts !
shortcut: aClass message: selector for: anObject withArguments: values
	" Private - Returns an action to handle an optimized message send (or nil). "

	| pool pair |
	(self isObject: anObject) ifFalse: [ ^nil ].
	pool := self shortcuts.
	pool isEmpty ifTrue: [
		self triggerEvent: #getShortcuts: with: pool.
		pool isEmpty ifTrue: [ self registerDefaultShortcuts ].
	].
	pair := Array with: anObject with: values.
	^((pool at: selector ifAbsent: [#()])
	 	detect: [:assoc| assoc key evaluateWithArguments: pair ]
		ifNone: [ #halt -> nil ]) value ! !

! S8Simulator methodsFor: #breakpoints !
break: why
	" Breakpoint hit. "

	self triggerEvent: #breakpoint: with: why.
	state := #paused! !

! S8Simulator methodsFor: #breakpoints !
testPoint: test
	" Private - Test matching of a breakpoint. "

	(test isNil or: [breakpoints size = 0]) ifTrue: [ ^self ].
	(breakpoints includes: test) ifTrue: [ ^self break: test ]. "method selector or cls"
	(test isKindOf: Behavior) ifTrue: [
		breakpoints do: [:each|
			((each isArray and: [each includes: test]) "#(cls1 cls2 cls3)"
			or: [ (each isKindOf: Behavior) and: [ test inheritsFrom: each ]]) "cls w/subclasses"
			ifTrue: [ ^self break: test ].
		].	^self
	].
	test isString ifTrue: [
		breakpoints do: [:each| "sel*"
			((each isString and: [each endsWith: $*])
			and: [test startsWith: (each upTo: $*)]) ifTrue: [
				^self break: test
			]
		]
	]! !

! S8Simulator methodsFor: #execute !
activateBasic: anObject message: selector withArguments: values
	" Private - Returns true if selector sent to anObject has been activated as basic message sent. "

	(anObject isKindOf: SimulationBlock) ifTrue: [
		self value: (anObject
			perform: selector
			withArguments: values).
		^true
	].
	selector = #halt ifTrue: [
		self value: anObject; pause.
		^true
	].
	selector = #|| ifTrue: [
		self value: anObject || values first.
		^true
	].
	^false! !

! S8Simulator methodsFor: #execute !
callNative: anObject message: selector withArguments: values
	" Private - Returns true if selector sent to anObject has been activated as native call. "

	| mth |
	{'self' jl' mth=eval(Symbol(asSelector(selector))) '}.
	mth isNil ifTrue: [
		"self note: 'proto-lookup for ',selector asSeletor."
		mth := anObject basicAt: selector asSelector.
	].
	(mth notNil and: [ mth respondsTo: #valueWithArguments: ]) ifTrue: [
		self value: (self callNative: mth on: anObject withArguments: values).
		^true
	].
	(self isObject: anObject) ifFalse: [
		^self error: 'Native object DNU ' ,selector asLiteral
	].
	^false! !

! S8Simulator methodsFor: #execute !
activateNative: anObject message: selector withArguments: values
	" Private - Returns true if selector sent to anObject has been activated as native call. "

	^self
		triggerEvent: #activateNative:
		with: (Message receiver: anObject selector: selector arguments: values)
		ifNotHandled: [
			self callNative: anObject message: selector withArguments: values
		] ! !

! S8Simulator methodsFor: #execute !
callDNU: anObject message: selector withArguments: values
	" Private - Call DNU error condition. "

	| mth array |
	mth := anObject class >>> #dnu:withArguments:.
	mth isNil ifTrue: [
		mth := anObject class >>> #doesNotUnderstand:.
		array := Array with: selector.
	] ifFalse: [
		array := Array with: selector with: values
	].
	self callMethod: mth for: anObject withArguments: array ! !

! S8Simulator methodsFor: #execute !
activate: aClass message: selector for: anObject withArguments: values
	" Activate a method context simulation for sending message to anObject with given arguments. "

	| mth |
	self testPoint: ((aClass isNil and: [self isObject: anObject])
		ifTrue: [ anObject class ] ifFalse: [aClass]).
	self testPoint: selector.
	(self	activateBasic: anObject
		message: selector withArguments: values)
		ifTrue: [ ^self ].
	aClass notNil ifTrue: [
		mth := self shortcut: aClass message: selector for: anObject withArguments: values.
		mth notNil ifTrue: [
			^mth evaluateWithArguments: (Array
				with: anObject with: selector
				with: values with: aClass)
		].
		mth := aClass >>> selector.
	].
	(aClass notNil and: [mth isNil]) ifTrue: [
		(self	activateNative: anObject
			message: selector withArguments: values)
			ifTrue: [ ^self ]
	].
	mth isNil ifTrue: [
		^self	callDNU: anObject
			message: selector
			withArguments: values
	].
	self testPoint: mth.
	^self callMethod: mth for: anObject withArguments: values ! !

! S8Simulator methodsFor: #execute !
callNativeLua: evaluable on: anObject withArguments: values
	" Private - Returns the result of evaluating an evaluable function on anObject. "

	^evaluable valueWithArguments: (Array with: anObject) ,values! !

! S8Simulator methodsFor: #execute !
callNativeJulia: evaluable on: anObject withArguments: values
	" Private - Returns the result of evaluating an evaluable function on anObject. "

	(anObject isNil and: [evaluable isMemberOf: BlockClosure]) ifTrue: [
		^{'' jl'evaluable(handle(values)...)'}
	].

"DEBUG-"	self note: '[S8Simulator] native-oo call with ',values asLiteral.
	^evaluable valueWithArguments: (Array with: anObject) ,values! !

! S8Simulator methodsFor: #execute !
callNative: evaluable on: anObject withArguments: values
	" Private - Returns the result of evaluating an evaluable function on anObject.
	e.g.	Smalltalk current classes
	see also	Behavior>>#allSubclassesDo:
	"

	self nativeLanguage = #lua ifTrue: [ "transform to a lua oo call"
		^self callNativeLua: evaluable on: anObject withArguments: values
	].
	self nativeLanguage = #jl ifTrue: [
		^self callNativeJulia: evaluable on: anObject withArguments: values
	].
	^evaluable valueWithArguments: values! !

! S8Simulator methodsFor: #execute !
send: selector to: anObject withArguments: values
	" Add message send to anObject with given arguments. "

	^self	activate: anObject class message: selector
		for: anObject withArguments: values ! !

! S8Simulator methodsFor: #execute !
repeat: aSimulationBlock while: condition doing: action then: epilogue
	" Private - Execute evaluations in a loop. "

	self add: aSimulationBlock node step: [
		condition value ifTrue: [
			self add: aSimulationBlock node step: [
				self triggerEvent: #aboutToLoop: with: aSimulationBlock.
				self repeat: aSimulationBlock while: condition doing: action then: epilogue
			] in: aSimulationBlock context.
			self add: aSimulationBlock node step: [
				action notNil ifTrue: [ action value ].
			] in: aSimulationBlock context.
		] ifFalse: [ epilogue value ]
	] in: aSimulationBlock context.
	aSimulationBlock value ! !

! S8Simulator methodsFor: #shortcuts !
ignoreShortcuts
	" Private - Set the receiver to do NOT use shortcuts. "

	| pool |
	pool := PoolDictionary new.
	pool at: #halt put: #().
	self shortcuts: pool ! !

! S8Simulator methodsFor: #shortcuts !
useShortcuts
	" Private - Set the receiver to use shortcuts. "

	self shortcuts: nil ! !

! S8Simulator methodsFor: #shortcuts !
register: selector shortcut: assoc
	" Private - Register the shortcut for selector with assoc. "

	(self shortcuts
		at: selector ifAbsentPut: [#()])
		add: assoc ! !

! S8Simulator methodsFor: #shortcuts !
registerInlined: messages
	" Private - Register messages to be inlined. "

	messages do: [:selector|
		self register: selector shortcut:
			[:receiver :args| receiver respondsTo: selector ]
			-> [:receiver :selector :args :cls|
				self value: (receiver perform: selector withArguments: args)
			]
	] ! !

! S8Simulator methodsFor: #shortcuts !
registerInlined: messages for: classOrArray
	" Private - Register messages to be inlined. "

	| test |
	classOrArray isClass ifTrue: [
		test := [:receiver :args| receiver isKindOf: classOrArray ]
	] ifFalse: [
		test := [:receiver :args|
			(classOrArray detect: [:cls|
				receiver isKindOf: cls
			] ifNone: []) notNil
		]
		
	].
	messages do: [:selector|
		self register: selector shortcut:
			test -> [:receiver :selector :args :cls|
				self value: (receiver perform: selector withArguments: args)
			]
	] ! !

! S8Simulator methodsFor: #shortcuts !
registerDefaultShortcuts
	" Private - Register the default shortcuts. "

	self	registerInlined: #(
			json handle size asLiteral
			@ -> asNativeSelector
			toString );
		register: #yourself shortcut: true -> [:receiver | self value: receiver ];
		register: #timesRepeat: shortcut: self timesRepeatShortcut;
		register: #do: shortcut: self doShortcut;
		register: #reverseDo: shortcut: self reverseDoShortcut;
		register: #withIndexDo: shortcut: self withIndexDoShortcut;
		register: #to:do: shortcut: self toDoShortcut;
		register: #to:by:do: shortcut: self toByDoShortcut;
		register: #keysAndValuesDo: shortcut: self keysAndValuesDoShortcut;
		register: #ifTrue: shortcut: self ifTrueShortcut;
		register: #ifFalse: shortcut: self ifFalseShortcut;
		register: #at:ifAbsent: shortcut: self atIfAbsentShortcut;
		register: #ifTrue:ifFalse: shortcut: self ifTrueIfFalseShortcut;
		register: #ifFalse:ifTrue: shortcut: self ifFalseIfTrueShortcut;
		register: #indexOf:ifAbsent: shortcut: self indexOfIfAbsentShortcut;
		register: #sorted: shortcut: self sortedShortcut;
		register: #sort: shortcut: self sortShortcut;
		yourself ! !

! S8Simulator methodsFor: #shortcuts !
timesRepeatShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		receiver isNumber and: [
		args last isKindOf: SimulationBlock ] ]
	-> [:receiver :selector :args :cls| | i blk loop |
		blk := args last. i := 1.
		self add: #timesRepeat-exit step: [ self value: receiver ] in: nil.
		loop := [ i < receiver ifTrue: [ self add: #timesRepeat-loop step: loop in: nil ].
			blk value: i. i := i + 1.
		].
		i <= receiver ifTrue: [ self add: #timesRepeat-loop step: loop in: nil ].
	] ! !

! S8Simulator methodsFor: #shortcuts !
doShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		(receiver isKindOf: SequenceableCollection) and: [
		args last isKindOf: SimulationBlock ] ]
	-> [:receiver :selector :args :cls| | i to blk loop |
		to := receiver size. blk := args last. i := 1.
		self add: #do-exit step: [ self value: receiver ] in: nil.
		loop := [ i < to ifTrue: [ self add: #do-loop step: loop in: nil ].
			blk value: (receiver at: i). i := i + 1.
		].
		i <= to ifTrue: [ self add: #do-loop step: loop in: nil ].
	] ! !

! S8Simulator methodsFor: #shortcuts !
reverseDoShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		(receiver isKindOf: SequenceableCollection) and: [
		args last isKindOf: SimulationBlock ] ]
	-> [:receiver :selector :args :cls| | i blk loop |
		i := receiver size. blk := args last.
		self add: #reverseDo-exit step: [ self value: receiver ] in: nil.
		loop := [ i > 1 ifTrue: [ self add: #reverseDo-loop step: loop in: nil ].
			blk value: (receiver at: i). i := i - 1.
		].
		i > 0 ifTrue: [ self add: #reverseDo-loop step: loop in: nil ].
	] ! !

! S8Simulator methodsFor: #shortcuts !
withIndexDoShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		(receiver isKindOf: SequenceableCollection) and: [
		args last isKindOf: SimulationBlock ] ]
	-> [:receiver :selector :args :cls| | i to blk loop |
		i := 1. to := receiver size. blk := args last.
		self add: #withIndexDo-exit step: [ self value: receiver ] in: nil.
		loop := [ i < to ifTrue: [ self add: #withIndexDo-loop step: loop in: nil ].
			blk value: (receiver at: i) value: i. i := i + 1. 
		].
		i <= to ifTrue: [ self add: #withIndexDo-loop step: loop in: nil ].
	] ! !

! S8Simulator methodsFor: #shortcuts !
keysAndValuesDoShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		(receiver isKindOf: KeyedCollection) and: [
		args last isKindOf: SimulationBlock ] ]
	-> [:receiver :selector :args :cls| | keys blk loop |
		keys := receiver keys asArray copy. blk := args last.
		self add: #keysAndValuesDo-exit step: [ self value: receiver ] in: nil.
		loop := [ keys size > 1 ifTrue: [ self add: #keysAndValuesDo-loop step: loop in: nil ].
			blk value: keys first value: (receiver at: keys removeFirst).
		].
		keys notEmpty ifTrue: [ self add: #keysAndValuesDo-loop step: loop in: nil ].
	] ! !

! S8Simulator methodsFor: #shortcuts !
toDoShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		receiver isNumber and: [
		args first isNumber and: [
		args last isKindOf: SimulationBlock ]] ]
	-> [:receiver :selector :args :cls| | i to blk loop |
		to := args first. blk := args last. i := receiver.
		self add: #toDo-exit step: [ self value: receiver ] in: nil.
		loop := [ i < to ifTrue: [ self add: #toDo-loop step: loop in: nil ].
			blk value: i. i := i + 1.
		].
		i <= to ifTrue: [ self add: #toDo-loop step: loop in: nil ].
	] ! !

! S8Simulator methodsFor: #shortcuts !
toByDoShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		receiver isNumber and: [
		args first isNumber and: [ args second isNumber and: [
		args last isKindOf: SimulationBlock ]]] ]
	-> [:receiver :selector :args :cls| | i to increment blk loop min max |
		to := args first. increment := args second. blk := args last. i := receiver.
		min := receiver min: to. max := receiver max: to.
		self add: #toByDo-exit step: [ self value: receiver ] in: nil.
		loop := [	(i + increment between: min and: max) ifTrue: [
				self add: #toByDo-loop step: loop in: nil
			].
			blk value: i. i := i + increment. 
		].
		(increment > 0
			ifTrue: [ i < to ]
			ifFalse: [ i > to ])
		ifTrue: [ self add: #toByDo-loop step: loop in: nil ].
	] ! !

! S8Simulator methodsFor: #shortcuts !
ifTrueShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		receiver isBoolean and: [
		args last isKindOf: SimulationBlock ] ]
	-> [:receiver :selector :args :cls|
		receiver == true
		ifTrue: [ args last value ]
		ifFalse: [ self value: nil ]
	] ! !

! S8Simulator methodsFor: #shortcuts !
ifFalseShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		receiver isBoolean and: [
		args last isKindOf: SimulationBlock ] ]
	-> [:receiver :selector :args :cls|
		receiver == false
		ifTrue: [ args last value ]
		ifFalse: [ self value: nil ]
	] ! !

! S8Simulator methodsFor: #shortcuts !
atIfAbsentShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args| args last isKindOf: SimulationBlock ]
	-> [:receiver :selector :args :cls| | failed result |
		failed := Object new.
		result := receiver at: args first ifAbsent: [ failed ].
		failed == result ifTrue: [
			self add: #atIfAbsent-failed step: [ args last value ] in: nil
		] ifFalse: [ self value: result ]
	] ! !

! S8Simulator methodsFor: #shortcuts !
ifTrueIfFalseShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		receiver isBoolean and: [
		(args last isKindOf: SimulationBlock) and: [
		args first isKindOf: SimulationBlock ]] ]
	-> [:receiver :selector :args :cls|
		receiver == true
		ifTrue: [ args first value ]
		ifFalse: [ args last value ]
	] ! ! 

! S8Simulator methodsFor: #shortcuts !
ifFalseIfTrueShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		receiver isBoolean and: [
		(args last isKindOf: SimulationBlock) and: [
		args first isKindOf: SimulationBlock ]] ]
	-> [:receiver :selector :args :cls|
		receiver == false
		ifTrue: [ args first value ]
		ifFalse: [ args last value ]
	] ! !

! S8Simulator methodsFor: #shortcuts !
indexOfIfAbsentShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args| receiver isKindOf: SequenceableCollection  ]
	-> [:receiver :selector :args :cls| | failed result |
		failed := Object new.
		result := receiver indexOf: args first ifAbsent: [ failed ].
		failed == result ifTrue: [
			self add: #indexOfIfAbsent-failed step: [ args last value ] in: nil
		] ifFalse: [ self value: result ]
	] ! !

! S8Simulator methodsFor: #shortcuts !
sortedShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		(receiver isKindOf: SequenceableCollection) and: [
		args first isKindOf: SimulationBlock ] ]
	-> [:receiver :selector :args :cls| self value: (receiver sorted: args first asBlock) ] ! !

! S8Simulator methodsFor: #shortcuts !
sortShortcut
	" Private - Returns the shortcut for the selector. "

	^[:receiver :args|
		(receiver isKindOf: SequenceableCollection) and: [
		args first isKindOf: SimulationBlock ] ]
	-> [:receiver :selector :args :cls| self value: (receiver sort: args first asBlock) ] ! !


