EventManager
 subclass: #SwikiCodeRobot
 instanceVariableNames: 'options categories loader verbose deferred aborted currentPages chunkStream'
 classVariableNames: 'Fetch'
 category: #swiki!
SwikiCodeRobot comment: '
	A robot walking swiki pages.
	The default behavior will select pages in the swiki by category,
	 sort the pages and evaluate/process code chunks in the pages.
	It can be used for testing, API validation, etc.
	See pages in s8-media swiki for more information.
	Use cases:
		(SwikiCodeRobot @ #test) process: #s8-media tagged: #metatest
		(SwikiCodeRobot @ #api) process: #s8-media tagged: #( #apple #WebKit )
'!
SwikiCodeRobot buildAccessors: #(
	#(options initialize 'self defaultOptions')
	#(categories initialize 'self defaultCategories')
	#(verbose initialize #true)
	#(deferred initialize 'self defaultDeferred')
	#(aborted initialize #false)
)!

! SwikiCodeRobot class methodsFor: #constants !
fetch: aBlock
	" Set the fetch block to get contents.
	aBlock will be a 2 arguments block:
		- the URL to access
		- the action to perform whith contents)
	Note that fetch block will set the robot to deferred mode.
	"

	Fetch := aBlock! !

! SwikiCodeRobot class methodsFor: #constants !
fetch
	" Returns the fetch block (or nil). "

	^Fetch! !

! SwikiCodeRobot class methodsFor: #instantiation !
@ what
	" Return an instance of the receiver. "

	^self new
		when: #needsCategories do: [
			what isString
			ifTrue: [ what asArrayOfSubstrings ]
			ifFalse: [ what asArray ]
		];
		yourself! !

! SwikiCodeRobot class methodsFor: #instantiation !
for: what
	" Return an instance of the receiver. "

	^self @ what! !

! SwikiCodeRobot class methodsFor: #process !
process: where
	" Process the contents at where. "

	^self @> where! !

! SwikiCodeRobot class methodsFor: #process !
@> where
	" Process the contents at where.
	See instance side comment.
	"

	^self new @> where! !

! SwikiCodeRobot methodsFor: #process !
@> where
	" Process the contents at where.
	where can be:
		aNumber or aString pointing to a page to process.
		anArray as
			#(swikiName category tag)
			#(category tag)
			#(page)
	"

	(where isNumber or: [ where isString ]) ifTrue: [
		^self processPageAt: where
	].
	where isArray ifFalse: [ ^self error: 'Invalid argument.' ].
	where size = 1 ifTrue: [ ^self processPageAt: where first ].
	where size = 2 ifTrue: [ ^self @> ((Array with: self defaultSwiki),where) ].
	^(self class @ where second) process: where first tagged: where last! !

! SwikiCodeRobot class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^#(	#evaluate: #running:
		#testLine: #needsActionLine: #endLines
		#startTimes: #endTimes: 
		#needsOptions #needsCategories
		#sortPages: #needsCommandsIn:
		#load: #needsFileContents:
		#startPages: #startPage:stream: #endPage:
	) , super constructEventsTriggered! !

! SwikiCodeRobot methodsFor: #defaults !
defaultSwikiOptions
	" Private - Return the literals used for swiki configuration. "

	| swikiHost swikiPort |
	swikiPort := 80.
	swikiHost := 'swiki.smalltalking.net'.
	(Smalltalk includesKey: #U8Object) ifTrue: [ | home |
		home := (U8Object swikiHostURL fromLast: '://') upTo: $/.
		(home includes: $:) ifTrue: [
			swikiPort := (home fromLast: $:) asNumber.
		].
		swikiHost := home upTo: $:
	].
	^(#host -> swikiHost) ,(#port -> swikiPort)! !

! SwikiCodeRobot methodsFor: #defaults !
defaultOptions
	" Private - Return the literals used for default configuration. "

	^self triggerEvent: #needsOptions ifNotHandled: [
		self defaultSwikiOptions
	]! !

! SwikiCodeRobot methodsFor: #defaults !
defaultSwiki
	" Private - Return the default swiki name. "

	^#s8-media! !

! SwikiCodeRobot methodsFor: #defaults !
defaultDeferred
	" Private - Return true if the receiver run in deferred mode. "

	^self hasFetch or: [ self hasNodeJS ]! !

! SwikiCodeRobot methodsFor: #defaults !
defaultCategories
	" Private - Return the literals used to filter pages of interest.
	The default answer is an empty array (all pages starting with [ are of interest).
	The subclasses can refine the result to select pages marked with a category before the [ (tag) field.
	"

	^self triggerEvent: #needsCategories ifNotHandled: [ Array new ]! !

! SwikiCodeRobot methodsFor: #testing !
hasFetch
	" Private - Returns true if the execution environment provides a fetch function. "

	^self class fetch notNil! !

! SwikiCodeRobot methodsFor: #testing !
hasNodeJS
	" Private - Returns true if the execution environment is running NodeJS. "

	| sentinel |
	sentinel := Smalltalk nativeObjectAt: #process ifAbsent: [ ^false ].
	(sentinel basicAt: #version) isNil ifTrue: [ ^false ].
	^true! !

! SwikiCodeRobot methodsFor: #http !
http: httpOptions do: dataBlock ifAbsent: aBlock
	" Private - Evaluates the dataBlock with data obtained accessing server options. "

	| request http |
	self hasFetch ifTrue: [
		^self class fetch
			value: (self urlFromOptions: httpOptions)
			value: dataBlock
	].
	self hasNodeJS ifFalse: [ ^aBlock evaluate ].
	http := # #require: #http. 
	http isNil ifTrue: [ ^aBlock evaluate ].
	request := http #request: httpOptions json doing: [:response| | content |
		content := ''.
		response
			#setEncoding: #utf8;
			#on: #data do: [:chunk| content := content, chunk ].
		response #on: #end do: [ dataBlock evaluateWith: content ].
	].
	request #on: #error do: [:ex| self error: "ex#message ," ex#stack ].
	{'request.end()' lua'self:invalidMessage()'}.
	^nil! !
	
! SwikiCodeRobot methodsFor: #http !
withHttp: httpOptions do: dataBlock
	" Private - Evaluates the dataBlock with data obtained accessing server options. "

	| request http |
	self hasFetch ifTrue: [
		^self class fetch
			value: (self urlFromOptions: httpOptions)
			value: dataBlock
	].
	self hasNodeJS ifFalse: [
		^self error: 'Accessing swiki using HTTP - Require Node.js execution environment.'
	].
	http := # #require: #http.
	request := http #request: httpOptions json doing: [:response| | content |		
		content := ''.
		response
			#setEncoding: #utf8;
			#on: #data do: [:chunk| content := content, chunk ].
		response #on: #end do: [ dataBlock evaluateWith: content ].
	].
	request #on: #error do: [:ex| self error: "ex#message ," ex#stack ].	
	{'request.end()' lua'self:invalidMessage()'}.
	^nil! !

! SwikiCodeRobot methodsFor: #utilities !
urlFromOptions: theOptions
	" Private - Returns the url set in theOptions. "

	^'http://'
		,(theOptions @ #host)
		,':' ,(theOptions @ #port)
		,'/' ,(theOptions @ #path)! !

! SwikiCodeRobot methodsFor: #utilities !
withOptions: theOptions do: dataBlock
	" Private - Evaluates the dataBlock with data obtained accessing server options. "

	self http: theOptions do: dataBlock ifAbsent: [
		| text |
		text := self fileContentsAt: (self urlFromOptions: theOptions).
		dataBlock value: text
	]! !

! SwikiCodeRobot methodsFor: #utilities !
withSwiki: swikiName page: pageNumber do: dataBlock
	" Private - Evaluates the dataBlock with data obtained from swikipage. "

	self aborted ifTrue: [ ^nil ].
	self	withOptions: self options
			,(#method -> 'GET')
			,(#path -> (swikiName,'/',pageNumber))
			,(#agent ->false)
		do: dataBlock! !

! SwikiCodeRobot methodsFor: #tags !
is: tag includedIn: tags
	" Private - Return true if the tag matches tags. "

	tag isArray ifTrue: [
		tag	detect: [:one| (self is: one includedIn: tags) not ]
			ifNone: [ ^true ].
		^false
	].
	^tags includes: tag! !

! SwikiCodeRobot methodsFor: #tags !
tagSeparator
	" Private - Return the tags separator. "

	^$,! !

! SwikiCodeRobot methodsFor: #tags !
tagsFrom: aStream
	" Private - Return the tags read from title stream. "

	^(self categoryAndTagsFrom: aStream) last! !

! SwikiCodeRobot methodsFor: #tags !
tagsInTitle: title
	" Private - Return the tags read from title string. "

	^(self categoryAndTagsInTitle: title) last! !

! SwikiCodeRobot methodsFor: #tags !
categoryAndTagsFrom: aStream
	" Private - Return the tags read from title stream. "

	| category |
	category := (aStream upTo: $[) trimBlanks.
	(self categories notEmpty
		and: [ (self categories includes: category) not])
		ifTrue: [ ^Array with: category with: #() ].
	^Array	with: category
		with: ((((aStream upTo: $])
			asArrayOfSubstringsSeparatedBy: self tagSeparator)
			collect: [:each| each trimBlanks ])
			select: [:each| each notEmpty ])! !

! SwikiCodeRobot methodsFor: #tags !
categoryAndTagsInTitle: title
	" Private - Return an array with category and the tags read from title string. "

	^self categoryAndTagsFrom: title stream! !

! SwikiCodeRobot methodsFor: #tags !
title: indexEntry matches: tag
	" Private - Return true if the indexEntry matches tag. "

	| tags |
	tags := self tagsInTitle: indexEntry last.
	tags isEmpty ifTrue: [ ^false ].
	tag isNil ifTrue: [ ^true ].
	^self is: tag includedIn: tags! !

! SwikiCodeRobot methodsFor: #tags !
pagesIn: stream matching: tag
	" Private - Returns the pages defined in index stream matching tag.
	The default implementation parse index stream selecting pages matching tag field.
	"

	| result line pageNumber boolean title |
	result := Array new.
	[stream atEnd] whileFalse: [
		pageNumber := (stream upTo: String tab) asNumber.
		boolean := stream upTo: String tab.
		title := stream nextLine trimBlanks.
		result add: (Array with: pageNumber with: title)
	].
	^self sortedPages: (result select: [:entry| self title: entry matches: tag ])! !

! SwikiCodeRobot methodsFor: #private !
htmlEncodings
	" Private - Return the encoded characters table. "

	^#(
		#( '&lt;' #< )
		#( '&gt;' #> )
		#( '&amp;' #& )
	)! !

! SwikiCodeRobot methodsFor: #private !
nextChunkIn: stream
	" Return the next chunk of code in stream. "

	| open tag close bar code |
	open := #<. close := #>. bar := #/. tag := #pre.
	code := stream upTo: open,tag,close; upTo: open,bar,tag,close.
	^self htmlEncodings inject: code
		into: [:total :each| total replace: each first with: each second ]! !

! SwikiCodeRobot methodsFor: #lineTests !
lineTest: expression
	" Private - Evaluate the test expression. "

	| result |
	result := expression stream fileInto: self.
	true = result ifTrue: [ ^#ok ].
	false = result ifTrue: [ ^self note: 'FAILED - ' ,expression ].
	^result! !

! SwikiCodeRobot methodsFor: #lineTests !
running: aTitle
	" Test - Notify we are running test aTitle. "

	self triggerEvent: #running: with: aTitle ifNotHandled: [
		self note: aTitle.
	].
	^true! !

! SwikiCodeRobot methodsFor: #lineTests !
testLine: expression
	" Private - Evaluate the test expression. "

	self triggerEvent: #testLine: with: expression ifNotHandled: [
		self lineTest: expression
	]! !

! SwikiCodeRobot methodsFor: #timeTests !
startTimes: code
	" Private - Start tracking time evaluating line tests. "

	self triggerEvent: #startTimes: with: self savedTimes ! !

! SwikiCodeRobot methodsFor: #timeTests !
endTimes: code
	" Private - Finish tracking time evaluating line tests. "

	self triggerEvent: #endTimes: with: self savedTimes ! !

! SwikiCodeRobot methodsFor: #timeTests !
resetTimes
	" Reset the times collected. "

	^self savedTimes removeAll ! !

! SwikiCodeRobot methodsFor: #timeTests !
savedTimes
	" Return the collected times spent evaluating line tests. "

	^self basicAt: '$savedTimes' ifAbsentPut: [PoolDictionary new] ! !

! SwikiCodeRobot methodsFor: #timeTests !
jlTime: action count: count
	" Private - Returns the time spent evaluating expression in a loop (count times). "

	| blk |
	blk := self callableOf: action.
	{''jl' Base.invokelatest(blk) '}.
	Smalltalk gc.
	^[ count timesRepeat: [
		{''jl' Base.invokelatest(blk) '}.
	] ] ms ! !

! SwikiCodeRobot methodsFor: #timeTests !
ms: expression count: count
	" Private - Returns the time spent evaluating expression in a loop (count times).
	Note expression must be a valid statements expression.
	"

	| nativeSelector |
	nativeSelector := (self nativeLanguage ,'Time:count:') asSymbol.
	(self respondsTo: nativeSelector) ifTrue: [ | action |
		action := ('[ ' ,expression ,' ]') doIt.
		^self perform: nativeSelector withArguments: (Array
			with: action with: count)
	].
	^('Smalltalk gc. ['
	,count ,' timesRepeat: [' ,expression ,']
	]ms rounded') doIt! !

! SwikiCodeRobot methodsFor: #timeTests !
saveTime: expression
	" Private - Track time evaluating expression. "

	| count stream |
	count := 1000.
	stream := expression stream.
	(stream peekForAll: '##') ifTrue: [ count := Number readFrom: stream ].
	stream skipSeparators.
	stream atEnd ifTrue: [ ^nil ].
	self savedTimes at: expression put: (self ms: stream upToEnd count: count) ! !

! SwikiCodeRobot methodsFor: #timeTests !
timesAction: code
	" Private - Return the action to evaluate to track time spent evaluating line tests. "

	self startTimes: code.
	^[:expression|
		expression isNil
		ifTrue: [ self endTimes: code ]
		ifFalse: [ self saveTime: expression ]
	]! !

! SwikiCodeRobot methodsFor: #lineTests !
testLines: code in: page
	" Private - Evaluate the test lines in code stream. "

	| action |
	action := self triggerEvent: #needsActionLine: with: code ifNotHandled: [
		(code peekForAll: #times) ifTrue: [ self timesAction: code]
	].
	[code atEnd] whileFalse: [ | line |
		code skipSeparators.
		line := code peek = $!!
			ifTrue: [ code next; nextChunk ]
			ifFalse: [ code nextLine upTo: $" ].
		line trimBlanks notEmpty ifTrue: [
			action isNil
			ifTrue: [ self testLine: line ]
			ifFalse: [ action evaluateWith: line ]
		]
	].
	action notNil ifTrue: [ action evaluateWith: nil ].
	self triggerEvent: #endLines! !

! SwikiCodeRobot methodsFor: #private !
fileInCode: code in: page
	" Private - FileIn the code in code string. "

	| codeStream commentStart |
	codeStream := code stream.
	codeStream peek = $. ifTrue: [ "code starting with $. enter line mode"
		codeStream next.
		^self testLines: codeStream in: page
	].

	commentStart := $  asComment upTo: $ .
	(codeStream peekForAll: commentStart) ifTrue: [ "native eval mode"
		self note: codeStream nextLine.
		^Smalltalk current eval: codeStream upToEnd
	].

	codeStream peek = $" ifTrue: [
		self note: (codeStream next; upTo: $")
	].
	^codeStream fileInto: self! !

! SwikiCodeRobot methodsFor: #private !
evaluateCode: code in: page
	" Private - Evaluate the code in code string.
	The default action fileIn the code.
	"

	[	self	triggerEvent: #evaluate: with: code
			ifNotHandled: [ self fileInCode: code in: page ]
	] on: Error do: [:ex|
		self note: 'Error evaluating code in '
			,page printString ,':'
			,ex messageText
	]! !

! SwikiCodeRobot methodsFor: #private !
processPage: page contents: stream
	" Private - Process the page contents (in stream). "

	chunkStream := stream.
	self	triggerEvent: #startPage:stream:
		withArguments: (Array with: page with: chunkStream).
	[stream atEnd] whileFalse: [ | chunk |
		chunk := self nextChunkIn: stream.
		self evaluateCode: chunk in: page.
	].
	self triggerEvent: #endPage: with: page! !

! SwikiCodeRobot methodsFor: #private !
processPagesDeferred: pages in: swikiName
	" Private - Process the pages found in swiki swikiName. "

	| current |
	pages isEmpty ifTrue: [ ^self ].
	self deferred ifFalse: [ ^self processPages: pages in: swikiName ].
	current := pages removeFirst.
	self withSwiki: swikiName page: current first do: [:data|
		self verbose ifTrue: [ self note: current last printString ].
		data isNil ifTrue: [ self error: 'Can''t read contents at ',current last printString ].
		self processPage: current contents: data stream.
		[ self processPages: pages in: swikiName ] valueDeferred: 300
	]! !

! SwikiCodeRobot methodsFor: #private !
processPagesIterative: pages in: swikiName
	" Private - Process the pages found in swiki swikiName. "

	| current |
	[pages notEmpty and: [ self deferred not ]] whileTrue: [
		current := pages removeFirst.
		self withSwiki: swikiName page: current first do: [:data|
			self verbose ifTrue: [ self note: current last printString ].
			data isNil ifTrue: [ self error: 'Can''t read contents at ',current last printString ].
			self processPage: current contents: data stream.
		].
	].
	pages notEmpty ifTrue: [
		self processPages: pages in: swikiName
	]! !

! SwikiCodeRobot methodsFor: #private !
processPages: pages in: swikiName
	" Private - Process the pages found in swiki swikiName. "

	currentPages := pages.
	self triggerEvent: #startPages: with: currentPages.
	^self deferred
		ifTrue: [ self processPagesDeferred: pages in: swikiName ]
		ifFalse: [ self processPagesIterative: pages in: swikiName ]! !

! SwikiCodeRobot methodsFor: #sorting !
sortedPages: entries
	" Private - Return the page entries sorted by scanning precedence. "

	^self	triggerEvent: #sortPages: with: entries
		ifNotHandled: [ self defaultSortedPages: entries ]! !

! SwikiCodeRobot methodsFor: #commands !
commandsIn: title
	" Private - Return the commands obtained from title. "

	^self triggerEvent: #needsCommandsIn: with: title ifNotHandled: [
		self defaultCommandsIn: title
	]! !

! SwikiCodeRobot methodsFor: #commands !
defaultCommandsIn: title
	" Private - Return the commands obtained from title. "

	^(title asArrayOfSubstrings inject: ''
		into: [:total :each| total ,each ,' ' ]) trimBlanks! !

! SwikiCodeRobot methodsFor: #sorting !
prologueKeys
	" Private - Return the commands used to sort pages. "

	^#(	prologue
		home index
		definitions definition
	)! !

! SwikiCodeRobot methodsFor: #sorting !
epilogueKeys
	" Private - Return the commands used to sort pages. "

	^#(	body methods
		code doIt evaluate
		fail epilogue
	)! !

! SwikiCodeRobot methodsFor: #sorting !
defaultSortedPages: entries
	" Private - Return the page entries sorted by default scanning precedence. "

	| map result add addResult epilogue addEpilogue |
	result := Array new.
	epilogue := Array new. 
	map := Dictionary new.
	entries do: [:tuple| | title |
		title := tuple last stream upTo: $]; upToEnd.
		map at: (self commandsIn: title) put: tuple
	].
	add := [:key :where| (map includesKey: key) ifTrue: [
			where add: (map at: key).
			map removeKey: key
		].
	].
	addResult := [:key| add value: key value: result ].
	addEpilogue := [:key| add value: key value: epilogue ].

	self prologueKeys do: [:each| addResult value: each ].
	self epilogueKeys do: [:each| addEpilogue value: each ].

	map keys sorted do: [:key| result add: (map at: key) ].
	result := result ,epilogue.

	"self note: 'Sorted pages = ' ,result printString."
	^result! !

! SwikiCodeRobot methodsFor: #executing !
processPageAt: url
	" Process code in the swiki page at url. "

	url isNumber ifTrue: [ ^self processPageAt: url printString ].
	(url includes: $/) ifFalse: [ | swikiName page |
		swikiName := self defaultSwiki.
		page := Array with: url with: url.
		^self	withSwiki: swikiName page: page first
			do: [:data| | pages |
				pages := Array with: page.
				self processPages: pages in: swikiName
			]
	].

	^self	processPage: (Array with: url)
		contents: (self fileNameAt: url) fileContents stream! !

! SwikiCodeRobot methodsFor: #executing !
process: swikiName
	" Access the swiki to process code from pages. "

	^self process: swikiName tagged: nil! !

! SwikiCodeRobot methodsFor: #executing !
process: swikiName tagged: tag
	" Access swiki to run code from pages matching tag. "

	self	withPagesIn: swikiName tagged: tag
		do: [:pages| self processPages: pages in: swikiName ].
	^#Ok! !

! SwikiCodeRobot methodsFor: #executing !
withPagesIn: swikiName tagged: tag do: aBlock
	" Evaluate aBlock with the pages matching tag in swikiName. "

	self withSwiki: swikiName page: '.pages' do: [:data| | pages |
		data isNil ifTrue: [ self error: 'Can''t read pages of ',swikiName ].
		pages := self pagesIn: data stream matching: tag.
		aBlock evaluateWith: pages
	]! !

! SwikiCodeRobot methodsFor: #executing !
withPagesIn: swikiName category: categ do: aBlock
	" Evaluate aBlock with the pages matching category in swikiName. "

	self withSwiki: swikiName page: '.pages' do: [:data| | pages |
		data isNil ifTrue: [ self error: 'Can''t read pages of ',swikiName ].
		pages := self pagesIn: data stream matchingCategory: categ.		
		aBlock evaluateWith: pages
	]! !

! SwikiCodeRobot methodsFor: #tags !
pagesIn: stream matchingCategory: categ
	" Private - Returns the pages defined in index stream matching category.
	The default implementation parse index stream selecting pages matching category field.
	"

	| result cat pageNumber boolean title tuple tag |
	result := Array new.
	[stream atEnd] whileFalse: [ 
		pageNumber := (stream upTo: String tab) asNumber.
		boolean := stream upTo: String tab.
		title := stream nextLine trimBlanks.
		tuple := self categoryAndTagsFrom:  title stream.
		cat := tuple first.
		tag := tuple second.
		cat = categ ifTrue:[
			result add: (Array with: pageNumber with: cat with: tag)
		]
	].
	^result! !

! SwikiCodeRobot methodsFor: #obsolete !
do: aBlock withPagesIn: swikiName tagged: tag
	" Evaluate aBlock with the pages matching tag in swikiName. "

	^self withPagesIn: swikiName tagged: tag do: aBlock! !

! SwikiCodeRobot methodsFor: #filenames !
u8ServicePathName
	" Private - Returns the root path name for U8 service files. "

	(Smalltalk includesKey: #U8Object) ifTrue: [
		^U8Object home ,'/profile/'
	].
	^'http://u8.smalltalking.net/profile/'! !

! SwikiCodeRobot methodsFor: #filenames !
swikiServicePathName
	" Private - Returns the root path name for swiki service files. "

	^(Smalltalk includesKey: #U8Object)
	ifTrue: [ U8Object swikiHostURL ]
	ifFalse: [ 'http://swiki.smalltalking.net/' ]! !

! SwikiCodeRobot methodsFor: #filenames !
fileNameAt: aPlace
	" Private - Return the complete path for file at aPlace. "

	(aPlace indexOf: #u8:) = 1 ifTrue: [
		^self u8ServicePathName
			,(aPlace copyFrom: 4 to: aPlace size)
	].
	(aPlace indexOf: #swiki:) = 1 ifTrue: [
		^self swikiServicePathName
			,((aPlace copyFrom: 7 to: aPlace size)
				copyWithout: $ )
	].
	^aPlace! !

! SwikiCodeRobot methodsFor: #loading !
load: fileName
	" Private - Load the file at fileName. "

	| index |
	index := fileName lastIndexOf: $. .
	index = 0 ifTrue: [ ^self load: fileName ,'.st.js' ].
	^self triggerEvent: #load: with: fileName ifNotHandled: [
		self basicLoad: (self fileNameAt: fileName)
	]! !

! SwikiCodeRobot methodsFor: #loading !
fileContentsAt: fileName
	" Private - Return the contents of the file at fileName. "

	^self
		triggerEvent: #needsFileContents:
		with: fileName
		ifNotHandled: [ fileName fileContents ]! !

! SwikiCodeRobot methodsFor: #loading !
loadedWithConfiguration: fileName
	" Private - Return true if the file at fileName has been loaded with U8Configuration. "

	(Smalltalk includesKey: #U8Configuration) ifFalse: [ ^false ].
	(U8Configuration respondsTo: #default) ifFalse: [ ^false ].
	^U8Configuration default loadFile: fileName! !

! SwikiCodeRobot methodsFor: #loading !
basicLoad: fileName
	" Private - Load the file at fileName. "

	| ext index code |
	(self loadedWithConfiguration: fileName) ifTrue: [ ^self ].
	code := self fileContentsAt: fileName.
	code isNil ifTrue: [
		^self error: 'Code not found at ', fileName
	].
	index := fileName lastIndexOf: $..
	ext := (fileName copyFrom: index + 1 to: fileName size) asLowercase.
	ext = #st ifTrue: [ ^code stream fileIn ].
	ext = #js ifTrue: [ ^Compiler new eval: code ].
	^self error: 'Invalid file extension ',ext asLiteral! !

! SwikiCodeRobot methodsFor: #tools !
require: names in: places ifAbsent: errorBlock
	" Ensure the globals in names are present.
	If there is a name missing, will try to load code from places
	 until all the names are present.
	Return the objects with that names or the result of evaluating errorBlock.
	"

	| what result remaining |
	names isString ifTrue: [
		what := names asArrayOfSubstrings.
		result := self require: what in: places ifAbsent: errorBlock.
		(result notNil and: [ what size = 1 ]) ifTrue: [ ^result first ].
		^result
	].
	places isString ifTrue: [
		^self
			require: names
			in: (Array with: places)
			ifAbsent: errorBlock
	].
	what := names asArray copy.
	remaining := places collect: [:each| self fileNameAt: each ].
	[what notEmpty and: [remaining notEmpty]] whileTrue: [
		what := what reject: [:each| Smalltalk includesKey: each ].
		what notEmpty ifTrue: [ self load: remaining removeFirst ].
	].
	result := names collect: [:each|
		Smalltalk at: each ifAbsent: [ ^errorBlock evaluateWith: each ]
	].
	^result! !

! SwikiCodeRobot methodsFor: #tools !
require: names ifAbsent: aBlock
	" Ensure the globals in names are present.
	If a name do not exist returns the result of evaluating aBlock
	"

	^self require: names in: #() ifAbsent: aBlock! !

! SwikiCodeRobot methodsFor: #tools !
require: names in: places
	" Ensure the globals in names are present. "

	^self require: names in: places ifAbsent: [:aName|
		self error: 'Missing global #' ,aName
			,(places notEmpty
				ifTrue: [' in ',places printString]
				ifFalse: [''])
	]! !

! SwikiCodeRobot methodsFor: #tools !
require: names
	" Ensure the globals in names are present. "

	^self require: names in: #()! !

! SwikiCodeRobot methodsFor: #executing !
abort
	" Abort current execution. "

	self aborted: true.
	^self cancelFileIn! !

! SwikiCodeRobot methodsFor: #executing !
abortPage
	" Abort execution of the page. "

	chunkStream notNil ifTrue: [ chunkStream upToEnd ].
	^self error: 'Page execution aborted'! !

! SwikiCodeRobot methodsFor: #pages !
selectPages: aBlock
	" Set the pages of the receiver with the pages selected with aBlock. "

	aBlock isString ifTrue: [
		^self selectPages: [:page| page last includes: aBlock ]
	].
	currentPages isNil ifTrue: [ ^self ].
	currentPages removeAll: (currentPages reject: aBlock)! !

! SwikiCodeRobot methodsFor: #pages !
rejectPages: aBlock
	" Set the pages of the receiver with the pages rejected with aBlock. "

	aBlock isString ifTrue: [
		^self rejectPages: [:page| page last includes: aBlock ]
	].
	currentPages isNil ifTrue: [ ^self ].
	currentPages removeAll: (currentPages select: aBlock)! !