
EventManager
	subclass: #SystemChangesManager
	instanceVariableNames: #systemChanges
	category: #SystemServer !

! SystemChangesManager methodsFor: 'tracking-classes' !
classChanged: aBehavior
	" Private - aBehavior has been changed. "

	^self systemHasChanged: (Array with: #classChanged: with: aBehavior)! !

! SystemChangesManager methodsFor: 'tracking-methods' !
methodBound: aMethod to: aClass
	" Private - The method has been bound to aClass. "

	(aMethod isNil or: [ aMethod selector = String doItSelector ]) ifTrue: [ ^self "ignored" ].
	^self systemHasChanged: (Array with: #methodBound:to: with: aMethod with: aClass)! !

! SystemChangesManager methodsFor: 'tracking-methods' !
methodRemoved: aMethod from: aClass
	" Private - The method has been removed from aClass. "

	(aMethod isNil or: [ aMethod selector = String doItSelector ]) ifTrue: [ ^self "ignored" ].
	^self systemHasChanged: (Array with: #methodRemoved:from: with: aMethod with: aClass)! !

! SystemChangesManager methodsFor: 'tracking-methods' !
memoryWarning: anArgument
	" Private - A memory warning event has been thrown. "

	^self systemHasChanged: (Array with: #memoryWarning: with: anArgument)! !

! SystemChangesManager methodsFor: 'system changes' !
processSystemChanges: changes
	" Private - Update reflecting system changes. "

	self note: #processSystemChanges:
		,' ', changes asString
		,' MUST be implemented at '
		,self class name! !

! SystemChangesManager methodsFor: #changes !
changesDo: aOneArgumentBlock
	" Private - Process the changes with aOneArgumentBlock. "
	
	| changes |
	changes := systemChanges.
	(changes notNil and: [changes notEmpty]) ifTrue: [
		systemChanges := nil.
		aOneArgumentBlock value: changes
	].! !

! SystemChangesManager methodsFor: #changes !
processTheChanges
	" Private - Process the changes. "
	
	^self changesDo: [:changes| self processSystemChanges: changes ]! !

! SystemChangesManager methodsFor: 'system changes' !
processSystemChanges
	" Private - The receiver has changes to be processed.
	The default action is to defer proccessing of the changes.
	This method can be overwritten/refined by subclasses to implement alternatives of change management.
	"
	
	[ self processTheChanges ] valueDeferred: 750.! !

! SystemChangesManager methodsFor: 'system changes' !
systemHasChanged: aSystemChange
	" Private - Signal a event of system change to be resolved in the future. "

	systemChanges notNil ifTrue: [
		^systemChanges add: aSystemChange
	].
	systemChanges := Array with: aSystemChange.
	self processSystemChanges.! !

! SystemChangesManager methodsFor: 'system changes' !
bindSystemManager
	" Private - Set the receiver to start listening current system changes.
	Note: Only the events implemented by the receiver will be hooked.
	"

	(SystemManager eventsTriggered copyWithout: #changed) do: [:each|
		(self respondsTo: each) ifTrue: [
			SystemManager current when: each send: each to: self
		]
	]! !

! SystemChangesManager methodsFor: 'system changes' !
unbindSystemManager
	" Private - rest the link from the receiver to system changes. "

	| aManager |
	aManager := SystemManager currentOrNil.
	aManager notNil ifTrue: [ aManager removeAllActionsWithReceiver: self ]! !

SystemChangesManager
	subclass: #SystemServer
	instanceVariableNames: 'server port methodReferences cache'
	category: #SystemServer !

SystemServer class instanceVariableNames: 'current silent platform port'!
SystemServer class buildAccessors: #(
	#( current initialize 'self new' )
	#( silent initialize #true )
	#( platform initialize 'self detectedPlatform' )
	#( port initialize #8088 )
)!
SystemServer buildAccessors: #(
	#( port initialize 'self class port' )
	#( methodReferences initialize 'Smalltalk current methodReferences' )
	#( cache initialize 'PoolDictionary new' )
)!

! SystemServer class methodsFor: #platform!
detectedPlatform
	" Private - Returns the platform we are running (or nil).
	Note that the platform will be used to solve transport functionality and the associated messages MUST be implemented by the receiver.
	"

	#(	#(NodeJS node)
		#(GCDWebServer gcd)
		#(Turbo turbo)
		#(SimpleHTTPServer luaSimple)
	) do: [:tuple| (Smalltalk includesKey: tuple first) ifTrue: [ ^tuple last ] ].
	^self nativePlatform! !

! SystemServer class methodsFor: #platform!
nativePlatform
	" Private - Returns the name of platform based on native language support (or nil).
	Note: detect implementor support prefixed by [nativeLanguage] or [nativeLanguage]Simple
	e.g. luaSimple* messages.
	"

	| prefix |
	prefix := self nativeLanguage.
	(self implements: prefix ,#RequestBlock) ifTrue: [ ^prefix ].
	prefix := prefix ,#Simple.
	(self implements: prefix ,#RequestBlock) ifTrue: [ ^prefix ].
	^nil! !

! SystemServer methodsFor: #platform!
platformSend: message withArguments: anArray
	" Private - Returns the result of sending message (platform dependent) to receiver.
	Trigger an error if the message is not implemented.
	"

	| token msg |
	token := self class platform.
	token isNil ifTrue: [ token := #default ].
	msg := token , message.
"	(self respondsTo: msg) ifFalse: [ ^self error: 'Missing server support for ' ,msg ]. "
	^self perform: msg withArguments: anArray! !

! SystemServer class methodsFor: #singleton!
currentOrNil
	" Returns the current instance (or nil). "

	^current! !

! SystemServer class methodsFor: #server!
start
	" Start the server. "

	^self current start! !

! SystemServer class methodsFor: #server!
close
	" Close the current server. "

	^current notNil ifTrue: [ current close ]! !

! SystemServer class methodsFor: #server!
stop
	" Stop the current server. "

	^self close! !

! SystemServer class methodsFor: #server!
restart
	" Close the current the server and start a new. "

	self close; current: nil.
	^self start! !

! SystemServer class methodsFor: #html !
htmlSelectorsFor: className
	" Return an html fragment containing the selectors implemented by className. "

	| stream aClass classDef |
	aClass := Smalltalk at: className ifAbsent:[ ^'' ].
	stream := '' writeStream.	
	stream nextPutAll: '<BR><BR><a href=',$'
		,'/SystemServer/htmlHierarchy',$'
		,'><< S8 Hierarchy</a><BR>'.
	classDef := Exporter new fileOutDefinitionOf: aClass.
	classDef := classDef replace:'\r\n' with:'<BR>'.
	classDef := classDef replace:'\t' with:'&nbsp;&nbsp;&nbsp;'.
	stream nextPutAll: '<BR>' ,classDef ,'<H4>' ,className ,' class methods</H4>'.
	aClass class selectors do: [:each| 
		stream nextPutAll: '<a href=',$'
			,'/SystemServer/htmlFileOut:selector:side:/', aClass name
			,'/',each,'/class', $', '>', each, '</a><BR>'
	].	
	stream nextPutAll: '<H4>', className,' instance methods</H4>'.
	aClass selectors do:[:each| 
		stream nextPutAll: '<a href=',$'
			,'/SystemServer/htmlFileOut:selector:/', aClass name
			,'/',each,$', '>', each, '</a><BR>'
	].
	^stream contents! !

! SystemServer class methodsFor: #html !
htmlFileOut: className selector: selector side: side
	" Return an html fragment of selector source code. "

	| stream aClass mth source |
	aClass := Smalltalk at: className ifAbsent:[ ^'' ].
	mth := (side = #class) ifTrue:[ aClass class >> selector 
	] ifFalse: [ aClass >> selector ].
	source := Exporter new fileOutMethod: mth of: aClass.
	source := source replace:'\r\n' with:'<BR>'.
	source := source replace:'\t' with:'&nbsp;&nbsp;&nbsp;'.
	source := source replace: $\,$" with: '&quot;'.
	source := '<BR><BR><a href='
		,'/SystemServer/htmlSelectorsFor:/'
		,className ,'><< ' ,className
		,' methods</a><BR>',source
		,'<BR>'.
	^source! 

! SystemServer class methodsFor: #html !
htmlFileOut: className selector: selector
	" Return an html fragment of selector source code. "

	^self htmlFileOut: className 
		selector: selector 
		side: #instance! !

! SystemServer class methodsFor: #html !
htmlHierarchy
	" Returns the system hierarchy in html fasion. "

	| stream |
	stream := '' writeStream.
	stream nextPutAll:'<H2>S8 Hierarchy</H2>'.
	Smalltalk rootClasses do: [:cls|
		self htmlWalk: cls blanks: 0 into: stream
	].
	^stream contents! !
	
! SystemServer class methodsFor: #html !
htmlWalk: aClass blanks: blanks into: stream
	" Private - Returns the description for hierarchy level @ blanks. "

	stream nextPutAll:'<BR>'. 
	blanks timesRepeat: [ stream nextPutAll: '&nbsp;&nbsp;' ].
	stream nextPutAll: (self htmlDescriptionForClass: aClass).
	aClass subclasses do: [:cls|
		self htmlWalk: cls blanks: blanks + 1 into: stream
	].! !

! SystemServer class methodsFor: #html !
htmlDescriptionForClass: aClass
	" Private - Returns the contents to show for aClass. "

	^'<a href=', $', '/SystemServer/htmlSelectorsFor:/', aClass name, $' ,$>, aClass name,'</a> [',aClass category, ']'! !

! SystemServer methodsFor: #query!
isRunning
	" Returns true if the receiver is online. "

	^server notNil! !

! SystemServer methodsFor: #runnning!
step
	" Run a server step. "

	^self isRunning ifTrue: [ server step ]! !

! SystemServer methodsFor: #runnning!
listen
	" Run a server step. "

	^server listen! !

! SystemServer methodsFor: #runnning!
run
	" Run the receiver in a loop until stopped. "

	[self isRunning] whileTrue: [ self listen ]! !

! SystemServer methodsFor: #server!
online
	" Returns true if the receiver is online. "

	^server notNil! !

! SystemServer methodsFor: #server!
start
	" Start the receiver. "

	^self startServer! !

! SystemServer methodsFor: #server!
close
	" Close the receiver. "

	self server: nil! !

! SystemServer methodsFor: #private!
server: aServer
	" Private - Set the server of the receiver. "

	server == aServer ifTrue: [ ^self ].
	self unbindSystemManager.
	(server respondsTo: #close) ifTrue: [ server close ].
	server := aServer.
	server notNil ifTrue: [ self bindSystemManager ]! !

! SystemServer methodsFor: #server!
startServer
	" Start the server. "

	^self startServerOn: self port! !
	
! SystemServer methodsFor: #server!
startServerOn: aPort
	" Start the server. "

	| query |
	query := self systemRequestBlock.
	query isNil ifTrue: [ ^false ].
	^self startWith: [:req :resp|
		'<html><body><pre>'
		,(query value: req value: resp)
		,'</pre></body></html>'.		
	] on: aPort! !
	
! SystemServer methodsFor: #server!
startWith: processBlock on: aPort
	
	^self	startBlock: processBlock 
		port: aPort! !
	
! SystemServer methodsFor: #server!
startBlock: aBlock port: aPort
	" Private - Start the server. "

	server := self
		start: self
		processBlock: aBlock
		port: aPort.
	^server notNil! !

! SystemServer methodsFor: #preparation !
aboutToStart
	" Private - Prepare the environment to start the receiver. "

	self	unbindSystemManager;
		bindSystemManager! !

! SystemServer methodsFor: #server!
start: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	self aboutToStart.
	^self	platformSend: #Start:processBlock:port:
		withArguments: (Array with: aServer with: aBlock with: aPort)! !

! SystemServer methodsFor: #server !
systemRequestBlock
	" Private - Return the request processing block used for remote system tools.
	Note that the block is evaluated for each request and must return the code of the body of HTML formated result.
	"

	^self platformSend: #RequestBlock withArguments: #()! !

! SystemServer methodsFor: #server !
jsonParse: string
	" Private - Return the result of parsing JSON string. "

	^self platformSend: #Parse: withArguments: (Array with: string)! !

! SystemServer methodsFor: #server !
jsonStringify: literal
	" Private - Return the result of JSON string. "

	^self platformSend: #Stringify: withArguments: (Array with: literal)! !

! SystemServer methodsFor: #ui !
serverStartedAt: url
	" Private - Show a notification to the user; the server has been started. "

	| string |
	string := 'Server started at ' ,url.
	(self respondsTo: #alert:) ifTrue: [ ^self alert: string ].
	^self note: string! !

! SystemServer methodsFor: #defaults !
defaultRequestBlock
	" Private - Return the request processing block used for remote system tools. "

	^self error: 'Missing support for server request'! !

! SystemServer methodsFor: #defaults !
defaultStart: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	^self error: 'Missing support for server start'! !

! SystemServer methodsFor: #defaults !
defaultParse: string
	" Private - Return the result of parsing JSON string. "

	^NativeHandle support @ (#{JSON} #parse: string)! !

! SystemServer methodsFor: #defaults !
defaultStringify: literal
	" Private - Return the result of JSON string. "

	^#{JSON} #stringify: literal! !

! SystemServer methodsFor: #json !
checkArrays: aHandle
	" Private,Lua - Scan native handle values for arrayed contents recursivelly.
	Note that JSON data is hierarchical non-recursive structures we walk the structure descending recursivelly in the structure mutating array candidates.
	"

	| remaining |
	remaining := Array with: aHandle.
	[remaining notEmpty] whileTrue: [ | literal |
		(literal := remaining removeFirst) notNil ifTrue: [
			(literal basicAt: 1 ifAbsent: []) notNil "mutate to Array"
			ifTrue: [ {'literal' lua'smalltalk:newArray(literal)'}. remaining addAll: literal ]
			ifFalse: [ "walk normal object contents"
				(self isObject: literal) ifFalse: [
					NativeHandle support @ literal propertiesDo: [:key :value|
						(value basicAt: 1 ifAbsent: []) notNil ifTrue: [
							{'value' lua'smalltalk:newArray(value)'}. remaining addAll: value
						] ifFalse: [ remaining add: value ]
					]
				]
			]
		]
	].! !

! SystemServer methodsFor: #notification !
say: text
	" Show text to the user. "

	self class silent ifTrue: [ ^self ].
	text isArray ifTrue: [ text do: [:line| self say: line ] ].
	self note: text! !

! SystemServer methodsFor: #private !
getCommand: command
	" Private - Return the result of evaluating the command. "

	(command isEmpty or: [command first isEmpty]) ifTrue:  [
		^self systemHomePage
	].
	(command size = 1 and: [ self respondsTo: command first ]) ifTrue: [
		^self perform: command first
	].
	^self resultForCommand: command stream! !

! SystemServer methodsFor: #private !
postCommand: command stream: stream
	" Private - Return the result of evaluating the command. "

	stream upTo: $=.
	^(command isEmpty or: [command first isEmpty])
	ifTrue:  [ self note: stream upToEnd ]
	ifFalse: [
		command add: stream upToEnd.
		self resultForCommand: command stream.
	]! !

! SystemServer methodsFor: #private !
requestBlockFor: urlBlock method: methodBlock post: postBlock
	" Private - Return the request processing block used for remote system tools. "

	^[:arg1 :arg2| | command method | self jsonValue: [
		command := urlBlock value: arg1 value: arg2.
		method := methodBlock value: arg1 value: arg2.
		self say: '' ,method ,' ' ,command.
		command := command asArrayOfSubstringsSeparatedBy: $/.
		command removeFirst.
		method = #GET ifTrue: [ self getCommand: command ]
		ifFalse: [
			postBlock value: arg1 value: [:stream|
				self postCommand: command stream: stream
			]
		]
	]]! !
	
! SystemServer methodsFor: '#commands' !
systemHomePage
	" Private - Returns the contents for system home page. "

	^'<h1>S8 System server</h1>'
	,'Try commands like <a href=',$' ,'/SystemServer/htmlHierarchy', $', '>hierarchy</a>&nbsp;'
	,'or more complex expressions: <a href=', $', '/Object/htmlSelectors', $', '>Object selectors</a>'! !

! SystemServer methodsFor: #private !
objectFromStream: stream
	" Private - Return the object received from client. "

	stream next = $!! ifFalse: [ self error: 'Invalid code stream' ].
	stream peek = $@ ifTrue: [
		stream next.
		^self at: stream upToEnd
	].
	^stream fileIn! !

! SystemServer methodsFor: #private !
objectsFromArguments: array
	" Private - Return the objects to use as arguments received from client. "

	^array collect: [:each|
		each notNil ifTrue: [
		each isArray ifTrue: [ self objectsFromArguments: each ]
		ifFalse: [ (each isString and: [ each startsWith: $!! ])
		ifTrue: [ self objectFromStream: each stream ]
		ifFalse: [ each ]]]
	]! !

! SystemServer methodsFor: #private !
perform: stream on: receiver
	" Private - Returns the result of performing messages in stream on receiver. "

	| result |
	result := receiver.
	[stream atEnd] whileFalse: [ | selector |
		selector := stream next.
		selector isEmpty ifTrue: [ selector := $/ ].
		selector first = ${ ifTrue: [ | json |
			[selector last = $}] whileFalse: [
				stream atEnd ifTrue: [ ^self error: 'Corrupted message ' ,selector ].
				selector := selector ,$/ ,stream next.
			].
			json := self jsonParse: selector.
			selector := json @ #selector.
			selector isNil ifTrue: [ self error: 'Missing selector' ].
			result := (Message
				receiver: result selector: selector toString
				arguments: (self objectsFromArguments: json @ #arguments)
				) perform
		] ifFalse: [ | args argCount |
			argCount := selector argumentCount.
			args := (1 to: argCount) collect: [:i|
				stream atEnd ifFalse: [ stream next ]
			].
			result := result perform: selector withArguments:
				(self objectsFromArguments: args)
		]
	].
	^result! !

! SystemServer methodsFor: #commands !
evalNative: stream
	" Private - Returns the result of evaluating native command in stream. "

	| source lang target |
	target := Smalltalk current.
	source := stream next.
	lang := stream peek.
	(lang notNil and: [ lang ~= target nativeLanguage ]) ifTrue: [
		^self error:
			 'Invalid native language #' ,lang
			,' must be #' ,target nativeLanguage
	].
	^target eval: source! !
 
! SystemServer methodsFor: #cache !
at: key put: anObject
	" Register anObject in the cache. "

	anObject isNil ifTrue: [
		^self cache removeKey: key ifAbsent: []
	].
	^self cache at: key put: anObject! !
 
! SystemServer methodsFor: #cache !
at: key
	" Returns the registered object at key in the cache (or nil). "

	^self cache at: key ifAbsent: [ nil ]! !
 
! SystemServer methodsFor: #cache !
clearCache
	" Reset the cache.
	Note that #persistent pool is retained in the cache.
	"

	| persistent |
	persistent := self at: #persistent.
	self initializeCache.
	persistent notNil ifTrue: [ self at: #persistent put: persistent ]! !
 
! SystemServer methodsFor: #cache !
cacheIt: smalltalkExpression to: aReceiver
	" Returns the id of cached result of evaluating smalltalkExpression on aReceiver. "

	| object id key |
	object := Compiler
		doIt: smalltalkExpression
		to: aReceiver.
	id := self at: #lastID.
	id isNil ifTrue: [ id := 0 ].
	key := id toString.
	self
		at: #lastID put: id + 1;
		at: key put: object.
	^key! !
 
! SystemServer methodsFor: #cache !
@ path
	" Returns the object at path (or nil). "

	| result |
	path isEmpty ifTrue: [ ^self ].
	path isString ifTrue: [
		^self @ (path asArrayOfSubstringsSeparatedBy: self rootStart)
	].
	(self cache includesKey: path first) ifTrue: [ ^self cache @ path ].
	result := Smalltalk at: path removeFirst ifAbsent: [ ^nil ].
	path do: [:each|
		result := result @ each.
		result isNil ifTrue: [ ^nil ].
	].
	^result! !
 
! SystemServer methodsFor: #roots !
rootStart
	" Private - Returns the tag to start root "

	^$@! !
 
! SystemServer methodsFor: #roots !
root: command stream: stream ifAbsent: aBlock
	" Private - Returns the root of path in command stream (or the result of evaluating aBlock). "

	command isEmpty ifTrue: [ ^self ].
	command first = self rootStart ifTrue: [
		^self @ (command copyFrom: 2 to: command size)
	].
	command = #self ifTrue: [ ^self ].
	command = #smalltalk ifTrue: [ ^self smalltalk ].
	^Smalltalk at: command ifAbsent: aBlock! !

! SystemServer methodsFor: #commands !
resultForCommand: stream
	" Private - Return the result of processing the command in stream. "

	| command origin |
	command := stream next.
	command = #doIt		ifTrue: [ ^stream next doIt ].
	command = #fileIn		ifTrue: [ ^stream next fileIn ].
	command = #eval		ifTrue: [ ^self evalNative: stream ].
	command = #fileOut	ifTrue: [ ^stream next fileOut ].
	(#(fileOut category) includes: command) ifTrue: [ ^stream next fileOut ].
	origin := self root: command stream: stream ifAbsent: [
		^'Unknown command ',command
	].
	^self perform: stream on: origin! !
 
! SystemServer methodsFor: #commands !
categories
	" Private - Returns the categories in the system. "

	^Class categories sorted printString! !

! SystemServer methodsFor: #commands !
classes
	" Private - Returns the contents for this command. "

	^Smalltalk classes inject: '' into: [:total :cls|
		total,'
',(self shortDescriptionForClass: cls)
	]! !

! SystemServer methodsFor: #commands !
hierarchy
	" Private - Returns the contents for this command. "

	| stream |
	stream := '' writeStream.
	Smalltalk rootClasses do: [:cls|
		self walk: cls blanks: 0 into: stream
	].
	^stream contents! !

! SystemServer methodsFor: #commands !
walk: aClass blanks: blanks into: stream
	" Private - Returns the description for hierarchy level @ blanks. "

	stream cr. blanks timesRepeat: [ stream nextPut: $ ].
	stream nextPutAll: (self shortDescriptionForClass: aClass);cr.
	aClass subclasses do: [:cls|
		self walk: cls blanks: blanks + 1 into: stream
	].! !
	
! SystemServer methodsFor: #private !
shortDescriptionForClass: aClass
	" Private - Returns the contents to show for aClass. "

	^aClass name,' ',aClass category! !

! SystemServer methodsFor: #output !
jsonValue: aBlock
	" Private - Returns a JSON package string with result of evaluating aBlock. "

	| response jResult |
	[ | result |
		result := self literalFor: aBlock evaluate.
		response := self changesWith: #result -> result
	] on: Exception do: [:ex|
		response := #error -> ex.
		self say: #error -> ex description.
		ex stack size > 0 ifTrue: [ self say: ex stack ]
	].
	^[	jResult := response json.
		self jsonStringify: jResult
	] on: Exception do: [:ignored| jResult toString ]! !

! SystemServer methodsFor: #output !
escape
	" Private - Returns the escape token. "

	^$!! ! !

! SystemServer methodsFor: #output !
referenceTo: anObject
	" Private - Returns the reference to link anObject on the client side. "

	(anObject respondsTo: #systemServerReferenceIn:) ifTrue: [
		^self escape ,$@ ,(anObject systemServerReferenceIn: self) toString
	].
	^anObject printString! !

! SystemServer methodsFor: #output !
literalFor: anObject
	" Private - Returns the literal to use as reference to anObject on client side. "

	anObject isNil ifTrue: [ ^self escape ,#nil ].
	(self isObject: anObject) ifTrue: [
		(anObject isNumber or: [ anObject isMemberOf: Boolean ]) ifTrue: [ ^self escape ,anObject ].
		anObject isString ifTrue: [
			(anObject startsWith: self escape) ifTrue: [ ^self escape ,anObject ].
			^anObject
		].
		anObject isArray ifTrue: [ ^anObject collect: [:each| self literalFor: each ] ].
		(anObject isClass or: [ anObject isMetaclass
		or: [anObject isKindOf: CompiledMethod]])
		ifTrue: [ ^self escape ,anObject printString ].
		^self referenceTo: anObject
	].
	^anObject! !

! SystemServer methodsFor: #output !
changesWith: results
	" Private - Returns the results with logged changes. "

	| changes |
	changes := systemChanges.
	systemChanges := nil.
	(changes isNil or: [changes isEmpty]) ifTrue: [ ^results ].
	changes := self literalFor: changes.
	^results ,(#changes -> changes)! !

! SystemServer methodsFor: 'system changes' !
processSystemChanges
	" Private - The receiver has changes to be processed.
	We do nothing, the changes will be sent to client(s?) on demmand.
	"! !

"--NodeJS customization----------------------------------------------"!
! SystemServer methodsFor: #node !
nodeStart: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	| localIP http httpServer |
	[ http := NodeJS http ] on: Error do: [:ignored|
		self note: self class name ,' missing HTTP interface.'.
		^nil
	].
	localIP := NodeJS os localIP.
	localIP isNil ifTrue: [
		self note: self class name ,' missing network connection. (localIP is nil, check network connection)'.
		^nil
	].
	httpServer := http createServer: [:req :resp| | request response |
		request  := NodeIncomingMessage @ req.
		response := NodeHTTPServerResponse @ resp.
		response 
			setHeader:'Access-Control-Allow-Origin' value:'*';
			writeHead: 200;
			end: (aBlock value: request value: response).
	].
	[	httpServer listen: aPort hostname: nil json.
		self serverStartedAt: 'http://' ,localIP ,':' ,aPort
	 ] on: Exception do: [:ex|
		self note: self class name ,' ' ,ex toString.
		^nil
	].
	^httpServer! !

! SystemServer methodsFor: #node !
nodeRequestBlock
	" Private - Return the request processing block used for remote system tools. "

	^self
		requestBlockFor: [:req| req url asDecodedURI ]
		method: [:req| req method ]
		post: [:req :action| | body |
			body := ''.
			req
				on: #data do: [:data| body := body ,data ];
				on: #end do: [ action value: body asDecodedURI stream ].
			#ok
		] ! !

! SystemServer methodsFor: #node !
nodeParse: string
	" Private - Return the result of parsing JSON string. "

	^self defaultParse: string! !

! SystemServer methodsFor: #node !
nodeStringify: literal
	" Private - Return the result of JSON string. "

	^self defaultStringify: literal! !

"--GCD customization----------------------------------------------"!
! SystemServer methodsFor: #gcd !
gcdStart: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	| instance |
	instance := GCDWebServer startHTML: aBlock withPort: aPort.
	"self serverStartedAt: instance serverURL absoluteString" ! !

! SystemServer methodsFor: #gcd !
gcdRequestBlock
	" Private - Return the request processing block used for remote system tools. "

	^self
		requestBlockFor: [:req| req path ]
		method: [:req| #GET ]
		post: [:req :action| | body |
			body := ''.
			req	notImplementedYet: #gcdRequestBlock ,' (POST)';
				on: #data do: [:data| body := body ,data ];
				on: #end do: [ action value: body asDecodedURI stream ].
			#ok
		]! !

! SystemServer methodsFor: #gcd !
gcdParse: string
	" Private - Return the result of parsing JSON string. "

	^self defaultParse: string! !

! SystemServer methodsFor: #gcd !
gcdStringify: literal
	" Private - Return the result of JSON string. "

	^self defaultStringify: literal! !

"--Turbo(Lua) customization----------------------------------------------"!
! SystemServer methodsFor: #turbo !
turboStart: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	| http |
	http := HTTPServer new: [:hRequest| | hdr request msg buf crlf |
		crlf := String crlf.
		request := HTTPRequest @ hRequest.
		request basicAt: #path put: request path.
		msg := aBlock value: request.
		hdr := HTTPHeaders new.
		hdr
			add: 'Content-Length' value: '',msg size;
			add: 'Connection' value: 'Keep-Alive';
			add: 'Access-Control-Allow-Origin' value: $*;
			setStatusCode: 200;
			setVersion: 'HTTP/1.1';
			yourself.
		buf := hdr stringifyAsResponse ,crlf ,msg.
		request write: buf; finish
	].
	[	http listen: aPort.
		self serverStartedAt: 'port ',aPort.
		"Ioloop instance start"
	 ] on: Exception do: [:ex|
		self note: self class name ,' ' ,ex toString.
		^nil
	].
	^http! !

! SystemServer methodsFor: #turbo !
turboRequestBlock
	" Private - Return the request processing block used for remote system tools. "

	^self
		requestBlockFor: [:req| Escape new unescape: req#path ]
		method: [:req| #GET ]
		post: [:req :action|
			req notImplementedYet: #gcdRequestBlock ,' (POST)'.
			#ok
		]! !

! SystemServer methodsFor: #turbo !
turboParse: string
	" Private - Return the result of parsing JSON string. "

	| result |
	result := NativeHandle support @ (Escape new jsonDecode: string).
	result notNil ifTrue: [ self checkArrays: result handle ].
	^result! !

! SystemServer methodsFor: #turbo !
turboStringify: literal
	" Private - Return the result of JSON string. "

	^Escape new jsonEncode: NativeHandle support @ literal! !

"--luaSimple customization----------------------------------------------"!
! SystemServer methodsFor: #luaSimple !
luaSimpleStart: aServer processBlock: aBlock port: aPort
	" Private - Start the server. "

	| http |
	(Smalltalk includesKey: #Json) ifFalse: [
		self note: 'Missing Json support (SimpleJson.st required)'.
		^self error: 'Missing Json support (SimpleJson.st required)'
	].
	http := SimpleHTTPServer new.
	[  http	port: aPort;
		when: #needsContents:with: do: [:uri :headers|
			aBlock value: uri value: headers
		]; start.
	  self serverStartedAt: http host toString ,':' ,http port.
	 ] on: Exception do: [:ex|
		self note: self class name ,' ' ,ex toString.
		^nil
	].
	^http! !

! SystemServer methodsFor: #luaSimple !
luaSimpleRequestBlock
	" Private - Return the request processing block used for remote system tools. "

	^self
		requestBlockFor: [:uri :headers| uri unescaped ]
		method: [:uri :headers| #GET ]
		post: [:req :action|
			req notImplementedYet: #luaSimpleRequestBlock ,' (POST)'.
			#ok
		]! !

! SystemServer methodsFor: #luaSimple !
luaSimpleParse: string
	" Private - Return the result of parsing JSON string. "

	| json js args |
	js := Json parse: string.
	js isNil ifTrue: [ ^self error: 'Broken Json :',string ].
	args := js @ #arguments.
	{'args' lua'smalltalk:newArray(args)'}.
	json := ((#selector -> (js @ #selector) toString)
		,(#arguments -> args)) json.
	^NativeHandle support @ json! !

! SystemServer methodsFor: #luaSimple !
luaSimpleStringify: literal
	" Private - Return the result of JSON string. "

	^Json stringify: NativeHandle support @ literal! !