
Object
 subclass: #APIBuilder
 instanceVariableNames: ' parents token bound '
 category: #Builders!
APIBuilder class instanceVariableNames: ' default saveBuilders '!

! APIBuilder class methodsFor: #accessing !
default
	" Return the default instance of the receiver. "

	default isNil ifTrue: [ default := self new ].
	^default! !

! APIBuilder class methodsFor: #accessing !
saveBuilders
	" Return true if the receiver is set to save builders in snapshot. "

	^saveBuilders == true! !

! APIBuilder class methodsFor: #accessing !
saveBuilders: aBoolean
	" Set to true if the receiver save builders in snapshot. "

	saveBuilders := aBoolean! !

! APIBuilder class methodsFor: #building !
build: token doing: aBlock
	" Process a (build...: spec) definition marking methods generated while evaluating aBlock.
	Return the result of evaluating aBlock. "

	| builder result success |
	builder := self default.
	builder open: token.
	success := false.
	[	result := aBlock value.
		success := true.
	] ensure: [ builder close: token success: success ].
	^result! !

! APIBuilder class methodsFor: #building !
register: aMethod
	" Register aMethod as related to current builder. "

	^default notNil and: [ default register: aMethod ]! !

! APIBuilder class methodsFor: #snapshot !
imageSaved: snapshot on: stream
	" Private - A system snapshot has been taken, the receiver can restore state. "

	| code |
	(self saveBuilders and: [
		(code := self codeForBuilders) notEmpty ])
		ifTrue: [ stream cr;nextPutAll: code ].
	^super imageSaved: snapshot on: stream! !

! APIBuilder class methodsFor: #code !
codeForBuilders
	" Return the (native) code to export the builders of the system. "

	^self codeFor: Smalltalk classes! !

! APIBuilder class methodsFor: #code !
codeFor: classes
	" Return the (native) code to export the builders defined by a collection of classes. "

	| builders |
	builders := self buildersFor: classes.
	builders isEmpty ifTrue: [ ^'' ].
	^'smalltalk.', self name ,'.builders_(
'	,(Exporter new exportObject: builders) ,'
);'! !

! APIBuilder class methodsFor: #private !
buildersFor: classes
	" Private - Return the builders defined by a collection of classes.
	#( className #(--classbuilders--) #(--instanceBuilders---))
	"

	^(classes collect: [:cls|
		Array
			with: cls name
			with: (self buildersIn: cls class)
			with: (self buildersIn: cls)
	]) reject: [:tuple|
		tuple second isEmpty and: [tuple third isEmpty]
	]! !

! APIBuilder class methodsFor: #private !
builders
	" Private - Return the builders defined in this system. "

	^self buildersFor: Smalltalk classes! !

! APIBuilder class methodsFor: #private !
buildersIn: aClass
	" Private - Return the builders used in aClass.
	#( builderType builderSpec #( selectors ) )
	"

	| map |
	map := Array new.
	aClass methodsDo: [:mth| | assoc |
		assoc := mth builder.
		assoc notNil ifTrue: [
			(map	detect: [:tuple| tuple first = assoc ]
				ifNone: [ map add: (Array with: assoc with: Array new) ])
				second add: mth selector
		]
	].
	^map collect: [:tuple| Array
		with: tuple first key
		with: tuple first value
		with: tuple last
	]! !

! APIBuilder class methodsFor: #private !
builders: spec
	" Private - Set the builders on this system according to bindings defined in spec.
	Note that missing bindings will be ignored and no check is done to ensure the bindings are in sync with target.
	"

	spec do: [:builder| | cls |
		cls := Smalltalk at: builder first ifAbsent: [].
		cls notNil ifTrue: [
			self setBuilders: builder second in: cls class.
			self setBuilders: builder third in: cls.
		].
	]! !

! APIBuilder class methodsFor: #private !
setBuilders: array in: aClass
	" Private - Set the builders in array on aClass. "

	| builder |
	array do: [:map|
		builder := Association key: map first value: map second.
		map last do: [:selector| | mth |
			mth := aClass compiledMethodAt: selector.
			mth notNil ifTrue: [ mth builder: builder ].
		]
	]! !

! APIBuilder methodsFor: #private !
register: aMethod
	" Private - Register aMethod as related to the receiver. "

	token isNil ifTrue: [ ^false ].
	bound isNil ifTrue: [ bound := Array new ].
	bound add: aMethod.
	^true! !

! APIBuilder methodsFor: #private !
open: spec
	" Private - The spec will be generated.
	Prepare the receiver to track generated methods while building spec.
	"

	parents isNil ifTrue: [ parents := Array new ].
	token notNil ifTrue: [ parents addLast: token ].
	token := spec.! !

! APIBuilder methodsFor: #private !
close: spec success: success
	" Private - The spec has been implemented, backtrack to parent or mark collected methods. "

	token = spec ifFalse: [ ^self error: 'Builders out of sync.' ].
	(parents notNil and: [ parents notEmpty ]) ifTrue: [
		token := parents removeLast.
		^success
	].
	bound notNil ifTrue: [
		bound do: [:each| self mark: each ].
		token := parents := bound := nil.
	].
	^success! !

! APIBuilder methodsFor: #private !
mark: aMethod
	" Private - Mark aMethod as dependent of current token. "

	aMethod builder: token! !

! APIBuilder methodsFor: #fileOut !
fileOut: exporter selectors: selectors of: aClass on: stream
	" Fileout clustered sources for selectors onto stream.
	Note: Remove the clustered message selectors from selectors collection while fileOut.
	"

	| specs |
	specs := Set new.
	selectors copy do: [:selector| | assoc |
		assoc := (aClass compiledMethodAt: selector) builder.
		assoc notNil ifTrue: [
			specs add: assoc.
			selectors remove: selector.
		]
	].
	(specs asArray sorted: [:a :b| a key <= b key ]) do: [:assoc|
		stream nextPutAll: aClass name ,' build' ,assoc key.
		assoc value notEmpty ifTrue: [
			stream nextPut: $:.
			self fileOutSpec: assoc value on: stream
		].
		stream nextChunkPut: $ ;cr
	].! !

! APIBuilder methodsFor: #fileOut !
fileOutSpec: array on: stream
	" Private - Fileout API spec in array onto stream. "

	stream nextPutAll: ' #(';cr.
	array do: [:each| stream tab; nextPutAll: each asLiteral;cr ].
	stream nextPut: $).! !

! CompiledMethod methodsFor: 'Builders-accessing' !
builder: cookie
	" Private - Set the builder cookie of the receiver. "

	self basicAt: '$builder' put: cookie! !

! CompiledMethod methodsFor: 'Builders-accessing' !
builder
	" Private - Return the builder cookie of the receiver (or nil). "

	^self basicAt: '$builder'! !

! Behavior methodsFor: #private !
aboutToAdd: aMethod
	" Private - Prepare aMethod to be added onto the receiver. "

	APIBuilder register: aMethod. ! !

! CompiledMethod methodsFor: 'Builders-tools' !
removeSources
	" Remove all development sources of the receiver. "

	self source: ''; category: #s8; builder: nil! !

! Behavior methodsFor: 'Builders-implementing' !
implement: messageHeader
	comment: aComment
	as: sourceCode
	category: aCategory
	" Private - Implement a method in the receiver.
	Note: sourceCode must be the source code in chunk format.
	"

	| code |
	code := '!! ',self name,
		(aCategory isNil
			ifTrue: [ ' methods' ]
			ifFalse: [ ' methodsFor: ', aCategory asLiteral ])
,' !!
',messageHeader, (aComment isNil ifTrue: [ '' ] ifFalse: ['
	" ', aComment, ' "']),'

	',sourceCode ,'!! !!

'.	^code fileIn; yourself! !

! Behavior methodsFor: 'Builders-implementing' !
implement: messageHeader comment: aComment as: sourceCode
	" Private - Implement a method in the receiver. "

	^self
		implement: messageHeader
		comment: aComment
		as: sourceCode
		category: nil! !

! Behavior methodsFor: 'Builders-implementing' !
implement: messageHeader as: sourceCode
	" Private - Implement a method in the receiver. "

	^self
		implement: messageHeader
		comment: nil
		as: sourceCode
		category: nil! !

! Behavior methodsFor: 'Builders-implementing' !
implement: messageHeader as: sourceCode category: aCategory
	" Private - Implement a method in the receiver. "

	^self
		implement: messageHeader
		comment: nil
		as: sourceCode
		category: aCategory! !

! Behavior methodsFor: 'Builders-implementing' !
class: aClass implement: message
	comment: aComment as: sourceCode
	category: aCategory
	" Private - Implement a method in aClass. "

	aClass isString ifTrue: [
		^self
			class: (aClass evaluateIn: nil)
			implement: message
			comment: aComment
			as: sourceCode
			category: aCategory
	].
	^aClass
		implement: message
		comment: aComment
		as: sourceCode
		category: aCategory! !

! Behavior methodsFor: 'Builders-private' !
build: token doing: aBlock
	" Process a (build...: spec) definition marking methods generated while evaluating aBlock.
	Return the result of evaluating aBlock. "

	^APIBuilder build: token doing: aBlock! !

! Behavior methodsFor: 'Builders-obsolete' !
buildAPI: specs
	" Generate code to implement API protocoll defined by specs.
	WARNING: OBSOLETE - this builder generate changes that will not be filedOut to builder code.
	specs must be an array of :
		- string : it will be considered as an evaluable expression to be executed in the context of the receiver.
		- anArray : of #( selector arg1 arg2 arg3 arg4... ) to be performed for code generation/setup of the receiver
			where selector is frequently a build...: selector.
	"

	specs do: [:spec|
		spec isString
		ifTrue: [ spec evaluateIn: self ]
		ifFalse: [
			self	perform: spec first
				withArguments: (spec copyFrom: 2 to: spec size)
		]
	]! !

! Behavior methodsFor: 'Builders-private' !
stMessageHeader: msgHeader
	" Private - Return a message header string with nice smalltalk looking.
	Default implementation leave the message header as is.
	This method can be refined by subclasses to transform message pattern for better code generation.
	"

	^msgHeader! !

! Behavior methodsFor: 'Builders-api' !
buildAccessor: accessor
	" Build (instance) method to access native property. "

	^self build: #Accessor -> accessor doing: [
		(accessor isString or: [
			accessor isArray and: [
			accessor size > 1 and: [
			accessor second = #initialize ] ]
		]) ifTrue: [
			(self buildGetter: accessor)
			,(self buildSetter: accessor)
		] ifFalse: [
			accessor compose: [:tuple|
				self implement: tuple first comment: (tuple at: 2)
				      as: (tuple at: 3) category: #accessing
			]
		]
	]! !

! Behavior methodsFor: 'Builders-private' !
buildGetterCode: aSelector
	" Private - Return source code to get aSelector. "

	^'^' ,aSelector! !

! Behavior methodsFor: 'Builders-api' !
buildGetter: spec
	" Build method for getter. "

	^self build: #Getter -> spec doing: [
		(spec isArray and: [ spec size > 1 and: [
		 spec second = #initialize ]]) ifTrue: [
			self buildGetter: spec first initialize: spec last
		] ifFalse: [
			self
				implement: (self stMessageHeader: spec)
				comment: 'Return the ' ,spec ,' of the receiver.'
				as: (self buildGetterCode: spec) category: #accessing
		]
	]! !

! Behavior methodsFor: 'Builders-private' !
buildInitialize: aSelector with: initializer
	" Private - Build method to initialize instance variable aSelector. "

	^self
		implement: #initialize ,(aSelector asSmalltalkName: true)
		comment: 'Private - Initialize the ' ,aSelector ,' of the receiver.'
		as: initializer
		category: #initialize! !

! Behavior methodsFor: 'Builders-private' !
buildGetter: aSelector initialize: initializer
	" Private - Build method to get instance aSelector w/deferred initialization mode.
	Note: #initializer can be:
		#initialize to generate default/deferred implementation.
		empty string to do NOT generate #initialize... method.
		smalltalk code starting with $. to use that code as method implementation.
		smalltalk expression for asignment (to be used as assignment to instance).
		GlobalName if code do not have blanks and start with uppercase, it will be used as class to instantiate during initialization.
	"

	| code |
	code := initializer = #initialize
	  ifTrue: [ '^self implementedBySubclass: #initialize' ] ifFalse: [
		initializer notEmpty ifTrue: [
			initializer first = $.
			ifTrue: [ initializer copyFrom: 2 to: initializer size ] ifFalse: [
				aSelector ,' := ' ,(
					(initializer first isUppercase and: [ initializer asArrayOfSubstrings size = 1 ])
					ifTrue: [ initializer ,' new' ]
					ifFalse: [ initializer ])
			]
		]
	  ].
	^(code notNil ifTrue: [ self buildInitialize: aSelector with: code ] ifFalse: [ '' ])
	,(self
		implement: (self stMessageHeader: aSelector)
		comment: 'Return the ' ,aSelector ,' of the receiver.'
		as: aSelector,' isNil ifTrue: [ self ' , #initialize ,(aSelector asSmalltalkName: true) ,' ].
	^' ,aSelector category: #accessing)! !

! Behavior methodsFor: 'Builders-private' !
buildSetterCode: aSelector
	" Private - Return source code to set aSelector. "

	^aSelector ,' := anObject'! !

! Behavior methodsFor: 'Builders-api' !
buildSetter: spec
	" Build method for setter. "

	(spec isArray and: [ spec size > 1 and: [
	 spec second = #initialize ]]) ifTrue: [
		^self buildSetter: spec first
	].
	^self build: #Setter -> spec doing: [
		self
			implement: (self stMessageHeader: spec) ,': anObject'
			comment: 'Set the ' ,spec ,' of the receiver.'
			as: (self buildSetterCode: spec)
			category: #accessing
	]! !

! Behavior methodsFor: 'Builders-api' !
buildGetters: specs
	" Build methods to get instances. "

	specs isString ifTrue: [ ^self buildGetters: specs asArrayOfSubstrings ].
	^self build: #Getters -> specs doing: [
		specs compose: [:each| self buildGetter: each ]
	]! !

! Behavior methodsFor: 'Builders-api' !
buildSetters: specs
	" Build methods to set instances. "

	specs isString ifTrue: [ ^self buildSetters: specs asArrayOfSubstrings ].
	^self build: #Setters -> specs doing: [
		specs compose: [:each| self buildSetter: each ]
	]! !

! Behavior methodsFor: 'Builders-api' !
buildAccessors: specs
	" Build methods to access instances. "

	specs isString ifTrue: [ ^self buildAccessors: specs asArrayOfSubstrings ].
	^self build: #Accessors -> specs doing: [
		specs compose: [:each| self buildAccessor: each ]
	]! !

! Behavior methodsFor: 'Builders-api' !
buildAccessors
	" Private - Build methods to access all instance variables. "

	^self build: #Accessors -> #() doing: [
		self buildAccessors: self instanceVariableNames
	]! !

! Behavior methodsFor: 'Builders-api' !
buildPool: poolName with: spec
	" Private - Build a PoolDictionary with spec (and log definition in log stream). "

	^self build: #Pool -> spec doing: [
		Smalltalk current
			definePool: poolName with: spec;
			printDefinePool: poolName with: spec;
			yourself
	]! !

! Object methodsFor: 'Builders-emmiting' !
isFile: fileName olderThan: referenceFileName
	" Private - Return true if the file at pathName is older than file at referenceFileName. "

	{'self'}.
	(Smalltalk global basicAt: #isFileOlderThan) isNil ifTrue: [ ^false ].
	^# #isFileOlderThan: fileName with: referenceFileName! !

! Object methodsFor: 'Builders-emmiting' !
emmit: fileName startingWith: initial
	" FileIn the contents of file at fileName redirecting output(emmited code) to a native file. "

	| output nativeFileName text |
	nativeFileName := fileName ,$. ,String nativeFileExtension.
	(self isFile: fileName olderThan: nativeFileName) ifTrue: [
		self fileInJS: nativeFileName.
		^# #fileContents: nativeFileName
	].
	output := '' writeStream.
	initial notNil ifTrue: [ output nextPutAll: initial ].
	self note: 'emmit(''' ,fileName ,''')'.
[	self	printingWith: [:what|
			text := '' ,what.
			output cr; nextPutAll: text
		] do: [ self fileIn: fileName ].
	text := output contents.
	text size > 0 ifTrue: [ text outputToFile: nativeFileName ].
]	on: Exception do: [:ex| | reason |
	reason := ex toString.
	(output cr; nextPutAll: reason) contents outputToFile: fileName ,'.error.txt'.
	self print: reason.
	^reason
].	^text! !

! Object methodsFor: 'Builders-emmiting' !
emmit: fileName
	" FileIn the contents of file at fileName redirecting output(emmited code) to a native file. "

	fileName isString ifFalse: [ ^fileName asArray emmit ].
	^self emmit: fileName startingWith: nil ! !

! Collection methodsFor: 'Builders-emmiting' !
emmit
	" Emmit (fileIn & save generated code of) the fileNames in the receiver. "

	^self asArray compose: [:each| self emmit: each ] with: '
'! !

! Collection methodsFor: 'Builders-emmiting' !
emmitSt
	" Emmit the code of source modules (st files) in the receiver. "

	^(self collect: [:module| module ,$. ,String sourceFileExtension ]) emmit! !

! Array methodsFor: 'Builders-emmiting' !
emmitAllIn: folder
	" Emmit(fileIn & save generated code of) the fileNames in the receiver.
	Note the receiver MUST include the file separator at end
	"

	^(self collect: [:each| folder , each ]) emmitAll! !

! Array methodsFor: 'Builders-emmiting' !
emmitAll
	" Deferred (consuming) emmit of fileNames in the receiver. "

	self isEmpty ifTrue: [ ^self ].
	[	self emmit: (Array with: self removeFirst).
		self emmitAll
	] valueDeferred: 500! !

! String methodsFor: 'Builders-emmiting' !
emmitFiles
	" Emmit all the files listed in emmit.txt "

	| folder fileNames stream line |
	folder := self.
	fileNames := #().
	stream := (folder ,'emmit.txt') fileContents stream.
	[ stream atEnd ] whileFalse: [
		line := (stream nextLine upTo: $;) trimBlanks.
		line notEmpty ifTrue: [
			line first = $@ ifTrue: [
				folder := (line copyFrom: 2 to: line size) trimBlanks.
				folder isEmpty ifTrue: [ folder := self ]
			] ifFalse: [ fileNames add: folder ,line ]
		]
	].
	fileNames emmitAll! !
