" S8 Compiler framework
Copyright (C) 2012-today Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
In case of questions and bug reports please join us at S8 Facebook group
	https://www.facebook.com/groups/s8smalltalk/
"!

Object
	subclass: #CompilerNode
	instanceVariableNames: ' nodes start end annotations client'
	category: #Compiler-Nodes !

! CompilerNode class methodsFor: #instantiation !
with: aNode
	^self new
		addNode: aNode;
		yourself! !

! CompilerNode methodsFor: #accessing !
nodes
	nodes isNil ifTrue: [ nodes := Array new ].
	^nodes! !

! CompilerNode methodsFor: #accessing !
nodes: aCollection
	nodes := aCollection! !

! CompilerNode methodsFor: #testing !
isEmpty
	" Return true if the receiver is empty. "

	^nodes size = 0 ! !

! CompilerNode methodsFor: #accessing !
annotations
	annotations isNil ifTrue: [ annotations := Array new ].
	^annotations! !

! CompilerNode methodsFor: #accessing !
annotations: aCollection
	annotations := aCollection! !

! CompilerNode methodsFor: #accessing !
addNode: aNode
	" Add a Node to the nodes of the receiver.
	Return the receiver. "

	self nodes add: aNode! !

! CompilerNode methodsFor: #accessing !
start
	^start! !

! CompilerNode methodsFor: #accessing !
end
	^end! !

! CompilerNode methodsFor: #accessing !
start: startOffset end: endOffset
	" Set the offsets of the receiver in source code. "

	start := startOffset.
	end := endOffset.! !

! CompilerNode methodsFor: #accessing !
sourceRange
	" Return the receiver's sourceRange. "

	^start to: end ! !

! CompilerNode methodsFor: #accessing !
client
	" Private - Return an object that require the result of evaluation of the receiver.
	Note that client can be:
		aNode that will demmand result on runtime
		nil in case of not known
		#none if granted that the result of evaluation of the receiver will be discarded/unused.
	"

	^client ! !

! CompilerNode methodsFor: #accessing !
client: anObject
	" Private - Set an object that require the result of evaluation of the receiver. "

	client := anObject ! !

! CompilerNode methodsFor: #prepare !
prepareIn: aNode for: anEncoder
	" Prepare the receiver to be encoded by anEncoder in aNode. "

	nodes size < 1 ifTrue: [ ^self ].
	nodes do: [:each| each client: #none ].
	nodes last client: self! !

! CompilerNode methodsFor: #accessing !
sourceRange: interval
	" Set the source range of the receiver. "

	self start: interval start end: interval end ! !

! CompilerNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitNode: self! !

! CompilerNode methodsFor: #visiting !
acceptMacroAssignment: anExpressionNode with: aVisitor
	" Return true if the receiver can resolve assignment on aVisitor. "

	^aVisitor visitMacroAssignment: self to: anExpressionNode! !

! CompilerNode methodsFor: #testing !
isSuper
	" Return true if the receiver is super. "

	^false! !

! CompilerNode methodsFor: #testing !
hasNumber
	" Return true if the receiver has a numeric literal value. "

	^false! !

! CompilerNode methodsFor: #testing !
hasString
	" Return true if the receiver has a string literal value. "

	^false! !

! CompilerNode methodsFor: #testing !
isAssignment
	" Return true if the receiver is of this type. "

	^false! !

! CompilerNode methodsFor: #testing !
isBlock
	" Return true if the receiver is of this type. "

	^false! !

! CompilerNode methodsFor: #testing !
isCascade
	" Return true if the receiver is of this type. "

	^false! !

! CompilerNode methodsFor: #testing !
isInline
	" Return true if the receiver is of this type. "

	^false! !

! CompilerNode methodsFor: #testing !
isReturn
	" Return true if the receiver is of this type. "

	^false! !

! CompilerNode methodsFor: #testing !
isSend
	" Return true if the receiver is of this type. "

	^false! !

! CompilerNode methodsFor: #testing !
isSequence
	" Return true if the receiver is of this type. "

	^false! !

! CompilerNode methodsFor: #testing !
isValue
	" Return true if the receiver is of this type. "

	^false! !

! CompilerNode methodsFor: #testing !
isVariable
	" Return true if the receiver is of this type. "

	^false! !

! CompilerNode methodsFor: #bindings !
bindings
	" Return the bindings of the receiver. "

	^#()! !

! CompilerNode methodsFor: #iteration !
do: aBlock
	" Evaluate aBlock for all contents of the receiver. "

	self do: aBlock in: nil ! !

! CompilerNode methodsFor: #iteration !
do: aBlock in: aNode
	" Evaluate aBlock for all contents of the receiver. "

	aBlock evaluateWithArguments: (Array with: self with: aNode).
	nodes notNil ifTrue: [
		nodes do: [:each| each do: aBlock in: self ]
	] ! !

! CompilerNode methodsFor: #iteration !
with: testBlock except: except do: aBlock
	" Evaluate aBlock for all nodes (returning true when evaluated with testBlock) in the local context of receiver. "

	(except evaluateWith: self) == true ifTrue: [ ^nil ].
	nodes notNil ifTrue: [
		nodes do: [:each|
			(testBlock value: each)
			ifTrue: [ aBlock value: each value: self ]
			ifFalse: [ each with: testBlock except: except do: aBlock ]
		]
	] ! !

! CompilerNode methodsFor: #enumerating  !
select: aBlock
	" Returns the nodes in the receiver that returns true when evaluated with aBlock. "

	| result |
	result := #().
	self do: [:each|
		(aBlock evaluateWith: each) == true ifTrue: [
			result add: each
		]
	].
	^result! !

CompilerNode
 subclass: #MethodNode
 instanceVariableNames: 'selector arguments source'
 category: #Compiler-Nodes!

! MethodNode methodsFor: #accessing !
selector
	^selector! !

! MethodNode methodsFor: #accessing !
selector: aString
	selector := aString! !

! MethodNode methodsFor: #accessing !
arguments
	^arguments isNil ifTrue: [ #() ] ifFalse: [ arguments ]! !

! MethodNode methodsFor: #accessing !
arguments: aCollection
	arguments := aCollection! !

! MethodNode methodsFor: #accessing !
argumentNames
	" Return the names of the arguments of the receiver. "

	^self arguments collect: [:each|
		each isString
		ifTrue: [each]
		ifFalse: [each value]
	]! !

! MethodNode methodsFor: #accessing !
source
	^source! !

! MethodNode methodsFor: #accessing !
source: aString
	source := aString! !

! MethodNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitMethodNode: self! !

! MethodNode methodsFor: #validation !
validateArguments
	" Ensure the nodes of the receiver can be emmited with the defined arguments. "

	| names |
	names := HashedSet withAll: self argumentNames.
	names size = self arguments size ifFalse: [
		^self error: 'Invalid argument names ' ,self argumentNames asLiteral
	].
	self nodes do: [:each| each validateNames: names ]! !

! MethodNode methodsFor: #accessing !
sendNodes
	" Returns the send nodes in the receiver. "

	^self select: [:each| each isSend ]! !

! MethodNode methodsFor: #accessing !
messages
	" Returns the messages potentially sent when evaluating the receiver. "

	^(self sendNodes
		collect: [:each| each selector ])
		asSet sorted! !

! MethodNode methodsFor: #accessing !
values
	" Returns the values bound in the receiver. "

	| result |
	result := #().
	result addAll: self arguments.
	self do: [:each|
		each isSequence ifTrue: [ result addAll: each temps ].
		each isBlock ifTrue: [ result addAll: each parameters ].
	].
	^result! !

CompilerNode
 subclass: #SendNode
 instanceVariableNames: 'selector arguments receiver action '
 category: #Compiler-Nodes!

! SendNode methodsFor: #accessing !
selector
	^selector! !

! SendNode methodsFor: #accessing !
selector: aString
	selector := aString! !

! SendNode methodsFor: #accessing !
arguments
	^arguments isNil ifTrue: [ #() ] ifFalse: [ arguments ]! !

! SendNode methodsFor: #accessing !
arguments: aCollection
	arguments := aCollection! !

! SendNode methodsFor: #accessing !
receiver
	^receiver! !

! SendNode methodsFor: #accessing !
receiver: aNode
	receiver := aNode! !

! SendNode methodsFor: #accessing !
valueForReceiver: anObject
	^SendNode new
	    receiver: (self receiver isNil ifTrue: [anObject]
		ifFalse: [self receiver valueForReceiver: anObject]);
	    selector: self selector;
	    arguments: self arguments;
	    start: self start end: self end;
	    yourself! !

! SendNode methodsFor: #accessing !
cascadeNodeWithMessages: aCollection
	| first sends |
	first := SendNode new
	    selector: self selector;
	    arguments: self arguments;
	    start: self start end: self end;
	    yourself.
	sends := (Array with: first), aCollection.
	^CascadeNode new
	    receiver: self receiver;
	    nodes: sends;
	    start: sends first start end: sends last end;
	    yourself! !

! SendNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitSendNode: self! !

! SendNode methodsFor: #testing !
isSend
	" Return true if the receiver is of this type. "

	^true! !

! SendNode methodsFor: #iteration !
do: aBlock in: aNode
	" Evaluate aBlock for all contents of the receiver. "

	super do: aBlock in: aNode.
	receiver notNil ifTrue: [ receiver do: aBlock in: self ].
	self arguments do: [:each| each do: aBlock in: self ].! !

! SendNode methodsFor: #iteration !
with: testBlock except: except do: aBlock
	" Evaluate aBlock for all nodes (returning true when evaluated with testBlock) in the local context of receiver. "

	(super with: testBlock except: except do: aBlock) isNil ifTrue: [ ^nil ].
	receiver notNil ifTrue: [
		(testBlock value: receiver)
		ifTrue: [ aBlock value: receiver value: self ]
		ifFalse: [ receiver with: testBlock except: except do: aBlock ]
	].
	self arguments do: [:each|
		(testBlock value: each)
		ifTrue: [ aBlock value: each value: self ]
		ifFalse: [ each with: testBlock except: except do: aBlock ]
	] ! !

! SendNode methodsFor: #prepare !
prepareIn: aNode for: anEncoder
	" Prepare the receiver to be encoded by anEncoder in aNode. "

	super prepareIn: aNode for: anEncoder.
	receiver notNil ifTrue: [ receiver client: self ].
	self arguments do: [:each| each client: self ] ! !

CompilerNode
 subclass: #ScopedNode
 category: #Compiler-Nodes!

ScopedNode
 subclass: #CascadeNode
 instanceVariableNames: #receiver
 category: #Compiler-Nodes!

! CascadeNode methodsFor: #accessing !
receiver
	^receiver! !

! CascadeNode methodsFor: #accessing !
receiver: aNode
	receiver := aNode! !

! CascadeNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitCascadeNode: self! !

! CascadeNode methodsFor: #testing !
isCascade
	" Return true if the receiver is of this type. "

	^true! !

! CascadeNode methodsFor: #iteration !
do: aBlock in: aNode
	" Evaluate aBlock for all contents of the receiver. "

	super do: aBlock in: aNode.
	receiver notNil ifTrue: [ receiver do: aBlock in: self ].! !

! CascadeNode methodsFor: #iteration !
with: testBlock except: except do: aBlock
	" Evaluate aBlock for all nodes (returning true when evaluated with testBlock) in the local context of receiver. "

	(super with: testBlock except: except do: aBlock) isNil ifTrue: [ ^nil ].
	receiver notNil ifTrue: [
		(testBlock value: receiver)
		ifTrue: [ aBlock value: receiver value: self ]
		ifFalse: [ receiver with: testBlock except: except do: aBlock ]
	] ! !

CompilerNode
 subclass: #AssignmentNode
 instanceVariableNames: 'left right'
 category: #Compiler-Nodes!

! AssignmentNode methodsFor: #accessing !
left
	^left! !

! AssignmentNode methodsFor: #accessing !
left: aNode
	left := aNode! !

! AssignmentNode methodsFor: #accessing !
right
	^right! !

! AssignmentNode methodsFor: #accessing !
right: aNode
	right := aNode! !

! AssignmentNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitAssignmentNode: self! !

! AssignmentNode methodsFor: #testing !
isAssignment
	" Return true if the receiver is an assignment node. "

	^true! !

! AssignmentNode methodsFor: #iteration !
do: aBlock in: aNode
	" Evaluate aBlock for all contents of the receiver. "

	super do: aBlock in: aNode.
	left  notNil ifTrue: [ left  do: aBlock in: self ].
	right notNil ifTrue: [ right do: aBlock in: self ].! !

! AssignmentNode methodsFor: #iteration !
with: testBlock except: except do: aBlock
	" Evaluate aBlock for all nodes (returning true when evaluated with testBlock) in the local context of receiver. "

	(super with: testBlock except: except do: aBlock) isNil ifTrue: [ ^nil ].
	right notNil ifTrue: [
		(testBlock value: right)
		ifTrue: [ aBlock value: right value: self ]
		ifFalse: [ right with: testBlock except: except do: aBlock ]
	] ! !

! AssignmentNode methodsFor: #prepare !
prepareIn: aNode for: anEncoder
	" Prepare the receiver to be encoded by anEncoder in aNode. "

	super prepareIn: aNode for: anEncoder.
	right client: self! !

ScopedNode
 subclass: #BlockNode
 instanceVariableNames: #parameters
 category: #Compiler-Nodes!

! BlockNode methodsFor: #accessing !
parameterNames
	" Return the names of the parameters of the receiver. "

	| result |
	result := self parameters collect: [:each|
		each isString
		ifTrue: [each]
		ifFalse: [each value]
	].
	result asSet size = result size ifFalse: [
		^self error: 'Invalid names ' ,result asLiteral
	].
	^result! !

! BlockNode methodsFor: #accessing !
parameters
	parameters isNil ifTrue: [ parameters := Array new ].
	^parameters! !

! BlockNode methodsFor: #accessing !
parameters: aCollection
	parameters := aCollection! !

! BlockNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitBlockNode: self! !

! BlockNode methodsFor: #bindings !
bindings
	" Return the bindings of the receiver. "

	^self parameterNames! !

! BlockNode methodsFor: #testing !
isBlock
	" Return true if the receiver is of this type. "

	^true! !

! BlockNode methodsFor: #iteration !
do: aBlock in: aNode
	" Evaluate aBlock for all contents of the receiver. "

	super do: aBlock in: aNode.
	self parameters do: [:each| each do: aBlock in: self ].! !

ScopedNode
 subclass: #SequenceNode
 instanceVariableNames: #temps
 category: #Compiler-Nodes!

! SequenceNode methodsFor: #accessing !
tempNames
	" Return the names of the temps in the receiver. "

	^self temps collect: [:each|
		each isString
		ifTrue: [each]
		ifFalse: [each value]
	]! !

! SequenceNode methodsFor: #accessing !
temps
	^temps isNil ifTrue: [ #() ] ifFalse: [ temps ]! !

! SequenceNode methodsFor: #accessing !
temps: aCollection
	temps := aCollection! !

! SequenceNode methodsFor: #converting !
asBlockSequenceNode
	^BlockSequenceNode new
	    nodes: self nodes;
	    temps: self temps;
	    start: self start end: self end;
	    yourself! !

! SequenceNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitSequenceNode: self! !

! SequenceNode methodsFor: #bindings !
bindings
	" Return the bindings of the receiver. "

	^self tempNames! !

! SequenceNode methodsFor: #testing !
isSequence
	" Return true if the receiver is of this type. "

	^true! !

! SequenceNode methodsFor: #return !
shouldReturn
	" Ensure the receiver includes a return node. "

	(self nodes notEmpty and: [
		self nodes last isReturn
	]) ifTrue: [ ^self].

	self addNode: (ReturnNode variable: #self at: self end).! !

! SequenceNode methodsFor: #validation !
validateNames: names
	" Validate the receiver can be emmited in a context with names. "

	self tempNames do: [:temp|
		(names includes: temp) ifTrue: [
			^self error: 'Invalid temp (',temp,')'
		].
	]! !

SequenceNode
 subclass: #BlockSequenceNode
 category: #Compiler-Nodes!

! BlockSequenceNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitBlockSequenceNode: self! !

SequenceNode
 subclass: #ExtendedSequenceNode
 instanceVariableNames: #directives
 category: #Compiler-Nodes!

! ExtendedSequenceNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitExtendedSequenceNode: self! !

! ExtendedSequenceNode methodsFor: #accessing !
directives
	" Return the directives of the receiver. "

	^directives isNil
		ifTrue: [ #() ]
		ifFalse: [directives]!

directives: aCollection
	" Set the directives of the receiver. "

	directives := aCollection! !

CompilerNode
 subclass: #ReturnNode
 instanceVariableNames: ' exitBlock '
 category: #Compiler-Nodes!

! ReturnNode class methodsFor: #instantiation !
variable: aName at: offset
	" Return an instance of the receiver. "

	^self new
		addNode: (VariableNode new
			value: aName;
			start: offset end: offset;
			yourself);
		start: offset end: offset;
		yourself! !

! ReturnNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitReturnNode: self! !

! ReturnNode methodsFor: #testing !
isReturn
	" Return true if the receiver is of this type. "

	^true! !

CompilerNode
 subclass: #ValueNode
 instanceVariableNames: ' value binding '
 category: #Compiler-Nodes!

! ValueNode class methodsFor: #instantiation !
with: aValue
	" Return an instance of the receiver. "

	^self new value: aValue! !

! ValueNode methodsFor: #accessing !
value
	^value! !

! ValueNode methodsFor: #accessing !
value: anObject
	value := anObject! !

! ValueNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitValueNode: self! !

! ValueNode methodsFor: #testing !
hasNumber
	" Return true if the receiver has a numeric literal value. "

	^value isNumber! !

! ValueNode methodsFor: #testing !
hasString
	" Return true if the receiver has a string literal value. "

	^value isString! !

! ValueNode methodsFor: #testing !
isValue
	" Return true if the receiver is of this type. "

	^true! !

ValueNode
 subclass: #VariableNode
 instanceVariableNames: #indices
 category: #Compiler-Nodes!

! VariableNode methodsFor: #accessing !
localNameIn: aCompiler
	" Return the name of the receiver to be used by aCompiler locally. "

	^aCompiler localName: self value ! !

! VariableNode methodsFor: #testing !
needsVerification
	" Return true if the receiver need verification. "

	^true! !

! VariableNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitVariableNode: self! !

! VariableNode methodsFor: #testing !
isSuper
	" Return true if the receiver is super. "

	^value = 'super'! !

! VariableNode methodsFor: #accessing !
indices
	" Return the indexing references of the receiver. "

	^indices isNil ifTrue: [ Array new ] ifFalse: [ indices ]! !

! VariableNode methodsFor: #visiting !
reference: chain
	" The receiver is a reference defined by chain. "

	indices := (chain size > 1 and: [ chain second notEmpty ]) ifTrue: [
		chain second collect: [:tuple|
			tuple second isString
			ifTrue: [ ValueNode new value: tuple second ]
			ifFalse: [ tuple second ]
		]
	].
	(chain size > 2 and: [ chain third notEmpty ]) ifTrue: [
		self annotations: chain third
	].
	^self! !

! VariableNode methodsFor: #testing !
isVariable
	" Return true if the receiver is of this type. "

	^true! !

VariableNode
 subclass: #GlobalReferenceNode
 category: #Compiler-Nodes!

! GlobalReferenceNode methodsFor: #visiting !
accept: aVisitor
	aVisitor visitGlobalReferenceNode: self! !

! CompilerNode methodsFor: #accessing !
messageSendConnector
	" Return the connector used to encode message sent to the receiver (or nil). "

	^$. ! !

VariableNode
 subclass: #NoReceiverNode
 category: #Compiler-Nodes!

! NoReceiverNode methodsFor: #visiting !
accept: aVisitor! !

! NoReceiverNode methodsFor: #accessing !
messageSendConnector
	" Return the connector used to encode message sent to the receiver (or nil). "

	^nil ! !

VariableNode
 subclass: #AlienNode
 category: #Compiler-Nodes!

! AlienNode methodsFor: #accessing !
localNameIn: aCompiler
	" Return the name of the receiver to be used by aCompiler locally. "

	^aCompiler alienName: self value ! !

! AlienNode methodsFor: #testing !
needsVerification
	" Return true if the receiver need verification. "

	^false! !

CompilerNode
 subclass: #InlineNode
 instanceVariableNames: #sources
 category: #Compiler-Nodes!

! InlineNode class methodsFor: #instantiation !
sources: tuples parser: parser
	^self basicNew
		initialize: tuples
		parser: parser ! !

! InlineNode methodsFor: #initialize !
initialize: tuples parser: parser
	" Private - Initialize the receiver "

	sources := tuples.
	self capture: #s8 with: parser. ! !

! InlineNode methodsFor: #accessing !
sourceFor: lang in: encoder
	" Return the native source code to inline (or nil). "

	| default |
	sources do: [:tuple| | code |
		code := tuple second value.
		tuple first = lang ifTrue: [ ^code ].
		tuple first isNil ifTrue: [ default := code ].
	].
	encoder notNil ifTrue: [ | result |
		result := encoder inline: self source: default.
		result notNil ifTrue: [ ^result ]
	].
	^default! !

! InlineNode methodsFor: #visiting !
accept: aVisitor
	^aVisitor visitInlineNode: self! !

! InlineNode methodsFor: #testing !
isInline
	" Return true if the receiver is of this type. "

	^true! !

! InlineNode methodsFor: #internal !
capture: lang with: parser
	" Private - Capture internal contents from sources. "

	sources do: [:tuple| tuple first = lang ifTrue: [
		self parse: tuple second value at: tuple second start with: parser	
	]]! !

! InlineNode methodsFor: #internal !
parse: code at: offset with: parser
	" Private - Parse internal contents from code using parser. "

	| stream array |
	stream := '' stream.
	offset timesRepeat: [ stream nextPut: $ ].
	stream nextPutAll: code; position: offset.
	array := parser ,PPEOFParser new parse: stream.
	(array isArray and: [array size > 1]) ifFalse: [
		self error: 'Invalid inline code ' ,code asString
	].
	self addNode: array second! !

CompilerNode
 subclass: #DirectiveNode
 instanceVariableNames: ''
 category: #Compiler-Nodes!

! DirectiveNode methodsFor: #visiting !
description
	" Return the description of the receiver. "

	^self class name!

selector
	" Private - Return the internal selector used to execute directive. "

	^#directive:arguments: asSelector!

literal
	" Private - Return the literal to embedd in directive call. "

	^self implementedBySubclass! !

! DirectiveNode methodsFor: 'code generation' !
asNative: selector for: anEncoder on: aStream
	" Return the native code to emmit the receiver. "

	| call |
	call := anEncoder
		sourceToDirectCall: self selector
		with: self literal.
	^anEncoder sourceToTry: call ! !

DirectiveNode
 subclass: #JSDirectiveNode
 instanceVariableNames: 'code protected'
 category: #Compiler-Nodes!

! JSDirectiveNode class methodsFor: #instantiation !
code: aCode protected: useTry
	" Return an instance of the receiver. "

	^self new
		code: aCode;
		protected: useTry;
		yourself! !

! JSDirectiveNode methodsFor: #accessing !
code: aCode
	" Set the code of the receiver. "

	code := aCode! !

! JSDirectiveNode methodsFor: #accessing !
nativeCodeFor: anEncoder
	" Private - Return the native code of the receiver. "

	code isString ifTrue: [ ^code ].
	^code evaluateWith: anEncoder ! !

! JSDirectiveNode methodsFor: #accessing !
protected: aBoolean
	" Set the receiver to use try/catch to handle failure. "

	protected := aBoolean! !

! JSDirectiveNode methodsFor: 'code generation' !
asNative: selector for: anEncoder on: aStream
	" Return the native code to emmit the receiver. "

	| string |
	string := self nativeCodeFor: anEncoder.
	protected == true ifTrue: [
		^anEncoder sourceToTry: string
	].
	^string ! !

DirectiveNode
	subclass: #InlineOptimizationNode
	instanceVariableNames: ' args remove '
	category: #Compiler-Nodes!
InlineOptimizationNode buildAccessors!

! InlineOptimizationNode methodsFor: 'code generation' !
asNative: selector for: anEncoder on: aStream
	" Return the native code to emmit the receiver. "

	self remove
		ifTrue: [ anEncoder removeOptimizations: args ]
		ifFalse: [ anEncoder addOptimizations: args ].
	^nil ! !

EventManager
	subclass: #S8Encoder
	instanceVariableNames: ' compiler '
	category: #Compiler-Encoding!

! S8Encoder class methodsFor: #instantiation !
for: aCompiler
	" Private - Return an instance of the receiver. "

	^self support new compiler: aCompiler! !

! S8Encoder class methodsFor: #instantiation !
support
	" Private - Return the support to encode for native platform. "

	| lang |
	lang := Smalltalk nativeLanguage.
	^self allSubclasses
		detect: [:cls| cls supportsLanguage: lang ]
		ifNone: [ self invalidMessage: #support ]! !

! S8Encoder class methodsFor: #instantiation !
supportsLanguage: aLanguage
	" Private - Return true if instances of the receiver can encode to aLanguage. "

	^false! !

! S8Encoder methodsFor: #encoding !
messageSendConnector
	" Private - Return the connector to use for sending a message. "

	^self implementedBySubclass: #messageSendConnector ! !

! S8Encoder methodsFor: #encoding !
propertyAccessConnector
	" Private - Return the connector to use for accessing a native property. "

	^self messageSendConnector ! !

! S8Encoder methodsFor: #encoding !
codeToGetGlobal: aName
	" Private - Return the native code to get global at aName. "

	^#smalltalk ,self propertyAccessConnector ,aName ! !

! S8Encoder methodsFor: #encoding !
nativeUndefined
	" Private - Return the code for undefined. "

	^self implementedBySubclass: #nativeUndefined ! !

! S8Encoder methodsFor: #encoding !
nativeNameOfMetaClass: aClass
	" Private - Return the native name for metaclass aClass. "

	^self implementedBySubclass: #nativeNameOfMetaClass: ! !

! S8Encoder methodsFor: #encoding !
nativeNameOfClass: aClass
	" Private - Return the native name for aClass. "

	aClass isMetaclass ifTrue: [
		^self nativeNameOfMetaClass: aClass
	].
	^aClass name ! !

! S8Encoder methodsFor: #export !
smalltalkNameFor: aClass
	" Private - Return the global name of aClass (can be nil). "

	aClass isNil ifTrue: [ ^self nativeUndefined ].
	^self codeToGetGlobal: (self nativeNameOfClass: aClass) ! !

! S8Encoder methodsFor: #accessing !
compiler
	" Return the compiler of the receiver. "

	^compiler! !

! S8Encoder methodsFor: #accessing !
compiler: aCompiler
	" Set the compiler of the receiver. "

	compiler := aCompiler! !

! S8Encoder methodsFor: #events !
compilerEvent: selector
	" Return the result of triggering the compiler's event. "

	^compiler triggerEvent: selector! !

! S8Encoder methodsFor: #events !
compilerEvent: selector ifNotHandled: aBlock
	" Return the result of triggering the compiler's event. "

	^compiler triggerEvent: selector ifNotHandled: aBlock! !

! S8Encoder methodsFor: #events !
compilerEvent: selector with: anObject
	" Return the result of triggering the compiler's event with anObject. "

	^compiler
		triggerEvent: selector
		with: anObject! !

! S8Encoder methodsFor: #events !
compilerEvent: selector with: anObject ifNotHandled: aBlock
	" Return the result of triggering the compiler's event with anObject. "

	^compiler
		triggerEvent: selector
		with: anObject
		ifNotHandled: aBlock! !

! S8Encoder methodsFor: #events !
compilerEvent: selector withArguments: array
	" Return the result of triggering the compiler's event with the arguments. "

	^compiler
		triggerEvent: selector
		withArguments: array! !

! S8Encoder methodsFor: #events !
compilerEvent: selector withArguments: array ifNotHandled: aBlock
	" Return the result of triggering the compiler's event with the arguments. "

	^compiler
		triggerEvent: selector
		withArguments: array
		ifNotHandled: aBlock! !

! S8Encoder methodsFor: #accessing !
currentClass
	" Return the current class of the receiver. "

	^compiler currentClass! !

! S8Encoder methodsFor: #visiting !
ignore: aNode! !

! S8Encoder methodsFor: #converting !
asSelector: aSelector
	" Return the native representation of aSelector. "

	^aSelector asSelector! !

! S8Encoder methodsFor: #variables !
knownVariables
	" Return the knownVariables of the receiver. "

	^compiler knownVariables! !

! S8Encoder methodsFor: #private !
localName: aName
	" Private - Return a valid local name for aName. "

	(self pseudoVariables includes: aName) ifTrue: [ ^aName ].
	^self asSelector: aName! !

! S8Encoder methodsFor: #private !
alienName: aName
	" Private - Return a valid native/alien name for aName. "

	^aName! !

! S8Encoder methodsFor: #accessing !
nativeLanguage
	" Return the nativeLanguage of the receiver.
	The default action return nil (the default native language).
	"

	^nil! !

! S8Encoder methodsFor: #results !
emmitedCode
	" Return the emmited code during compilation (or nil). "

	^self implementedBySubclass: #emmitedCode! !

! S8Encoder methodsFor: #private !
endCompiling: aMethodNode
	" Private - Finish emmiting code for aMethodNode. "! !

! S8Encoder methodsFor: #private !
startCompiling: aNode
	" Private - Set the receiver to start emmiting code for aNode. "! !

! S8Encoder methodsFor: #prepare !
prepare: aNode
	" Private - Prepare aNode for traversal. "

	aNode isParseFailure ifTrue: [ ^nil ].
	aNode do: [:each :parent|
		each prepareIn: parent for: self
	]! !

! S8Encoder methodsFor: #variables !
pseudoVariables
	" Return the pseudoVariables of the receiver. "

	^compiler pseudoVariables! !

! S8Encoder methodsFor: #variables !
tempVariables
	" Return the tempVariables of the receiver. "

	^compiler tempVariables! !

! S8Encoder methodsFor: #variables !
unknownVariables
	" Return the unknownVariables of the receiver. "

	^compiler unknownVariables! !

! S8Encoder methodsFor: #visiting !
visit: aNode
	^aNode accept: self! !

! S8Encoder methodsFor: #visiting !
visitAssignmentNode: aNode
	self visitNode: aNode! !

! S8Encoder methodsFor: #visiting !
visitBlockNode: aNode
	self visitNode: aNode! !

! S8Encoder methodsFor: #visiting !
visitBlockSequenceNode: aNode
	self visitSequenceNode: aNode! !

! S8Encoder methodsFor: #visiting !
visitCascadeNode: aNode
	self visitNode: aNode! !

! S8Encoder methodsFor: #visiting !
visitMacroAssignment: aVariableNode to: anExpressionNode
	" Return true if the receiver generated code to a macro assignment to aVariableNode. "

	^false! !

! S8Encoder methodsFor: #visiting !
visitMethodNode: aNode
	" Visit a method node. "

	self visitNode: aNode! !

! S8Encoder methodsFor: #visiting !
visitNode: aNode
	self ignore: aNode! !

! S8Encoder methodsFor: #visiting !
visitReturnNode: aNode
	self visitNode: aNode! !

! S8Encoder methodsFor: #visiting !
visitSendNode: aNode
	self visitNode: aNode! !

! S8Encoder methodsFor: #visiting !
visitSequenceNode: aNode
	self visitNode: aNode! !

! S8Encoder methodsFor: #visiting !
visitValueNode: aNode
	self visitNode: aNode! !

! S8Encoder methodsFor: #visiting !
visitVariableNode: aNode
	self ignore: aNode! !

! S8Encoder methodsFor: #private !
translate: aNode by: delta
	" Private - Translate the contents of aNode by delta characters in source code. "

	delta = 0 ifTrue: [ ^self ].
	self	compilerEvent: #move:by:
		withArguments: (Array with: aNode with: delta).! !

! S8Encoder methodsFor: #export !
loadRequirementsOf: aClass
	" Private - Return the requirements to load aClass. "

	^self implementedBySubclass: #loadRequirementsOf: ! !

! S8Encoder methodsFor: #export !
exportDefinitionOf: aClass on: aStream
	" Export definition of aClass onto aStream. "

	^self implementedBySubclass: #exportDefinitionOf:on: ! !

! S8Encoder methodsFor: #export !
exportMetaDefinitionOf: aClass on: aStream
	" Export class side definition of aClass onto aStream. "

	^self implementedBySubclass: #exportMetaDefinitionOf:on: ! !

! S8Encoder methodsFor: #export !
exportCommentOf: aClass on: aStream
	" Export the comment of aClass onto aStream. "

	^self implementedBySubclass: #exportCommentOf:on: ! !

! S8Encoder methodsFor: #export !
export: aClass method: aMethod on: aStream
	" Export a method definition of aClass onto aStream. "

	^self implementedBySubclass: #export:method:on: ! !

! S8Encoder methodsFor: #literals !
arrayLiteralStart
	" Private - Return the code string to start exporting an array literal. "

	^self implementedBySubclass: #arrayLiteralStart! !

! S8Encoder methodsFor: #literals !
arrayLiteralEnd
	" Private - Return the code string to end exporting an array literal. "

	^self implementedBySubclass: #arrayLiteralEnd! !

! S8Encoder methodsFor: #literals !
escaped: string
	" Private - Return the code string to export string literal. "

	^$" ,string ,$"! !

! S8Encoder methodsFor: #constants !
argumentsSeparatorOp
	" Private - Return the OP for separating arguments. "

	^$, ! !

! S8Encoder methodsFor: #literals !
export: tags literals: anArray on: aStream
	" Private - Export the native code for elements in anArray between tags. "

	tags notNil ifTrue: [ aStream nextPutAll: tags first ].
	anArray
	    do: [:each | aStream nextPutAll: (
		each isString ifTrue: [ $" ,each ,$" ] ifFalse: [ each printString ]
	    ) ] separatedBy: [ aStream nextPutAll: self argumentsSeparatorOp ].
	tags notNil ifTrue: [ aStream nextPutAll: tags last ]! !

! S8Encoder methodsFor: #literals !
exportArrayLiteral: anArray on: aStream newLine: newLine
	" Private - Export the native code for elements in anArray. "

	self	export: (Array
			with: self arrayLiteralStart
			with: self arrayLiteralEnd)
		literals: anArray on: aStream.
	newLine ifTrue: [ aStream cr ]! !

! S8Encoder methodsFor: #directives !
sourceToTry: sourceCode
	" Return the code to emmit to try execution of sourceCode. "

	^sourceCode ! !

! S8Encoder methodsFor: #directives !
sourceToDirectCall: nativeSelector with: arg
	" Return the code to emmit to direct native call to selector with argument. "

	^self implementedBySubclass: #sourceToDirectCall:with: ! !

S8Encoder
	subclass: #OptimizingEncoder
	instanceVariableNames: ' optimizationMap knownOptimizations inlines '
	category: #Compiler-Encoding !
OptimizingEncoder class instanceVariableNames: ' inlines ' !

! OptimizingEncoder methodsFor: #optimization !
addOptimizations: array
	" Private - Add an array of optimizations to the receiver. "

	array do: [:opt|
		self optimizationMap addAll:
			(self optimizationLiteralsFor: opt)
	]! !

! OptimizingEncoder methodsFor: #optimization !
allBlockNoArguments: aNode
	" Private - Return true if receiver and arguments of aNode are blocks with no arguments. "

	^(self receiverIsBlockNoArguments: aNode) and: [
		self hasBlockNoArguments: aNode
	]! !

! OptimizingEncoder methodsFor: #optimization !
arg1HasNumber: aNode
	" Private - Return true if argument of aNode is a number literal. "

	^(aNode arguments at: 1) hasNumber! !

! OptimizingEncoder methodsFor: #optimization !
arg2HasNumber: aNode
	" Private - Return true if argument of aNode is a number literal. "

	^(aNode arguments at: 2) hasNumber! !

! OptimizingEncoder methodsFor: #optimization !
arg3HasNumber: aNode
	" Private - Return true if argument of aNode is a number literal. "

	^(aNode arguments at: 3) hasNumber! !

! OptimizingEncoder methodsFor: #optimization !
argumentsHasNumber: aNode
	" Private - Return true if any argument of aNode is a number literal. "

	aNode arguments do: [:each|
		each hasNumber ifTrue: [ ^true ]
	].
	^false! !

! OptimizingEncoder methodsFor: #optimization !
argumentsHasString: aNode
	" Private - Return true if any argument of aNode is a string literal. "

	aNode arguments do: [:each|
		each hasString ifTrue: [ ^true ]
	].
	^false! !

! OptimizingEncoder methodsFor: #defaults !
defaultKnownOptimizations
	" Private - Return the default knownOptimizations of the receiver. "

	^#()! !

! OptimizingEncoder methodsFor: #defaults !
defaultOptimizations
	" Private - Return the default optimizations of the receiver. "

	^#()! !

! OptimizingEncoder methodsFor: #optimization !
hasBlockArguments: aNode
	" Private - Return true if arguments of aNode are blocks. "

	aNode arguments do: [:each|
		(each isKindOf: BlockNode) ifFalse: [ ^false ].
	].
	^true! !

! OptimizingEncoder methodsFor: #optimization !
hasBlockNoArguments: aNode
	" Private - Return true if arguments of aNode are blocks with no parameters. "

	(self hasBlockArguments: aNode) ifFalse: [ ^false ].
	aNode arguments do: [:each|
		each parameters isEmpty ifFalse: [ ^false ].
	].
	^true! !

! OptimizingEncoder methodsFor: #optimization !
hasBlockOneArguments: aNode
	" Private - Return true if arguments of aNode are blocks with ONE parameter. "

	(self hasBlockArguments: aNode) ifFalse: [ ^false ].
	aNode arguments do: [:each|
		each parameters size = 1 ifFalse: [ ^false ].
	].
	^true! !

! OptimizingEncoder methodsFor: #optimization !
hasNumber: aNode
	" Private - Return true if the receiver or arguments in aNode has numeric value. "

	^(self receiverHasNumber: aNode) or: [
		self argumentsHasNumber: aNode
	]! !

! OptimizingEncoder methodsFor: #optimization !
hasString: aNode
	" Private - Return true if the receiver or arguments in aNode has string value. "

	^(self receiverHasString: aNode) or: [
		self argumentsHasString: aNode
	]! !

! OptimizingEncoder methodsFor: #optimization !
noClient: aNode
	" Private - Return true if aNode has no client. "

	^aNode client = #none ! !

! OptimizingEncoder methodsFor: #optimization !
hasClient: aNode
	" Private - Return true if aNode has a client. "

	| client |
	client := aNode client.
	^client notNil and: [ client ~= #none ] ! !

! OptimizingEncoder methodsFor: #initialize !
initializeKnownOptimizations
	" Private - Initialize the knownOptimizations of the receiver. "

	knownOptimizations := self defaultKnownOptimizations! !

! OptimizingEncoder methodsFor: #initialize !
initializeOptimizationMap
	" Private - Initialize the optimization map of the receiver. "

	optimizationMap := self
		compilerEvent: #needsOptimizationMap
		ifNotHandled: [ #() ].
	self compilerEvent: #initializeOptimizations ifNotHandled: [
		self addOptimizations: self defaultOptimizations
	].
	self compilerEvent: #customizeOptimizations! !

! OptimizingEncoder methodsFor: #accessing !
knownOptimizations
	" Return the knownOptimizations of the receiver. "

	knownOptimizations isNil ifTrue: [ self initializeKnownOptimizations ].
	^knownOptimizations! !

! OptimizingEncoder methodsFor: #accessing !
knownOptimizations: anObject
	" Set the knownOptimizations of the receiver. "

	knownOptimizations := anObject! !

! OptimizingEncoder methodsFor: #optimization !
matchOptimizedCondition: condition with: aNode
	" Private - Return true if aNode matches the condition. "

	(#(
	hasNumber: receiverHasNumber: argumentsHasNumber:
	hasString: receiverHasString: argumentsHasString:
	arg1HasNumber: arg2HasNumber: arg3HasNumber:
	hasBlockArguments: hasBlockNoArguments: hasBlockOneArguments:
	allBlockNoArguments: receiverIsBlockNoArguments: receiverIsBlock:
	noClient: hasClient:
	) includes: condition) ifTrue: [
		^self perform: condition withArguments: (Array with: aNode)
	].
	condition isArray ifTrue: [
		condition detect: [:one|
			false = (self matchOptimizedCondition: one with: aNode)
		] ifNone: [ ^true ].
		^false
	].
	condition = #true ifTrue: [ ^true ].
	condition = #false ifTrue: [ ^false ].
	self note: 'Unknown optimize-condition: ',condition.
	^false! !

! OptimizingEncoder methodsFor: #optimization !
optimizationLiteralsFor: opt
	" Private - Return the optimization literals array defined for opt.
	opt can be an optimization family or message selector
	"

	| array |
	opt isString ifFalse: [ ^Array with: opt ].
	array := self knownOptimizations.
	array do: [:pair|
		pair first = opt ifTrue: [
			^pair last inject: #() into: [:total :tuple|
				total addAll: (self optimizationLiteralsFor: tuple); yourself
			]
		].
	].
	array do: [:pair|
		pair last do: [:tuple|
			(tuple isArray and: [tuple first = opt]) ifTrue: [ ^Array with: tuple ]
		].
	].
	^#()! !

! OptimizingEncoder methodsFor: #accessing !
optimizationMap
	" Return the message send optimization map of the receiver. "

	optimizationMap isNil ifTrue: [ self initializeOptimizationMap ].
	^optimizationMap! !

! OptimizingEncoder methodsFor: #accessing !
optimizationMap: aSpec
	" Set the current optimizationMap of the receiver. "

	optimizationMap := aSpec! !

! OptimizingEncoder methodsFor: #generation !
emmitCode: code
	" Private - Emmit code onto output stream. "

	^self implementedBySubclass: #emmitCode: ! !

! OptimizingEncoder methodsFor: #generation !
emmitInlinedCode: code
	" Private - Emmit inlined code onto output stream. "

	^self emmitCode: code ! !

! OptimizingEncoder methodsFor: #macros !
macroToken
	" Private - Return the (default) macro token. "

	^$% ! !

! OptimizingEncoder methodsFor: #macros !
macroStart
	" Private - Return the macro start token. "

	^self macroToken ! !

! OptimizingEncoder methodsFor: #macros !
macroEnd
	" Private - Return the macro end token. "

	^self macroToken ! !

! OptimizingEncoder methodsFor: #macros !
macroStream: macro
	" Private - Return a code stream from macro string. "

	^macro readStream ! !

! OptimizingEncoder methodsFor: #macros !
expandMacrosIn: input with: aNode
	" Private - Parse input expanding macros.
	Expandable macros are defined between macroToken
	"

	[input atEnd] whileFalse: [ | code |
		code := input upTo: self macroStart.
		code size > 0 ifTrue: [ self emmitInlinedCode: code ].
		input atEnd ifFalse: [ self visitMacroIn: input with: aNode ]
	] ! !

! OptimizingEncoder methodsFor: #macros !
optimize: aNode macro: template
	" Private - Generate the code for aNode using a macro template.
	Return true if expanded the macro, or leave the receiver unchanged returning false
	"

	template isArray ifTrue: [ "conditional templates"
		template detect: [:one| self visitOptimizedNode: aNode as: one ]
			ifNone: [ ^false ].
		^true
	].
	template isString ifFalse: [ "an evaluable template (must return nil if failed to expand)"
		^(template value: aNode value: self) notNil
	].
	self expandMacrosIn: (self macroStream: template) with: aNode.
	^true! !

! OptimizingEncoder methodsFor: #optimization !
receiverHasNumber: aNode
	" Private - Return true if the receiver of aNode is a number literal. "

	^aNode receiver hasNumber! !

! OptimizingEncoder methodsFor: #optimization !
receiverHasString: aNode
	" Private - Return true if the receiver of aNode is a string literal. "

	^aNode receiver hasString! !

! OptimizingEncoder methodsFor: #optimization !
receiverIsBlock: aNode
	" Private - Return true if receiver of aNode is block."

	^aNode receiver isKindOf: BlockNode ! !

! OptimizingEncoder methodsFor: #optimization !
receiverIsBlockNoArguments: aNode
	" Private - Return true if receiver of aNode is block with no arguments. "

	^(self receiverIsBlock: aNode) and: [
		aNode receiver parameters isEmpty
	]! !

! OptimizingEncoder methodsFor: #optimization !
removeOptimizationLiteral: aLiteral
	" Private - Remove the optimization from the receiver. "

	self optimizationMap remove: aLiteral ! !

! OptimizingEncoder methodsFor: #optimization !
removeOptimizations: array
	" Private - Remove an array of optimizations to the receiver. "

	array do: [:opt|
		(self optimizationLiteralsFor: opt) do: [:each|
			self removeOptimizationLiteral: each
		]
	]! !

! OptimizingEncoder class methodsFor: #inlines !
initializeInlines
	" Private - Initialize the pool of inlines known by the receiver.
	Note that initialize defaults from methods implementing #initialize*Inlines: in the receiver
	"

	inlines := PoolDictionary new.
	(self class selectors
		select: [:msg| (msg startsWith: #initialize) and: [ msg endsWith: #Inlines: ] ])
		do: [:each| self perform: each with: inlines ].
	^inlines ! !

! OptimizingEncoder class methodsFor: #inlines !
inlines
	" Private - Return the pool of inlines. "

	inlines isNil ifTrue: [ self initializeInlines ].
	^inlines! !

! OptimizingEncoder class methodsFor: #inlines !
register: literals into: pool
	" Private - Register literals array onto pool dictionary.
	literals must be a collection of arrays like:
		#(token inlines)
		#(ClassName #(
			#(selector inlines)
		))
	inlines can be a string or array of strings to be used to encode inlines in source
	token must be lowercase and ClassName must be uppercase names
	token & className can be arrays for multiple registrations
	"

	literals do: [:tuple| | tokens value |
		tuple isArray ifFalse: [ self error: 'Must be array ' ,tuple ].
		tuple size = 2 ifFalse: [ self error: 'Invalid tuple size (must be 2)' ].
		tokens := tuple first. value := tuple last.
		tokens isString ifTrue: [ tokens := Array with: tokens ].
		tokens do: [:token|
			token isString ifFalse: [ self error: 'Must be string ' ,token asLiteral ].
			(token isEmpty or: [ token first isLowercase ]) ifTrue: [ "#(token inlines) "
				pool at: token put: value
			] ifFalse: [ | options | "#(ClassName pairs)"
				options := pool at: token ifAbsentPut: [PoolDictionary new].
				value do: [:pair| "#(selector inlines)"
					pair isArray ifFalse: [ self error: 'Must be array ' ,tuple ].
					pair size = 2 ifFalse: [ self error: 'Invalid pair size (must be 2) - ' ,pair asLiteral ].
					value := pair last.
					(pair first isString ifTrue: [Array with: pair first] ifFalse: [pair first]) do: [:selector|
						selector isString ifFalse: [ self error: 'Must be string ' ,selector asLiteral ].
						options at: selector put: value
					]
				]
			]
		]
	] ! !

! OptimizingEncoder class methodsFor: #inlines !
inlinesFor: token compiler: aCompiler
	" Private - Return the inlines stream with native source chunks to inject while generating code. "

	| result |
	token notNil ifTrue: [
		result := self inlines at: token ifAbsent: [].
		result notNil ifTrue: [ ^result ].
	].
	^self inlines @ (Array
		with: aCompiler currentClass name
		with: aCompiler selector) ! !

! OptimizingEncoder methodsFor: #inlines !
inlineToken: string
	" Private - Return the inline key to search for contents (or nil). "

	^string copyFrom: 2 to: string size ! !

! OptimizingEncoder methodsFor: #inlines !
inlineKey: string
	" Private - Return the inline key to search for inline contents (or nil). "

	string size = 0 ifTrue: [ ^'' ].
	(string startsWith: $#) ifTrue: [ ^self inlineToken: string ].
	^nil ! !

! OptimizingEncoder methodsFor: #inlines !
inlinesFor: string
	" Private - Return the inlines stream with native source chunks to inject while generating code. "

	| chunks |
	inlines notNil ifTrue: [ ^inlines ].
	chunks := self class inlinesFor: (self inlineKey: string) compiler: self compiler.
	chunks isNil ifTrue: [ chunks := #() ].
	chunks isString ifTrue: [ chunks := Array with: chunks ].
	inlines := chunks stream.
	^inlines! !

! OptimizingEncoder methodsFor: #inlines !
missingNativeCodeFor: aNode as: lang
	" Private - Return the native language code for missing expression (or nil). "

	^self error: 'Missing native code for language ' ,lang ! !
	
! OptimizingEncoder methodsFor: #inlines !
inline: aNode source: string
	" Return the string source for inline node. "

	| chunks |
	chunks := self inlinesFor: string.
	^chunks atEnd ifFalse: [ chunks next ] ! !

! OptimizingEncoder methodsFor: #inlines !
visitInlineNode: aNode
	" Private - Encode an inline node containing native language code. "

	| string lang |
	lang := self nativeLanguage.
	lang isNil ifTrue:  [ lang := Smalltalk nativeLanguage ].
	string := aNode sourceFor: lang in: self.
	string = #self ifTrue: [ ^nil ]. "inlined 'self' is encoded a no-operation (optional inline)"
	string size = 0 ifTrue:  [ string := self missingNativeCodeFor: aNode as: lang ].
	^string notNil ifTrue: [ self optimize: aNode macro: string ] ! !

! OptimizingEncoder methodsFor: #private !
startCompiling: aNode
	" Private - Set the receiver to start emmiting code for aNode. "

	inlines := nil.
	^super startCompiling: aNode! !

! OptimizingEncoder methodsFor: #optimization !
visitInlined: aNode
	" Private - Encode aNode as inlined sequence of statements. "

	(aNode isKindOf: BlockNode) ifFalse: [
		^self error: 'Invalid block node (' ,aNode ,').'
	].
	aNode nodes do: [:each | self visit: each ].! !

! OptimizingEncoder methodsFor: #optimizing !
visitInlinedMacro: macro in: input with: aNode
	" Private - Visit inlined macro in input. "

	macro = #inlinedReceiver ifTrue: [ ^self visitInlined: aNode receiver ].
	(macro startsWith: #inlined) ifTrue: [ | args |
		args := aNode arguments.
		1 to: args size do: [:i|
			#inlined ,i asString = macro ifTrue: [
				^self visitInlined: (args at: i)
			].
		].
	].
	^self error: 'Invalid macro [',macro,']' ! !

! OptimizingEncoder methodsFor: #optimizing !
visitSeqMacro: macro in: input with: aNode
	" Private - Visit argument macro in input. "

	| args i |
	macro = #sequence ifTrue: [
		^self visitSequenceNode: aNode
	].
	i := (macro copyFrom: 4 to: macro size) asNumber.
	args := aNode arguments.
	(i between: 1 and: args size) ifTrue: [ | block |
		block := args at: i.
		block parameters size = 0 ifFalse: [
			self error: 'Invalid block (has parameters)'
		].
		^self visitSequenceNode: block nodes first
	].
	^self error: 'Invalid macro [',macro,']' ! !

! OptimizingEncoder methodsFor: #optimizing !
visitArgMacro: macro in: input with: aNode
	" Private - Visit argument macro in input. "

	| args i |
	i := (macro copyFrom: 4 to: macro size) asNumber.
	args := aNode arguments.
	(i between: 1 and: args size) ifTrue: [
		^self visit: (args at: i)
	].
	^self error: 'Invalid macro [',macro,']' ! !

! OptimizingEncoder methodsFor: #macros !
basicMacro: macro in: input with: aNode
	" Private - Visit aNode encoding/expanding basic macros in input. "

	macro = #lang ifTrue: [
		^self nextPutAll: self nativeLanguage
	].
	macro = #receiver ifTrue: [
		^self visitReceiver: aNode receiver
	].
	(macro startsWith: #arg) ifTrue: [
		^self visitArgMacro: macro in: input with: aNode
	].
	(macro startsWith: #seq) ifTrue: [
		^self visitSeqMacro: macro in: input with: aNode
	].
	(macro startsWith: $!!) ifTrue: [
		^macro stream next; fileInto: self
	].
	(macro startsWith: #inlined) ifTrue: [
		^self visitInlinedMacro: macro in: input with: aNode
	].
	(self respondsTo: macro) ifTrue: [ | args count |
		count := macro argumentCount.
		count = 0 ifTrue: [ ^self perform: macro ].
		args := (Array with: aNode with: input)
			copyFrom: 1 to: (count min: 2).
		^self perform: macro withArguments: args
	].
	^self error: 'Missing macro [' ,macro ,$] ! !

! OptimizingEncoder methodsFor: #optimizing !
visitUppercaseMacro: macro in: input with: aNode
	" Private - Expand global macro in input. "

	macro = #Lang ifTrue: [
		^self nextPutAll: self nativeLanguage asGlobalName
	].
	Smalltalk at: macro ifAbsent: [
		^self basicMacro: macro in: input with: aNode 
	].
	^self nextPutAll: (self codeToGetGlobal: macro) ! !

! OptimizingEncoder methodsFor: #optimizing !
visitMacroSend: macro in: input with: aNode
	" Private - Expand root macro in input. "

	^self nextPutAll: #smalltalk ,macro ! !

! OptimizingEncoder methodsFor: #optimizing !
visitMacro: macro in: input with: aNode
	" Private - Visit aNode encoding/expanding macro in input. "

	(macro startsWith: self messageSendConnector) ifTrue: [ " %.msgAPI% "
		^self visitMacroSend: macro in: input with: aNode
	].
	macro first isUppercase ifTrue: [ " %GlobalName% "
		^self visitUppercaseMacro: macro in: input with: aNode 
	].
	^self basicMacro: macro in: input with: aNode ! !

! OptimizingEncoder methodsFor: #optimizing !
visitMacroIn: input with: aNode
	" Private - Visit next macro in input. "

	| macro |
	macro := input upTo: self macroEnd.
	macro isEmpty
		ifTrue: [ self emmitCode: self macroEnd ]
		ifFalse: [ self visitMacro: macro in: input with: aNode ] ! !

! OptimizingEncoder methodsFor: #visiting !
visitOptimizedNode: aNode as: mapping
	" Private - Generate the code for aNode using custom mapping.
	Return true if the code can be generated.
	If the optimization fails, the receiver remains intact.
	"

	| macro |
	macro := mapping last.

	mapping size = 2 ifTrue: [ "(selector macro)"
		^self optimize: aNode macro: macro
	].
	mapping size = 3 ifTrue: ["(selector condition macro)"
		^(self matchOptimizedCondition: (mapping at:2) with: aNode)
			and: [ self optimize: aNode macro: macro ]
	].

	self note: 'Compiler optimization'
		,' for #',aNode selector
		,' - not implemented '
		,mapping asLiteral.
	^false! !

! OptimizingEncoder methodsFor: #visiting !
visitOptimizedSend: aNode
	" Private - Generate optimal code for aNode using optimization map.
	Return true if optimization was successfull.
	"

	^self visitOptimizedSend: aNode with: self optimizationMap! !

! OptimizingEncoder methodsFor: #visiting !
visitOptimizedSend: aNode with: mapping
	" Private - Generate optimal code for aNode using mapping.
	Return true if the code has been generated.
	Leave the receiver intact if the optimization do not apply.
	"

	| argCount |
	argCount := aNode arguments size.
	mapping do: [:tuple|
		(tuple first isNumber and: [ tuple first = argCount ]) ifTrue: [ "arguments size match"
			(self visitOptimizedSend: aNode with: tuple last) ifTrue: [ ^true ].
		] ifFalse: [
			tuple first = aNode selector ifTrue: [ "selector match"
				(self visitOptimizedNode: aNode as: tuple) ifTrue: [ ^true ].
			].
		].
	].
	^false! !

! OptimizingEncoder methodsFor: #visiting !
visitReceiver: aNode
	" Private - Visit the node (as receiver) emmitting output. "

	^self implementedBySubclass: #visitReceiver:! !

! OptimizingEncoder methodsFor: #visiting !
visitSendNode: aNode
	" Private - Visit the node (message send node) emmitting output. "

	^self implementedBySubclass: #visitSendNode:! !

! OptimizingEncoder methodsFor: #visiting !
visitNode: aNode index: index
	" Private - Generate code for indexed access. "

	^self implementedBySubclass: #visitNode:index: ! !

! OptimizingEncoder methodsFor: #visiting !
visitNodeIndices: aNode doing: aBlock
	" Private - Generate code for indexed access. "

	aBlock evaluate.
	aNode indices do: [:each| self visitNode: aNode index: each ].! !

! OptimizingEncoder class methodsFor: #variables !
unknownVariable: aNode in: anEncoder
	" Private - anEncoder has found an unknown name reference.
	The default action will thrown an error (stop compilation),
	 this method can be refined/changed to silently ignore the
	 unknown to let the compiler produce code to refer code
	 in outer native contexts.
	See implementors of #visitVariableNode:
	"

self note: 'Unknown variable ' ,aNode value asString.
	^self error: 'Unknown variable ',aNode value! !

! OptimizingEncoder class methodsFor: #variables !
bind: aName in: anEncoder
	" Private - Return true if aName has been bound to anEncoder.
	Note that the process of binding must register aName onto knownVariables of encoder. "

	"we use underscores as native cookies"
	aName detect: [:c| c ~= #_ ] ifNone: [ "all underscores"
		Smalltalk nativeObjectAt: aName ifAbsent: [
			Smalltalk global basicAt: aName put: nil
		].
		anEncoder knownVariables add: aName.
		^true
	].
	^false! !

! OptimizingEncoder methodsFor: #variables !
missingVariable: aNode
	" Private - Try to bind to access to variable in aNode. "

	self	compilerEvent: #missingVariable:in:
		withArguments: (Array with: aNode with: self)
		ifNotHandled: [ | unknown |
			unknown := aNode value.
			(self class bind: unknown in: self) ifFalse: [
				self unknownVariables add: unknown.
				self class unknownVariable: aNode in: self.
			]
	]! !

! OptimizingEncoder methodsFor: #variables !
codeVisitingLocal: aNode
	" Private - Returns code to access local value. "

	^aNode localNameIn: self! !

! OptimizingEncoder methodsFor: #variables !
codeVisitingVariableNode: aNode
	" Private - Returns code to access variable value. "

	| code |
	code := self currentClass codeFor: self accessing: aNode.
	(code isNil and: [aNode needsVerification]) ifTrue: [
		(self knownVariables includes: aNode value) ifFalse: [
			self missingVariable: aNode
		].
	].
	(code isNil or: [self knownVariables includes: aNode value]) ifTrue: [
		code := self codeVisitingLocal: aNode
	].
	^code! !

! OptimizingEncoder methodsFor: #variables !
visitVariableNode: aNode
	" Private - Generate code to access variable aNode. "

	self visitNodeIndices: aNode doing: [
		self emmitCode: (self codeVisitingVariableNode: aNode)
	]! !

OptimizingEncoder
	subclass: #NativeEncoder
	instanceVariableNames:
		 ' stream indentation startOffset'
		,' nestedBlocks earlyReturn'
		,' functionName '
	category: #Compiler-Encoding !

! NativeEncoder methodsFor: #indentation !
indentation
	" Private - Return the indentation of the receiver. "

	indentation isNil ifTrue: [ indentation := 0 ].
	^indentation ! !

! NativeEncoder methodsFor: #indentation !
indent
	" Private - Enter the next level of indentation. "

	indentation := self indentation + 1! !

! NativeEncoder methodsFor: #indentation !
dedent
	" Private - Exit the current level of indentation. "

	indentation := self indentation - 1 max: 0! !

! NativeEncoder methodsFor: #writing !
cr
	" Private - Emmit a new line onto output stream. "

	stream cr.
	self indentation timesRepeat: [ self tab ]! !

! NativeEncoder methodsFor: #writing !
tab
	" Private - Emmit a tab onto output stream. "

	stream tab! !

! NativeEncoder methodsFor: #writing !
space
	" Private - Emmit a space onto output stream. "

	stream space! !

! NativeEncoder methodsFor: #writing !
nextPutAll: aString
	" Private - Emmit aString onto output stream. "

	stream nextPutAll: aString! !

! NativeEncoder methodsFor: #constants !
assignmentOp
	" Private - Return the OP for encoding assignment. "

	^$= ! !

! NativeEncoder methodsFor: #constants !
separatorOp
	" Private - Return the OP for separating statements. "

	^$; ! !

! NativeEncoder methodsFor: #constants !
returnOpStart
	" Private - Return the OP for starting return expressions. "

	^'return ' ! !

! NativeEncoder methodsFor: #constants !
returnOpEnd
	" Private - Return the OP for ending return expressions. "

	^nil ! !

! NativeEncoder methodsFor: #constants !
returnOps
	" Private - Return the OPs for return expression. "

	^Array	with: self returnOpStart
		with: self returnOpEnd ! !

! NativeEncoder methodsFor: #writing !
between: separators do: aBlock
	" Private - Emmit aBlock between separators onto output stream. "

	separators first notNil ifTrue: [ stream nextPutAll: separators first ].
	aBlock evaluate.
	separators last notNil ifTrue: [ stream nextPutAll: separators last ] ! !

! NativeEncoder methodsFor: #writing !
nextArgumentsPut: aBlock
	" Private - Emmit arguments list evaluating aBlock onto output stream. "

	self between: '()' do: aBlock ! !

! NativeEncoder methodsFor: #writing !
nextIndexPut: aBlock
	" Private - Emmit a series between indexing notation evaluating aBlock onto output stream. "

	self between: '[]' do: aBlock ! !

! NativeEncoder methodsFor: #writing !
nextBlockPut: aBlock
	" Private - Evaluates aBlock on the next level of indentation. "

	self indent;cr.
	aBlock evaluate.
	self dedent;cr ! !

! NativeEncoder methodsFor: #writing !
nextReturn: early put: aBlock
	" Private - Emmit a return OP evaluating aBlock onto output stream. "

	self between: self returnOps do: aBlock ! !

! NativeEncoder methodsFor: #generation !
emmitCode: code
	" Private - Emmit code onto output stream. "

	self nextPutAll: code! !

! NativeEncoder methodsFor: #encoding !
codeToAccessProperty: aName in: aNode
	" Private - Return the source code to access property aName. "

	^'self["',aName,'"]'! !

! NativeEncoder methodsFor: #encoding !
codeToAccessInstanceVariable: aName in: aNode
	" Private - Return the source code to access instance variable aName. "

	^self codeToAccessProperty: '@', aName in: aNode! !

! NativeEncoder methodsFor: #export !
defaultRequirementsFor: aClass
	" Private - Return the default (collection of) requirements to load aClass.
	Return an empty collection.
	This method can be refined by subclasses.
	"

	^#() ! !

! NativeEncoder methodsFor: #assignment !
acceptAssignment: aNode
	" Private - Return true if assignment aNode can be accepted. "

	^self	compilerEvent: #acceptAssignmentTo:
		with: aNode left
		ifNotHandled: [ true ]! !

! NativeEncoder methodsFor: #generation !
emmitFunctionStart: fName
	" Private - Emmit the code of function start definition. "

	self nextPutAll: #function ! !

! NativeEncoder methodsFor: #generation !
emmitFunctionArguments: argList
	" Private - Emmit the code of function arguments. "

	self nextArgumentsPut: [
		self nextPutAll: argList
	] ! !

! NativeEncoder methodsFor: #generation !
emmitFunctionBody: aBlock
	" Private - Emmit the code of function body evaluating aBlock. "

	^self implementedBySubclass: #emmitFunctionBody: ! !

! NativeEncoder methodsFor: #generation !
emmitFunction: fName arguments: argList doing: aBlock
	" Private - Emmit function on receiver's code stream. "

	self
		compilerEvent: #startFunction:with:
		withArguments: (Array with: fName with: argList)
		ifNotHandled: [
			self emmitFunctionStart: fName.
			self compilerEvent: #startFunctionName: with: fName.
			self emmitFunctionArguments: argList
		].
	self emmitFunctionBody: [
		self	compilerEvent: #startFunctionBody:with:
			withArguments: (Array with: fName with: argList).
		aBlock evaluate
	].
	self compilerEvent: #endFunction: with: fName.! !

! NativeEncoder methodsFor: #generation !
emmitFunction: fName node: aNode doing: aBlock
	" Private - Generate function for aNode. "

	^self	emmitFunction: fName
		node: aNode arguments: aNode argumentNames
		doing: aBlock! !

! NativeEncoder methodsFor: #generation !
emmitFunctionDoing: aBlock
	" Private - Emmit function on receiver's code stream. "

	^self emmitFunction: '' arguments: '' doing: aBlock! !

! NativeEncoder methodsFor: #generation !
emmitReturn: aBlockOrString early: early
	" Private - Generate return code. "

	self compilerEvent: #aboutReturn: with: aBlockOrString.
	self nextReturn: early put: [
		self compilerEvent: #return: with: aBlockOrString.
		aBlockOrString isString
			ifTrue: [ self nextPutAll: aBlockOrString ]
			ifFalse: [ aBlockOrString evaluate ]
	].
	aBlockOrString notNil ifTrue: [
		self	compilerEvent: #returned:early:
			withArguments: (Array with: aBlockOrString with: early).
	].! !

! NativeEncoder methodsFor: #generation !
emmitReturn: aBlockOrString
	" Private - Generate return code. "

	^self emmitReturn: aBlockOrString early: false! !

! NativeEncoder methodsFor: #defaults !
defaultOptimizations
	" Private - Return the default optimizations of the receiver.
	Note that returned optimizations MUST be provided by the subclasses. "

	^#( s8 "magnitude math" )! !

! NativeEncoder methodsFor: #results !
emmitedCode
	" Return the emmited code during compilation (or nil). "

	^stream notNil ifTrue: [ stream contents ]! !

! NativeEncoder methodsFor: #literals !
arrayLiteralStart
	" Private - Return the code string to start exporting an array literal. "

	^$[! !

! NativeEncoder methodsFor: #literals !
arrayLiteralEnd
	" Private - Return the code string to end exporting an array literal. "

	^$]! !

! NativeEncoder methodsFor: #literals !
escaped: string
	" Private - Return the code string to export string literal. "

	^string escapedCode! !

! NativeEncoder methodsFor: #private !
nativeSourceFor: aMethod
	" Private - Return the code of native function implementing aMethod. "

	^aMethod compiledSource! !

! NativeEncoder methodsFor: #generation !
emmitMethodContents: aNode
	" Private - Generate code for contents of method aNode. "

	| code delta |
	code := self generatedCodeFor: aNode while: [
		self compilerEvent: #enterMethod: withArguments: (Array with: aNode)
			ifNotHandled: [ self emmitEnterMethod: aNode ].
		aNode nodes do: [:each | self visit: each ].
	].
	delta := self generate: functionName early: earlyReturn return: code.
	self translate: aNode by: delta.! !

! NativeEncoder methodsFor: #generation !
emmitMethod: aNode
	" Private - Generate code of method aNode. "

	nestedBlocks := 0.
	earlyReturn := false.
	self compiler resetVariables.

	functionName := self asSelector: aNode selector.
	aNode validateArguments.
	self emmitFunction: functionName node: aNode doing: [
		self compilerEvent: #startMethod: with: aNode.
		self emmitMethodContents: aNode.
		self compilerEvent: #endMethod: with: aNode.
	]! !

! NativeEncoder methodsFor: #state !
saveState
	" Private - Return the state of the receiver. "

	^Array	with: stream
		with: startOffset
		with: self indentation ! !

! NativeEncoder methodsFor: #state !
restoreState: array
	" Private - Restore the state of the receiver. "

	stream := array first.
	startOffset := array second.
	indentation := array third. ! !

! NativeEncoder methodsFor: #generation !
generatedCodeFor: aNode while: aBlock
	" Private - Return the code generated during evaluation of aBlock. "

	| backup result |
	backup := self saveState.
	startOffset := stream position.
	stream := '' writeStream.
	indentation := 0.
	self visit: aNode doing: aBlock.
	result := stream contents.
	self restoreState: backup.
	^result ! !

! NativeEncoder methodsFor: #generation !
includeSourceCode
	" Private - Return true if source code is included as method argument. "

	^true! !

! NativeEncoder methodsFor: #operations !
isOperation: aNode
	" Private - Return true if aNode is an operation. "

	^aNode arguments size = 1 and: [
		aNode selector size > 1 and: [
		aNode selector first = $# and: [
		'|&\\+*/=><,@%~-' includes: (aNode selector at: 2)
		]]]! !

! NativeEncoder methodsFor: #export !
loadRequirementsOf: aClass
	" Private - Return the requirements to load aClass.
	It must be a collection of requirements.
	aRequirement must implement #fileOutOn: or must be aString(native code) to execute just before defining aClass.
	"

	| requirements |
	requirements := self defaultRequirementsFor: aClass.
	aClass loadRequirementsInto: requirements for: self.
	^requirements! !

! NativeEncoder methodsFor: #accessing !
position
	" Return the current position in output stream. "

	^stream position + startOffset! !

! NativeEncoder methodsFor: #private !
startCompiling: aNode
	" Private - Set the receiver to start emmiting code for aNode. "

	super startCompiling: aNode.
	stream := '' writeStream.
	startOffset := stream position.
	indentation := nil. ! !

! NativeEncoder methodsFor: #accessing !
stream
	" Return the current stream (or nil). "

	^stream! !

! NativeEncoder methodsFor: #visiting !
visit: aNode
	^self visit: aNode doing: [ super visit: aNode ]! !

! NativeEncoder methodsFor: #visiting !
visit: aNode doing: aBlock
	| cookie result |
	cookie := self
		compilerEvent: #needsCookie: with: aNode
		ifNotHandled: [ nil ].
	cookie notNil ifTrue: [ self compilerEvent: #startVisit: with: cookie ].
	result := aBlock evaluate.
	cookie notNil ifTrue: [ self compilerEvent: #endVisit: with: cookie ].
	^result! !

! NativeEncoder methodsFor: #visiting !
visitArgument: aNode
	" Private - Visit the argument emmitting output. "

	aNode isAssignment ifTrue: [ ^self visitLocalAssignment: aNode ].
	^self visit: aNode! !

! NativeEncoder methodsFor: #generation !
emmitAssignment: aNode
	" Private - Emmit the code for assignment. "

	self
		visit: aNode left;
		nextPutAll: self assignmentOp;
		visit: aNode right! !

! NativeEncoder methodsFor: #visiting !
visitAssignmentNode: aNode

	(self visitMacroAssignment: aNode) ifTrue: [ ^self ].
	(self acceptAssignment: aNode) ifTrue: [
		self emmitAssignment: aNode
	] ifFalse: [
		self	compilerEvent: #invalidAssignment: with: aNode
			ifNotHandled: [ self error: 'Invalid assignment to ' ,aNode left value ]
	]! !

! NativeEncoder methodsFor: #visiting !
visitNode: aNode in: aParentNode cr: withCr
	" Private - Visit aNode as part of sequence in aParentNode. "

	(self visit: aNode) notNil ifTrue: [
		self nextPutAll: self separatorOp.
		withCr ifTrue: [ self cr ].
	]! !

! NativeEncoder methodsFor: #visiting !
visitEmptyNode: aNode
	" Private - Visit an empty node, aNode. " ! !

! NativeEncoder methodsFor: #visiting !
visitNodesIn: aNode cr: withCr
	" Private - Visit the nodes contained in aNode. "

	| array |
	array := aNode nodes.
	array isEmpty ifTrue: [ ^self visitEmptyNode: aNode ].
	array do: [:one|
		self visitNode: one in: aNode cr: false
	] separatedBy: [ withCr ifTrue: [ self cr ] ]! !

! NativeEncoder methodsFor: #optimizing !
visitArgMacro: macro in: input with: aNode
	" Private - Visit argument macro in input. "

	macro = #argList ifTrue: [
		^self visitArguments: aNode arguments
	].
	^super visitArgMacro: macro in: input with: aNode! !

! NativeEncoder methodsFor: #encoding !
defineTemporary: aName
	" Private - Return the code for defining temporary variable with aName. "

	self nextPutAll: 'var ' ,(self localName: aName) ,'=nil;' ;cr ! !

! NativeEncoder methodsFor: #visiting !
withNodesIn: aNode do: aBlock atEnd: endBlock
	" Evaluates aBlock for nodes in aNode (without last).
	Return the result of evaluating lastBlock with the last node in aNode (or nil).
	"

	| array |
	array := aNode nodes asArray.
	1 to: array size - 1 do: [:index|
		aBlock evaluateWith: (array at: index)
	].
	^endBlock evaluateWith: (array notEmpty ifTrue: [ array last ]) ! !

! NativeEncoder methodsFor: #visiting !
visitTemporaries: aNode
	" Private - Emmit code for local variables in aNode. "

	aNode tempNames do: [:aName |
		self tempVariables add: aName.
		self defineTemporary: aName
	] ! !

! NativeEncoder methodsFor: #contexts !
startSequence: aNode
	" Private - Prepare context for sequence aNode. " ! !

! NativeEncoder methodsFor: #contexts !
endSequence: aNode
	" Private - Leave context for sequence aNode. " ! !

! NativeEncoder methodsFor: #contexts !
sequenceSpace: aNode
	" Private - aNode has been visited in a sequence. " ! !

! NativeEncoder methodsFor: #contexts !
sequenceExit: aNode
	" Private - A block sequence is exiting with aNode. "

	self emmitReturn: [ self visit: aNode ] ! !

! NativeEncoder methodsFor: #visiting !
visitBlockSequenceNode: aNode
	" Private - Visit the block sequence emmitting output. "

	nestedBlocks := nestedBlocks + 1.
	aNode nodes isEmpty
	ifTrue: [ self emmitReturn: #nil ]
	ifFalse: [
		self	startSequence: aNode;
			visitTemporaries: aNode;
			withNodesIn: aNode do: [:each|
				self visitNode: each in: aNode cr: false.
				self sequenceSpace: each
			] atEnd: [:last| self sequenceExit: last ];
			endSequence: aNode.
	].
	nestedBlocks := nestedBlocks - 1 ! !

! NativeEncoder methodsFor: #visiting !
visitSequenceNode: aNode
	" Private - Visit the sequence emmitting output. "

	self	startSequence: aNode;
		visitTemporaries: aNode;
		visitNodesIn: aNode cr: true;
		endSequence: aNode. ! !

! NativeEncoder methodsFor: #visiting !
visitExtendedSequenceNode: aNode
	" Private - Visit an extended sequence node. "

	aNode directives do: [:each|
		self visitDirectiveNode: each
	].
	^self visitSequenceNode: aNode! !

! NativeEncoder methodsFor: #visiting !
visitMacroAssignment: aNode
	^aNode left acceptMacroAssignment: aNode right with: self! !

! NativeEncoder methodsFor: #visiting !
visitMacroAssignment: aVariableNode to: anExpressionNode
	" Return true if the receiver generated code to a macro assignment to aVariableNode. "

	| aName aClass |
	aVariableNode indices notEmpty ifTrue: [ ^false ].
	aName := aVariableNode value.
	aClass := self currentClass instanceClass.
	aClass classVariableNamesDo: [:each|
		each = aName ifTrue: [
			self visitClassVariableAssignment: each
				in: aClass
				to: anExpressionNode.
			^true
		]
	].
	aClass sharedPoolsAndNamesDo: [:poolName :pool|
		(pool includesKey: aName) ifTrue: [
			self visitPoolVariableAssignment: aName
				in: poolName
				to: anExpressionNode.
			^true
		].
	].
	^false! !

! NativeEncoder methodsFor: #visiting !
visitReceiver: aNode
	" Private - Visit the node (as receiver) emmitting output. "

	aNode isAssignment ifTrue: [ ^self visitLocalAssignment: aNode. ].
	^self visit: aNode! !

! NativeEncoder methodsFor: #visiting !
visitLocalAssignment: aNode
	" Private - Visit a node that is an assignment in a local expression. "

	self nextArgumentsPut: [ self visit: aNode ] ! !

! NativeEncoder methodsFor: #visiting !
visitArguments: anArray

	anArray
	    do: [:each | self visitArgument: each ]
	    separatedBy: [ self nextPutAll: self argumentsSeparatorOp ] ! !

! NativeEncoder methodsFor: #visiting !
visitSendArguments: aSendNode

	self visitArguments: aSendNode arguments ! !

! NativeEncoder methodsFor: #visiting !
visitSendNode: aNode
	" Private - Visit the node (message send node) emmitting output. "

	aNode receiver isSuper ifTrue: [ ^self visitSendToSuper: aNode ].
	(self visitOptimizedSend: aNode) ifTrue: [ ^self ].
	(self isOperation: aNode) ifTrue: [ ^self visitOperation: aNode ].
	^self visitNormalSend: aNode! !

! NativeEncoder methodsFor: #visiting !
visitValueNode: aNode
	" Private - Visit the node emmitting output. "

	self nextPutAll: (aNode value asNative: self nativeLanguage)! !

! NativeEncoder methodsFor: #generation !
emmitEnterMethod: aMethodNode
	" Private - Generate home context definitions, entering aMethodNode. "

	^self implementedBySubclass: #emmitEnterMethod: ! !

! NativeEncoder methodsFor: #generation !
emmitFunction: fName node: aNode arguments: argumentNames doing: aBlock
	" Private - Generate function for aNode. "

	^self implementedBySubclass: #emmitFunction:node:arguments:doing: ! !

! NativeEncoder methodsFor: #generation !
generate: fName early: early return: code
	" Private - Generate code to catch early return of code expression in fName. "

	^self implementedBySubclass: #generate:early:return: ! !

! NativeEncoder methodsFor: #visiting !
visitBlockNode: aNode

	^self implementedBySubclass: #visitBlockNode: ! !

! NativeEncoder methodsFor: #visiting !
visitReturnNode: aNode

	^self implementedBySubclass: #visitReturnNode: ! !

! NativeEncoder methodsFor: #visiting !
visitCascadeNode: aNode

	^self implementedBySubclass: #visitCascadeNode: ! !

! NativeEncoder methodsFor: #visiting !
visitClassVariableAssignment: aClassVariableName in: aClass to: anExpressionNode

	^self implementedBySubclass: #visitClassVariableAssignment:in:to: ! !

! NativeEncoder methodsFor: #visiting !
visitPoolVariableAssignment: aName in: poolName to: anExpressionNode

	^self implementedBySubclass: #visitPoolVariableAssignment:in:to: ! !

! NativeEncoder methodsFor: #visiting !
visitDirectiveNode: aNode
	" Private - Visit a directive node. "

	^self implementedBySubclass: #visitDirectiveNode: ! !

! NativeEncoder methodsFor: #visiting !
visitMethodNode: aMethodNode
	" Visit a method node. "

	^self implementedBySubclass: #visitMethodNode: ! !

! NativeEncoder methodsFor: #visiting !
visitOperation: aNode
	" Private - aNode is an operation and selector is mapped to target operation. "

	^self implementedBySubclass: #visitOperation: ! !

! NativeEncoder methodsFor: #query !
isNativeCall: aNode
	" Private - Return true if aNode is a send calling to native function. "

	^aNode selector startsWith: $# ! !

! NativeEncoder methodsFor: #encoding !
encodeNormalSend: aNode
	" Private - Generate code for direct/inlined message send."

	^self implementedBySubclass: #encodeNormalSend: ! !

! NativeEncoder methodsFor: #encoding !
encodeNormalSend: aNode
	" Private - Generate code for direct/inlined message send."

	^self implementedBySubclass: #encodeNormalSend: ! !

! NativeEncoder methodsFor: #encoding !
encodeSuperSend: aNode
	" Private - Generate code for super message send."

	^self implementedBySubclass: #encodeSuperSend: ! !

! NativeEncoder methodsFor: #visiting !
visitNormalSend: aNode
	" Visit direct/inlined message send."

	self compilerEvent: #aboutSend: with: aNode.
	self encodeNormalSend: aNode.
	self compilerEvent: #send: with: aNode. ! !

! NativeEncoder methodsFor: #visiting !
visitSendToSuper: aNode
	" Generate code for super message send."

	self compilerEvent: #aboutSend: with: aNode.
	self encodeSuperSend: aNode.
	self compilerEvent: #send: with: aNode. ! !

! NativeEncoder methodsFor: #debug !
currentMethodInfo
	" Private - Returns the debug info for current method. "

	^self compiler currentClass name 
		,'>>#' ,self compiler selector! !


EventManager
	subclass: #Compiler
	instanceVariableNames: ' currentClass selector encoder unknownVariables tempVariables transformers useTemplates '
	category: #Compiler !

! Compiler class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^#(	#needsCookie: #startVisit: #endVisit:
		#startMethod: #enterMethod: #endMethod:
		#startFunction:with: #startFunctionName: #startFunctionBody:with: #endFunction:
		#aboutReturn: #return: #returned:early:
		#aboutSend: #send:
		#aboutOperation: #operation:
		#aboutCascade: #cascade:
		#aboutToCompile: #setTransformers:
		#needsParser #aboutToParse:
		#needsEncoder #aboutToEncode:
		#aboutToRecompile: #recompiled:
		#startCompiling: #endCompiling:
		#needsOptimizationMap #initializeOptimizations #customizeOptimizations
		#unknownVariable: #acceptAssignmentTo: #invalidAssignment:
		#move:by:
	) ,super constructEventsTriggered! !

! Compiler class methodsFor: #private !
debugCookieFrom: aStream
	" Private - Return the next debug cookie encoded in aStream (or nil). "

	| nextOffset |
	nextOffset := [ aStream nextChunk asNumber ].
	^Array	"ref. #debugInfo:forClass: "
		with: aStream nextChunk
		with: nextOffset value @ nextOffset value
		with: nextOffset value @ nextOffset value
		with: aStream nextChunk asArrayOfSubstrings! !

! Compiler class methodsFor: #private !
debugCookiesFrom: aStream
	" Private - Return the debug cookies encoded in aStream. "

	| result cookie |
	result := Array new.
	[ aStream skipSeparators; atEnd ] whileFalse: [
		cookie := self debugCookieFrom: aStream.
		cookie notNil ifTrue: [ result add: cookie ].
	].
	^result! !

! Compiler class methodsFor: #debug !
debugInfo: aMethod
	" Return the debug information of aMethod. "

	^aMethod basicAt: '$debugInfo' ifAbsentPut: [
		self debugInfo: aMethod source forClass: aMethod classField
	]! !

! Compiler class methodsFor: #debug !
debugInfo: aString forClass: aClass
	" Private - Return aString with information of compiling aString as a method implemented in aClass. "

	| compiler output cookies aNode delta |
	output := '' writeStream.
	cookies := #(). delta := 0.
	compiler := self new.
	compiler
		useTemplates: false;
		when: #needsCookie: do: [:one| cookies add: (Array with: one with: compiler encoder position) ];
		when: #endVisit: do: [:cookie| cookie add: compiler encoder position ];
		when: #move:by: do: [:one :distance| delta := delta + distance ].
	output
		nextChunkPut: aClass name;
		nextChunkPut: aString;
		nextChunkPut: (compiler compile: aString forClass: aClass);
		yourself.
	cookies do: [:tuple|
		aNode := tuple first.
		output
			nextChunkPut: aNode class name;
			nextChunkPut: aNode start asString;
			nextChunkPut: aNode end asString;
			nextChunkPut: (tuple second + delta) asString;
			nextChunkPut: (tuple third + delta) asString;
			yourself.
		aNode bindings do: [:each| output nextPutAll: ' ' ,each ].
		output nextChunkPut: ' '.
	].
	^output contents! !

! Compiler class methodsFor: #debug !
debugInfoFrom: aString
	" Private - Return the debug information decoding aString contents.
	Note: aString must be result of sending #getInfo: to the receiver.
	"

	^self debugInfoFromStream: aString stream! !

! Compiler class methodsFor: #private !
debugInfoFromStream: aStream
	" Private - Return the debug information encoded in aStream. "

	^Array	"ref. #debugInfo:forClass: "
		with: aStream nextChunk
		with: aStream nextChunk
		with: aStream nextChunk
		with: (self debugCookiesFrom: aStream)! !

! Compiler class methodsFor: #evaluation !
doIt: smalltalkExpression
	" Return the result of evaluating the smalltalk expression in the global context or a description of an error situation occurred evaluating the code. "

	^self doIt: smalltalkExpression to: nil! !

! Compiler class methodsFor: #evaluation !
doItOBSOLETE: smalltalkExpression to: aReceiver
	" Return the result of evaluating the smalltalk expression in the context of aReceiver or a description of an error situation occurred evaluating the code. "

	| result |
	[ result := aReceiver isNil
			ifTrue:  [ self evaluate: smalltalkExpression ]
			ifFalse: [ self evaluate: smalltalkExpression to: aReceiver ]
	] on: Exception do: [:error|
		result := error.
		((self isObject: error) and: [ error isString not ])
		ifTrue: [ result := error class name ,': ' ,error messageText ].
		self note: result
	].
	^result! !

! Compiler class methodsFor: #evaluation !
doIt: smalltalkExpression to: aReceiver
	" Return the result of evaluating the smalltalk expression in the context of aReceiver or a description of an error situation occurred evaluating the code. "

	| result |
	[ result := aReceiver isNil
			ifTrue:  [ self evaluate: smalltalkExpression ]
			ifFalse: [ self evaluate: smalltalkExpression to: aReceiver ]
	] on: Exception do: [:error|
		result := error.
		((self isObject: error) and: [ error respondsTo: #description ])
		ifTrue: [ result := error description ].
		self note: result
	].
	^result! !

! Compiler class methodsFor: #evaluation !
evaluate: smalltalkExpression
	" Return the result of evaluating the smalltalk expression in a global context. "

	^self new evaluate: smalltalkExpression! !

! Compiler class methodsFor: #evaluation !
evaluate: expression in: aClass to: aReceiver
	" Return the result of evaluating the expression with aReceiver. "

	^aClass compiler evaluate: expression in: aClass to: aReceiver! !

! Compiler class methodsFor: #evaluation !
targetFor: aReceiver
	" Return the object to use for compiling expressions to be evaluated in aReceiver. "

	^(self isObject: aReceiver)
		ifTrue: [ aReceiver ]
		ifFalse: [ NativeHandle support @ aReceiver ] ! !

! Compiler class methodsFor: #evaluation !
evaluate: expression to: aReceiver
	" Return the result of evaluating the expression with aReceiver. "

	| target |
	target := self targetFor: aReceiver.
	^self evaluate: expression in: target class to: target! !

! Compiler class methodsFor: #instantiation !
for: aBehavior
	" Return an instance of the receiver. "

	^self new
		currentClass: aBehavior;
		yourself! !

! Compiler class methodsFor: #debug !
getInfo: aString
	" Private - Return the information of compiling aString as a chunked spec. "

	| stream aClass selector source aMethod |
	stream := aString stream.
	aClass := stream nextChunk doIt.
	selector := stream nextChunk.
	source := stream nextChunk.
	aMethod := aClass >> selector.
	(aMethod notNil and: [ source trimBlanks isEmpty or: [
		aMethod source = source ]]) ifTrue: [
		^self debugInfo: aMethod
	].
	^self debugInfo: source forClass: aClass! !

! Compiler methodsFor: #parsing !
parseMethod: aString
	" Return the result of parsing aString as a method source code. "

	| result |
	self	triggerEvent: #aboutToParse:
		with: aString ifNotHandled: [
			result := self nodeFromTemplateIn: aString.
			result notNil ifTrue: [ ^result ]
		].
	result := self parser parse: aString readStream.
	^result! !

! Compiler methodsFor: #parsing !
parse: aString
	" OBSOLETE - Return the result of parsing aString. "

	self error: 'OBSOLETE message #parse: ' ,aString printString.
	^self parseMethod: aString ! !

! Compiler methodsFor: #private !
compile: methodSource
	" Return the native code result of compiling methodSource for currentClass. "

	| aNode |
	aNode := self parseMethod: methodSource.
	self triggerEvent: #aboutToCompile: with: aNode.
	^self emmitNode: aNode! !

! Compiler methodsFor: #compiling !
compile: methodSource forClass: aClass
	" Return the native code result of compiling methodSource as a method implemented by aClass. "

	self currentClass: aClass.
	^self compile: methodSource ! !

! Compiler methodsFor: #compiling !
compile: methodSource forClass: aClass category: category
	" Return the compiled method result of installing a new method compiled from methodSource in aClass. "

	| method |
	(methodSource isNil or: [ methodSource isEmpty ]) ifTrue: [ ^nil ].
	method := self load: methodSource forClass: aClass.
	method category: category.
	aClass addCompiledMethod: method.
	^method! !

! Compiler methodsFor: #accessing !
selector
	" Return the selector of the receiver (or nil). "

	^selector ! !

! Compiler methodsFor: #accessing !
currentClass
	" Return the current class of the receiver. "

	^currentClass! !

! Compiler methodsFor: #accessing !
currentClass: aClass
	" Set the current class of the receiver. "

	currentClass := aClass! !

! Compiler methodsFor: #private !
emmitNode: aNode
	" Private - Return the native code for aNode. "

	self startCompiling: aNode.
	self prepare: aNode.
	self triggerEvent: #aboutToEncode: with: aNode.
	self encoder visit: aNode.
	self endCompiling: aNode.
	^self encoder emmitedCode! !

! Compiler methodsFor: #accessing !
encoder
	" Return the encoder of the receiver. "

	encoder isNil ifTrue: [ self initializeEncoder ].
	^encoder! !

! Compiler methodsFor: #accessing !
encoder: anEncoder
	" Set the encoder of the receiver. "

	encoder := anEncoder! !

! Compiler methodsFor: #private !
endCompiling: aMethodNode
	" Private - Finish emmiting code for aMethodNode. "

	self encoder endCompiling: aMethodNode.
	self triggerEvent: #endCompiling: with: aMethodNode.! !

! Compiler methodsFor: #evaluation !
eval: nativeExpression
	" Return the result of evaluating a native expression in global context. "

	^Smalltalk current eval: nativeExpression! !

! Compiler methodsFor: #evaluation !
eval: nativeExpression as: aState
	" Return the result of evaluating a native expression in global context.
	The state of the system change to state while evaluating the expression.
	"

	^Smalltalk current eval: nativeExpression as: aState! !

! Compiler methodsFor: #doIt !
evaluate: smalltalkExpression
	" Return the result of evaluating the smalltalk expression in global context. "

	^self evaluate: smalltalkExpression in: nil class to: nil! !

! Compiler methodsFor: #evaluation !
compiledMethodToEvaluate: smalltalkExpression in: aClass
	" Return a compiled method to evaluate the smalltalk expression in aClass. "

	| result doIt node code |
	doIt := String doItSelector.
	self currentClass: aClass.
	node := self parseMethod: doIt,' ^[', smalltalkExpression, '] value'.
	code := self emmitNode: node.
	result := self eval: code.
	result category: doIt.
	aClass addCompiledMethod: result.
	^result! !

! Compiler methodsFor: #evaluation !
evaluate: smalltalkExpression in: aClass to: aReceiver
	" Return the result of evaluating the smalltalk expression with aReceiver. "

	| result cm |
	cm := self compiledMethodToEvaluate: smalltalkExpression in: aClass.
	[ result := aReceiver perform: cm selector.
	] ensure: [ cm removeFromSystem ].
	^result! !

! Compiler methodsFor: #initialize !
initialize
	" Private - Initialize the receiver. "

	super initialize.
	self resetVariables.! !

! Compiler methodsFor: #private !
initializeEncoder
	" Private - Initialize the encoder of the receiver. "

	encoder := self
		triggerEvent: #needsEncoder
		ifNotHandled: [ S8Encoder for: self ]! !

! Compiler methodsFor: #variables !
knownVariables
	" Return the knownVariables of the receiver. "

	^self pseudoVariables , self tempVariables! !

! Compiler methodsFor: #compiling !
load: aString forClass: aClass
	" Return the CompiledMethod result of compiling aString as a method implemented in aClass. "

	| nativeCode |
	nativeCode := self compile: aString forClass: aClass.
	^self eval: nativeCode! !

! Compiler methodsFor: #private !
localName: aName
	" Private - Return a valid (native, local) name for aName. "

	^self encoder localName: aName! !

! Compiler methodsFor: #private !
alienName: aName
	" Private - Return a valid native name for aName. "

	^self encoder alienName: aName! !

! Compiler methodsFor: #accessing !
nativeLanguage

	^self encoder nativeLanguage! !

! Compiler methodsFor: #templates !
nodeFromTemplateIn: aString
	" Return the nodes syntetized from inlined templates in aString (or nil).
	Inlined templates are detected from last comment in source.
	"

	| src driver msg count found |
	(aString isEmpty or: [(aString last = $") not]) ifTrue: [ ^nil ].
	src := ((aString upToLast: $") fromLast: $") stream.
	src peek = $# ifTrue: [ src next ] ifFalse: [
		#template = (src nextLine upTo: $ ) ifFalse: [ ^nil ]
	].
	driver := src upTo: $ .
	driver = #self  ifTrue: [ driver := self ].
	driver = #class ifTrue: [ driver := self currentClass ].
	driver = #encoder ifTrue: [ driver := self encoder ].
	driver = #api   ifTrue: [ driver := #APIBuilder ].
	driver isString ifTrue: [ driver := Smalltalk at: driver ifAbsent: [] ].
	msg := src upTo: $ .
	(driver respondsTo: msg) ifFalse: [ ^nil ].
	count := msg occurrencesOf: $:.
	found := [:mth| mth notNil ifTrue: [ mth source: aString; yourself ] ].
	count = 0 ifTrue: [ ^found value: (driver perform: msg) ].
	count = 1 ifTrue: [ ^found value: (driver perform: msg with: src upToEnd) ].
	^nil! !

! Compiler methodsFor: #private !
parser
	" Private - Return the parser of the receiver. "

	^self	triggerEvent: #needsParser
		ifNotHandled: [ SmalltalkParser new ]! !

! Compiler methodsFor: #prepare !
prepare: aMethodNode
	" Private - Apply tranformations/optimization to aMethodNode. "

	self encoder prepare: aMethodNode.
	self transformers do: [:each| each visit: aMethodNode ].! !

! Compiler methodsFor: #variables !
pseudoVariables
	" Return the pseudoVariables of the receiver. "

	^#( #self #super #true #false #nil #thisContext )! !

! Compiler methodsFor: #compiling !
recompile: aClass
	" Re-compile all the methods defined by aClass. "

	self triggerEvent: #aboutToRecompile: with: aClass.
	aClass methodDictionary do: [:each |
"Debug		self note: each."
		self compile: each source forClass: aClass category: each category
	].
	self triggerEvent: #recompiled: with: aClass.
	aClass isMetaclass ifFalse: [ self recompile: aClass class ].! !

! Compiler methodsFor: #variables !
resetVariables
	" Reset the variables of the receiver. "

	unknownVariables := #().
	tempVariables := #().! !

! Compiler methodsFor: #names !
smalltalkNameFor: aClass
	" Return the global name of aClass (can be nil). "

	^self encoder smalltalkNameFor: aClass! !

! Compiler methodsFor: #private !
startCompiling: aNode
	" Private - Set the receiver to start emmiting code for aNode. "

	aNode isParseFailure ifFalse: [
		selector := aNode selector
	].
	self triggerEvent: #startCompiling: with: aNode.
	self encoder startCompiling: aNode.! !

! Compiler methodsFor: #variables !
tempVariables
	" Return the tempVariables of the receiver. "

	^tempVariables! !

! Compiler methodsFor: #accessing !
transformers
	" Return the node transformers installed in the receiver. "

	transformers isNil ifTrue: [
		transformers := Array new.
		self triggerEvent: #setTransformers: with: transformers.
	].
	^transformers! !

! Compiler methodsFor: #variables !
unknownVariables
	" Return the unknownVariables of the receiver. "

	^unknownVariables! !

! Compiler methodsFor: #accessing !
useTemplates
	" Return true if the receiver use templates. "

	^useTemplates notNil and: [ useTemplates ]! !

! Compiler methodsFor: #accessing !
useTemplates: aBoolean
	" Set the receiver to consider inlined templates. "

	useTemplates := aBoolean! !


Object
 subclass: #PPFailure
 instanceVariableNames: 'position reason'
 category: #Compiler-Parsing!

! PPFailure methodsFor: #accessing !
position
	" Return the position of the receiver. "

	position isNil ifTrue: [ ^0 ].
	^position! !

! PPFailure methodsFor: #accessing !
position: aNumber
	" Set the position of the receiver. "

	position := aNumber! !

! PPFailure methodsFor: #printing !
toString
	" Return the receiver converted to String. "

	^'Failure ',self reason printString ,'(@',self position,')'! !

! PPFailure methodsFor: #accessing !
reason
	" Return the reason of the receiver. "

	reason isNil ifTrue: [ ^String new ].
	^reason! !

! PPFailure methodsFor: #accessing !
reason: aString
	" Set the reason of the receiver. "

	reason := '' ,aString! !

! PPFailure methodsFor: #accessing !
reason: aString at: anInteger
	" Set the reason and position of the receiver. "

	self
		reason: aString; 
		position: anInteger! !

! PPFailure class methodsFor: #instantiation !
reason: aString at: anInteger
	" Return an instance of the receiver. "

	^self new reason: aString at: anInteger! !

! Object methodsFor: 'Compiler-testing' !
isParseFailure
	" Return true if the receiver is a parse failure. "

	^false! !

! PPFailure methodsFor: #testing !
isParseFailure
	" Return true if the receiver is a parse failure. "

	^true! !

!PPFailure methodsFor: #bug!
accept: anEncoder
	" Report compilation failure and error. "

	| description |
	description := 'PPFailure: ', self reason
		,' at: ' ,self position asString
		,' source: ' ,anEncoder stream contents.
	self print: description.
	self error: description! !

Object
 subclass: #PPParser
 instanceVariableNames: #memo
 category: #Compiler-Parsing!

! PPParser class methodsFor: #instantiation !
, anObject
	" Return the receiver composed with the argument. "

	^anObject asParser! !

! PPParser class methodsFor: #instantiation !
optional: anObject
	" Return an instance of the receiver to parse an optional element anObject. "

	^(self ,anObject) optional! !

! PPParser class methodsFor: #debug !
debug: action or: aBlock
	" Log evaluation of action if in debugging mode,
	 else returns the result of evaluation of aBlock.
	"

	| show result |
	show := PPParser basicAt: #debugOn ifAbsent:[ result := aBlock evaluate. false ].
	show = true ifTrue: [ result := action evaluate ].
	^result! !

! PPParser class methodsFor: #debug !
debugWhile: aBlock
	^[	PPParser basicAt: #debugOn put: true.
		aBlock evaluate
	] ensure: [
		PPParser basicAt: #debugOn put: false
	]! !

! PPParser methodsFor: #private !
set: anElement start: start end: end
	" Private - Set start&end source offest of anElement. "

	(anElement respondsTo: #start:end:) ifTrue: [
		anElement start: start end: end
	]! !

! PPParser methodsFor: #operations !
flatten
	" Return a flattened version of the receiver. "

	^PPFlattenParser on: self! !

! PPParser methodsFor: #operations !
withSource
	" Return a source parser on the receiver. "

	^PPSourceParser on: self! !

! PPParser methodsFor: #operations !
==> aBlock
	" Return an action parser on the receiver. "

	^PPActionParser on: self block: aBlock! !

! PPParser methodsFor: #operations !
, aParser
	" Return the receiver composed with the argument. "

	^PPSequenceParser with: self with: aParser! !

! PPParser methodsFor: #operations !
/ aParser
	" Return a choice composite version of the receiver and aParser. "

	^PPChoiceParser with: self with: aParser! !

! PPParser methodsFor: #debug !
@ tag
	" DEBUG - Return a notification wrapper of the receiver. "

	self basicAt: #debugTag put: tag.
	^self ==> [:node| self note: tag ,'==>',node printString. node ]! !

! PPParser methodsFor: #debug !
@@ tag
	" DEBUG - Tag the receiver for debugging. "

	self basicAt: #debugTag put: tag! !

! PPParser methodsFor: #operations !
plus
	" Return an iterative version of the receiver. "

	^(PPRepeatingParser on: self min: 1) @@#plus! !

! PPParser methodsFor: #operations !
star
	" Return an iterative version of the receiver. "

	^(PPRepeatingParser on: self min: 0) @@#star! !

! PPParser methodsFor: #operations !
not
	" Return the negation of the receiver. "

	^(PPNotParser on: self) @@#not! !

! PPParser methodsFor: #operations !
optional
	" Return an optional version of the receiver. "

	^self / PPEpsilonParser new @@#opt! !

! PPParser methodsFor: #test !
isParseFailure: what
	" Private - Return true if what is a parsing failure. "

	^what isParseFailure! !

! PPParser methodsFor: #memo !
valueToMemoize: aStream start: start
	" Private - Return the value to memoize as result of parsing en element in aStream. "

	| end node |
	node := self parse: aStream.
	end := aStream position.
	self set: node start: start end: end.
	^Array with: node with: end! !

! PPParser methodsFor: #memo !
memoizedParse: aStream
	" Private - Evaluate the receiver. "

	| start value |
	memo isNil ifTrue: [ memo := PoolDictionary new ].
	start := aStream position.
	value := memo at: start ifAbsentPut: [
		self valueToMemoize: aStream start: start
	].
	aStream position: value second.
	^value first! !

! PPParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	self subclassResponsibility: #basicParse:! !

! PPParser methodsFor: #parsing !
parseLog: aStream
	" Parse contents in aStream. "

	| result old |
	old := aStream position.
	result := self basicParse: aStream.
	(self isParseFailure: result) ifFalse: [
		self note: $[ ,old ,$- ,aStream position ,'] ' ,self printString
	].
	^result! !

! PPParser methodsFor: #parsing !
parse: aStream
	" Parse contents in aStream. "

	^PPParser
		debug: [self parseLog: aStream]
		or: [self basicParse: aStream]! !

! PPParser methodsFor: #parsing !
parseAll: aStream
	" Parse contents in aStream. "

	| result |
	result := (PPSequenceParser
		with: self with: PPEOFParser new)
		memoizedParse: aStream.
	^(self isParseFailure: result)
	    ifTrue: [ self error: (result messageFor: aStream contents) ]
	    ifFalse: [ result first ]! !

! PPParser methodsFor: #converting !
asParser
	" Return a parser on the receiver. "

	^self! !

! PPParser methodsFor: #printing !
printStringDetail
	" Private - Return a printable detail of the receiver. "

	^''! !

! PPParser methodsFor: #printing !
printString
	" Return a printable representation of the receiver. "

	| tag |
	tag := self basicAt: #debugTag.
	tag notNil ifTrue: [ ^#@@ ,tag ,#@@ ]. "compact notation"

	^#a ,self class name ,(tag isNil
		ifTrue: [ self printStringDetail ]
		ifFalse: [ '@@',tag ])! !

PPParser
 subclass: #PPEOFParser
 category: #Compiler-Parsing!

! PPEOFParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	aStream atEnd ifTrue: [ ^nil ] .
	^PPFailure reason: 'EOF expected (or invalid character in input)' at: aStream position! !

PPParser
 subclass: #PPAnyParser
 category: #Compiler-Parsing!

! PPAnyParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	aStream atEnd ifFalse: [ ^aStream next ].
	^PPFailure reason: 'End of stream' at: aStream position! !

PPParser
 subclass: #PPEpsilonParser
 category: #Compiler-Parsing!

! PPEpsilonParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	^nil! !

PPParser
 subclass: #PPStringParser
 instanceVariableNames: #string
 category: #Compiler-Parsing!

! PPStringParser methodsFor: #accessing !
string
	" Return the string of the receiver. "

	^string! !

! PPStringParser methodsFor: #accessing !
string: aString
	" Set the string of the receiver. "

	string := aString! !

! PPStringParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	| back reason |
	string isEmpty ifTrue: [ ^string ].
	back := aStream position.
	(aStream peekForAll: string) ifTrue: [ ^string ].
	reason := 'Expecting ', string printString
		,' at ',back printString
		,' read ',(aStream next: string size) printString.
	aStream position: back.
	^PPFailure reason: reason at: back! !

! PPStringParser methodsFor: #printing !
printStringDetail
	" Private - Return a printable detail of the receiver. "

	^string printString! !

PPParser
 subclass: #PPStStringParser
 category: #Compiler-Parsing!

! PPStStringParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	| position result |
	position := aStream position.
	result := ''.
	[ result := result , (aStream upTo: $').
	  aStream atEnd ] whileFalse: [
		aStream peek = $' ifFalse: [ ^result ].
		result := result ,aStream next.
	].
	aStream position > position ifTrue: [
		aStream position: aStream position - 1.
		aStream next = $' ifTrue: [ ^result ].
	].
	aStream position: position.
	^PPFailure reason: 'Unfinished String' at: position! !

PPParser
 subclass: #PPCharacterParser
 instanceVariableNames: 'regexp theString'
 classVariableNames: #RegExpCache
 category: #Compiler-Parsing!

! PPCharacterParser class methodsFor: #constants !
separator
	" Return the parser for separator. "

	^self new separator! !

! PPCharacterParser class methodsFor: #constants !
in: aCollection
	" Return the parser for input matching any element in aCollection. "

	| tag |
	tag := aCollection asString.
	aCollection size = 1 ifTrue: [
		tag := aCollection first.
		^self for: tag asString doing: [:c| tag = c ]
	].
	^self for: tag doing: [:c| aCollection includes: c ]! !

! PPCharacterParser class methodsFor: #constants !
notIn: aCollection
	" Return the parser for input matching any element in aCollection. "

	| tag |
	tag := aCollection asString.
	aCollection size = 1 ifTrue: [
		tag := aCollection first.
		^self for: tag asString doing: [:c| tag ~= c ]
	].
	^self for: tag doing: [:c| (aCollection includes: c) not ]! !

! PPCharacterParser class methodsFor: #constants !
between: start and: end
	" Return the parser for input between start and end. "

	^self	for: '',start,'-',end
		doing: [:c| c between: start and: end ]! !

! PPCharacterParser class methodsFor: #constants !
for: label doing: aBlock
	" Return a parser that evaluates aBlock to match label. "

	^self new do: aBlock for: label! !

! PPCharacterParser class methodsFor: #constants !
regExpCache
	" Private - Return the regexp cache. "

	RegExpCache isNil ifTrue: [
		RegExpCache := PoolDictionary new
	].
	^RegExpCache! !

! PPCharacterParser methodsFor: #accessing !
do: aBlock for: aString
	" Set the receiver to match aString using aBlock. "

	theString := aString.
	regexp := aBlock! !

! PPCharacterParser methodsFor: #accessing !
string: aString
	" Set the string of the receiver. "

	theString := aString.
	regexp := self class regExpCache at: aString ifAbsentPut: [
		RegularExpression fromString: '[', aString, ']'
	]! !

! PPCharacterParser methodsFor: #private !
match: aString
	" Private - Match the receiver to string. "

	aString isNil ifTrue: [ ^false ].
	regexp isNil ifTrue: [ ^aString isSeparator ].
	regexp isValuable ifTrue: [ ^regexp value: aString ].
	theString = aString ifTrue: [ ^true ].
	^aString match: regexp! !

! PPCharacterParser methodsFor: #accessing !
matchFailureReason
	" Private - Return the reason for mismatch. "

	^'Could not match ' ,(theString isNil
		ifTrue: [#separator]
		ifFalse: [theString printString])! !

! PPCharacterParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	^(self match: aStream peek)
	 ifTrue:  [ aStream next ]
	 ifFalse: [ PPFailure
		reason: self matchFailureReason
		at: aStream position
	]! !

! PPCharacterParser methodsFor: #separator !
separator
	" Set the receiver to parse separator. "

	theString := regexp := nil.
	self @@#separator! !

! PPCharacterParser methodsFor: #printing !
printStringDetail
	" Private - Return a printable detail of the receiver. "

	theString isCharacter ifTrue: [
		(theString between: $0 and: $z) ifTrue: [
			^$$ ,theString
		].
		^$# ,theString asciiValue
	].
	^theString printString! !

PPParser
 subclass: #PPListParser
 instanceVariableNames: #parsers
 category: #Compiler-Parsing!

! PPListParser class methodsFor: #instantiation !
withAll: aCollection
	" Return an instance of the receiver. "

	    ^self basicNew
		parsers: aCollection;
		yourself! !

! PPListParser class methodsFor: #instantiation !
with: aParser with: anotherParser
	" Return an instance of the receiver. "

	^self withAll: (Array with: aParser with: anotherParser)! !

! PPListParser methodsFor: #accessing !
parsers
	" Return the parsers of the receiver. "

	parsers isNil ifTrue: [ ^#() ].
	^parsers! !

! PPListParser methodsFor: #accessing !
parsers: aCollection
	" Set the parsers of the receiver. "

	parsers := aCollection collect: [:each| each asParser ]! !

! PPListParser methodsFor: #copying !
copyWith: aParser
	" Return a copy of the receiver with aParser. "

	^self class withAll: (self parsers copyWith: aParser)! !

! PPListParser methodsFor: #printing !
op
	" Private - Return the operator of the receiver. "

	^$+! !

! PPListParser methodsFor: #printing !
printString
	" Return a printable representation of the receiver. "

	| stream |
	(self basicAt: #debugTag) notNil ifTrue: [ ^super printString ].
	stream := ${ writeStream.
	self parsers
		do: [:each| stream nextPutAll: each printString ]
		separatedBy: [stream nextPutAll: self op].
	^stream contents ! !

PPListParser
 subclass: #PPSequenceParser
 category: #Compiler-Parsing!

! PPSequenceParser methodsFor: #copying !
, aRule
	" Return the receiver composed with the argument. "

	^self copyWith: aRule! !

! PPSequenceParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	| start result element |
	result := #().
	start := aStream position.
	self parsers do: [:each |
		element := each memoizedParse: aStream.
		result add: element.
		(self isParseFailure: element) ifTrue: [
			aStream position: start.
			^element
		]
	].
	^result! !

! PPSequenceParser methodsFor: #printing !
op
	" Private - Return the operator of the receiver. "

	^$,! !

PPListParser
 subclass: #PPChoiceParser
 category: #Compiler-Parsing!

! PPChoiceParser methodsFor: #copying !
/ aRule
	" Return the receiver composed with aRule. "

	^self copyWith: aRule! !

! PPChoiceParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	| result |
	self parsers detect: [:each |
		result := each memoizedParse: aStream.
		(self isParseFailure: result) not
	] ifNone: [].
	^result! !

! PPChoiceParser methodsFor: #printing !
op
	" Private - Return the operator of the receiver. "

	^$/! !

PPParser
 subclass: #PPDelegateParser
 instanceVariableNames: #parser
 category: #Compiler-Parsing!

! PPDelegateParser methodsFor: #accessing !
parser
	" Return the parser of the receiver. "

	^parser! !

! PPDelegateParser methodsFor: #accessing !
parser: aParser
	" Set the parser of the receiver. "

	parser := aParser! !

! PPDelegateParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	^self parser memoizedParse: aStream! !

! PPDelegateParser class methodsFor: #instantiation !
on: aParser
	" Return an instance of the receiver. "

	    ^self new
		parser: aParser;
		yourself! !

! PPDelegateParser methodsFor: #printing !
printStringDetail
	" Private - Return a printable detail of the receiver. "

	^'[' ,parser printString ,']'! !

PPDelegateParser
 subclass: #PPAndParser
 category: #Compiler-Parsing!

! PPAndParser methodsFor: #parsing !
opParse: aStream
	" Parse contents in aStream. "

	| element position |
	position := aStream position.
	element := self parser memoizedParse: aStream.
	aStream position: position.
	^element! !

! PPAndParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	^self opParse: aStream! !

PPAndParser
 subclass: #PPNotParser
 category: #Compiler-Parsing!

! PPNotParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	| element |
	element := self opParse: aStream.
	(self isParseFailure: element) ifTrue: [ ^nil ].
	^PPFailure reason: element at: aStream position! !

PPDelegateParser
 subclass: #PPActionParser
 instanceVariableNames: #block
 category: #Compiler-Parsing!

! PPActionParser methodsFor: #accessing !
block
	" Return the block of the receiver. "

	^block! !

! PPActionParser methodsFor: #accessing !
block: aBlock
	" Set the block of the receiver. "

	block := aBlock! !

! PPActionParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	| element |
	element := self parser memoizedParse: aStream.
	(self isParseFailure: element) ifTrue: [ ^element ].
	^self block value: element! !

! PPActionParser class methodsFor: #instantiation !
on: aParser block: aBlock
	" Return an instance of the receiver. "

	^self new
		parser: aParser;
		block: aBlock;
		yourself! !

! PPActionParser methodsFor: #printing !
printString
	" Return a printable representation of the receiver. "

	^${ ,parser printString ,$} ! !

PPDelegateParser
 subclass: #PPFlattenParser
 category: #Compiler-Parsing!

! PPFlattenParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	| start element |
	start := aStream position.
	element := self parser memoizedParse: aStream.
	(self isParseFailure: element) ifTrue: [ ^element ].
	^aStream collection 
		copyFrom: start + 1 
		to: aStream position! !

PPDelegateParser
 subclass: #PPSourceParser
 category: #Compiler-Parsing!

! PPSourceParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	| start element result |
	start := aStream position.
	element := self parser memoizedParse: aStream.
	(self isParseFailure: element) ifTrue: [ ^element ].
	result := aStream collection
		copyFrom: start + 1
		to: aStream position.
	^Array with: element with: result! !

PPDelegateParser
 subclass: #PPRepeatingParser
 instanceVariableNames: #min
 category: #Compiler-Parsing!

! PPRepeatingParser methodsFor: #accessing !
min
	" Return the min of the receiver. "

	^min! !

! PPRepeatingParser methodsFor: #accessing !
min: aNumber
	" Set the min of the receiver. "

	min := aNumber! !

! PPRepeatingParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	| start element result failure |
	start := aStream position.
	result := Array new.
	[ (result size < self min) and: [ failure isNil ]
	] whileTrue: [
	    element := self parser memoizedParse: aStream.
	    (self isParseFailure: element)
		ifFalse: [ result addLast: element ]
		ifTrue: [ aStream position: start.
			  failure := element ]
	].
	failure notNil ifTrue: [ ^failure ].
	[failure isNil] whileTrue: [
		element := self parser memoizedParse: aStream.
	 	(self isParseFailure: element)
			ifTrue: [ failure := element ]
			ifFalse: [ result addLast: element ]
	].
	^result! !

! PPRepeatingParser class methodsFor: #instantiation !
on: aParser min: aNumber
	" Return an instance of the receiver. "

	^self new
		parser: aParser;
		min: aNumber;
		yourself! !

! PPRepeatingParser methodsFor: #printing !
printStringDetail
	" Private - Return a printable detail of the receiver. "

	^'[min=' ,min ,']'! !

PPParser
 subclass: #PPInlineParser
 instanceVariableNames: #terminator
 category: #Compiler-Parsing!

! PPInlineParser class methodsFor: #instantiation !
upTo: aString
	" Return an instance of the receiver. "

	^self new terminator: aString; yourself! !

! PPInlineParser methodsFor: #accessing !
terminator: aString
	" Set terminator of the receiver. "

	terminator := aString! !

! PPInlineParser methodsFor: #parsing !
basicParse: aStream
	" Private - Parse contents in aStream. "

	| result |
	result := aStream upTo: terminator.
	aStream atEnd ifFalse: [ ^result ].
	^PPFailure
		reason: 'Missing inline terminator (',terminator,')'
		at: aStream position! !

EventManager
 subclass: #SourceCodeParser
 category: #Compiler-Parsing!

! SourceCodeParser class methodsFor: #instantiation !
parse: aStream
	" Return the result of parsing contents in aStream. "

	^self new parse: aStream ! !

! SourceCodeParser class methodsFor: #debug !
debug: aStream
	^PPParser debugWhile: [
		self parse: aStream
	]! !

! SourceCodeParser methodsFor: #parsing !
parse: aStream
	" Return the result of parsing contents in aStream. "

	^self parser parse: aStream! !

! SourceCodeParser methodsFor: #gc !
gc
	" Private - Perform (preventive) garbage collection.
	Note: JavascriptCore VM can crash reporting
	 'memory issues' if we do not GC asap.
	 This issue will persist while Apple continue
	  distributing a version of JSC with issue reported
		swiki:U8/232
		https://bugs.webkit.org/show_bug.cgi?id=160027
	"

	(Smalltalk includesKey: #JSC) ifTrue: [ Smalltalk gc ].! !

! SourceCodeParser methodsFor: #parsers !
separator
	" Return the parser for separator. "

	^PPCharacterParser separator! !

! SourceCodeParser methodsFor: #parsers !
parser
	" Return the method parser. "

	^self implementedBySubclass: #parser! !

! SourceCodeParser methodsFor: #parsers !
charParser: options
	" Private - Return the character parser. "

	^PPCharacterParser for: 'a-zA-Z0-9',options doing: [:c|
		c isAlphaNumeric or: [ options includes: c ]
	]! !

! SourceCodeParser methodsFor: #parsers !
charParserOr: aCharacter
	" Private - Return the character parser. "

	^PPCharacterParser for: 'a-zA-Z0-9',aCharacter doing: [:c|
		c isAlphaNumeric or: [ c = aCharacter ]
	]! !

! SourceCodeParser methodsFor: #parsers!
, anObject
	" Private - Return the parser to start parsing anObject. "

	^PPParser , anObject! !

! SourceCodeParser methodsFor: #parsers!
optional: anObject
	" Private - Return the parser to parse an optional element anObject. "

	^PPParser optional: anObject! !

! SourceCodeParser methodsFor: #parsers!
gc: anObject
	" Private - Return the parser to parse anObject. "

	^(self ,anObject) ==> [:node| self gc. node ]! !

SourceCodeParser subclass: #SmalltalkParser !

! SmalltalkParser class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^#(	#needsDirectiveParser:ws:
	) , super constructEventsTriggered! !

! SmalltalkParser methodsFor: #directives!
missingDirective: number ws: ws
	" Private - Return the parser for missing support of directives. "

	^(PPCharacterParser notIn: $>) star flatten
	, $> @@#missingDirective
	==> [:node| self error: 'Invalid directive - ' ,node first ]! !

! SmalltalkParser methodsFor: #directives !
jsDirective: number ws: ws
	" Private - Return the parser for native directives.
	"

	^self	, #js: / #javascript:
		, (PPInlineParser upTo: ':>')
		@@#jsDirective
	==> [:node| JSDirectiveNode
		code: node second
		protected: node first = #javascript:
	]! !

! SmalltalkParser methodsFor: #directives !
optDirective: number ws: ws
	" Private - Return the parser for compiler optimizations directive.
	"

	^self	, #opt: / #opt+: / #opt-:
		,(PPInlineParser upTo: ':>')
		@@#optDirective
	==> [:node| InlineOptimizationNode new
		args: node second asArrayOfSubstrings;
		remove: node first = #opt-:;
		yourself
	]! !

! SmalltalkParser methodsFor: #directives!
defaultDirective: number ws: ws
	" Private - Return the parser for default support of directives.
	We support inlined native directives.
	"

	^ (self optDirective: number ws: ws)
	/ (self jsDirective: number ws: ws)
	/ (self missingDirective: number ws: ws)! !

! SmalltalkParser methodsFor: #directives!
directive: number ws: ws
	" Private - Return the parser for directives. "

	| parser |
	parser := self
		triggerEvent: #needsDirectiveParser:ws:
		withArguments: (Array with: number with: ws)
		ifNotHandled: [ self defaultDirective: number ws: ws ].

	^self	, $< ,(ws, parser, ws) @@#directive
		==> [:node | node second second ]! !

! SmalltalkParser methodsFor: #constants !
binarySelectorChars
	" Return the binary selector characters mask. "

	^String binarySelectorChars ! !

! SmalltalkParser methodsFor: #parsers !
binaryCharParser
	" Return the binary selector character parser. "

	^(PPCharacterParser in: self binarySelectorChars) @@#binaryChar! !

! SmalltalkParser methodsFor: #parsers !
symbolCharParser
	" Return the symbol character parser. "

	^(self charParserOr: $_) @@#symbolChar! !

! SmalltalkParser methodsFor: #parsers !
jsSymbolCharParser
	" Private - Return the js symbol character parser. "

	^(self charParser: '_$') @@#jsSymbolChar! !

! SmalltalkParser methodsFor: #parsers !
extendedSymbolCharParser
	" Private - Return the extended symbol character parser. "

	^(self charParser: '_:' , self binarySelectorChars) @@#extendedSymbolChar! !

! SmalltalkParser methodsFor: #parsers !
globalNameParser
	" Return the parser for global names. "

	^((PPCharacterParser between: $A and: $Z)
		, self symbolCharParser star)
		flatten @@#globalNameParser! !

! SmalltalkParser methodsFor: #parsers !
comment
	" Return the parser for comment. "

	^($" asParser
	,($" asParser not, PPAnyParser new) star
	, $" @@#stComment) flatten
	@@#comment! !

! SmalltalkParser methodsFor: #parsers !
nameStart
	" Return the parser for name start."

	^(PPCharacterParser for: 'a-z_' doing: [:c|
		(c between: $a and: $z)
		or: [ c = $_ ]
	]) @@#nameStart! !

! SmalltalkParser methodsFor: #parsers !
upNameStart
	" Return the parser for name start (can be uppercase)."

	^(PPCharacterParser for: 'A-Za-z_' doing: [:c|
		c isLetter or: [ c = $_ ]
	]) @@#upNameStart! !

! SmalltalkParser methodsFor: #parsers !
digits
	" Return the parser for numeric digits."

	^(PPCharacterParser between: $0 and: $9) plus! !

! SmalltalkParser methodsFor: #parsers !
decimals
	" Return the parser for number decimals."

	^self	, $.
		, self digits
		@@#decimals! !

! SmalltalkParser methodsFor: #parsers !
exponent
	" Return the parser for number exponent."

	^self	, $e
		, (self optional: $-)
		, self digits
		@@#exponent! !

! SmalltalkParser methodsFor: #parsers !
radix
	" Return the parser for number radix."

	^self	, $r
		, (PPCharacterParser for: '0-9A-Z' doing: [:c|
			(c between: $0 and: $9)
			or: [ c between: $A and: $Z ]
		]) plus
		@@#radix! !

! SmalltalkParser methodsFor: #parsers !
number
	" Return the parser for number."

	^(self optional: $-)
		, self digits , self decimals optional
		, (self exponent / self radix) optional
		@@#number! !

! SmalltalkParser methodsFor: #private !
annotationStart
	" Private - Return the parser for starting annotations. "

	^self ,#:: ! !

! SmalltalkParser methodsFor: #private !
methodParserParts
	" Private - Return a parser for parts of a method. "

 | expression separator comment ws identifier keyword stSelectorPart
  string symbol number variable reference globalReference
  literal returnExpression expressionParser keyword unarySelector
  binarySelector keywordPattern unaryPattern binaryPattern assignment
  temps blockParamList block expression expressions subexpression
  statements sequence operand unaryMessage unarySend unaryTail
  binaryMessage binarySend binaryTail keywordMessage keywordSend
  keywordPair cascade message inline jsIndex character directive
  arrayLiteral arrayVariable literalArray aSymbol
  stIdentifier jsIdentifier noReference alienReference
  alienIdentifier emptyArray hashedString annotation argument
 |

 separator := self separator.
 comment := self comment.
 ws := (separator / comment @@#sp) star @@#ws.
 
 string := self ,$' ,PPStStringParser new @@#string
  ==> [:node| ValueNode with: node second ].
 
 hashedString := self ,($#,$') ,PPStStringParser new @@#hashedString
  ==> [:node| ValueNode with: node second ].

 stIdentifier := (self nameStart, self symbolCharParser star @@#stId) flatten @@#stIdentifier.
 stSelectorPart := (self upNameStart, self symbolCharParser star @@#stSel) flatten @@#stSelectorPart.
 jsIdentifier := (self ,$# ,self jsSymbolCharParser plus @@#jsId) flatten @@#jsIdentifier.
 alienIdentifier := self jsSymbolCharParser plus flatten @@#alienIdentifier.
 identifier := jsIdentifier / stIdentifier @@#identifier.

 keyword := (stSelectorPart / identifier, $:) flatten @@#keyword.

 annotation := self annotationStart
	,(string / stIdentifier / self globalNameParser) @@#annotation
	==>[:node| node last ].
 argument := identifier ,annotation star ==> [:node|
	(ValueNode with: node first) annotations: node last
 ].

 symbol := self ,$# ,self extendedSymbolCharParser plus flatten @@#symbol
  ==> [:node | ValueNode with: node second ].

 number := self number flatten
  ==> [:node | ValueNode with: node asNumber ].

 character := self ,$$ ,PPAnyParser new @@#character
  ==> [:node | ValueNode with: (node at: 2) ].

 directive := self directive: number ws: ws.

 variable := identifier ==> [:token | VariableNode new value: token ].
 noReference := self ,$# @@#noReference ==> [:token | NoReceiverNode new ].
 literal := PPDelegateParser new @@#literalDelegate.
 expression := PPDelegateParser new @@#expressionDelegate.

 arrayVariable :=
	  (self ,#true  ==> [:node| ValueNode with: true ])
	/ (self ,#false ==> [:node| ValueNode with: false ])
	/ (self ,#nil   ==> [:node| ValueNode with: nil ])
	@@#arrayVariable.

 aSymbol := self extendedSymbolCharParser plus flatten @@#aSymbol
  ==> [:node| ValueNode with: node ].

 arrayLiteral := literal / arrayVariable / aSymbol / variable.
 literalArray := (self gc: '#(')
	,(ws ,arrayLiteral ,ws) star ,$)
	@@#literalArray
  ==> [:node | ValueNode with: (Array
	withAll: (node second
		collect: [:each | each second value])) ].
 emptyArray := self ,'#(' ,ws ,$) @@#emptyArray
  ==> [:node | ValueNode with: Array new ].

 globalReference := self globalNameParser @@#globalReference
  ==> [:token | self gc. GlobalReferenceNode new value: token ].
 alienReference := self ,($# ,${) ,alienIdentifier, $} @@#alienReference
  ==> [:node | AlienNode new value: node second ].

 jsIndex := (self gc: $[) ,expression, $] / (self ,$# ,alienIdentifier).
 reference := (variable / globalReference / alienReference @@#ref)
	,jsIndex star ,annotation star @@#reference
  ==> [:nodes | nodes first reference: nodes ].

 binarySelector := (
	(self optional: $#)
	,self binaryCharParser plus
	) flatten @@#binarySelector.

 unarySelector := stSelectorPart / identifier @@#unarySelector.

 keywordPattern :=
	(ws, keyword  @@#kw ,
	 ws, argument @@#kwId
	) plus @@#keywordPattern
  ==> [:nodes | Array
   with: ((nodes collect: [:each | each at: 2]) join: '')
   with: (nodes collect: [:each | each at: 4]) ].

 binaryPattern := ws, binarySelector, ws, argument @@#binaryPattern
  ==> [:node | Array with: node second with: (Array with: node fourth)].

 unaryPattern := ws, unarySelector @@#unaryPattern
  ==> [:node | Array with: node second with: Array new].
 
 expressions := expression, ((ws, $., ws, expression) @@#anExpression
  ==> [:node | node fourth]) star @@#expressions
  ==> [:node | (Array with: node first) ,node second ].

 assignment := reference, ws, ':=', ws, expression @@#assignment
  ==> [:node | AssignmentNode new left: node first; right: (node at: 5)].

returnExpression := (self gc: $^) @@#caret
	, ws, expression ,annotation star
	, ws, $. asParser optional	@@#returnExpression
     ==> [:node | ReturnNode new
	addNode: (node at: 3);
	annotations: (node at: 4);
	yourself ].

 temps := self, $|, (ws, argument) star, ws, $| @@#temps
  ==> [:node | node second collect: [:each | each second ] ].

 blockParamList := (self ,$: ,ws ,argument ,ws) plus ,$| @@#blockParamList
  ==> [:node | node first collect: [:each | each third ] ].

 subexpression := (self gc: $() ,ws ,expression ,ws ,$) @@#subexpression
  ==> [:node | node third ].

 statements := (returnExpression @@#stat1 ==> [:node | Array with: node])
  / (expressions ,ws ,$. ,ws ,returnExpression @@#stat2 ==> [:node | node first add: (node at: 5); yourself ])
  / (expressions ,(self optional: $.) @@#stat3 ==> [:node | node first ]).

 sequence := temps optional @@#temps
	, ws, directive star
	, ws, statements optional
	, ws @@#sequence
  ==> [:node | ExtendedSequenceNode new
	temps: node first;
	directives: node third;
	nodes: (node at: 5);
	yourself ].

 block := (self gc: $[)
	,ws ,blockParamList optional
	,ws ,sequence optional
	,ws ,$] @@#block
  ==> [:node | BlockNode new
	parameters: node third;
	addNode: (node at: 5) asBlockSequenceNode ].

 operand := block / literal / reference / noReference / subexpression @@#operand.

 literal parser: symbol / number / string / character
	/ emptyArray / literalArray / hashedString @@#literalDelegate.

 unaryMessage := ws ,unarySelector ,(self ,$:) not @@#unaryMessage
  ==> [:node | SendNode new selector: node second ].

 unaryTail := PPDelegateParser new @@#unaryTailDelegate.
 unaryTail parser: (unaryMessage, unaryTail optional @@#unaryTailDelegate
  ==> [:node | node second isNil ifTrue: [ node first ]
	ifFalse: [ node second valueForReceiver: node first ] ]).

 unarySend := operand, unaryTail optional @@#unarySend
  ==> [:node | node second isNil ifTrue: [ node first ]
	ifFalse: [ node second valueForReceiver: node first ] ].

 binaryMessage := ws, binarySelector, ws, (unarySend / operand) @@#binaryMessage
  ==> [:node | SendNode new
	selector: node second;
	arguments: (Array with: node fourth) ].

 binaryTail := PPDelegateParser new @@#binaryTailDelegate.
 binaryTail parser: (binaryMessage, binaryTail optional
  ==> [:node | node second isNil ifTrue: [ node first ]
	ifFalse: [ node second valueForReceiver: node first ] ]).

 binarySend := unarySend, binaryTail optional @@#binarySend
  ==> [:node | node second isNil ifTrue: [ node first ]
	ifFalse: [ node second valueForReceiver: node first ] ].

 keywordPair := keyword, ws, binarySend @@#keywordPair.

 keywordMessage := (ws, keywordPair) plus @@#keywordMessage
  ==> [:nodes | SendNode new
	selector: ((nodes collect: [:each | each second first]) join: '');
	arguments: (nodes collect: [:each | each second third])].

 keywordSend := binarySend, keywordMessage @@#keywordSend
  ==> [:node | node second valueForReceiver: node first].

 message := binaryMessage / unaryMessage / keywordMessage @@#message.

 cascade := (keywordSend / binarySend @@#head)
	,(ws, $;, message) plus @@#cascade
  ==> [:node | node first cascadeNodeWithMessages: 
	(node second collect: [:each | each third ])].

 inline := self ,${
	,(stIdentifier optional ,string ,ws) plus
	,$} @@#inline
  ==> [:node | InlineNode sources: node second parser: ws ,sequence optional ].

 expression parser: assignment / cascade / keywordSend / binarySend / inline @@#expression.

^Array	with: ws
	with: keywordPattern / binaryPattern / unaryPattern @@#header
	with: sequence optional @@#body! !

! SmalltalkParser methodsFor: #parsers !
parser
	" Return the method parser. "

	| parts ws header body method |
	parts := self methodParserParts.
	ws := parts first.
	header := parts second.
	body := parts third.

	method := (ws ,header
		,ws ,body
		,ws @@#impl) withSource
	  ==> [:node | MethodNode new
		selector: node first second first;
		arguments: node first second second;
		addNode: node first fourth shouldReturn;
		source: node second;
		yourself ].
 
	self gc.
 	^method ,PPEOFParser new @@#method
	   ==> [:nodes | nodes first ]! !

! SmalltalkParser methodsFor: #parsers !
methodHeaderParser
	" Return a parser for method header. "

	| parts ws header |
	parts := self methodParserParts.
	ws := parts first.
	header := parts second.
	^ws ,header @@#methodHeader
		==> [:nodes| nodes second ]! !

! Object methodsFor: 'Compiler-converting' !
asParser
	" Return a parser on the receiver. "

	^self subclassResponsibility: #asParser! !

! Collection methodsFor: 'Compiler-converting' !
asParser
	" Return a parser on the receiver. "

	^PPChoiceParser withAll: self asArray! !

! String methodsFor: 'Compiler-converting' !
asParser
	" Return a parser on the receiver. "

	^self isCharacter
	ifTrue:  [ PPCharacterParser in: self ]
	ifFalse: [ PPStringParser new string: self ]! !

! String methodsFor: 'Compiler-converting' !
asChoiceParser
	" Return a parser of choices for each character the receiver. "

	^PPChoiceParser withAll: (self asArray collect: [:each | each asParser ])! !

! String methodsFor: 'Compiler-converting' !
asCharacterParser
	" Return a parser on the characters of the receiver. "

	^PPCharacterParser new string: self! !

! Object methodsFor: 'Compiler-inlines' !
ifUndefined: aBlock
	" Return the receiver or the result of evaluating aBlock if it is undefined.
	Notes:	-this message is inlined by compiler (it is not sent)
		-aBlock is evaluated ONLY in case of undefined.
	"

	^self! !

! Object methodsFor: 'Compiler-inlines' !
orNil
	" Return the receiver or nil.
	Notes:	-this message is inlined by compiler (it is not sent)
		-nil, null and undefined returns nil.
	"

	^self! !

! Behavior methodsFor: 'Compiler-generation' !
codeFor: anEncoder accessing: aNode
	" Return the code to access variable in aNode (aVariableNode).
	Return nil if the code to access variable can't be generated.
	The default implementation return code to access instance variables.
	"

	| aName |
	aName := aNode value.
	self allInstVarNamesDo: [:each|
		aName = each ifTrue: [
			^anEncoder codeToAccessInstanceVariable: aName in: aNode
		].
	].
	self allPropertyNamesDo: [:each|
		aName = each ifTrue: [
			^anEncoder codeToAccessProperty: aName in: aNode
		].
	].
	^nil! !

! Smalltalk methodsFor: 'Compiler-compiling' !
recompileAll
	" Recompile all classes in the receiver. "

	self classes do: [:each | each recompile ]! !

! Smalltalk methodsFor: 'Compiler-accessing' !
compilerSupport
	" Return the compiler support for the receiver. "

	^Compiler! !

! Object methodsFor: 'Compiler-parser-bug' !
accept: what
	" Private - Report buggy send by compiler. "

	self note: '[bug] ' ,self asString ,' accept: ' ,what halt.
	^self error: '#accept: out of control.'! !

! Object methodsFor: 'Compiler-accessing' !
compilerSupport
	" Return the compiler of the receiver "

	^self smalltalk compilerSupport! !

! Behavior methodsFor: 'Compiler-compiling' !
compiler
	" Return a compiler for receiver's source code. "

	^self compilerSupport for: self! !

! Behavior methodsFor: 'Compiler-compiling' !
recompile
	" Recompile all the methods of the receiver. "

	self compiler recompile: self! !

! Behavior methodsFor: 'Compiler-compiling' !
recompileAll
	" Recompile all the methods of the receiver and its subclasses.	"

	self instanceClass withAllSubclassesDo: [:each| each recompile ]! !

! CompiledMethod methodsFor: 'Compiler-compiling' !
recompile
	" Recompile the source of the receiver and install the new compiled method. "

	self classField compiler
		compile: self source
		forClass: self classField
		category: self category! !

! CompiledMethod methodsFor: 'Compiler-inspecting' !
parseNodes
	" Return the result of parsing the source of the receiver. "

	^self classField compiler parseMethod: self source! !

! Behavior methodsFor: 'Compiler-compiling' !
recompile: what
	" Recompile the method(s) at selector(s) of the receiver.	"

	what isString ifTrue: [ ^self recompile: (Array with: what) ].
	what do: [:selector| (self >> selector) recompile ]! !

! CompiledMethod methodsFor: 'Compiler-transforms' !
changeCommentWith: aBlock
	" Private - Change the comment of the receiver with aBlock.
	aBlock must be a one argument block to be evaluated with current comment.
	aBlock must return the new/empty comment (or nil, to cancel change).
	"

	| stream header old new nl |
	stream := self source stream.
	header := SmalltalkParser new methodHeaderParser parse: stream.
	header isParseFailure ifTrue: [ ^false ].
	stream skipSeparators.
	header := stream collection copyFrom: 1 to: stream position.
	stream peek = $" ifTrue: [
		stream next.
		old := stream upTo: $".
		new := aBlock evaluateWith: old.
	] ifFalse: [
		new := aBlock evaluateWith: ''.
	].
	new isNil ifTrue: [ ^false ].
	nl := old isNil ifTrue: [ '
	' ] ifFalse: [ '' ].
	new := nl ,'" ' ,new trimBlanks ,' "' ,nl.

	self
		source: header ,new ,stream upToEnd;
		recompile! !

! CompiledMethod methodsFor: 'Compiler-tags' !
tags
	" Development - Return the tags stored in the receiver's comment. "

	self changeCommentWith: [:old|
		^((old upTo: $-)
			asArrayOfSubstringsSeparatedBy: $,)
			collect: [:tag| tag trimBlanks ]
	].
	^Array new! !

! CompiledMethod methodsFor: 'Compiler-accessing' !
comment: aString
	" Change the comment in the receiver to aString. "

	^self changeCommentWith: [:old| aString ]! !

! CompiledMethod methodsFor: 'Compiler-accessing' !
messageComment
	" Return the comment of the message implemented by the receiver.
	The messageComment is obtained from first line of comment of the receiver.
	Note: it is expected that all the methods implementing the same message in this system have the same messageComment.
	"

	| line |
	line := self comment stream nextLine.
	(line includes: $-) ifFalse: [ ^line ].
	^(line stream upTo: $-; upToEnd) trimBlanks! !

! CompiledMethod methodsFor: 'Compiler-transforms' !
messageComment: aString
	" Set the comment of the message implemented by the receiver. "

	^self changeCommentWith: [:old| | line tags stream |
		stream := old stream.
		line := stream nextLine.
		(line includes: $-) ifTrue: [
			tags := (line upTo: $-) ,'- '
		] ifFalse: [ tags := '' ].
		tags	,aString ,(stream atEnd ifTrue: [''] ifFalse: ['
'			,stream upToEnd])
	]! !

! Behavior methodsFor: 'Compiler-templates' !
templateFrom: host doing: aBlock on: methodBlock
	" Return the result of collecting mutations in template cached in host (or nil). "

	| template |
	host isNil ifTrue: [ ^nil ].
	template := host basicAt: #s8Template ifAbsent: [].
	(template isNil and: [host isKindOf: CompiledMethod]) ifTrue: [
		template := Compiler new
			useTemplates: false;
			currentClass: host classField;
			parseMethod: host source.
		host basicAt: #s8Template put: template.
	].
	^template notNil ifTrue: [
		template := template deepCopy.
		aBlock notNil ifTrue: [ template do: aBlock ].
		methodBlock notNil
			ifTrue: [ methodBlock value: template ]
			ifFalse: [ template ]
	]! !

! Behavior methodsFor: 'Compiler-templates' !
templateFrom: host with: methodBlock
	" Return the template cached in host (or nil). "

	^self templateFrom: host doing: nil on: methodBlock! !