" Adaptions to run S8 on Julia execution machinery.
Copyright (C) 2021-today Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
In case of questions and bug reports please join us at S8 Facebook group
	https://www.facebook.com/groups/s8smalltalk/
"!

"------------------------------------------------------------------------------
	Julia code generation
 ------------------------------------------------------------------------------"!
NativeEncoder
	subclass: #JuliaEncoder
	instanceVariableNames: ''
	category: #Julia !

! JuliaEncoder class methodsFor: #instantiation !
supportsLanguage: aLanguage
	" Private - Return true if instances of the receiver can encode to aLanguage. "

	^aLanguage = #jl ! !

! JuliaEncoder methodsFor: #accessing !
nativeLanguage
	" Return the nativeLanguage of the receiver. "

	^#jl ! !

! JuliaEncoder methodsFor: #encoding !
nativeUndefined
	" Private - Return the code for undefined. "

	^#nil ! !

! JuliaEncoder methodsFor: #encoding !
nativeNameOfMetaClass: aClass
	" Private - Return the native name for metaclass aClass. "

	^'classOf_(smalltalk,smalltalk.' ,aClass instanceClass name ,$) ! !

! JuliaEncoder methodsFor: #encoding !
nativeNameOfClass: aClass
	" Private - Return the native name for aClass. "

	aClass isMetaclass ifTrue: [
		^self nativeNameOfMetaClass: aClass
	].
	^'smalltalk.' ,aClass name ! !

! JuliaEncoder methodsFor: #encoding !
nativeUnion: aClassName with: typeName
	" Private - Return the name of union of class & type. "

	^'Union{' ,aClassName ,$, ,typeName ,$} ! !

! JuliaEncoder methodsFor: #encoding !
nativeTypeOfClass: aClass
	" Private - Return the name of (abstract) type for aClass. "

	| result clsName |
	aClass isMetaclass ifTrue: [
		^#St ,aClass instanceClass name ,#_class
	].
	clsName := aClass name.
	result := #St ,clsName.
	clsName = #Object ifTrue: [ ^self nativeUnion: result with: #Any ].
	clsName = #Class ifTrue: [ ^self nativeUnion: result with: #StObject_class ].
"	clsName = #Metaclass ifTrue: [ ^self nativeUnion: result with: #StObject_class ]."
	(String inheritsFrom: aClass) ifTrue: [ ^self nativeUnion: result with: #String ].
	(aClass inheritsFrom: Number) ifTrue: [ ^self nativeUnion: result with: #Number ].
	(aClass inheritsFrom: Boolean) ifTrue: [ ^self nativeUnion: result with: #Bool ].
	(aClass inheritsFrom: BlockClosure) ifTrue: [ ^self nativeUnion: result with: #Function ].
	(aClass inheritsFrom: Smalltalk) ifTrue: [ ^self nativeUnion: result with: #StSystemDictionary ].
	^result ! !

! JuliaEncoder methodsFor: #export !
smalltalkNameFor: aClass
	" Private - Return the global name of aClass (can be nil). "

	aClass isNil ifTrue: [ ^self nativeUndefined ].
	^self nativeNameOfClass: aClass ! !

! JuliaEncoder methodsFor: #export !
smalltalkTypeFor: aClass
	" Private - Return the global name of type of aClass (can be nil). "

	aClass isNil ifTrue: [ ^#Nothing ].
	^self nativeTypeOfClass: aClass ! !

! JuliaEncoder methodsFor: #constants !
separatorOp
	" Private - Return the OP for separating statements. "

	^'
' ! !

! JuliaEncoder methodsFor: #accessing !
callEscape
	" Private - Return the selector sequence to escape OO message send selector, and produce code for function call.
	Note that an alien selector starting with two $$ are understood as a function call to function starting with $
	"

	^'#_'! !

! JuliaEncoder methodsFor: #inlines !
missingNativeCodeFor: aNode as: lang
	" Private - Return the native language code for missing expression (or nil). "

	^'invalidMessage_(self,"' ,compiler selector ,'")'! !

! JuliaEncoder methodsFor: #private !
canExportMethod: aMethod
	" Private - Return true if the receiver can/want to export method aMethod. "

	^true! !

! JuliaEncoder methodsFor: #converting !
asSelector: aSelector
	" Return the native representation of aSelector. "

	^aSelector asJuliaSelector! !

! JuliaEncoder methodsFor: #private !
nativeSourceFor: aMethod
	" Private - Return the code of native function implementing aMethod. "

	^(self canExportMethod: aMethod) ifTrue: [
		self privateSourceFor: aMethod
	]! !

! JuliaEncoder methodsFor: #export !
export: aClass method: aMethod on: aStream
	" Export a method definition of aClass onto aStream. "

	| fName selectorName bar source |
	source := self nativeSourceFor: aMethod.
	source isNil ifTrue: [ ^false ].
	aMethod isNil ifTrue: [ self error: 'Method is missing' ].
	bar := $".
	fName := aMethod selector.
	fName first isBackSlash ifTrue: [
		selectorName := self escaped: aMethod selector.
		fName := self escaped: (self asSelector: fName).
	] ifFalse: [
		selectorName := bar ,aMethod selector, bar.
		(self asSelector: fName) = (aMethod selector replace: $: with: $_) ifTrue: [
			fName := 0
		] ifFalse: [
			fName := bar ,(self asSelector: fName) ,bar
		].
	].
	aStream
		nextPutAll: 'bind(smalltalk,' ,(self smalltalkNameFor: aClass)
			,',' ,selectorName ,',' ,fName ,'
,'			,source.
	(aMethod source notNil and: [ aMethod source notEmpty ]) ifTrue: [
	    aStream cr; "warning: this cr is important for debugger"
		nextPutAll: ',' ,(self escaped: aMethod category)
			,',' ,(self escaped: aMethod source)
			,',' ,(self escaped: source).
	].
	((aMethod respondsTo: #stamp) and: [ aMethod stamp notNil ]) ifTrue: [
		aStream nextPutAll: ',' ,(self escaped: aMethod stamp)
	].
	aStream nextPutAll: ')
'.	^true! !

! JuliaEncoder methodsFor: #arrays !
arrayLiteralStart
	" Private - Return the code string to start exporting an array literal. "

	^'newArray(smalltalk,' ,$[! !

! JuliaEncoder methodsFor: #arrays !
arrayLiteralEnd
	" Private - Return the code string to end exporting an array literal. "

	^$] ,$)! !

! JuliaEncoder methodsFor: #literals !
escaped: string
	" Private - Return the code string to export string literal. "

	^string asJulia! !

! JuliaEncoder methodsFor: #export !
exportIndexEq: key
	" Returns the code to export setIndex expression to key. "

	^$[ ,key asJulia ,']=' ! !

! JuliaEncoder methodsFor: #export !
exportCommentOf: aClass on: aStream
	" Export the comment of aClass onto aStream. "

	aClass comment size > 0 ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: (self exportIndexEq: '$comment') ,(self escaped: aClass comment); cr
	].! !
 
! JuliaEncoder methodsFor: #export !
exportDefinitionOf: aClass on: aStream
	" Export definition of aClass onto aStream. "

	(self loadRequirementsOf: aClass) do: [:each|
		each isString
		ifTrue: [ aStream nextPutAll: each;cr ]
		ifFalse: [ aStream fileOutOn: aStream ].
	].
	aStream 
	    nextPutAll: 'addClass(smalltalk,"', aClass name, '",';
	    nextPutAll: (self smalltalkNameFor: aClass superclass);
	    nextPutAll: ','.
	self exportArrayLiteral: aClass instanceVariableNames on: aStream newLine: false.
	aStream
	    nextPutAll: ',"' ,aClass category, $";
	    nextPutAll: ')'; cr.
	aClass classVariables notNil ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: (self exportIndexEq: '$classVariableNames') ,(self escaped: aClass classVariablesString); cr
	].
	aClass poolDictionaries notNil ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: (self exportIndexEq: '$poolDictionaries').
	   self exportArrayLiteral: aClass poolDictionaries on: aStream newLine: true.
	].
	aClass propertyNames notEmpty ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: (self exportIndexEq: '$propertyNames').
	   self exportArrayLiteral: aClass propertyNames on: aStream newLine: true.
	].
	self exportCommentOf: aClass on: aStream.! !

! JuliaEncoder methodsFor: #export !
exportMetaDefinitionOf: aClass on: aStream
	" Export class side definition of aClass onto aStream. "

	aClass class instanceVariableNames notEmpty ifTrue: [
		aStream 
			nextPutAll: (self smalltalkNameFor: aClass class);
			nextPutAll: (self exportIndexEq: '$iVarNames').
		self exportArrayLiteral: aClass class instanceVariableNames on: aStream newLine: true.
	].

	aClass class propertyNames notEmpty ifTrue: [
		aStream 
			nextPutAll: (self smalltalkNameFor: aClass class);
			nextPutAll: (self exportIndexEq: '$propertyNames').
		self exportArrayLiteral: aClass class propertyNames on: aStream newLine: true.
	].! !

! JuliaEncoder methodsFor: #jl !
privateSourceFor: aMethod
	" Private - Return the native code implementing aMethod. "

	| code |
	(Smalltalk isRunningOn: self nativeLanguage) ifTrue: [
		code := aMethod basicAt: '$nativeSource' ifAbsent: [].
		code notNil ifTrue: [ ^code ].
	].

	code := [ self privateCodeFor: aMethod ] on: Exception do: [:ex|
		self note: aMethod.
		self privateNIYSourceFor: aMethod with: ex description
	].
	^'(
'	,code ,'
)'! !

! JuliaEncoder methodsFor: #niy !
privateNIYSourceFor: aMethod with: description
	" Private - Return the native code implementing aMethod. "

	^(self privateMethodDefinition: aMethod)
	,(self privateMethodBody: aMethod with: description)! !

! JuliaEncoder methodsFor: #niy !
privateMethodDefinition: aMethod
	" Private - Return the (Julia) method definition of aMethod. "

	| args |
	args := aMethod compiledSource stream upTo: $(; upTo: $).
	args notEmpty ifTrue: [
		args := (args asArrayOfSubstringsSeparatedBy: $,)
			inject: '' into: [:total :each| total ,',' ,each asJuliaName ].
	].
	^'function(self' ,args ,')'! !

! JuliaEncoder methodsFor: #niy !
privateMethodBody: aMethod with: description
	" Private - Return the (Julia) body of aMethod. "

	^'
	print("NIY:#privateMethodBody: ',aMethod,' - ',description,'")
	return self
end'! !

! JuliaEncoder methodsFor: #compiling !
privateCodeFor: aMethod
	" Private - Return the native code translated from sources of aMethod. "

	self compiler: aMethod classField compiler.
	^self compiler
		encoder: self;
		compile: aMethod source! !

! JuliaEncoder methodsFor: #generation !
emmitEnterMethod: aMethodNode
	" Private - Generate home context definitions, entering aMethodNode. "

"DEBUG"	self tab;nextPutAll: $# ,self currentMethodInfo; cr ! !

! JuliaEncoder methodsFor: #generation !
emmitFunctionStart: fName
	" Private - Emmit the code of function start definition. "

	self nextPutAll: #function.
	fName notEmpty ifTrue: [ | unescaped |
		unescaped := fName replace: #_bk_bk with: #_bk.
		self nextPutAll: $  ,unescaped
	] ! !

! JuliaEncoder methodsFor: #generation !
emmitFunctionBody: aBlock
	" Private - Emmit the code of function body evaluating aBlock. "

	self cr.
	aBlock evaluate.
	self cr; nextPutAll: #end ! !

! JuliaEncoder methodsFor: #generation !
newMethodFunctionName
	" Private - Return the function to call when installing a new method. "

	^#newMethod ! !

! JuliaEncoder methodsFor: #generation !
argumentNamed: aName ofType: aClass
	" Private - Return the code for argument with (abstract) type. "

	^aName ,#:: ,(self smalltalkTypeFor: aClass) ! !

! JuliaEncoder methodsFor: #generation !
emmitFunction: fName node: aNode withSelf: aBoolean arguments: argumentNames doing: aBlock
	" Private - Generate function for aNode. "

	| aStream |
	aStream := '' writeStream.
	aBoolean ifTrue: [
		aStream nextPutAll: (self argumentNamed: #self ofType: self currentClass).
		argumentNames notEmpty ifTrue: [ aStream nextPutAll: self argumentsSeparatorOp ].
	].
	argumentNames do: [:each | 
		self tempVariables add: each.
		aStream nextPutAll: (self localName: each) ]
	    separatedBy: [ aStream nextPutAll: self argumentsSeparatorOp ].
	self	emmitFunction: fName
		arguments: aStream contents
		doing: aBlock! !

! JuliaEncoder methodsFor: #generation !
emmitFunction: fName node: aNode arguments: argumentNames doing: aBlock
	" Private - Generate function for aNode. "

	^self	emmitFunction: fName node: aNode
		withSelf: false arguments: argumentNames
		doing: aBlock! !

! JuliaEncoder methodsFor: #generation !
emmitFunction: fName node: aNode doing: aBlock
	" Private - Generate function for aNode. "

	^self	emmitFunction: fName node: aNode
		withSelf: true arguments: aNode argumentNames
		doing: aBlock! !

! JuliaEncoder methodsFor: #visiting !
visitMethodNode: aNode
	" Visit a method node. "

	| fnStart fnEnd nativeSource |
	(Smalltalk isRunningOn: self nativeLanguage) ifFalse: [ ^self emmitMethod: aNode ].

	"return the code for method ready to be injected in the system"
	self nextPutAll: self newMethodFunctionName.
	self nextArgumentsPut: [
		self	nextPutAll: 'smalltalk,';
			nextPutAll: (self smalltalkNameFor: self currentClass);
			nextPutAll: ',"', aNode selector, '",'; cr.
		fnStart := stream position.
		self emmitMethod: aNode.
		fnEnd := stream position.
		self includeSourceCode ifTrue: [
			nativeSource := stream contents copyFrom: fnStart+1 to: fnEnd+1.
			self	cr; "warning: this cr is important for debugger"
				nextPutAll: ',' ,(self escaped: aNode source)
				;nextPutAll: ',' ,(self escaped: nativeSource)
		]
	] ! !

! JuliaEncoder methodsFor: #visiting !
visitDirectiveNode: aNode
	" Private - Visit a directive node. "

	(aNode respondsTo: #asNative:for:on:) ifTrue: [
		self nextPutAll: (aNode
			asNative: functionName
			for: self on: stream); cr.
		^self
	].

	self visit: aNode.
	self nextPutAll: self separatorOp; cr! !

! JuliaEncoder methodsFor: #results !
emmitedCode
	" Return the emmited code during compilation (or nil). "

	stream isNil ifTrue: [ ^nil ].
	(Smalltalk isRunningOn: self nativeLanguage) ifTrue: [
		"the code must be valid native code including return expression"
		^'return ' ,stream contents
	].
	^stream contents! !

! JuliaEncoder methodsFor: #constants !
returnOpStart
	" Private - Return the OP for starting return expressions. "

	^'return ' ! !

! JuliaEncoder methodsFor: #constants !
returnOpEnd
	" Private - Return the OP for ending return expressions. "

	^'' ! !

! JuliaEncoder methodsFor: #visiting !
visitValueNode: aNode
	" Private - Visit the node emmitting output. "

	self nextPutAll: aNode value asJulia! !

! JuliaEncoder methodsFor: #visiting !
visitNode: aNode index: index
	" Private - Generate code for indexed access. "

"ToDo:	((index isMemberOf: ValueNode) and: [ index hasString ]) ifTrue: [
		| selector |
		selector := self asSelector: index value.
		((selector includes: $ ) or: [ selector includes: $_ ]) ifFalse: [
			self nextPutAll: '.'; nextPutAll: selector.
			^self
		].
	].
"	self nextPutAll: '['.
	self visit: index.
	self nextPutAll: ']'.! !

! JuliaEncoder methodsFor: #encoding !
messageSendConnector
	" Private - Return the connector to use for sending a message. "

	^$. ! !

! JuliaEncoder methodsFor: #encoding !
functionCallConnector
	" Private - Return the connector to use for calling a function. "

	^$. ! !

! JuliaEncoder methodsFor: #accessing !
messageSendConnector: aNode
	" Private - Return the caracter to use for sending message. "

	aNode receiver messageSendConnector size = 0 ifTrue: [ ^'' ].
	(self isNativeCall: aNode) ifTrue: [ "calling an alien"
		(aNode selector startsWith: self callEscape) ifTrue: [
			^self functionCallConnector
		].
		^self messageSendConnector
	].
	^self messageSendConnector ! !

! JuliaEncoder methodsFor: #accessing !
asPrivateSelector: aSelector
	" Private - Validate and returns a valid native selector "

	| result |
	result := self asSelector: aSelector.
	(aSelector startsWith: self callEscape) ifTrue: [ "call to an alien starting with _"
		^result copyFrom: 2 to: result size
	].
	^result! !

! JuliaEncoder methodsFor: #visiting !
visitPrivateMessage: aNode
	" Private - Generate code for native message send."

	self nextPutAll: (self asPrivateSelector: aNode selector)! !

! JuliaEncoder methodsFor: #encoding !
isAPICall: aNode
	" Private - Return true if aNode is native API call."

	| cls |
	(self isNativeCall: aNode) ifFalse: [ ^false ].
	cls := self currentClass instanceClass.
	(cls respondsTo: #isAPICall:encoder:) ifFalse: [ ^false ].
	^cls isAPICall: aNode encoder: self! !

! JuliaEncoder methodsFor: #encoding !
encodeAPICall: aNode
	" Private - Generate code for native API call."

	self	visitReceiver: aNode receiver;
		nextPutAll: (self messageSendConnector: aNode);
		nextPutAll: (self asSelector: aNode selector);
		nextArgumentsPut: [ self visitSendArguments: aNode ] ! !

! JuliaEncoder methodsFor: #encoding !
encodeNormalSend: aNode
	" Private - Generate code for direct/inlined message send."

	| link |
	(self isAPICall: aNode) ifTrue: [
		^self encodeAPICall: aNode
	].
	link := self messageSendConnector: aNode.
	self visitPrivateMessage: aNode.
	self nextArgumentsPut: [
		self visitReceiver: aNode receiver.
		aNode arguments notEmpty ifTrue: [
			link notEmpty ifTrue: [ self nextPutAll: self argumentsSeparatorOp ].
			self visitSendArguments: aNode
		]
	] ! !

! JuliaEncoder methodsFor: #visiting !
visitReceiver: aNode
	" Private - Visit the node (as receiver) emmitting output. "

	aNode isAssignment ifTrue: [
		self nextArgumentsPut: [ self visitLocalAssignment: aNode ]
	] ifFalse: [ self visit: aNode ] ! !

! JuliaEncoder methodsFor: #private !
localName: aName
	" Private - Return a valid (native, local) name for aName. "

	(self pseudoVariables includes: aName) ifTrue: [ ^aName ].
	^self asSelector: aName! !

! JuliaEncoder methodsFor: #visiting !
visitGlobalReferenceNode: aNode

	| aName aClass lookupPools |
	aName := aNode value.
	aClass := self currentClass instanceClass.
	aClass classVariableNamesDo: [:each|
		each = aName ifTrue: [
			self
				nextPutAll: 'classVariableAt_(';
				nextPutAll: (self smalltalkNameFor: aClass);
				nextPutAll: ',"';
				nextPutAll: aName;
				nextPutAll: '")'.
			^self
		]
	].
	aClass sharedPoolsAndNamesDo: [:poolName :pool|
		(pool includesKey: aName) ifTrue: [
			self	nextPutAll: 'at_(';
				nextPutAll: 'poolDictionary_(smalltalk,"';
				nextPutAll: poolName;
				nextPutAll: '"),"';
				nextPutAll: aName;
				nextPutAll: '")'.
			^self
		].
	].
	self
		nextPutAll: 'smalltalk.';
		nextPutAll: aNode value.! !

! JuliaEncoder methodsFor: #visiting !
visitBlockNode: aNode

	self nextArgumentsPut: [
		self	emmitFunction: '' node: aNode
			arguments: aNode parameterNames
			doing: [ self visitNodesIn: aNode cr: true ]
	] ! !

! JuliaEncoder methodsFor: #encoding !
defineTemporary: aName
	" Private - Return the code for defining temporary variable with aName. "

	self nextPutAll: (self localName: aName), '=nil' ;cr ! !

! JuliaEncoder methodsFor: #visiting !
visitCascadeNode: aNode

	| target |
	target := #_s8_rec.
	(self tempVariables includes: target) ifFalse: [ self tempVariables add: target ].
	self compilerEvent: #aboutCascade: with: aNode.
	self nextArgumentsPut: [
		self cr.
		self emmitFunction: '' arguments: target doing: [ | index rec |
			index := 0.
			aNode nodes do: [:each |
			    index := index + 1.
			    index = aNode nodes size ifTrue: [ self nextPutAll: self returnOpStart ].
			    rec := VariableNode new.
			    rec value: target; start: each start end: each start.
			    each receiver: rec.
			    self visit: each.
			    index = aNode nodes size ifTrue: [
					self	compilerEvent: #returned:early:
						withArguments: (Array with: nil with: false).
					self nextPutAll: self returnOpEnd
			    	] ifFalse: [ self nextPutAll: self separatorOp; cr ]
			]
		].
		self cr
	]; nextArgumentsPut: [ self visit: aNode receiver ].
	self compilerEvent: #cascade: with: aNode.! !

! JuliaEncoder methodsFor: #visiting !
visitClassVariableAssignment: aClassVariableName in: aClass to: anExpressionNode

	self
		nextPutAll: 'classVariableAt_put_(';
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: ',"';
		nextPutAll: aClassVariableName; nextPutAll: '",'.
	self visit: anExpressionNode.
	self	nextPutAll: ')'! !

! JuliaEncoder methodsFor: #visiting !
visitPoolVariableAssignment: aName in: poolName to: anExpressionNode

	self	nextPutAll: 'at_put_(';
		nextPutAll: 'poolDictionary_(smalltalk,"';
		nextPutAll: poolName; nextPutAll: '"),"';
		nextPutAll: aName; nextPutAll: '",'.
	self visit: anExpressionNode.
	self	nextPutAll: ')'! !

! JuliaEncoder methodsFor: #visiting !
visitOperation: aNode
	" Private - aNode is an operation and selector is mapped to target operation. "

	self compilerEvent: #aboutOperation: with: aNode.
	self nextArgumentsPut: [ self visitReceiver: aNode receiver ].
	self nextPutAll: (self asPrivateSelector: aNode selector).
	self visitSendArguments: aNode.
	self compilerEvent: #operation: with: aNode.! !

! JuliaEncoder methodsFor: #encoding !
encodeSuperSend: aNode
	" Private - Generate code for super message send."

	| classType argsTypes |
	classType := #St ,self currentClass name.
	classType := classType replace: $  with: $_.
	argsTypes := aNode arguments isEmpty ifTrue: [ '' ]
		ifFalse: [ $, ,((aNode arguments collect: [:each| #Any ]) join: $,) ].
	self
		nextPutAll: 'Base.invoke('
			,(self asPrivateSelector: aNode selector)
			,',Tuple{supertype(' ,classType ,')' ,argsTypes ,'}'
			,',self'.
	aNode arguments notEmpty ifTrue: [
		self nextPutAll: self argumentsSeparatorOp.
		self visitSendArguments: aNode
	].
	self nextPutAll: ')' ! !

! JuliaEncoder methodsFor: #visiting !
visitReturnNode: aNode

	| theReturn |
	nestedBlocks > 0 ifTrue: [ earlyReturn := true ].
"OBSOLETE
	theReturn := [ self emmitReturn: [
		self visitNodesIn: aNode cr: true
	] ].
"	theReturn := [
		self withNodesIn: aNode do: [:each|
			self visitNode: each in: aNode cr: true
		] atEnd: [:last| | result |
			result := self prepareValue: last.
			self emmitReturn: [ self visitNode: result in: aNode cr: false ]
		]
	].
	earlyReturn ifFalse: [ ^theReturn value ].

	self nextArgumentsPut: [
		self emmitFunctionDoing: [
			self nextPutAll: '_s8Ret_.hasResult=true';cr.
			self nextPutAll: '_s8Ret_.result='.
			self nextArgumentsPut: [
				self emmitFunctionDoing: theReturn; cr
			]; nextArgumentsPut: [].
			self cr;nextPutAll: 'throw(_s8Ret_)'
		]
	]; nextArgumentsPut: [] ! !

! JuliaEncoder methodsFor: #generation !
generate: fName early: early return: code
	" Private - Generate code to catch early return of code expression in fName. "

	| prev target |
	early ifFalse: [
		self nextPutAll: code.
		^0
	].
	prev := stream position.
	target := #_s8Ret_.
	self nextPutAll: target ,'=nativeNew_(smalltalk.Object,nil)' ;cr.
	self nextPutAll: target ,'.name="stReturn" ' ;cr.
	self nextPutAll: #try ;cr.
	prev := stream position - prev.
	self nextPutAll: code ;cr.
	self nextPutAll: 'catch __ex' ;cr.
	self nextPutAll: '	if __ex == ' ,target ,' return ' ,target ,'.result end' ;cr.
	self nextPutAll: '	rethrow()#throw(__ex)' ;cr.
	self nextPutAll: #end.
	^prev ! !

! JuliaEncoder methodsFor: #prepare !
prepare: aMethodNode
	" Private - Prepare aMethodNode for traversal. "

	(super prepare: aMethodNode) isNil ifTrue: [ ^nil ].
	self prepareValuesIn: aMethodNode ! !

! JuliaEncoder methodsFor: #prepare !
prepareValuesIn: aMethodNode
	" Private - Rename locals when there is a collision with messages potentially sent.
	Messages in Julia are global functions, so a collision will mask a global (function) name.
	"

	| forbidden problems |
	forbidden := aMethodNode messages collect: [:msg| msg asSelector ].
	problems := aMethodNode values select: [:each|
		forbidden includes: each value
	].
	problems isEmpty ifTrue: [ ^self ].
	self rename: problems in: aMethodNode ! !

! JuliaEncoder methodsFor: #refactoring !
rename: values in: aMethodNode
	" Private - Rename node values in aMethodNode to avoid name clashes. "

	values do: [:each| | aName |
		aName := each value.
		self rename: each to: #jl8_ ,aName in: aMethodNode.
		"self note: 'Name collision ' ,aName
			,' renamed to ' ,each value
			,' @ ' ,aMethodNode selector asLiteral"
	].! !

! JuliaEncoder methodsFor: #refactoring !
rename: aNode to: aName in: aMethodNode
	" Private - Rename aNode in aMethodNode to avoid name clashes. "

	| old |
	old := aNode value.
	aNode value: aName.
	aMethodNode do: [:each|
		(each isVariable and: [ each value = old ]) ifTrue: [
			each value: aName
		].
	].! !

! JuliaEncoder methodsFor: #visiting !
visitSendNode: aNode
	" Private - Visit the node (message send node) emmitting output. "

	self prepareSendNode: aNode.
	^super visitSendNode: aNode! !

! JuliaEncoder methodsFor: #prepare !
prepareSendNode: aNode
	" Private - Prepare a Send node to be used as value.
	Return the node to be emmited.
	"

	self prepareValue: aNode receiver.
	aNode arguments do: [:each| self prepareValue: each ].
	^aNode ! !

! JuliaEncoder methodsFor: #prepare !
prepareValue: aNode
	" Private - Prepare aNode to be used as a return value.
	Return the node to be emmited.
	Note that code can be generated as side effect of preparation
	e.g. for asignments in return expression or arguments resolution
	"

	aNode isAssignment ifTrue: [
		^self prepareAssignment: aNode
	].
	aNode isSend ifTrue: [
		aNode isCascade ifTrue: [ ^aNode ].
		^self prepareSendNode: aNode
	].
	^aNode ! !

! JuliaEncoder methodsFor: #visiting !
visitAssignmentNode: aNode
	" Visit the assignment aNode.
	Note that aNode can be visited during preparation;
	 in that case an ID has been already assigned
	See implementation of #prepareAssignment:
	"

	aNode basicAt: #ID ifAbsent: [
		^super visitAssignmentNode: aNode
	].
	^self emmitAssignment: aNode ! !

! JuliaEncoder methodsFor: #prepare !
prepareAssignment: aNode
	" Private - Assignments are not valid expressions here,
	 so we must encode the assignment in advance
	 and mark the assignment to do not encode when pushed as value.
	"

	aNode basicAt: #ID ifAbsentPut: [
		self prepareValue: aNode right.
		self visitAssignmentNode: aNode; cr.
		self prepareValue: aNode left
	].
	^aNode ! !

! JuliaEncoder methodsFor: #assignments !
emmitAssignment: aNode

	| id |
	id := aNode basicAt: #ID ifAbsent: [].
	id notNil ifTrue: [ ^self visit: id ].
	^super emmitAssignment: aNode ! !

! JuliaEncoder methodsFor: #contexts !
sequenceExit: aNode
	" Private - A block sequence is exiting with aNode. "

	self prepareValue: aNode.
	^super sequenceExit: aNode ! !

! JuliaEncoder methodsFor: #optimization !
coreOptimizations
	" Private - Return the optimization set. "

	^#(core  #(
		#(class	'classOf_(smalltalk,%receiver%)' )
		#(isNil	'isNil_(smalltalk,%receiver%)' )
		#(notNil	'notNil_(smalltalk,%receiver%)' )
		#(=	'isObject_equalTo_(smalltalk,%receiver%,%arg1%)' )
		#(~=	'!!(isObject_equalTo_(smalltalk,%receiver%,%arg1%))' )
		#(==	'%receiver%===%arg1%' )
		#(~~	'!!(%receiver%===%arg1%)' )
		#(===	'%receiver%===%arg1%' )
		#(||	'object_or_(smalltalk,%receiver%,%arg1%)' )
		#(,	'object_comma_(smalltalk,%receiver%,%arg1%)' )
		#(basicAt:	'basic_at_or_(smalltalk,%receiver%,%arg1%,nil)' )
		#(basicAt:ifAbsent: 'basic_at_ifAbsent_(smalltalk,%receiver%,%arg1%,%arg2%)' )
		#(basicAt:put: 'basic_at_put_(smalltalk,%receiver%,%arg1%,%arg2%)' )
		#(ifUndefined: 'ifUndefined_do_(smalltalk,%receiver%)' )
	))! !

! JuliaEncoder methodsFor: #optimization !
s8Optimizations
	" Private - Return the optimization set. "

	^#(s8	#(
		core
		#(ifTrue: hasBlockNoArguments: #(
			#(ifTrue: noClient: 'if %receiver%%cr%%seq1%%cr%end' )
			#(ifTrue: 'if_true_false_(smalltalk,%receiver%,%arg1%,nil)' )
		))
		#(ifFalse: hasBlockNoArguments: #(
			#(ifFalse: noClient: 'if !!(%receiver%)%cr%%seq1%%cr%end' )
			#(ifFalse: 'if_true_false_(smalltalk,%receiver%,nil,%arg1%)' )
		))
		#(ifTrue:ifFalse: hasBlockNoArguments: #(
			#(ifTrue:ifFalse: noClient: 'if %receiver%%cr%%seq1%%cr%else%cr%%seq2%%cr%end' )
			#(ifTrue:ifFalse: 'if_true_false_(smalltalk,%receiver%,%arg1%,%arg2%)' )
		))
		#(ifFalse:ifTrue: hasBlockNoArguments: #(
			#(ifFalse:ifTrue: noClient: 'if %receiver%%cr%%seq2%%cr%else%cr%%seq1%%cr%end' )
			#(ifFalse:ifTrue: 'if_true_false_(smalltalk,%receiver%,%arg2%,%arg1%)' )
		))
		#(whileTrue receiverIsBlockNoArguments: #(
			#(whileTrue noClient: 'while %receiver%()%cr%end' )
			#(whileTrue 'while_do_(smalltalk,%receiver%,nil)' )
		))
		#(whileFalse receiverIsBlockNoArguments: #(
			#(whileFalse noClient: 'while !!%receiver%()%cr%end' )
			#(whileFalse 'whileNot_do_(smalltalk,%receiver%,nil)' )
		))
		#(whileTrue: receiverIsBlockNoArguments: #(
			#(whileTrue: noClient: 'while %receiver%()%cr%%seq1%%cr%end' )
			#(whileTrue: 'while_do_(smalltalk,%receiver%,%arg1%)' )
		))
		#(whileFalse: receiverIsBlockNoArguments: #(
			#(whileFalse: noClient: 'while !!%receiver%()%cr%%seq1%%cr%end' )
			#(whileFalse: 'whileNot_do_(smalltalk,%receiver%,%arg1%)' )
		))
		#(timesRepeat: 'times_repeat_(smalltalk,%receiver%,%arg1%)' )
		#(orNil 'orNil_(smalltalk,%receiver%)' )
		#(isKindOf: 'is_kindOf_(smalltalk,%receiver%,%arg1%)' )
		#(respondsTo: 'responds_to_(smalltalk,%receiver%,%arg1%)' )
		#(ifNil: 'if_nil_notNil_(smalltalk,%receiver%,%arg1%,nil)' )
		#(ifNotNil: 'if_nil_notNil_(smalltalk,%receiver%,nil,%arg1%)' )
		#(ifNil:ifNotNil: 'if_nil_notNil_(smalltalk,%receiver%,%arg1%,%arg2%)' )
		#(ifNotNil:ifNil: 'if_nil_notNil_(smalltalk,%receiver%,%arg2%,%arg1%)' )
	))! !

! JuliaEncoder methodsFor: #optimization !
magnitudeOptimizations
	" Private - Return the optimization set. "

	^#(magnitude #(
		#(> 	'(%receiver% > %arg1%)' )
		#(< 	'(%receiver% < %arg1%)' )
		#(>=	'(%receiver% >= %arg1%)' )
		#(<=	'(%receiver% <= %arg1%)' )
	))! !

! JuliaEncoder methodsFor: #optimization !
mathOptimizations
	" Private - Return the optimization set. "

	^#(math #(
		#(max:	'math_max_(smalltalk,%receiver%,%arg1%)' )
		#(min:	'math_min_(smalltalk,%receiver%,%arg1%)' )
		#(mustBeNumber	'mustBeNumber_(smalltalk,%receiver%)' )
		#(+	'(%receiver% + %arg1%)' )
		#(#-	'(%receiver% - %arg1%)' )
		#(*	'(%receiver% * %arg1%)' )
		#(/	'(%receiver% / %arg1%)' )
	))! !

! JuliaEncoder methodsFor: #optimization !
printingOptimizations
	" Private - Return the optimization set. "

	^#(printing #(
		#(toString	'to_String(smalltalk,%receiver%)' )
		#(printString	'printString_(smalltalk,%receiver%)' )
		#(asString	'asString_(smalltalk,%receiver%)' )
		#(asCharacter	'asCharacter_(smalltalk,%receiver%)' )
	))! !

! JuliaEncoder methodsFor: #optimization !
booleanOptimizations
	" Private - Return the optimization set. "

	^#(boolean #(
		#(isBoolean 'typeof(%receiver%)==Bool' )
		#(and: hasBlockNoArguments: 'boolean_and_(smalltalk,%receiver%,%arg1%)' )
		#(or: hasBlockNoArguments: 'boolean_or_(smalltalk,%receiver%,%arg1%)' )
		#(not 'booleanNot_(smalltalk,%receiver%)' )
	))! !

! JuliaEncoder methodsFor: #optimization !
valueOptimizations
	" Private - Return the optimization set. "

	^#(value #(
		#(value 'evaluate(%receiver%)' )
		#(value: 'evaluateWith_(%receiver%,%arg1%)' )
		#(value:value: 'evaluateWithArguments_(%receiver%,with_with_(smalltalk.Array,%arg1%,%arg2%))' )
		#(value:value:value: 'evaluateWithArguments_(%receiver%,with_with_with_(smalltalk.Array,%arg1%,%arg2%,%arg3%))' )
		#(value:value:value:value: 'evaluateWithArguments_(%receiver%,with_with_with_with_(smalltalk.Array,[%arg1%,%arg2%,%arg3%,%arg4%))' )
	))! !

! JuliaEncoder methodsFor: #defaults !
defaultKnownOptimizations
	" Private - Return the default knownOptimizations of the receiver. "

	^(Array	with: self coreOptimizations)
		add: self s8Optimizations;
"		add: self magnitudeOptimizations;
		add: self mathOptimizations;"
		add: self booleanOptimizations;
		add: self printingOptimizations;
		add: self valueOptimizations;
		yourself! !

! JuliaEncoder methodsFor: #defaults !
defaultOptimizations
	" Private - Return the default optimizations of the receiver. "

	^#( s8 magnitude "math" boolean "value" printing )! !

! JuliaEncoder class methodsFor: #inlines !
initializeKernelInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self kernelInlines into: pool ! !

! JuliaEncoder class methodsFor: #inlines !
kernelInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(Object #(
#(halt		" '@enter yourself(self)') "	'# @bp #uncomment for #halt support (using Debugger) ')
#(toString	'self')
#(toString:	'string(anObject)')
#(isObject:	'true && return notNil(klass(anObject))')
#(typeOf:	'typeof(anObject)')
#(callableOf:	'true && return aBlock')
#(basicDelete:	'basic_delete_(smalltalk,self,aString)')
#(basicPerform:withArguments:	#('t=self ,handle(anArray)...
fn=eval(Symbol(jsSelector))
#true && return fn(t...)
true && return Base.invokelatest(fn,t...)' 'error(self,"Should not happen")'))
#(asJSONObject	'newHash(smalltalk)')
#(print:	'println(aString)
true && return self')
#(fileIn: 'true && return fileIn(smalltalk,fileName)')
#(fileInJS: 'true && return fileInJS(smalltalk,fileName)')
#(isFile:olderThan: 'true && return isFileOlderThan(smalltalk,fileName,referenceFileName)')
#(printingWith:do: 'true && return printBlock_do_(smalltalk,printing,aBlock)')
))
#(Smalltalk #(
#(includesKey:	'newHash(smalltalk)')
#(eval:		'true && return eval(Meta.parse(anExpression))')
))
#('Smalltalk class' #(
#(global	'globalScope(smalltalk)')
#(gc	'GC.gc()
true && return true')
))
#(Behavior #(
#(basicNew	'basicNewInstance_(self)')
#(fn		'invalidMessage_(self,"fn")')
#(prototype	'invalidMessage_(self,"prototype")')
#(privateMethods	'true && return self["\$methods"]')
))
#(CompiledMethod #(
#(asClosure	'true && return nativeNew_(smalltalk.BlockClosure,self)')
))
#(Boolean #(
#(not 'self == false')
#(= 'true && return self===anObject')
#(ifTrue:ifFalse: 'if self==true
	return value(aBlock)
	else return value(anotherBlock)
end')
))
#(Number #(
#(= '	isa(anObject,Bool) && return false
	isa(anObject,Number) && self==anObject
	true && return false')
#(> 'self > aNumber')
#(>= 'self >= aNumber')
#(< 'self < aNumber')
#(<= 'self <= aNumber')
#(+ 'self + aNumber')
#(- 'self - aNumber')
#(* 'self * aNumber')
#(/ 'self / aNumber')
#(sin 'Base.sin(self)')
#(cos 'Base.cos(self)')
#(tan 'Base.tan(self)')
#(arcCos 'Base.acos(self)')
#(arcSin 'Base.asin(self)')
#(arcTan 'Base.atan(self)')
#(atan2: 'Base.atan2(self,aNumber)')
#(sqrt 'Base.sqrt(self)')
#(<< 'self << aNumber')
#(>> 'self >> aNumber')
#(raisedTo: 'self ^ aNumber')
#(max: 'Base.max(self,aNumber)')
#(min: 'Base.min(self,aNumber)')
#(rounded 'Base.round(Int,self*1)')
#(truncated 'Base.trunc(Int,self*1)')
#(timesRepeat: 'for i = 1:self fn(i) end')
#(to:do: 'for i = self:aNumber fn(i) end')
#(to:by:do: 'for i = self:increment:aNumber fn(i) end')
#(printString 'tostring(self)')
#(toString: 'string(self,base=radix)')
#(isNumber 'isa(self,Bool) && return false
	if self!!==self return false else return true end')
#(clearInterval 'if true notImplementedYet_(self,"Number>>#clearInterval") end')
#(clearTimeout 'if true notImplementedYet_(self,"Number>>#clearTimeout") end')
#(int32 'if true return Int32(self) end')
#(int64 'if true return Int64(self) end')
#(float32 'if true return Float32(self) end')
#(float64 'if true return Float64(self) end')
))
#('Number class' #(
#(pi 'Base.MathConstants.pi')
))
#(BlockClosure #(
#(whileTrue: 'while self() fn() end')
#(whileFalse: 'while !!self() fn() end')
#(try:catch: 'try return fn()
catch ex
	sex=basicNew(smalltalk.Exception)
	nativeException_(sex,ex)
	return ab(sex)
end')
#(valueWithTimeout: '@async begin
	msInSecond = 1_000.0::Float64
	sleep(aNumber / msInSecond)
	value(self)
    end')
#(valueWithInterval: 'notImplementedYet_(self, "BlockClosure>>valueWithInterval:")')
#(ensure: 'try return evaluate(self) finally aBlock() end')
#(ifCurtailed: 'try
	result = value(self)
	curtailed = false
finally
	if curtailed value(exceptionBlock) end
end
return result')
#(argumentCount 'block_argumentCount(smalltalk,self)')
#(value 'block_valueWithPossibleArguments_(smalltalk,self,[])')
#(value: 'block_valueWithPossibleArguments_(smalltalk,self,[anObject])')
#(value:value: 'block_valueWithPossibleArguments_(smalltalk,self,[arg1,arg2])')
#(value:value:value: 'block_valueWithPossibleArguments_(smalltalk,self,[arg1,arg2,arg3])')
#(value:value:value:value: 'block_valueWithPossibleArguments_(smalltalk,self,[arg1,arg2,arg3,arg4])')
#(valueWithPossibleArguments: 'block_valueWithPossibleArguments_(smalltalk,self,array["@handle"])')
))
#(UndefinedObject  #(
#(asJSONObject 'nothing')
))
#(Collection  #(
#(wrap: 'isNil_(smalltalk,anObject) && return nothing #alias(smalltalk.UndefinedObject)')
#(unwrap: 'isNil_(smalltalk,anObject) && return nil
anObject == alias(smalltalk.UndefinedObject) && return nil')
))
#(HashedSet #(
#(initializeHash 'newHash(smalltalk)')
#(includes: '(self["@hash"][cookie])!!=nil')
#(privateAdd: 'self')
#(privateRemove: 'self["@hash"][cookie] = nil')
))
#(SequenceableCollection #(
#(indexOf:ifAbsent: 'i=0
for x in handle(self)
	i = i + 1
	isObject_equalTo_(smalltalk,anObject,unwrap_(self,x)) && return i
end
return value(aBlock)')
#(do: 'for x in handle(self) aBlock(unwrap_(self,x)) end')
#(detect:ifNone: 'for x in handle(self)
	aBlock(unwrap_(self,x)) && return x
end
return value(anotherBlock)')
#(reversed 'result=basicNew(species(self))
result["@handle"] = reverse(handle(self))
true && return result')
#(isEmpty 'size(self)==0')
#(reverseDo:  'i=size(self)
while i>0
	fn(at_(self,i))
	i = i - 1
end')
#(includes: #('cookie=wrap_(self,anObject)
true && return cookie in handle(self)' 'if true end'))
#(occurrencesOf: 'i = 0
for x in handle(self)
	if isObject_equalTo_(smalltalk,anObject,unwrap_(self,x))
		i = i + 1
	end
end
true && return i')
))
#('String class' #(
#(fromString: 'true && return tostring(aString)')
#(new	'""')
#(cr 	'"\r"')
#(lf 	'"\n"')
#(crlf 	'"\r\n"')
#(space	'" "')
#(tab	'"\t"')
#(zero	'string(Char(0))')
#(fromCharCode: 'true && return string(Char(anArrayOrNumber))')
))
#(String #(
#(=	'true && return self==anObject')
#(size	'length(self)')
#(at:	'string(self[anIndex])')
#(do: 	'for c in self fn(string(c)) end')
#(withIndexDo: 'for i = 1:size(self) fn(string(self[i]),i) end')
#(detect:ifNone: '
	for ch in self
		c = string(ch)
		fn(c) && return c
	end
	true && return value(anotherBlock)')
#(isLiteral 'length(self)==0 && return false
isBinarySelector_(smalltalk.String,self) && return true
true && return isNil_(smalltalk,match(r"[^a-zA-Z0-9_:]",self))')
#(occurrencesOf: 'isString(aString) && return count(aString,self)
	true && return 0')
#(startsWith: 'isString(aString) && return startswith(self, aString)')
#(escaped 'true && return HTTP.URIs.escapeuri(self)')
#(unescaped 'true && return HTTP.URIs.unescapeuri(self)')
#(asEncodedURI 'true && return notImplementedYet_(self, "String>>#asEncodedURI")')
#(asDecodedURI 'true && return notImplementedYet_(self, "String>>#asDecodedURI")')
#(, '	isNil_(smalltalk,aCollection) && return self
	true && return self * asString_(smalltalk,aCollection)')
#(copyWith: 'self * asString_(smalltalk,anObject)')
#(copyFrom:to: 'string_copyFrom_to_(smalltalk,self,anIndex,anotherIndex)')
#(includes: '!!isString(anObject) && return false
	isEmpty(anObject) && true
	true && return findfirst(anObject,self) !!= nothing ')
#(reversed 'true && return reverse(self)')
#(indexOf: '!!isString(aString) && return 0
	isEmpty(aString) && return 1
	index = findfirst(aString,self)
	if isnothing(index) return 0
	else return index[1] end')
#(indexOf:startingAt: '
	!!isString(aString) && return 0
	isEmpty(aString) && return position
	s = self[position:end]
	index = findfirst(aString,s)
	if isnothing(index)
		return 0
		else return index[1]+position-1 end')
#(lastIndexOf: '!!isString(aString) && return 0
	isEmpty(aString) && return size(self)+1
	index = findlast(aString,self)
	if isnothing(index)
		return 0
		else return index[1] end')
#(lastIndexOf:startingAt: '
	!!isString(aString) && return 0
	isEmpty(aString) && return position
	s = self[position:end]
	index = findlast(aString,s)
	if isnothing(index)
		return 0
		else return index[1]+position-1 end')
#(asNativeName 'asJuliaName(self)')
#(asPrivateSelector 'result = self
	result = replace(result,":"=>"_")
	result = replace(result,"+"=>"_plus")
	result = replace(result,"-"=>"_minus")
	result = replace(result,"*"=>"_star")
	result = replace(result,"/"=>"_slash")
	result = replace(result,"|"=>"_bar")
	result = replace(result,"&"=>"_amp")
	result = replace(result,">"=>"_gt")
	result = replace(result,"<"=>"_lt")
	result = replace(result,"="=>"_eq")
	result = replace(result,","=>"_comma")
	result = replace(result,"@"=>"_at")
	result = replace(result,"~"=>"_no")
	result = replace(result,"\\"=>"_bk")')
#(asNative: 'true && return asJulia(self) #was: slashEncoded(self)
')
#(replace:with: '
if size(aString)==2 && first(aString)=="\\"
	return replace(self,second(aString) => anotherString)
end
true && return replace(self,aString => anotherString)')
#(replaceRegexp:with: 'notImplementedYet_(self, "String>>#replaceRegexp:with:")')
#(tokenize: 'true && return newArray(smalltalk,string.(split(self,string(aString))))')
#(match: 'notImplementedYet_(self, "String>>#match:")')
#(isLetter 'size(self)==1 && isletter(self[1])')
#(isDigit 'size(self)==1 && isdigit(self[1])')
#(isUpperCase 'if true
	if size(self)==0 return false end
	for c in self
		if !!(isuppercase(c)) return false end
	end
	return true
end')
#(isLowerCase 'if true
	if size(self)==0 return false end
	for c in self
		if !!(islowercase(c)) return false end
	end
	return true
end')
#(asUppercase 'uppercase(self)')
#(asLowercase 'lowercase(self)')
#(asNumber 'string_toNumber(smalltalk,self)')
#(> 'self > aString')
#(< 'self < aString')
#(>= 'self >= aString')
#(<= 'self <= aString')
#(isCharacter 'size(self) == 1')
#(asciiValue 'Int(self[1])')
#(isSeparator 'size(self)==1 && isspace(self[1])')
#(fileContents 'true && return fileContents(smalltalk,self)')
))
#('Array class' #(
#(basicNew: 'newArrayOfSize(smalltalk,size)')
#(with: 'self')
#(with:with: 'self')
#(with:with:with: 'self')
#(with:with:with:with: 'self')
#(withAll: '
	result = asArray(aCollection)
	if aCollection==result
		return copy(result)
		else return result
	end')
#(slice: 'notImplementedYet_(self, "Array class>>#slice:")')
))
#(Array #(
#(asJSONObject #('self' 'true && return handle(result)'))
#(json #('self' 'true && return handle(result)'))
#(size 'length(handleOr(self,[]))')
#(at: 'true && return unwrap_(self, handle(self)[anIndex])')
#(at:ifAbsent: 'array = handle(self)
	anIndex < 1 && return value(aBlock)
	anIndex > size(self) && return value(aBlock)
	true && return unwrap_(self, array[anIndex])')
#(at:put: 'anIndex - 1 == size(self) && return add_(self,anObject)
	o = wrap_(self, anObject)
	arrayAdapt(smalltalk,self,o)
	handle(self)[anIndex] = o
	true && return anObject')
#(do: 'for x in handle(self) aBlock(unwrap_(self,x)) end')
#(addAll: '
	array = asArray(aCollection)
	size(array) = 0 && return aCollection
	Base.append!!(handle(self),handle(array))
	arrayPack(smalltalk,self)')
#(shallowCopy '
	result = new(species(self))
	addAll_(result,self)')
#(copyWithAll: '
	result = copy(self)
	addAll_(result,aCollection)
	true && return result')
#(, '	result = copy(self)
	addAll_(result,aCollection)
	true && return result')
#(join: 'join(handle(self),aString)')
#(sort: '
	if isNil_(smalltalk,fn)
		fn = function(a,b) return _lt(to_String(smalltalk,a),to_String(smalltalk,b)) end
	end
	Base.sort!!(handle(self),lt=fn)')
#(remove:ifAbsent: '
	found = true
	function done()
		found = false
		return aBlock.value()
	end
	i = indexOf_ifAbsent_(self,anObject,done)
	if found
		removeIndex_(self,i)
		return anObject
	end
	return i
')
#(add: 'Base.push!!(handle(self),wrap_(self, anObject))')
#(addFirst: 'Base.pushfirst!!(handle(self),wrap_(self, anObject))')
#(addLast: 'Base.push!!(handle(self),wrap_(self, anObject))')
#(removeAll 'Base.empty!!(handle(self))')
#(removeFirst 'unwrap_(self,Base.popfirst!!(handle(self)))')
#(removeIndex: 'Base.deleteat!!(handle(self),index)')
#(removeLast 'unwrap_(self,Base.pop!!(handle(self)))')
#(atAllPut: 'o=wrap_(self, anObject)
	arrayAdapt(smalltalk,self,o)
	Base.fill!!(handle(self),o)')
#(first '	value=self[1]
	value!!=nil && return unwrap_(self, value)')
#(second '
	value=self[2]
	value!!=nil && return unwrap_(self, value)')
#(third '	value=self[3]
	value!!=nil && return unwrap_(self, value)')
#(fourth '	
	value=self[4]
	value!!=nil && return unwrap_(self, value)')
#(fifth '	value=self[5]
	value!!=nil && return unwrap_(self, value)')
#(sixth '	value=self[6]
	value!!=nil && return unwrap_(self, value)')
))
#('RegularExpression class' #(
#(fromString:flag: 'notImplementedYet_(self, "RegularExpression class>>#fromString:flag:")')
))
#(Exception #(
#(captureStack 'self["@stack"]=stacktrace(smalltalk)')
#(signal 'error(self)')
))
#(ComposedAssociations #(
#(hashObject 'newHash(smalltalk)')
))
#('KeyedCollection class' #(
#(@ 'if isObject_(self,jsObject) return jsObject else return _at_at(self,jsObject) end')
))
#(KeyedCollection #(
#(newHash 'newHash(smalltalk)')
))
#('Dictionary class' #(
#(basicNew 'true && return new(smalltalk.PoolDictionary)')
))
#(PoolDictionary #(
#(at:put: 'true && return pool_at_put_(smalltalk,self, aKey, aValue)')
#(removeKey:ifAbsent: 'pool_removeKey_(smalltalk,self, aKey)')
#(at:ifAbsent: 'true && return pool_at_ifAbsent_(smalltalk,self, aKey, aBlock)')
#(includesKey: 'pool_includesKey_(smalltalk,self, aKey)')
#(keys 'true && return pool_keys(smalltalk,self)')
))
#(ClassBuilder #(
#(addSubclassOf:named:instanceVariableNames: '
	addClass(smalltalk, aString, aClass, aCollection)
	true && return smalltalk[aString]')
))
#(StringStream #(
#(peek '	i = position(self) + 1
	contents = self["@collection"]
	size(contents) < i && return nil
	true && return at_(contents,i)')
#(peekForAll: 'i = position(self) + 1
	contents = self["@collection"]
	!!startswith(contents[i:end],string(aCollection)) && return false')
))
#(Random #(
#(next 'true && return Base.rand()')
))
#('DateTime class' #(
#(new: 'true && return nativeNew_(self,datesSupport(smalltalk).DateTime(anObject,DateFormat("Y/m/d H:M:S")))')
#(today 'true && return nativeNew_(self,datesSupport(smalltalk).now(datesSupport(smalltalk).UTC))')
#(now 'true && return today(self)')
#(msNow 'true && return round(Int64, datesSupport(smalltalk).time() * 1000)')
))
#(DateTime #(
#(asJSONObject 'if true return handle(self) end')
#(asString 'if true return asLocaleString(self) end')
#(< 'handle(self) < handle(aDateTime)')
#(> 'handle(self) > handle(aDateTime)')
#(<= 'handle(self) <= handle(aDateTime)')
#(>= 'handle(self) >= handle(aDateTime)')
#(- 'nativeNew_(species(self),handle(self) - handle(aDateTime))')
#(+ 'nativeNew_(species(self),handle(self) + handle(aDateTime))')
#(isValid 'true')
#(year 'datesSupport(smalltalk).year(handle(self))')
#(year: 'handle_(self,datesSupport(smalltalk).DateTime(aNumber,month(self),day(self)))')
#(month 'datesSupport(smalltalk).month(handle(self))')
#(month: 'handle_(self,datesSupport(smalltalk).DateTime(year(self),aNumber,day(self)))')
#(day 'datesSupport(smalltalk).day(handle(self))')
#(day: 'handle_(self,datesSupport(smalltalk).DateTime(year(self),month(self),aNumber))')
#(dayOfMonth 'day(self)')
#(dayOfMonth: 'day_(self,aNumber)')
#(asLocaleString 'handle(self).strftime("%%c")')
#(asDateString 'handle(self).strftime("%%x")')
#(asTimeString 'handle(self).strftime("%%X")')
#(time 'self.notImplementedYet_("time")')
#(time: 'self.notImplementedYet_("time:")')
#(hours: 'self.notImplementedYet_("hours:")')
#(minutes: 'self.notImplementedYet_("minutes:")')
#(seconds: 'self.notImplementedYet_("seconds:")')
#(milliseconds: 'self.invalidMessage("milliseconds:")')
#(hours 'datesSupport(smalltalk).hour(handle(self))')
#(minutes 'datesSupport(smalltalk).minute(handle(self))')
#(seconds 'datesSupport(smalltalk).second(handle(self))')
#(milliseconds 'self.notImplementedYet_("milliseconds")')
))
	) ! !

! JuliaEncoder class methodsFor: #inlines !
initializeNativeInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self nativeInlines into: pool ! !

! JuliaEncoder class methodsFor: #inlines !
nativeInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#('NativeObject class' #(
#(undefined 'nil')
#(codeToNumber '" "')
#(newInstanceOf:withAll: 'if true return fn(handle(handlesArray)...) end')
))
#(NativeObject #(
#(jsObject 'invalidMessage_(self, "jsObject")')
#(ownPropertyNames 'h = handle(self)
isa(h,StObject) && return keys(_at_at(smalltalk.PoolDictionary,h.dict))
true && return newArray(smalltalk,fieldnames(typeof(h)))')
))
#(NativeHandle #(
#(hasOwnProperty: 'rawget(self["@handle"],aName)!!=nil')
#(delete: 'self["@handle"][aName]=nil')
))
#('NativeLibraryObject class' #(
#(basicLoad: 'require(url)')
))
#(NativeLibraryObject  #(
#(newHandleOf: 'result = aConstructor()')
#(newHandleOf:with: 'result = aConstructor(arg1)')
#(newHandleOf:with:with: 'result = aConstructor(arg1,arg2)')
#(newHandleOf:with:with:with: 'result = aConstructor(arg1,arg2,arg3)')
#(newHandleOf:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4)')
#(newHandleOf:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5)')
#(newHandleOf:with:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5,arg6)')
#(newHandleOf:with:with:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5,arg6,arg7)')
#(newHandleOf:with:with:with:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8)')
))
#(Array #(
#(addAllJulia: 'invalidMessage_(self, "addAllJulia:")')
))
#('Array class' #(
#(@@ 'notNil_(smalltalk,hArray) && return newArray(smalltalk,hArray)')
#(@ 'addAll_(result,newArray(smalltalk,natives))')
))
#('Boolean class' #(
#(@ '	aBOOL==true && return true
	aBOOL==1 && return true
	aBOOL==false && return false
	(aBOOL+0==1) && return true')
))
	) ! !

! JuliaEncoder class methodsFor: #inlines !
initializeToolsInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self toolsInlines into: pool ! !

! JuliaEncoder class methodsFor: #inlines !
toolsInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(NativeReader #(
#(initializeObject 'newHash(smalltalk)')
))
#(String  #(
#(asComment #('"#= "' '" =#"' '"# "'))
))
#(MethodReferences #(
#(dot		'"."')
#(localDef	'""')
#(superSendPatternOf:	'superJuliaSendPatternOf_(self, aSelector)')
#(normalSendStarter:	'if true return "" end')
#(compiledSourceStart	'"("')
))
#(Class #(
#(hackDNU 'notImplementedYet_(self, "hackDNU")')
))
	) ! !

! JuliaEncoder class methodsFor: #inlines !
initializeSnapshotInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self snapshotInlines into: pool ! !

! JuliaEncoder class methodsFor: #inlines !
snapshotInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(Smalltalk #(
#(exit	'quit()')
))
#(Snapshot #(
#(bootHeader '""')
#(bootFooter '""')
#(launchExpression '"launchImage(smalltalk)"')
))
	) ! !

! JuliaEncoder class methodsFor: #inlines !
initializeOtherInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self otherInlines into: pool ! !

! JuliaEncoder class methodsFor: #inlines !
otherInlines
	" Private - Return the inlines for methods in the category. "

	^#(
"#(ClassName #(
#(selector '...')
))"
	) ! !

"------------------------------------------------------------------------------
	Extension to Kernel to support Julia encoding
 ------------------------------------------------------------------------------"!

! String class methodsFor: #Julia-jl !
reservedJuliaWords
	" Private - Return the reserved names of Julia language. "

	^#(
	#true #false
	baremodule begin break catch const 
	continue do else elseif end export false 
	finally for function global if import let 
	local macro module quote return 
	struct try using while abstract
	mutable primitive type
	)! !

! String class methodsFor: #Julia-jl !
forbiddenJuliaSelectors
	" Private - Return the platform forbidden selectors. "

	^#( smalltalk nil methods
	self true false
	"Base." map show
	)! !

! String methodsFor: #Julia-jl !
asJuliaName
	" Private - Return the receiver as a valid Julia name. "

	self first = $_ ifTrue: [ ^self ].
	(String reservedJuliaWords includes: self) ifTrue: [ ^'__',self ].
	^self! !

! String methodsFor: #Julia-jl !
asJuliaSelector
	" Private - Return the receiver as a valid native selector. "

	(String forbiddenJuliaSelectors includes: self) ifTrue: [ ^$_ ,self ].
	(self startsWith: $#) ifTrue: [
		^self upTo: $: startingAt: 2
	].
	^self asPrivateSelector asJuliaName ! !

! Object methodsFor: #Julia-jl !
asJulia
	" Return the native representation of the receiver.
	The default implementation return the receiver as a String.
	This method must be refined by subclasses implementing native literal representation.
	"

	^self asString! !

! Number methodsFor: #Julia-jl !
asJulia
	" Return the native representation of the receiver. "

	^self printString ! !

! String methodsFor: 'Julia-jl' !
asJulia
	" Return the native representation of the receiver. "

	| encoded bar s |
	s := 36 asCharacter.
	bar := 92 asCharacter.
	encoded := self.
	encoded := encoded replace: bar with: bar ,bar.
	encoded := encoded replace: bar ,s with: bar ,s.
	encoded := encoded replace: String lf with: bar ,$n.
	encoded := encoded replace: String cr with: bar ,$r.
	encoded := encoded replace: String tab with: bar ,$t.
	encoded := encoded replace: $" with: bar ,$".
	^$" ,encoded ,$"! !

! Array methodsFor: #Julia-jl !
asJulia
	" Return the native representation of the receiver. "

	^'newArray' ,$( ,#smalltalk ,$, ,$[
		,((self collect: [:each | each asJulia ]) join: ', ')
	,$] ,$) ! !

! Set methodsFor: #Julia-jl !
asJulia
	" Return the native representation of the receiver. "

	^'newContents_(smalltalk.',self class name ,','
		, contents asJulia
	,')'! !

! Dictionary methodsFor: #Julia-jl !
asJulia
	" Return the native representation of the receiver. "

	^'fromKeys_and_(smalltalk.',self class name ,','
		, keys asArray asJulia
		,',' ,values asJulia
	,')'! !

! PoolDictionary methodsFor: #Julia-jl !
asJulia
	" Return the native representation of the receiver. "

	^'fromTuples_(smalltalk.',self class name ,','
		, self tuples asJulia
	,')'! !


" Custom conversion methods
Note: We implement __tostring delegating to #toString implementors"
#toString implementors do: [:mth| | source |
	source := '__tostring
	" Return the printable representation of the receiver. "

	^self toString'.
	(mth classField
		methodsFor: #Julia-printing)
		scanFrom: source stream
]!

! Array methodsFor: #Julia-printing !
__tostring
	" Return the printable representation of the receiver. "

	^self asLiteral! !

"------------------------------------------------------------------------------
	Native handlers for Julia
 ------------------------------------------------------------------------------"!

! NativeObject class methodsFor: 'Julia-builders' !
buildJuliaFunctions: specs
	" OBSOLETE - Generate code from specs as raw functions implemented by instance handle. "

	self buildNativeFunctions: specs! !

! NativeObject class methodsFor: 'Julia-builders' !
buildJuliaStaticFunctions: specs
	" OBSOLETE - Generate code from specs as raw functions implemented by class handle. "

	self buildNativeStaticFunctions: specs! !

NativeHandle
	subclass: #Julia
	category: #Julia!
Julia comment: '
	Implementation of wrappers to native Julia objects.
	@2021 Alejandro F. Reimondo
	aleReimondo@smalltalking.net'!

! Julia class methodsFor: #instantiation !
supportsLanguage: language
	" Private - Return true if instances of the receiver implement native wrappers for language. "

	^language = #jl! !

! Julia class methodsFor: #calling !
nativeCall: jlFunction args: args kwargs: kwargs
	" Private - Return the native result of calling jlFunction with args and kwargs. "

	^{''jl' jlFunction(args...;kwargs...)'}! !

! Julia class methodsFor: #calling !
call: jlFunction args: anArray kwargs: keywords
	" Return the result of calling jlFunction with positional arguments and keywords arguments. "

	| kwargs |
	kwargs := keywords isArray
		ifTrue: [Kwargs fromTuples: keywords]
		ifFalse: [keywords].
	^self	nativeCall: jlFunction
		args:  (# #Tuple: anArray handle)
		kwargs: kwargs json! !

! Julia methodsFor: #calling !
call: jlFunction args: anArray kwargs: keywords
	" Return the result of calling jlFunction with positional arguments and keywords arguments. "

	((self isObject: jlFunction) and: [jlFunction isString]) ifTrue: [
		^self	call: self @ jlFunction
			args: anArray kwargs: keywords
	].
	^self class call: jlFunction args: anArray kwargs: keywords! !

! Julia class methodsFor: #calling !
call: selector on: receiver args: anArray kwargs: keywords
	" Return the native result of calling selector on receiver with positional arguments and keywords arguments. "

	^self	call: (receiver basicAt: selector asSelector) handle
		args: anArray kwargs: keywords! !

! Julia class methodsFor: #calling !
send: selector to: receiver withArguments: anArray keywords: keywords
	" Return the result of sending message selector on receiver with positional arguments and keywords arguments. "

	| hResult |
	hResult := self
		call: selector on: receiver
		args: anArray kwargs: keywords.
	^#{smalltalk} asObject: hResult! !

! Julia class methodsFor: #streamming !
on: streamable do: aBlock
	" Stream on the contents of streamable. "

	| generator |
	generator := [ (self isObject: streamable)
		ifTrue: [streamable json]
		ifFalse: [streamable] ].
{''jl'	evaluate(generator) do element evaluateWith_(aBlock,element) end'}! !

! Julia class methodsFor: #iterating !
in: iterable do: aBlock
	" Stream on the contents of streamable. "

	| hIterator |
	hIterator := (self isObject: iterable)
		ifTrue: [iterable json]
		ifFalse: [iterable].
{''jl'	for element in hIterator evaluateWith_(aBlock,element) end'}! !

! Julia methodsFor: #macros !
expand: macro
	" Return the result of expanding the macrocall. "

	^self eval: #@ ,macro ! !

! Julia methodsFor: #accessing !
propertyNames
	" Return the property names of the receiver. "

	^self ownPropertyNames! !

! Julia methodsFor: #accessing !
@ propertyName
	" Return the property at propertyName of the receiver. "

	^self handle basicAt: propertyName! !

! Julia class methodsFor: #accessing !
global
	" Return an instance wrapping global. "

	^self @ Smalltalk global! !

! Julia class methodsFor: #accessing !
version
	" Return the host version. "

	^self global @ #_VERSION! !

! Julia class methodsFor: #accessing !
arg
	" Return the main arguments (anArray of strings). "

	^Array @ (self global @ #arg)! !

! Julia class methodsFor: #modules !
require: aName
	" Return an instance of the receiver obtained from loading module aName. (or nil) "

	^self @ (# #require: aName)! !

! Julia class methodsFor: #import !
installPackage: key
	" Install the package key. "

	self smalltalk eval: 'using Pkg; Pkg.add("',key,'")'! !

! Julia class methodsFor: #import !
using: key
	" Private - Start using key and return an instance of the receiver.
	If the package is not present, try to load it also installing if required.
	"

	| h st code |
	st := self smalltalk.
	[ h := st eval: key ] on: Exception do: [:ex| ].
	h notNil ifTrue: [ ^self @ h ].
	code := 'using ',key.
	[	st eval: code
	] on: Exception do: [:ex|
		self installPackage: key.
		st eval: code
	].
	^self @ (st eval: key)! !

! Julia methodsFor: #iteration !
do: aBlock
	" Evaluates aBlock on each element of the receiver. "

{''js'	self.invalidMessage_("do:")'
jl'	for item in handle(self) aBlock(item) end'}.! !

! Julia methodsFor: #iteration !
pairsDo: aBlock
	" Evaluates aBlock on each key/value pair of the receiver. "

{''js'	self.invalidMessage_("pairsDo:")'
jl'	invalidMessage_(self,"pairsDo:")'}.! !

! Julia methodsFor: #iteration !
keysDo: aBlock
	" Evaluates aBlock on each key of the receiver. "

{''js'	self.invalidMessage_("keysDo:")'
jl'	invalidMessage_(self,"keysDo:")'}.! !

! Julia methodsFor: #iteration !
valuesDo: aBlock
	" Evaluates aBlock on each value of the receiver. "

{''js'	self.invalidMessage_("valuesDo:")'
jl'	invalidMessage_(self,"valuesDo:")'}.! !

! Julia methodsFor: #converting !
asDictionary
	" Convert the receiver to a(Pool)Dictionary. "

	^PoolDictionary @@ self handle! !

Julia subclass: #JuliaObject!
JuliaObject comment: '
	Implementation of wrappers to native Julia objects.
	@2021 Alejandro F. Reimondo
	aleReimondo@smalltalking.net'!

! JuliaObject class methodsFor: 'Builders-private' !
stMessageHeader: msgHeader
	" Private - Return a message header string with nice smalltalk looking. "

	(msgHeader includes: $:) ifTrue: [ ^msgHeader ].
	^msgHeader asSmalltalkName: false! !

! JuliaObject class methodsFor: #builders !
buildConstants: constants
	" Build the methods for constants. "

	^self build: #Constants -> constants doing: [
		constants isString ifTrue: [ | stream |
			stream := constants stream.
			[stream atEnd] whileFalse: [
				self buildConstant: stream nextLine trimBlanks
			]
		] ifFalse: [ constants do: [:each| self buildConstant: each ] ]
	]! !

! JuliaObject class methodsFor: #builders !
buildConstant: spec
	" Private - Build the accessor for constant spec. "

	| selector |
	spec isEmpty ifTrue: [ ^self ].
	selector := (spec fromLast: $.) asSmalltalkName: false.
	self class
		implement: selector
		comment: 'Return the constant value for ' ,spec
		as: $^ ,(self codeToGetConstant: spec) category: #constants! !

! JuliaObject class methodsFor: #private !
library
	" Private - Returns the library of the receiver.
	A new instance is returned by default.
	This method can be refined by subclasses (e.g. implementing singleton). "

	^self new ! !

! JuliaObject class methodsFor: #builders !
codeToGetConstant: spec
	" Private - Returns the code to get the constant value for spec. "

	spec isArray ifTrue: [
		^spec last ,' @ (' ,(self codeToGetConstant: spec first) ,$)
	].
	^'self library @ ' ,spec asLiteral ! !

! JuliaObject class methodsFor: #builders !
constants
	" Returns the constants implemented by receiver's library. "

	^self constantsIn: self library! !

! JuliaObject class methodsFor: #builders !
constantsIn: aJuliaObject
	" Returns the constants implemented by aJuliaObject. "

	^((aJuliaObject keys
		select: [:key| #(Number String) includes: (self typeOf: aJuliaObject@key) ])
		reject: [:key| key startsWith: $_ ]) sorted! !

! JuliaObject class methodsFor: #encoding !
isAPICall: aNode encoder: anEncoder
	" Return true if code generated by aNode should be encoded as standard native call. "

	^true! !

ComposedAssociations
	subclass: #Kwargs
	instanceVariableNames: ''
	category: #Julia !
Kwargs comment: 'Keyword arguments container'! !

! Kwargs class methodsFor: #instantiation !
fromTuples: tuples
	" Return an instance of the receiver. "

	^tuples inject: self new into: [:total :each|
		total ,(each first -> each last)
	]! !

! Kwargs methodsFor: #private !
hashObject
	" Private - Return an empty hash object. "

	| h |
	{''jl' h = Dict{Symbol,Any}()'}.
	^PoolDictionary @@ h! !

! Kwargs methodsFor: #converting !
asJSONObject
	" Return the receiver as a native object. "

	^super asJSONObject instVarAt: #hash! !

! Kwargs methodsFor: #private !
install: key value: value into: jsObject
	" Private - Install a key/value pair into jsObject. "

	^jsObject
		at: (self toString: key)
		put: ((self isObject: value)
			ifTrue: [ value json ]
			ifFalse: [ value ])! !

! Snapshot class methodsFor: #Julia-precompile !
precompilesDo: aBlock
	" Evaluates aBlock with each precompile information of the current system. "

	| fns all |
	all := JuliaSet new.
	fns := JuliaSet new.
	Smalltalk allMethodsDo: [:mth| fns add: mth fn ].
	fns do: [:fun| | mths |
		mths := Array @@ (# #methods: fun).
		mths do: [:mth| | ops |
			ops := Array @@ (mth basicAt: #specializations).
			ops do: [:spec| all add: spec ].
		]
	].
	aBlock notNil ifTrue: [ all do: aBlock ].
	^all! !

! Snapshot class methodsFor: #Julia-precompile !
precompileDictionary
	" Private - Returns a dictionary with precompile information of the current system. "

	| result |
	result := PoolDictionary new.
	self precompilesDo: [:each| | spec |
		spec := '' ,(each basicAt: #specTypes).
		(spec includes: 'var"') ifFalse: [ | fnName argTypes |
			spec := spec readStream.
			fnName := spec upTo: 'typeof('; upTo: $).
			fnName trimBlanks notEmpty ifTrue: [
				spec next.
				argTypes := (spec upToEnd upToLast: $}) trimBlanks.
				(result at: fnName ifAbsentPut: [#()])
					add: argTypes
			]
		].
	].
	^result! !

! Snapshot class methodsFor: #Julia-precompile !
precompiles
	" Returns the code to save precompile information of the current system. "

	| output dict |
	output := '' stream.
	output nextPutAll: '#See implementors of #precompiles'.
	dict := self precompileDictionary.
	dict keys sorted do: [:fnName| | argTypes |
		(fnName includes: $.) ifFalse: [ "dump only local messages"
			(dict at: fnName) do: [:each|
				argTypes := each.
				(argTypes asArrayOfSubstringsSeparatedBy: $,) do: [:part| | arg |
					arg := part trimBlanks.
					((arg startsWith: #S8)
					or: [#(String Any) includes: arg])
					ifFalse: [ argTypes := '' ].
				].
				argTypes notEmpty ifTrue: [
					output	cr;
						nextPutAll: 'Base.precompile('; nextPutAll: fnName;
						nextPutAll: ',('; nextPutAll: argTypes; nextPutAll: ',))'
				]
			].
		]
	].
	^output contents! !

"------------------------------------------------------------------------------
	Development Utilities & tools for Julia environment
 ------------------------------------------------------------------------------"!

! Smalltalk methodsFor: #Julia-printing !
printBlock: printBlock do: aBlock
	" Private - Return the result of evaluating aBlock, redirecting #print: to printBlock.
	Note: #print: message will be routed at s8 level to printBlock.
	The native print() function is unmodified and will be evaluated as usual during aBlock.
	"

	| old result |
	[	old := Object basicAt: #printBlock.
		Object basicAt: #printBlock put: printBlock.
		result := aBlock value.
	] ensure: [ Object basicAt: #printBlock put: old ].
	^result! !

! MethodReferences methodsFor: #Tools-patterns !
superJuliaSendPatternOf: aSelector
	" Private - Return the search pattern for senders (to super) of aSelector. "

	| encoder |
	encoder := JuliaEncoder new.
	^'Base.invoke('
		,(encoder asPrivateSelector: aSelector)
		,',Tuple{supertype('! !

! MethodReferences methodsFor: #Tools-messages !
jlNextMessageIn: stream
	" Private - Return the next sent encoded in native source stream. "

	| maerts fn |
	maerts := (stream upTo: $() reversed stream.
	fn := ''.
	[maerts atEnd] whileFalse: [ | c |
		c := maerts next.
		(c isAlphaNumeric or: [ '_$' includes: c ]) ifTrue: [
			fn := fn ,c
		] ifFalse: [ maerts upToEnd ]
	].
	^fn reversed asNativeSelector! !