" Adaptions to run S8 on Lua execution machinery.
Copyright (C) 2018-today Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
In case of questions and bug reports please join us at S8 Facebook group
	https://www.facebook.com/groups/s8smalltalk/
"!

"------------------------------------------------------------------------------
	Lua code generation
 ------------------------------------------------------------------------------"!
NativeEncoder
	subclass: #LuaEncoder
	instanceVariableNames: ''
	category: #Lua !

! LuaEncoder class methodsFor: #instantiation !
supportsLanguage: aLanguage
	" Private - Return true if instances of the receiver can encode to aLanguage. "

	^aLanguage = #lua ! !

! LuaEncoder methodsFor: #accessing !
nativeLanguage
	" Return the nativeLanguage of the receiver. "

	^#lua ! !

! LuaEncoder methodsFor: #encoding !
nativeUndefined
	" Private - Return the code for undefined. "

	^#nil ! !

! LuaEncoder methodsFor: #encoding !
nativeNameOfMetaClass: aClass
	" Private - Return the native name for metaclass aClass. "

	^aClass instanceClass name ,'["$klass"]' ! !

! LuaEncoder methodsFor: #accessing !
callEscape
	" Private - Return the selector sequence to escape OO message send selector, and produce code for function call.
	Note that an alien selector starting with two $$ are understood as a function call to function starting with $
	"

	^'#_'! !

! LuaEncoder methodsFor: #inlines !
missingNativeCodeFor: aNode as: lang
	" Private - Return the native language code for missing expression (or nil). "

	^'(function() self:invalidMessage_("' ,compiler selector ,'") end)()'! !

! LuaEncoder methodsFor: #private !
canExportMethod: aMethod
	" Private - Return true if the receiver can/want to export method aMethod. "

	^true! !

! LuaEncoder methodsFor: #converting !
asSelector: aSelector
	" Return the native representation of aSelector. "

	^aSelector asLuaSelector! !

! LuaEncoder methodsFor: #private !
nativeSourceFor: aMethod
	" Private - Return the code of native function implementing aMethod. "

	^(self canExportMethod: aMethod) ifTrue: [
		self privateSourceFor: aMethod
	]! !

! LuaEncoder methodsFor: #export !
export: aClass method: aMethod on: aStream
	" Export a method definition of aClass onto aStream. "

	| fName selectorName bar source |
	source := self nativeSourceFor: aMethod.
	source isNil ifTrue: [ ^false ].
	aMethod isNil ifTrue: [ self error: 'Method is missing' ].
	bar := $".
	fName := aMethod selector.
	fName first isBackSlash ifTrue: [
		selectorName := self escaped: aMethod selector.
		fName := self escaped: (self asSelector: fName).
	] ifFalse: [
		selectorName := bar ,aMethod selector, bar.
		(self asSelector: fName) = (aMethod selector replace: $: with: $_) ifTrue: [
			fName := 0
		] ifFalse: [
			fName := bar ,(self asSelector: fName) ,bar
		].
	].
	aStream
		nextPutAll: 'smalltalk:bind(' ,(self smalltalkNameFor: aClass)
			,',' ,selectorName ,',' ,fName ,'
,'			,source.
	(aMethod source notNil and: [ aMethod source notEmpty ]) ifTrue: [
	    aStream cr; "warning: this cr is important for debugger"
		nextPutAll: ',' ,(self escaped: aMethod category)
			,',' ,(self escaped: aMethod source)
			,',' ,(self escaped: source).
	].
	((aMethod respondsTo: #stamp) and: [ aMethod stamp notNil ]) ifTrue: [
		aStream nextPutAll: ',' ,(self escaped: aMethod stamp)
	].
	aStream nextPutAll: ')
'.	^true! !

! LuaEncoder methodsFor: #arrays !
arrayLiteralStart
	" Private - Return the code string to start exporting an array literal. "

	^'smalltalk:newArray' ,$( ,${! !

! LuaEncoder methodsFor: #arrays !
arrayLiteralEnd
	" Private - Return the code string to end exporting an array literal. "

	^$} ,$)! !

! LuaEncoder methodsFor: #literals !
escaped: string
	" Private - Return the code string to export string literal. "

	^string asLua! !

! LuaEncoder methodsFor: #export !
exportCommentOf: aClass on: aStream
	" Export the comment of aClass onto aStream. "

	aClass comment size > 0 ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: '["$comment"]=' ,(self escaped: aClass comment); cr
	]! !

! LuaEncoder methodsFor: #export !
exportDefinitionOf: aClass on: aStream
	" Export definition of aClass onto aStream. "

	(self loadRequirementsOf: aClass) do: [:each|
		each isString
		ifTrue: [ aStream nextPutAll: each;cr ]
		ifFalse: [ aStream fileOutOn: aStream ].
	].
	aStream 
	    nextPutAll: 'smalltalk:addClass("', aClass name, '",';
	    nextPutAll: (self smalltalkNameFor: aClass superclass);
	    nextPutAll: ','.
	self exportArrayLiteral: aClass instanceVariableNames on: aStream newLine: false.
	aStream
	    nextPutAll: ',"' ,aClass category, $";
	    nextPutAll: ')'; cr.
	aClass classVariables notNil ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: '["$classVariableNames"]=' ,(self escaped: aClass classVariablesString); cr
	].
	aClass poolDictionaries notNil ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: '["$poolDictionaries"]='.
	   self exportArrayLiteral: aClass poolDictionaries on: aStream newLine: true.
	].
	aClass propertyNames notEmpty ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: '["$propertyNames"]='.
	   self exportArrayLiteral: aClass propertyNames on: aStream newLine: true.
	].
	self exportCommentOf: aClass on: aStream.! !

! LuaEncoder methodsFor: #export !
exportMetaDefinitionOf: aClass on: aStream
	" Export class side definition of aClass onto aStream. "

	aClass class instanceVariableNames notEmpty ifTrue: [
		aStream 
			nextPutAll: (self smalltalkNameFor: aClass class);
			nextPutAll: '["$iVarNames"]='.
		self exportArrayLiteral: aClass class instanceVariableNames on: aStream newLine: true.
	].

	aClass class propertyNames notEmpty ifTrue: [
		aStream 
			nextPutAll: (self smalltalkNameFor: aClass class);
			nextPutAll: '["$propertyNames"]='.
		self exportArrayLiteral: aClass class propertyNames on: aStream newLine: true.
	].! !

! LuaEncoder methodsFor: #lua !
privateSourceFor: aMethod
	" Private - Return the native code implementing aMethod. "

	| code |
	(Smalltalk isRunningOn: self nativeLanguage) ifTrue: [
		code := aMethod basicAt: '$nativeSource' ifAbsent: [].
		code notNil ifTrue: [ ^code ].
	].

	code := [ self privateCodeFor: aMethod ] on: Exception do: [:ex|
		self note: aMethod.
		self privateNIYSourceFor: aMethod with: ex description
	].
	^'(
'	,code ,'
)'! !

! LuaEncoder methodsFor: #niy !
privateNIYSourceFor: aMethod with: description
	" Private - Return the native code implementing aMethod. "

	^(self privateMethodDefinition: aMethod)
	,(self privateMethodBody: aMethod with: description)! !

! LuaEncoder methodsFor: #niy !
privateMethodDefinition: aMethod
	" Private - Return the (Lua) method definition of aMethod. "

	| args |
	args := aMethod compiledSource stream upTo: $(; upTo: $).
	args notEmpty ifTrue: [
		args := (args asArrayOfSubstringsSeparatedBy: $,)
			inject: '' into: [:total :each| total ,',' ,each asLuaName ].
	].
	^'function(self' ,args ,')'! !

! LuaEncoder methodsFor: #niy !
privateMethodBody: aMethod with: description
	" Private - Return the (Lua) body of aMethod. "

	^'
	print("NIY:#privateMethodBody: ',aMethod,' - ',description,'")
	return self
end'! !

! LuaEncoder methodsFor: #compiling !
privateCodeFor: aMethod
	" Private - Return the native code translated from sources of aMethod. "

	self compiler: aMethod classField compiler.
	^self compiler
		encoder: self;
		compile: aMethod source! !

! LuaEncoder methodsFor: #generation !
emmitEnterMethod: aMethodNode
	" Private - Generate home context definitions, entering aMethodNode. "! !

! LuaEncoder methodsFor: #generation !
emmitFunctionBody: aBlock
	" Private - Emmit the code of function body evaluating aBlock. "

	self cr.
	aBlock evaluate.
	self cr; nextPutAll: #end ! !

! LuaEncoder methodsFor: #generation !
newMethodFunctionName
	" Private - Return the function to call when installing a new method. "

	^'smalltalk:newMethod'! !

! LuaEncoder methodsFor: #generation !
emmitFunction: fName node: aNode withSelf: aBoolean arguments: argumentNames doing: aBlock
	" Private - Generate function for aNode. "

	| aStream |
	aStream := '' writeStream.
	aBoolean ifTrue: [
		aStream nextPutAll: #self.
		argumentNames notEmpty ifTrue: [ aStream nextPutAll: self argumentsSeparatorOp ].
	].
	argumentNames do: [:each | 
		self tempVariables add: each.
		aStream nextPutAll: (self localName: each) ]
	    separatedBy: [ aStream nextPutAll: self argumentsSeparatorOp ].
	self	emmitFunction: fName
		arguments: aStream contents
		doing: aBlock! !

! LuaEncoder methodsFor: #generation !
emmitFunction: fName node: aNode arguments: argumentNames doing: aBlock
	" Private - Generate function for aNode. "

	^self	emmitFunction: fName node: aNode
		withSelf: false arguments: argumentNames
		doing: aBlock! !

! LuaEncoder methodsFor: #generation !
emmitFunction: fName node: aNode doing: aBlock
	" Private - Generate function for aNode. "

	^self	emmitFunction: fName node: aNode
		withSelf: true arguments: aNode argumentNames
		doing: aBlock! !

! LuaEncoder methodsFor: #visiting !
visitMethodNode: aNode
	" Visit a method node. "

	| fnStart fnEnd nativeSource |
	(Smalltalk isRunningOn: self nativeLanguage) ifFalse: [ ^self emmitMethod: aNode ].

	"return the code for method ready to be injected in the system"
	self nextPutAll: self newMethodFunctionName.
	self nextArgumentsPut: [
		self
			nextPutAll: (self smalltalkNameFor: self currentClass);
			nextPutAll: ',"', aNode selector, '",'; cr.
		fnStart := stream position.
		self emmitMethod: aNode.
		fnEnd := stream position.
		self includeSourceCode ifTrue: [
			nativeSource := stream contents copyFrom: fnStart+1 to: fnEnd+1.
			self	cr; "warning: this cr is important for debugger"
				nextPutAll: ',' ,(self escaped: aNode source)
				;nextPutAll: ',' ,(self escaped: nativeSource)
		]
	] ! !

! LuaEncoder methodsFor: #visiting !
visitDirectiveNode: aNode
	" Private - Visit a directive node. "

	(aNode respondsTo: #asNative:for:on:) ifTrue: [
		self nextPutAll: (aNode
			asNative: functionName
			for: self on: stream); cr.
		^self
	].

	self visit: aNode.
	self nextPutAll: self separatorOp; cr! !

! LuaEncoder methodsFor: #results !
emmitedCode
	" Return the emmited code during compilation (or nil). "

	stream isNil ifTrue: [ ^nil ].
	(Smalltalk isRunningOn: self nativeLanguage) ifTrue: [
		"the code must be valid native code including return expression"
		^'return ' ,stream contents
	].
	^stream contents! !

! LuaEncoder methodsFor: #constants !
returnOpStart
	" Private - Return the OP for starting return expressions. "

	^'return (' ! !

! LuaEncoder methodsFor: #constants !
returnOpEnd
	" Private - Return the OP for ending return expressions. "

	^')' ! !

! LuaEncoder methodsFor: #visiting !
visitValueNode: aNode
	" Private - Visit the node emmitting output. "

	self nextPutAll: aNode value asLua! !

! LuaEncoder methodsFor: #visiting !
visitNode: aNode index: index
	" Private - Generate code for indexed access. "

"ToDo:	((index isMemberOf: ValueNode) and: [ index hasString ]) ifTrue: [
		| selector |
		selector := self asSelector: index value.
		((selector includes: $ ) or: [ selector includes: $_ ]) ifFalse: [
			self nextPutAll: '.'; nextPutAll: selector.
			^self
		].
	].
"	self nextPutAll: '['.
	self visit: index.
	self nextPutAll: ']'.! !

! LuaEncoder methodsFor: #encoding !
propertyAccessConnector
	" Private - Return the connector to use for accessing a native property. "

	^$. ! !

! LuaEncoder methodsFor: #encoding !
messageSendConnector
	" Private - Return the connector to use for sending a message. "

	^$: ! !

! LuaEncoder methodsFor: #encoding !
functionCallConnector
	" Private - Return the connector to use for calling a function. "

	^$. ! !

! LuaEncoder methodsFor: #encoding !
messageSendConnector: aNode
	" Private - Return the connector to use for sending message. "

	aNode receiver messageSendConnector size = 0 ifTrue: [ ^'' ].
	(self isNativeCall: aNode) ifTrue: [ "calling an alien"
		(aNode selector startsWith: self callEscape) ifTrue: [
			^self functionCallConnector
		].
		^self messageSendConnector
	].
	^self messageSendConnector ! !

! LuaEncoder methodsFor: #encoding !
asPrivateSelector: aSelector
	" Private - Validate and returns a valid native selector "

	| result |
	result := self asSelector: aSelector.
	(aSelector startsWith: self callEscape) ifTrue: [ "call to an alien starting with _"
		^result copyFrom: 2 to: result size
	].
	^result! !

! LuaEncoder methodsFor: #visiting !
visitPrivateMessage: aNode
	" Private - Generate code for native message send."

	self
		nextPutAll: (self messageSendConnector: aNode);
		nextPutAll: (self asPrivateSelector: aNode selector)! !

! LuaEncoder methodsFor: #encoding !
encodeNormalSend: aNode
	" Private - Generate code for direct/inlined message send."

	self visitReceiver: aNode receiver.
	self visitPrivateMessage: aNode.
	self nextArgumentsPut: [ self visitSendArguments: aNode ] ! !

! LuaEncoder methodsFor: #visiting !
visitReceiver: aNode
	" Private - Visit the node (as receiver) emmitting output. "

	self nextArgumentsPut: [ | last |
		last := stream position.
		aNode isAssignment
			ifTrue: [ self visitLocalAssignment: aNode ]
			ifFalse: [ self visit: aNode ].
		stream position = last ifTrue: [
			stream removeLastInput.
			^self
		]
	]! !

! LuaEncoder methodsFor: #private !
localName: aName
	" Private - Return a valid (native, local) name for aName. "

	(self pseudoVariables includes: aName) ifTrue: [ ^aName ].
	^self asSelector: aName! !

! LuaEncoder methodsFor: #visiting !
visitGlobalReferenceNode: aNode

	| aName aClass lookupPools |
	aName := aNode value.
	aClass := self currentClass instanceClass.
	aClass classVariableNamesDo: [:each|
		each = aName ifTrue: [
			self
				nextPutAll: (self smalltalkNameFor: aClass);
				nextPutAll: ':classVariableAt_("';
				nextPutAll: aName;
				nextPutAll: '")'.
			^self
		]
	].
	aClass sharedPoolsAndNamesDo: [:poolName :pool|
		(pool includesKey: aName) ifTrue: [
			self
				nextPutAll: '(smalltalk:poolDictionary_("';
				nextPutAll: poolName;
				nextPutAll: '")):at_("';
				nextPutAll: aName;
				nextPutAll: '")'.
			^self
		].
	].
	self
		nextPutAll: 'smalltalk.';
		nextPutAll: aNode value.! !

! LuaEncoder methodsFor: #visiting !
visitBlockNode: aNode

	self nextArgumentsPut: [
		self	emmitFunction: '' node: aNode
			arguments: aNode parameterNames
			doing: [ aNode nodes do: [:each | self visit: each ] ]
	] ! !

! LuaEncoder methodsFor: #encoding !
defineTemporary: aName
	" Private - Return the code for defining temporary variable with aName. "

	self nextPutAll: 'local ', (self localName: aName), '=nil' ;cr ! !

! LuaEncoder methodsFor: #visiting !
visitCascadeNode: aNode

	| target |
	target := #_s8_rec.
	(self tempVariables includes: target) ifFalse: [ self tempVariables add: target ].
	self compilerEvent: #aboutCascade: with: aNode.
	self nextArgumentsPut: [
		self cr.
		self emmitFunction: '' arguments: target doing: [ | index rec |
			index := 0. self tab.
			aNode nodes do: [:each |
			    index := index + 1.
			    index = aNode nodes size ifTrue: [ self nextPutAll: self returnOpStart ].
			    rec := VariableNode new.
			    rec value: target; start: each start end: each start.
			    each receiver: rec.
			    self visit: each.
			    index = aNode nodes size ifTrue: [
					self	compilerEvent: #returned:early:
						withArguments: (Array with: nil with: false).
					self nextPutAll: self returnOpEnd
			    	] ifFalse: [ self nextPutAll: self separatorOp; cr;tab ]
			]
		].
		self cr
	]; nextArgumentsPut: [ self visit: aNode receiver ].
	self compilerEvent: #cascade: with: aNode.! !

! LuaEncoder methodsFor: #generation !
emmitAssignment: aNode
	" Private - Emmit the code for assignment. "

	self nextArgumentsPut: [ | tab |
		tab := String tab.
		self nextPutAll: 'function()'; cr.
		self tab;nextPutAll: 'local _s8_value='.
		self nextArgumentsPut: [ self visit: aNode right ].
		self cr; tab.
		self visit: aNode left.
		self	nextPutAll: '=_s8_value' "; '";
			 cr; tab;
			nextPutAll: 'return _s8_value'; cr.
		self nextPutAll: #end
	]; nextArgumentsPut: [] ! !

! LuaEncoder methodsFor: #visiting !
visitClassVariableAssignment: aClassVariableName in: aClass to: anExpressionNode

	self
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: ':classVariableAt_put_("';
		nextPutAll: aClassVariableName; nextPutAll: '",'.
	self visit: anExpressionNode.
	self	nextPutAll: ')'! !

! LuaEncoder methodsFor: #visiting !
visitPoolVariableAssignment: aName in: poolName to: anExpressionNode

	self
		nextPutAll: 'smalltalk:poolDictionary_("';
		nextPutAll: poolName; nextPutAll: '"):at_put_("';
		nextPutAll: aName; nextPutAll: '",'.
	self visit: anExpressionNode.
	self	nextPutAll: ')'! !

! LuaEncoder methodsFor: #visiting !
visitOperation: aNode
	" Private - aNode is an operation and selector is mapped to target operation. "

	self compilerEvent: #aboutOperation: with: aNode.
	self nextArgumentsPut: [ self visitReceiver: aNode receiver ].
	self nextPutAll: (self asPrivateSelector: aNode selector).
	self visitSendArguments: aNode.
	self compilerEvent: #operation: with: aNode.! !

! LuaEncoder methodsFor: #encoding !
encodeSuperSend: aNode
	" Private - Generate code for super message send."

	self
		nextPutAll: 'smalltalk:superImplementor('
			,(self smalltalkNameFor: self currentClass)
			,',"',(self asPrivateSelector: aNode selector)
			,'")(self'.
	aNode arguments notEmpty ifTrue: [
		self nextPutAll: self argumentsSeparatorOp.
		self visitSendArguments: aNode
	].
	self nextPutAll: ')'. ! !

! LuaEncoder methodsFor: #visiting !
visitReturnNode: aNode

	| theReturn |
	nestedBlocks > 0 ifTrue: [ earlyReturn := true ].
	theReturn := [ self emmitReturn: [
		aNode nodes do: [:each | self visit: each ]
	] ].
	earlyReturn ifFalse: [ ^theReturn value ].

	self nextArgumentsPut: [
		self emmitFunctionDoing: [
			self nextPutAll: '_s8Ret_.hasResult=true';cr.
			self nextPutAll: '_s8Ret_.result='.
			self nextArgumentsPut: [
				self emmitFunctionDoing: theReturn; cr
			]; nextArgumentsPut: [].
			self cr;nextPutAll: 'error(_s8Ret_,0)'
		]
	]; nextArgumentsPut: [] ! !

! LuaEncoder methodsFor: #generation !
generate: fName early: early return: code
	" Private - Generate code to catch early return of code expression in fName. "

	| prev |
	early ifFalse: [ self nextPutAll: code. ^0 ].
	prev := '
local _s8Ret_={["name"]="stReturn"}
local _s8_f = function()
'.
	self	nextPutAll: prev;
		nextPutAll: code;
		nextPutAll: '
end
local _s8_status, _s8_result = pcall(_s8_f)
if _s8_status then return _s8_result
else
    if (_s8Ret_.hasResult) then
      '.	self emmitReturn: '_s8Ret_.result' early: true.
	self nextPutAll: '
    end
end
error(_s8_result,0)
'.
	^prev size! !

! LuaEncoder methodsFor: #optimization !
coreOptimizations
	" Private - Return the optimization set. "

	^#(core  #(
		#(class	'%receiver%["$klass"]' )
		#(isNil	'%:isNil_%(%receiver%)' )
		#(notNil	'%:notNil_%(%receiver%)' )
		#(=	'%:isObject_equalTo_%(%receiver%,%arg1%)' )
		#(~=	'(not %:isObject_equalTo_%(%receiver%,%arg1%))' )
		#(==	'(%receiver%==%arg1%)' )
		#(~~	'(%receiver%~=%arg1%)' )
		#(===	'(%receiver% == %arg1%)' )
		#(||	'(%receiver% or %arg1%)' )
		#(,	'%:object_comma_%(%receiver%,%arg1%)' )
		#(basicAt:	'%receiver%[%arg1%]' )
		#(basicAt:ifAbsent: '%:basic_at_ifAbsent_%(%receiver%,%arg1%,%arg2%)' )
		#(basicAt:put:	'%:basic_at_put_%(%receiver%,%arg1%,%arg2%)' )
		#(ifUndefined:	'%:ifUndefined_do_%(%receiver%,%arg1%)' )
	))! !

! LuaEncoder methodsFor: #optimization !
s8Optimizations
	" Private - Return the optimization set. "

	^#(s8	#(
		core
		#(ifTrue: hasBlockNoArguments: #(
			#(ifTrue: noClient: 'if %receiver%==true then %seq1% end' )
			#(ifTrue: '%:if_true_false_%(%receiver%,%arg1%,nil)' )
		))
		#(ifFalse: hasBlockNoArguments: #(
			#(ifFalse: noClient: 'if not (%receiver%)==true then %seq1% end' )
			#(ifFalse: '%:if_true_false_%(%receiver%,nil,%arg1%)' )
		))
		#(ifTrue:ifFalse: hasBlockNoArguments: #(
			#(ifTrue:ifFalse: noClient: 'if %receiver%==true then %seq1% else %seq2% end' )
			#(ifTrue:ifFalse: '%:if_true_false_%(%receiver%,%arg1%,%arg2%)' )
		))
		#(ifFalse:ifTrue: hasBlockNoArguments: #(
			#(ifFalse:ifTrue: noClient: 'if %receiver%==true then %seq2% else %seq1% end' )
			#(ifFalse:ifTrue: '%:if_true_false_%(%receiver%,%arg2%,%arg1%)' )
		))
		#(whileTrue receiverIsBlockNoArguments: #(
			#(whileTrue noClient: 'while %receiver%() do end' )
			#(whileTrue '%:while_do_%(%receiver%,nil)' )
		))
		#(whileFalse receiverIsBlockNoArguments: #(
			#(whileFalse noClient: 'while not %receiver%() do end' )
			#(whileFalse '%:whileNot_do_%(%receiver%,nil)' )
		))
		#(whileTrue: receiverIsBlockNoArguments: #(
			#(whileTrue: noClient: 'while %receiver%() do %seq1% end' )
			#(whileTrue: '%:while_do_%(%receiver%,%arg1%)' )
		))
		#(whileFalse: receiverIsBlockNoArguments: #(
			#(whileFalse: noClient: 'while not %receiver%() do %seq1% end' )
			#(whileFalse: '%:whileNot_do_%(%receiver%,%arg1%)' )
		))
		#(orNil '%:orNil_%(%receiver%)' )
		#(timesRepeat: '%:times_repeat_%(%receiver%,%arg1%)' )
		#(isKindOf: '%:is_kindOf_%(%receiver%,(%arg1%))' )
		#(respondsTo: '%:responds_to_%(%receiver%,(%arg1%))' )
		#(ifNil: '%:if_nil_notNil_%(%receiver%,%arg1%,nil)' )
		#(ifNotNil: '%:if_nil_notNil_%(%receiver%,nil,%arg1%)' )
		#(ifNil:ifNotNil: '%:if_nil_notNil_%(%receiver%,%arg1%,%arg2%)' )
		#(ifNotNil:ifNil: '%:if_nil_notNil_%(%receiver%,%arg2%,%arg1%)' )
	))! !

! LuaEncoder methodsFor: #optimization !
magnitudeOptimizations
	" Private - Return the optimization set. "

	^#(magnitude #(
		#(> 	'(%receiver% > %arg1%)' )
		#(< 	'(%receiver% < %arg1%)' )
		#(>=	'(%receiver% >= %arg1%)' )
		#(<=	'(%receiver% <= %arg1%)' )
	))! !

! LuaEncoder methodsFor: #optimization !
mathOptimizations
	" Private - Return the optimization set. "

	^#(math #(
		#(max:	'%:math_max_%(%receiver%,%arg1%)' )
		#(min:	'%:math_min_%(%receiver%,%arg1%)' )
		#(mustBeNumber	'%:mustBeNumber_%(%receiver%)' )
		#(+	'(%receiver% + %arg1%)' )
		#(#-	'(%receiver% - %arg1%)' )
		#(*	'(%receiver% * %arg1%)' )
		#(/	'(%receiver% / %arg1%)' )
	))! !

! LuaEncoder methodsFor: #optimization !
printingOptimizations
	" Private - Return the optimization set. "

	^#(printing #(
		#(toString	'%:toString_%(%receiver%)' )
		#(printString	'%:printString_%(%receiver%)' )
		#(asString	'%:asString_%(%receiver%)' )
		#(asCharacter	'%:asCharacter_%(%receiver%)' )
	))! !

! LuaEncoder methodsFor: #optimization !
booleanOptimizations
	" Private - Return the optimization set. "

	^#(boolean #(
		#(isBoolean '(type(%receiver%)=="boolean")' )
		#(and: hasBlockNoArguments: '%:boolean_and_%(%receiver%,%arg1%)' )
		#(or: hasBlockNoArguments:  '%:boolean_or_%(%receiver%,%arg1%)' )
		#(not '%:booleanNot_%(%receiver%)' )
	))! !

! LuaEncoder methodsFor: #defaults !
defaultKnownOptimizations
	" Private - Return the default knownOptimizations of the receiver. "

	^(Array	with: self coreOptimizations)
		add: self s8Optimizations;
		add: self magnitudeOptimizations;
"		add: self mathOptimizations;"
		add: self booleanOptimizations;
		add: self printingOptimizations;
		yourself! !

! LuaEncoder methodsFor: #defaults !
defaultOptimizations
	" Private - Return the default optimizations of the receiver. "

	^#( s8 magnitude "math" boolean "evaluation" printing )! !

! LuaEncoder class methodsFor: #inlines !
initializeKernelInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self kernelInlines into: pool ! !

! LuaEncoder class methodsFor: #inlines !
kernelInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(Object #(
#(halt		'debug.debug()')
"#(toString	'local unused')"
#(toString:	'tostring(anObject)')
#(isObject:	'
	local status, result = pcall(function() return type(anObject.yourself)=="function" end)
	if true then return status and result end')
#(typeOf:	'type(anObject)')
"#(callableOf:	'local unused')"
#(basicDelete:	'if self==nil then return else self[aString]=nil end')
#(basicPerform:withArguments:	#('self' 'fn(self,(table.unpack or unpack)(anArray))'))
#(asJSONObject	'{}')
))
#(Smalltalk #(
#(includesKey:	'{}')
#(eval:	'
	-- (req v5.3) local fn,err  = load(anExpression)
	function f() local r = anExpression; anExpression= ""; return r end -- (req v5.1)
	local fn,err  = load(f)
	if fn then return fn() else
		self:note_(" eval: Error "..err) --the expression must be a lua expression (including return)
		self:note_(" expression= "..anExpression)
		return nil
	end')
))
#('Smalltalk class' #(
#(global	'_G')
#(gc	'if collectgarbage() then return true end')
))
#(Behavior #(
#(basicNew	'self["$fn"]()')
"#(fn		'local unused')"
"#(prototype	'local unused')"
"#(privateMethods	'local unused')"
))
#(CompiledMethod #(
"#(asClosure	'local unused')"
))
#(Boolean #(
#(not 'self == false')
#(=	'
	if anObject==nil then return false end
	if true then return self==anObject end')
#(ifTrue:ifFalse: '
	if (self==true)
		then return aBlock:value()
		else return anotherBlock:value() end')
))
#(Number #(
#(= '
	if (type(anObject)=="number")
		then return self==anObject
		else return false
		end')
#(> 'self > aNumber')
#(>= 'self >= aNumber')
#(< 'self < aNumber')
#(<= 'self <= aNumber')
#(+ 'self + aNumber')
#(- 'self - aNumber')
#(* 'self * aNumber')
#(/ 'self / aNumber')
#(sin 'math.sin(self)')
#(cos 'math.cos(self)')
#(tan 'math.tan(self)')
#(arcCos 'math.acos(self)')
#(arcSin 'math.asin(self)')
#(arcTan 'math.atan(self)')
#(atan2: 'math.atan2(self,aNumber)')
#(sqrt 'math.sqrt(self)')
#(<< 'bit.lshift(self,aNumber)')
#(>> 'bit.rshift(self,aNumber)')
#(raisedTo: 'math.pow(self,aNumber)')
#(max: 'math.max(self,aNumber)')
#(min: 'math.min(self,aNumber)')
#(rounded 'self + 0.5 - (self + 0.5) %% 1')
#(truncated 'math.floor(self)')
#(timesRepeat: 'for i=1,self do fn(i) end')
#(to:do: 'for i=self,aNumber do fn(i) end')
#(to:by:do: 'for i=self,aNumber,increment do fn(i) end')
#(printString 'tostring(self)')
#(toString: '%:basen%(self,radix)')
#(isNumber 'if (self~=self) then return false else return (self~=math.huge and self~=-math.huge) end')
#(clearInterval 'if true then self:notImplementedYet_("clearInterval") end')
#(clearTimeout 'if true then self:notImplementedYet_("clearTimeout") end')
))
#('Number class' #(
#(pi 'math.pi')
))
#(BlockClosure #(
#(whileTrue: 'while self() do fn() end')
#(whileFalse: 'while (self()==false) do fn() end')
#(valueWithPossibleArguments: 'self((table.unpack or unpack)(array))')
#(try:catch: '
	local success, result = pcall(function() return fn() end)
	if success
		then return result
		else return ab( self:smalltalkErrorOf_(result) ) end')
#(valueWithTimeout: 'self:notImplementedYet_("valueWithTimeout:")')
#(valueWithInterval: 'self:notImplementedYet_("valueWithInterval:")')
#(ensure: '
	local success, result = pcall(function() return self() end)
	fn()
	if true then return result end')
#(ifCurtailed: '
	local curtailed = true
	local success, result = pcall(function()
		local r = self()
		curtailed = false
		return r
	end)
	if (curtailed)
		then exceptionBlock:value()
		else return result
		end')
#(argumentCount 'self:notImplementedYet_("argumentCount")')
))
#(UndefinedObject  #(
#(asJSONObject 'nil')
))
#(Collection  #(
#(wrap: 'if %:isNil_%(anObject) --anObject==nil
	then return smalltalk.UndefinedObject:alias()
	else return anObject end')
#(unwrap: '--if anObject==nil then return nil end
	--if (type(anObject)=="table") and %:isNil_%(anObject)
	if %:isNil_%(anObject)
		then return nil
		else return anObject end')
))
#(HashedSet #(
#(initializeHash '{}')
#(includes: '(self["@hash"][cookie])~=nil')
"#(privateAdd: 'local unused')"
#(privateRemove: 'self["@hash"][cookie] = nil')
))
#(SequenceableCollection #(
#(indexOf:ifAbsent: '
	for i,value in ipairs(self) do
		if (%:isObject_equalTo_%(self:unwrap_(value),anObject)) then return i end
	end
	if true then return aBlock:value() end')
#(do: '
	for _,value in ipairs(self) do
		fn(self:unwrap_(value))
	end')
#(detect:ifNone: '
	for _,value in ipairs(self) do
		local v=self:unwrap_(value)
		if (fn(v)) then return v end
	end
	if true then return anotherBlock:value() end')
#(reversed '
	local result = smalltalk.Array:new()
	local top = #self + 1
	for i,v in ipairs(self) do result[top - i] = v end
	if true then return result end')
#(isEmpty '#self==0')
#(reverseDo:  'for i = #self,1,-1 do fn(self:unwrap_(self[i])) end')
#(includes: #('
	for _,value in ipairs(self) do
		if (%:isObject_equalTo_%(self:unwrap_(value),anObject)) then return true end
	end
	if true then return false end' 'if true then end'))
#(occurrencesOf: '
	local count = 0
	for _,value in ipairs(self) do
		if (%:isObject_equalTo_%(self:unwrap_(value),anObject)) then count = count + 1 end
	end
	if true then return count end')
))
#('String class' #(
#(fromString: 'if true then tostring(aString) end')
#(new	'""')
#(cr 	'"\r"')
#(lf 	'"\n"')
#(crlf 	'"\r\n"')
#(space	'" "')
#(tab	'"\t"')
#(zero	'string.char(0)')
#(fromCharCode: '
	if (type(anArrayOrNumber)=="number") then
		return string.char(anArrayOrNumber)
	end
	if true then
		local fup = table.unpack or unpack
		return (string.char(fup(anArrayOrNumber)))
	end')
))
#(String #(
#(=	'if true then return self==anObject end')
#(size	'#self')
#(at:	'self:sub(anIndex,anIndex)')
#(do: 	'for c in self:gmatch(".") do fn(c) end')
#(withIndexDo: 'for i=1,#self do fn(self:sub(i,i),i) end')
#(detect:ifNone: '
	for c in self:gmatch(".") do
		if (fn(c)) then return c end
	end
	if true then return anotherBlock:value() end')
#(isLiteral '
	if #self==0 then return false end
	if smalltalk.String:isBinarySelector_(self) then return true end
	if true then return string.find(self,"[^a-zA-Z0-9_:]+")==nil end')
#(occurrencesOf: '
	local j
	local k,count = 0,0
	repeat	j, k = self:find(aString, k + 1, true)
		if (j~=nil) then count=count+1 end
	until j == nil
	if true then return count end')
#(startsWith: 'if true then return (self:sub(1, #aString) == aString) end')
#(escaped 'self:notImplementedYet_("escaped")')
#(unescaped '
	local x16 = function(x) return string.char(tonumber(x, 16)) end
	if true then return self:gsub("%%%%(%%x%%x)", x16) end')
#(asEncodedURI 'self:notImplementedYet_("asEncodedURI")')
#(asDecodedURI 'self:notImplementedYet_("asDecodedURI")')
#(, 'if true then return self..(%:asString_%(aCollection)) end')
#(copyWith: 'self..(%:asString_%(anObject))')
#(copyFrom:to: 'self:sub(anIndex,anotherIndex)')
#(reversed 'if true then return self:reverse() end')
#(includes: 'local ignored')
#(indexOf: '
	local index = self:find(aString,1,true)
	if not index
		then return 0
		else return index end')
#(indexOf:startingAt: '
	local index = self:find(aString,position,true)
	if not index
		then return 0
		else return index end')
#(lastIndexOf: '
	local i = 0
	local j = 0
	local k = 0
	repeat	i = j
		j, k = self:find(aString, k + 1, true)
	until j == nil
	if true then return i end')
#(lastIndexOf:startingAt: '
	local i, j
	local k = position
	repeat	i = j
		j, k = self:find(aString, k + 1, true)
	until j == nil
	if true then return i end')
#(asNativeName 'self:asLuaName()')
#(asPrivateSelector 'local map = {
		[":"]="_",
		["+"]="_plus",
		["-"]="_minus",
		["*"]="_star",
		["/"]="_slash",
		["|"]="_bar",
		["&"]="_amp",
		[">"]="_gt",
		["<"]="_lt",
		["="]="_eq",
		[","]="_comma",
		["@"]="_at",
		["~"]="_no",
		["\\"]="_bk"
	};
	result = self:gsub(".",function(c) return map[c] end)')
#(asNative: 'if true then return self:slashEncoded() end')
#(replace:with: '
	local result=""
	local i,j
	local k = 0
	repeat	i=k+1
		j, k = self:find(aString, i, true)
		if (j==nil) then
			result=result..(self:sub(i))
		else
			result=result..(self:sub(i,j-1)..anotherString)
		end
	until j == nil
	if true then return result end')
#(replaceRegexp:with: 'self:notImplementedYet_("replaceRegexp:with:")')
#(tokenize: '
	local result=smalltalk.Array:new()
	local i,j
	local k = 0
	repeat	i=k+1
		j, k = self:find(aString, i, true)
		if (j==nil) then
			result:add_(self:sub(i))
		else
			result:add_(self:sub(i,j-#aString))
		end
	until j == nil
	if true then return result end')
#(match: 'self:notImplementedYet_("match:")')
#(isLetter '#self==1 and string.find(self,"%%a")==1')
#(isDigit '#self==1 and string.find(self,"%%d")==1')
"#(isAlphaNumeric 'local unused')"
#(isUpperCase '
	if (#self==0) then return false end
	for c in self:gmatch(".") do
		if (c<"A" or c>"Z") then return false end
	end
	if true then return true end')
#(isLowerCase '
	if (#self==0) then return false end
	for c in self:gmatch(".") do
		if (c<"a" or c>"z") then return false end
	end
	if true then return true end')
#(asUppercase 'self:upper()')
#(asLowercase 'self:lower()')
#(asNumber 'tonumber(self)')
#(> 'self > aString')
#(< 'self < aString')
#(>= 'self >= aString')
#(<= 'self <= aString')
#(isCharacter '#self == 1')
#(asciiValue 'string.byte(self)')
#(isSeparator '#self==1 and self:find("%%s")==1')
))
#('Array class' #(
#(basicNew: '%:newArray%()')
#(with: 'self')
#(with:with: 'self')
#(with:with:with: 'self')
#(with:with:with:with: 'self')
#(withAll: '
	local result = aCollection:asArray()
	if aCollection==result
		then return result:copy()
		else return result
	end')
#(slice: 'self:notImplementedYet_("slice:")')
))
#(Array #(
#(size '#self')
#(at: '
	local value = self[anIndex]
	if value==nil then return self:objectNotFoundAt_(anIndex) end
	if true then return self:unwrap_(value) end')
#(at:ifAbsent: '
	local value = self[anIndex]
	if (value==nil) then return aBlock:value() end
	if true then return self:unwrap_(value) end')
#(at:put: '
	self[anIndex]=self:wrap_(anObject)
	if true then return anObject end')
#(do: '
	for _,value in ipairs(self) do
		fn(self:unwrap_(value))
	end')
#(addAll: '
	local array = aCollection:asArray()
	for i=1,#array do self[#self+1]=self:wrap_(array[i]) end')
#(shallowCopy '
	result = self:class():new()
	for i=1,#self do result[#result+1]=self[i] end')
#(copyWithAll: '
	local result = self:copy()
	result:addAll_(aCollection)
	if true then return result end')
#(, '
	local result = self:copy()
	result:addAll_(aCollection)
	if true then return result end')
#(join: 'table.concat(self,aString)')
#(sort: 'table.sort(self,fn)')
#(remove:ifAbsent: '
	for i,value in ipairs(self) do
		if (%:isObject_equalTo_%(self:unwrap_(value),anObject)) then return table.remove(self,i) end
	end
	if true then return aBlock:value() end')
#(add: 'self[#self+1]=self:wrap_(anObject)')
#(addFirst: 'table.insert(self,1,self:wrap_(anObject))')
#(addLast: 'self[#self+1]=self:wrap_(anObject)')
#(removeAll 'for i=#self,1,-1 do self[i]=nil end')
#(removeFirst 'table.remove(self,1)')
#(removeIndex: 'table.remove(self,index)')
#(removeLast 'self:unwrap_(table.remove(self))')
#(atAllPut: '
	local what=self:wrap_(anObject)
	for i=1,#self do self[i]=what end')
#(first '
	local value=self[1]
	if (value~=nil) then return self:unwrap_(value) end')
#(second '
	local value=self[2]
	if (value~=nil) then return self:unwrap_(value) end')
#(third '
	local value=self[3]
	if (value~=nil) then return self:unwrap_(value) end')
#(fourth '	
	local value=self[4]
	if (value~=nil) then return self:unwrap_(value) end')
#(fifth '
	local value=self[5]
	if (value~=nil) then return self:unwrap_(value) end')
#(sixth '
	local value=self[6]
	if (value~=nil) then return self:unwrap_(value) end')
))
#('RegularExpression class' #(
#(fromString:flag: 'self:notImplementedYet_("fromString:flag:")')
))
#(Exception #(
#(signal 'error(self)')
))
#(ComposedAssociations #(
#(hashObject '{}')
))
#('KeyedCollection class' #(
#(@ '
	parts = smalltalk.Array:new()
	for k,_ in pairs(jsObject) do parts:add_(k) end')
))
#(KeyedCollection #(
#(newHash '{}')
))
#('Dictionary class' #(
#(basicNew 'if true then return smalltalk.PoolDictionary:new() end')
))
#(PoolDictionary #(
"#(validKey: 'local unused')"
"#(at:put: 'local unused')"
#(removeKey:ifAbsent: 'self["@hash"][self:wrap_(aKey)]=nil')
#(at:ifAbsent: '
	local found = self["@hash"][self:wrap_(aKey)]
	if found~=nil
		then return self:unwrap_(found)
		else return aBlock:value()
	end')
#(includesKey: 'self["@hash"][self:wrap_(aKey)]~=nil')
#(keys '
	local result=smalltalk.Array:new()
	for k,_ in pairs(self["@hash"]) do result:add_(k) end
	if true then return result end')
))
#(ClassBuilder #(
#(addSubclassOf:named:instanceVariableNames: '
	%:addClass%(aString, aClass, aCollection)
	if true then return smalltalk[aString] end')
))
#(StringStream #(
#(peek '
	local c = self["@collection"]:at_(self:position()+1)
	if c=="" then return nil else return c end')
#(peekForAll: '
	local start=self:position()
	if (self["@collection"]:sub(start+1,start+#aCollection)~=aCollection) then return false end')
))
#(Random #(
#(next 'if true then return math.random() end')
))
#('DateTime class' #(
#(new: 'if true then return self:notImplementedYet_("new:") end')
#(today 'if true then return self:new() end')
#(now 'if true then return self:today() end')
#(msNow 'if true then return os.clock()*1000 end')
))
#(DateTime #(
#(< 'self:notImplementedYet_("<")')
#(> 'self:notImplementedYet_(">")')
#(<= 'self:notImplementedYet_("<=")')
#(>= 'self:notImplementedYet_(">=")')
#(- 'self:notImplementedYet_("-")')
#(+ 'self:notImplementedYet_("+")')
#(isValid 'self:notImplementedYet_("isValid")')
))
	) ! !

! LuaEncoder class methodsFor: #inlines !
initializeNativeInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self nativeInlines into: pool ! !

! LuaEncoder class methodsFor: #inlines !
nativeInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#('NativeObject class' #(
#(undefined 'nil')
#(codeToNumber '" "')
))
#(NativeObject #(
#(jsObject 'self:invalidMessage_("jsObject")')
#(ownPropertyNames '
	local result = smalltalk.Array:new()
	for k,_ in pairs(self["@handle"]) do
		result:add_(k)
	end
	if true then return result end')
))
#(NativeHandle #(
#(hasOwnProperty: 'rawget(self["@handle"],aName)~=nil')
#(delete: 'self["@handle"][aName]=nil')
))
#('NativeLibraryObject class' #(
#(basicLoad: 'require(url)')
))
#(NativeLibraryObject  #(
#(newHandleOf: 'result = aConstructor()')
#(newHandleOf:with: 'result = aConstructor(arg1)')
#(newHandleOf:with:with: 'result = aConstructor(arg1,arg2)')
#(newHandleOf:with:with:with: 'result = aConstructor(arg1,arg2,arg3)')
#(newHandleOf:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4)')
#(newHandleOf:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5)')
#(newHandleOf:with:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5,arg6)')
#(newHandleOf:with:with:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5,arg6,arg7)')
#(newHandleOf:with:with:with:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8)')
))
#(Array #(
#(addAllLua: 'for _,value in pairs(natives) do self:add_(value) end')
))
#('Array class' #(
#(@@ 'self')
#(@ 'result:addAllLua_(natives)')
))
#('Boolean class' #(
#(@ '
	if (aBOOL==true) then return true end
	if (aBOOL==1) then return true end
	if (aBOOL==false) then return false end
	if (aBOOL+0==1) then return true end')
))
	) ! !

! LuaEncoder class methodsFor: #inlines !
initializeToolsInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self toolsInlines into: pool ! !

! LuaEncoder class methodsFor: #inlines !
toolsInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(NativeReader #(
#(initializeObject '{}')
))
#(String  #(
#(asComment #('"--[[ "' '" --]]"' '"-- "'))
))
#(MethodReferences #(
#(dot		'":"')
#(localDef	'"local "')
#(superSendPatternOf:	'self:superLuaSendPatternOf_(aSelector)')
#(compiledSourceStart	'"("')
))
#(Class #(
#(hackDNU 'self:notImplementedYet_("hackDNU")')
))
	) ! !

! LuaEncoder class methodsFor: #inlines !
initializeSnapshotInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self snapshotInlines into: pool ! !

! LuaEncoder class methodsFor: #inlines !
snapshotInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(Smalltalk #(
#(exit	'quit()')
))
#(Snapshot #(
#(bootHeader '[[smalltalk = (function() 
]]')
#(bootFooter '[[
end)()
]]')
#(launchExpression '"%:launchImage%()"')
))
	) ! !

! LuaEncoder class methodsFor: #inlines !
initializeOtherInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self otherInlines into: pool ! !

! LuaEncoder class methodsFor: #inlines !
otherInlines
	" Private - Return the inlines for methods in the category. "

	^#(
"#(ClassName #(
#(selector '...')
))"
	) ! !

"------------------------------------------------------------------------------
	Extension to Kernel to support Lua encoding
 ------------------------------------------------------------------------------"!

! String class methodsFor: #Lua-lua !
reservedLuaWords
	" Private - Return the reserved names of Lua language. "

	^#(
	and	break	do	else	elseif
	end	#false	for	function	if
	in	local	#nil	not	or
	repeat	return	then	#true	until
	while
	)! !

! String class methodsFor: #Lua-lua !
forbiddenLuaSelectors
	" Private - Return the platform forbidden selectors. "

	^#( "class new" )! !

! String methodsFor: #Lua-lua !
asLuaName
	" Private - Return the receiver as a valid Lua name. "

	self first = $_ ifTrue: [ ^self ].
	(String reservedLuaWords includes: self) ifTrue: [ ^'__',self ].
	^self! !

! String methodsFor: #Lua-lua !
asLuaSelector
	" Private - Return the receiver as a valid native selector. "

	(String forbiddenLuaSelectors includes: self) ifTrue: [ ^$_ ,self ].
	(self startsWith: $#) ifTrue: [
		^self upTo: $: startingAt: 2
	].
	^self asPrivateSelector asLuaName ! !

! Object methodsFor: #Lua-lua !
asLua
	" Return the native representation of the receiver.
	The default implementation return the receiver as a String.
	This method must be refined by subclasses implementing native literal representation.
	"

	^self asString! !

! Number methodsFor: #Lua-lua !
asLua
	" Return the native representation of the receiver. "

	^self printString ! !

! String methodsFor: #Lua-lua !
asLua
	" Return the native representation of the receiver. "

	| encoded bar |
	bar := 92 asCharacter.
	encoded := self.
	encoded := encoded replace: bar with: bar ,bar.
	encoded := encoded replace: String lf with: bar ,$n.
	encoded := encoded replace: String cr with: bar ,$r.
	encoded := encoded replace: String tab with: bar ,$t.
	encoded := encoded replace: $" with: bar ,$".
	^$" ,encoded ,$"! !

! Array methodsFor: #Lua-lua !
asLua
	" Return the native representation of the receiver. "

	^'smalltalk:newArray' ,$( ,${
		,((self collect: [:each | each asLua ]) join: ', ')
	,$} ,$) ! !

! Set methodsFor: #Lua-lua !
asLua
	" Return the native representation of the receiver. "

	^'smalltalk.' ,self class name
	, ':newContents_('
		, contents asLua
	,')'! !

! Dictionary methodsFor: #Lua-lua !
asLua
	" Return the native representation of the receiver. "

	^'smalltalk.' ,self class name
	, ':fromKeys_and_('
		, keys asArray asLua
		,',' ,values asLua
	,')'! !

! PoolDictionary methodsFor: #Lua-lua !
asLua
	" Return the native representation of the receiver. "

	^'smalltalk.',self class name
	, ':fromTuples_('
		, self tuples asLua
	,')'! !


" Custom conversion methods
Note: We implement __tostring delegating to #toString implementors"
#toString implementors do: [:mth| | source |
	source := '__tostring
	" Return the printable representation of the receiver. "

	^self toString'.
	(mth classField
		methodsFor: #Lua-printing)
		scanFrom: source stream
]!

Boolean removeSelector: #__tostring!

! Array methodsFor: #Lua-printing !
__tostring
	" Return the printable representation of the receiver. "

	^self asLiteral! !

! Array methodsFor: 'Lua-private' !
addAllLua: natives
	" Private - Add the native values to the receiver. "

	(self typeOf: natives) = #userdata ifTrue: [ | count |
		count := JavaArray getLength: Lua@natives.
		1 to: count do: [:i|
			self add: (JavaArray nativeClass handle
				#get: natives index: i - 1).
		].
		^self
	].
	{''js'self.addAll_(natives)'}.
	^self! !

"------------------------------------------------------------------------------
	Native handlers for Lua
 ------------------------------------------------------------------------------"!

! NativeObject class methodsFor: 'Lua-builders' !
buildLuaFunctions: specs
	" OBSOLETE - Generate code from specs as raw functions implemented by instance handle. "

	self buildNativeFunctions: specs! !

! NativeObject class methodsFor: 'Lua-builders' !
buildLuaStaticFunctions: specs
	" OBSOLETE - Generate code from specs as raw functions implemented by class handle. "

	self buildNativeStaticFunctions: specs! !

NativeHandle
	subclass: #Lua
	category: #Lua!
Lua comment: '
	Implementation of wrappers to native Lua objects.
	@2011 Alejandro F. Reimondo
	aleReimondo@smalltalking.net'!

! Lua class methodsFor: #instantiation !
supportsLanguage: language
	" Private - Return true if instances of the receiver implement native wrappers for language. "

	^language = #lua! !

! Lua methodsFor: #accessing !
propertyNames
	" Return the property names of the receiver. "

	^self ownPropertyNames! !

! Lua class methodsFor: #accessing !
global
	" Return an instance wrapping global. "

	^self @ Smalltalk global! !

! Lua class methodsFor: #accessing !
version
	" Return the host version. "

	^self global @ #_VERSION! !

! Lua class methodsFor: #accessing !
arg
	" Return the main arguments (anArray of strings). "

	^Array @ (self global @ #arg)! !

! Lua class methodsFor: #modules !
require: aName
	" Return an instance of the receiver obtained from loading module aName. (or nil) "

	^self @ (# #require: aName)! !

! Lua methodsFor: #iteration !
do: aBlock
	" Evaluates aBlock on each element of the receiver. "

{''js'	self.invalidMessage_("do:")'
lua'	for item in self:handle() do aBlock(item) end'}.! !

! Lua methodsFor: #iteration !
pairsDo: aBlock
	" Evaluates aBlock on each key/value pair of the receiver. "

{''js'	self.invalidMessage_("pairsDo:")'
lua'	for k,v in pairs(self:handle()) do aBlock(k,v) end'}.! !

! Lua methodsFor: #iteration !
keysDo: aBlock
	" Evaluates aBlock on each key of the receiver. "

{''js'	self.invalidMessage_("keysDo:")'
lua'	for k,_ in pairs(self:handle()) do aBlock(k) end'}.! !

! Lua methodsFor: #iteration !
valuesDo: aBlock
	" Evaluates aBlock on each value of the receiver. "

{''js'	self.invalidMessage_("valuesDo:")'
lua'	for _,v in pairs(self:handle()) do aBlock(k,v) end'}.! !


"------------------------------------------------------------------------------
	Development Utilities & tools for Lua environment
 ------------------------------------------------------------------------------"!

! MethodReferences methodsFor: #Tools-patterns !
superLuaSendPatternOf: aSelector
	" Private - Return the search pattern for senders (to super) of aSelector. "

	^$" ,aSelector asSelector ,$" ,')(self'! !

! Object methodsFor: #Lua-fileIn !
fileInOBSOLETE: fileName
	" FileIn contents of file at fileName. "

	self note: 'fileIn("', fileName, '");'.
	^# #fileIn: fileName! !

! Object methodsFor: #Lua-fileIn !
fileInJSOBSOLETE: fileName
	" Load native code from file at fileName. "

	self note: 'fileInJS("', fileName, '");'.
	^# #fileInJS: fileName! !