
Object
 subclass: #NativeObject
 instanceVariableNames: #handle
 category: #Natives!
NativeObject comment: '
	Wrapper to native object.
	@2011 Alejandro F. Reimondo
	aleReimondo@smalltalking.net
'!

! NativeObject class methodsFor: #constants !
undefined
	" Private - Return the native representation for undefined. "

	^{''}! !

! NativeObject class methodsFor: #builders !
nativeNameFromSelector: aSymbol
	" Private - Return the property name for selector aSymbol. "

	^aSymbol! !

! NativeObject class methodsFor: 'builders-private' !
buildGetterCode: aSelector
	" Private - Return source code to get aSelector. "

	(self instanceVariableNames includes: aSelector) ifTrue: [
		^super buildGetterCode: aSelector
	].
	^'^self handle basicAt: '
		,(self nativeNameFromSelector: aSelector) asLiteral! !

! NativeObject class methodsFor: #builders !
buildGetter: spec
	" Build method for getter. "

	(spec isString or: [
		spec isArray and: [
		spec second = #initialize ]]) ifTrue: [
		^super buildGetter: spec
	].
	^self buildTypedGetter: spec! !

! NativeObject class methodsFor: 'builders-private' !
buildSetterCode: aSelector
	" Private - Return source code to set aSelector. "

	(self instanceVariableNames includes: aSelector) ifTrue: [
		^super buildSetterCode: aSelector
	].
	^'self handle basicAt: '
		,(self nativeNameFromSelector: aSelector) asLiteral
		, ' put: anObject'! !

! NativeObject class methodsFor: #builders !
buildSetter: spec
	" Private - Return source code for setter. "

	(spec isString or: [
		spec isArray and: [
		spec second = #initialize ]]) ifTrue: [
		^super buildSetter: spec
	].
	^self buildTypedSetter: spec! !

! NativeObject class methodsFor: #builders !
buildTypedGetter: tuple
	" Build (instance) method to get property from spec array. "

	tuple isString ifTrue: [ ^self buildGetter: tuple ].
	^self build: #TypedGetter -> tuple doing: [
		self
			implement: (self stMessageHeader: tuple first)
			comment: 'Return the ' ,tuple first ,' of the receiver.'
			as: 	 (self apiPrologueFor: tuple last)
				,'self handle basicAt: '
				,(self nativeNameFromSelector: tuple first) asLiteral
				,(self apiEpilogueFor: tuple last)
			category: #accessing
	]! !

! NativeObject class methodsFor: #builders !
buildTypedSetter: tuple
	" Build (instance) method to set property to spec array. "

	tuple isString ifTrue: [ ^self buildSetter: tuple ].
	^self build: #TypedSetter -> tuple doing: [
		self
			implement: (self stMessageHeader: tuple first , ': anObject')
			comment: 'Set the ' ,tuple first ,' of the receiver.'
			as: 'self handle basicAt: '
				,(self nativeNameFromSelector: tuple first) asLiteral
				,' put: anObject handle'
			category: #accessing
	]! !

! NativeObject class methodsFor: #builders !
buildTypedGetters: specs
	" Build (instance) methods to get properties. "

	^self build: #TypedGetters -> specs doing: [
		specs compose: [:each| self buildTypedGetter: each ]
	]! !

! NativeObject class methodsFor: #builders !
buildTypedAccessors: tuples
	" Build (instance) methods to get/set property from/to tuples. "
	
	^self build: #TypedAccessors -> tuples doing: [
		tuples compose: [:each|
			(self buildTypedGetter: each)
			,(self buildTypedSetter: each)
		]
	]! !

! NativeObject class methodsFor: #builders !
buildFastGetter: aSelector
	" Build (instance) method to get native property. "

	| type jsName |
	aSelector isString
		ifTrue: [ type := ''. jsName := self nativeNameFromSelector: aSelector ]
		ifFalse: [ type := aSelector last. jsName := self nativeNameFromSelector: aSelector first ].
	^self build: #FastGetter -> aSelector doing: [
		self
			implement: (self stMessageHeader: aSelector)
			comment: 'Return the ' ,aSelector ,' of the receiver.'
			as: '^' ,type ,'handle#' ,jsName category: #accessing
	]! !

! NativeObject class methodsFor: #builders !
buildFastSetter: aSelector
	" Build (instance) method to set native property. "

	| jsName |
	aSelector isString
		ifTrue: [ jsName := self nativeNameFromSelector: aSelector ]
		ifFalse: [ jsName := self nativeNameFromSelector: aSelector first ].
	^self build: #FastSetter -> aSelector doing: [
		self
			implement: (self stMessageHeader: aSelector) ,': anObject'
			comment: 'Set the ' ,aSelector ,' of the receiver.'
			as: 'handle#' ,jsName ,' := anObject' category: #accessing
	]! !

! NativeObject class methodsFor: #builders !
buildFastAccessor: accessor
	" Build (instance) method to access native property. "

	accessor isString ifTrue: [
		^self build: #FastAccessor -> accessor doing: [
			(self buildFastGetter: accessor)
			,(self buildFastSetter: accessor)
		]
	].
	^self error: 'Invalid (Fast)accessors.'! !

! NativeObject class methodsFor: #builders !
buildFastGetters: specs
	" Build methods to get properties. "

	specs isString ifTrue: [ ^self buildFastGetters: specs asArrayOfSubstrings ].
	^self build: #FastGetters -> specs doing: [
		specs compose: [:each| self buildFastGetter: each ]
	]! !

! NativeObject class methodsFor: #builders !
buildFastAccessors: specs
	" Build methods to access properties. "

	specs isString ifTrue: [ ^self buildFastAccessors: specs asArrayOfSubstrings ].
	^self build: #FastAccessors -> specs doing: [
		specs compose: [:each| self buildFastAccessor: each ]
	]! !

! NativeObject class methodsFor: #constants !
codeToNumber
	" Private - Return the code to transform value to a number. "

	^{''}! !

! NativeObject class methodsFor: #constants !
handleAt: aSelector
	" Private - Return the code for accessing instance handle. "

	^'handle#' ,(self nativeNameFromSelector: aSelector)! !

! NativeObject class methodsFor: #builders !
buildNumberGetter: aSelector
	" Build (instance) method to get native property. "

	^self build: #NumberGetter -> aSelector doing: [
		self
			implement: (self stMessageHeader: aSelector)
			comment: 'Return the ' ,aSelector ,' of the receiver.'
			as: $^ ,(self handleAt: aSelector) ,self codeToNumber
			category: #accessing
	]! !

! NativeObject class methodsFor: #builders !
buildNumberAccessor: accessor
	" Build (instance) method to access native property. "

	accessor isString ifTrue: [
		^self build: #NumberAccessor -> accessor doing: [
			(self buildNumberGetter: accessor)
			,(self buildSetter: accessor)
		]
	].
	^self error: 'Invalid (Number)accessors.'! !

! NativeObject class methodsFor: #builders !
buildNumberGetters: specs
	" Build methods to get properties. "

	specs isString ifTrue: [ ^self buildNumberGetters: specs asArrayOfSubstrings ].
	^self build: #NumberGetters -> specs doing: [
		specs compose: [:each| self buildNumberGetter: each ]
	]! !

! NativeObject class methodsFor: #builders !
buildNumberAccessors: specs
	" Build methods to access properties. "

	specs isString ifTrue: [ ^self buildNumberAccessors: specs asArrayOfSubstrings ].
	^self build: #NumberAccessors -> specs doing: [
		specs compose: [:each| self buildNumberAccessor: each ]
	]! !

! NativeObject class methodsFor: #builders !
buildBooleanGetter: aSelector
	" Build (instance) method to get native property. "

	^self build: #BooleanGetter -> aSelector doing: [
		self
			implement: (self stMessageHeader: aSelector)
			comment: 'Return the ' ,aSelector ,' of the receiver.'
			as: '^true == ' ,(self handleAt: aSelector)
			category: #accessing
	]! !

! NativeObject class methodsFor: #builders !
buildBooleanAccessor: accessor
	" Build (instance) method to access native property. "

	accessor isString ifTrue: [
		^self build: #BooleanAccessor -> accessor doing: [
			(self buildBooleanGetter: accessor)
			,(self buildSetter: accessor)
		]
	].
	^self error: 'Invalid (Boolean)accessors.'! !

! NativeObject class methodsFor: #builders !
buildBooleanGetters: specs
	" Build methods to get properties. "

	specs isString ifTrue: [ ^self buildBooleanGetters: specs asArrayOfSubstrings ].
	^self build: #BooleanGetters -> specs doing: [
		specs compose: [:each| self buildBooleanGetter: each ]
	]! !

! NativeObject class methodsFor: #builders !
buildBooleanAccessors: specs
	" Build methods to access properties. "

	specs isString ifTrue: [ ^self buildBooleanAccessors: specs asArrayOfSubstrings ].
	^self build: #BooleanAccessors -> specs doing: [
		specs compose: [:each| self buildBooleanAccessor: each ]
	]! !

! NativeObject class methodsFor: #builders !
buildIsBooleanGetter: aSelector
	" Build (instance) method to get native property. "

	^self build: #IsBooleanGetter -> aSelector doing: [ | isSelector |
		isSelector :=  #is ,aSelector asGlobalName.
		self
			implement: (self stMessageHeader: isSelector)
			comment: 'Return the ' ,aSelector ,' of the receiver.'
			as: '^true == handle#' ,(self nativeNameFromSelector: isSelector)
			category: #accessing
	]! !

! NativeObject class methodsFor: #builders !
buildIsBooleanAccessor: accessor
	" Build (instance) method to access native property. "

	accessor isString ifTrue: [
		^self build: #IsBooleanAccessor -> accessor doing: [
			(self buildIsBooleanGetter: accessor)
			,(self buildSetter: accessor)
		]
	].
	^self error: 'Invalid (IsBoolean)accessors.'! !

! NativeObject class methodsFor: #builders !
buildIsBooleanGetters: specs
	" Build methods to get properties. "

	specs isString ifTrue: [ ^self buildIsBooleanGetters: specs asArrayOfSubstrings ].
	^self build: #IsBooleanGetters -> specs doing: [
		specs compose: [:each| self buildIsBooleanGetter: each ]
	]! !

! NativeObject class methodsFor: #builders !
buildIsBooleanAccessors: specs
	" Build methods to access properties. "

	specs isString ifTrue: [ ^self buildIsBooleanAccessors: specs asArrayOfSubstrings ].
	^self build: #IsBooleanAccessors -> specs doing: [
		specs compose: [:each| self buildIsBooleanAccessor: each ]
	]! !

! NativeObject class methodsFor: #builder !
buildSubclasses: aCollection
	" Private - Build aCollection of subclasses of the receiver as wrappers to the library.
	Return the last class created (or nil).
	"

	| result |
	aCollection do: [:tuple|
		result := tuple isString
		ifTrue: [ self buildSubclass: tuple arguments: #() typed: #() ]
		ifFalse: [
			tuple size = 2
			ifTrue: [ self buildSubclass: tuple first arguments: tuple second typed: #() ]
			ifFalse: [ self buildSubclass: tuple first arguments: tuple second typed: tuple last ]
		 ]
	].
	^result! !

! NativeObject class methodsFor: #builder !
buildHierarchy: aCollection
	" Private - Build aCollection of subclasses of the receiver as wrappers to the library. "
 
	| stream aClass |
	stream := aCollection stream.
	[ stream atEnd ] whileFalse: [
		aClass := self buildSubclasses: (Array with: stream next).
		stream peek isArray ifTrue: [
			aClass buildHierarchy: stream next
		].
	]! !

! NativeObject class methodsFor: #builder !
buildSubclass: globalClassName arguments: argList typed: typedArguments
	" Private - Build a subclass of the receiver implementing a wrapper to the library.
	Note: argList and typed arguments are ignored; see subclasses to find concrete implementors.
	Return the created class.
	"

	| cls clsName |
	clsName := globalClassName first asUppercase
			,(globalClassName copyFrom: 2 to: globalClassName size).
	clsName := self classToken ,clsName.
	self
		subclass: clsName
		instanceVariableNames: ''
		category: self category.
	^Smalltalk at: clsName! !

! NativeObject class methodsFor: #builder !
classToken
	" Private - Return the classToken to prepend/tag the receiver.
	The default answer is empty.
	The subclasses can subclass to build tagged subclasses.
	"

	^''! !

! NativeObject class methodsFor: #builder !
classToken: aToken
	" Private - Generate code to return the classToken of the receiver. "

	^self buildClassToken: aToken! !

! NativeObject class methodsFor: #builder !
buildClassToken: aToken
	" Generate code to return the classToken of the receiver. "

	self build: #ClassToken -> aToken doing: [
		self
			implement: #classToken
			comment: 'Private - Return the classToken to prepend/tag the receiver.'
			as: $^ ,aToken asLiteral
			category: #builder
	]! !

! NativeObject class methodsFor: #builders !
buildFunctions: specs
	" Generate code from specs as functions implemented by instance handle. "

	self build: #Functions -> specs doing: [
		specs do: [:spec| self buildFunction: spec ]
	]! !

! NativeObject class methodsFor: #builders !
buildBooleanFunctions: specs
	" Build methods to evaluate simple functions returning a Boolean. "

	specs isString ifTrue: [ ^self buildBooleanFunctions: specs asArrayOfSubstrings ].
	^self build: #BooleanFunctions -> specs doing: [
		self build: #Boolean functions: specs
	]! !

! NativeObject class methodsFor: #builders !
build: aClassName functions: specs
	" Build methods to evaluate simple functions returning an instance of aClassName. "

	specs isString ifTrue: [ ^self build: aClassName functions: specs asArrayOfSubstrings ].
	^self buildFunctions: (specs collect: [:each|
		each isString
		ifTrue: [ Array with: each with: #() with: aClassName ]
		ifFalse: [
			(each isArray and: [each size = 2]) ifFalse: [
				self error: 'Invalid literal spec ' ,each
			].
			each copyWith: aClassName
		]
	])! !

! NativeObject methodsFor: #eval !
eval: code
	" Private - Return the result of evaluating (native) code. "

	^Smalltalk current eval: code! !

! NativeObject methodsFor: #api !
ignoreReturn: value
	" Private - Ignore the argument value.
	Return the receiver. "! !

! NativeObject class methodsFor: #api !
apiSupportFor: literal
	" Private - Return the support to encode literal as native api.
	Note: literal can be a String or Array of api annotation literals.
	"

	| support |
	support := NativeHandle support.
	((support respondsTo: #handleAPIFor:)
		and: [support handleAPIFor: literal])
		ifTrue: [ ^support ].
	^self! !

! NativeObject class methodsFor: #builders !
apiEpilogueFor: aReturnType
	" Private - Return the epilogue code to use to wrap a return. "

	^(self apiSupportFor: aReturnType)
		nativeEpilogueFor: aReturnType! !

! NativeObject class methodsFor: #builders !
apiPrologueFor: aReturnType
	" Private - Return the prologue to use to wrap a return. "

	^(self apiSupportFor: aReturnType)
		nativePrologueFor: aReturnType! !

! NativeObject class methodsFor: #builders !
nativeEpilogueFor: aReturnType
	" Private - Return the epilogue code to use to wrap a return. "

	^$)! !

! NativeObject class methodsFor: #builders !
nativePrologueFor: aReturnType
	" Private - Return the prologue to use to wrap a return.
	aReturnType can be:
		- empty string to ignore api return value (will return the receiver)
		- suffixed by [...] or + to return an arrayed result
		- suffixed by type spec GlobalName
		- smalltalk code ending with (
		- a global name understanding #@
	"

	aReturnType isEmpty ifTrue: [ ^'^self ignoreReturn: (' ].
	aReturnType last = $( ifTrue: [ ^$^ ,aReturnType ].
	aReturnType last = $+ ifTrue: [
		^self apiPrologueFor: (aReturnType upTo: $+) ,'[]'
	].
	(aReturnType includes: $[) ifTrue: [
		^$^ ,(self apiPrologueForArray: aReturnType)
	].
	(aReturnType includes: $<) ifTrue: [
		^$^ ,(self apiPrologueForTuple: aReturnType)
	].

	^$^ ,(self apiWrapperFor: aReturnType) ,' @ ('! !

!NativeObject class methodsFor: #builders!
apiPrologueForArray: aReturnType
	" Private - Return the prologue to use to wrap an arrayed return.
	The returned code will be like:
		GlobalName withAll: (
		GlobalName of: ElementClass withAll: (
		GlobalName of: 'Type1,Type2,Type3' withAll: (
	"

	| type stream kind |
	stream := aReturnType stream.
	type := self apiWrapperFor: (stream upTo: $[).
	kind := stream upTo: $].
	stream atEnd ifFalse: [ ^self error: 'Invalid brackets for array' ].
	kind isEmpty ifTrue: [ ^type ,' withAll: ('].
	(kind includes: $,) ifFalse: [
		^type ,' of: ' ,(self apiWrapperFor: kind) ,' withAll: ('
	].	
	^type ,' of: ' ,kind printString ,' withAll: ('! !

!NativeObject class methodsFor: #builders!
apiPrologueForTuple: aReturnType
	" Private - Return the prologue to use to wrap a tuple return.
	The returned code will be like:
		GlobalName of: 'Type1,Type2,Type3' with: (
	"

	| type stream kind |
	stream := aReturnType stream.
	type := self apiWrapperFor: (stream upTo: $<).
	kind := stream upTo: $>.
	stream atEnd ifFalse: [ ^self error: 'Invalid code for typle' ].
	kind isEmpty ifTrue:  [ ^self error: 'Empty types for tuple' ].
	(kind includes: $,) ifFalse: [
		^type ,' of: ' ,(self apiWrapperFor: kind) ,' with: ('
	].	
	^type ,' of: ' ,kind printString ,' with: ('! !

! NativeObject class methodsFor: #builders !
apiWrapperFor: aName
	" Private - Return the name to use to wrap API structures of type aName. "

	Smalltalk at: aName ifAbsent: [ ^self defaultAPIWrapperFor: aName ].
	^aName! !

! NativeObject class methodsFor: #builders !
defaultAPIWrapperFor: aName
	" Private - Return the default name to use to wrap API structures of type aName.
	Default implementation will defer lookup upto runtime.
	"

	^aName! !

! NativeObject class methodsFor: #builders !
buildFunction: spec
	" Private - Generate code from spec as functions on instance handle.
	spec can be:
		aSymbol (e.g. any smalltalk message selector)
		#( selector #ReturnType )
		#( selector #( argumentNames ) #ReturnType )
		#( #( selector #( argumentNames ) ) #ReturnType )
		#( selector #( argumentNames ) )
	argumentNames must match selector pattern
	argumentNames can include kwargsStart to define keyword arguments
	if argumentNames is not given, it will be computed from selector pattern
	The function will return the result of calling the handle, wrapped to a ReturnType
	See implementation of #apiPrologueFor: for documentation on #ReturnType
	"

	spec isArray ifTrue: [
	    (spec size = 3 and: [spec last isString]) ifTrue: [ "#( selector #( argumentNames ) #ReturnType )"
		^self buildFunction: (Array
			with: (Array with: spec first with: spec second)
			with: spec last)
	    ]. 
	    (spec size = 2 and: [spec last isString]) ifTrue: [ "#( selector #ReturnType )"
		^self
			buildFunction: spec first
			prologue: (self apiPrologueFor: spec second)
			epilogue: (self apiEpilogueFor: spec second)
	    ]. 
	].
	^self buildFunction: spec prologue: $^ epilogue: ''! !

! NativeObject class methodsFor: #builders !
buildFunction: spec prologue: prologue epilogue: epilogue 
	" Private - Generate code for spec to apply on instance handle. "

	| args |
	spec isArray ifTrue: [ "#( selector #(argumentNames) )"
		^self	buildFunction: spec first arguments: spec last
			prologue: prologue epilogue: epilogue
	].
	args :=  (1 to: (spec occurrencesOf: $:)) collect: [:i| #arg , i asString ].
	args size = 1 ifTrue: [ args := #( #anObject ) ].
	^self
		buildFunction: spec arguments: args
		prologue: prologue epilogue: epilogue  ! !

! NativeObject class methodsFor: #builders !
apiSelector: callName forArguments: args
	" Private - Return the api selector to implement a native call with arguments. "

	| parts |
	args size = callName argumentCount ifTrue: [ ^callName ].
	parts := callName asArrayOfSubstringsSeparatedBy: $:.
	parts last isEmpty ifTrue: [ parts removeLast ].
	parts size + 1 to: args size do: [:i| | argName |
		argName := args at: i.
		argName isArray ifTrue: [ argName := argName first ].
		parts add: (argName asSmalltalkName: false)
	].
	^parts inject: '' into: [:total :part| total ,part ,$: ]! !

! NativeObject class methodsFor: #api !
apiArgumentsFrom: array
	" Private - Return the arguments for API call. "

	^(self apiSupportFor: array)
		nativeArgumentsFrom: array! !

! NativeObject class methodsFor: #builders !
nativeArgumentsFrom: array
	" Private - Return the arguments for API call. "

	| tag |
	tag := $_.
	^array collect: [:each|
		(each isString and: [ each first = tag ]) ifTrue: [ | varName |
			varName := each copyFrom: 2 to: each size.
			each last = tag
			ifTrue: [ Array with: (varName copyFrom: 1 to: varName size - 1) with: #handles ]
			ifFalse: [ Array with: varName with: #json ]
		] ifFalse: [ each ]
	]! !

! NativeObject class methodsFor: #builders !
apiSelectorAndFunctionName: callName forArguments: args
	" Private - Return the selector and api call to implement a native call with arguments. "

	| msgSelector apiSelector |
	(callName isString or: [callName isArray]) ifFalse: [
		^self error: 'CallName must be aString or anArray'
	].
	msgSelector := self apiSelector: (callName isString
		ifTrue: [ self apiMessage: callName for: args ]
		ifFalse: [ self apiMessage: callName first for: args ]) forArguments: args.
	apiSelector := self apiSelector: (callName isString
		ifTrue: [ callName ] ifFalse: [ callName last ]) forArguments: args.

	^Array with: msgSelector with: apiSelector! !

! NativeObject class methodsFor: #builders !
apiMessage: callName for: args
	" Private - Return the message to implement for calling function callName.
	The default action use the callName as local name.
	It can be refined by subclasses to edit the name for better smalltalk looking of the message.
	"

	| result |
	result := callName copyWithout: $!!.
	self nativeLanguage = #lua ifTrue: [
		(callName startsWith: $_) ifTrue: [ result := callName copyFrom: 2 to: callName size ].
	].
	result argumentCount > 0 ifTrue: [ ^result ].
	^result asLocalName ! !

! NativeObject class methodsFor: #builders !
apiEncode: delegate
	call: aSelector args: array
	on: code header: header
	" Private - Generate code to call function onto code&header streams. "

	| pair args msgStream apiStream argStream msg space |
	args := self apiArgumentsFrom: array.
	pair := self apiSelectorAndFunctionName: aSelector forArguments: args.
	msgStream := pair first stream.
	space := (pair first includes: $:) ifTrue: [ ': ' ] ifFalse: [ $ ].
	apiStream := pair last stream.
	argStream := args stream.

	code	nextPutAll: delegate.
	code	nextPut: apiStream next.
	header	nextPut: msgStream next.

	[ argStream atEnd ] whileFalse: [ | param expr |
		argStream peek isString
		ifTrue: [ param := argStream next. expr := param ]
		ifFalse: [ param := argStream peek first. expr := param ,' ' ,argStream next last. ].
		code nextPutAll: (apiStream upTo: $:) ,': ' ,expr ,' '.
		header nextPutAll: (msgStream upTo: $:) ,space ,param ,' '.
	].

	code   nextPutAll: apiStream upToEnd.
	header nextPutAll: msgStream upToEnd.! !

! NativeObject class methodsFor: #builders !
apiEncode: delegate
	call: aSelector args: array kwargs: kwarray
	on: code header: header
	" Private - Generate code to call function onto code&header streams. "

	| pair args kwargs msgStream argStream msg space |
	args := self apiArgumentsFrom: array.
	kwargs := self apiArgumentsFrom: kwarray.
	pair := self apiSelectorAndFunctionName: aSelector forArguments: args ,kwargs.
	msgStream := pair first stream.
	space := (pair first includes: $:) ifTrue: [ ': ' ] ifFalse: [ $ ].
	argStream := (args ,kwargs) stream.

	header	nextPut: msgStream next.
	[ argStream atEnd ] whileFalse: [ | param |
		argStream peek isString
		ifTrue: [ param := argStream next ]
		ifFalse: [ param := argStream next first ].
		header nextPutAll: (msgStream upTo: $:) ,space ,param ,' '.
	].
	header nextPutAll: msgStream upToEnd.

	code nextPut: $(.
	code nextPutAll: (self
		apikw: delegate call: pair last
		args: args kwargs: kwargs
		header: header).
	code nextPut: $).! !

! NativeObject class methodsFor: #builders !
apikw: delegate call: aSelector args: array kwargs: kwarray header: header
	" Private - Returns the code for a keyword function call. "

	| callArgs |
	callArgs := 'self call: ' ,(self apikw: aSelector delegate: delegate) ,'
	args: '     ,(self apikwArguments: array named: false).
	kwarray isEmpty ifTrue: [
		(aSelector includes: $!!)
		ifTrue: [ ^callArgs ,' kwargs: nil' ]
		ifFalse: [
			array notEmpty ifTrue: [ header nextPutAll: #kwargs ].
			header nextPutAll: ': kwargs'.
			^callArgs ,' kwargs: kwargs'
		]
	].
	^callArgs ,' kwargs: ' ,(self apikwArguments: kwarray named: true)! !

! NativeObject class methodsFor: #builders !
apikw: aSelector delegate: delegate
	" Private - Returns the code to access a keyword function. "

	^$(
	,(delegate upTo: $#)
	,'basicAt: '	,(aSelector upTo: #:) asLiteral
	,$)! !

! NativeObject class methodsFor: #builders !
kwargsStart
	" Private - Returns the tag used for keyword arguments. "

	^#|! !

! NativeObject class methodsFor: #builders !
apikwArguments: array named: named
	" Private - Returns the code to pass arguments to kw function call. "

	| element |
	element := [:arg| | kw v |
		arg isArray
		ifTrue: [ kw := arg first. v := arg first ,' ',arg last ]
		ifFalse: [ kw := v := arg ].
		named ifTrue: ['(Array with: ' ,kw asLiteral,' with: ' ,v ,$) ]
		ifFalse: [ v ]
	].
	array isEmpty ifTrue: [ ^'#()'].
	array size <= 3 ifTrue: [
		^(array inject: $( ,#Array into: [:total :each|
			total,' with: ',(element value: each)
		]) ,$)
	].
	^'(Array new'
	,(array inject: '' into: [:total :each|
		total,' add: ',(element value: each) ,$;
	]) ,' yourself)' ! !

! NativeObject class methodsFor: #builders !
apiEncode: delegate
	call: aSelector arguments: array
	on: code header: header
	" Private - Generate code to call function onto code&header streams. "

	((aSelector includes: $!!) or: [array includes: self kwargsStart]) ifTrue: [
		| stream | stream := array readStream.
		^self	apiEncode: delegate call: aSelector
			args: (stream upTo: self kwargsStart) kwargs: stream upToEnd
			on: code header: header
	].
	^self	apiEncode: delegate
		call: aSelector args: array
		on: code header: header! !

! NativeObject class methodsFor: #private !
codeForInstanceReceiver
	" Private - Return the code for the receiver of instance function call.
	Default answer asumes a wrapper calling with syntax sugar.
	This method can be refined to adapt to custom platform calls.
	"

	^'self handle #'! !

! NativeObject class methodsFor: #builders !
buildFunction: spec arguments: args prologue: prologue epilogue: epilogue 
	" Private - Generate code to delegate message on instance handle. "

	| header code |
	(args includes: self kwargsStart) ifTrue: [ | array |
		array := args stream upTo: self kwargsStart.
		args last = self kwargsStart
		ifTrue: [ self	buildFunction: spec arguments: array
				prologue: prologue epilogue: epilogue ]
		ifFalse: [
			array add: self kwargsStart.
			self	buildFunction: spec arguments: array
				prologue: prologue epilogue: epilogue 
		].
	].
	header := '' writeStream. code := '' writeStream.
	self
	  apiEncode: self codeForInstanceReceiver
		call: spec arguments: args
		on: code header: header;
	  implement: (self stMessageHeader: header contents)
		comment: 'Delegate to handle.'
		as:	((prologue
			, code contents
			, epilogue) replace: $!! with: $!! ,$!!)
		category: #wrappers! !

! NativeObject class methodsFor: #private !
codeForStaticReceiver
	" Private - Return the code for the receiver of static function call.
	Default answer asumes a nativeClass wrapper calling with syntax sugar.
	This method can be refined to adapt to custom platform calls.
	"

	^'self nativeClass handle #'! !

! NativeObject class methodsFor: #private !
nativeClass
	" Private - Return the wrapper for native binding of the receiver. "

	^self implementedBySubclass: #nativeClass! !

! NativeObject class methodsFor: #builders !
buildConstants
	" Build the methods for constants defined in the receiver's library. "

	^self buildConstants: self constants! !

! NativeObject class methodsFor: #builders !
buildConstants: constants
	" Build the methods for constants. "

	^self build: #Constants -> constants doing: [
		constants isString ifTrue: [ | stream |
			stream := constants stream.
			[stream atEnd] whileFalse: [
				self buildConstant: stream nextLine trimBlanks
			]
		] ifFalse: [ constants do: [:each| self buildConstant: each ] ]
	]! !

! NativeObject class methodsFor: #builders !
buildConstant: spec
	" Private - Build the accessor for constant spec. "

	| selector |
	spec isEmpty ifTrue: [ ^self ].
	selector := (spec fromLast: $.) asSmalltalkName: false.
	self class
		implement: selector
		comment: 'Return the constant value for ' ,spec
		as: $^ ,(self codeToGetConstant: spec) category: #constants! !

! NativeObject class methodsFor: #private !
library
	" Private - Returns the library of the receiver.
	A new instance is returned by default.
	This method can be refined by subclasses (e.g. implementing singleton). "

	^self new ! !

! NativeObject class methodsFor: #builders !
codeToGetConstant: spec
	" Private - Returns the code to get the constant value for spec. "

	spec isArray ifTrue: [
		^spec last ,' @ (' ,(self codeToGetConstant: spec first) ,$)
	].
	^'self library @ ' ,spec asLiteral ! !

! NativeObject class methodsFor: #builders !
constants
	" Returns the constants implemented by receiver's library. "

	^self constantsIn: self library! !

! NativeObject class methodsFor: #builders !
constantsIn: aNativeObject
	" Returns the constants implemented by aNativeObject. "

	^((aNativeObject keys
		select: [:key| #(Number String) includes: (self typeOf: aNativeObject@key) ])
		reject: [:key| key startsWith: $_ ]) sorted! !

! NativeObject class methodsFor: #api !
newInstanceOf: aName withAll: handlesArray
	" Private - Returns the handle result of instatiating native object of type aName. "

	| fn |
	fn := self instantiatorFor: aName.
	{'self'}.
	^fn evaluateWithArguments: handlesArray! !

! NativeObject class methodsFor: #api !
instantiatorFor: aName
	" Private - Returns the function to evaluate when creating native instance. "

	^self library @ aName! !

! NativeObject class methodsFor: #api !
instantiate: aNativeObjectClass withAll: handlesArray
	" Returns the result of instatiating aCVObjectClass. "

	^aNativeObjectClass @ (self
		newInstanceOf: aNativeObjectClass name
		withAll: handlesArray)! !

! NativeObject class methodsFor: #builders !
buildSubclasses: classNames in: packageName
	" Private - Generate subclasses of the receiver. "

	^self buildSubclasses: classNames in: packageName reference: nil! !

! NativeObject class methodsFor: #builders !
buildSubclasses: classSpecs in: packageName reference: urlSite
	" Private - Generate subclasses of the receiver. "

	classSpecs do: [:spec|
		self buildSubclass: spec in: packageName reference: urlSite
	]! !

! NativeObject class methodsFor: #builders !
buildSubclass: spec in: packageName
	" Private - Generate a subclass of the receiver. "

	^self buildSubclass: spec in: packageName reference: nil! !

! NativeObject class methodsFor: #builders !
urlReference
	" Private - Returns the URL of package site. "

	^#none ! !

! NativeObject class methodsFor: #builders !
urlSelector
	" Private - Returns the selector for urlSite. "

	^nil ! !

! NativeObject class methodsFor: #builders !
urlSelectors
	" Private - Returns the selector for urlSite. "

	^Array with: self urlSelector ! !

! NativeObject class methodsFor: #builders !
referenceTo: topic in: packageName at: urlSite
	" Private - Returns the URL to find documentation on topic. "

	urlSite = self urlSelector ifTrue: [
		^self	referenceTo: topic in: packageName
			at: self urlReference
	].
	^urlSite ,$/
		,(packageName notNil ifTrue: [ (packageName replace: $. with: $/) ,$/ ] ifFalse: [''])
		,topic ,'.html'! !

! NativeObject class methodsFor: #builders !
buildSubclass: spec in: packageName reference: urlSite
	" Private - Generate a subclass of the receiver. "

	| cls className |
	spec isString ifFalse: [ "spec can be #(superClassName className...) "
		cls := spec first.
		cls isString ifTrue: [
			cls := Smalltalk at: cls ifAbsent: [
				self buildSubclass: cls in: packageName reference: urlSite
			]
		] ifFalse: [ self error: 'Invalid superClass name: ',cls ].
		^(spec copyFrom: 2 to: spec size) collect: [:each|
			cls buildSubclass: each in: packageName reference: urlSite
		]
	].
	className := spec.
	cls := packageName notNil
		ifTrue: [ self subclass: className category: packageName ]
		ifFalse: [ self subclass: className ].
	(urlSite isNil and: [packageName notNil and: [
		self urlSelectors includes: (packageName upTo: $.)
		]]) ifTrue: [ urlSite := self urlSelector ].
	(urlSite notNil and: [urlSite ~= #none]) ifTrue: [
		cls comment: 'See ' ,(self referenceTo: className in: packageName at: urlSite)
	].
	(cls respondsTo: #packagePrefix:) ifTrue: [
		cls packagePrefix: packageName
	].
	^cls! !

! NativeObject class methodsFor: #builders !
buildConstructors: selectors
	" Generate code for constructors implemented by native class. "

	self build: #Constructors -> selectors doing: [
		selectors do: [:each| self buildConstructor: each ]
	]! !

! NativeObject class methodsFor: #builders !
buildConstructor: selector
	" Private - Generate code for constructor of native class. "

	selector isArray ifTrue: [
	    (selector size = 3 and: [selector last isString]) ifTrue: [ "#( selector #( argumentNames ) #RETURNClass )"
		^self buildStaticFunction: selector
	    ]. 
	    (selector size = 2 and: [selector last isString]) ifTrue: [ "#( selector #RETURNClass )"
		^self buildStaticFunction: selector
	    ]. 
	].
	^self buildConstructorFor: selector! !

! NativeObject class methodsFor: #builders !
buildConstructorFor: selector
	" Private - Generate code for constructor of native class. "

	| args |
	selector isArray ifTrue: [ "#( selector #(argumentNames) )"
		^self	buildConstructor: selector first
			arguments: selector last
	].
	args :=  (1 to: (selector occurrencesOf: $:)) collect: [:i| #arg , i asString ].
	args size = 1 ifTrue: [ args := #( #anObject ) ].
	^self buildConstructor: selector arguments: args! !

! NativeObject class methodsFor: #builders !
constructorSupport
	" Private - Returns the instantiation support for constructor. "

	^#self! !

! NativeObject class methodsFor: #builders !
buildConstructor: aSelector arguments: array
	" Private - Generate code for constructor of native class. "

	| selector args header code input argStream msg long adder comma |
	args := self apiArgumentsFrom: array.
	selector := self apiSelector: aSelector forArguments: args.
	long := args size > 4.
	input := selector stream.
	argStream := args stream.
	header := '' writeStream. code := '' writeStream.
	code nextPutAll: (long ifTrue: [ '#()' ] ifFalse: [ #Array ]).
	adder := long ifTrue: [' add: '] ifFalse: [' with: '].
	comma := long ifTrue: [$;] ifFalse: [''].
	header nextPut: input next asLowercase.
	argStream atEnd ifTrue: [ code nextPutAll: ' new' ].
	[ argStream atEnd ] whileFalse: [ | param expr |
		argStream peek isString
		ifTrue: [ param := argStream next. expr := param ]
		ifFalse: [ param := argStream peek first. expr := param ,' ' ,argStream next last. ].
		msg := (input upTo: $:) ,': '.
		code   nextPutAll: adder ,expr ,comma.
		header nextPutAll: msg ,param ,' '.
	].
	code nextPutAll: (long ifTrue: [ ' yourself' ] ifFalse: [ '' ]).
	msg := input upToEnd.
	code nextPutAll: msg.
	header nextPutAll: msg.
	self class
		implement: (self stMessageHeader: header contents)
		  comment: 'Returns an instance of the receiver.'
		  as: $^ ,self constructorSupport ,' instantiate: self withAll: (' ,code contents ,')'
		  category: #instantiation! !

! NativeObject class methodsFor: #builders !
buildStaticFunctions: specs
	" Generate code from specs as functions implemented implemented by native class. "

	self build: #StaticFunctions -> specs doing: [
		specs do: [:each| self buildStaticFunction: each ]
	]! !

! NativeObject class methodsFor: #builders !
buildStaticFunction: spec
	" Private - Generate code to delegate message implemented as functions on native class. "

	spec isArray ifTrue: [
	    (spec size = 3 and: [spec last isString]) ifTrue: [ "#( selector #( argumentNames ) #RETURNClass )"
		^self buildStaticFunction: (Array
			with: (Array with: spec first with: spec second)
			with: spec last)
	    ]. 
	    (spec size = 2 and: [spec last isString]) ifTrue: [ "#( selector #RETURNClass )"
		^self
			buildStaticFunction: spec first
			prologue: (self apiPrologueFor: spec second)
			epilogue: (self apiEpilogueFor: spec second)
	    ].
	].
	^self buildStaticFunction: spec prologue: $^ epilogue: ''! !

! NativeObject class methodsFor: #builders !
buildStaticFunction: spec prologue: prologue epilogue: epilogue 
	" Private - Generate code to delegate message on native class. "

	| args |
	spec isArray ifTrue: [ "#( selector #(argumentNames) )"
		^self	buildStaticFunction: spec first
			arguments: spec last
			prologue: prologue epilogue: epilogue
	].
	args :=  (1 to: (spec occurrencesOf: $:)) collect: [:i| #arg , i asString ].
	args size = 1 ifTrue: [ args := #( #anObject ) ].
	^self
		buildStaticFunction: spec arguments: args
		prologue: prologue epilogue: epilogue ! !

! NativeObject class methodsFor: #builders !
buildStaticFunction: spec arguments: args prologue: prologue epilogue: epilogue 
	" Private - Generate code to delegate message on native class. "

	| header code |
	(args includes: self kwargsStart) ifTrue: [ | array |
		array := args stream upTo: self kwargsStart.
		args last = self kwargsStart
		ifTrue: [ self	buildStaticFunction: spec arguments: array
				prologue: prologue epilogue: epilogue ]
		ifFalse: [
			array add: self kwargsStart.
			self	buildStaticFunction: spec arguments: array
				prologue: prologue epilogue: epilogue 
		].
	].
	header := '' writeStream. code := '' writeStream.
	self
	  apiEncode: self codeForStaticReceiver
		call: spec arguments: args
		on: code header: header.
	self class
		implement: (self stMessageHeader: header contents)
		  comment: 'Delegated to native class.'
		  as:	((prologue
			, code contents
			, epilogue) replace: $!! with: $!! ,$!!)
		  category: #wrappers! !

! NativeObject class methodsFor: #builders !
buildNativeFunctions: specs
	" Generate code from specs as raw functions implemented by instance handle. "

	self build: #NativeFunctions -> specs doing: [ | tuples |
		tuples := specs collect: [:spec|
			spec isString ifTrue: [ $_ ,spec ] ifFalse: [ | selector |
				selector := spec first isString ifTrue: [$_ ,spec first] ifFalse: [
					spec first copy at: 1 put: ($_ ,spec first first); yourself
				].
				(Array with: selector), (spec copyFrom: 2 to: spec size)
			]
		].
		tuples do: [:spec| self buildFunction: spec ]
	]! !

! NativeObject class methodsFor: #builders !
buildNativeStaticFunctions: specs
	" Generate code from specs as raw functions implemented by class handle. "

	self build: #NativeStaticFunctions -> specs doing: [ | tuples |
		tuples := specs collect: [:spec|
			spec isString ifTrue: [ $_ ,spec ] ifFalse: [ | selector |
				selector := spec first isString ifTrue: [$_ ,spec first] ifFalse: [
					spec first copy at: 1 put: ($_ ,spec first first); yourself
				].
				(Array with: selector), (spec copyFrom: 2 to: spec size)
			]
		].
		tuples do: [:spec| self buildStaticFunction: spec ]
	]! !

! NativeObject class methodsFor: #builders !
buildMethods: specs
	" Generate code from specs as OO methods implemented by instance handle. "

	self build: #Methods -> specs doing: [
		self buildFunctions: specs
	]! !

! NativeObject class methodsFor: #instantiation !
@ handle
	" Return an instance of the receiver. "

	^self basicNew handle: handle! !

! NativeObject methodsFor: #accessing !
@@ path
	" Return the property at path of the receiver (or nil). "

	path isString ifTrue: [
		^self @@ (path asArrayOfSubstringsSeparatedBy: $.)
	].
	^path inject: self handle into: [:h :each|
		h notNil ifTrue: [ h basicAt: each ]
	]! !

! NativeObject methodsFor: #accessing !
<@ path
	" Return the native property at path of the receiver (or nil). "

	^NativeHandle @ (self @@ path)! !

! NativeObject class methodsFor: #instantiation !
empty
	" Return an empty instance of the receiver. "

	^self @ Object new json! !

! NativeObject class methodsFor: #instantiation !
for: handle
	" Return an instance of the receiver. "

	^self basicNew handle: handle! !

! NativeObject class methodsFor: #instantiation !
withAll: handles
	" Return a collection of instance of the receiver with handles. "

	^Array @ handles collect: [:each| self @ each ]! !

! NativeObject class methodsFor: #instantiation !
of: aType withAll: handles
	" Return an instance of the receiver wrapping handles with a type (or nil).
	Default implementation ignore the type.
	This method can be refined by subclasses to support typed wrappers.
	aType can be a class for type of elements or aString of comma separated type names.
	"

	^self withAll: Array @ handles! !

! NativeObject class methodsFor: #instantiation !
of: aType with: aHandle
	" Return an instance of the receiver wrapping aHandle with a type (or nil). "

	aHandle isNil ifTrue: [ ^nil ].
	^self basicOf: aType with: aHandle! !

! NativeObject class methodsFor: #instantiation !
basicOf: aType with: aHandle
	" Private - Return an instance of the receiver wrapping aHandle with a type.
	aType can be a class for type of element or aString of comma separated type names.
	"

	^self implementedBySubclass: #basicOf:with:! !

! NativeObject methodsFor: #converting !
as: aNativeObjectClass
	" Return the receiver casted to aNativeObjectClass. "

	| result |
	(self isKindOf: aNativeObjectClass) ifTrue: [ ^self ].
	result := aNativeObjectClass @ self handle.
	result == self ifTrue: [ ^self ].
	result copyPropertiesFrom: self.
	^result! !

! NativeObject methodsFor: #accessing !
handle
	" Return the native handle. "

	handle isNil ifTrue: [ self initializeHandle ].
	^handle! !

! NativeObject methodsFor: #private !
handleOrNil
	" Return the native handle (or nil). "

	^handle! !

!NativeObject methodsFor: #testing!
= anObject

	self == anObject ifTrue: [ ^true ].
	(self isObject: anObject) ifFalse: [ ^false ].
	^(anObject respondsTo: #handleOrNil) and: [
		handle == anObject handleOrNil
	] ! !

! NativeObject methodsFor: #accessing !
handle: anAlien
	" Set the native handle to a potentialy native(non-S8) object.
	Return the receiver, or nil if the receiver is void.
	"

	handle := anAlien.
	handle isNil ifTrue: [ ^nil ].! !

! NativeObject methodsFor: #converting !
jsonHiddenNames
	" Private - Return the names of the instance variables to skip when mapping the receiver to JSON structure. "

	^#( #handle )! !

! NativeObject methodsFor: #converting !
basicJson
	" Convert the receiver to JSON structure. "

	^super json! !

! NativeObject methodsFor: #converting !
json
	" Convert the receiver to JSON structure. "

	^self handle! !

! NativeObject methodsFor: #private !
jsObject
	" Private - Return the (native)Object. "

	^{''}! !

! NativeObject methodsFor: #accessing !
ownPropertyNames
	" Return the owned property names of the receiver. "

	{''}! !

! NativeObject methodsFor: #accessing !
@ propertyName
	" Return the property at propertyName of the receiver. "

	^self handleAt: propertyName! !

! NativeObject methodsFor: #accessing !
handleAt: propertyName
	" Return the property at propertyName of the receiver. "

	^handle[propertyName]! !

! NativeObject methodsFor: #accessing !
handleAt: propertyName put: anObject
	" Set the property at propertyName of the receiver. "

	^handle[propertyName] := anObject! !

! NativeObject methodsFor: #initialize !
initializeHandle
	" Private - Initialize the native handle.
	This method must be implemented by subclasses.
	"

	^self error: 'Missing '
		,self class name
		,'>>#initializeHandle'! !

! NativeObject methodsFor: #release !
releaseHandle
	" Private - Release the native handle.
	The default implementation clear handle.
	This method can be refined by subclasses to free external resources.
	"

	handle := nil! !

! NativeObject methodsFor: #release !
release
	" Private - Release the receiver. "

	self releaseHandle.
	^super release! !

! NativeObject methodsFor: #copying !
copyPropertiesFrom: aNativeObject
	" Private - Copy receiver's properties from aNativeObject. "

	aNativeObject ownPropertyNames do: [:key|
		self basicAt: key put: (aNativeObject basicAt: key)
	]! !

! NativeObject methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self! !

! NativeObject methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self! !

! NativeObject class methodsFor: #events !
html5Events
	" Private - Return the html5 events+notifications managed by instances of the receiver.
	Default answer is empty, this method can be refined by subclasses to handle html5 events.
	"
	^#()! !

! NativeObject class methodsFor: #events !
localEvents
	" Private - Return the smalltalk events managed by instances of the receiver. "

	^#()! !

! NativeObject class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^self html5Events
	,self localEvents
	,super constructEventsTriggered! !

! NativeObject class methodsFor: #builder !
buildHTML5Events: selectors
	" Private - Generate code to route events to handle. "

	self build: #HTML5Events -> selectors doing: [
		self class
			implement: #html5Events
			comment: 'Private - Return the html5 events managed by instances of the receiver.'
			as: '^#( ' ,(selectors inject: '' into: [:total :each|
				total ,'#',each ,' '
				]),') , super html5Events'
			category: #events.
		self buildGetters: (selectors collect: [:each|
			self selectorForHTML5Event: each
		]).
	]! !

! NativeObject methodsFor: #events !
closureForEvent: selector
	" Private - Return the closure to handle event selector. "

	| argumentCount |
	argumentCount := selector argumentCount.
	argumentCount = 0 ifTrue: [ ^[ self triggerEvent: selector ] ].
	argumentCount = 1 ifTrue: [ ^[:arg1| self triggerEvent: selector with: arg1 ] ].
	argumentCount = 2 ifTrue: [ ^[:arg1 :arg2| self triggerEvent: selector with: arg1 with: arg2 ] ].
	argumentCount = 3 ifTrue: [ ^[:arg1 :arg2 :arg3| self triggerEvent: selector withArguments: (Array with: arg1 with: arg2 with: arg3) ] ].
	argumentCount = 4 ifTrue: [ ^[:arg1 :arg2 :arg3 :arg4| self triggerEvent: selector withArguments: (Array with: arg1 with: arg2 with: arg3 with: arg4) ] ].
	^self error: 'Invalid argument count'! !

! NativeObject methodsFor: #events !
when: selector evaluate: anAction
	" Register anAction to the be evaluated when the receiver triggers the event. "

	| result needsRegistration |
	needsRegistration := (self eventTable at: selector ifAbsent: [nil]) isNil.
	result := super when: selector evaluate: anAction.
	needsRegistration ifTrue: [
		self	registerEvent: selector
			doing: (self closureForEvent: selector).
	].
	^result! !

! NativeObject methodsFor: #events !
registerEvent: selector doing: closure
	" Private - Register basic event handler. "

	| onSelector |
	onSelector := self class selectorForEvent: selector.
	onSelector isNil ifTrue: [ ^self ]. "local/smalltalk event"
	^self bindEvent: onSelector doing: closure! !

! NativeObject methodsFor: #events !
bindEvent: onSelector doing: closure
	" Private - Bind the event handler. "

	| native |
	native := self handle.
	native[onSelector] isNil ifTrue: [
		native[onSelector] := closure
	] ifFalse: [
		native[onSelector] #addListener: closure json
	]! !

! NativeObject class methodsFor: #events !
selectorForEvent: selector
	" Private - Return the on... selector to use to register binding to selector. "

	(self localEvents includes: selector) ifTrue: [ ^nil ].
	^(self html5Events includes: selector) ifTrue: [
		self selectorForHTML5Event: selector
	]! !

! NativeObject class methodsFor: #events !
selectorForHTML5Event: selector
	" Private - Return the on... selector to use to register binding to selector. "

	^#on , ((selector upTo: $:) asSmalltalkName: true)! !


"----------------------------------------------------------------------------------------"
NativeObject
	subclass: #NativeHandle
	instanceVariableNames: ''
	classVariableNames: #Support
	category: #Natives!
NativeHandle comment: '
	Wrapper for native objects.
	@2018 Alejandro F. Reimondo
	aleReimondo@smalltalking.net
'!

! NativeHandle class methodsFor: #instantiation !
supportsLanguage: language
	" Private - Return true if instances of the receiver implement native wrappers for language.
	Returns false. This method may be refined by subclasses. "

	^false! !

! NativeHandle class methodsFor: #instantiation !
support
	" Private - Return the support for hosting execution environment. "

	| lang |
	Support notNil ifTrue: [ ^Support ].
	lang := self nativeLanguage.
	NativeHandle allSubclasses do: [:cls|
		(cls supportsLanguage: lang) ifTrue: [
			Support := cls.
			^Support
		].
	].
	^self error: 'Native handle support missing'! !

! NativeHandle class methodsFor: #converting !
@> nativeObject
	" Return the description of nativeObject. "

	| h |
	h := self @ nativeObject.
	^h notNil
		ifTrue: [ h description ]
		ifFalse: [ #nil ]! !

! NativeHandle class methodsFor: #private !
instanceHandleFromString: aString
	" Private - Return a (native)handle obtained from aString (or nil). "

	^Smalltalk nativeObjectAt: aString ifAbsent: []! !

! NativeHandle class methodsFor: #instantiation !
@ nativeObject
	" Return an instance of the receiver. "

	self == NativeHandle ifTrue: [ ^self support @ nativeObject ].
	((self isObject: nativeObject) and: [ nativeObject isString ]) ifTrue: [
		^super @ (self instanceHandleFromString: nativeObject)
	].
	^super @ nativeObject! !

! NativeHandle class methodsFor: #utilities !
chunks
	" Return a NativeReader to fileIn properties in new native object. "

	^NativeReader new! !

! NativeHandle methodsFor: #accessing !
keys
	" Return the sorted property names of the receiver. "

	^self propertyNames sorted! !

! NativeHandle methodsFor: #accessing !
properties
	" Return an array with property name and values. "

	^self keys collect: [:each|
		Array with: each with: (self handleAt: each)
	]! !

! NativeHandle methodsFor: #iterating !
propertiesDo: aBlock
	" Evaluate aTwoArgumentBlock with each property (name & value) of the receiver. "

	self propertyNames do: [:each| aBlock value: each value: (self handleAt: each) ]! !

! NativeHandle class methodsFor: #converting !
@>> nativeObject
	" Return the (owned) description of nativeObject. "

	| h |
	h := self @ nativeObject.
	^h notNil
		ifTrue: [ h description: 1 owned: true ]
		ifFalse: [ #nil ]! !

! NativeHandle methodsFor: #debugging !
description: level
	" Return the description of the receiver. "

	^self description: level owned: false ! !

! NativeHandle methodsFor: #debugging !
description: level owned: owned
	" Return the description of the receiver. "

	| text data cr | 
	text := [:h|
		h isNil ifFalse: [
			(self isObject: h) ifTrue: [
				(self typeOf: h) == #function
				ifTrue: [ $( ,(h toString upTo: ${) ,'{...})' ]
				ifFalse: [ h printString ]
			] ifFalse: [
				[ level > 1
					ifTrue: [ (self class @ h) description: level - 1 owned: owned ]
					ifFalse: [ self toString: h ]
				] on: Exception do: [:err| '???' ]
			]
		]
	].
	data := [:key| | value |
		[ value := handle basicAt: key ] on: Exception do: [:ex|
			value := '[',key,'] ' ,ex description
		]. value
	].
	cr := '
'.	^((owned
	ifTrue: [self ownPropertyNames]
	ifFalse: [self propertyNames]) sorted
		inject: ${ into: [:total :each| total ,cr
 			,each ,': '
			,(text value: (data value:each))
			,';' ]
		)	,cr ,$}! !

! NativeHandle methodsFor: #debugging !
description
	" Return the description of the receiver. "

	^self description: 1! !

! NativeObject methodsFor: #debugging !
description
	" Return a string description from receiver's properties. "

	^self asNativeHandle description! !

! NativeHandle methodsFor: #private !
isFunction: aHandle
	" Return true if aHandle is a function. "

	^(self typeOf: aHandle) == #function! !

! NativeHandle methodsFor: #accessing !
functionNames
	" Return the names of property of the receiver of type function. "

	^self propertyNames select: [:each| self isFunction: (self handleAt: each) ]! !

! NativeHandle methodsFor: #accessing !
functions
	" Return the functions of the receiver as a collection of tuples #(functionName aClosure). "

	^self functionNames collect: [:each|
		Array with: each with: (self handleAt: each)
	]! !

! NativeHandle methodsFor: #accessing !
ownFunctionNames
	" Return the names of functions owned/implemented by the receiver. "

	^self ownPropertyNames select: [:each| self isFunction: (self handleAt: each) ]! !

! NativeHandle methodsFor: #testing !
hasOwnProperty: aName
	" Return true if receiver own a property named aName. "

	^{''}! !

! NativeHandle methodsFor: #actions !
delete: aName
	" Delete the property named aName in the receiver. "

	{''}! !

! NativeObject methodsFor: #converting !
asDictionary
	" Convert the receiver to a(Pool)Dictionary. "

	| result |
	result := PoolDictionary new.
	self ownPropertyNames do: [:key| result at: key put: handle[key] ].
	^result! !

! NativeObject methodsFor: #converting !
asNativeHandle
	" Convert the receiver to NativeHandle instance (for reflection). "

	^NativeHandle @ self handle! !

! NativeObject methodsFor: #converting !
asJS
	" Convert the receiver to JS instance (for reflection). "

	^NativeHandle @ self handle! !

! Object methodsFor: 'Natives-converting' !
asJS
	" Convert the receiver to JS instance (for reflection). "

	^NativeHandle @ self! !

! Object methodsFor: 'Natives-converting' !
asNativeHandle
	" Convert the receiver to NativeHandle instance (for reflection). "

	^NativeHandle @ self! !

! Array methodsFor: 'Natives-converting' !
handles
	" Return the result of collecting the handles of the receiver. "

	^self collect: [:each| each handle ]! !

! Collection methodsFor: 'Natives-converting' !
handles
	" Return an array with the handles of the receiver. "

	^self asArray handles! !

! UndefinedObject methodsFor: 'Natives-converting' !
handles
	" Return the result of collecting the handles of the receiver. "

	^self! !

! Collection class methodsFor: 'Natives-instantiation' !
of: aType withAll: handles
	" Return an instance of the receiver wrapping handles with a type (or nil). "

	(aType includes: $,) ifTrue: [ | stream |
		stream := aType stream.
		^handles collect: [:h| | cls |
			cls := stream upTo: $,.
			cls isEmpty ifTrue: [ h ] ifFalse: [
				cls := Smalltalk at: cls ifAbsent: [
					self error: 'Missing global ' ,cls
				].
				cls @ h
			]
		]
	].
	^self withAll: (Array @ handles collect: [:each| aType @ each ])! !

! Array class methodsFor: 'Natives-instantiation' !
@@ hArray
	" Return an instance of the receiver wrapping native handle. "

	{''}.
	^hArray! !

! Array class methodsFor: 'Natives-instantiation' !
@ natives
	" Return an instance of the receiver containing the values in native object. "

	| result |
	result := self new.
	natives isNil ifTrue: [ ^result ].
	{''}.
	^result! !

! Object methodsFor: 'Natives-converting' !
handleOf: nativeObject
	" Private - Return the handle of nativeObject (or the nativeObject). "

	^nativeObject notNil ifTrue: [
		((self isObject: nativeObject)
		and: [ nativeObject[#handle] notNil ])
		ifTrue: [ nativeObject handle ]
		ifFalse: [ nativeObject ]
	] ifFalse: [ NativeObject undefined ]! !

! UndefinedObject methodsFor: 'Natives-converting' !
handle
	" Return the handle of the receiver. "

	^self asJSONObject! !

! Boolean class methodsFor: 'Natives-converting' !
@ aBOOL
	" Return aBOOL as a Boolean instance. "

	{''}.
	^false! !

! String class methodsFor: 'Natives-converting' !
@ aHandle
	" Return an instance of the receiver from aHandle. "

	^aHandle notNil ifTrue: [ self toString: aHandle ]! !

! Number class methodsFor: 'Natives-converting' !
@ aHandle
	" Return an instance of the receiver from aHandle. "

	^aHandle notNil ifTrue: [ aHandle asNumber ]! !

! Number methodsFor: 'Natives-converting' !
int32
	" Return the receiver to be passed to API call as 32 bits integer value. "

	{'self'}.
	^self! !

! Number methodsFor: 'Natives-converting' !
int64
	" Return the receiver to be passed to API call as 64 bits integer value. "

	{'self'}.
	^self! !

! Number methodsFor: 'Natives-converting' !
float32
	" Return the receiver to be passed to API call as 32 bits float value. "

	{'self'}.
	^self! !

! Number methodsFor: 'Natives-converting' !
float64
	" Return the receiver to be passed to API call as 64 bits float value. "

	{'self'}.
	^self! !