" Adaptions to run S8 on Python execution machinery.
Copyright (C) 2021-today Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
In case of questions and bug reports please join us at S8 Facebook group
	https://www.facebook.com/groups/s8smalltalk/
"!

"------------------------------------------------------------------------------
	Python code generation
 ------------------------------------------------------------------------------"!
NativeEncoder
	subclass: #PythonEncoder
	instanceVariableNames: ' blocks scopes '
	category: #Python !

! PythonEncoder class methodsFor: #instantiation !
supportsLanguage: aLanguage
	" Private - Return true if instances of the receiver can encode to aLanguage. "

	^aLanguage = #py ! !

! PythonEncoder methodsFor: #accessing !
nativeLanguage
	" Return the nativeLanguage of the receiver. "

	^#py ! !

! PythonEncoder methodsFor: #encoding !
nativeUndefined
	" Private - Return the code for undefined. "

	^#nil ! !

! PythonEncoder methodsFor: #encoding !
nativeNameOfMetaClass: aClass
	" Private - Return the native name for metaclass aClass. "

	^aClass instanceClass name ,'.__class__' ! !

! PythonEncoder methodsFor: #private !
startCompiling: aNode
	" Private - Set the receiver to start emmiting code for aNode. "

	blocks := scopes := nil.
	^super startCompiling: aNode! !

! PythonEncoder methodsFor: #constants !
separatorOp
	" Private - Return the OP for separating statements. "

	^$ ! !

! PythonEncoder methodsFor: #optimizing !
macroStream: macro
	" Private - Return a code stream from macro string. "

	^(super macroStream: macro)
		skipSeparators;
		yourself! !

! PythonEncoder methodsFor: #optimizing !
basicMacro: macro in: input with: aNode
	" Private - Visit aNode encoding/expanding basic macros in input. "

	(#( #true #false #> #< #, ) includes: macro) ifTrue: [
		macro = #> ifTrue: [ ^self indent; cr ].
		macro = #< ifTrue: [ ^self dedent; cr ].
		macro = #, ifTrue: [ ^self nextPutAll: 'smalltalk.object_comma_' ].
		^self nextPutAll: 'smalltalk.' ,macro "boolean constants"
	].
	(macro first isDigit or: [macro startsWith: $-]) ifTrue: [ " %aNumber% "
		^self nextPutAll: 'smalltalk.Number(' ,macro ,$)
	].
	^super basicMacro: macro in: input with: aNode ! !

! PythonEncoder methodsFor: #generation !
emmitInlinedCode: code
	" Private - Emmit inlined code onto output stream. "

	| input |
	input := code readStream.
	[input atEnd] whileFalse: [
		self emmitCode: input nextLine.
		input atEnd ifFalse: [ self cr ]
	]! !

! PythonEncoder methodsFor: #generation !
emmitFunctionStart: fName
	" Private - Emmit the code of function start definition. " ! !

! PythonEncoder methodsFor: #inlines !
missingNativeCodeFor: aNode as: lang
	" Private - Return the native language code for missing expression (or nil). "

	^'self.invalidMessage_(smalltalk.String("' ,compiler selector ,'"))'! !

! PythonEncoder methodsFor: #accessing !
callEscape
	" Private - Return the selector sequence to escape OO message send selector, and produce code for function call.
	Note that an alien selector starting with two $$ are understood as a function call to function starting with $
	"

	^'#_'! !

! PythonEncoder methodsFor: #private !
canExportMethod: aMethod
	" Private - Return true if the receiver can/want to export method aMethod. "

	^true! !

! PythonEncoder methodsFor: #converting !
asSelector: aSelector
	" Return the native representation of aSelector. "

	^aSelector asPythonSelector! !

! PythonEncoder methodsFor: #private !
nativeSourceFor: aMethod
	" Private - Return the code of native function implementing aMethod. "

	^(self canExportMethod: aMethod) ifTrue: [
		self privateSourceFor: aMethod
	]! !

! PythonEncoder methodsFor: #export !
export: aClass method: aMethod on: aStream
	" Export a method definition of aClass onto aStream. "

	| fName selectorName bar source |
	source := self nativeSourceFor: aMethod.
	source isNil ifTrue: [ ^false ].
	aMethod isNil ifTrue: [ self error: 'Method is missing' ].
	bar := $".
	fName := aMethod selector.
	fName first isBackSlash ifTrue: [
		selectorName := self escaped: aMethod selector.
		fName := self escaped: (self asSelector: fName).
	] ifFalse: [
		selectorName := bar ,aMethod selector, bar.
		(self asSelector: fName) = (aMethod selector replace: $: with: $_) ifTrue: [
			fName := 0
		] ifFalse: [
			fName := bar ,(self asSelector: fName) ,bar
		].
	].
	aStream
		nextPutAll: 'smalltalk.bind(' ,(self smalltalkNameFor: aClass)
			,',' ,selectorName ,',' ,fName;
		cr;nextPutAll: ','; nextPutAll: source.
	(aMethod source notNil and: [ aMethod source notEmpty ]) ifTrue: [
		| compiledSource |
		compiledSource := "self escaped: source." #nil.
		aStream cr; "warning: this cr is important for debugger"
			nextPutAll: ',' ,(self escaped: aMethod category)
				,',' ,(self escaped: aMethod source)
				,',' ,compiledSource
	].
	((aMethod respondsTo: #stamp) and: [ aMethod stamp notNil ]) ifTrue: [
		aStream nextPutAll: ',' ,(self escaped: aMethod stamp)
	].
	aStream nextPutAll: ')'; cr.
	^true! !

! PythonEncoder methodsFor: #code !
arrayLiteralStart
	" Private - Return the code string to start exporting an array literal. "

	^'smalltalk.newArray(['! !

! PythonEncoder methodsFor: #code !
arrayLiteralEnd
	" Private - Return the code string to end exporting an array literal. "

	^'])'! !

! PythonEncoder methodsFor: #code !
functionStart
	" Private - Return the code string to start a function. "

	^'smalltalk.__f__('! !

! PythonEncoder methodsFor: #code !
functionEnd
	" Private - Return the code string to end a function. "

"DEBUG"	true ifTrue: [ "add debug tag to functions"
	| tag |
	tag := self compiler currentClass name 
		,$_ ,(self asSelector: self compiler selector).
	^',"' ,(tag replace: $ with: $_) ,'")'
].

	^$)! !

! PythonEncoder methodsFor: #literals !
escaped: string
	" Private - Return the code string to export string literal. "

	^string asPythonString ! !

! PythonEncoder methodsFor: #export !
exportCommentOf: aClass on: aStream
	" Export the comment of aClass onto aStream. "

	aClass comment size > 0 ifTrue: [
	    aStream
		nextPutAll: 'setattr(' ,(self smalltalkNameFor: aClass);
		nextPutAll: ',"$comment",' ,(self escaped: aClass comment) ,')'; cr
	].! !

! PythonEncoder methodsFor: #export !
exportDefinitionOf: aClass on: aStream
	" Export definition of aClass onto aStream. "

	(self loadRequirementsOf: aClass) do: [:each|
		each isString
		ifTrue: [ aStream nextPutAll: each;cr ]
		ifFalse: [ aStream fileOutOn: aStream ].
	].
	aStream 
	    nextPutAll: 'smalltalk.addClass("', aClass name, '",';
	    nextPutAll: (self smalltalkNameFor: aClass superclass);
	    nextPutAll: ','.
	aClass instanceVariableNames isEmpty
		ifTrue: [ aStream nextPutAll: #nil ]
		ifFalse: [ self export: '[]' literals: aClass instanceVariableNames on: aStream ].
	aStream
	    nextPutAll: ',"' ,aClass category, $";
	    nextPutAll: ')'; cr.
	aClass classVariables notNil ifTrue: [
	    aStream
		nextPutAll: 'setattr(' ,(self smalltalkNameFor: aClass);
		nextPutAll: ',"$classVariableNames",' ,(self escaped: aClass classVariablesString) ,')'; cr
	].
	aClass poolDictionaries notNil ifTrue: [
	    aStream
		nextPutAll: 'setattr(' ,(self smalltalkNameFor: aClass);
		nextPutAll: ',"$poolDictionaries",'.
	   self exportArrayLiteral: aClass poolDictionaries on: aStream newLine: false.
	   aStream nextPutAll: ')'; cr
	].
	aClass propertyNames notEmpty ifTrue: [
	    aStream
		nextPutAll: 'setattr(' ,(self smalltalkNameFor: aClass);
		nextPutAll: ',"$propertyNames",'.
	   self exportArrayLiteral: aClass propertyNames on: aStream newLine: false.
	   aStream nextPutAll: ')'; cr
	].
	self exportCommentOf: aClass on: aStream.! !

! PythonEncoder methodsFor: #export !
exportMetaDefinitionOf: aClass on: aStream
	" Export class side definition of aClass onto aStream. "

	aClass class instanceVariableNames notEmpty ifTrue: [
		aStream	nextPutAll: 'smalltalk.setVarNames(';
			nextPutAll: (self smalltalkNameFor: aClass class);
			nextPutAll: $,.
		self exportArrayLiteral: aClass class instanceVariableNames on: aStream newLine: false.
		aStream nextPutAll: ')'; cr
	].

	aClass class propertyNames notEmpty ifTrue: [
		aStream	nextPutAll: 'smalltalk.setPropertyNames(';
			nextPutAll: (self smalltalkNameFor: aClass class);
			nextPutAll: $,.
		self exportArrayLiteral: aClass class propertyNames on: aStream newLine: false.
		aStream nextPutAll: ')'; cr
	].! !

! PythonEncoder methodsFor: #py !
privateSourceFor: aMethod
	" Private - Return the native code implementing aMethod. "

	| code |
	(Smalltalk isRunningOn: self nativeLanguage) ifTrue: [
		code := aMethod basicAt: '$nativeSource' ifAbsent: [].
		code notNil ifTrue: [ ^code ].
	].

	code := [ self privateCodeFor: aMethod ] on: Exception do: [:ex|
		self note: aMethod.
		self privateNIYSourceFor: aMethod with: ex description
	].
	^self functionStart ,(self escaped: code) ,self functionEnd ! !

! PythonEncoder methodsFor: #niy !
privateNIYSourceFor: aMethod with: description
	" Private - Return the native code implementing aMethod. "

	^(self privateMethodDefinition: aMethod)
	,(self privateMethodBody: aMethod with: description)! !

! PythonEncoder methodsFor: #niy !
privateMethodDefinition: aMethod
	" Private - Return the (Python) method definition of aMethod. "

	| args |
	args := aMethod compiledSource stream upTo: $(; upTo: $).
	args notEmpty ifTrue: [
		args := (args asArrayOfSubstringsSeparatedBy: $,)
			inject: '' into: [:total :each| total ,',' ,each asPythonName ].
	].
	^'(self' ,args ,')'! !

! PythonEncoder methodsFor: #niy !
privateMethodBody: aMethod with: description
	" Private - Return the (Python) body of aMethod. "

	^'
print("NIY:#privateMethodBody: ',aMethod,' - ',description,'")
return self'! !

! PythonEncoder methodsFor: #compiling !
privateCodeFor: aMethod
	" Private - Return the native code translated from sources of aMethod. "

	self compiler: aMethod classField compiler.
	^self compiler
		encoder: self;
		compile: aMethod source! !

! PythonEncoder methodsFor: #generation !
codeToAccessProperty: aName in: aNode
	" Private - Return the source code to access property aName. "

	^'getattr(self,"',aName,'",nil)'! !

! PythonEncoder methodsFor: #generation !
emmitFunctionBody: aBlock
	" Private - Emmit the code of function body evaluating aBlock. "

	self nextPutAll: $:.
	self nextBlockPut: aBlock ! !

! PythonEncoder methodsFor: #generation !
emmitEnterMethod: aMethodNode
	" Private - Generate home context definitions, entering aMethodNode. "

"DEBUG"	self nextPutAll: $# ,self currentMethodInfo; cr ! !

! PythonEncoder methodsFor: #generation !
newMethodFunctionName
	" Private - Return the function to call when installing a new method. "

	^'smalltalk.newMethod'! !

! PythonEncoder methodsFor: #generation !
emmitFunction: fName node: aNode withSelf: withSelf arguments: argumentNames doing: aBlock
	" Private - Generate function for aNode. "

	| aStream |
	aStream := '' writeStream.
	withSelf ifTrue: [
		aStream nextPutAll: #self.
		argumentNames notEmpty ifTrue: [ aStream nextPutAll: self argumentsSeparatorOp ].
	].
	argumentNames do: [:each | 
		self tempVariables add: each.
		aStream nextPutAll: (self localName: each) ]
	    separatedBy: [ aStream nextPutAll: self argumentsSeparatorOp ].
	self	emmitFunction: fName
		arguments: aStream contents
		doing: aBlock! !

! PythonEncoder methodsFor: #generation !
emmitFunction: fName node: aNode arguments: argumentNames doing: aBlock
	" Private - Generate function for aNode. "

	^self	emmitFunction: fName node: aNode
		withSelf: true arguments: argumentNames
		doing: aBlock! !

! PythonEncoder methodsFor: #generation !
emmitFunction: fName node: aNode doing: aBlock
	" Private - Generate function for aNode. "

	^self	emmitFunction: fName node: aNode
		withSelf: true arguments: aNode argumentNames
		doing: aBlock! !

! PythonEncoder methodsFor: #visiting !
visitEmptyNode: aNode
	" Private - Visit an empty node, aNode. "

	self nextPutAll: #pass ! !

! PythonEncoder methodsFor: #visiting !
visitMethodNode: aNode
	" Visit a method node. "

	| nativeSource |
	self emmitMethod: aNode.
	(Smalltalk isRunningOn: self nativeLanguage) ifFalse: [ ^self ].

	nativeSource := stream contents.
	self restoreState: (Array with: '' writeStream with: 0 with: 0).
	self nextPutAll: self newMethodFunctionName.
	self nextArgumentsPut: [
		self
			nextPutAll: (self smalltalkNameFor: self currentClass);
			nextPutAll: ',"', aNode selector, '",'; cr.
		self	nextPutAll: self functionStart;
			nextPutAll: (self escaped: nativeSource);
			nextPutAll: self functionEnd.
		self includeSourceCode ifTrue: [
			self	cr; "warning: this cr is important for debugger"
				nextPutAll: ',' ,(self escaped: aNode source)
				;nextPutAll: ',nil' ",(self escaped: nativeSource)"
		]
	] ! !

! PythonEncoder methodsFor: #visiting !
visitDirectiveNode: aNode
	" Private - Visit a directive node. "

	(aNode respondsTo: #asNative:for:on:) ifTrue: [
		self nextPutAll: (aNode
			asNative: functionName
			for: self on: stream); cr.
		^self
	].

	self visit: aNode.
	self nextPutAll: self separatorOp; cr! !

! PythonEncoder methodsFor: #results !
emmitedCode
	" Return the emmited code during compilation (or nil). "

	stream isNil ifTrue: [ ^nil ].
	(Smalltalk isRunningOn: self nativeLanguage) ifTrue: [
		"the code must be valid native code including return expression"
		^'return ' ,stream contents
	].
	^stream contents! !

! PythonEncoder methodsFor: #constants !
returnOpStart
	" Private - Return the OP for starting return expressions. "

	earlyReturn ifTrue: [
		^'smalltalk.raise_result_(_s8Ret_,'
	].
	^'return ' ! !

! PythonEncoder methodsFor: #constants !
returnOpEnd
	" Private - Return the OP for ending return expressions. "

	earlyReturn ifTrue: [
		^$)
	].
	^nil ! !

! PythonEncoder methodsFor: #literals !
asPythonLiteral: value
	" Private - Return the literal representation of value. "

	^value asPython ! !

! PythonEncoder methodsFor: #visiting !
visitValueNode: aNode
	" Private - Visit the node emmitting output. "

	self nextPutAll: (self asPythonLiteral: aNode value) ! !

! PythonEncoder methodsFor: #visiting !
visitNode: aNode index: index
	" Private - Generate code for indexed access. "

	((index isMemberOf: ValueNode) and: [ index hasString ]) ifTrue: [
		| selector |
		selector := self asSelector: index value.
		((selector includes: $ ) or: [ selector includes: $_ ]) ifFalse: [
			self nextPutAll: '.'; nextPutAll: selector.
			^self
		].
	].
	self nextIndexPut: [ self visit: index ] ! !

! PythonEncoder methodsFor: #writing !
nextIndexPut: aBlock
	" Private - Emmit a series between indexing notation evaluating aBlock onto output stream. "

	self between: #('[smalltalk.handleOf(' ')]') do: aBlock ! !

! PythonEncoder methodsFor: #variables !
visitVariableNode: aNode
	" Private - Generate code to access variable aNode. "

	| indices |
	indices := aNode indices.
	(indices isEmpty or: [
		(aNode basicAt: #onAssignment) == true
	]) ifTrue: [ ^super visitVariableNode: aNode ].
	self between: #('smalltalk.asObject_(' ')') do: [
		super visitVariableNode: aNode
	]! !

! PythonEncoder methodsFor: #encoding !
messageSendConnector
	" Private - Return the connector to use for sending a message. "

	^$. ! !

! PythonEncoder methodsFor: #encoding !
functionCallConnector
	" Private - Return the connector to use for calling a function. "

	^$. ! !

! PythonEncoder methodsFor: #accessing !
messageSendConnector: aNode
	" Private - Return the caracter to use for sending message. "

	aNode receiver messageSendConnector size = 0 ifTrue: [ ^'' ].
	(self isNativeCall: aNode) ifTrue: [ "calling an alien"
		(aNode selector startsWith: self callEscape) ifTrue: [
			^self functionCallConnector
		].
		^self messageSendConnector
	].
	^self messageSendConnector ! !

! PythonEncoder methodsFor: #accessing !
asPrivateSelector: aSelector
	" Private - Validate and returns a valid native selector "

	| result |
	result := self asSelector: aSelector.
	(aSelector startsWith: self callEscape) ifTrue: [ "call to an alien starting with _"
		^result copyFrom: 2 to: result size
	].
	^result! !

! PythonEncoder methodsFor: #encoding !
encodeMessageSend: aNode
	" Private - Generate code for message send."

	self
		nextPutAll: (self messageSendConnector: aNode);
		nextPutAll: (self asPrivateSelector: aNode selector)! !

! PythonEncoder methodsFor: #visiting !
visitSendArguments: aSendNode

	aSendNode arguments 
	    do: [:each | self visitArgument: each ]
	    separatedBy: [ self nextPutAll: self argumentsSeparatorOp ].! !

! PythonEncoder methodsFor: #visiting !
visitReceiver: aNode
	" Private - Visit the receiver of aNode emmitting output. "

	aNode isAssignment
	ifTrue: [ self visitLocalAssignment: aNode ]
	ifFalse: [ self visit: aNode ] ! !

! PythonEncoder methodsFor: #encoding !
encodeNativeCall: aNode
	" Private - Generate code for native call (alien message send).
	Note: the call is routed to smalltalk for adaption of arguments and return value.
	"

	self nextPutAll: 'smalltalk.nativeCall'; nextArgumentsPut: [
		aNode receiver messageSendConnector isNil ifTrue: [
			self nextPutAll: (self asPrivateSelector: aNode selector). "global function name"
			self nextPutAll: self argumentsSeparatorOp.
			self nextPutAll: 'nil'. "selector=nil to call global function"
		] ifFalse: [
			self visitReceiver: aNode receiver.
			self nextPutAll: self argumentsSeparatorOp.
			self nextPutAll: (self asPythonLiteral: (self asPrivateSelector: aNode selector)).
		].
		aNode arguments notEmpty ifTrue: [
			self nextPutAll: self argumentsSeparatorOp.
			aNode arguments 
				do: [:each | self visitArgument: each ]
				separatedBy: [ self nextPutAll: self argumentsSeparatorOp ]
		]
	]! !

! PythonEncoder methodsFor: #encoding !
encodeNormalSend: aNode
	" Private - Generate code for message send."

	(self isNativeCall: aNode) ifTrue: [ "encode native call (converting arguments & return value)"
		^self encodeNativeCall: aNode
	].

	self visitReceiver: aNode receiver.
	self encodeMessageSend: aNode.
	self nextArgumentsPut: [ self visitSendArguments: aNode ] ! !

! PythonEncoder methodsFor: #visiting !
visitGlobalReferenceNode: aNode

	| aName aClass lookupPools |
	aName := aNode value.
	aClass := self currentClass instanceClass.
	aClass classVariableNamesDo: [:each|
		each = aName ifTrue: [
			self
				nextPutAll: (self smalltalkNameFor: aClass);
				nextPutAll: '.classVariableAt_("';
				nextPutAll: aName;
				nextPutAll: '")'.
			^self
		]
	].
	aClass sharedPoolsAndNamesDo: [:poolName :pool|
		(pool includesKey: aName) ifTrue: [
			self
				nextPutAll: '(smalltalk.poolDictionary_("';
				nextPutAll: poolName;
				nextPutAll: '")).at_("';
				nextPutAll: aName;
				nextPutAll: '")'.
			^self
		].
	].
	self 
		nextPutAll: 'smalltalk.';
		nextPutAll: aNode value.! !

! PythonEncoder methodsFor: #variables !
knownVariables
	" Return the knownVariables of the receiver. "

	| locals |
	scopes isNil ifTrue: [
		^super knownVariables
	].
	^super knownVariables
	,(scopes inject: #() into: [:total :scope|
		total ,scope value
	])! !

! PythonEncoder methodsFor: #assignments !
emmitInternalAssignment: aNode
	" Private - Return true if the receiver emmited the code for internal (i.v. & property) assignment. "

	| cls aName tags |
	cls := self currentClass.
	aName := aNode left value.
	tags := Array with: 'smalltalk.basic_at_put(self,"@',aName,'",' with: $).
	cls allInstVarNamesDo: [:each|
		aName = each ifTrue: [
			self between:tags do: [ self visit: aNode right ].
			^true
		].
	].
	tags := Array with: 'smalltalk.basic_at_put(self,"',aName,'",' with: $).
	cls allPropertyNamesDo: [:each|
		aName = each ifTrue: [
			self between:tags do: [ self visit: aNode right ].
			^true
		].
	].
	^false ! !

! PythonEncoder methodsFor: #assignments !
emmitAssignment: aNode

	| id |
	aNode left basicAt: #onAssignment put: true.
	id := aNode basicAt: #ID ifAbsent: [].
	id notNil ifTrue: [ ^self visit: id ].
	(self emmitInternalAssignment: aNode) ifTrue: [ ^self ].
	^super emmitAssignment: aNode ! !

! PythonEncoder methodsFor: #prepare !
prepareAssignment: aNode
	" Private - Assignments are not valid expressions here,
	 so we must encode the assignment in advance
	 and mark the assignment to do not encode when pushed as value.
	"

	aNode basicAt: #ID ifAbsentPut: [
		self emmitAssignment: aNode; cr.
		self prepareValue: aNode left
	].
	^aNode ! !

! PythonEncoder methodsFor: #prepare !
prepareSendNode: aNode
	" Private - Prepare a Send node to be used as value. "

	self prepareValue: aNode receiver.
	aNode arguments do: [:each| self prepareValue: each ].
	^aNode ! !

! PythonEncoder methodsFor: #prepare !
prepareValue: aNode
	" Private - Prepare aNode to be used as a return value.
	Note that code can be generated as side effect of preparation
	e.g. for asignments in return expression or arguments resolution
	Returns the node to be emmited.
	"

	aNode isAssignment ifTrue: [ ^self prepareAssignment: aNode ].
	aNode isSend ifTrue: [
		aNode isCascade ifTrue: [ ^aNode ].
		^self prepareSendNode: aNode
	].
	^aNode ! !

! PythonEncoder methodsFor: #visiting !
visitSendNode: aNode
	" Private - Visit the node (message send node) emmitting output. "

	self prepareSendNode: aNode.
	^super visitSendNode: aNode! !

! PythonEncoder methodsFor: #contexts !
registerBlock: aNode id: aName
	" Private - Dump the block aNode with name aName. "

	self nextPutAll: 'def ' ,aName.
	self	emmitFunction: aName node: aNode
		withSelf: false arguments: aNode parameterNames
		doing: [ self visitNodesIn: aNode cr: true ]  ! !

! PythonEncoder methodsFor: #contexts !
isSendIf: aNode
	" Private - Return true if the block aNode is an If expression. "

	^(self hasBlockNoArguments: aNode) and: [
	#(ifTrue: ifFalse: ifTrue:ifFalse: ifFalse:ifTrue:) includes: aNode selector ] ! !

! PythonEncoder methodsFor: #contexts !
isSendWhile: aNode
	" Private - Return true if the block aNode is a while expression. "

	^(self allBlockNoArguments: aNode) and: [
	#(whileTrue: whileFalse:) includes: aNode selector ] ! !

! PythonEncoder methodsFor: #contexts !
willBeInlined: aNode
	" Private - Return true if the block aNode will be optimized to an inline code.
	Note that the implementation is related with optimization of well known messages.
	"

	| client |
	client := aNode client.
	^(client respondsTo: #isSend)
	and: [	client isSend
	and: [	(self noClient: client)
	and: [	(self isSendIf: client)
	or:  [	(self isSendWhile: client) and: [
		aNode ~= client receiver
	]] ]]] ! !

! PythonEncoder methodsFor: #prepare !
prepareBlock: aNode
	" Private - Register the block aNode. "

	| id |
	blocks isNil ifTrue: [ blocks := Array new ].
	(self willBeInlined: aNode) ifTrue: [ ^false ].
	blocks add: aNode.
	id := '__blk' ,blocks size ,'__'.
	aNode basicAt: #ID put: id.
	self registerBlock: aNode id: id.
	^true ! !

! PythonEncoder methodsFor: #prepare !
prepareBlocksIn: aNode
	" Private - Prepare block contexts in aNode. "

	aNode	with: [:each| each isBlock ]
		except: [:each| each isBlock ]
		do: [:each| self prepareBlock: each ] ! !

! PythonEncoder methodsFor: #visiting !
visitBlockNode: aNode

	| id count debug |
	id := aNode basicAt: #ID ifAbsent: [
		^self error: 'Broken block - must be registered'
	].
	(aNode basicAt: #noWrap ifAbsent: [false]) ifTrue: [
		^self nextPutAll: id
	].
	count := $, ,aNode parameters size.
	debug := self blockHome: aNode info: id.
	debug notEmpty ifTrue: [ debug := $, ,debug ].
	self nextPutAll: 'smalltalk.blk(' ,id ,count ,debug ,$) ! !

! PythonEncoder methodsFor: #debug !
blockHome: aNode info: id
	" Private - Returns the debug info for block aNode (or empty String).
	Note: edit this method to emmit blocks w/o $info link to method
	"

	true ifTrue: [ ^'' ].
"DEBUG"	^$" ,self currentMethodInfo ,$@ ,id ,$"! !

! PythonEncoder methodsFor: #contexts !
createContext: aName
	" Private - Emmit code for new context with local name aName. "

	self nextPutAll: aName ,'=smalltalk.newContext()'; cr ! !

! PythonEncoder methodsFor: #contexts !
startSequence: aNode
	" Private - Prepare context for sequence aNode. "

	| key vars |
	super startSequence: aNode.
	vars := aNode tempNames.
	scopes isNil ifTrue: [ scopes := #() ].
	key := '__' ,scopes size asString ,'__'.
	scopes add: key -> vars.
	vars notEmpty ifTrue: [ self createContext: key ].
	self prepareBlocksIn: aNode. ! !

! PythonEncoder methodsFor: #contexts !
endSequence: aNode
	" Private - Leave context for sequence aNode. "

	super endSequence: aNode.
	scopes removeLast ! !

! PythonEncoder methodsFor: #contexts !
sequenceSpace: aNode
	" Private - aNode has been visited in a sequence. "

	self cr ! !

! PythonEncoder methodsFor: #contexts !
localReturnOps
	" Private - Return the OPs for local return expression. "

	^Array with: 'return ' with: nil ! !

! PythonEncoder methodsFor: #contexts !
sequenceExit: aNode
	" Private - A block sequence is exiting with aNode. "

	| result |
	result := self prepareValue: aNode.
	self between: self localReturnOps do: [ self visit: result ] ! !

! PythonEncoder methodsFor: #encoding !
defineTemporary: aName
	" Private - Return the code for defining temporary variable with aName. "

	self nextPutAll: (self localName: aName) ,'=nil' ;cr ! !

! PythonEncoder methodsFor: #locals !
pseudoName: aName
	" Private - Return the literal to encode constant aName. "

	(#(#true #false) includes: aName) ifTrue: [ ^'smalltalk.' ,aName ].
	aName = #thisContext ifTrue: [
		^scopes isNil ifTrue: [ #nil ] ifFalse: [
			scopes last key
		]
	].
	^aName ! !

! PythonEncoder methodsFor: #locals !
localName: aName
	" Private - Return a valid (native, local) name for aName. "

	| result |
	result := super localName: aName.
	scopes notNil ifTrue: [
		scopes reverseDo: [:assoc|
			(assoc value includes: aName) ifTrue: [
				^assoc key isNil
					ifTrue: [ result ]
					ifFalse: [ assoc key ,'.' ,result ]
			].
		].
	].
	^self pseudoName: result ! !

! PythonEncoder methodsFor: #contexts !
startBlock: aNode
	" Private - Prepare arguments for block aNode. "

	scopes isNil ifTrue: [ scopes := #() ].
	scopes add: nil -> aNode parameterNames. ! !

! PythonEncoder methodsFor: #contexts !
endBlock: aNode
	" Private - Leave a block aNode. "

	scopes removeLast ! !

! PythonEncoder methodsFor: #visiting !
visitCascadeNode: aNode

	| evaluation blk |
	aNode nodes size = 1 ifFalse: [ ^self error: 'Broken cascade - needs transform' ].
	aNode nodes first selector = #value: ifFalse: [ ^self error: 'Broken cascade - must be #value:' ].
	evaluation := aNode nodes first.
	self compilerEvent: #aboutCascade: with: aNode.
	blk := aNode receiver.
	blk basicAt: #noWrap put: true.
	self visit: blk.
	self nextArgumentsPut: [ self visit: evaluation arguments first ].
	self compilerEvent: #cascade: with: aNode.! !

! PythonEncoder methodsFor: #cascade !
transformCascade: cascade in: aParent
	" Private - Cascades are mutated to block semantics. "

	| target block evaluation sequence |
	target := VariableNode with: #_s8_rec.
	target start: cascade receiver start end: cascade receiver end.
	sequence := BlockSequenceNode new
	    nodes: cascade nodes; temps: #();
	    start: cascade start end: cascade end;
	    yourself.
	block := BlockNode new
		parameters: (Array with: (ValueNode with: target value));
		addNode: sequence;
		start: cascade nodes first start end: cascade nodes last end;
		yourself.
	cascade nodes do: [:send| send receiver: target ].
	evaluation := SendNode new
	    selector: #value:;
	    arguments: (Array with: cascade receiver);
	    start: cascade start end: cascade end;
	    yourself.
	cascade
		receiver: block;
		nodes: (Array with: evaluation) ! !

! PythonEncoder methodsFor: #prepare !
prepare: aMethodNode
	" Private - Prepare aMethodNode for traversal. "

	(super prepare: aMethodNode) isNil ifTrue: [ ^nil ].
	aMethodNode do: [:each :parent|
		each isCascade ifTrue: [
			self transformCascade: each in: parent
		]
	].! !

! PythonEncoder methodsFor: #visiting !
visitClassVariableAssignment: aClassVariableName in: aClass to: anExpressionNode

	self
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: '.classVariableAt_put_("';
		nextPutAll: aClassVariableName; nextPutAll: '",'.
	self visit: anExpressionNode.
	self	nextPutAll: ')'! !

! PythonEncoder methodsFor: #visiting !
visitPoolVariableAssignment: aName in: poolName to: anExpressionNode

	self
		nextPutAll: 'smalltalk.poolDictionary_("';
		nextPutAll: poolName; nextPutAll: '").at_put_("';
		nextPutAll: aName; nextPutAll: '",'.
	self visit: anExpressionNode.
	self	nextPutAll: ')'! !

! PythonEncoder methodsFor: #visiting !
visitOperation: aNode
	" Private - aNode is an operation and selector is mapped to target operation. "

	self compilerEvent: #aboutOperation: with: aNode.
	self nextArgumentsPut: [ self visitReceiver: aNode receiver ].
	self nextPutAll: (self asPrivateSelector: aNode selector).
	self visitSendArguments: aNode.
	self compilerEvent: #operation: with: aNode.! !

! PythonEncoder methodsFor: #encoding !
encodeSuperSend: aNode
	" Private - Generate code for super message send."

	self nextPutAll: 'smalltalk.superReceiver('
		,(self smalltalkNameFor: self currentClass)
		,',self).',(self asPrivateSelector: aNode selector).
	self nextArgumentsPut: [
		self visitSendArguments: aNode
	] ! !

! PythonEncoder methodsFor: #writing !
nextReturn: early put: aBlock
	" Private - Emmit a return OP evaluating aBlock onto output stream. "

	early ifFalse: [ ^aBlock value ]. "the return OPs will dump while visiting the return node"
	^super nextReturn: early put: aBlock ! !

! PythonEncoder methodsFor: #generation !
defineBlock: aName arguments: argumentNames doing: aBlock
	" Private - Generate code for block function named aName.  "

	self nextPutAll: 'def ' ,aName.
	self	emmitFunction: aName node: nil
		withSelf: false arguments: argumentNames
		doing: aBlock ! !

! PythonEncoder methodsFor: #visiting !
visitReturnNode: aNode

	| result |
	nestedBlocks > 0 ifTrue: [ earlyReturn := true ].
	self withNodesIn: aNode do: [:each|
		self visitNode: each in: aNode cr: true
	] atEnd: [:last|
		result := self prepareValue: last
	].
	result isNil ifTrue: [ ^self error: 'Missing result value' ].
	^self between: self returnOps do: [
		self emmitReturn: [ self visit: result ] early: false
	] ! !

! PythonEncoder methodsFor: #writing !
nextCodePut: code
	" Private - Emmit code (preserving indent) onto output stream. "

	| lines |
	lines := code stream.
	[lines atEnd] whileFalse: [
		self nextPutAll: lines nextLine.
		lines atEnd ifFalse: [ self cr ]
	]! !

! PythonEncoder methodsFor: #generation !
generate: fName early: early return: code
	" Private - Generate code to catch early return of code expression in fName. "

	| prev target |
	early ifFalse: [
		self nextCodePut: code.
		^0
	].
	prev := stream position.
	target := #_s8Ret_.
	self createContext: target.
	self nextPutAll: #try:.
	prev := stream position - prev.
	self nextBlockPut: [ self nextCodePut: code ].
	self nextPutAll: 'except Exception as __ex:'.
	self nextBlockPut: [
		self nextPutAll: 'if smalltalk.exception_at(__ex,',target ,'):'.
		self nextBlockPut: [
			self nextPutAll: 'return smalltalk.exceptionResult_(__ex)'
		].
		self nextPutAll: 'raise __ex'
	].
	^prev ! !

! PythonEncoder methodsFor: #annotations !
annotate: aNode variable: literal
	" Private - Returns code to annotate variable in list. "

	| referenceTo |
	referenceTo := [:ref|
		(VariableNode with: ref) localNameIn: self
	].
	(literal includes: $@) ifTrue: [ | globl locl |
		globl := literal upTo: $@.
		locl := referenceTo value: (literal fromLast: $@).
		^'smalltalk.' ,globl ,'._at(' ,locl ,')'
	].
	^referenceTo value: literal ! !

! PythonEncoder methodsFor: #annotations !
annotate: code for: aNode variables: array
	" Private - Returns code annotated with variables array. "

	^array inject: code into: [:total :each|
		code ,$, ,(self annotate: aNode variable: each)
	]! !

! PythonEncoder methodsFor: #annotations !
annotate: code for: aNode on: stream
	" Private - Returns code annotated from stream encoding (or nil). "

	stream next = $, ifFalse: [ ^nil "unknown annotation format" ].
	^self annotate: code for: aNode variables: (stream upToEnd
		asArrayOfSubstringsSeparatedBy: $,)! !

! PythonEncoder methodsFor: #annotations !
annotate: code for: aNode with: annotation
	" Private - Returns code to apply annotation to aNode. "

	| literal |
	literal := annotation.
	(annotation isKindOf: ValueNode) ifTrue: [
		literal := annotation value
	].
	literal isEmpty ifTrue: [ ^code ].
	aNode isValue ifTrue: [ | result |
		result := self annotate: code for: aNode on: literal stream.
		result notNil ifTrue: [ ^result ].
	].
	^self invalidAnnotation: annotation for: aNode! !

! PythonEncoder methodsFor: #annotations !
invalidAnnotation: literal for: aNode
	" Private - The annotation literal is invalid/unknown for aNode. "

	^self error: 'Invalid/wrong annotation ' ,literal! !

! PythonEncoder methodsFor: #variables !
codeVisitingLocal: aNode
	" Private - Returns code to access local value. "

	| code |
	code := super codeVisitingLocal: aNode.
	aNode annotations do: [:each|
		code := self annotate: code for: aNode with: each
	].
	^code! !

! PythonEncoder methodsFor: #optimization !
coreOptimizations
	" Private - Return the optimization set. "

	^#(core  #(
		#(class	'(%receiver%).__class__' )
		#(isNil	'%.isNil_%(%receiver%)' )
		#(notNil	'%.notNil_%(%receiver%)' )
		#(=	'%.isObject_equalTo_%(%receiver%,%arg1%)' )
		#(~=	'%.isObject_equalTo_%(%receiver%,%arg1%)._8not()' )
"BROKEN		#(==	'%.b_%(%receiver% is %arg1% or (%.handleOf%(%receiver%) is %.handleOf%(%arg1%)))' )		"
"BROKEN		#(~~	'%.b_%(not (%receiver% is %arg1% or (%.handleOf%(%receiver%) is %.handleOf%(%arg1%))))' )	"
"BROKEN		#(==	'%.isObject_eqeq_%(%receiver%,%arg1%)' )		"
"BROKEN		#(~~	'%.isObject_eqeq_%(%receiver%,%arg1%)._8not()' )	"
"	Broken on Windows produces a VM fast exit while compiling nested blocks.
	See https://bugs.python.org/issue44906					"
		#(==	'%.b_%(%receiver% is %arg1%)' )
		#(~~	'%.b_%(not(%receiver% is %arg1%))' )
		#(||	'%.b_%(%receiver% or %arg1%)' )
		#(,	'%,%(%receiver%,%arg1%)' )
		#(basicAt:	'%.basic_at_%(%receiver%,%arg1%)' )
		#(basicAt:ifAbsent: '%.basic_at_ifAbsent_%(%receiver%,%arg1%,%arg2%)' )
		#(basicAt:put: '%.basic_at_put_%(%receiver%,%arg1%,%arg2%)' )
		#(ifUndefined: '%.ifUndefined_do_%(%receiver%,%arg1%)' )
	))! !

! PythonEncoder methodsFor: #optimization !
s8Optimizations
	" Private - Return the optimization set. "

	^#(s8	#(
		core
		#(ifTrue: hasBlockNoArguments: #(
			#(ifTrue: noClient: 'if %.b%(%receiver%):%>%%seq1%%<%' )
			#(ifTrue: '%.if_true_false_%(%receiver%,%arg1%,nil)' )
		))
		#(ifFalse: hasBlockNoArguments: #(
			#(ifFalse: noClient: 'if not %.b%(%receiver%):%>%%seq1%%<%' )
			#(ifFalse: '%.if_true_false_%(%receiver%,nil,%arg1%)' )
		))
		#(ifTrue:ifFalse: hasBlockNoArguments: #(
			#(ifTrue:ifFalse: noClient: 'if %.b%(%receiver%):%>%%seq1%%<%else:%>%%seq2%%<%' )
			#(ifTrue:ifFalse: '%.if_true_false_%(%receiver%,%arg1%,%arg2%)' )
		))
		#(ifFalse:ifTrue: hasBlockNoArguments: #(
			#(ifFalse:ifTrue: noClient: 'if %.b%(%receiver%):%>%%seq2%%<%else:%>%%seq1%%<%' )
			#(ifFalse:ifTrue: '%.if_true_false_%(%receiver%,%arg2%,%arg1%)' )
		))
		#(whileTrue receiverIsBlockNoArguments: #(
			#(whileTrue noClient: 'while %.b%(%receiver%()):%>%continue%<%' )
			#(whileTrue '%.while_do_%(%receiver%,nil)' )
		))
		#(whileFalse receiverIsBlockNoArguments: #(
			#(whileFalse noClient: 'while not %.b%(%receiver%()):%>%continue%<%' )
			#(whileFalse '%.whileNot_do_%(%receiver%,nil)' )
		))
		#(whileTrue: receiverIsBlockNoArguments: #(
			#(whileTrue: noClient: 'while %.b%(%receiver%()):%>%%seq1%%<%' )
			#(whileTrue: '%.while_do_%(%receiver%,%arg1%)' )
		))
		#(whileFalse: receiverIsBlockNoArguments:  #(
			#(whileFalse: noClient: 'while not %.b%(%receiver%()):%>%%seq1%%<%' )
			#(whileFalse: '%.whileNot_do_%(%receiver%,%arg1%)' )
		))
		#(orNil '(%receiver% or nil)' )
		#(timesRepeat: '%.times_repeat_%(%receiver%,%arg1%)' )
		#(isKindOf: '%.is_kindOf_%(%receiver%,(%arg1%))' )
		#(respondsTo: '%.responds_to_%(%receiver%,(%arg1%))' )
		#(ifNil: '%.if_nil_notNil_%(%receiver%,%arg1%,nil)' )
		#(ifNotNil: '%.if_nil_notNil_%(%receiver%,nil,%arg1%)' )
		#(ifNil:ifNotNil: '%.if_nil_notNil_%(%receiver%,%arg1%,%arg2%)' )
		#(ifNotNil:ifNil: '%.if_nil_notNil_%(%receiver%,%arg2%,%arg1%)' )
	))! !

! PythonEncoder methodsFor: #optimization !
magnitudeOptimizations
	" Private - Return the optimization set. "

	^#(magnitude #(
		#(> 	'%.b_%(%receiver% > %arg1%)' )
		#(< 	'%.b_%(%receiver% < %arg1%)' )
		#(>=	'%.b_%(%receiver% >= %arg1%)' )
		#(<=	'%.b_%(%receiver% <= %arg1%)' )
	))! !

! PythonEncoder methodsFor: #optimization !
mathOptimizations
	" Private - Return the optimization set. "

	^#(math #(
		#(max:	'max(%receiver%,%arg1%)' )
		#(min:	'min(%receiver%,%arg1%)' )
		#(mustBeNumber	'%.mustBeNumber_%(%receiver%)' )
		#(+	'(%receiver% + %arg1%)' )
		#(#-	'(%receiver% - %arg1%)' )
		#(*	'(%receiver% * %arg1%)' )
		#(/	'(%receiver% / %arg1%)' )
	))! !

! PythonEncoder methodsFor: #optimization !
printingOptimizations
	" Private - Return the optimization set. "

	^#(printing #(
		#(toString	'%.to_String%(%receiver%)' )
		#(printString	'%.printString_%(%receiver%)' )
		#(asString	'%.asString_%(%receiver%)' )
		#(asCharacter	'%.asCharacter_%(%receiver%)' )
	))! !

! PythonEncoder methodsFor: #optimization !
booleanOptimizations
	" Private - Return the optimization set. "

	^#(boolean #(
		#(isBoolean '%.isBoolean_%(%receiver%)' )
		#(and: hasBlockNoArguments: '%.boolean_and_%(%receiver%,%arg1%)' )
		#(or: hasBlockNoArguments:  '%.boolean_or_%(%receiver%,%arg1%)' )
		#(not '%.booleanNot_%(%receiver%)' )
	))! !

! PythonEncoder methodsFor: #defaults !
defaultKnownOptimizations
	" Private - Return the default knownOptimizations of the receiver. "

	^(Array	with: self coreOptimizations)
		add: self s8Optimizations;
"		add: self magnitudeOptimizations;
		add: self mathOptimizations;"
		add: self booleanOptimizations;
		add: self printingOptimizations;
		yourself! !

! PythonEncoder methodsFor: #defaults !
defaultOptimizations
	" Private - Return the default optimizations of the receiver. "

	^#( s8 magnitude "math" boolean "evaluation" printing )! !

! PythonEncoder class methodsFor: #inlines !
initializeKernelInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self kernelInlines into: pool ! !

! PythonEncoder class methodsFor: #inlines !
kernelInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(Object #(
#(halt		'import pdb; pdb.set_trace()')
#(toString	'if self.__class__ == smalltalk.String or self.__class__ == smalltalk.Number: return smalltalk.stObject().toString(self)')
#(toString:	'%String%(str(anObject))')
#(isObject:	'return %.b_%(%.isObject%(anObject))')
#(typeOf:	'%String%(type(anObject).__name__)')
#(callableOf:	'if True:%>%return %.handleOf%(aBlock)%<%')
#(basicDelete:	'key = smalltalk.handleOf(aString)%cr%try:%>%del self[key]%cr%return self%<%except Exception as ex:%>%pass%<%try:%>%delattr(self,key)%cr%return self%<%except Exception as ex:%>%setattr(self,key,nil)%<%')
#(basicPerform:withArguments: #('self' '__0__.fn(*%.objects_toCall_%(anArray,__0__.fn))'))
#(asJSONObject	'%.newHash%()')
))
#(Smalltalk #(
#(includesKey:	'StObject(nil)')
#(eval:	'return %.__f__%("():"+%.handleOf%(anExpression))()')
))
#('Smalltalk class' #(
#(global	'globals()')
#(gc	'if True:%>%import gc%cr%return gc.collect()%<%')
))
#(Behavior #(
#(basicNew	'self(nil)')
#(fn		'if True:%>%return %.fn_Of%(self)%<%')
#(prototype	'if True:%>%return %.prototype_Of%(self)%<%')
#(privateMethods	'if True:%>%return %.methods_Of%(self)%<%')
))
#(CompiledMethod #(
#(asClosure	'if True:%>%return %BlockClosure%(self.fn())%<%')
))
#(Boolean #(
#(asJSONObject 'if True: return %.handleOf%(self)')
#(not 	'%false% if %.handleOf%(self) else %true%')
#(=	'if %.b%(%.is_kindOf_%(anObject, %Boolean%)):%>%return %.b_%(%.handleOf%(self) == %.handleOf%(anObject))%<%if True:%>%return %.b_%(%.handleOf%(self) == anObject)%<%')
#(ifTrue:ifFalse: 'if %.b%(self):%>%return %.handleOf%(aBlock)()%<%return %.handleOf%(anotherBlock)()')
))
#(Number #(
#(asJSONObject 'if True: return %.handleOf%(self)')
#(= 'if True: return %.b_%(%.handleOf%(anObject) is %.handleOf%(self))')
#(> '%.b_%(%.handleOf%(self) > %.handleOf%(aNumber))')
#(< '%.b_%(%.handleOf%(self) < %.handleOf%(aNumber))')
#(>= '%.b_%(%.handleOf%(self) >= %.handleOf%(aNumber))')
#(<= '%.b_%(%.handleOf%(self) <= %.handleOf%(aNumber))')
#(+ '%Number%(%.handleOf%(self) + %.handleOf%(aNumber))')
#(- '%Number%(%.handleOf%(self) - %.handleOf%(aNumber))')
#(* '%Number%(%.handleOf%(self) * %.handleOf%(aNumber.mustBeNumber()))')
#(/ '%Number%(%.handleOf%(self) / %.handleOf%(aNumber))')
#(sin '%Number%(%.math%.sin(%.handleOf%(self)))')
#(cos '%Number%(%.math%.cos(%.handleOf%(self)))')
#(tan '%Number%(%.math%.tan(%.handleOf%(self)))')
#(arcCos '%Number%(%.math%.acos(%.handleOf%(self)))')
#(arcSin '%Number%(%.math%.asin(%.handleOf%(self)))(self)')
#(arcTan '%Number%(%.math%.atan(%.handleOf%(self)))')
#(atan2: '%Number%(%.math%.atan2(%.handleOf%(self)))')
#(sqrt '%Number%(%.math%.sqrt(%.handleOf%(self)))')
#(<< '%Number%(%.handleOf%(self) << %.handleOf%(aNumber))')
#(>> '%Number%(%.handleOf%(self) >> %.handleOf%(aNumber))')
#(raisedTo: '%Number%(%.math%.pow(%.handleOf%(self),%.handleOf%(aNumber)))')
#(max: '%Number%(max(%.handleOf%(self),%.handleOf%(aNumber)))')
#(min: '%Number%(min(%.handleOf%(self),%.handleOf%(aNumber)))')
#(rounded '%Number%(round(%.handleOf%(self)))')
#(truncated '%Number%(%.math%.trunc(%.handleOf%(self)))')
#(timesRepeat: 'for i in range(%.handleOf%(self)): aBlock(%Number%(i+1))')
#(to:do: 'for i in range(%.handleOf%(self),%.handleOf%(aNumber)+1): aBlock(%Number%(i))')
#(to:by:do: 'incr = %.handleOf%(increment)%cr%gap = 1 if incr>0 else -1%cr%for i in range(%.handleOf%(self),%.handleOf%(aNumber)+gap,incr): aBlock(%Number%(i))')
#(printString '%String%(str(%.handleOf%(self)))')
#(toString: '%.basen%(self,radix)')
#(isNumber 'if True:%>%return %.b_%(%.is_Number%(%.handleOf%(self)))%<%')
#(clearInterval 'self.notImplementedYet_(%String%("clearInterval"))')
#(clearTimeout 'self.notImplementedYet_(%String%("clearTimeout"))')
))
#('Number class' #(
#(pi 'self(%.math%.pi)')
))
#(BlockClosure #(
#(whileTrue: 'while %.handleOf%(self)(): %.handleOf%(aBlock)()')
#(whileFalse: 'while not %.handleOf%(self)(): %.handleOf%(aBlock)()')
#(valueWithPossibleArguments: 'self(*%.arrayContents%(array))')
#(try:catch: 'try:%>%return aBlock()%<%except Exception as e:%>%stEx = smalltalk.Exception(e)%cr%stEx.nativeException_(e)%cr%stEx.messageText_(str(e))%cr%return anotherBlock(self.smalltalkErrorOf_(stEx))%<%')
#(valueWithTimeout: '%.value_withTimeout_%(self,aNumber)')
#(valueWithInterval: '%.value_withInterval_%(self,aNumber)')
#(ensure: 'try:%>%return self()%<%finally:%>%aBlock()%<%')
#(ifCurtailed: 'try:%>%result = self()%cr%curtailed = smalltalk.false%<%finally:%>%if %.b%(curtailed):%>%%.handleOf%(exceptionBlock)()%<%%<%return result')
#(argumentCount '%.argumentCount_%(self)')
))
#(UndefinedObject  #(
#(asJSONObject 'None')
))
#(Collection  #(
#(wrap: 'return %.asNativeValue_%(anObject)')
#(unwrap: 'return %.asObject_%(anObject)')
))
#(HashedSet #(
#(initializeHash 'dict()')
#(includes: '%.b_%(str(smalltalk.handleOf(__0__.cookie)) in getattr(self,"@hash"))')
#(privateAdd: '%.hashedSet_privateAdd_%(self,cookie)%cr%if True:%>%return%<%')
#(privateRemove: '%.hashedSet_privateRemove_%(self,cookie)')
))
#(SequenceableCollection #(
#(indexOf:ifAbsent: 'cookie=self.wrap_(anObject)%cr%for i, x in enumerate(%.handleOf%(self)):%>%if %.isObject_equalTo%(x,cookie):%>%return %Number%(i+1)%<%%<%if True:%>%return %.handleOf%(aBlock)()%<%')
#(do: 'for x in %.handleOf%(self):%>%aBlock(self.unwrap_(x))%<%')
#(detect:ifNone: 'for x in %.handleOf%(self):%>%each=self.unwrap_(x)%cr%if %.b%(aBlock(each)):%>%return each%<%%<%if True:%>%return %.handleOf%(anotherBlock)()%<%')
#(reversed 'return self.__class__(list(%.handleOf%(self))[::-1])')
#(isEmpty '%.b_%(len(%.handleOf%(self)) == 0)')
#(reverseDo:  'for x in reversed(%.handleOf%(self)):%>%aBlock(self.unwrap_(x))%<%')
#(includes: #('cookie=self.wrap_(anObject)%cr%for x in %.handleOf%(self):%>%if %.isObject_equalTo%(x,cookie):%>%return %true%%<%%<%if True:%>%return %false%%<%' 'self'))
#(occurrencesOf: 'count = 0%cr%cookie=self.wrap_(anObject)%cr%for x in %.handleOf%(self):%>%if %.isObject_equalTo%(x,cookie):%>%count = count + 1%<%%<%if True:%>%return %Number%(count)%<%')
))
#('String class' #(
#(fromString: 'if isinstance(aString,self):%>%return aString%<%else:%>%return self(aString)%<%')
#(new	'self("")')
#(cr 	'self(chr(13))')
#(lf 	'self(chr(10))')
#(crlf 	'self(chr(13)+chr(10))')
#(space	'self(" ")')
#(tab	'self("\t")')
#(zero	'self(chr(0))')
#(fromCharCode: 'return self(chr(%.handleOf%(anArrayOrNumber)))')
))
#(String #(
#(asJSONObject 'if True: return %.handleOf%(self)')
#(=	'if True:%>%return %.b_%(%.handleOf%(anObject) == %.handleOf%(self))%<%')
#(size	'%Number%(len(%.handleOf%(self)))')
#(at:	'self.__class__(%.handleOf%(self)[%.handleOf%(anIndex) - 1])')
#(do: 	'for c in %.handleOf%(self):%>%aBlock(self.__class__(c))%<%')
#(withIndexDo: 'for i in range(%.handleOf%(self.size())):%>%aBlock(%.handleOf%(self)[i],%Number%(i+1))%<%')
#(detect:ifNone: 'for c in %.handleOf%(self):%>%each = self.__class__(c)%cr%if %.b%(%.handleOf%(aBlock(each))):%>%return each%<%%<%return %.handleOf%(anotherBlock)()')
#(isLiteral 'if True: return %.string_isLiteral_%(self)')
#(occurrencesOf: 'if True: return %.string_occurrencesOf_%(self,aString)')
#(startsWith: 'if True: return %.string_startsWith_%(self,aString)')
#(escaped 'if True:%>%try:%>%from urllib import quote%<%except ImportError:%>%from urllib.parse import quote%<%return %String%(quote(%.handleOf%(self)))%<%')
#(unescaped 'if True:%>%try:%>%from urllib import unquote%<%except ImportError:%>%from urllib.parse import unquote%<%return %String%(unquote(%.handleOf%(self)))%<%')
#(asEncodedURI 'if True: return self.notImplementedYet_(%String%("asEncodedURI"))')
#(asDecodedURI 'if True: return self.notImplementedYet_(%String%("asDecodedURI"))')
#(, 'if True:%>%return self.__class__(%.handleOf%(self)+str(%.handleOf%(aCollection)))%<%')
#(copyWith: 'self.__class__(%.handleOf%(self)+%.handleOf%(anObject))')
#(copyFrom:to: 'self.__class__(%.handleOf%(self)[%.handleOf%(anIndex) - 1:%.handleOf%(anotherIndex)])')
#(reversed 'if True:%>%return %String%(%.handleOf%(self)[::-1])%<%')
#(includes: 'try:%>%return %.b_%(%.handleOf%(anObject) in %.handleOf%(self))%<%except TypeError:%>%pass%<%')
#(indexOf: 'try:%>%return %Number%(%.handleOf%(self).find(%.handleOf%(aString))+ 1)%<%except TypeError:%>%return %Number%(0)%<%return %Number%(0)')
#(indexOf:startingAt: 'try:%>%return %Number%(%.handleOf%(self).find(%.handleOf%(aString),%.handleOf%(position) - 1)+ 1)%<%except TypeError:%>%return %Number%(0)%<%return %Number%(0)')
#(lastIndexOf: 'offset = %.handleOf%(self).rfind(%.handleOf%(aString))%cr%if True:%>%return %Number%(offset + 1)%<%')
#(lastIndexOf:startingAt: 'offset = %.handleOf%(self).rfind(%.handleOf%(aString),%.handleOf%(position) - 1)%cr%if True:%>%return %Number%(offset + 1)%<%')
#(asNativeName 'self.asPythonName()')
#(asPrivateSelector '__0__.result = %.string_asPrivateSelector_%(self)')
#(asNative: 'if True: return self.asPython()')
#(replace:with: 'if True: return %String%(%.handleOf%(self).replace(%.handleOf%(aString),%.handleOf%(anotherString)))')
#(replaceRegexp:with: 'self.notImplementedYet_(%String%("String>>#replaceRegexp:with:"))')
#(tokenize: 'result = %.handleOf%(self).split(%.handleOf%(aString))%cr%return %.newArray%(result)')
#(match: '%.notNil_%(%.string_match%(self,aRegexp))')
#(isLetter '%true% if %.handleOf%(self).isalpha() else %false%')
#(isDigit '%true% if %.handleOf%(self).isdigit() else %false%')
#(isAlphaNumeric 'if %.handleOf%(self).isalnum():%>%return %true%%<%else:%>%return %false%%<%')
#(isUpperCase 'if True:%>%return %.b_%(%.handleOf%(self).isupper())%<%')
#(isLowerCase 'if True:%>%return %.b_%(%.handleOf%(self).islower())%<%')
#(asUppercase 'self.__class__(%.handleOf%(self).upper())')
#(asLowercase 'self.__class__(%.handleOf%(self).lower())')
#(asNumber '%.string_asNumber_%(self)')
#(> '%true% if %.handleOf%(self) > %.handleOf%(aString) else %false%')
#(< '%true% if %.handleOf%(self) < %.handleOf%(aString) else %false%')
#(>= '%true% if %.handleOf%(self) >= %.handleOf%(aString) else %false%')
#(<= '%true% if %.handleOf%(self) <= %.handleOf%(aString) else %false%')
#(isCharacter '%true% if len(%.handleOf%(self)) == 1 else %false%')
#(asciiValue '%Number%(ord(%.handleOf%(self)))')
#(isSeparator '%true% if %.handleOf%(self).isspace() else %false%')
))
#('Array class' #(
#(basicNew: 'self([nil] * %.handleOf%(size))')
#(with: 'if True:%>%return %.newArray%([%.asNativeValue_%(anObject)])%<%')
#(with:with: 'if True:%>%return %.newArray%([%.asNativeValue_%(firstObject), %.asNativeValue_%(secondObject)])%<%')
#(with:with:with: 'if True:%>%return %.newArray%([%.asNativeValue_%(firstObject), %.asNativeValue_%(secondObject), %.asNativeValue_%(thirdObject)])%<%')
#(with:with:with:with: 'if True:%>%return %.newArray%([%.asNativeValue_%(firstObject), %.asNativeValue_%(secondObject), %.asNativeValue_%(thirdObject), %.asNativeValue_%(fourthObject)])%<%')
#(withAll: 'if True:%>%return self(%.handleOf%(aCollection.asArray()))%<%')
#(slice: 'self.notImplementedYet_(%String%("Array class>>#slice:"))')
))
#(Array #(
#(asJSONObject #('if True: return %.handleOf%(self)' 'self'))
#(json #('if True: return %.handleOf%(self)' 'self'))
#(at: 'if True:%>%return %.array_at_%(self,anIndex)%<%')
#(at:ifAbsent: 'if True:%>%return %.array_at_ifAbsent_%(self,anIndex,aBlock)%<%')
#(at:put: 'if True:%>%return %.array_at_put_%(self,anIndex,anObject)%<%')
"#(do: 'for x in %.handleOf%(self):%>%aBlock(%.asObject_%(x))%<%')"
#(do: 'for x in %.handleOf%(self):%>%aBlock(self.unwrap_(x))%<%')
#(addAll: 'result = %.handleOf%(self) + %.handleOf%(aCollection.asArray())%cr%setattr(self,"@handle",result)')
#(shallowCopy 'aType = type(%.handleOf%(self))%cr%objects = aType(%.handleOf%(self))%cr%if True:%>%return self.__class__(objects)%<%')
#(copyWithAll: 'result = self.copy()%cr%result.addAll_(aCollection.asArray())%cr%if True:%>%return result%<%')
#(, 'result = self.copy()%cr%result.addAll_(aCollection.asArray())%cr%if True:%>%return result%<%')
#(join: '%.array_join_%(self,aString)')
#(size '%Number%(len(%.handleOf%(self)))')
#(sort: '%.array_sort_%(self,aBlock)')
#(remove:ifAbsent: 'try:%>%%.handleOf%(self).remove(%.asNativeValue_%(anObject))%<%except ValueError:%>%return %.handleOf%(aBlock)()%<%else:%>%return anObject%<%')
#(add: '%.handleOf%(self).append(%.asNativeValue_%(anObject))')
#(addFirst: '%.handleOf%(self).insert(0,%.asNativeValue_%(anObject))')
#(addLast: '%.handleOf%(self).append(%.asNativeValue_%(anObject))')
#(removeAll '%.handleOf%(self).clear()')
#(removeFirst '%.asObject_%(%.handleOf%(self).pop(0))')
#(removeIndex: '%.asObject_%(%.handleOf%(self).pop(%.handleOf%(index)-1))')
#(removeLast '%.asObject_%(%.handleOf%(self).pop())')
#(atAllPut: 'array = %.handleOf%(self)%cr%o = %.asNativeValue_%(anObject)%cr%for i in range(0,%.handleOf%(self.size())):%>%array[i] = o%<%')
#(first 'array = %.handleOf%(self)%cr%if array:%>%return %.asObject_%(array[0])%<%')
#(second 'array = %.handleOf%(self)%cr%if len(array) > 1:%>%return %.asObject_%(array[1])%<%')
#(third 'array = %.handleOf%(self)%cr%if len(array) > 2:%>%return %.asObject_%(array[2])%<%')
#(fourth '	array = %.handleOf%(self)%cr%if len(array) > 3:%>%return %.asObject_%(array[3])%<%')
#(fifth 'array = %.handleOf%(self)%cr%if len(array) > 4:%>%return %.asObject_%(array[4])%<%')
#(sixth 'array = %.handleOf%(self)%cr%if len(array) > 5:%>%return %.asObject_%(array[5])%<%')
))
#('RegularExpression class' #(
#(fromString:flag: 'self.notImplementedYet_(%String%("RegularExpression>>#fromString:flag:"))')
))
#(Exception #(
#(signal 'raise Exception(self)')
))
#(ComposedAssociations #(
#(hashObject '%.newHash%()')
))
#('KeyedCollection class' #(
#(@ 'if True:%>%return %PoolDictionary%._at_at(jsObject)%<%')
))
#(KeyedCollection #(
#(newHash '%.newHash%()')
))
#('Dictionary class' #(
#(basicNew 'if True:%>%return %PoolDictionary%.new()%<%')
))
#(PoolDictionary #(
#(at:put: 'if True:%>%return %.pool_at_put_%(self, aKey, aValue)%<%')
#(removeKey:ifAbsent: '%.pool_removeKey_%(self, aKey)')
#(at:ifAbsent: 'if True:%>%return %.pool_at_ifAbsent_%(self, aKey, aBlock)%<%')
#(includesKey: '%.pool_includesKey_%(self, aKey)')
#(keys 'if True:%>%return %.pool_keys%(self)%<%')
))
#(ClassBuilder #(
#(addSubclassOf:named:instanceVariableNames: 'aName = str(%.handleOf%(aString))%cr%%.addClass%(aName, aClass, aCollection)%cr%if True:%>%return getattr(smalltalk,aName,nil)%<%')
))
#(StringStream #(
#(peek 'try:%>%s = %.handleOf%(self.collection())%cr%offs = %.handleOf%(self.position())%cr%if offs>=len(s):%>%return nil%<%return %String%(s[offs])%<%except:%>%return nil%<%')
#(peekForAll: 'start = %.handleOf%(self.position())%cr%s = %.handleOf%(self.collection())%cr%if not s.startswith(%.handleOf%(aCollection),start):%>%return %false%%<%')
))
#(Random #(
#(next 'if True:%>%import random%cr%return %Number%(random.random())%<%')
))
#('DateTime class' #(
#(new: 'iso = %.handleOf%(anObject).replace("/","-")%cr%if True:%>%import datetime%cr%return self(datetime.fromisoformat(iso))%<%')
#(today 'if True:%>%from datetime import date%cr%return self(date.today())%<%')
#(now 'if True:%>%from datetime import datetime%cr%return self(datetime.now())%<%')
#(msNow 'if True:%>%import time%cr%return %Number%(int(time.time_ns() / 1000000))%<%')
))
#(DateTime #(
#(asJSONObject 'if True: return %.handleOf%(self)')
#(asString 'if True: return self.asLocaleString()')
#(< '%.b_%(%.handleOf%(self) < %.handleOf%(aDateTime))')
#(> '%.b_%(%.handleOf%(self) > %.handleOf%(aDateTime))')
#(<= '%.b_%(%.handleOf%(self) <= %.handleOf%(aDateTime))')
#(>= '%.b_%(%.handleOf%(self) >= %.handleOf%(aDateTime))')
#(- 'self.__class__(%.handleOf%(self) - %.handleOf%(aDateTime))')
#(+ 'self.__class__(%.handleOf%(self) + %.handleOf%(aDateTime))')
#(isValid '%true%')
#(year '%.handleOf%(self).year')
#(year: '%.handleOf%(self).year = aNumber')
#(month '%.handleOf%(self).month')
#(month: '%.handleOf%(self).month = aNumber')
#(day '%.handleOf%(self).day')
#(day: 'self.day = aNumber')
#(dayOfMonth '%.handleOf%(self).day')
#(dayOfMonth: '%.handleOf%(self).day = aNumber')
#(asLocaleString '%.handleOf%(self).strftime("%%c")')
#(asDateString '%.handleOf%(self).strftime("%%x")')
#(asTimeString '%.handleOf%(self).strftime("%%X")')
#(time 'self.invalidMessage_(%String%("time"))')
#(time: 'self.invalidMessage_(%String%("time:"))')
#(hours: 'self.invalidMessage_(%String%("hours:"))')
#(minutes: 'self.invalidMessage_(%String%("minutes:"))')
#(seconds: 'self.invalidMessage_(%String%("seconds:"))')
#(milliseconds: 'self.invalidMessage_(%String%("milliseconds:"))')
#(hours 'self.notImplementedYet_(%String%("hours"))')
#(minutes 'self.notImplementedYet_(%String%("minutes"))')
#(seconds 'self.notImplementedYet_(%String%("seconds"))')
#(milliseconds 'self.notImplementedYet_(%String%("milliseconds"))')
))
	) ! !

! PythonEncoder class methodsFor: #inlines !
initializeNativeInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self nativeInlines into: pool ! !

! PythonEncoder class methodsFor: #inlines !
nativeInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#('NativeObject class' #(
#(undefined 'None')
#(codeToNumber '" "')
#(newInstanceOf:withAll: 'if True:%>%return __0__.fn(*%.arrayHandles%(handlesArray))%<%')
))
#(NativeObject #(
#(jsObject 'self.invalidMessage_(%String%("jsObject"))')
#(ownPropertyNames 'if True:%>%return %PoolDictionary%._at_at(self.handle()).keys()%<%')
))
#(NativeHandle #(
#(hasOwnProperty: '%true% if hasattr(getattr(self,"@handle"),aName) else %false%')
#(delete: 'delattr(getattr(self,"@handle"),aName)')
))
#('NativeLibraryObject class' #(
#(basicLoad: 'require(url)')
))
#(NativeLibraryObject  #(
#(newHandleOf: 'result = aConstructor()')
#(newHandleOf:with: 'result = aConstructor(arg1)')
#(newHandleOf:with:with: 'result = aConstructor(arg1,arg2)')
#(newHandleOf:with:with:with: 'result = aConstructor(arg1,arg2,arg3)')
#(newHandleOf:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4)')
#(newHandleOf:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5)')
#(newHandleOf:with:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5,arg6)')
#(newHandleOf:with:with:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5,arg6,arg7)')
#(newHandleOf:with:with:with:with:with:with:with:with: 'result = aConstructor(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8)')
))
#(Array #(
#(addAllLua: 'self.addAll_(natives)')
))
#('Array class' #(
#(@@ 'if True: return self(hArray)')
#(@ '__0__.result=self(natives)')
))
#('Boolean class' #(
#(@ 'h=%.handleOf%(aBOOL)%cr%if h==1 or h==True:%>%return %true%%<%')
))
	) ! !

! PythonEncoder class methodsFor: #inlines !
initializeToolsInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self toolsInlines into: pool ! !

! PythonEncoder class methodsFor: #inlines !
toolsInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(NativeReader #(
#(initializeObject '%.newHash%()')
))
#(String  #(
#(asComment #('%String%.pythonCommentStart()' 	'%String%.pythonCommentEnd()' 	'"# "'))
))
#(MethodReferences #(
#(dot		'%String%(".")')
#(localDef	'%String%(" ")')
#(superSendPatternOf:	'self.superPythonSendPatternOf_(aSelector)')
#(compiledSourceStart	'%String%("(")')
))
#(Class #(
#(hackDNU 'self.notImplementedYet_(%String%("Class>>#hackDNU"))')
))
	) ! !

! PythonEncoder class methodsFor: #inlines !
initializeSnapshotInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self snapshotInlines into: pool ! !

! PythonEncoder class methodsFor: #inlines !
snapshotInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(Smalltalk #(
#(exit	'self.quit()')
))
#(Snapshot #(
#(bootHeader	'%String%("")')
#(bootFooter	'%String%("")')
#(launchExpression '%String%("%.launchImage%()")')
))
	) ! !

! PythonEncoder class methodsFor: #inlines !
initializeOtherInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self otherInlines into: pool ! !

! PythonEncoder class methodsFor: #inlines !
otherInlines
	" Private - Return the inlines for methods in the category. "

	^#(
"#(ClassName #(
#(selector '...')
))"
	) ! !

"------------------------------------------------------------------------------
	Extension to Kernel to support Python encoding
 ------------------------------------------------------------------------------"!

! String class methodsFor: #Python-py !
pythonCommentStart
	" Private - Return the (multiline) comment start sequence. "

	^$" ,$" ,$"! !

! String class methodsFor: #Python-py !
pythonCommentEnd
	" Private - Return the (multiline) comment end sequence. "

	^$" ,$" ,$"! !

! String class methodsFor: #Python-py !
reservedPythonWords
	" Private - Return the reserved names of Python language. "

	^#(
	False	None	True
	and	as	assert	async
	await	break	class	continue
	def	del	elif	else
	except	finally	for	from
	global	if	import	in
	is	lambda	nonlocal	not
	or	pass	raise	return
	try	while	with	yield
	__peg_parser__
	)! !

! String class methodsFor: #Python-py !
forbiddenPythonSelectors
	" Private - Return the platform forbidden selectors. "

	^#( "class new" )! !

! String methodsFor: #Python-py !
asPythonName
	" Private - Return the receiver as a valid Python name. "

	self first = $_ ifTrue: [ ^self ].
	(String reservedPythonWords includes: self) ifTrue: [ ^'_8',self ].
	^self! !

! String methodsFor: #Python-py !
asPythonSelector
	" Private - Return the receiver as a valid native selector. "

	(String forbiddenPythonSelectors includes: self) ifTrue: [ ^$_ ,self ].
	(self startsWith: $#) ifTrue: [
		^self upTo: $: startingAt: 2
	].
	^self asPrivateSelector asPythonName ! !

! Object methodsFor: #Python-py !
asPython
	" Return the native representation of the receiver.
	The default implementation return the receiver as a String.
	This method must be refined by subclasses implementing native literal representation.
	"

	^self asString ! !

! Object methodsFor: #Python-py !
asPythonContents
	" Return the native representation of the receiver in contents of aCollection. "

	^self asPython ! !

! Boolean methodsFor: #Python-py !
asPython
	" Return the native representation of the receiver. "

	^'smalltalk.' ,(self == true
		ifTrue: [#true]
		ifFalse: [#false]) ! !

! Boolean methodsFor: #Python-py !
asNative: lang
	" Return the native representation of the receiver. "

	lang = #py ifTrue: [ ^self asPython ].
	^super asNative: lang ! !

! Number methodsFor: #Python-py !
asPython
	" Return the native representation of the receiver. "

	^'smalltalk.Number(' ,self printString ,$) ! !

! Number methodsFor: #Python-py !
asPythonContents
	" Return the native representation of the receiver in contents of aCollection. "

	^self printString ! !

! String methodsFor: #Python-py !
isPythonLines
	" Private - Return true if the receiver is a complex string. "

	(self includes: String cr) ifTrue: [ ^true ].
	(self includes: String lf) ifTrue: [ ^true ].
	"(self includes: String tab) ifTrue: [ ^true ]."
	(self includes: 92 asCharacter) ifTrue: [ ^true ].
	^false! !

! String methodsFor: #Python-py !
asPython
	" Return the native representation of the receiver. "

	^'smalltalk.String(' ,self asPythonString ,$) ! !

! String methodsFor: #Python-py !
asPythonContents
	" Return the native representation of the receiver in contents of aCollection. "

	^self asPythonString ! !

! String methodsFor: #Python-py !
asPython: bar cuoted: tag
	" Private - Return the Python representation of the receiver couted with tag (or nil). "

	| result |
	(self includes: tag) ifTrue: [ ^nil ].
	self isPythonLines ifFalse: [ ^nil ].
	result := self replace: bar with: bar ,bar.
	(self endsWith: tag last) ifTrue: [
		^tag	,(result copyFrom: 1 to: result size - 1)
			,bar ,tag last ,tag
	].
	^tag ,result ,tag! !

! String methodsFor: #Python-py !
asPythonString
	" Return the Python representation of the receiver. "

	| tag encoded bar |
	self isEmpty ifTrue: [ ^'""'].
	bar := 92 asCharacter.
	self = bar ifTrue: [ ^$" ,bar ,bar ,$" ].
	encoded := self asPython: bar cuoted: $",$",$".
	encoded notNil ifTrue: [ ^encoded ].
	encoded := self asPython: bar cuoted: $',$',$'.
	encoded notNil ifTrue: [ ^encoded ].

	tag := $".
	(self includes: tag) ifTrue: [ tag := $' ].
	encoded := self replace: bar with: bar ,bar.
	encoded := encoded replace: String crlf with: bar ,$n.
	encoded := encoded replace: String cr with: bar ,$r.
	encoded := encoded replace: String lf with: bar ,$n.
	"encoded := encoded replace: String tab with: bar ,$t."
	encoded := encoded replace: tag with: bar ,tag.
	^tag ,encoded ,tag! !

! Array methodsFor: #Python-py !
asPython
	" Return the native representation of the receiver. "

	^'smalltalk.newArray(' ,self asPythonContents ,$) ! !

! Array methodsFor: #Python-py !
asPythonContents
	" Return the native representation of the receiver in contents of aCollection. "

	^$[ ,((self collect: [:each | each asPythonContents ]) join: $,) ,$] ! !

! Set methodsFor: #Python-py !
asPython
	" Return the native representation of the receiver. "

	^'smalltalk.' ,self class name
	, '.newContents_('
		, contents asPython
	,')'! !

! Dictionary methodsFor: #Python-py !
asPython
	" Return the native representation of the receiver. "

	^'smalltalk.' ,self class name
	, '.fromKeys_and_('
		, keys asArray asPython
		,',' ,values asPython
	,')'! !

! PoolDictionary methodsFor: #Python-py !
asPython
	" Return the native representation of the receiver. "

	^'smalltalk.',self class name
	, '.fromTuples_('
		, self tuples asPython
	,')'! !

" Custom conversion methods
Note: We implement __tostring delegating to #toString implementors"
#toString implementors do: [:mth| | source |
	source := '__tostring
	" Return the printable representation of the receiver. "

	^self toString'.
	(mth classField
		methodsFor: #Python-printing)
		scanFrom: source stream
]!

! Array methodsFor: #Python-printing !
__tostring
	" Return the printable representation of the receiver. "

	^self asLiteral! !

! Array class methodsFor: #Python-instantiation !
basicNew
	" Return an instance of the receiver. "

	{'self'
	py'if True: return self([])'}.
	^super basicNew! !

! String class methodsFor: #Python-instantiation !
basicNew
	" Return an instance of the receiver. "

	{'self'
	py'if True: return self("")'}.
	^super basicNew! !

! Number class methodsFor: #Python-instantiation !
basicNew
	" Return an instance of the receiver. "

	{'self'
	py'if True: return self(0)'}.
	^super basicNew! !

"------------------------------------------------------------------------------
	Native handlers for Python
 ------------------------------------------------------------------------------"!

NativeHandle
	subclass: #Python
	category: #Python!
Python comment: '
	Generic wrapper of native Python object.
	@2021 Alejandro F. Reimondo
	aleReimondo@smalltalking.net'!

! Python class methodsFor: #instantiation !
supportsLanguage: language
	" Private - Return true if instances of the receiver implement native wrappers for language. "

	^language = #py! !

! Python class methodsFor: #calling !
nativeCall: pyFunction args: args kwargs: kwargs
	" Private - Return the native result of calling pyFunction with args and kwargs. "

"
# #print: 'Native call to ' with: pyFunction.
# #print: 'args=' with: args.
# #print: 'kwargs=' with: kwargs.
"
	^{''py'smalltalk.handleOf(pyFunction)(*args,**kwargs)'}! !

! Python class methodsFor: #calling !
call: pyFunction args: anArray kwargs: keywords
	" Return the result of calling pyFunction with positional arguments and keywords arguments. "

	| args kwargs |
	{''py'__0__.args = tuple(anArray.handle())'}.
	kwargs := (keywords isArray
		ifTrue: [PoolDictionary fromTuples: keywords]
		ifFalse: [keywords]) json.
	^self	nativeCall: pyFunction
		args: args kwargs: kwargs! !

! Python methodsFor: #calling !
call: pyFunction args: anArray kwargs: keywords
	" Return the result of calling pyFunction with positional arguments and keywords arguments. "

	^self class call: pyFunction args: anArray kwargs: keywords! !

! Python class methodsFor: #calling !
call: selector on: receiver args: anArray kwargs: keywords
	" Return the native result of calling selector on receiver with positional arguments and keywords arguments. "

	^self	call: (receiver basicAt: selector asSelector) handle
		args: anArray kwargs: keywords! !

! Python class methodsFor: #calling !
send: selector to: receiver withArguments: anArray keywords: keywords
	" Return the result of sending message selector on receiver with positional arguments and keywords arguments. "

	| hResult |
	hResult := self
		call: selector on: receiver
		args: anArray kwargs: keywords.
	^#{smalltalk} asObject: hResult! !

! Python class methodsFor: #accessing !
global
	" Return an instance wrapping global. "

	^self @ Smalltalk global! !

! Python class methodsFor: #accessing !
version
	" Return the host version. "

	^self global @ #_VERSION! !

! Python class methodsFor: #accessing !
arg
	" Return the main arguments (anArray of strings). "

	^self global @ #arg! !

! Python class methodsFor: #modules !
import: aName
	" Return an instance of the receiver obtained from importing module aName. "

	^self @ (self smalltalk import: aName)! !

! Python class methodsFor: #modules !
from: libName import: aName
	" Return an instance of the receiver obtained from importing aName from library libName. "

	^self @ (self smalltalk from: libName import: aName)! !

! Python class methodsFor: #modules !
require: aName
	" Return an instance of the receiver obtained from loading module aName(or nil) "

	^[ self import: aName ] on: Exception do: []! !

! Python class methodsFor: #modules !
from: libName require: aName
	" Return an instance of the receiver obtained from loading aName(or nil) "

	^[ self from: libName import: aName ] on: Exception do: []! !

! Python methodsFor: #accessing !
propertyNames
	" Return the property names of the receiver. "

	^# #dir: self handle! !

! Python methodsFor: #accessing !
handleAt: propertyName
	" Return the property at propertyName of the receiver. "

	^self handle basicAt: propertyName! !

! Python methodsFor: #accessing !
handleAt: propertyName put: anObject
	" Set the property at propertyName of the receiver. "

	^self handle basicAt: propertyName put: anObject! !

! Python methodsFor: #private !
isFunction: aHandle
	" Return true if aHandle is a function. "

	^#( BlockClosure builtin_function_or_method function )
		includes: (self typeOf: aHandle)! !

! Python methodsFor: #iteration !
do: aBlock
	" Evaluates aBlock on each element of the receiver. "

{'' js'	self.invalidMessage_("do:")'
py'	for item in self.handle():%>%aBlock(item)%<%'}.! !

! Python methodsFor: #iteration !
pairsDo: aBlock
	" Evaluates aBlock on each key/value pair of the receiver. "

{'' js'	self.invalidMessage_("pairsDo:")'
py'	h = self.handle()%cr%for k in dir(h):%>%aBlock(k,getattr(h,k))%<%'}.! !

! Python methodsFor: #iteration !
keysDo: aBlock
	" Evaluates aBlock on each key of the receiver. "

{'' js'	self.invalidMessage_("keysDo:")'
py'	for k in dir(self.handle()):%>%aBlock(%String%(k))%<%'}.! !

! Python methodsFor: #iteration !
valuesDo: aBlock
	" Evaluates aBlock on each value of the receiver. "

{'' js'	self.invalidMessage_("valuesDo:")'
py'	h = self.handle()%cr%for k in dir(h):%>%aBlock(getattr(h,k))%<%'}.! !

! Python methodsFor: #converting !
asDictionary
	" Convert the receiver to a(Pool)Dictionary. "

	^PoolDictionary @@ self handle! !

Python
	subclass: #PythonObject
	category: #Python!
PythonObject comment: '
	Implementation of wrappers to native Python objects.
	@2021 Alejandro F. Reimondo
	aleReimondo@smalltalking.net'!
PythonObject subclass: #PythonClass!

! Python methodsFor: #python-accessing !
getClass
	" Returns the native class of the receiver. "

	^PythonClass @ (self @#__class__) handle! !

! Python class methodsFor: #instantiation !
handle: aHandle prefix: prefix
	" Returns a concrete wrapper for native object aHandle.
	The wrapper support will be lookup with prefixed native name.
	"

	| delegates result |
	aHandle isNil ifTrue: [ ^nil ].
	result := self @ (self handleOf: aHandle).
	delegates := Array with: result getClass.
	delegates addAll: delegates first mro.
	delegates do: [:each| | support |
		support := Smalltalk at: prefix ,each name asGlobalName ifAbsent: [].
		support notNil ifTrue: [ ^support @ aHandle ]
	].
	^result! !

! Python class methodsFor: #instantiation !
handle: aHandle
	" Returns a concrete wrapper for native object aHandle. "

	^self handle: aHandle prefix: self wrapperPrefix! !

! Python class methodsFor: #private !
wrapperPrefix
	" Private - Returns the prefix for searching wrapper implementor. "

	^''! !

! Python methodsFor: #accessing !
<@ path
	" Return the most concrete instance at path of the receiver (or nil). "

	^self class handle: self @@ path! !

! Python methodsFor: #python-accessing !
getClassName
	" Returns the name of the native class of the receiver. "

	^self getClass @#__name__ ! !

! PythonClass class methodsFor: #instantiation !
named: globalName ifAbsent: aBlock
	" Returns the global object (at globalName) as a instance of the receiver or the result of evaluating aBlock. "

	| hResult |
	hResult := Smalltalk nativeObjectAt: globalName ifAbsent: [ ^aBlock evaluate ].
	^self @ hResult ! !

! PythonClass class methodsFor: #instantiation !
named: globalName
	" Returns the global object (at globalName) as a instance of the receiver (or nil). "

	^self named: globalName ifAbsent: [] ! !

! PythonClass methodsFor: #accessing !
name
	" Returns the name of the receiver. "

	^self @ #__name__ ! !

! PythonClass methodsFor: #accessing !
mro
	" Returns the mro delegates of the receiver. "

	^self @ #__mro__ collect: [:each| PythonClass @ each handle ]! !

! Object class methodsFor: #Python-converting !
@ aHandle
	" Return a S8 Object instance with aHandle . "

	^#{smalltalk} asObject: aHandle! !

! String methodsFor: #native !
sendTo: aReceiver withArguments: args keywords: kwargs
	" Send the message to aReceiver with args and kwargs.
	Returns the (object) result of calling function (given by the receiver) on aReceiver.
	args is anArray and kwargs can be a PoolDictionary or ComposedAssociation. "

	^Python send: self to: aReceiver withArguments: args keywords: kwargs! !

! Object methodsFor: #native !
call: selector withArguments: args keywords: kwargs
	" Send the message(selector) to the receiver's handle with args and kwargs. "

	^Python	send: selector to: self handle
		withArguments: args keywords: kwargs! !

! Object methodsFor: #native !
perform: selector withArguments: args keywords: kwargs
	" Send the message(selector) to the receiver with args and kwargs. "

	^Python send: selector to: self withArguments: args keywords: kwargs! !

! Class methodsFor: #native !
linkTo: spec
	" Bind the receiver to native objects in spec.
	spec can be a global name of a native class
	or an array of tuples of module and attribute name to bind (see senders of #nativeLinkage)
	Note that the linking process will remove the receiver and create a NEW class bound to link referents.
	The linking will persist in exports/snapshots and referents must be available at loading time.
	In case of problems resolving linkage during loading, resulting behavior of instances is undefined.
	"

	| backup st clsName |
	clsName := self name.
	st := self smalltalk.
	self class
		implement: #nativeLinkage
		comment: 'Return the native linkage of the receiver'
		as: $^ ,spec asLiteral
		category: #native.
	backup := Exporter exportAll: self withAllSubclasses.
	self removeFromSystem.
	st eval: backup.
	^st at: clsName! !

! PythonEncoder methodsFor: #export !
encode: aClass linkageCheck: spec on: stream
	" Private - Generate the code to check native linkage when loading. "

	spec isArray ifFalse: [ ^self ].
	stream nextPutAll: 'if smalltalk.available(',
		(spec inject: '' into: [:total :pair|
			total, $" ,pair first ,$" ,$,
		]) ,'):' ;cr! !

! PythonEncoder methodsFor: #export !
encode: aClass native: nativeExpression on: stream
	" Private - Generate the code to map aClass to nativeExpression when loading. "

	| code |
	code := 'smalltalk.mapClassName("'
		,aClass name ,'","'
		,aClass category, '",'
		,nativeExpression, ','
		,(self smalltalkNameFor: aClass superclass)
		,$).
	stream nextPutAll: code;cr! !

! PythonEncoder methodsFor: #export !
encode: aClass linkageMap: spec on: stream
	" Private - Generate the code to map native linkage when loading. "

	spec isArray ifFalse: [ ^self encode: aClass native: spec on: stream ].
	stream	tab;nextPutAll: 'native = (';
		nextPutAll: (spec inject: '' into: [:total :pair| total,
			'smalltalk.moduleAt("' ,pair first ,'","' ,pair last ,'"),'
		]); nextPutAll: $);
		cr;tab.
	self encode: aClass native: #native on: stream! !

! PythonEncoder methodsFor: #export !
encode: aClass linkage: spec on: stream
	" Private - Generate the code to bind to native linkage when loading. "

	spec size = 0 ifTrue: [ ^self ].
	self encode: aClass linkageCheck: spec on: stream.
	self encode: aClass linkageMap: spec on: stream.! !

! PythonEncoder methodsFor: #export !
defaultRequirementsFor: aClass
	" Private - Return the default (collection of) requirements to load aClass. "

	| result |
	result := super defaultRequirementsFor: aClass.
	(aClass class implements: #nativeLinkage) ifTrue: [
		| stream |
		stream := '' writeStream.
		self encode: aClass linkage: aClass nativeLinkage on: stream.
		result add: stream contents
	].
	^result ! !

"------------------------------------------------------------------------------
	Extension for Python API builders
 ------------------------------------------------------------------------------"!

! Python class methodsFor: #builders !
varTag
	" Private - Return the character separator used to declare multiple variable returns. "

	^$: ! !

! Python class methodsFor: #builders !
handleAPIFor: literal
	" Private - Return true when the receiver support encoding of literal.
	See NativeObject class>>#apiSupportFor:
	Note: we support multiple variables encoding
	"

	^literal isString and: [ literal includes: self varTag ]! !

! Python class methodsFor: #builders !
nativePrologueFor: aReturnType
	" Private - Return the prologue to use to wrap a return. "

	(aReturnType includes: self varTag) ifTrue: [
		^self apiPrologueForVariables: aReturnType
	].
	^super nativePrologueFor: aReturnType! !

! Python class methodsFor: #builders !
nativeEpilogueFor: aReturnType
	" Private - Return the epilogue code to use to wrap a return. "

	(aReturnType includes: self varTag) ifTrue: [
		^self apiEpilogueForVariables: aReturnType
	].
	^super nativeEpilogueFor: aReturnType! !

! Python class methodsFor: #builders !
apiVariablesAndTypes: aReturnType
	" Private - Return an array of variables w/type for api call. "

	^(aReturnType asArrayOfSubstringsSeparatedBy: self varTag) collect: [:each|
		| token | token := $@.
		(each includes: token) ifTrue: [
			Array	with: (each fromLast: token)
				with: (each upTo: token)
		] ifFalse: [ Array with: each ]
	]! !

! Python class methodsFor: #builders !
apiPrologueForVariables: aReturnType
	" Private - Return the prologue to return variables. "

	| vars |
	vars := self apiVariablesAndTypes: aReturnType.
	^(vars inject: $| into: [:total :each| total ,$  ,each first ]) ,' |
	',(vars inject: nil into: [:total :each|
		total isNil ifTrue: [ each first ,'::' ,$' ]
		ifFalse: [ total ,$, ,each first ]
	]) ,$' ,' := '! !

! Python class methodsFor: #builders !
apiEpilogueForVariables: aReturnType
	" Private - Return the epilogue to return variables. "

	^(self apiVariablesAndTypes: aReturnType) inject: '.
	^Array' into: [:total :each|
		total ,' with: ' ,(self apiCodeForVariable: each)
	]! !

! Python class methodsFor: #builders !
apiCodeForVariable: tuple
	" Private - Return the code to push variable in tuple. "

	tuple size = 1 ifTrue: [ ^tuple first ].
	^tuple second ,$@ ,tuple first ! !

"------------------------------------------------------------------------------
	Development Utilities & tools for Python environment
 ------------------------------------------------------------------------------"!

! MethodReferences methodsFor: #Tools-patterns !
superPythonSendPatternOf: aSelector
	" Private - Return the search pattern for senders (to super) of aSelector. "

	| encoder |
	encoder := PythonEncoder new.
	^',self).'
		,(encoder asPrivateSelector: aSelector)
		,'('! !

! MethodReferences methodsFor: #Tools-patterns !
pyMessagesStreamOn: aCompiledMethod
	" Private - Return the stream for reading messages on compiled code of aCompiledMethod. "

	| src |
	src := aCompiledMethod compiledSource
		replace: 'smalltalk.' with: ''. "remove global refs"
	^src stream
		upTo: self compiledSourceStart;
		yourself ! !

! Object methodsFor: #Python-fileIn !
fileIn: fileName
	" FileIn contents of file at fileName. "

	self note: 'fileIn("', fileName, '")'.
	^# #fileIn: fileName! !

! Object methodsFor: #Python-fileIn !
fileInJS: fileName
	" Load native code from file at fileName. "

	self note: 'fileInJS("', fileName, '")'.
	^# #fileInJS: fileName! !

