
Object
	subclass: #Snapshot
	category: #Snapshot!
Snapshot comment: '
	Dump a binary representation of this S8 system onto an image (String).
	The default implementation export full/runnable images containing boot sections.
	The subclasses can refine behavior to select/customize exports and format.
	@2011-today Alejandro Reimondo aleReimondo@smalltalking.net'!

! String methodsFor: 'Snapshot-io' !
outputToFile: fileName
	" Save the receiver's contents onto a file at fileName. "

	^# #fileWrite: fileName string: self mode: #wb! !

! String methodsFor: 'Snapshot-io'  !
appendToFile: fileName
	" Append the receiver's contents onto a file at fileName. "

	^# #fileWrite: fileName string: self mode: #ab! !

! Class methodsFor: 'Snapshot-startup' !
launch
	" Private - The system has been launched and start execution.
	The default action is to do nothing.
	Subclasses can refine this method to execute actions on startup.
	"! !

! Class methodsFor: 'Snapshot-snapshot' !
aboutToSaveImage: snapshot on: stream
	" Private - A system snapshot will be taken, the receiver can prepare the contents to be stored on snapshot.
	Must return the receiver or a replacement to act as the receiver when persisting data.
	The default action return the receiver.
	"

	^self! !

! Class methodsFor: 'Snapshot-snapshot' !
imageSaved: snapshot on: stream
	" Private - A system snapshot has been taken, the receiver can emmit code to restore state.
	The default action do nothing.
	This method can be refined by the subclasses.
	"! !

! Class methodsFor: 'Snapshot-snapshot' !
aboutToExit: aSystemDictionary
	" Private - A system will exit, and the receiver can act executing actions just before exiting.
	The default action do nothing.
	This method can be refined by the subclasses.
	"! !

! Smalltalk class methodsFor: 'Snapshot-snapshot' !
saveImageAs: fileName
	" Dump the image contents onto file at fileName. "

	self current saveImageAs: fileName! !

! Smalltalk class methodsFor: 'Snapshot-snapshot' !
saveImage
	" Dump the image contents onto image file. "

	self current saveImage! !

! Smalltalk class methodsFor: 'Snapshot-exiting'!
exit
	" Exit to OS system. "

	self current exit! !

! Smalltalk methodsFor: 'Snapshot-exiting'!
exit
	" Exit to OS system. "

	self classes do: [:each| each aboutToExit: self ].
	{'' doc'quit to OS'}.! !

! Smalltalk methodsFor: 'Snapshot-accessing' !
product
	" Private - Return the name of the system. "

	^#S8! !

! Smalltalk methodsFor: 'Snapshot-accessing' !
description
	" Return the description of the receiver. "

	^'A modern Smalltalk environment running on the "invisible machine".'! !

! Smalltalk methodsFor: 'Snapshot-license' !
licenseComment
	" Private - Comment note about license of this product. "

	^'
The text you are reading here is the residue of something that has run before.
Keep the system running with responsibility. Smalltalk is not the words,
 nor any definition, nor it current contents...
Smalltalk include your feelings when you see it running.
Smalltalk is what happens when time flows and you "are" here.
Stop reading/writing code (thinking you are outside the system),
 change your world with responsibility, one object at a time, with your hands,
 during all your life, in family.
'! !

! Smalltalk methodsFor: 'Snapshot-license' !
licenseTerms
	" Private - Return the license terms to use this system. "

	^self product ,' - ' ,self description ,'
' ,self licenseMoreHolders ,'
' ,self licenseHolders ,'
Permission is hereby granted, free of charge, to any person obtaining
 a copy of this software and associated documentation files (the "Software"),
 to deal in the Software without restriction, including without limitation
 the rights to use, copy, modify, merge, publish, distribute, sublicense,
 and/or sell copies of the Software, and to permit persons to whom the
 Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included
 in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
 IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
 CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
 TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
 OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

' ,self licenseComment! !

! Smalltalk methodsFor: 'Snapshot-license' !
licenseYear
	" Private - Return the year license information. "

	^[ DateTime today year toString
	] on: Exception do: [ #today ] ! !

! Smalltalk methodsFor: 'Snapshot-license' !
licenseMoreHolders
	" Private - Return additional license holders information. "

	^''! !

! Smalltalk methodsFor: 'Snapshot-license' !
licenseHolders
	" Private - Return the license holders information. "

	^'Copyright (C) '
	,self licenseYear asString ,' '
	,'Alejandro F. Reimondo <aleReimondo@smalltalking.net>

Parts of source code written for Jtalk (Sept 2011),
Copyright (C) 2011 by Nicolas Petton <petton.nicolas@gmail.com>
Copyright (C) 2011 Jtalk contributors
	David Graham <dgraham@iphouse.com>
	Goran Krampe <goran@krampe.se>
Also code and ideas from Clamato (http://clamato.net), written by Avi Byrant.
The PetitParser library, published by Lukas Renggli (http://lukas-renggli.ch) and released under the MIT license.
And people that contributed to Smalltalk on diverse media years(decades) before this license holders claimed ownership.
'! !

! Smalltalk methodsFor: 'Snapshot-evaluating' !
globalsDo: aBlock
	" Private - Evaluates aBlock for each globals defined in the receiver. "

	^self #globalsDo: aBlock! !

! Smalltalk methodsFor: 'Snapshot-evaluating' !
globalsDo: aBlock separatedBy: anotherBlock
	" Private - Evaluates aBlock for each globals defined in the receiver. "

	| first | first := true.
	^self #globalsDo: [:each|
	    first
		ifTrue: [ first := false ]
		ifFalse: [ anotherBlock value ].
	    aBlock value: each
	]! !

! Smalltalk methodsFor: 'Snapshot-writing' !
saveImageOn: stream
	" Dump the receiver's contents onto stream. "

	| exporter allClasses |
	stream nextPutAll: self licenseTerms asComment; cr.
	allClasses := ((self rootClasses
		inject: #() into: [:total :each| total addAll: each withAllSubclasses. total ])
		collect: [:each| each aboutToSaveImage: self on: stream ])
		select: [:each| each notNil ].
	exporter := Exporter new.
"	stream nextPutAll: 'Smalltalk Globals (pre)definition.' asComment; cr.
	self
		globalsDo: [:key| stream nextPutAll: 'smalltalk.',key ]
		separatedBy: [ stream nextPutAll: ' , ' ].
	stream nextPutAll: '=nil'; cr.
"	stream cr; nextPutAll: 'Smalltalk Classes.' asComment; cr.
	exporter exportClasses: allClasses on: stream.
	stream cr; nextPutAll: 'Smalltalk Globals set-up.' asComment.
	self globalsDo: [:key|
		stream cr; nextPutAll: 'smalltalk.',key,'='
			,(exporter exportObject: (self basicAt: key))
	].
	allClasses do: [:each| each imageSaved: self on: stream ].! !

! Smalltalk methodsFor: 'Snapshot-contents' !
imageContents
	" Return a dump of the system contents. "

	| stream |
	stream := String new writeStream.
	self saveImageOn: stream.
	^stream contents! !

! Smalltalk methodsFor: 'Snapshot-contents' !
snapshot
	" Return a system snapshot. "

	^Snapshot contents! !

! Smalltalk methodsFor: 'Snapshot-fileName' !
imageFileName
	" Return the image file name. "

	^self basicAt: '$imageFileName'! !

! Smalltalk methodsFor: 'Snapshot-fileName' !
imageFileName: fileName
	" Set the image file name. "

	^self basicAt: '$imageFileName' put: fileName! !

! Smalltalk methodsFor: 'Snapshot-save' !
saveImage
	" Dump the image contents onto default image file. "

	self saveImageAs: self imageFileName! !

! Smalltalk methodsFor: 'Snapshot-save' !
saveImageAs: fileName
	" Dump the image contents onto file at fileName. "

	String doItSelector implementors do: [:each| each removeFromSystem ].
	self imageContents outputToFile: fileName.
	self imageFileName: fileName.! !

! Smalltalk methodsFor: 'Snapshot-startup' !
launch
	" The system has been launched and start execution.
	The default action is to propagate #launch to current classes and evaluate #LaunchExpression.
	"

	self classes do: [:each| each launch ].
	^(Smalltalk at: #LaunchExpression ifAbsent: []) evaluate! !

! Snapshot class methodsFor: #accessing !
contents
	" Return a snapshot of current system contents. "

	^self new contents! !

! Snapshot class methodsFor: #saveImage !
outputToFile: fileName
	" Save a complete system snapshot in a file at fileName. "

	String doItSelector implementors do: [:each| each removeFromSystem ].
	self contents outputToFile: fileName! !

! Snapshot methodsFor: #accessing !
contents
	" Return a snapshot of current system contents. "

	| stream |
	stream := String new writeStream.
	self saveImageOn: stream.
	^stream contents! !

! Snapshot methodsFor: #saveImage !
saveImageOn: stream
	" Save the current system contents onto stream. "

	^self saveImage: self smalltalk on: stream! !

! Snapshot methodsFor: #saveImage !
saveImage: aSmalltalk on: stream
	" Save aSmalltalk system contents onto stream. "

	self
		writeBoot: aSmalltalk on: stream;
		write: #Prologue on: stream;
		writeContents: aSmalltalk on: stream;
		write: #Epilogue on: stream;
		writeLaunchExpressionOn: stream;
		yourself! !

! Snapshot methodsFor: #writing !
writeLicense: aSmalltalk on: aStream
	" Private - Write the license terms onto aStream. "

	aStream nextPutAll: aSmalltalk licenseTerms asComment; cr! !

! Snapshot methodsFor: #js !
writeJSBootOn: aStream
	" Private - Write the jsBoot section of the image onto aStream. "

	self writeLicense: Smalltalk current on: aStream.
	self writeGlobals: #(
		#StObject #StBehavior #StClass #StMetaclass
		#StMethod
		#StUndefinedObject
		) on: aStream.
	self writeNilDefinitionOn: aStream.
	self writeGlobals: #( #StSystemDictionary ) on: aStream.
	self writeSystemDefinitionOn: aStream.! !

! Snapshot methodsFor: #encoding !
bootHeader
	" Private - Return the header for boot contents. "

	^{''}! !

! Snapshot methodsFor: #encoding !
bootFooter
	" Private - Return the footer for boot contents. "

	^{''}! !

! Snapshot methodsFor: #writing !
writeBoot: aSmalltalk on: aStream
	" Private - Write the boot section of the image onto aStream. "

	| boot bootName save |
	save := [ | text |
		self writeLicense: aSmalltalk on: aStream.
		text := self bootHeader ,boot ,self bootFooter.
		self writeString: text on: aStream.
		text
	].

	boot := aSmalltalk at: #BootSector ifAbsent: [].
	(boot respondsTo: #for:) ifTrue: [
		boot := (boot for: self nativeLanguage) contents.
		boot notNil ifTrue: [
			^self writeString: boot on: aStream
		].
	].

	bootName := 'S8_Boot_' ,String nativeFileExtension.
	boot := aSmalltalk at: bootName ifAbsent: [].
	boot notNil ifTrue: [
		"embedded boot sector"
		^save value
	].

	(self write: #Boot on: aStream) ifTrue: [
		"custom implementation by the receiver"
		^self
	].

	bootName := 's8/start/boot.' ,String nativeFileExtension.
	boot := bootName fileContents.
	boot notNil ifTrue: [
		"embedding from external boot file"
		^aSmalltalk at: bootName put: save value
	].
	^self error: 'Missing boot resource'! !

! Snapshot methodsFor: #writing !
write: section on: aStream
	" Private - Write the section of the image onto aStream. "

	| msg |
	aStream cr.
	msg:= #write ,self nativeLanguage asUppercase ,section ,#On:.
	(self respondsTo: msg) ifTrue: [
		self perform: msg with: aStream.
		^true
	].
	^false! !

! Snapshot methodsFor: #js !
writeJSPrologueOn: aStream
	" Private - Write the js prologue section of the image onto aStream. "

	self writeString: 'smalltalk.$state = "js-toolset"' on: aStream.
	self writeGlobals: #(
		'smalltalk.st' 'smalltalk.doIt'
		#fileIn #fileInJS
		#doit #exportCategory
		#saveImage #saveImageAs
		) on: aStream.
	self writeString: 'smalltalk.$state = "initialized"' on: aStream.
	aStream cr.! !

! Snapshot methodsFor: #encoding !
launchExpression
	" Private - Return the native launch expression. "

	^{''}! !

! Snapshot methodsFor: #writing !
writeLaunchExpressionOn: aStream
	" Private - Write the epilogue section of the image onto aStream. "

	| launch |
	launch := self launchExpression.
	launch notNil ifTrue: [ self writeString: launch on: aStream ].! !

! Snapshot methodsFor: #writing !
writeString: aString on: aStream
	" Private - Write aString onto aStream. "

	aStream cr; nextPutAll: aString! !

! Snapshot methodsFor: #error !
globalNotFound: aGlobal
	" Private - A global at native level has not been found. "

	^self error: 'Global not found: ',aGlobal! !

! Snapshot methodsFor: #writing !
writeGlobals: globals on: aStream
	" Private - Write the prologue section of the image onto aStream. "

	| object varName |
	globals do: [:each|
		object := Smalltalk current eval: each.
		object isNil ifTrue: [ object := self globalNotFound: each ].
		object notNil ifTrue: [
			varName := (each includes: $.) ifTrue: [ each ] ifFalse: [ 'var ',each ].
			self writeString: varName,' = ', object toString on: aStream.
		]
	]! !

! Snapshot methodsFor: #js !
writeNilDefinitionOn: aStream
	" Private - Write the definition of nil object onto aStream. "

	| sample |
	sample := nil.
	self writeString: 'var nil = new StUndefinedObject()' on: aStream.
	self writeString: 'nil.toString = ', sample#toString toString on: aStream.
	self writeString: 'nil.isNil_ = ', sample#isNil_ toString on: aStream.
	self writeString: 'nil.is_eqeq_ = ', sample#is_eqeq_ toString on: aStream.! !

! Snapshot methodsFor: #js !
writeSystemDefinitionOn: aStream
	" Private - Write the definition of smalltalk object onto aStream. "

	self writeString: 'var smalltalk = new StSystemDictionary()' on: aStream.
	self writeString: 'smalltalk.toString = function() { return "smalltalk" }' on: aStream.
	self writeString: 'smalltalk.startup();' on: aStream.! !

! Snapshot methodsFor: #writing !
writeContents: aSmalltalk on: aStream
	" Private - Write the image section of aSmalltalk system onto aStream. "

	| exporter globals classes |
	aStream cr.
	exporter := self exporterFor: aSmalltalk.
	globals := self globalsIn: aSmalltalk.
	classes := ((self classesToExportFrom: aSmalltalk with: exporter)
		collect: [:each| each aboutToSaveImage: self on: aStream ])
		select: [:each| each notNil ].
"	self exportGlobalNames: globals keys sorted in: aSmalltalk on: aStream.
"	exporter exportClasses: classes on: aStream.
	self exportGlobals: globals with: exporter on: aStream.
	classes do: [:each| each imageSaved: self on: aStream ].! !

! Snapshot methodsFor: #exporting !
exporterFor: aSmalltalk
	" Private - Return the exporter to use to write contents of aSmalltalk system. "

	^Exporter new! !

! Snapshot methodsFor: #exporting !
globalsIn: aSmalltalk
	" Private - Return the globals defined in aSmalltalk system. "

	| result |
	result := PoolDictionary new.
	aSmalltalk globalsDo: [:key| result at: key put: (aSmalltalk basicAt: key) ].
	^result! !

! Snapshot methodsFor: #exporting !
classesToExportFrom: aSmalltalk with: exporter
	" Private - Return the classes to export from aSmalltalk system. "

	^aSmalltalk rootClasses
		inject: #() into: [:total :each| total addAll: each withAllSubclasses. total ]! !

! Snapshot methodsFor: #exporting !
exportGlobalNames: names in: aSmalltalk on: aStream
	" Private - Export global names in aSmalltalk system. "

	names
		do: [:key| aStream nextPutAll: 'smalltalk["',key,'"]' ]
		separatedBy: [ aStream nextPutAll: ' , ' ].
	aStream nextPutAll: '=nil'; cr.! !

! Snapshot methodsFor: #exporting !
exportGlobals: globals with: exporter on: aStream
	" Private - Export global objects onto aStream. "

	globals associationsDo: [:binding|
		self writeString: 'smalltalk["',binding key,'"]='
			,(exporter exportObject: binding value)
			on: aStream
	]! !

EventManager
	subclass: #SystemManager
	instanceVariableNames: ''
	classVariableNames: ' Current '
	poolDictionaries: ''
	category: #Snapshot!

! SystemManager class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^super constructEventsTriggered ,#(
		#aboutToBind:to:
		#methodBound:to:
		#aboutToRemove:from:
		#methodRemoved:from:
		#aboutToChange:to:
		#aboutToCreateClass:with:
		#classChanged:
		#classShapeChanged:
		#object:at:notFoundIn:
		#log:
		#memoryWarning:
	)! !

! SystemManager class methodsFor: #singleton !
currentOrNil
	" Return the current instance of the receiver. "

	^self classVariableAt: #Current ifAbsent: [ nil ]! !

! SystemManager class methodsFor: #singleton !
current
	" Return the current instance of the receiver. "

	Current isNil ifTrue: [ self current: self new ].
	^Current! !

! SystemManager class methodsFor: #singleton !
current: aSystemManager
	" Set the current instance of the receiver. "

	Current := aSystemManager.
	aSystemManager install.! !

! SystemManager class methodsFor: #log !
log: aString
	" Log aString onto loggin stream. "

	Current notNil ifTrue: [ Current log: aString ]! !

! SystemManager class methodsFor: #memory !
memoryWarning: anArgument
	" Trigger a memory warning event. "

	Current notNil ifTrue: [ Current memoryWarning: anArgument ]! !

! SystemManager methodsFor: #install !
install
	" Private - Install the receiver as the source manager of the system.
	The default action is to do nothing.
	This method can be refined by subclasses.
	"! !

! SystemManager methodsFor: 'tracking-methods' !
aboutToBind: aMethod to: aClass
	" Private - The method will be bound to aClass. "

	^self triggerEvent: #aboutToBind:to: with: aMethod with: aClass ! !

! SystemManager methodsFor: 'tracking-methods' !
methodBound: aMethod to: aClass
	" Private - The method has been bound to aClass. "

	^self triggerEvent: #methodBound:to: with: aMethod with: aClass ! !

! SystemManager methodsFor: 'tracking-methods' !
aboutToRemove: aMethod from: aClass
	" Private - The method will be removed from aClass. "

	^self triggerEvent: #aboutToRemove:from: with: aMethod with: aClass ! !

! SystemManager methodsFor: 'tracking-methods' !
methodRemoved: aMethod from: aClass
	" Private - The method has been removed from aClass. "

	^self triggerEvent: #methodRemoved:from: with: aMethod with: aClass ! !

! SystemManager methodsFor: 'tracking-classes' !
aboutToChange: aClass superclass: superclass iVarNames: iVarNames category: category
	" Private - aClass is about to be changed. "

	^self	triggerEvent: #aboutToChange:to:
		with: aClass
		with: (Array with: superclass with: iVarNames with: category)! !

! SystemManager methodsFor: 'tracking-classes' !
aboutToCreateClass: globalName superclass: superclass iVarNames: iVarNames category: category
	" Private - Attemp to create Class with name globalName. "

	^self	triggerEvent: #aboutToCreateClass:with:
		with: globalName
		with: (Array with: superclass with: iVarNames with: category)! !

! SystemManager methodsFor: 'tracking-classes' !
classChanged: aBehavior
	" Private - aBehavior has been changed. "

	^self triggerEvent: #classChanged: with: aBehavior ! !

! SystemManager methodsFor: 'tracking-classes' !
classShapeChanged: aBehavior
	" Private - aBehavior has changed shape. "

	^self	triggerEvent: #classShapeChanged:
		with: aBehavior
		ifNotHandled: [ aBehavior recompileAll ]! !

! SystemManager methodsFor: #log !
log: aString
	" Private - Log aString onto loggin device. "

	^self triggerEvent: #log: with: aString! !

! SystemManager methodsFor: #streaming !
nextPutAll: aString
	" Log aString when streaming. "

	^self log: aString! !

! SystemManager methodsFor: #streaming !
cr
	" Ignore the output, we log as line oriented streaming service. "! !

! SystemManager methodsFor: #memory !
memoryWarning: anArgument
	" Trigger a memory warning event.
	The execution environment can signal memory warning/low conditions, we trigger an event to be handled at application level.
	Note: It is frequent procedure to clean cached resources that can be reconstructed later, if required.
	"

	^self triggerEvent: #memoryWarning: with: anArgument! !

! SystemManager methodsFor: #failures !
object: anObject at: indexOrKey notFoundIn: aCollection
	" Trigger a missing object event. "

	^self triggerEvent: #object:at:notFoundIn:
		withArguments: (Array with: anObject with: indexOrKey with: aCollection)
		ifNotHandled: [
			self print: 'Object/key('
				,(indexOrKey isNil ifTrue: [anObject] ifFalse: [indexOrKey]) toString
				,') not found in ',aCollection printString
		]! !

! Object methodsFor: #evaluating !
logEvaluate: source
	" Return the result of evaluating the receiver.
	Log the source expression as reference of evaluation.
	"

	| result |
	result := self evaluate.
	SystemManager log: source.
	^result! !