
! Stream methodsFor: #Tools-fileIn!
nextChunk
	" FileIn - Return the next chunk read from the receiver.
	Trailing non-printing characters are skipped. "

	| result |
	result := self upTo: $!!.
	[ self peek = $!! ] whileTrue: [
		self next.
		result := result, '!!', (self upTo: $!!)
	].
	^result trimFirst: [:c| c asciiValue <= 32 ]! !

! Stream methodsFor: #Tools-fileIn!
fileInto: aReceiver
	" FileIn - FileIn the receiver's contents (in chunk format) using aReceiver as default scanner. "

	| scanner |
	[self atEnd] whileFalse: [
		scanner isNil ifTrue: [ scanner := aReceiver ].
		scanner := scanner scanFrom: self
	].
	^scanner! !

! Stream methodsFor: #Tools-fileIn!
fileIn
	" FileIn - FileIn the receiver's contents (in chunk format). "

	^self fileInto: nil! !

! Stream methodsFor: #Tools-fileOut !
nextChunkPut: aString
	" Output aString in chunk format. "

	| eoc parts |
	eoc := $!!.
	aString isNil ifTrue: [  self nextPut: eoc. ^nil ].
	aString isEmpty ifTrue: [ self nextPut: $ ].
	(aString asArrayOfSubstringsSeparatedBy: eoc)
		do: [:each| self nextPutAll: each ]
		separatedBy: [ self nextPutAll: eoc,eoc ].
	self nextPut: eoc.
	^aString! !

! Object methodsFor: #Tools-fileIn!
scanFrom: aStream
	" FileIn - Return the result of evaluating the next chunk of data from aStream.
	The method can be refined by chunk readers to support custom fileIn features.
	Must return the next reader to use for reading remaining contents from aStream.
	"

	| chunk |
	chunk := aStream nextChunk.
	chunk isEmpty ifTrue: [ ^nil ].
	^chunk evaluateIn: self! !

Object
 subclass: #FileInReader
 category: #Tools-fileIn!
FileInReader comment: '
	Perform sequential reading from aStream in chunk format.
	Instances implement #scanFrom: message, advancing aStream upto next reader instantiation chunk.'!

! FileInReader methodsFor: #fileIn !
scanFrom: aStream
	" FileIn - Return the result of evaluating the next chunk of data from aStream.
	The method must be implemented by chunk readers to support custom fileIn features.
	"

	^self subclassResponsibility: #scanFrom:! !

FileInReader
 subclass: #ChunkReader
 instanceVariableNames: 'actionBlock terminationBlock'
 category: #Tools-fileIn!

ChunkReader comment: '
	Pluggable FileInReader.
	actionBlock is evaluated with input stream as argument.
	Read chunks until empty chunk, terminationBlock is nil or evaluates to true.'!

! ChunkReader class methodsFor: #instantiation !
doing: aBlock until: terminationBlock
	" Return an instance of the receiver. "

	^self new
		actionBlock: aBlock;
		terminationBlock: terminationBlock;
		yourself! !

! ChunkReader class methodsFor: #instantiation !
doing: aBlock
	" Return an instance of the receiver. "

	^self doing: aBlock until: [ true ]! !

! ChunkReader class methodsFor: #instantiation !
until: terminationBlock
	" Return an instance of the receiver. "

	^self doing: nil until: terminationBlock! !

! ChunkReader class methodsFor: #instantiation !
ignore
	" Return an instance of the receiver.
	The returned reader will ignore/absorb next chunk.
	"

	^self new! !

! ChunkReader class methodsFor: #instantiation !
javascriptChunk
	" Obsolete - use #nativeChunk "

	^self nativeChunk ! !

! ChunkReader class methodsFor: #instantiation !
javascript
	" Obsolete - use #native "

	^self native ! !

! ChunkReader class methodsFor: #instantiation !
nativeChunk
	" Return an instance of the receiver.
	The returned reader will evaluate next chunk as native code (in global context).
	"

	^self doing: [:aStream| Smalltalk current eval: aStream nextChunk ]! !

! ChunkReader class methodsFor: #instantiation !
native
	" Return an instance of the receiver.
	The returned reader will evaluate following contents as native code (in global context).
	The stream will be consumed up to end.
	"

	^self doing: [:aStream| Smalltalk current eval: aStream upToEnd ]! !

! ChunkReader methodsFor: #accessing !
actionBlock
	" Return the actionBlock of the receiver. "

	^actionBlock!

actionBlock: aBlock
	" Set the actionBlock of the receiver. "

	actionBlock := aBlock!

terminationBlock
	" Return the terminationBlock of the receiver. "

	^terminationBlock!

terminationBlock: aBlock
	" Set the terminationBlock of the receiver. "

	terminationBlock := aBlock! !

! ChunkReader methodsFor: #fileIn !
scanFrom: aStream
	" Scan the next chunk of data from aStream. "

	| chunk |
	self actionBlock isNil ifTrue: [
		chunk := aStream nextChunk.
		chunk isEmpty ifTrue: [ ^nil ].
	] ifFalse: [ self actionBlock value: aStream ].
	(self terminationBlock isNil
		or: [ true == self terminationBlock value ])
		ifTrue: [ ^nil ].
	^self! !

! Object methodsFor: #Tools-fileIn !
javascriptCodeFollows
	" Obsolete - use #nativeCodeFollows"

	^self nativeCodeFollows! !

! Object methodsFor: #Tools-fileIn !
nativeCodeFollows
	" FileIn - FileIn native code on current stream.
	Consume stream contents upto end of stream evaluating the contents as native code.
	"

	^ChunkReader native! !

FileInReader
	subclass: #ClassCategoryReader
	instanceVariableNames: 'class category stamp'
	category: #Tools-fileIn!

! ClassCategoryReader methodsFor: #accessing !
class: aClass category: aString
	" Private - Set the class and category of the receiver. "

	class := aClass.
	category := aString! !

! ClassCategoryReader methodsFor: #accessing !
class: aClass category: aString stamp: aStamp
	" Private - Set the class and category of the receiver. "

	self class: aClass category: aString.
	stamp := aStamp! !

! ClassCategoryReader methodsFor: #fileIn !
scanFrom: aStream
	" Scan the next chunk of data from aStream. "

	| chunk |
	chunk := aStream nextChunk.
	^[ chunk isEmpty ifFalse: [ self compileMethod: chunk ]
	] logEvaluate: chunk! !

! ClassCategoryReader methodsFor: #private !
compileMethod: aString
	" Compile aString as a method source. "

	| aMethod |
	aMethod := class compiler
		compile: aString
		forClass: class
		category: category.
	stamp size > 0 ifTrue: [ aMethod stamp: stamp ] ! !

! Behavior methodsFor: #Tools-fileIn !
methodsFor: aString
	" FileIn - Return a chunk reader to fileIn methods onto the receiver. "

	^ClassCategoryReader new class: self category: aString! !

! Behavior methodsFor: #Tools-fileIn !
methodsFor: aString stamp: stamp
	" FileIn - Return a chunk reader to fileIn methods onto the receiver. "

	^ClassCategoryReader new class: self category: aString stamp: stamp! !

!Behavior methodsFor: #Tools-fileIn!
methods
	" FileIn - Return a chunk reader to fileIn methods onto the receiver. "

	^self methodsFor: #unclassified ! !

FileInReader
 subclass: #ClassCommentReader
 instanceVariableNames: #class
 category: #Tools-fileIn!

! ClassCommentReader methodsFor: #accessing !
class: aClass
	" Set the destination class of the receiver. "

	class := aClass! !

! ClassCommentReader methodsFor: #fileIn !
scanFrom: aStream
	" FileIn - Parse next chunk in aStream as a comment for target class. "

	self comment: aStream nextChunk.
	^nil! !

! ClassCommentReader methodsFor: #private !
comment: aString
	" Set the comment of target class. "

	class comment: (aString isEmpty ifFalse: [ aString ])! !

! Behavior methodsFor: #Tools-fileIn !
commentStamp
	" Return a comment reader for the receiver. "

	^ClassCommentReader new class: self! !

FileInReader
 subclass: #SkipFileInReader
 instanceVariableNames: #limit
 category: #Tools-fileIn!

! SkipFileInReader class methodsFor: #instantiation !
upTo: aToken
	" Return an instance of the receiver. "

	^self new limit: aToken! !

! SkipFileInReader methodsFor: #fileIn !
scanFrom: aStream
	" Consume/Ignore contents in aStream. "

	limit isNil
		ifTrue: [ aStream setToEnd ]
		ifFalse: [ aStream skipTo: limit ].
	^nil! !

! SkipFileInReader methodsFor: #accessing !
limit: aToken
	" Set the token limit. "

	limit := aToken! !

! Object methodsFor: #Tools-fileIn !
cancelFileIn
	" FileIn - Cancel fileIn on current stream. "

	^self ignoreFileInUpTo: nil! !

! Object methodsFor: #Tools-fileIn !
ignoreFileInUpTo: aToken
	" FileIn - Ignore fileIn on current stream upto token. "

	^SkipFileInReader upTo: aToken! !

FileInReader
 subclass: #AssociationsReader
 category: #Tools-fileIn!

! AssociationsReader methodsFor: #reading !
keyFrom: aStream
	" Private - Return the name of the current binding in aStream. "

	aStream peek = $# ifTrue: [ ^(aStream next; upTo: $ ) asSymbol ].
	aStream peek = $' ifTrue: [ ^aStream next; upTo: $' ].
	^aStream upTo: $ ! !

! AssociationsReader methodsFor: #reading !
valueFrom: aStream
	" Private - Return the value of the current binding in aStream. "

	^self compilerSupport evaluate: aStream upToEnd ! !

! AssociationsReader methodsFor: #evaluating !
compilerSupport
	" Private - Return compiler in use to get values from stream. "

	^self smalltalk compilerSupport! !

! AssociationsReader methodsFor: #binding !
bind: tuples
	" Private - Bind the filedIn tuples to target.
	Must return the result of the fileIn/binding process.
	"

	^self subclassResponsibility: #bind: ! !

! AssociationsReader methodsFor: #fileIn !
scanFrom: aStream
	" Consume chunks in aStream until empty chunk.
	The chunks are considered associative bindings with the syntax:
		<key> <expression>
	Where
		key is the keyword of the binding (name, string or symbol).
		expression is a valid smalltalk expression that will be evaluated to generate the value of the binding.
	"

	| chunk tuples |
	tuples := Array new.
	[ aStream atEnd not and: [
	  chunk := aStream nextChunk trimBlanks.
	  chunk notEmpty ]] whileTrue: [
	  chunk := chunk stream.
	  tuples add: (Array
		with: (self keyFrom: chunk)
		with: (self valueFrom: chunk)).
	].
	^self bind: tuples! !

AssociationsReader
 subclass: #PoolDictionaryReader
 instanceVariableNames: #scope
 category: #Tools-fileIn!

! PoolDictionaryReader methodsFor: #accessing !
scope: aScope
	" Set the scope of the receiver. "

	scope := aScope! !

! PoolDictionaryReader methodsFor: #binding !
bind: tuples
	" Private - Bind the filedIn tuples to target.
	Must return the result of the fileIn/binding process.
	"

	scope isString ifTrue: [
		^Smalltalk definePool: scope with: tuples
	].
	scope isNil ifTrue: [ scope := PoolDictionary new ].
	tuples do: [:each| scope at: each first put: each last ].
	^scope! !

! PoolDictionaryReader methodsFor: #reading !
valueFrom: aStream
	" Private - Return the value of the current binding in aStream. "

	| code result word |
	code := aStream upToEnd.
	word := code trimBlanks.
	(word = #nil or: [ word isEmpty ]) ifTrue: [ ^nil ].
	result := word asNumber.
	result isNumber ifTrue: [ ^result ].
	^super valueFrom: code stream ! !

! Smalltalk class methodsFor: #Tools-variables !
bindingsFor: aScope
	" FileIn - Return a reader to fileIn bindings in a (global) scope.
	aScope can be a string or a dictionary.
	In case of aScope been a String, it will be considered a PoolDictionaryName
	See PoolDictionaryReader>>#scanFrom:
	"

	^PoolDictionaryReader new scope: aScope! !

AssociationsReader
	subclass: #NativeReader
	instanceVariableNames: #object
	category: #Tools-fileIn!

! NativeReader methodsFor: #binding !
bind: tuples
	" Private - Bind the filedIn tuples to target.
	Must return the result of the fileIn/binding process.
	"

	object isNil ifTrue: [ self initializeObject ].
	tuples do: [:each| object basicAt: each first put: each last ].
	^object! !

! NativeReader methodsFor: #accessing !
object: anObject
	" Set the object where the bindings will be stored. "

	object := anObject! !

! NativeReader methodsFor: #initialize !
initializeObject
	" Private - Initialize the object where the bindings will be stored. "

	object := {''}! !

EventManager
 subclass: #Exporter
 instanceVariableNames: #encoder
 category: #Tools-fileOut!

! Exporter class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^#(	#isBehavior:
		#needsClasses #needsRoots
		#needsEncoderFor:
	) ,super constructEventsTriggered! !

! Exporter methodsFor: #accessing !
encoder: anEncoder
	" Set the encoder of the receiver. "

	encoder := anEncoder! !

! Exporter methodsFor: #accessing !
setEncoderFor: aClass
	" Set the encoder of the receiver to emmit code for aClass. "

	| anEncoder |
	anEncoder := self triggerEvent: #needsEncoderFor: with: aClass ifNotHandled: [
		aClass compiler encoder
	].
	self encoder: anEncoder ! !

! Exporter methodsFor: #accessing !
encoderFor: aClass
	" Return the encoder of the receiver.
	If the encoder is nil, use the default encoder for aClass.
	"

	encoder isNil ifTrue: [ self setEncoderFor: aClass ].
	^encoder! !

! Exporter methodsFor: #accessing !
encoder
	" Return the encoder of the receiver. "

	encoder isNil ifTrue: [ self setEncoderFor: Object ].
	^encoder! !

! Exporter class methodsFor: #fileOut !
fileOutCategories: categories
	" Return the result of fileOut all the classes in categories. "

	^self new fileOutCategories: categories! !

! Exporter class methodsFor: #fileOut !
fileOutCategory: aString
	" Return the result of fileOut all the classes in category aString. "

	^self new fileOutCategory: aString! !

! Exporter class methodsFor: #fileOut !
fileOutAll: elements
	" Return the result of fileOut elements (classes, methods and/or categories). "

	^self new fileOutAll: elements! !

! Exporter class methodsFor: #fileOut !
fileOutAll
	" Return the result of fileOut of all classes in the system. "

	^self new fileOutAll ! !

! Exporter methodsFor: #fileOut !
fileOutAll
	" Return the result of fileOut of all classes in the system. "

	^self fileOutAll: self allClasses! !

! Exporter class methodsFor: #fileOut !
fileOut: spec
	" Return the result of fileOut of literals in spec. "

	^self new fileOutSpec: spec! !

! Exporter methodsFor: #testing !
isBehavior: anObject
	" Private - Return true if anObject is a Behavior. "

	^self triggerEvent: #isBehavior: with: anObject ifNotHandled: [
		(anObject isKindOf: Behavior) or: [ anObject isClass ]
	]! !

! Exporter methodsFor: #fileOut !
fileOutCollection: array
	" Return the result of fileOut of array contents. "

	| cr |
	cr := '
'.	^(array inject: #() into: [:total :each|
		total add: cr; add: (self fileOutSpec: each); yourself
		]) join: ''! !

! Exporter methodsFor: #fileOut !
fileOutSpec: spec
	" Return the result of fileOut of literals in spec. "

	| aName |
	spec isNil ifTrue: [ ^'' ].
	spec isString ifFalse: [
		(self isBehavior: spec) ifTrue: [ ^self fileOut: spec ].
		^self fileOutCollection: spec
	].
	(spec includes: $,) ifTrue: [ ^self fileOutSpec: (spec asArrayOfSubstringsSeparatedBy: $,) ].
	(spec includes: $ ) ifTrue: [ ^self fileOutSpec: (spec asArrayOfSubstringsSeparatedBy: $ ) ].
	aName := spec trimBlanks.
	(Smalltalk includesKey: aName) ifTrue: [ ^self fileOutSpec: (Smalltalk at: aName) ].
	^self fileOutCategories: aName! !

! Exporter class methodsFor: #export !
exportCategories: categories
	" Return the result of exporting all the classes in categories. "

	^self new exportCategories: categories! !

! Exporter class methodsFor: #export !
exportCategory: aString
	" Return the result of exporting all the classes in category aString. "

	^self new exportCategory: aString! !

! Exporter class methodsFor: #export !
export: spec
	" Return the result of exporting literals in spec. "

	^self new exportSpec: spec! !

! Exporter methodsFor: #export !
exportCollection: array
	" Return the result of exporting array contents. "

	| cr |
	cr := '
'.	^(array inject: #() into: [:total :each|
		total add: cr; add: (self exportSpec: each); yourself
		]) join: ''! !

! Exporter methodsFor: #export !
exportSpec: spec
	" Return the result of exporting literals in spec. "

	| aName |
	spec isNil ifTrue: [ ^'' ].
	spec isString ifFalse: [
		(self isBehavior: spec) ifTrue: [ ^self export: spec ].
		^self exportCollection: spec
	].
	(spec includes: $,) ifTrue: [ ^self exportSpec: (spec asArrayOfSubstringsSeparatedBy: $,) ].
	(spec includes: $ ) ifTrue: [ ^self exportSpec: (spec asArrayOfSubstringsSeparatedBy: $ ) ].
	aName := spec trimBlanks.
	(Smalltalk includesKey: aName) ifTrue: [ ^self export: (Smalltalk at: aName) ].
	^self exportCategories: aName! !

! Exporter class methodsFor: #export !
exportAll
	" Return the result of exporting all classes in the system. "

	^self new exportAll! !

! Exporter class methodsFor: #export !
exportAll: elements
	" Return the result of exporting all elements (classes, methods and/or categories). "

	^self new exportAll: elements! !

! Exporter class methodsFor: #accessing !
allClasses
	" OBSOLETE, Private - Return all the classes in the system (in fileIn order). "

	^self new allClasses! !

! Exporter methodsFor: #accessing !
rootClasses
	" Private - Return all the root classes in the system (in fileIn order). "

	^self triggerEvent: #needsRoots ifNotHandled: [
		Smalltalk rootClasses
	]! !

! Exporter methodsFor: #accessing !
allClasses
	" Private - Return all the classes in the system (in fileIn order). "

	^self triggerEvent: #needsClasses ifNotHandled: [
		self rootClasses inject: #() into: [:total :each|
			total addAll: each withAllSubclasses; yourself
		]
	]! !

! Exporter methodsFor: #categories !
matchCategory: aString
	" Return the result of filtering all the classes (and orphan methods) in category aString. "

	^self matchCategory: aString from: Smalltalk rootClasses! !

! Exporter methodsFor: #categories !
matchCategory: aString from: rootClasses
	" Return the result of filtering all the classes (and orphan methods) in category aString. "

	| selection selected rejected orphanMethods fileOutMethods |
	selected := Array new.
	rejected := Array new.
	orphanMethods := Array new.
	selection := (aString includes: $-)
		ifTrue: [ [:each| each category asString = aString ] ]
		ifFalse: [ [:each| (each category asString upTo: $-) = aString ] ].
	rootClasses do: [:root|
		root instanceClass withAllSubclasses do: [:each|
			((selection value: each)
				ifTrue: [ selected ] ifFalse: [ rejected ])
				add: each
		]
	].
	fileOutMethods := [:cls|
		cls methodsDo: [:aMethod |
			(selection value: aMethod) ifTrue: [ orphanMethods add: aMethod ]
		].
	].
	rejected do: [:each| fileOutMethods value: each; value: each class ].
	^Array with: selected with: rejected with: orphanMethods! !

! Exporter methodsFor: #fileOut !
fileOutCategories: categories
	" Return the result of fileOut all the classes in categories. "

	categories isString ifTrue: [ ^self fileOutCategories: categories asArrayOfSubstrings ].
	^(categories asArray inject: #() into: [:total :each|
		total add: (self fileOutCategory: each); yourself
	]) join: ''! !

! Exporter methodsFor: #fileOut !
fileOutCategory: aString
	" Return the result of fileOut all the classes in category aString. "

	| stream match |
	stream := '' writeStream.
	match := self matchCategory: aString.
	self fileOutAll: match first on: stream.
	self fileOutMethods: match last on: stream.
	^stream contents! !

! Exporter methodsFor: #fileOut !
fileOut: aClass
	" Return fileOut definition and code of a class. "

	| stream |
	stream := '' writeStream.
	self fileOut: aClass on: stream.
	^stream contents! !

! Exporter methodsFor: #fileOut !
fileOutClass: aClass
	" Return fileOut definition and code of a class. "

	^self fileOut: aClass! !

! Exporter methodsFor: #fileOut !
fileOutAll: elements
	" Return fileOut definition and code of elements (classes, methods and/or categories). "

	| stream |
	stream := '' writeStream.
	self fileOutAll: elements on: stream.
	^stream contents! !

! Exporter methodsFor: #fileOut !
fileOutMethod: aMethod of: aClass
	" Return fileOut method of a class. "

	| stream |
	stream := '' writeStream.
	self fileOutMethod: aMethod of: aClass on: stream.
	^stream contents! !

! Exporter methodsFor: #fileOut !
fileOutAllMethodsOf: aClass on: stream
	" FileOut all methods of aClass (and metaclass) onto stream. "

	self fileOutMethodsOf: aClass class on: stream.
	self fileOutMethodsOf: aClass on: stream.! !

! Exporter methodsFor: #fileOut !
fileOutAll: elements on: stream
	" FileOut definition and code of elements (classes, methods and/or categories) onto stream. "

	elements isString ifTrue: [ ^self fileOutAll: elements asArrayOfSubstrings on: stream ].
	elements do: [:each|
		(self isBehavior: each) ifTrue: [
			self fileOutDefinitionOfClass: each on: stream
		].
	].	stream cr.
	elements do: [:each|
		(self isBehavior: each) ifTrue: [
			self fileOutAllMethodsOf: each on: stream
		] ifFalse: [
			each isString
			ifTrue: [ stream nextPutAll: (self fileOutCategory: each);cr ]
			ifFalse: [ self fileOutMethod: each of: each classField on: stream ]
		]
	].! !

! Exporter methodsFor: #fileOut !
fileOutDefinitionOfClass: aClass on: stream
	" FileOut definition of aClass (and metaclass) onto stream. "

	self fileOutDefinitionOf: aClass on: stream.
	self fileOutDefinitionOf: aClass class on: stream! !

! Exporter methodsFor: #fileOut !
fileOutDefinitionOf: aClass
	" Return the definition of a class. "

	| stream |
	stream := '' writeStream.
	self fileOutDefinitionOfClass: aClass instanceClass on: stream.
	^stream contents! !

! Exporter methodsFor: #fileOut !
fileOut: aClass on: stream
	" FileOut definition and code of a class onto stream. "

	self fileOutAll: (Array with: aClass) on: stream.! !

! Exporter methodsFor: #fileOut !
fileOutDefinitionOf: aClass on: stream
	" FileOut definition of a class onto stream. "

	aClass isMetaclass ifTrue: [
		(self fileOutStrings: #instanceVariableNames of: aClass on: stream) ifTrue: [
			stream nextChunkPut: '';cr.
		].
		(self fileOutArray: #propertyNames of: aClass on: stream) ifTrue: [
			stream nextChunkPut: '';cr.
		].
		^self
	].

	stream
		nextPutAll: aClass superclass asString; cr;tab;
		nextPutAll: 'subclass: #'.
	(self fileOutStrings: #instanceVariableNames of: aClass on: stream) ifFalse: [
		stream
			nextPutAll: (self globalNameFor: aClass); cr;tab;
			nextPutAll: 'instanceVariableNames: ';
			nextPutAll: '' printString
	].
	aClass classVariables notNil ifTrue: [
	    stream	 cr;tab;
		nextPutAll: 'classVariableNames: ';
		nextPutAll: aClass classVariablesString printString
	].
	aClass poolDictionaries notNil ifTrue: [
	    stream	 cr;tab;
		nextPutAll: 'poolDictionaries: ';
		nextPutAll: aClass poolDictionariesString printString
	].
	stream cr;tab;
		nextPutAll: 'category: ';
		nextPutAll: aClass category asLiteral.
	stream nextChunkPut: '';cr.
	(self fileOutArray: #propertyNames of: aClass on: stream) ifTrue: [
		stream nextChunkPut: '';cr.
	].
	(aClass comment notNil and: [aClass comment notEmpty]) ifTrue: [
		stream nextPutAll: (self globalNameFor: aClass),' comment: ', aClass comment printString.
		stream nextChunkPut: '';cr.
	].! !

! Exporter methodsFor: #fileOut !
fileOutStrings: selector of: aClass on: stream
	" Private - FileOut selector strings of a class onto stream.
	Return true if code has been filedOut.
	"

	| strings |
	strings := aClass perform: selector.
	strings isEmpty ifTrue: [ ^false ].
	stream
		nextPutAll: (self globalNameFor: aClass); cr;tab;
		nextPutAll: selector,': '''.
	strings do: [:each| stream nextPutAll: ' ',each ].
	stream nextPutAll: ' '''.
	^true! !

! Exporter methodsFor: #fileOut !
fileOutArray: selector of: aClass on: stream
	" Private - FileOut selector arrayed strings of a class onto stream.
	Return true if code has been filedOut.
	"

	| strings |
	strings := aClass perform: selector.
	strings isEmpty ifTrue: [ ^false ].
	stream
		nextPutAll: (self globalNameFor: aClass);
		nextPutAll: ' ',selector,': ', strings asLiteral.
	^true! !

! Exporter methodsFor: #fileOut !
fileOutMethodsOf: aClass on: stream
	" FileOut methods of a class onto stream. "

	| selectors |
	selectors := aClass selectors.
	selectors isEmpty ifTrue: [ ^self ].
	stream cr.
	self
		fileOutSelectors: selectors
		of: aClass on: stream! !

! Exporter methodsFor: #private !
fileOutClusters
	" Private - Return the fileOut chustering dumpers of the receiver.
	A fileOutCluster is an object that potentially dump clustered actions for selectors implemented by aClass.
	The fileOut write (smalltalk)code for multiple selectors removing them from selectors collection.
	Note: the selectors collection is modified during fileOut.
	The default answer is an empty collection of custed or a collection with default APIBuilder.
	This method can be refined by the subclasses.
	"

	^(#( #APIBuilder )
		select: [:key| Smalltalk includesKey: key ])
		collect: [:each| (Smalltalk at: each) default ]! !

! Exporter methodsFor: #fileOut !
fileOutSelectors: selectors of: aClass on: stream
	" FileOut methods implementing selectors of a class onto stream. "

	self fileOutClusters do: [:each|
		each
			fileOut: self selectors: selectors
			of: aClass on: stream
	].
	self	fileOutMethods: (selectors collect: [:each|
			aClass compiledMethodAt: each ])
		on: stream! !

! Exporter methodsFor: #fileOut !
fileOutMethods: array on: stream
	" FileOut array of methods of a class onto stream. "

	| sorted |
	sorted := array sorted: [:a :b|
		a category < b category
		or: [ a category = b category
		and: [ a selector < b selector ]]
	].

	sorted do: [:each|
		self	fileOutMethod: each
			of: each classField on: stream
	].! !

! Exporter methodsFor: #private !
isASymbol: aString
	" Private - Return true if aString can be output with symbol notation. "

	^((RegularExpression fromString: '[^a-zA-Z0-9_:]') exec: aString) isNil! !

! Exporter methodsFor: #fileOut !
fileOutMethod: aMethod on: aStream
	" Return the fileOut of aMethod. "

	^self
		fileOutMethod: aMethod
		of: aMethod classField
		on: aStream! !

! Exporter methodsFor: #fileOut !
fileOutMethod: aMethod
	" Return the fileOut of aMethod. "

	| stream |
	stream := '' writeStream.
	self fileOutMethod: aMethod on: stream.
	^stream contents! !

! Exporter methodsFor: #fileOut !
fileOutMethod: aMethod of: aClass on: stream
	" FileOut method of a class onto stream. "

	| stamp |
	stamp := aMethod stamp size > 0
		ifTrue: [ ' stamp: ' , aMethod stamp asLiteral ]
		ifFalse: [''].
	stream	cr; nextChunkPut: nil;
		nextChunkPut: ' '
			,(self globalNameFor: aClass)
			,' methodsFor: ' ,aMethod category asLiteral
			,stamp ,' '; cr;
		nextChunkPut: aMethod source;
		nextChunkPut: ' '; cr! !

! Exporter methodsFor: #export !
exportCategory: aString
	" Return the result of exporting all the classes in category aString. "

	| stream match |
	stream := '' writeStream.
	match := self matchCategory: aString.
	match first do: [:each| stream nextPutAll: (self export: each) ].
	match last do: [:aMethod |
		self export: aMethod classField method: aMethod on: stream
	].
	^stream contents! !

! Exporter methodsFor: #export !
exportCategories: categories
	" Return the result of exporting all the classes in categories. "

	categories isString ifTrue: [ ^self exportCategories: categories asArrayOfSubstrings ].
	^(categories asArray inject: #() into: [:total :each|
		total add: (self exportCategory: each); yourself
	]) join: ''! !

! Exporter methodsFor: #export !
exportAllMethodsOf: aClass on: stream
	" Export all methods of a class (and metaclass) onto stream. "

	self exportMethodsOf: aClass class on: stream.
	self exportMethodsOf: aClass on: stream.! !

! Exporter methodsFor: #export !
exportAllMethodsOn: stream
	" Export all methods in the system onto stream. "

	self allClasses do: [:aClass|
		self exportAllMethodsOf: aClass on: stream
	]! !

! Exporter methodsFor: #export !
exportAllMethods
	" Return the result of exporting all methods in the system. "

	| stream |
	stream := '' writeStream.
	self exportAllMethodsOn: stream.
	^stream contents! !

! Exporter methodsFor: #export !
exportAllDefinitionsOn: stream
	" Export all class definitions in the system onto stream. "

	self allClasses do: [:aClass|
		self exportDefinitionOfClass: aClass on: stream
	]! !

! Exporter methodsFor: #export !
exportAllDefinitions
	" Return the result of exporting all class definitions in the system. "

	| stream |
	stream := '' writeStream.
	self exportAllDefinitionsOn: stream.
	^stream contents! !

! Exporter methodsFor: #export !
exportDefinitionOfClass: aClass on: stream
	" Export definition of aClass (and metaclass) onto stream. "

	self exportDefinitionOf: aClass on: stream.
	self exportMetaDefinitionOf: aClass on: stream.! !

! Exporter methodsFor: #export !
export: aClass
	" Return the result of exporting definition and source code of a class. "

	| stream |
	stream := '' writeStream.
	self exportDefinitionOfClass: aClass on: stream.
	self exportAllMethodsOf: aClass on: stream.
	^stream contents! !

! Exporter methodsFor: #export !
exportNew: aClass
	" Return the result of exporting definition of aClass. "

	^self exportDefinitionOf: aClass! !

! Exporter methodsFor: #export !
exportCommentOf: aClass
	" Return the result of exporting comment of a class. "

	| stream |
	stream := '' writeStream.
	self exportCommentOf: aClass on: stream.
	^stream contents! !

! Exporter methodsFor: #private !
exportCommentOf: aClass on: aStream
	" Export comment of aClass onto aStream. "

	^(self encoderFor: aClass) exportCommentOf: aClass on: aStream! !

! Exporter methodsFor: #private !
exportDefinitionOf: aClass on: aStream
	" Export definition of aClass onto aStream. "

	^(self encoderFor: aClass) exportDefinitionOf: aClass on: aStream! !

! Exporter methodsFor: #private !
exportMetaDefinitionOf: aClass on: aStream
	" Export class side definition of aClass onto aStream. "

	^(self encoderFor: aClass) exportMetaDefinitionOf: aClass on: aStream! !

! Exporter methodsFor: #private !
smalltalkNameFor: aClass
	" Private - Return the native reference expression of aClass (or nil). "

	^(self encoderFor: aClass) smalltalkNameFor: aClass! !

! Exporter methodsFor: #private !
globalNameFor: aClass
	" Private - Return the smalltalk name of aClass (or nil). "

	^aClass isNil ifTrue: [ #nil ] ifFalse: [ aClass name ]! !

! Exporter methodsFor: #export !
exportAll
	" Return the result of exporting all classes in the system. "

	^self exportClasses: self allClasses! !

! Exporter methodsFor: #export !
exportAllCategories
	" Return the result of exporting all categories in the system. "

	| categories |
	categories := Array new.
	self allClasses do: [:each |
		(categories includes: each category) ifFalse: [
			categories add: each category
		]
	].
	^(categories inject: #() into: [:total :each | total add: (self exportCategory: each); yourself ]) join: ''! !

! Exporter methodsFor: #export !
exportDefinitionOf: aClass
	" Return the result of exporting definition of aClass. "

	| stream |
	stream := String new writeStream.
	self exportDefinitionOf: aClass on: stream.
	^stream contents! !

! Exporter methodsFor: #export !
exportMetaDefinitionOf: aClass
	" Return the result of exporting class side definition of aClass. "

	| stream |
	stream := String new writeStream.
	self exportMetaDefinitionOf: aClass on: stream.
	^stream contents! !

! Exporter methodsFor: #export !
exportClasses: classes on: stream
	" Return the result of exporting class definitions onto stream. "

	^self exportAll: classes on: stream! !

! Exporter methodsFor: #export !
exportAll: elements on: stream
	" Return the result of exporting elements (classes, methods and/or categories) onto stream. "

	elements isString ifTrue: [ ^self exportAll: elements asArrayOfSubstrings on: stream ].
	elements do: [:each|
		(self isBehavior: each) ifTrue: [
			self exportDefinitionOfClass: each on: stream
		].
	].
	stream cr.
	elements do: [:each|
		(self isBehavior: each) ifTrue: [
			self exportAllMethodsOf: each on: stream
		] ifFalse: [
			each isString
			ifTrue: [ stream nextPutAll: (self exportCategory: each);cr ]
			ifFalse: [ self exportMethod: each on: stream ]
		]
	].! !

! Exporter methodsFor: #export !
exportAll: elements
	" Return the result of exporting elements (classes, methods and/or categories). "

	| stream |
	stream := String new writeStream.
	self exportAll: elements on: stream.
	^stream contents! !

! Exporter methodsFor: #export !
exportClasses: classes
	" Return the result of exporting classes. "

	| stream |
	stream := String new writeStream.
	self exportClasses: classes on: stream.
	^stream contents! !

! Exporter methodsFor: #export !
exportObject: anObject on: stream
	" Export a (native)literal representation of anObject onto stream. "

	stream nextPutAll: anObject asNative! !

! Exporter methodsFor: #export !
exportObject: anObject
	" Return the result of exporting a literal representation of anObject. "

	| stream |
	stream := String new writeStream.
	self exportObject: anObject on: stream.
	^stream contents! !

! Exporter methodsFor: #export !
export: aClass method: aMethod
	" Return the result of exporting aClass method. "

	| stream |
	stream := String new writeStream.
	self export: aClass method: aMethod on: stream.
	^stream contents! !

! Exporter methodsFor: #export !
exportMethod: aMethod
	" Return the result of exporting aMethod. "

	^self export: aMethod classField method: aMethod! !

! Exporter methodsFor: #export !
exportMethod: aMethod on: aStream
	" Return the result of exporting aMethod on stream. "

"	(aMethod respondsTo: #classField) ifFalse: [ ^self note: 'Warning: broken method ',aMethod ].
"	^self export: aMethod classField method: aMethod on: aStream! !

! Exporter methodsFor: #export !
exportMethods: aCollection on: aStream
	" Return the result of exporting aCollection of methods on stream. "

	aCollection do: [:each| self exportMethod: each on: aStream ]! !

! Exporter methodsFor: #Tools-export !
exportMethods: aCollection
	" Return the result of exporting aCollection of methods. "

	| stream |
	stream := String new writeStream.
	self exportMethods: aCollection on: stream.
	^stream contents! !

! Exporter methodsFor: #Tools-private !
exportMethodsOf: aClass on: aStream
	" Export methods of aClass onto aStream. "

	aClass methodsDo: [:each|
		self export: aClass method: each on: aStream
	]! !

! Exporter methodsFor: #Tools-export !
export: aClass method: aMethod on: aStream
	" Export a method definition of aClass onto aStream. "

	(aMethod respondsTo: #exportFor:on:) ifTrue: [ ^aMethod exportFor: self on: aStream ].
	^(self encoderFor: aClass) export: aClass method: aMethod on: aStream! !

! String methodsFor: #Tools-evaluating !
showIt
	" Show/Print the result of evaluating the receiver. "

	^self print: self logEvaluate! !

! String methodsFor: #Tools-evaluating !
logEvaluate
	" Return the result of evaluating the receiver.
	The source of evaluation will be logged.
	"

	^[ self evaluate ] logEvaluate: self ! !

! String methodsFor: #Tools-evaluating !
evaluate
	" Return the result of evaluating the receiver. "

	^self smalltalk compilerSupport evaluate: self ! !

! String methodsFor: #Tools-evaluating !
evaluateIn: aReceiver
	" Return the result of evaluating the receiver in the context of aReceiver. "

	| target |
	target := ((self isObject: aReceiver) and: [aReceiver notNil]) ifTrue: [ aReceiver ].
	^target smalltalk compilerSupport evaluate: self to: target! !

! String methodsFor: #Tools-evaluating !
doIt
	" Return the result of evaluating the receiver or the error occurred during compilation or evaluation. "

	^self smalltalk compilerSupport doIt: self! !

! Smalltalk methodsFor: #Tools-fileIn !
fileInCode: code
	" FileIn the code into the receiver. "

	^code notNil ifTrue: [ code stream fileIn ]! !

! String methodsFor: #Tools-fileIn !
fileIn
	" FileIn the contents of the receiver. "

	^self stream fileIn! !

! String methodsFor: #Tools-fileOut !
fileOut
	" FileOut the classes/categories in the receiver. "

	^Exporter fileOut: self! !

! String methodsFor: #Tools-fileOut !
export
	" Export the classes/categories in the receiver. "

	^Exporter export: self! !

! String methodsFor: #Tools-converting !
asComment
	" Return the receiver as a source code comment. "

	| prologue epilogue |
	((self includes: String cr) or: [self includes: String lf]) ifTrue: [
		prologue := {''}.
		epilogue := {''}.
	] ifFalse: [
		prologue := {''}.
		epilogue := ''
	].
	^prologue ,self ,epilogue! !

! Behavior methodsFor: #Tools-fileOut !
fileOut
	" Return fileOut definition and code of a class. "

	^Exporter fileOut: self instanceClass! !

! Behavior methodsFor: #Tools-export !
export
	" Return the result of exporting definition and source code of a class. "

	^Exporter export: self instanceClass! !

! Collection methodsFor: #Tools-fileOut !
fileOut
	" Return fileOut definition and code of classes in the receiver. "

	^Exporter fileOut: self! !

! Collection methodsFor: #Tools-export !
export
	" Return the result of exporting definition and source code of classes in the receiver. "

	^Exporter export: self! !

! S8Encoder class methodsFor: #Tools-export !
export: spec
	" Return the native code result of encoding spec. "

	^self new export: spec! !

! S8Encoder methodsFor: #Tools-export !
export: spec
	" Return the native code result of encoding spec. "

	| exporter |
	exporter := Exporter new.
	exporter encoder: self.
	spec = #all ifTrue: [ ^exporter exportAll ].
"OBSOLETE-
	spec isArray ifTrue: [ ^exporter exportAll: spec ].
	^exporter exportAll: (Array with: spec)
"	^exporter export: spec! !

EventManager
 subclass: #MethodReferences
 instanceVariableNames: 'rootClasses results'
 category: #Tools-references!

! Smalltalk class methodsFor: #Tools-query !
sendersOf: selector
	" Return the senders of message selector. "

	^self current sendersOf: selector! !

! Smalltalk methodsFor: #Tools-accessing !
methodReferences
	" Return the methodReferences tool of the receiver. "

	^MethodReferences new! !

! Smalltalk methodsFor: #Tools-query !
sendersOf: selector
	" Return the senders of message selector. "

	^self methodReferences sendersOf: selector! !

! String methodsFor: #Tools-query !
senders
	" Return the senders of the receiver. "

	^Smalltalk sendersOf: self! !

! String methodsFor: #Tools-query !
implementors
	" Return the implementors of the receiver. "

	^Smalltalk implementorsOf: self! !

! MethodReferences class methodsFor: #Tools-events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver. "

	^#(	#found: #found:pattern:in:
		#needsMethods
		#searchStarted #searchFinished
	) , super constructEventsTriggered! !

! MethodReferences class methodsFor: #Tools-testing !
of: selector
	" Return the references to selector. "

	^self new referencesOf: selector! !

! MethodReferences class methodsFor: #Tools-testing !
ofGlobal: globalName
	" Return the references to global globalName. "

	^self new referencesOfGlobal: globalName! !

! MethodReferences class methodsFor: #Tools-testing !
findClasses: what
	" Find classes in the system matching what. "

	^self new findClasses: what! !

! MethodReferences class methodsFor: #Tools-testing !
globals
	" Return the references to all global names. "

	| tool result |
	tool := self new.
	result := PoolDictionary new.
	tool when: #found:pattern:in: do: [:mth :mask :stream| | global |
		global := SmalltalkParser new globalNameParser parse: stream.
		global isParseFailure ifFalse: [
			(result at: global ifAbsentPut: [Array new]) add: mth
		].
	].
	tool searchGlobalReferences.
	^result! !

! MethodReferences class methodsFor: #Tools-testing !
methodOverwrites
	" Find the methods that are implemented/overwritten by subclasses. "

	^self new methodOverwrites! !

! MethodReferences class methodsFor: #Tools-testing !
variableOverwrites
	" Find the methods that has potential overwrite of variable names. "

	^self new variableOverwrites! !


! MethodReferences methodsFor: #Tools-initialize !
initializeResults
	" Private - Initialize the results of the receiver. "

	results := Array new! !

! MethodReferences methodsFor: #Tools-initialize !
reset
	" Reset the receiver. "

	self results: nil! !

! MethodReferences methodsFor: #Tools-accessing !
rootClasses: array
	" Set the rootClasses of the receiver. "

	rootClasses := array! !

! MethodReferences methodsFor: #accessing !
results
	" Return the results of the receiver. "

	results isNil ifTrue: [ self initializeResults ].
	^results! !

! MethodReferences methodsFor: #Tools-accessing !
results: array
	" Set the results of the receiver. "

	results := array! !

! MethodReferences methodsFor: #Tools-accessing !
rootClasses
	" Return the rootClasses of the receiver. "

	rootClasses isNil ifTrue: [ ^Smalltalk rootClasses ].
	^rootClasses! !

! MethodReferences methodsFor: #Tools-accessing !
allMethodsIn: aCollection
	" Return the methods defined by aCollection of classes. "

	| result |
	result := Array new.
	self rootClasses do: [:root|
		root instanceClass withAllSubclasses do: [:cls|
			cls methodsDo: [:each| result add: each ].
			cls class methodsDo: [:each| result add: each ].
		]
	].
	^result! !

! MethodReferences methodsFor: #Tools-accessing !
allMethods
	" Return the methods to search the receiver. "

	^self	triggerEvent: #needsMethods
		ifNotHandled: [ self allMethodsIn: self rootClasses ]! !

! MethodReferences methodsFor: #Tools-searching !
search: what with: targetBlock
	" Find what on target of selected methods. "

	| atoms |
	atoms := self allMethods.
	self triggerEvent: #searchStarted.
	atoms := atoms select: [:aMethod| | source |
		source := targetBlock value: aMethod.
		(source notNil and: [source toString includes: what]) ifTrue: [
			self results add: aMethod.
			self triggerEvent: #found: with: aMethod.
			(self hasActionForEvent: #found:pattern:in:) ifTrue: [ | stream args |
				stream := source toString stream.
				args := Array with: aMethod with: what with: stream.
				[ stream upTo: what; atEnd ] whileFalse: [
					self triggerEvent: #found:pattern:in: withArguments: args
				]
			].
			true
		] ifFalse: [ false ]
	].
	self triggerEvent: #searchFinished.
	^atoms! !

! MethodReferences methodsFor: #Tools-searching !
search: what
	" Find references to what in selected methods.
	Note: search is realized on native source of native code of methods.
	"

	^self search: what with: [:aMethod| aMethod compiledSource ]! !

! MethodReferences methodsFor: #Tools-searching !
searchSource: what
	" Find what in source of selected methods. "

	^self search: what with: [:aMethod| aMethod source ]! !

! MethodReferences methodsFor: #Tools-searching !
inlines
	" Find native inlines in selected methods. "

	^self searchSource: ${ ,$' ! !

! MethodReferences methodsFor: #Tools-searching !
searchCategory: aName
	" Find (source) references to category starting from root classes. "

	| found |
	self triggerEvent: #searchStarted.
	found := Exporter new matchCategory: aName from: self rootClasses.
	found := found first , found last.
	found do: [:each|
			self results add: each.
			self triggerEvent: #found: with: each.
	].
	self triggerEvent: #searchFinished.
	^found! !

! MethodReferences methodsFor: #Tools-patterns !
sendPatternsOf: aSelector
	" Private - Return the search patterns for references of aSelector. "

	^Array
		with: (self normalSendPatternOf: aSelector)
		with: (self superSendPatternOf: aSelector)! !

! MethodReferences methodsFor: #Tools-patterns !
dot
	" Private - Return the message send pattern starter. "

	^{''}! !

! MethodReferences methodsFor: #Tools-patterns !
localDef
	" Private - Return the variable definition starter. "

	^{''}! !

! MethodReferences methodsFor: #Tools-patterns !
normalSendStarter: aSelector
	" Private - Return the pattern for detecting normal senders of aSelector. "

	{'self'}.
	^self dot! !

! MethodReferences methodsFor: #Tools-patterns !
normalSendPatternOf: aSelector
	" Private - Return the search pattern for normal senders of aSelector. "

	^(self normalSendStarter: aSelector)
		,aSelector asSelector
		,$( ! !

! MethodReferences methodsFor: #Tools-patterns !
superJSSendPatternOf: aSelector
	" Private - Return the search pattern for senders (to super) of aSelector. "

	^$" ,aSelector asSelector ,$" ,').apply(self,'! !

! MethodReferences methodsFor: #Tools-patterns !
superSendPatternOf: aSelector
	" Private - Return the search pattern for senders (to super) of aSelector. "

	^{''}! !

! MethodReferences methodsFor: #Tools-patterns !
instVarPatternOf: instVarName
	" Private - Return the search pattern for instance variable access. "

	^'self["@' ,instVarName asNativeName ,'"]'! !

! MethodReferences methodsFor: #Tools-patterns !
globalPatternOf: globalName
	" Private - Return the search pattern for global variable access. "

	^'smalltalk.' ,globalName! !

! MethodReferences methodsFor: #Tools-types !
types
	" Private - Return the types of search implemented by the receiver. "

	^#(	#referencesOf: #referencesOfGlobal:
		#referencesOfClassVariable: 
		#referencesOfInstanceVariable:
		#findClasses: #findMessages:
		#searchCategory:
		#search: #searchSource:
		#globalReferences #missingGlobals #inlines
		#methodOverwrites #variableOverwrites
	)! !

! MethodReferences methodsFor: #Tools-searching !
search: aName type: typeSelector
	" Return the result of searching for aName as a search of specified type (or nil). "

	| selector |
	selector := (typeSelector includes: $:)
	ifTrue: [ (typeSelector upTo: $:) ,$: ]
	ifFalse: [ typeSelector upTo: $  ].
	^(self respondsTo: selector) ifTrue: [
		self perform: selector with: aName trimBlanks
	]! !

! MethodReferences methodsFor: #Tools-searching !
sendersOf: aSelector
	" Find the requested elements from root classes. "

	^(self sendPatternsOf: aSelector) inject: #() into: [:total :each|
		total addAll: (self search: each); yourself
	]! !

! MethodReferences methodsFor: #Tools-searching !
referencesOf: aSelector
	" Find the references to message (aSelector). "

	self results
		add: '--- Implementors of #' ,aSelector ,' ---';
		addAll: (self implementorsOf: aSelector);
		add: '--- Senders of #' ,aSelector ,' ---'.
	self sendersOf: aSelector.
	^results! !

! MethodReferences methodsFor: #Tools-searching !
referencesOfInstanceVariable: instVarName
	" Find the requested elements from root classes. "

	^self search: (self instVarPatternOf: instVarName)! !

! MethodReferences methodsFor: #Tools-searching !
settersOfInstanceVariable: instVarName
	" Find the requested elements from root classes. "

	^self search: (self instVarPatternOf: instVarName) ,'='! !

! MethodReferences methodsFor: #Tools-searching !
referencesOfGlobal: globalName
	" Find the requested elements from root classes. "

	^self search: (self globalPatternOf: globalName)! !

! MethodReferences methodsFor: #Tools-searching !
searchGlobalReferences
	" Find the requested elements from root classes. "

	^self search: (self globalPatternOf: '')! !

! MethodReferences methodsFor: #Tools-searching !
settersOfGlobalVariable: globalName
	" Find the requested elements from root classes. "

	^self search: (self globalPatternOf: globalName) ,'='! !

! MethodReferences methodsFor: #Tools-searching !
referencesOfClassVariable: classVarName
	" Find the requested elements from root classes. "

	^self results
		add: '--- Getters of ' ,classVarName ,' ---';
		addAll: (self gettersOfClassVariable: classVarName);
		add: '--- Setters of ' ,classVarName ,' ---';
		addAll: (self settersOfClassVariable: classVarName);
		yourself! !

! MethodReferences methodsFor: #Tools-searching !
gettersOfClassVariable: classVarName
	" Find the requested elements from root classes. "

	^self search: self dot,'classVariableAt_("',classVarName,'")'! !

! MethodReferences methodsFor: #Tools-searching !
settersOfClassVariable: classVarName
	" Find the requested elements from root classes. "

	^self search: self dot,'classVariableAt_put_("',classVarName,'",'! !

! MethodReferences methodsFor: #Tools-searching !
globalReferences
	" Find the references to globals. "

	| pool |
	pool := self class globals.
	pool keys sorted do: [:aName|
		self results
			add: '--- References to: ',aName,' ---';
			addAll: (pool at: aName)
	].
	^results! !

! MethodReferences methodsFor: #Tools-searching !
missingGlobals
	" Find the references to missing globals. "

	| pool |
	pool := self class globals.
	(pool keys sorted
		reject: [:key| (Smalltalk includesKey: key) ])
		do: [:aName|
			self results
				add: '--- Missing global: ',aName,' ---';
				addAll: (pool at: aName)
		].
	self results isEmpty ifTrue: [
		self results add: '--- All global references are bound to Smalltalk ---'
	].
	^results! !

! MethodReferences methodsFor: #Tools-private !
implementorsOf: aSelector
	" Private - Return the implementors of aSelector. "

	^Smalltalk current implementorsOf: aSelector! !

! MethodReferences methodsFor: #Tools-patterns !
compiledSourceStart
	" Private - Return the token for start of compiled code. "

	^{''}! !

! MethodReferences methodsFor: #Tools-patterns !
messagesStreamOn: aCompiledMethod
	" Private - Return the stream for reading messages on compiled code of aCompiledMethod. "

	| stream nativeMessage |
	nativeMessage := self nativeLanguage ,#MessagesStreamOn:.
	(self respondsTo: nativeMessage) ifTrue: [
		^self perform: nativeMessage with: aCompiledMethod
	].
	stream := aCompiledMethod compiledSource stream.
	stream upTo: self compiledSourceStart.
	^stream ! !

! MethodReferences methodsFor: #Tools-messages !
nextMessageIn: stream
	" Private - Return the next sent encoded in native source stream. "

	| dot |
	dot := self dot.
	^((stream upTo: dot; upTo: $(
	) asArrayOfSubstringsSeparatedBy: dot)
	last asNativeSelector! !

! MethodReferences methodsFor: #Tools-accessing !
messagesOf: aCompiledMethod
	" Return all the messages that can be sent during execution of aCompiledMethod. "

	| stream selectors getMessage |
	stream := self messagesStreamOn: aCompiledMethod.
	stream atEnd ifTrue: [ ^#() ].
	getMessage := self nativeLanguage ,#NextMessageIn:.
	(self respondsTo: getMessage) ifFalse: [
		getMessage := #nextMessageIn:
	].
	selectors := HashedSet new.
	[ stream atEnd ] whileFalse: [
		selectors add: (self perform: getMessage with: stream)
	].
	selectors remove: '' ifAbsent: [].
	^(selectors asArray select: [:each| each isLiteral ]) sorted! !

! MethodReferences methodsFor: #Tools-accessing !
globalReferencesOf: aCompiledMethod
	" Return all the global reference names in aCompiledMethod. "

	| result |
	result := HashedSet new.
	self
		when: #needsMethods do: [ Array with: aCompiledMethod ];
		when: #found:pattern:in: do: [:mth :mask :stream| | global |
			global := SmalltalkParser new globalNameParser parse: stream.
			global isParseFailure ifFalse: [ result add: global ].
		];
		yourself.
	self searchGlobalReferences.
	^result asArray sorted! !

! MethodReferences methodsFor: #Tools-searching !
findClasses: what
	" Find classes matching what. "

	| keys system |
	what isString ifTrue: [
		^self findClasses: what trimBlanks asArrayOfSubstrings
	].
	self triggerEvent: #searchStarted.
	keys := what collect: [:each| each asString asLowercase ].
	system := Smalltalk current.
	keys notEmpty ifTrue: [
		system classes do: [:aClass| | key |
			key := aClass name asLowercase.
			(keys detect: [:one| key includes: one ] ifNone: []) notNil ifTrue: [
				self results add: aClass.
				self triggerEvent: #found: with: aClass.
			]
		].
	].
	self triggerEvent: #searchFinished.
	^self results! !

! MethodReferences methodsFor: #Tools-private !
methodOverwritesOf: aClass
	" Private - Find the methods subimplemented by subclasses of aClass. "

	| map subclasses |
	subclasses := aClass allSubclasses.
	map := PoolDictionary new.
	subclasses isEmpty ifTrue: [ ^map ].
	aClass methodsDo: [:mth| | selector |
		selector := mth selector.
		map at: selector put: (Array with: mth).
		subclasses do: [:each|
			(each selectors includes: selector) ifTrue: [
				(map at: selector) add: each >> selector
			].
		].
	].
	subclasses do: [:each|
		(self methodOverwritesOf: each) keysAndValuesDo: [:key :value|
			(map at: key ifAbsentPut: [Array new]) addAll: value
		].
	].
	(map keys asArray
		select: [:key| (map at: key) size <= 1 ]
		) do: [:key| map removeKey: key ].
	^map! !

! MethodReferences methodsFor: #Tools-private !
addMethodOverwrites: map
	" Private - Register the methods overwrites in map. "

	map associationsDo: [:assoc|
		assoc value do: [:each|
			self triggerEvent: #found: with: each.
			self results add: each.
		]
	]! !

! MethodReferences methodsFor: #Tools-searching !
methodOverwrites
	" Find the methods that are implemented/overwritten by subclasses. "

	self triggerEvent: #searchStarted.
	Smalltalk rootClasses do: [:aClass|
		self addMethodOverwrites: (self methodOverwritesOf: aClass).
		self addMethodOverwrites: (self methodOverwritesOf: aClass class).
	].
	self triggerEvent: #searchFinished.
	^self results! !

! MethodReferences methodsFor: #Tools-private !
hasArgumentOverwrites: names in: stream
	" Private - Return true if stream of code has argument name conflicts. "

	| args |
	args := (stream upTo: $(; upTo: $)) asArrayOfSubstringsSeparatedBy: $,.
	args do: [:each|
		(names includes: each) ifTrue: [ ^true ].
	].
	^false! !

! MethodReferences methodsFor: #Tools-private !
hasLocalOverwrites: names in: stream
	" Private - Return true if stream of code has local name conflicts. "

	| candidate |
	[stream atEnd] whileFalse: [
		candidate := (stream upTo: self localDef; upTo: $=) trimBlanks.
		(names includes: candidate) ifTrue: [
			^true
		].
	].
	^false! !

! MethodReferences methodsFor: #Tools-private !
addVariableOverwrites: names in: mth
	" Private - Register the methods that has name overwrites. "

	| stream |
	stream := mth compiledSource stream.
	((self hasArgumentOverwrites: names in: stream)
	or: [ self hasLocalOverwrites: names in: stream ])
	ifTrue: [
		self triggerEvent: #found: with: mth.
		self results add: mth.
	]! !

! MethodReferences methodsFor: #Tools-searching !
variableOverwrites
	" Find the methods that has potential overwrite of variable names. "

	self triggerEvent: #searchStarted.
	Smalltalk classes do: [:aClass| | names |
		names := aClass allInstVarNames.
		aClass methodsDo: [:mth|
			self addVariableOverwrites: names in: mth.
		].
		names := aClass class allInstVarNames.
		aClass class methodsDo: [:mth|
			self addVariableOverwrites: names in: mth.
		].
	].
	self triggerEvent: #searchFinished.
	^self results! !

! MethodReferences class methodsFor: #Tools-testing !
findMessages: what
	" Find messages in the system matching what. "

	^self new findMessages: what! !

! MethodReferences methodsFor: #Tools-searching !
findMessages: what
	" Find messages matching what. "

	| keys |
	what isString ifTrue: [
		^self findMessages: what trimBlanks asArrayOfSubstrings
	].
	self triggerEvent: #searchStarted.
	keys := what collect: [:each| each asString ].
	keys notEmpty ifTrue: [ | all collect |
		all := self allMethods.
		all do: [:mth| | current |
			current := mth selector.
			(keys detect: [:one| current includes: one ] ifNone: []) notNil ifTrue: [
				self results add: mth.
				self triggerEvent: #found: with: mth.
			].
		].
	].
	self triggerEvent: #searchFinished.
	^self results! !

! CompiledMethod methodsFor: #Tools-accessing !
messages
	" Return all the messages that can be sent during execution of the receiver. "

	^self smalltalk methodReferences messagesOf: self! !

! CompiledMethod methodsFor: #Tools-accessing !
globalReferences
	" Return all the global references in the receiver. "

	^self smalltalk methodReferences globalReferencesOf: self! !

! CompiledMethod methodsFor: #Tools-accessing !
references
	" Return all the references in the receiver. "

	^PoolDictionary new
		at: #messages put: self messages;
		at: #globalReferences put: self globalReferences;
		yourself! !

! Behavior methodsFor: #Tools-references !
references
	" Return all the references in the receiver. "

	| result |
	result := PoolDictionary new.
	self methodsDo: [:mth| | refs |
		refs := mth references.
		refs keys do: [:key|
			(result at: key ifAbsentPut: [HashedSet new])
				addAll: (refs at: key)
		]
	].
	^result collect: [:each| each asArray ]! !

! Behavior methodsFor: #Tools-messages !
messagesSent
	" Return the messages sent by the receiver. "

	| result cls |
	result := HashedSet new.
	cls := self instanceClass.
	cls methodsDo: [:each| result addAll: each messages ].
	cls class methodsDo: [:each| result addAll: each messages ].
	^result! !

! Behavior methodsFor: #Tools-messages !
allMessagesSent
	" Return the messages sent by the receiver and its subclasses. "

	^self allSubclasses
		inject: self messagesSent
		into: [:total :each| total addAll: each messagesSent; yourself ]! !

! Behavior methodsFor: #Tools-messages !
implementedMessages
	" Return the messages implemented by the receiver. "

	| result cls |
	result := HashedSet new.
	cls := self instanceClass.
	result addAll: cls selectors.
	result addAll: cls class selectors.
	^result! !

! Behavior methodsFor: #Tools-messages !
allImplementedMessages
	" Return the messages implemented by the receiver and its subclasses. "

	^self allSubclasses
		inject: self implementedMessages
		into: [:total :each| total addAll: each implementedMessages; yourself ]! !

! Smalltalk methodsFor: #Tools-iterating !
implementorsDo: aBlock
	" Evaluates aBlock for each implementor in the receiver. "

	^self classes do: [:each| each implementorsDo: aBlock ]! !

! Smalltalk methodsFor: #Tools-implementors !
implementors
	" Return a collection with all implementors in the receiver. "

	| result |
	result := #().
	self implementorsDo: [:each| result add: each].
	^result! !

! Behavior methodsFor: #Tools-iterating !
implementorsDo: aBlock
	" Evaluates aBlock for each implementor in the receiver. "

	| cls |
	cls := self instanceClass.
	cls class methodsDo: aBlock.
	cls methodsDo: aBlock.! !

! Behavior methodsFor: #Tools-implementors !
implementors
	" Return a collection with all implementors in the receiver. "

	| result |
	result := #().
	self implementorsDo: [:each| result add: each].
	^result! !

! UndefinedObject methodsFor: #Tools-printing !
compiler
	" Return the compiler of the receiver. "

	^self smalltalk compilerSupport for: nil! !

! Object methodsFor: #Tools-comparing !
== anObject
	^anObject == self! !

! Object methodsFor: #Tools-comparing !
~~ anObject
	^(self == anObject) == false! !

! UndefinedObject methodsFor: #Tools-comparing !
== anObject
	^anObject isNil! !

! UndefinedObject methodsFor: #Tools-comparing !
~~ anObject
	^anObject notNil! !

! Class methodsFor:#Tools-hacking !
hackDNU
	" Catch DNU condition in the receiver and its subclasses.
	Note: hack will persist until #unhackDNU is sent to the receiver.
	"

	| all |
	all := HashedSet new.
	Smalltalk rootClasses do: [:root|
		root allSubclasses do: [:each|
			all addAll: each selectors; addAll: each class selectors
		].
	].
	Smalltalk rootClasses do: [:root| root selectors do: [:each| all remove: each ifAbsent: []] ].
	#( #handleOrNil #start:end: ) do: [:each| all remove: each ifAbsent: []].
	all do: [:selector| | mth |
		mth := CompiledMethod new.
		mth selector: selector; classField: self; category: #hackDNU; source: selector,'(dnu)'.
		{''}.
		self basicAddCompiledMethod: mth.
	].! !

! Class methodsFor: #Tools-hacking !
unhackDNU
	" Remove Catch DNU hacking on the receiver. "

	self removeSelectors: (self selectors
		select: [:each| (Object >> each) category = #hackDNU ])! !

! Class methodsFor: #Tools-hacking !
hackDNUWhile: aBlock
	" Catch DNU hacking on the receiver while evaluating aBlock. "

	| result |
	[	self hackDNU.
		result := aBlock evaluate
	] ensure: [ self unhackDNU ].
	^result! !

! BlockClosure methodsFor: #Tools-hacking !
valueWithDNU
	" Catch DNU hacking while evaluating the receiver. "

	^Object hackDNUWhile: self! !

! String methodsFor: #Tools-debug !
ms: aBlock
	" Print the time spent evaluating aBlock titled with the receiver. "

	self note: self ,' (' ,aBlock ms ,'ms)'! !
