﻿
! Object methodsFor: #parserBug !
annotations: what
	" Private - Report buggy send by compiler. "

	self print: '[bug] ',self printString,' annotations: ',what asString.! !


Object
	subclass: #CompilerNode
	instanceVariableNames: ' nodes start end annotations '
	category: #Compiler !
CompilerNode buildAccessors: #( annotations ).!

! ValueNode class methodsFor: #instantiation !
with: aValue
	" Returns an instance of the receiver. "

	^self new value: aValue! !


! SequenceNode methodsFor: #validation !
validateArguments: args
	" OBSOLETE - Ensure the receiver can be emmited with arguments. "

	self temps do: [:temp|
		(args includes: temp) ifTrue: [
			^self error: 'Invalid temp (',temp,')'
		].
	]! !

! SequenceNode methodsFor: #accessing !
tempNames
	" Return the names of the temps in the receiver. "

	^self temps collect: [:each|
		each isString
		ifTrue: [each]
		ifFalse: [each value]
	]! !

! BlockNode methodsFor: #accessing !
parameterNames
	" Return the names of the parameters of the receiver. "

	^self parameters collect: [:each|
		each isString
		ifTrue: [each]
		ifFalse: [each value]
	]! !

! MethodNode methodsFor: #accessing !
argumentNames
	" Return the names of the arguments of the receiver. "

	^self arguments collect: [:each|
		each isString
		ifTrue: [each]
		ifFalse: [each value]
	]! !

! JSEncoder methodsFor: #generation !
emmitFunction: fName node: aNode arguments: argumentNames doing: aBlock
	" Private - Generate function for aNode. "

	| args |
	args := '' writeStream.
	argumentNames do: [:each | 
		self tempVariables add: each.
		args nextPutAll: (self localName: each) ]
	    separatedBy: [ args nextPutAll: ',' ].
	self	emmitFunction: fName
		arguments: args contents
		doing: aBlock! !

! JSEncoder methodsFor: #visiting !
visitMethodNode: aMethodNode
	" Visit a method node. "

	nestedBlocks := 0.
	earlyReturn := false.
	self compiler resetVariables.
	stream
		nextPutAll: self newMethodFunctionName, '(';
		nextPutAll: (self smalltalkNameFor: self currentClass);
		nextPutAll: ',"', aMethodNode selector, '",'; cr.
	self emmitMethod: aMethodNode.
	self includeSourceCode ifTrue: [
		stream	cr; "warning: this cr is important for debugger"
			nextPutAll: ',' ,aMethodNode source escapedCode
	].
	stream nextPutAll: ')'! !

! JSEncoder methodsFor: #generation !
emmitMethod: aMethodNode
	" Private - Generate code of aMethodNode. "

	| code |
	functionName := aMethodNode selector asSelector.
	aMethodNode validateArguments.
	self emmitFunction: functionName node: aMethodNode doing: [
		self compilerEvent: #startMethod: with: aMethodNode.
		code := self generatedCodeFor: aMethodNode while: [
			self compilerEvent: #enterMethod: withArguments: (Array with: aMethodNode)
				ifNotHandled: [ self emmitEnterMethod: aMethodNode ].
			aMethodNode nodes do: [:each | self visit: each ].
		].
		self generate: functionName early: earlyReturn return: code.
		self compilerEvent: #endMethod: with: aMethodNode.
	]! !

! JSEncoder methodsFor: #visiting !
visitSendArguments: aSendNode

	aSendNode arguments 
	    do: [:each | self visitArgument: each ]
	    separatedBy: [ stream nextPutAll: ', ' ].! !

! OptimizingEncoder methodsFor: #optimizing !
visitMacroIn: input with: aNode
	" Private - Dump the value of next variable defined in input. "

	| varName char args |
	varName := String new.
	[ char := input next. char = $% ] whileFalse: [
		varName := varName ,char asString
	].

	varName = #receiver ifTrue: [ ^self visitReceiver: aNode receiver ].
	(varName startsWith: #arg) ifTrue: [
		args := aNode arguments.
		1 to: args size do: [:i|
			#arg ,i asString = varName ifTrue: [
				^self visit: (args at: i)
			].
		].
	].
	varName = #inlinedReceiver ifTrue: [ ^self visitInlined: aNode receiver ].
	(varName startsWith: #inlined) ifTrue: [
		args := aNode arguments.
		1 to: args size do: [:i|
			#inlined ,i asString = varName ifTrue: [
				^self visitInlined: (args at: i)
			].
		].
	].
	^self error: 'Invalid macro variable[',varName,'].'! !

! JSEncoder methodsFor: #generation !
emmitFunction: fName node: aNode doing: aBlock
	" Private - Generate function for aNode. "

	^self	emmitFunction: fName
		node: aNode arguments: aNode argumentNames
		doing: aBlock! !

! JSEncoder methodsFor: #visiting !
visitBlockNode: aNode
	stream nextPutAll: $(.
	self	emmitFunction: '' node: aNode
		arguments: aNode parameterNames
		doing: [ aNode nodes do: [:each | self visit: each ] ].
	stream nextPutAll: $).! !

! SequenceNode methodsFor: #validation !
validateNames: names
	" Validate the receiver can be emmited in a context with names. "

	self tempNames do: [:temp|
		(names includes: temp) ifTrue: [
			^self error: 'Invalid temp (',temp,')'
		].
	]! !

! MethodNode methodsFor: #validation !
validateArguments
	" Ensure the nodes of the receiver can be emmited with the defined arguments. "

	| names |
	names := HashedSet withAll: self argumentNames.
	self nodes do: [:each| each validateNames: names ]! !

! BlockNode methodsFor: #bindings !
bindings
	" Return the bindings of the receiver. "

	^self parameterNames! !

! JSEncoder methodsFor: #visiting !
visitBlockSequenceNode: aNode
	| index |
	nestedBlocks := nestedBlocks + 1.
	aNode nodes isEmpty
	    ifTrue: [ self emmitReturn: #nil ]
	    ifFalse: [
		aNode tempNames do: [:each |
		    self tempVariables add: each.
		    stream nextPutAll: 'var ', (self localName: each), '=nil;'
		].
		index := 0.
		aNode nodes do: [:each |
		    index := index + 1.
		    index = aNode nodes size
			ifTrue: [ self emmitReturn: [ self visit: each ] ]
		    	ifFalse: [ self visit: each. stream nextPutAll: $; ]
		].
	].
	nestedBlocks := nestedBlocks - 1! !

! JSEncoder methodsFor: #visiting !
visitSequenceNode: aNode
	aNode tempNames do: [:each |
		self tempVariables add: each.
		stream nextPutAll: 'var ', (self localName: each), '=nil;'
	].
	aNode nodes do: [:each |
		self visit: each.
		stream nextPutAll: ';'
	] separatedBy: [stream cr]! !

! SequenceNode methodsFor: #bindings !
bindings
	" Return the bindings of the receiver. "

	^self tempNames! !

"SequenceNode removeSelector: #validateArguments:"!
"------------------------------------------------------------------------------------------------"!

! SmalltalkParser methodsFor: #private !
methodParserParts
	" Private - Return a parser for parts of a method. "

 | expression separator comment ws identifier keyword stSelectorPart
  string symbol number variable reference globalReference
  literal returnExpression expressionParser keyword unarySelector
  binarySelector keywordPattern unaryPattern binaryPattern assignment
  temps blockParamList block expression expressions subexpression
  statements sequence operand unaryMessage unarySend unaryTail
  binaryMessage binarySend binaryTail keywordMessage keywordSend
  keywordPair cascade message inline jsIndex character directive
  arrayLiteral arrayVariable literalArray aSymbol
  stIdentifier jsIdentifier noReference alienReference
  alienIdentifier emptyArray hashedString annotation argument
 |

 separator := self separator.
 comment := self comment.
 ws := (separator / comment @@#sp) star @@#ws.
 
 string := self ,$' ,PPStStringParser new @@#string
  ==> [:node| ValueNode with: node second ].
 
 hashedString := self ,($#,$') ,PPStStringParser new @@#hashedString
  ==> [:node| ValueNode with: node second ].

 stIdentifier := (self nameStart, self symbolCharParser star @@#stId) flatten @@#stIdentifier.
 stSelectorPart := (self upNameStart, self symbolCharParser star @@#stSel) flatten @@#stSelectorPart.
 jsIdentifier := (self ,$# ,self jsSymbolCharParser plus @@#jsId) flatten @@#jsIdentifier.
 alienIdentifier := self jsSymbolCharParser plus flatten @@#alienIdentifier.
 identifier := jsIdentifier / stIdentifier @@#identifier.

 keyword := (stSelectorPart / identifier, $:) flatten @@#keyword.

 annotation := string / stIdentifier.
 argument := identifier ,(self ,$@ ,annotation) star ==> [:node|
	(ValueNode with: node first) annotations: node last
 ].

 symbol := self ,$# ,self extendedSymbolCharParser plus flatten @@#symbol
  ==> [:node | ValueNode with: node second ].

 number := self number flatten
  ==> [:node | ValueNode with: node asNumber ].

 character := self ,$$ ,PPAnyParser new @@#character
  ==> [:node | ValueNode with: (node at: 2) ].

 directive := self directive: number ws: ws.

 variable := identifier ==> [:token | VariableNode new value: token ].
 noReference := self ,$# @@#noReference ==> [:token | NoReceiverNode new ].
 literal := PPDelegateParser new @@#literalDelegate.
 expression := PPDelegateParser new @@#expressionDelegate.

 arrayVariable :=
    (self ,#true  ==> [:node| ValueNode with: true  ])
  / (self ,#false ==> [:node| ValueNode with: false ])
  / (self ,#nil   ==> [:node| ValueNode with: nil   ])
  @@#arrayVariable.

 aSymbol := self extendedSymbolCharParser plus flatten @@#aSymbol
  ==> [:node| ValueNode with: node ].

 arrayLiteral := literal / arrayVariable / aSymbol / variable.
 literalArray := (self gc: '#(')
	,(ws ,arrayLiteral ,ws) star ,$)
	@@#literalArray
  ==> [:node | ValueNode with: (Array
	withAll: (node second
		collect: [:each | each second value])) ].
 emptyArray := self ,'#(' ,ws ,$) @@#emptyArray
  ==> [:node | ValueNode with: Array new ].

 globalReference := self globalNameParser @@#globalReference
  ==> [:token | self gc. GlobalReferenceNode new value: token ].
 alienReference := self ,'#{' ,alienIdentifier, $} @@#alienReference
  ==> [:node | AlienNode new value: node second ].

 jsIndex := (self gc: $[) ,expression, $] / (self ,$# ,identifier).
 reference := (variable / globalReference / alienReference @@#ref)
	,jsIndex star @@#reference
  ==> [:nodes | nodes first indexedAs: nodes ].

 binarySelector := (
	(self optional: $#)
	,self binaryCharParser plus
	) flatten @@#binarySelector.

 unarySelector := stSelectorPart / identifier @@#unarySelector.

 keywordPattern :=
	(ws, keyword  @@#kw ,
	 ws, argument @@#kwId
	) plus @@#keywordPattern
  ==> [:nodes | Array
   with: ((nodes collect: [:each | each at: 2]) join: '')
   with: (nodes collect: [:each | each at: 4]) ].

 binaryPattern := ws, binarySelector, ws, argument @@#binaryPattern
  ==> [:node | Array with: node second with: (Array with: node fourth)].

 unaryPattern := ws, unarySelector @@#unaryPattern
  ==> [:node | Array with: node second with: Array new].
 
 expressions := expression, ((ws, $., ws, expression) @@#anExpression
  ==> [:node | node fourth]) star @@#expressions
  ==> [:node | (Array with: node first) ,node second ].

 assignment := reference, ws, ':=', ws, expression @@#assignment
  ==> [:node | AssignmentNode new left: node first; right: (node at: 5)].

returnExpression := (self gc: $^) @@#caret
	,(self ,$@ ,annotation) star
	, ws, expression, ws
	, $. asParser optional	@@#returnExpression
     ==> [:node | ReturnNode new
		addNode: ((node at:4) annotations: node second);
		yourself ].

 temps := self, $|, (ws, argument) star, ws, $| @@#temps
  ==> [:node | node second collect: [:each | each second ] ].

 blockParamList := (self ,$: ,ws ,argument ,ws) plus ,$| @@#blockParamList
  ==> [:node | node first collect: [:each | each third ] ].

 subexpression := (self gc: $() ,ws ,expression ,ws ,$) @@#subexpression
  ==> [:node | node third ].

 statements := (returnExpression @@#stat1 ==> [:node | Array with: node])
  / (expressions ,ws ,$. ,ws ,returnExpression @@#stat2 ==> [:node | node first add: (node at: 5); yourself ])
  / (expressions ,(self optional: $.) @@#stat3 ==> [:node | node first ]).

 sequence := temps optional @@#temps
	, ws, directive star
	, ws, statements optional
	, ws @@#sequence
  ==> [:node | ExtendedSequenceNode new
	temps: node first;
	directives: node third;
	nodes: (node at: 5);
	yourself ].

 block := (self gc: $[)
	,ws ,blockParamList optional
	,ws ,sequence optional
	,ws ,$] @@#block
  ==> [:node | BlockNode new
	parameters: node third;
	addNode: (node at: 5) asBlockSequenceNode ].

 operand := block / literal / reference / noReference / subexpression @@#operand.

 literal parser: symbol / number / string / character
  / emptyArray / literalArray / hashedString @@#literalDelegate.

 unaryMessage := ws ,unarySelector ,(self ,$:) not @@#unaryMessage
  ==> [:node | SendNode new selector: node second ].

 unaryTail := PPDelegateParser new @@#unaryTailDelegate.
 unaryTail parser: (unaryMessage, unaryTail optional @@#unaryTailDelegate
  ==> [:node | node second isNil ifTrue: [ node first ]
   ifFalse: [ node second valueForReceiver: node first ] ]).

 unarySend := operand, unaryTail optional @@#unarySend
  ==> [:node | node second isNil ifTrue: [ node first ]
   ifFalse: [ node second valueForReceiver: node first ] ].

 binaryMessage := ws, binarySelector, ws, (unarySend / operand) @@#binaryMessage
  ==> [:node | SendNode new
   selector: node second;
   arguments: (Array with: node fourth) ].

 binaryTail := PPDelegateParser new @@#binaryTailDelegate.
 binaryTail parser: (binaryMessage, binaryTail optional
  ==> [:node | node second isNil ifTrue: [ node first ]
   ifFalse: [ node second valueForReceiver: node first ] ]).

 binarySend := unarySend, binaryTail optional @@#binarySend
  ==> [:node | node second isNil ifTrue: [ node first ]
   ifFalse: [ node second valueForReceiver: node first ] ].

 keywordPair := keyword, ws, binarySend @@#keywordPair.

 keywordMessage := (ws, keywordPair) plus @@#keywordMessage
  ==> [:nodes | SendNode new
   selector: ((nodes collect: [:each | each second first]) join: '');
   arguments: (nodes collect: [:each | each second third])].

 keywordSend := binarySend, keywordMessage @@#keywordSend
  ==> [:node | node second valueForReceiver: node first].

 message := binaryMessage / unaryMessage / keywordMessage @@#message.

 cascade := (keywordSend / binarySend @@#preCascade), (ws, $;, message) plus @@#cascade
  ==> [:node | node first cascadeNodeWithMessages: 
   (node second collect: [:each | each third ])].

 inline := self ,${
	,(stIdentifier optional ,string ,ws) plus
	,$} @@#inline
  ==> [:node | InlineNode new sources: (node second collect: [:each|
		Array	with: each first with:
			(each second value replace: '''''' with: '''')
		])
	].

 expression parser: assignment / cascade / keywordSend / binarySend / inline @@#expression.

	^Array	with: ws
		with: keywordPattern / binaryPattern / unaryPattern @@#header
		with: sequence optional @@#body! !
