" Adaptions to run S8 on Javascript execution machinery.
Copyright (C) 2013-today Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
In case of questions and bug reports please join us at S8 Facebook group
	https://www.facebook.com/groups/s8smalltalk/
"!

"------------------------------------------------------------------------------
	Javascript code generation
 ------------------------------------------------------------------------------"!
NativeEncoder
	subclass: #JSEncoder
	instanceVariableNames: ' '
	category: #JS !

! JSEncoder class methodsFor: #instantiation !
supportsLanguage: aLanguage
	" Private - Return true if instances of the receiver can encode to aLanguage. "

	^aLanguage isNil or: [ #(js javascript) includes: aLanguage ] ! !

! JSEncoder methodsFor: #accessing !
nativeLanguage
	" Return the nativeLanguage of the receiver. "

	^#js ! !

! JSEncoder methodsFor: #encoding !
nativeUndefined
	" Private - Return the code for undefined. "

	^#undefined ! !

! JSEncoder methodsFor: #encoding !
nativeNameOfMetaClass: aClass
	" Private - Return the native name for metaclass aClass. "

	^aClass instanceClass name ,'.$klass' ! !

! JSEncoder methodsFor: #inlines !
missingNativeCodeFor: aNode as: lang
	" Private - Return the native language code for missing expression (or nil). "

	^'(self.invalidMessage_("' ,compiler selector ,'"))'! !

! JSEncoder methodsFor: #export !
defaultRequirementsFor: aClass
	" Private - Return the default requirements to load aClass.
	Detect native bindings and generate code to rebind before class definition.
	"

	| constructorName code |
	constructorName := aClass constructorName.
	constructorName isNil ifTrue: [ ^#() ].
	^Array with:
		'if(',constructorName,' && !!'
		, (self smalltalkNameFor: aClass)
		,') smalltalk.mapClassName("'
		,aClass name ,'","'
		,aClass category, '",'
		,constructorName, ','
		,(self smalltalkNameFor: aClass superclass)
		,');'! !

! JSEncoder methodsFor: #generation !
emmitEnterMethod: aMethodNode
	" Private - Generate home context definitions, entering aMethodNode. "

	self nextPutAll: 'var self=this;'; cr ! !

! JSEncoder methodsFor: #generation !
emmitFunctionBody: aBlock
	" Private - Emmit the code of function body evaluating aBlock. "

	self nextPutAll: ${.
	aBlock evaluate.
	self nextPutAll: $}. ! !

! JSEncoder methodsFor: #generation !
emmitFunction: fName node: aNode arguments: argumentNames doing: aBlock
	" Private - Generate function for aNode. "

	| aStream |
	aStream := '' writeStream.
	argumentNames do: [:each | 
		self tempVariables add: each.
		aStream nextPutAll: (self localName: each) ]
	    separatedBy: [ aStream nextPutAll: ',' ].
	self	emmitFunction: fName
		arguments: aStream contents
		doing: aBlock! !

! JSEncoder methodsFor: #export !
export: aClass method: aMethod on: aStream
	" Export a method definition of aClass onto aStream. "

	| fName selectorName bar source |
	source := self nativeSourceFor: aMethod.
	source isNil ifTrue: [ ^false ].
	aMethod isNil ifTrue: [ self error: 'Method is missing' ].
	bar := $" asString.
	fName := aMethod selector.
	fName first isBackSlash ifTrue: [
		selectorName := self escaped: aMethod selector.
		fName := self escaped: (self asSelector: fName).
	] ifFalse: [
		selectorName := bar ,aMethod selector, bar.
		(self asSelector: fName) = (aMethod selector replace: $: with: $_) ifTrue: [
			fName := '0' "we exploit the fact that 0 isNil to signal a compact selector"
		] ifFalse: [
			fName := bar ,(self asSelector: fName) ,bar
		].
	].
	aStream
		nextPutAll: 'smalltalk.bind(' ,(self smalltalkNameFor: aClass)
			,',' ,selectorName ,',' ,fName ,'
,'			,source.
	(aMethod source notNil and: [ aMethod source notEmpty ]) ifTrue: [
	    aStream cr; "warning: this cr is important for debugger"
		nextPutAll: ',' ,(self escaped: aMethod category)
			,',' ,(self escaped: aMethod source).
	].
	((aMethod respondsTo: #stamp) and: [ aMethod stamp notNil ]) ifTrue: [
		aStream nextPutAll: ',' ,(self escaped: aMethod stamp)
	].
	aStream nextPutAll: ');
'! !

! JSEncoder methodsFor: #export !
exportCommentOf: aClass on: aStream
	" Export the comment of aClass onto aStream. "

	aClass comment size > 0 ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: '.$comment=' ,(self escaped: aClass comment);
		nextPutAll: ';'; cr
	]. ! !

! JSEncoder methodsFor: #export !
exportDefinitionOf: aClass on: aStream
	" Export definition of aClass onto aStream. "

	(self loadRequirementsOf: aClass) do: [:each|
		each isString
		ifTrue: [ aStream nextPutAll: each;cr ]
		ifFalse: [ aStream fileOutOn: aStream ].
	].
	aStream 
	    nextPutAll: 'smalltalk.addClass("', aClass name, '",';
	    nextPutAll: (self smalltalkNameFor: aClass superclass);
	    nextPutAll: ','.
	self exportArrayLiteral: aClass instanceVariableNames on: aStream newLine: false.
	aStream
	    nextPutAll: ',"' ,aClass category, $";
	    nextPutAll: ');'; cr.
	aClass classVariables notNil ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: '.$classVariableNames=' ,(self escaped: aClass classVariablesString);
		nextPutAll: ';'; cr
	].
	aClass poolDictionaries notNil ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: '.$poolDictionaries='.
	   self exportArrayLiteral: aClass poolDictionaries on: aStream newLine: true.
	].
	aClass propertyNames notEmpty ifTrue: [
	    aStream
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: '.$propertyNames='.
	   self exportArrayLiteral: aClass propertyNames on: aStream newLine: true.
	].
	self exportCommentOf: aClass on: aStream.! !

! JSEncoder methodsFor: #export !
exportMetaDefinitionOf: aClass on: aStream
	" Export class side definition of aClass onto aStream. "

	aClass class instanceVariableNames notEmpty ifTrue: [
		aStream 
			nextPutAll: (self smalltalkNameFor: aClass class);
			nextPutAll: '.$iVarNames='.
		self exportArrayLiteral: aClass class instanceVariableNames on: aStream newLine: true.
	].

	aClass class propertyNames notEmpty ifTrue: [
		aStream 
			nextPutAll: (self smalltalkNameFor: aClass class);
			nextPutAll: '.$propertyNames='.
		self exportArrayLiteral: aClass class propertyNames on: aStream newLine: true.
	].! !

! JSEncoder methodsFor: #directives !
sourceToTry: sourceCode
	" Return the code to emmit to try execution of sourceCode. "

	^'try' ,${ ,sourceCode ,$} ,' catch($$e) {
	if ($$e.directiveError) { return(this.error_($$e.directiveError)); }
	if (!!$$e.directiveFailed) { throw($$e); } }'! !

! JSEncoder methodsFor: #directives !
sourceToDirectCall: nativeSelector with: arg
	" Return the code to emmit to direct native call to selector with argument. "

	^'return this.' ,nativeSelector ,'(' ,arg ,',' ,'arguments' ,');' ! !

! JSEncoder methodsFor: #generation !
generate: fName early: early return: code
	" Private - Generate code to catch early return of code expression in fName. "

	| prev |
	early ifFalse: [ self nextPutAll: code. ^0 ].
	prev := 'var $s8Ret$={name:"stReturn"};try' ,${.
	self	nextPutAll: prev;
		nextPutAll: code;
		nextPutAll: $} ,' catch($$ex) {if($$ex === $s8Ret$)' ,${.
	self emmitReturn: '$$ex.result' early: true.
	self nextPutAll: $} ,' throw($$ex);' ,$}.
	^prev size! !

! JSEncoder methodsFor: #generation !
newMethodFunctionName
	" Private - Return the function to call when installing a new method. "

	^'smalltalk.newMethod'! !

! JSEncoder methodsFor: #visiting !
visitBlockNode: aNode

	self nextArgumentsPut: [
		self	emmitFunction: '' node: aNode
			arguments: aNode parameterNames
			doing: [ self visitNodesIn: aNode cr: false ]
	] ! !

! JSEncoder methodsFor: #visiting !
visitCascadeNode: aNode

	(self tempVariables includes: '$rec') ifFalse: [ self tempVariables add: '$rec' ].
	self compilerEvent: #aboutCascade: with: aNode.
	self nextArgumentsPut: [
		self emmitFunction: '' arguments: '$rec' doing: [ | index rec |
			index := 0.
			aNode nodes do: [:each |
			    index := index + 1.
			    index = aNode nodes size ifTrue: [ self emmitReturn: nil ].
			    rec := VariableNode new.
			    rec value: '$rec'; start: each start end: each start.
			    each receiver: rec.
			    self visit: each.
			    index = aNode nodes size ifTrue: [
				self compilerEvent: #returned:early: withArguments: (Array with: nil with: false)
			    ].
			    self nextPutAll: ';'
			].
		]
	].
	self nextArgumentsPut: [ self visit: aNode receiver ].
	self compilerEvent: #cascade: with: aNode.! !

! JSEncoder methodsFor: #visiting !
visitClassVariableAssignment: aClassVariableName in: aClass to: anExpressionNode

	self
		nextPutAll: (self smalltalkNameFor: aClass);
		nextPutAll: '.classVariableAt_put_("';
		nextPutAll: aClassVariableName;
		nextPutAll: '",'.
	self visit: anExpressionNode.
	self	nextPutAll: ')'! !

! JSEncoder methodsFor: #visiting !
visitDirectiveNode: aNode
	" Private - Visit a directive node. "

	(aNode respondsTo: #asNative:for:on:) ifTrue: [
		| code |
		code := aNode
			asNative: functionName
			for: self on: stream.
		^code notNil ifTrue: [
			self nextPutAll: code; cr
		]
	].
	self visitNode: aNode in: nil cr: true! !

! JSEncoder methodsFor: #visiting !
visitGlobalReferenceNode: aNode

	| aName aClass lookupPools |
	aName := aNode value.
	aClass := self currentClass instanceClass.
	aClass classVariableNamesDo: [:each|
		each = aName ifTrue: [
			self
				nextPutAll: (self smalltalkNameFor: aClass);
				nextPutAll: '.classVariableAt_("';
				nextPutAll: aName;
				nextPutAll: '")'.
			^self
		]
	].
	aClass sharedPoolsAndNamesDo: [:poolName :pool|
		(pool includesKey: aName) ifTrue: [
			self
				nextPutAll: 'smalltalk.poolDictionary_("';
				nextPutAll: poolName;
				nextPutAll: '").at_("';
				nextPutAll: aName;
				nextPutAll: '")'.
			^self
		].
	].
	self nextPutAll: (self codeToGetGlobal: aNode value).! !

! JSEncoder methodsFor: #visiting !
visitMethodNode: aMethodNode
	" Visit a method node. "

	self
		nextPutAll: self newMethodFunctionName, '(';
		nextPutAll: (self smalltalkNameFor: self currentClass);
		nextPutAll: ',"', aMethodNode selector, '",'; cr.
	self emmitMethod: aMethodNode.
	self includeSourceCode ifTrue: [
		self	cr; "warning: this cr is important for debugger"
			nextPutAll: ',' ,(self escaped: aMethodNode source)
	].
	self nextPutAll: ')'! !

! JSEncoder methodsFor: #visiting !
visitNode: aNode index: index
	" Private - Generate code for indexed access. "

	| selector |
	((index isMemberOf: ValueNode) and: [ index hasString ]) ifTrue: [
		selector := self asSelector: index value.
		((selector includes: $ ) or: [ selector includes: $_ ]) ifFalse: [
			self nextPutAll: $.; nextPutAll: selector.
			^self
		].
	].
	self nextIndexPut: [ self visit: index ] ! !

! JSEncoder methodsFor: #encoding !
messageSendConnector
	" Private - Return the connector to use for sending a message. "

	^$. ! !

! JSEncoder methodsFor: #encoding !
messageSendConnector: aNode
	" Private - Return the connector to use for sending message. "

	aNode receiver messageSendConnector isNil ifTrue: [ ^'' ].
	^self messageSendConnector ! !

! JSEncoder methodsFor: #encoding !
encodeNormalSend: aNode
	" Private - Generate code for message send. "

	self	visitReceiver: aNode receiver;
		nextPutAll: (self messageSendConnector: aNode);
		nextPutAll: (self asSelector: aNode selector);
		nextArgumentsPut: [ self visitSendArguments: aNode ] ! !

! JSEncoder methodsFor: #visiting !
visitOperation: aNode
	" Private - aNode is an operation and selector is mapped to target operation. "

	self compilerEvent: #aboutOperation: with: aNode.
	self nextArgumentsPut: [ self visitReceiver: aNode receiver ].
	self nextPutAll: (self asSelector: aNode selector).
	self visitSendArguments: aNode.
	self compilerEvent: #operation: with: aNode.! !

! JSEncoder methodsFor: #visiting !
visitPoolVariableAssignment: aName in: poolName to: anExpressionNode

	self
		nextPutAll: 'smalltalk.poolDictionary_("';
		nextPutAll: poolName; nextPutAll: '").at_put_("';
		nextPutAll: aName; nextPutAll: '",'.
	self visit: anExpressionNode.
	self	nextPutAll: ')'! !

! JSEncoder methodsFor: #visiting !
visitReturnNode: aNode

	| theReturn |
	nestedBlocks > 0 ifTrue: [ earlyReturn := true ].
	theReturn := [ self emmitReturn: [
		aNode nodes do: [:each | self visit: each ]
	] ].
	earlyReturn ifFalse: [ ^theReturn value ].
	self nextArgumentsPut: [
		self emmitFunctionDoing: [
			self nextPutAll: '$s8Ret$.result='.
			self nextArgumentsPut: [
				self emmitFunctionDoing: theReturn.
			]; nextArgumentsPut: [].
			self nextPutAll: ';throw($s8Ret$)'
		]
	]; nextArgumentsPut: [] ! !

! JSEncoder methodsFor: #encoding !
encodeSuperSend: aNode
	" Private - Generate code for super message send."

	self nextPutAll: 'smalltalk.superImplementor'.
	self nextArgumentsPut: [
		self nextPutAll:
			(self smalltalkNameFor: self currentClass)
			,',"',(self asSelector: aNode selector) ,$"
	].
	self nextPutAll: '.apply(self, '.
	self nextIndexPut: [ self visitSendArguments: aNode ].
	self nextPutAll: ')'. ! !

! JSEncoder methodsFor: #optimization !
coreOptimizations
	" Private - Return the optimization set. "

	^#(core  #(
		#(class	'%receiver%.$klass' )
		#(isNil	'(nil.isNil_(%receiver%))' )
		#(notNil '(nil.isNil_(%receiver%)===false)' )
		#(==	'(nil.is_eqeq_(%receiver%,%arg1%))' )
		#(===	'(%receiver% === %arg1%)' )
		#(||	'(%receiver% || %arg1%)' )
		#(basicAt: '((function(){var $1$=(%receiver%[%arg1%]);if(nil.isNil_($1$))return nil;return $1$;})())' )
		#(ifUndefined: hasBlockNoArguments: '(function(){var $1$=(%receiver%);if (typeof($1$)=="undefined") return %arg1%();return $1$;})()' )
		#(basicAt:ifAbsent: '((function(){var $1$=(%receiver%[%arg1%]);if(nil.isNil_($1$))return (%arg2%.value());return $1$;})())' )
		#(basicAt:put: '(%receiver%[%arg1%]=(%arg2%))' )
	))! !

! JSEncoder methodsFor: #optimization !
s8Optimizations
	" Private - Return the optimization set. "

	^#(s8	#(
		core
		#(ifTrue: hasBlockNoArguments: #(
			#(ifTrue: noClient: 'if (%receiver%){%seq1%};' )
			#(ifTrue: '((%receiver%).mustBeBoolean()==true ? %arg1%() : nil)' )
		))
		#(ifFalse: hasBlockNoArguments:  #(
			#(ifFalse: noClient: 'if (!!(%receiver%)){%seq1%};' )
			#(ifFalse: '((%receiver%).mustBeBoolean()==true ? nil : %arg1%())' )
		))
		#(ifTrue:ifFalse: hasBlockNoArguments: #(
			#(ifTrue:ifFalse: noClient: 'if (%receiver%){%seq1%} else{%seq2%};' )
			#(ifTrue:ifFalse: '((%receiver%).mustBeBoolean()==true ? %arg1%() : %arg2%())' )
		))
		#(ifFalse:ifTrue: hasBlockNoArguments: #(
			#(ifFalse:ifTrue: noClient: 'if (%receiver%){%seq2%} else{%seq1%};' )
			#(ifFalse:ifTrue: '((%receiver%).mustBeBoolean()==true ? %arg2%() : %arg1%())' )
		))
		#(whileTrue receiverIsBlockNoArguments: '(function(){while(%receiver%()){};return nil})()' )
		#(whileFalse receiverIsBlockNoArguments: '(function(){while(!!%receiver%()){};return nil})()' )
		#(orNil '(function(){var $1$=(%receiver%);return (nil.isNil_($1$)) ? nil : $1$;})()' )
		#(whileTrue: allBlockNoArguments: #(
			#(whileTrue: noClient: 'while(%receiver%()){%seq1%};' )
			#(whileTrue: '(function(){while(%receiver%()){%arg1%()};return nil})()' )
		))
		#(whileFalse: allBlockNoArguments: #(
			#(whileFalse: noClient: 'while(!!%receiver%()){%seq1%};' )
			#(whileFalse: '(function(){while(!!%receiver%()){%arg1%()};return nil})()' )
		))
		#(ifNil: hasBlockArguments: '(function(){var $1$=(%receiver%);return (nil.isNil_($1$)) ? %arg1%() : $1$;})()' )
		#(ifNotNil: hasBlockArguments: '(function(){var $1$=(%receiver%);return (nil.isNil_($1$)===false) ? %arg1%(%receiver%) : $1$;})()' )
		#(timesRepeat: hasBlockNoArguments: '(function(){for(var $1$=(%receiver%);$1$>0;$1$--){%arg1%()};return nil})()' )
		#(isKindOf: 'smalltalk.is_kindOf_(%receiver%,(%arg1%))' )
		#(respondsTo: 'smalltalk.responds_to_(%receiver%,(%arg1%))' )
		#(ifNil:ifNotNil: hasBlockNoArguments: '(nil.isNil_(%receiver%)) ? %arg1%() : %arg2%()' )
		#(ifNotNil:ifNil: hasBlockNoArguments: '(nil.isNil_(%receiver%)===false) ? %arg1%() : %arg2%()' )
	)) ! !

! JSEncoder methodsFor: #optimization !
magnitudeOptimizations
	" Private - Return the optimization set. "

	^#(magnitude #(
		#(>  hasNumber: '(%receiver% > %arg1%)' )
		#(<  hasNumber: '(%receiver% < %arg1%)' )
		#(>= hasNumber: '(%receiver% >= %arg1%)' )
		#(<= hasNumber: '(%receiver% <= %arg1%)' )
	))! !

! JSEncoder methodsFor: #optimization !
mathOptimizations
	" Private - Return the optimization set. "

	^#(math #(
		#(+ hasNumber: '(%receiver% + %arg1%)' )
		#(#- hasNumber: '(%receiver% - %arg1%)' )
		#(* hasNumber: '(%receiver% * %arg1%)' )
		#(/ hasNumber: '(%receiver% / %arg1%)' )
		#(= hasNumber: '(%receiver% == %arg1%)' )
	))! !

! JSEncoder methodsFor: #defaults !
defaultKnownOptimizations
	" Private - Return the default knownOptimizations of the receiver. "

	^(Array	with: self coreOptimizations)
		add: self s8Optimizations;
		add: self magnitudeOptimizations;
		add: self mathOptimizations;
		yourself! !


! JSEncoder class methodsFor: #inlines !
initializeKernelInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self kernelInlines into: pool ! !

! JSEncoder class methodsFor: #inlines !
kernelInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(Object #(
#(print:		'self')
#(halt		'debugger')
#(toString	'self')
#(toString:	'""+anObject')
#(isObject:	'return (anObject==0 || anObject && anObject.$klass)?true:false')
#(typeOf:	'typeof anObject')
#(callableOf:	'self')
#(basicDelete:	'delete self[aString]')
#(basicPerform:withArguments: #('self' 'fn.apply(self,anArray)'))
#(asJSONObject	'{}')
#(printingWith:do:	'self')
#(isFile:olderThan: 'self')
))
#(Smalltalk #(
#(includesKey:	'undefined')
#(eval:	'return eval(anExpression)')
))
#('Smalltalk class' #(
#(global	'(function(){return this;}).call(null)')
#(gc	'if(typeof gc!!="undefined"){gc(); return true}')
))
#(Behavior #(
#(basicNew	'new self.$fn()')
#(fn		'self')
#(prototype	'self')
#(privateMethods	'self')
))
#(CompiledMethod #(
#(asClosure	'self')
#(printString	'self')
))
#(Number #(
#(= '
	if (nil.isNil_(anObject)) return false;
	if (!!(anObject.isNumber)) return false;
	if (!!(anObject.isNumber())) return false;
	if (self == anObject) return true;
	return ((0+self-anObject) == 0)')
#(> 'self > aNumber.mustBeNumber()')
#(< 'self < aNumber.mustBeNumber()')
#(>= 'self >= aNumber.mustBeNumber()')
#(<= 'self <= aNumber.mustBeNumber()')
#(+ 'self + aNumber.mustBeNumber()')
#(- 'self - aNumber.mustBeNumber()')
#(* 'self * aNumber.mustBeNumber()')
#(/ 'self / aNumber.mustBeNumber()')
#(sin 'Math.sin(self)')
#(cos 'Math.cos(self)')
#(tan 'Math.tan(self)')
#(arcCos 'Math.acos(self)')
#(arcSin 'Math.asin(self)')
#(arcTan 'Math.atan(self)')
#(atan2: 'Math.atan2(self,aNumber.mustBeNumber())')
#(sqrt 'Math.sqrt(self)')
#(<< 'self << aNumber.mustBeNumber()')
#(>> 'self >> aNumber.mustBeNumber()')
#(raisedTo: 'Math.pow(self,aNumber.mustBeNumber())')
#(max: 'Math.max(self,aNumber)')
#(min: 'Math.min(self,aNumber)')
#(rounded 'Math.round(self)')
#(truncated 'Math.floor(self)')
#(timesRepeat: 'for(var i=1;i<=self;i++){fn(i);}')
#(to:do: 'for(var i=self;i<=aNumber;i++){fn(i);}')
#(to:by:do: 'if (increment>0) { for(var i=self;i<=aNumber;i=i+increment){fn(i);} }
	else { for(var i=self;aNumber<=i;i=i+increment){fn(i);} }')
#(printString 'String(self)')
#(toString: 'self.toString(radix)')
#(isNumber 'return isNaN(self).not()')
#(asJSONObject 'self')
#(clearInterval 'clearInterval(self)')
#(clearTimeout 'clearTimeout(self)')
))
#('Number class' #(
#(pi 'Math.PI')
#(basicNew 'self')
))
#(BlockClosure #(
#(compiledSource 'self')
#(whileTrue: 'while(self()) {fn()}')
#(whileFalse: 'while(!!self()) {fn()}')
#(value 'self()')
#(value: 'self(anObject)')
#(value:value: 'self(arg1,arg2)')
#(value:value:value: 'self(arg1,arg2,arg3)')
#(value:value:value:value: 'self(arg1,arg2,arg3,arg4)')
#(valueWithPossibleArguments: 'self.apply(null,array)')
#(try:catch: '
	var result;
	try{ result = fn() } catch(e) {
		result = ab( self.smalltalkErrorOf_(e) )
	};
	return result')
#(valueWithTimeout: 'setTimeout(self,aNumber)')
#(valueWithInterval: 'setInterval(self,aNumber)')
#(ensure: 'try{return self();} finally {fn();}')
#(ifCurtailed: '
	try{	result = self.value();
		curtailed = false;
	} finally {
		if (curtailed) exceptionBlock.value();
	};
	return result')
#(argumentCount 'self.length')
))
#('BlockClosure class' #(
#(defaultSortBlock 'self')
))
#(Boolean #(
#(=	'
	if (nil.isNil_(anObject)) return false;
	if (!!(self.$klass === anObject.$klass)) return false;
	return (self.valueOf() == anObject.valueOf())')
#(ifTrue:ifFalse: 'if(self == true) { return aBlock.value(); } else { return anotherBlock.value(); }')
#(not 	'self == false')
#(asJSONObject 'self')
))
#(DateTime #(
#(year 'self.getFullYear()')
#(month 'self.getMonth() + 1')
#(month: 'self.setMonth(aNumber - 1)')
#(day 'self.getDay() + 1')
#(day: 'self.setDay(aNumber - 1)')
#(year: 'self.setFullYear(aNumber)')
#(dayOfMonth 'self.getDate()')
#(dayOfMonth: 'self.setDate(aNumber)')
#(asString 'self')
#(time 'self.getTime()')
#(time: 'self.setTime(aNumber)')
#(asDateString 'self.toDateString()')
#(asTimeString 'self.toTimeString()')
#(asLocaleString 'self.toLocaleString()')
#(hours: 'self.setHours(aNumber)')
#(minutes: 'self.setMinutes(aNumber)')
#(seconds: 'self.setSeconds(aNumber)')
#(milliseconds: 'self.setMilliseconds(aNumber)')
#(hours 'self.getHours()')
#(minutes 'self.getMinutes()')
#(seconds 'self.getSeconds()')
#(milliseconds 'self.getMilliseconds()')
#(< 'self < aDateTime')
#(> 'self > aDateTime')
#(<= 'self <= aDateTime')
#(>= 'self >= aDateTime')
#(- 'self - aDateTime')
#(+ 'self + aDateTime')
#(isValid 'isFinite(self)')
#(asJSONObject 'self')
))
#('DateTime class' #(
#(new: 'return new Date(anObject)')
#(today 'self')
#(now 'self')
#(msNow 'self')
))
#(UndefinedObject  #(
#(asJSONObject 'null')
))
#(Collection  #(
#(wrap: 'self')
#(unwrap: 'self')
))
#(SequenceableCollection #(
#(indexOf:ifAbsent: '
	for(var i=0;i<self.length;i++){ if (self[i]._eq(anObject)) { return i + 1; } }
	return aBlock.value()')
#(detect:ifNone: '
	for(var i = 0; i < self.length; i++) if(fn(self[i])) return self[i];
	return anotherBlock.value()')
#(reversed '
	var result = new Array();
	var len=self.length; var i=len;
	while(i--){result[len-i-1]= self[i];}
	return result')
#(isEmpty '(this.length)?false:true')
#(reverseDo:  'var i=self.length;while(i--){fn(self[i]);}')
#(includes: #('self' '
	if (anObject.isNumber() && (self.indexOf(anObject) > -1)) return true;
	if (anObject.isString() && (self.indexOf(anObject) > -1)) return true;
	var i = self.length;
	while (i--) { if (anObject._eq(self[i])) { return true; } }
	return false'))
#(occurrencesOf: '
	var i = self.length; var count = 0;
	while (i--) { if (self[i]._eq(anObject)) { count++; } }
	return count')
))
#(String #(
#(=	'
	if (nil.isNil_(anObject)) return false;
	if (!!(anObject.isString)) return false;
	if (!!(anObject.isString())) return false;
	if (self == anObject) return true;
	return ((""+self) == (""+anObject))')
#(size	'self.length')
#(at:	'self.charAt(anIndex - 1)')
#(do: 	'for(var i=0;i<self.length;i++){fn(self.charAt(i));}')
#(detect:ifNone: 'for(var i = 0; i < self.length; i++)
		if(fn(self.charAt(i))) return self.charAt(i);
	return anotherBlock.value()')
#(includes: 'self')
#(withIndexDo: 'for(var i=0;i<self.length;i++){fn(self.charAt(i), i+1);}')
#(isLiteral 'self')
#(occurrencesOf: 'self')
#(startsWith: 'if (self.startsWith) { return self.startsWith(aString) }')
#(escaped 'return escape(self)')
#(unescaped 'return unescape(self)')
#(asEncodedURI 'return encodeURIComponent(self)')
#(asDecodedURI 'return decodeURIComponent(self)')
#(, '
	if (nil.isNil_(aCollection)) return self+""; // for JSC under iOS
	return self + aCollection.toString()')
#(copyWith: 'self + anObject.toString()')
#(copyFrom:to: 'self.substring(anIndex - 1,anotherIndex)')
#(reversed 'self')
#(indexOf: 'return self.indexOf(aString) + 1')
#(indexOf:startingAt: 'return self.indexOf(aString,position-1) + 1')
#(lastIndexOf: 'return self.lastIndexOf(aString) + 1')
#(lastIndexOf:startingAt: 'return self.lastIndexOf(aString,position-1) + 1')
#(asNativeName 'self.asJsName()')
#(asPrivateSelector 'result = self
	.replace(/:/g,"_")
	.replace(/[+]/g,"_plus")
	.replace(/-/g,"_minus")
	.replace(/[*]/g,"_star")
	.replace(/[/]/g,"_slash")
	.replace(/[|]/g,"_bar")
	.replace(/[&]/g,"_amp")
	.replace(/>/g,"_gt")
	.replace(/</g,"_lt")
	.replace(/=/g,"_eq")
	.replace(/,/g,"_comma")
	.replace(/[@]/g,"_at")
	.replace(/~/g,"_no")
	.replace(/\\/g,"_bk")')
#(asNative: '
	var bar=unescape("%%22");
	if(self.search(/^[a-zA-Z0-9_:.$ ]*$/) == -1)
		return "unescape(" +bar +escape(self) +bar +")";
	else
		return bar +self +bar')
#(replace:with: 'self')
#(replaceRegexp:with: 'self.replace(aRegexp,aString)')
#(tokenize: 'return self.split(aString)')
#(match: 'self.search(aRegexp) !!= -1')
#(isLetter 'self.search("[A-Za-z]$")==0')
#(isDigit 'self.search("[0-9]$")==0')
#(isAlphaNumeric 'self')
#(isUpperCase 'self')
#(isLowerCase 'self')
#(asUppercase 'self.toUpperCase()')
#(asLowercase 'self.toLowerCase()')
#(asNumber 'Number(self)')
#(> 'String(self) > aString')
#(< 'String(self) < aString')
#(>= 'String(self) >= aString')
#(<= 'String(self) <= aString')
#(asJSONObject 'self')
#(isCharacter 'self.length == 1')
#(asciiValue 'self.charCodeAt(0)')
#(isSeparator '(new RegExp("\\s|\\u0085")).test(self)')
))
#('String class' #(
#(fromString: 'return new self.$fn(aString)')
#(new	'""')
#(cr 	'"\r"')
#(lf 	'"\n"')
#(crlf 	'"\r\n"')
#(space	'" "')
#(tab	'"\t"')
#(zero	'"\0"')
#(fromCharCode: '
	if (anArrayOrNumber.isNumber()) return String.fromCharCode(anArrayOrNumber);
	return String.fromCharCode.apply(null,anArrayOrNumber.asArray());')
#(basicNew 'self')
))
#(Array #(
#(size 'self.length')
#(at: 'var value = self[anIndex - 1];
	if(value === undefined) { return self.objectNotFoundAt_(anIndex); }
	return value')
#(at:ifAbsent: 'var value = self[anIndex - 1];
	if(value === undefined) { return aBlock.value(); }
	return value')
#(at:put: 'return self[anIndex - 1] = anObject')
#(do: 'for(var i=0;i<self.length;i++){fn(self[i]);}')
#(addAll: 'self.push.apply(self, aCollection.asArray())')
#(shallowCopy 'result = self.slice(0)')
#(copyWithAll: 'return self.concat(aCollection.asArray())')
#(, 'return self.concat(aCollection.asArray())')
#(join: 'self.join(aString)')
#(sort: 'if (nil.isNil_(aBlock)) { self.sort(); return self }
	var criteria = function(a, b) {
		if(fn(a,b)) {return -1} else {return 1}
	}
	self.sort(criteria)')
#(remove:ifAbsent: 'for(var i=0;i<self.length;i++) {
		if(self[i] == anObject) {
			self.splice(i,1); return anObject;
		}
	}; return aBlock.value()')
#(add: 'self.push(anObject)')
#(addFirst: 'self.splice(0,0,anObject)')
#(addLast: 'self.push(anObject)')
#(removeAll 'self.splice(0,self.size())')
#(removeFirst 'self.splice(0,1)')
#(removeIndex: 'self.splice(index-1,1)')
#(removeLast 'self.pop()')
#(atAllPut: 'for(var i=0;i<self.length;i++){ self[i] = anObject; }')
#(first 'var result=self[0];
	if (typeof(result)!!="undefined") return result')
#(second 'var result=self[1];
	if (typeof(result)!!="undefined") return result')
#(third 'var result=self[2];
	if (typeof(result)!!="undefined") return result')
#(fourth '	var result=self[3];
	if (typeof(result)!!="undefined") return result')
#(fifth 'var result=self[4];
	if (typeof(result)!!="undefined") return result')
#(sixth 'var result=self[5];
	if (typeof(result)!!="undefined") return result')
#(asJSONObject #('self' 'self'))
#(json #('self' 'self'))
))
#('Array class' #(
#(basicNew: 'new Array(size)')
#(with: 'return [ anObject ]')
#(with:with: 'return [ firstObject, secondObject ]')
#(with:with:with: 'return [ firstObject, secondObject, thirdObject ]')
#(with:with:with:with: 'return [ firstObject, secondObject, thirdObject, fourthObject ]')
#(withAll: 'return [].concat(aCollection.asArray())')
#(slice: 'Array.prototype.slice.call(anIterable)')
))
#(ComposedAssociations #(
#(hashObject '{}')
))
#(HashedSet #(
#(initializeHash '{}')
#(privateAdd: 'self')
#(privateRemove: 'delete self["@hash"][cookie]')
#(includes: 'Object.prototype.hasOwnProperty.call(self["@hash"],cookie)')
))
#(KeyedCollection #(
#(newHash '{}')
))
#('KeyedCollection class' #(
#(@ 'parts=Object.getOwnPropertyNames(jsObject)')
))
#('Dictionary class' #(
#(basicNew 'self')
))
#(PoolDictionary #(
#(validKey: 'self')
#(at:put: 'self')
#(removeKey:ifAbsent: 'delete self["@hash"][aKey]')
#(at:ifAbsent: 'var found = Object.prototype.hasOwnProperty.call(self["@hash"],aKey);
	if(found) return self["@hash"][aKey];
	return aBlock.value()')
#(includesKey: 'Object.prototype.hasOwnProperty.call(self["@hash"],aKey)')
#(keys 'return Object.getOwnPropertyNames(self["@hash"])')
))
#('RegularExpression class' #(
#(fromString:flag: 'new RegExp(aString, anotherString)')
))
#(Exception #(
#(captureStack 'self')
#(signal 'throw(self)')
))
#(ClassBuilder #(
#(addSubclassOf:named:instanceVariableNames: '
	smalltalk.addClass(aString, aClass, aCollection);
	return smalltalk[aString]')
))
#(StringStream #(
#(peek 'return self["@collection"].charAt(self.position())||nil')
#(peekForAll: 'var start = self.position();
	for (var i=0;i<aCollection.length; i++){
		if (self["@collection"].charCodeAt(start+i) !!= aCollection.charCodeAt(i)) return false;
	}')
))
#(Random #(
#(next 'return Math.random()')
))
#(Rectangle #(
#(printString 'self')
))
	) ! !

! JSEncoder class methodsFor: #inlines !
initializeNativeInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self nativeInlines into: pool ! !

! JSEncoder class methodsFor: #inlines !
nativeInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(NativeObject #(
#(jsObject 'Object')
#(ownPropertyNames 'return Object.getOwnPropertyNames(self["@handle"])')
))
#('NativeObject class' #(
#(undefined 'undefined')
#(codeToNumber '" || 0"')
#(newInstanceOf:withAll: 'self')
))
#(NativeHandle #(
#(hasOwnProperty: 'Object.prototype.hasOwnProperty.call(self["@handle"],aName)')
#(delete: 'delete self["@handle"][aName]')
))
#('NativeLibraryObject class' #(
#(basicLoad: 'load(url)')
))
#(NativeLibraryObject  #(
#(newHandleOf: 'result = new aConstructor()')
#(newHandleOf:with: 'result = new aConstructor(arg1)')
#(newHandleOf:with:with: 'result = new aConstructor(arg1,arg2)')
#(newHandleOf:with:with:with: 'result = new aConstructor(arg1,arg2,arg3)')
#(newHandleOf:with:with:with:with: 'result = new aConstructor(arg1,arg2,arg3,arg4)')
#(newHandleOf:with:with:with:with:with: 'result = new aConstructor(arg1,arg2,arg3,arg4,arg5)')
#(newHandleOf:with:with:with:with:with:with: 'result = new aConstructor(arg1,arg2,arg3,arg4,arg5,arg6)')
#(newHandleOf:with:with:with:with:with:with:with: 'result = new aConstructor(arg1,arg2,arg3,arg4,arg5,arg6,arg7)')
#(newHandleOf:with:with:with:with:with:with:with:with: 'result = new aConstructor(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8)')
))
#(Array #(
#(addAllLua: 'self.addAll_(natives)')
))
#('Array class' #(
#(@@ 'self')
#(@ 'result.addAll_(natives)')
#(basicNew 'self')
))
#('Boolean class' #(
#(@ '
	if (aBOOL === true) return true;
	if (aBOOL === 1) return true;
	if ((1 & aBOOL) + 0 == 1) return true')
))
#(Number #(
#(int32 'self')
#(int64 'self')
#(float32 'self')
#(float64 'self')
))
	) ! !

! JSEncoder class methodsFor: #inlines !
initializeToolsInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self toolsInlines into: pool ! !

! JSEncoder class methodsFor: #inlines !
toolsInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(NativeReader #(
#(initializeObject '{}')
))
#(String  #(
#(asComment #('"/* "' 	'"*/ "' 	'"// "'))
))
#(MethodReferences #(
#(dot		'"."')
#(localDef	'"var "')
#(normalSendStarter:	'self')
#(superSendPatternOf:	'self.superJSSendPatternOf_(aSelector)')
#(compiledSourceStart	'"{"')
))
#(Class #(
#(hackDNU 'mth.fn_(function(){
		this.print_("//DNU "+selector);
		this.doesNotUnderstand_(
			smalltalk.Message.receiver_selector_arguments_
				(this,selector,[].slice.call(arguments)))}
		);')
))
	) ! !

! JSEncoder class methodsFor: #inlines !
initializeSnapshotInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self snapshotInlines into: pool ! !

! JSEncoder class methodsFor: #inlines !
snapshotInlines
	" Private - Return the inlines for methods in the category. "

	^#(
#(Smalltalk #(
#(exit	'quit()')
))
#(Snapshot #(
#(bootHeader	'""')
#(bootFooter	'""')
#(launchExpression '"smalltalk.launchImage()"')
))
	) ! !

! JSEncoder class methodsFor: #inlines !
initializeOtherInlines: pool
	" Private - Initialize the inlines for methods in the category. "

	^self register: self otherInlines into: pool ! !

! JSEncoder class methodsFor: #inlines !
otherInlines
	" Private - Return the inlines for methods in the category. "

	^#(
"#(ClassName #(
#(selector '...')
))"
	) ! !

"------------------------------------------------------------------------------
	Extension to Kernel to support Javascript encoding
 ------------------------------------------------------------------------------"!

"------------------------------------------------------------------------------
	Development Utilities for Javascript environment
 ------------------------------------------------------------------------------"!

"------------------------------------------------------------------------------
	Native handlers for Javascript
 ------------------------------------------------------------------------------"!

NativeHandle
	subclass: #JS
	category: #JS!
JS comment: '
	Implementation of wrappers to native Javascript objects.
	@2011 Alejandro F. Reimondo
	aleReimondo@smalltalking.net'!

! JS class methodsFor: #instantiation !
supportsLanguage: language
	" Private - Return true if instances of the receiver implement native wrappers for language. "

	^language = #js! !

! JS methodsFor: #accessing !
propertyNames
	" Return the property names of the receiver. "

	{''js'return Object.keys(self.handle())'}! !

! JS methodsFor: #accessing !
prototype
	" Return the prototype of the receiver. "

	{''js'return smalltalk.JS._at(Object.getPrototypeOf(self["@handle"]))'}! !

! JS methodsFor: #accessing !
functionNamesAndArguments
	" Return the functions of the receiver as a collection of tuples #(functionName argumentsArray). "

	| argList |
	argList := [:source|
		((source upTo: $(; upTo: $)) asArrayOfSubstringsSeparatedBy: $,)
			collect: [:arg| arg trimBlanks ]
	].
	^self functions collect: [:tuple| Array with: tuple first with: (argList value: tuple last toString stream) ]! !

! JS methodsFor: #properties !
defineProperty: aName
	value: aValue
	writable: writable
	enumerable: enumerable
	configurable: configurable
	" Define the property named aName in the receiver. "

	^self defineProperty: aName as: ( Array
		with: (Array with: #value with: aValue)
		with: (Array with: #writable with: writable)
		with: (Array with: #enumerable with: enumerable)
		with: (Array with: #configurable with: configurable)
	)! !

! JS methodsFor: #properties !
defineAccessor: aName
	get: aBlock
	set: oneArgumentBlock
	enumerable: enumerable
	configurable: configurable
	" Define the Accessor property named aName in the receiver. "

	^self defineProperty: aName as: ( Array
		with: (Array with: #get with: aBlock)
		with: (Array with: #set with: oneArgumentBlock)
		with: (Array with: #enumerable with: enumerable)
		with: (Array with: #configurable with: configurable)
	)! !

! JS methodsFor: #properties !
defineProperty: aName
	value: aValue
	writable: writable
	enumerable: enumerable
	" Define the property named aName in the receiver. "

	^self defineProperty: aName as: ( Array
		with: (Array with: #value with: aValue)
		with: (Array with: #writable with: writable)
		with: (Array with: #enumerable with: enumerable)
	)! !

! JS methodsFor: #properties !
defineProperty: aName
	value: aValue
	writable: writable
	" Define the property named aName in the receiver. "

	^self defineProperty: aName as: ( Array
		with: (Array with: #value with: aValue)
		with: (Array with: #writable with: writable)
	)! !

! JS methodsFor: #properties !
defineProperty: aName value: aValue
	" Define the property named aName in the receiver. "

	^self defineProperty: aName as: ( Array
		with: (Array with: #value with: aValue)
	)! !

! JS methodsFor: #properties !
defineProperty: aName
	" Define the property named aName in the receiver. "

	^self defineProperty: aName as: Object new! !

! JS methodsFor: #properties !
defineProperty: aName as: descriptor
	" Define the property named aName in the receiver. "

	descriptor isArray ifTrue: [
		^self defineProperty: aName as: (descriptor
			inject: Object new into: [:result :tuple|
				tuple last notNil ifTrue: [
					result basicAt: tuple first put: tuple last
				].
				result
			])
	].

	{''js'return Object.defineProperty(self["@handle"],aName,descriptor)'}! !

! JS methodsFor: #properties !
defineProperties: descriptors
	" Define the properties in the receiver. "

	{''js'return Object.defineProperties(self["@handle"],descriptors)'}! !

! JS methodsFor: #actions !
preventExtensions
	" Set the receiver as non-extensible. "

	{''js'return Object.preventExtensions(self["@handle"])'}! !

! JS methodsFor: #actions !
freeze
	" Prevents the modification of existing property attributes and values, and prevents the addition of new properties. "

	{''js'return Object.freeze(self["@handle"])'}! !

! JS methodsFor: #actions !
seal
	" Prevents the modification of attributes of existing properties, and prevents the addition of new properties. "

	{''js'return Object.seal(self["@handle"])'}! !

! JS methodsFor: #testing !
isExtensible
	" Return true if the receiver is extensible. "

	{''js'return Object.isExtensible(self["@handle"])'}! !

! JS methodsFor: #testing !
isFrozen
	" Return true if the receiver is frozen. "

	{''js'return Object.isFrozen(self["@handle"])'}! !

! JS methodsFor: #testing !
isSealed
	" Return true if the receiver is sealed. "

	{''js'return Object.isSealed(self["@handle"])'}! !

! JS methodsFor: #creation !
create: prototype with: descriptors
	" Return a JS object that has been created with the specified prototype, and that optionally contains specified properties. "

	{''js'return smalltalk.JS._at(Object.create(prototype,descriptors))'}! !

! JS methodsFor: #creation !
createWith: descriptors
	" Return a JS object that has been created with the handle prototype, and that optionally contains specified properties. "

	^self create: handle with: descriptors! !


"------------------------------------------------------------------------------
	Simple library wrapper for Javascript
 ------------------------------------------------------------------------------"!

NativeObject
 subclass: #NativeLibraryObject
 instanceVariableNames: ''
 classVariableNames: ' BindingListeners '
 poolDictionaries: ''
 category: #JS!
NativeLibraryObject comment: '
	Abstract wrapper to javascript external library.
	@2012 Alejandro Reimondo - aleReimondo@smalltalking.net
'!

! NativeLibraryObject class methodsFor: #private !
globalName
	" Private - Return the name of the global to detect/access the library of the receiver. "

	^self subclassResponsibility: #globalName! !

! NativeLibraryObject class methodsFor: #private !
topGlobal
	" Private - Return the object to access the global during detection stage.
	The default implementation return window global (present when running in a web browser).
	This method must be refined by the subclasses requiring other root dependences.
	Can return nil to do not check dependents.
	"

	^Smalltalk nativeObjectAt: #window ifAbsent: [ self error: 'Missing window support' ].! !

! NativeLibraryObject class methodsFor: #private !
requiredFeatures
	" Private - Return the names of the required properties/features to execute the receiver.
	The default implementation do not check dependents.
	This method can be refined by subclasses to ensure features to run the receiver are present in this system.
	"

	^#()! !

! NativeLibraryObject class methodsFor: #private !
classNamed: className
	" Private - Return the class named className in the receiver's library.
	The default implementation get the class from library.
	Subclasses can refine to obtain classes from anywhere
		e.g. global scope with: ^self global: className.
	"

	| library |
	library := self library.
	^library[className]! !

! NativeLibraryObject class methodsFor: #private !
sourceCode
	" Private - Return the (native) source code of the library as a String.
	This method can be implemented by subclasses loading sources from external media or simply returning aString embedded in this method.
	A reference/link to source origin is appreciated by users.
	CAUTION: S8 is MIT, so you MUST embedd here MIT compatible sources.
	"

	^self subclassResponsibility: #sourceCode! !

! NativeLibraryObject class methodsFor: #binding !
libraryBound: aClassName
	" Private - The library of aClassName has just been bound to this system. "

	self bindingListeners do: [:each| each libraryBound: aClassName ]! !

! NativeLibraryObject class methodsFor: #binding !
bindingListeners
	" Private - Return the collection of listeners to library binding notifications of this system. "

	BindingListeners isNil ifTrue: [ BindingListeners := Array new ].
	^BindingListeners! !

! NativeLibraryObject class methodsFor: #builder !
globalName: aName
	" Generate code to return the name of the global to detect/access the library of the receiver. "

	^self class
		implement: #globalName
		comment: 'Private - Return the name of the global to detect/access the library of the receiver.'
		as: '^' ,aName printString
		category: #private! !

! NativeLibraryObject class methodsFor: #builder !
libraryURL: anURL
	" Generate the method to return libraryURL. "

	^self class
		implement: #libraryURL
		comment: 'Private - Return the URL to use when loading library of the receiver.'
		as: '^' ,anURL printString
		category: #private! !

! NativeLibraryObject class methodsFor: #private !
globalClassNameFrom: globalClassName
	" Private - Return the global class name for global globalClassName.
	It must return the (global) name of the constructor used to instantiate instance handles.
	The default action is to return the globalClassName.
	This method can be refined by subclasses to adapt name to (case sensitive) library constructor names.
	"

	^globalClassName! !

! NativeLibraryObject class methodsFor: #builder !
selectorForConstantNamed: aName
	" Private - Return the selector to use for constant aName. "

	aName detect: [:c| c isLowercase ] ifNone: [
		^aName asLowercase asSmalltalkName: false
	].
	^aName asSmalltalkName: false! !

! NativeLibraryObject class methodsFor: #builder !
constants: names
	" Private - Build methods to access library globals. "

	names do: [:each|
		self class
			implement: (self selectorForConstantNamed: each)
			comment: 'Return the value of ',each
			as: '^self library basicAt: #',each
			category: #accessing
	].! !

! NativeLibraryObject class methodsFor: #builder !
buildSubclass: globalClassName arguments: argList typed: typedArguments
	" Private - Build a subclass of the receiver implementing a wrapper to the library. "

	| cls clsName |
	clsName := globalClassName first asUppercase
			,(globalClassName copyFrom: 2 to: globalClassName size).
	clsName := self classToken ,clsName.
	self
		subclass: clsName
		instanceVariableNames: ''
		category: self category.
	(argList isNil or: [ argList = #nil ]) ifTrue: [ ^self ].

	cls := Smalltalk at: clsName.
	cls class
		implement: #globalClassName
		comment: 'Private - Return the internal class name.'
		as: '^self globalClassNameFrom: #' ,globalClassName
		category: #private.
	argList isEmpty ifTrue: [
		cls
			implement: #initializeHandle
			comment: 'Private - Initialize the receiver''s handle.'
			as: 'self handle: (self newHandleOf: self class globalClassName)'
			category: #initialize.
	] ifFalse: [ | selector args current |
		selector := argList inject: '' into: [:total :each| total ,each,': ',each,' ' ].
		args := argList inject: '' into: [:total :each|
			(typedArguments = #all
			  or: [ typedArguments includes: each ])
			  ifTrue: [ current := ' handle ' ]
			  ifFalse: [ current := ' ' ].
			total ,'with: ',each, current
		].
		cls class
			implement: selector
			comment: 'Return an instance of the receiver.'
			as: '^self basicNew ',selector
			category: #instantiation.
		cls
			implement: selector
			comment: 'Private - Initialize the receiver.'
			as: 'self handle: (self newHandleOf: self class globalClassName ' ,args ,')'
			category: #initialize.
	].! !

! NativeLibraryObject class methodsFor: #accessing !
global: globalName
	" Private - Return the native library global object (or nil). "

	| result |
	result := Smalltalk nativeObjectAt: globalName ifAbsent: [ nil ].
	result notNil ifTrue: [ ^result ].
	self isValid ifFalse: [ ^nil ].
	self libraryLoaded: self loadLibrary.
	^Smalltalk nativeObjectAt: globalName ifAbsent: [ nil ]! !

! NativeLibraryObject class methodsFor: #validation !
library
	" Check support to run this framework and try to load the library implementation.
	Return the native library object if success.
	"

	^self global: self globalName! !

! NativeLibraryObject class methodsFor: #validation !
libraryOrNil
	" Quick Check - Return true if the library is already bound. "

	^Smalltalk nativeObjectAt: self globalName ifAbsent: [ nil ]! !

! NativeLibraryObject class methodsFor: #private !
nativeClass
	" Private - Return the wrapper for native binding of the receiver.
	The default result is a NativeHandle instance wrapping the receiver's library,
	 so static(class side) builders generate bindings accessing receiver's library.
	"

	^NativeHandle @ self library! !

! NativeLibraryObject class methodsFor: #private !
libraryURLs
	" Private - Return a collection of URL to load library of the receiver.
	Return an empty collection, the subclasses can complement addresses to load library from external resources.
	"

	^(Array with: self libraryURL) select: [:each| each notNil ]! !

! NativeLibraryObject class methodsFor: #private !
libraryURL
	" Private - Return the URL to use when loading library of the receiver.
	Default implementation do not provide support to load library.
	"

	^nil! !

! NativeLibraryObject class methodsFor: #loading !
basicLoad: url
	" Private - Try to load the library at url. "

	{''js'load(url)'}! !

! NativeLibraryObject class methodsFor: #private !
onLoad: aBlock
	" Private - Load the library of the receiver and evaluates aBlock. "

	| dom success |
	self libraryOrNil notNil ifTrue: [ ^aBlock value: self library ].
	success := [
		NativeLibraryObject libraryBound: self name.
		aBlock value: self library
	].
	dom := Smalltalk at: #DOM ifAbsent: [
		self libraryURLs do: [:each| self basicLoad: each ].
		^success value
	].
	^dom document
		loadAll: self libraryURLs
		onLoad: [ success value ]! !

! NativeLibraryObject class methodsFor: #private !
loadLibrary
	" Private - Load the library of the receiver.
	Default implementation load the library from remote URL or sourceCode.
	"

	self libraryURLs notEmpty ifTrue: [ ^self onLoad: [:ignored| ] ].
	^Smalltalk current
		eval: self sourceCode
		as: (Array with: #loadLibrary with: self name)! !

! NativeLibraryObject class methodsFor: #private !
libraryLoaded: loadResult
	" Private - The library of the receiver has been loaded.
	Default implementation do nothing.
	This method can be refined by subclasses to perform postLoading actions.
	The loadResult argument is the result of the just evaluated load process.
	"! !

! NativeLibraryObject class methodsFor: #validation !
isValid
	" Return true if the receiver is valid in this (browser) context.
	Generate error if the minimum support to run the library is not available.
	"

	| global |
	global := self topGlobal.
	global isNil ifTrue: [ ^true ]. "always valid"
	self requiredFeatures do: [:each|
	  global[each] isNil ifTrue: [
	    self error: 'Missing support for ',each
	  ].
	].
	^true! !

! NativeLibraryObject methodsFor: #private !
library
	" Private - Return the library of the receiver. "

	^self class library! !

! NativeLibraryObject methodsFor: #private !
classNamed: className
	" Private - Return the class named className in the receiver's library. "

	^self class classNamed: className! !

! NativeLibraryObject methodsFor: #private !
newHandleOf: className
	" Private - Return a new instance of the class named className in the receiver's library. "

	| result aConstructor |
	aConstructor := self classNamed: className.
	{''js'result = new aConstructor()'}.
	^result! !

! NativeLibraryObject methodsFor: #private !
newHandleOf: className with: arg1
	" Private - Return a new instance of the class named className in the receiver's library. "

	| result aConstructor |
	aConstructor := self classNamed: className.
	{''js'result = new aConstructor(arg1)'}.
	^result! !

! NativeLibraryObject methodsFor: #private !
newHandleOf: className with: arg1 with: arg2
	" Private - Return a new instance of the class named className in the receiver's library. "

	| result aConstructor |
	aConstructor := self classNamed: className.
	{''js'result = new aConstructor(arg1,arg2)'}.
	^result! !

! NativeLibraryObject methodsFor: #private !
newHandleOf: className with: arg1 with: arg2 with: arg3
	" Private - Return a new instance of the class named className in the receiver's library. "

	| result aConstructor |
	aConstructor := self classNamed: className.
	{''js'result = new aConstructor(arg1,arg2,arg3)'}.
	^result! !

! NativeLibraryObject methodsFor: #private !
newHandleOf: className with: arg1 with: arg2 with: arg3 with: arg4
	" Private - Return a new instance of the class named className in the receiver's library. "

	| result aConstructor |
	aConstructor := self classNamed: className.
	{''js'result = new aConstructor(arg1,arg2,arg3,arg4)'}.
	^result! !

! NativeLibraryObject methodsFor: #private !
newHandleOf: className with: arg1 with: arg2 with: arg3 with: arg4 with: arg5
	" Private - Return a new instance of the class named className in the receiver's library. "

	| result aConstructor |
	aConstructor := self classNamed: className.
	{''js'result = new aConstructor(arg1,arg2,arg3,arg4,arg5)'}.
	^result! !

! NativeLibraryObject methodsFor: #private !
newHandleOf: className with: arg1 with: arg2 with: arg3 with: arg4 with: arg5 with: arg6
	" Private - Return a new instance of the class named className in the receiver's library. "

	| result aConstructor |
	aConstructor := self classNamed: className.
	{''js'result = new aConstructor(arg1,arg2,arg3,arg4,arg5,arg6)'}.
	^result! !

! NativeLibraryObject methodsFor: #private !
newHandleOf: className with: arg1 with: arg2 with: arg3 with: arg4 with: arg5 with: arg6 with: arg7
	" Private - Return a new instance of the class named className in the receiver's library. "

	| result aConstructor |
	aConstructor := self classNamed: className.
	{''js'result = new aConstructor(arg1,arg2,arg3,arg4,arg5,arg6,arg7)'}.
	^result! !

! NativeLibraryObject methodsFor: #private !
newHandleOf: className with: arg1 with: arg2 with: arg3 with: arg4 with: arg5 with: arg6 with: arg7 with: arg8
	" Private - Return a new instance of the class named className in the receiver's library. "

	| result aConstructor |
	aConstructor := self classNamed: className.
	{''js'result = new aConstructor(arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8)'}.
	^result! !