" S8 kernel implementation - minimal implementation for comfortable smalltalk experience
Copyright (C) 2012-today Alejandro F. Reimondo <aleReimondo@smalltalking.net>
MIT License - See license.txt file
In case of questions and bug reports please join us at S8 Facebook group
	https://www.facebook.com/groups/s8smalltalk/
More resources can be dynamically loaded/assembled from U8 service at
	http://u8.smalltalking.net
Read more on application, platforms, execution machinery and tests in our sWiki spaces
	swiki:U8
	swiki:s8-media
S8 is a smalltalk implementation for modern/social software development,
 it is not only the source code stored in files, it is about experiences
 and activities that can start now running on your mind sharing
 your development on virtual objects with others.
"!

nil
 subclass: #Object
 instanceVariableNames: ''
 category: #Kernel!

Object
 subclass: #Exception
 instanceVariableNames: 'messageText'
 category: #Kernel!

Exception
 subclass: #Error
 instanceVariableNames: ''
 category: #Kernel!

! Object methodsFor: #system !
smalltalk
	" Private - Return the hosting context of the receiver.
	The smalltalk host is used for evaluations in the receiver. "

	^self basicAt: '$smalltalk' ifAbsent: [ Smalltalk current ]! !

! Object methodsFor: #environment !
nativeLanguage
	" Private - Return the native language of the host. "

	^self smalltalk
		basicAt: '$nativeLanguage'
		ifAbsent: [ #js ]! !

! Object methodsFor: #debug !
halt
	" Private - Halt execution entering debugger. "

	{'' doc'Enter lowlevel debugger (or do nothing)'}! !

! Object methodsFor: #development !
note: aString
	" Print a note (aString) to the user. "

	self print: aString asString asComment! !

! Object methodsFor: #development !
notImplementedYet
	" Private - Halt execution and/or generate an error reporting implementation incomplete. "

	self error: 'Not implemented. Yet' halt! !

! Object methodsFor: #development !
notImplementedYet: aSelector
	" Private - Halt execution and/or generate an error reporting implementation incomplete. "

	self error: (aSelector asLiteral ,' Not implemented. Yet') halt! !

! Object methodsFor: #comparing !
= anObject
	" Return true if the receiver is equal to argument. "

	^self == anObject! !

! Object methodsFor: #comparing !
~= anObject
	" Return true if the receiver is not equal to anObject. "

	^(self = anObject) == false! !

! Object methodsFor: #initialize !
initialize
	" Private - Initialize the receiver.
	The default implementation do nothing.
	This method can be refined by subclasses to support creation time initialization of collaborators.
	"! !

! Object methodsFor: #identity  !
yourself
	" Return the receiver. "

	^self! !

! Object methodsFor: #species !
species
	" Return the species to instantiate objects like the receiver. "

	^self class ! !

! Object methodsFor: #accessing !
class
	" Return the class of the receiver. "

	^self basicAt: '$klass' ifAbsent: [ self klass ]! !

! Object methodsFor: #accessing !
size
	" Return the size of the receiver.
	The default implementation return zero.
	This method can be refined by subclasses to support indexed access.
	"

	^0! !

! Object methodsFor: #copying !
copy
	" Return a copy of the receiver. "

	^self shallowCopy! !

! Object methodsFor: #copying !
shallowCopy
	" Return a deep copy of the receiver. "

	| result |
	result := self class basicNew.
	self class allInstVarNamesDo: [:iv| | v |
		v := self basicAt: $@,iv.
		v notNil ifTrue: [ result basicAt: $@,iv put: v ]
	].
	^result! !

! Object methodsFor: #copying !
deepCopyOf: what
	" Private - Return a deep copy of what. "

	what isNil ifTrue: [ ^what ].
	(self isObject: what) ifTrue: [ ^what deepCopy ].
	^what! !

! Object methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	| result |
	result := self class basicNew.
	self class allInstVarNamesDo: [:iv| | v |
		v := self deepCopyOf: (self basicAt: $@,iv).
		v notNil ifTrue: [ result basicAt: $@,iv put: v ]
	].
	^result! !

! Object methodsFor: #converting !
-> anObject
	" Return an association of the receiver to anObject. "

	^Association key: self value: anObject! !

! Object methodsFor: #converting !
asString
	" Return the receiver represented as String. "

	^self printString! !

! Object methodsFor: #converting !
toString
	" Return the receiver as a String. "

	{'self'}.
	^'{st:a',self class name,$}! !

! Object methodsFor: #converting !
toString: anObject
	" Return anObject converted to a String. "

	anObject isNil ifTrue: [ ^#nil ].
	(self isObject: anObject) ifTrue: [ ^anObject toString ].
	^{'' doc'Native conversion (default goes to error)'}! !

! Object methodsFor: #converting !
asNative: lang
	" Return the native representation of the receiver in target language lang.
	The default implementation return the receiver as a String.
	This method must be refined by subclasses implementing native literal representation.
	"

	^self asString! !

! Object methodsFor: #converting !
asNative
	" Return the native representation of the receiver.
	The default implementation return the receiver as a String.
	This method must be refined by subclasses implementing native literal representation.
	"

	^self asNative: self nativeLanguage! !

! Object methodsFor: #json !
asJSON
	" Return the JSON representation of the receiver. "

	| json |
	json := Smalltalk nativeObjectAt: #JSON.
	^json #stringify: self asJSONObject! !

! Object methodsFor: #converting !
escapedCode
	" Return the code to reproduce receiver from code. "

	^self asNative! !

! Object methodsFor: #perform !
perform: aSymbol
	" Return the result of sending message aSymbol to the receiver. "

	^self perform: aSymbol withArguments: #()! !

! Object methodsFor: #perform !
perform: aSymbol with: argument
	" Return the result of sending message aSymbol to the receiver. "

	^self perform: aSymbol withArguments: (Array with: argument)! !

! Object methodsFor: #perform !
perform: aSymbol withArguments: aCollection
	" Return the result of sending message aSymbol to the receiver. "

"	Warning: To do NOT check arguments size is a feature
		if we do strict matching U8 tools will break because they are using the feature.
	aSymbol numberOfArguments = aCollection size ifFalse: [
		^self error: 'Invalid arguments size'
	].
"	^self basicPerform: aSymbol asSelector withArguments: aCollection asArray! !

! Object methodsFor: #testing !
isObject: anObject
	" Return true if anObject is a smalltalk object.
	Return false if the object is a native object.
	"

	{'' doc'Native detection (default goes to error)'}! !

! Object methodsFor: #testing !
isNumber: anObject
	" Return true if anObject is a number. "

	^(self isObject: anObject) and: [ anObject isNumber ]! !

! Object methodsFor: #accessing !
basicAt: aString
	" Return the property of the receiver known as aString (or nil). "

	| result |
	result := self[aString].
	result isNil ifTrue: [ ^nil ].
	^result! !

! Object methodsFor: #accessing !
basicAt: aString ifAbsent: aBlock
	" Return the property of the receiver known as aString (or the result of evaluating aBlock). "

	^self[aString] ifUndefined: [ aBlock value ]! !

! Object methodsFor: #accessing !
basicAt: aString ifAbsentPut: aBlock
	" Return the property of the receiver known as aString (or set it with the result of evaluating aBlock). "

	^self basicAt: aString ifAbsent: [ self basicAt: aString put: aBlock value ]! !

! Object methodsFor: #accessing !
instVarAt: instVarName
	" Return the internal collaborator of the receiver known as instVarName. "

	(self isNumber: instVarName) ifTrue: [
		^self instVarAt: (self class allInstVarNames at: instVarName)
	].
	^self basicAt: $@ ,instVarName! !

! Object methodsFor: #accessing !
instVarAt: instVarName put: anObject
	" Set the internal collaborator of the receiver known as instVarName to anObject. "

	(self isNumber: instVarName) ifTrue: [
		^self instVarAt: (self class allInstVarNames at: instVarName) put: anObject
	].
	^self basicAt: $@ ,instVarName put: anObject! !

! Object methodsFor: #accessing !
basicAt: aString put: anObject
	" Set the property of the receiver known as aString. "

	^self[aString] := anObject! !

! Object methodsFor: #error !
error: aString
	" Return the result of managing error(aString) in the receiver. "

	^Error signal: aString! !

! Object methodsFor: #error !
subclassResponsibility
	" Private - The concrete implementation on the receiver is not complete. "

	^self subclassResponsibility: nil! !

! Object methodsFor: #error !
implementedBySubclass
	" Private - The concrete implementation on the receiver is not complete. "

	^self subclassResponsibility! !

! Object methodsFor: #error !
shouldNotImplement
	" The receiver received a message that is not expected. "

	^self shouldNotImplement: nil! !

! Object methodsFor: #error !
invalidMessage
	" The receiver received a message that is invalid (not expected for this object). "

	^self invalidMessage: nil! !

! Object methodsFor: #error !
subclassResponsibility: aSelector
	" Private - The concrete implementation on the receiver is not complete. "

	^self halt error:
		 'Implementation of'
		,(aSelector isNil ifTrue: [' '] ifFalse: [' #',aSelector])
		,' is a responsibility of the subclass(',self class name,').'! !

! Object methodsFor: #error !
implementedBySubclass: aSelector
	" Private - The concrete implementation on the receiver is not complete. "

	^self subclassResponsibility: aSelector! !

! Object methodsFor: #error !
shouldNotImplement: aSelector
	" The receiver received a message that is not expected. "

	^self error:
		 'Message'
		,(aSelector isNil ifTrue: [' '] ifFalse: [' #',aSelector])
		,' should not be implemented for instances of ', self class name! !

! Object methodsFor: #error !
invalidMessage: aSelector
	" The receiver received a message that is invalid (not expected for this object). "

	^self error: 
		 'Message'
		,(aSelector isNil ifTrue: [' '] ifFalse: [' #',aSelector])
		,' should not be sent to instances of ', self class name! !

! Object methodsFor: #private !
smalltalkErrorOf: aNativeException
	" Private, Internal - Return anException associated with a native exception. "

	| result |
	(self isObject: aNativeException) ifTrue: [
		(aNativeException isKindOf: Exception) ifTrue: [ ^aNativeException ].
		^Error new messageText: aNativeException
	].
	result := aNativeException basicAt: #smalltalkError.
	result notNil ifTrue: [ ^result ].
	^Error new
		messageText: aNativeException#description || aNativeException#message;
		nativeException: aNativeException;
		yourself! !

! Object methodsFor: #inspection !
inspect
	" Open an inspector on the receiver. "

	^self basicInspect! !

! Object methodsFor: #inspection !
basicInspect
	" Private - Open an inspector on the receiver. "

	(self smalltalk includesKey: #NativeHandle) ifTrue: [ "print native representation"
		^self print: NativeHandle @> self
	].
	^self! !

! Exception methodsFor: #private !
captureStack
	" Private, Internal - Capture the native stack of the receiver. "

	{'self' doc'Must return the native stack'}! !

! Exception methodsFor: #private !
nativeStack
	" Private, Internal - Return the native stack of the receiver (or nil).
	Note that the native stack can be nativelly captured during signal.
	"

	^self basicAt: '@stack'! !

! Exception methodsFor: #private !
nativeException
	" Private, Internal - Return the native exception associated with the receiver (or nil). "

	^self basicAt: '@nativeException'! !

! Exception methodsFor: #private !
nativeException: nex
	" Private, Internal - Set the native exception associated with the receiver. "

	^self basicAt: '@nativeException' put: nex! !

! Exception methodsFor: #stack !
stack
	" Private - Return a stack trace string (or an empty string). "

	| nex stk |
	stk := self nativeStack.
	stk size > 0 ifTrue: [ ^stk ].
	nex := self nativeException.
	nex isNil ifTrue: [ ^'' ].
	stk := nex basicAt: #stack.
	^stk isNil ifTrue: [''] ifFalse: [stk]! !

! Exception methodsFor: #converting !
toString
	" Return the receiver as a String. "

	| txt stk |
	txt := self messageText.
	txt isNil ifTrue: [ txt := '{st: ' ,self description ,$} ].
	stk := self stack.
	stk size = 0 ifTrue: [ ^txt ].
	^'' writeStream
		nextPutAll: txt; cr;
		nextPutAll: stk;
		contents! !

! Object methodsFor: #printing !
printString
	" Return a printable representation of the receiver. "

	^'a ', self class name! !

! Object methodsFor: #testing !
isKindOf: aClass
	" Return true if the receiver is kind of aClass. "

	^self class inheritsFrom: aClass! !

! Object methodsFor: #testing !
isMemberOf: aClass
	" Return true if the receiver is member of aClass. "

	^self class == aClass! !

! Object methodsFor: #testing !
ifNil: aBlock
	" Return the receiver if it is not nil, or the result of evaluating aBlock (if it is nil). "

	^self! !

! Object methodsFor: #testing !
ifNil: aBlock ifNotNil: anotherBlock
	" Return the result of evaluating aBlock if the receiver is nil or the result of evaluating anotherBlock (if not nil). "

	^anotherBlock value! !

! Object methodsFor: #testing !
ifNotNil: aBlock
	" Return the result of evaluating aBlock if the receiver is not nil, or nil if it is nil. "

	^aBlock value! !

! Object methodsFor: #testing !
ifNotNil: aBlock ifNil: anotherBlock
	" Return the result of evaluating aBlock if the receiver is not nil, or the result of evaluating anotherBlock if it is nil. "

	^aBlock value! !

! Object methodsFor: #testing !
isNil
	" Return true if the receiver is nil. "

	^false! !

! Object methodsFor: #testing !
notNil
	" Return true if the receiver is not nil. "

	^self isNil not! !

! Object methodsFor: #testing !
isClass
	" Return true if the receiver is a Class. "

	^false! !

! Object methodsFor: #testing !
isMetaclass
	" Return true if the receiver is a metaclass. "

	^false! !

! Object methodsFor: #testing !
isNumber
	" Return true if the receiver is a Number. "

	^false! !

! Object methodsFor: #testing !
isString
	" Return true if the receiver is a String. "

	^false! !

! Object methodsFor: #testing !
isArray
	" Return true if the receiver is an Array. "

	^false! !

! Object methodsFor: #testing !
isValuable
	" Return true if the receiver can be evaluated. "

	^false! !

! Object methodsFor: #native !
typeOf: anObject
	" Return the (native) type of anObject. "

	^{'' doc'Get native type (default goes to error, use to be a String instance)'}! !

! Object methodsFor: #native !
callableOf: aBlock
	" Return a (native) callable object for evaluating aBlock. "

	{'self' s8' ^aBlock '}.
	aBlock isNil ifTrue: [ ^aBlock ].
	(self typeOf: aBlock) = #function ifTrue: [ ^aBlock ].
	^[:x :y :z| aBlock value: x value: y value: z ] ! !

! Object methodsFor: #messages !
methodFor: aSymbol 
	" Return the method implementing aSymbol (or nil). "

	^self class compiledMethodFor: aSymbol! !

! Object methodsFor: #messages !
respondsTo: aSymbol 
	" Return true if the receiver respond to message aSymbol. "

	^(self methodFor: aSymbol) notNil! !

! Object methodsFor: #accessing !
basicDelete: aString
	" Private - Basic implementation of deletion of property aString of the receiver. "

	{'' doc'native deletion'}! !

! Object methodsFor: #dnu !
doesNotUnderstand: aMessage
	" Private - Basic implementation of a message sent and not understood by the receiver. "

	^MessageNotUnderstood new
		receiver: self;
		message: aMessage;
		signal! !

! Object methodsFor: #dnu !
dnu: selector withArguments: anArray
	" Private - Basic implementation of a message sent and not understood by the receiver. "

	^self doesNotUnderstand: (Message selector: selector arguments: anArray)! !

! Object methodsFor: #private !
basicPerform: jsSelector withArguments: anArray
	" Private - Basic implementation of #perform... "

	| fn |
	{'self'}.
	fn := self basicAt: jsSelector ifAbsent: [
		^self dnu: jsSelector asNativeSelector withArguments: anArray
	].
	^{'' doc'dynamic message send (fn can be native function to call)'}! !

! Object methodsFor: #json !
json
	" Return the receiver as a hash object. "

	^self asJSONObject! !

! Object methodsFor: #json !
asJson: anObject
	" Return anObject as a hash object. "

	anObject isNil ifTrue: [ ^nil json ].
	^(self isObject: anObject)
		ifTrue: [ anObject json ]
		ifFalse: [ anObject ]! !

! Object methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	| result |
	result := {'' doc'get empty JSON collector'}.
	self	jsonFieldsDo: [:field :value|
		(value notNil or: [ self jsonExportNil ]) ifTrue: [
			result basicAt: field put: (self asJson: value)
		].
	]	ignoring: self jsonHiddenNames.
	^result! !

! Object methodsFor: #converting !
jsonHiddenNames
	" Private - Return the names of the fields to skip when mapping the receiver to JSON structure. "

	^#()! !

! Object methodsFor: #converting !
jsonFieldsDo: aBlock ignoring: ignored
	" Private - Evaluates a(TwoArguments)Block with the fields to map the receiver to JSON structure. "

	^self class allInstVarNamesDo: [:each|
		(ignored notNil and: [ ignored includes: each ]) ifFalse: [
			aBlock value: each value: (self instVarAt: each)
		].
	]! !

! Object methodsFor: #converting !
jsonExportNil
	" Private - Return true if the receiver export nil values. "

	^false! !

! Object methodsFor: #testing !
mustBeBoolean
	" Signal an error if the receiver is not a Boolean. "

	^self error: 'Must be boolean.'! !

! Boolean methodsFor: #testing !
mustBeBoolean
	" Signal an error if the receiver is not a Boolean. "

	^self! !

Object
 subclass: #Smalltalk
 instanceVariableNames: ''
 category: #Kernel!

! Smalltalk class methodsFor: #accessing !
current
	" Return the current instance of the receiver. "

	^#{smalltalk} ! !

! Smalltalk methodsFor: #accessing !
globalAt: aKey
	" Return the global at aKey (or nil). "

	^self basicAt: aKey! !

! Smalltalk methodsFor: #accessing !
globalAt: aKey ifAbsent: aBlock
	" Return the global at aKey (or the result of evaluating aBlock). "

	^self basicAt: aKey ifAbsent: aBlock! !

! Smalltalk methodsFor: #accessing !
rootClasses
	" Return the collection of topLevel classes in the receiver. "

	^self classes select: [:each| each superclass isNil ]! !

! Smalltalk methodsFor: #accessing !
at: aGlobalName
	" Return the global known as aGlobalName in the receiver (or nil). "

	^self basicAt: aGlobalName ifAbsent: [ nil ]! !

! Smalltalk methodsFor: #accessing !
at: aGlobalName ifAbsent: aBlock
	" Return the value at aKey or the result of evaluating aBlock. "

	^self basicAt: aGlobalName ifAbsent: aBlock! !

! Smalltalk methodsFor: #accessing !
at: aGlobalName ifAbsentPut: aBlock
	" Return the global known as aGlobalName in the receiver or the result of evaluating aBlock.
	Set the global to return value if was absent.
	"

	^self basicAt: aGlobalName ifAbsentPut: aBlock! !

! Smalltalk methodsFor: #accessing !
at: aGlobalName put: anObject
	" Set the global known as aGlobalName in the receiver top anObject. "

	^self basicAt: aGlobalName put: anObject! !

! Smalltalk methodsFor: #testing !
includesKey: aGlobalName
	" Return true if the global aGlobalName is defined in the receiver. "

	| none |
	none := {'' doc'absence sentinel'}.
	^((self
		basicAt: aGlobalName ifAbsent: [none])
		== none) not! !

! Smalltalk methodsFor: #accessing !
implementorsOf: aSymbol
	" Return the implementors of aSymbol. "

	| result |
	result := #().
	self rootClasses do: [:each|
		result addAll: (each implementorsOf: aSymbol)
	].
	^result! !

! Smalltalk methodsFor: #accessing !
objectAt: aName ifAbsent: aBlock
	" Return the (global scoped) smalltalk object at aName or the result of evaluating aBlock (if undefined or not an Object instance). "

	| result |
	result := self nativeObjectAt: aName ifAbsent: aBlock.
	^result#yourself notNil ifTrue: [ result ]! !

! Smalltalk methodsFor: #accessing !
jsObjectAt: aName ifAbsent: aBlock
	" OBSOLETE - Return the (global) object at aName or the result of evaluating aBlock (if undefined). "

self note: #OBSOLETE ,' - jsObjectAt:ifAbsent: '.
	^self nativeObjectAt: aName ifAbsent: aBlock! !

! Smalltalk methodsFor: #accessing !
jsObjectAt: aName
	" OBSOLETE - Return the (global) object at aName. "

self note: #OBSOLETE ,' - jsObjectAt: '.
	^self nativeObjectAt: aName ifAbsent: [ self error: 'Undefined ',aName ]! !

! Smalltalk methodsFor: #accessing !
nativeObjectAt: aName ifAbsent: aBlock
	" Return the (global) object at aName or the result of evaluating aBlock (if undefined). "

	^Smalltalk global basicAt: aName ifAbsent: aBlock! !

! Smalltalk methodsFor: #accessing !
nativeObjectAt: aName
	" Return the (global) object at aName. "

	^self nativeObjectAt: aName ifAbsent: [ self error: 'Undefined ',aName ]! !

! Smalltalk methodsFor: #removing !
removeKey: aKey ifAbsent: aBlock
	" Remove the binding at aKey. "

	self basicAt: aKey ifAbsent: [ ^aBlock value ].
	self basicDelete: aKey! !

! Smalltalk methodsFor: #accessing !
definePool: poolName
	" Define the pool dictionary (if it is not defined yet). "

	^self definePool: poolName with: Array new! !

! Smalltalk methodsFor: #accessing !
poolDictionary: aPoolName ifAbsent: aBlock
	" Return the pool dictionary named aPoolName or the result of evaluating aBlock. "

	^self basicAt: aPoolName ifAbsent: [ aBlock value ]! !

! Smalltalk methodsFor: #accessing !
poolDictionary: aPoolName
	" Return the pool dictionary named aPoolName. "

	^self poolDictionary: aPoolName ifAbsent: [
		"self error: 'Missing Pool dictionary: ',aPoolName. "
		self definePool: aPoolName
	]! !

! Smalltalk methodsFor: #compiling !
printDefinePool: poolName with: tuples
	" DEBUG - Dump definition of the pool dictionary. "

	self print: 'smalltalk.definePool_with_("',poolName,'",['.
	tuples do: [:tuple|
		tuple isArray ifTrue: [
			self print: '	[',tuple first printString,',',tuple last asNative,'],'
		] ifFalse: [
			self print: '	',tuple asNative ,','
		]
	].
	self print: '	]);'.! !

! Smalltalk methodsFor: #compiling !
definePool: poolName with: tuples
	" Define the pool dictionary. "

	| dictionary offset |
	dictionary := self basicAt: poolName ifAbsent: [
		self basicAt: poolName put: PoolDictionary new
	].
	offset := 0.
	tuples do: [:each|
		each isArray ifTrue: [
			dictionary at: each first put: each last.
			each last isNumber ifTrue: [ offset := each last + 1 ].
		] ifFalse: [
			each isNumber ifTrue: [ offset := each ]
			ifFalse: [
				dictionary at: each put: offset.
				offset := offset + 1.
			]
		]
	].
	^dictionary! !

! Smalltalk methodsFor: #accessing !
state
	" Return the state of the receiver (or nil). "

	^self basicAt: '$state'! !

! Smalltalk methodsFor: #accessing !
state: aState
	" Set the state of the receiver. "

	self basicAt: '$state' put: aState! !

! Smalltalk methodsFor: #accessing !
withState: aState do: aBlock
	" Set the state of the receiver to aState during execution of aBlock. "

	| old |
	old := self state.
	^[	self state: aState.
		aBlock value
	] ensure: [
		self state: old
	]! !

! Smalltalk methodsFor: #accessing !
@ keyOrPath
	" Return the result of accessing the receiver at key or path (or nil). "

	| result |
	keyOrPath isArray ifFalse: [ ^self class at: keyOrPath ifAbsent: [] ].
	result := keyOrPath inject: self into: [:pool :key | | value |
		(value := pool @ key) isNil ifTrue: [ ^nil ].
		value
	].
	^result! !

! Smalltalk methodsFor: #eval !
eval: anExpression
	" Return the result of evaluating a native expression in global context. "

	{'' doc'Native evaluation hook'}! !

! Smalltalk methodsFor: #eval !
eval: anExpression as: aState
	" Return the result of evaluating a native expression in global context.
	The state of the system change to state while evaluating the expression.
	"

	^self withState: aState do: [ self eval: anExpression ]! !

! Smalltalk methodsFor: #tools !
capture: keyOrPath with: aBlock
	" Capture keyOrPath of the receiver and evaluate aBlock.
	Default action will try to capture the object found and (potentially) evaluate aBlock with the result. "

	aBlock isNil ifTrue: [ ^self capture: keyOrPath with: [:result | result ] ].
	keyOrPath isNil ifTrue: [ ^aBlock evaluateWith: nil ].
	keyOrPath isString ifFalse: [ ^self @ keyOrPath ].
	(keyOrPath startsWith: $!!) ifTrue: [ ^aBlock evaluateWith: keyOrPath fileIn ].
	(keyOrPath startsWith: $@) ifTrue: [ ^self @ (keyOrPath asArrayOfSubstringsSeparatedBy: $@) ].
	(keyOrPath startsWith: $.) ifTrue: [ ^self @ (keyOrPath asArrayOfSubstringsSeparatedBy: $.) ].
	(self respondsTo: keyOrPath) ifFalse: [ ^aBlock evaluateWith: nil ].
	keyOrPath argumentCount = 1 ifTrue: [ ^self perform: keyOrPath with: aBlock ].
	^aBlock evaluateWith: (self perform: keyOrPath)! !

! Smalltalk methodsFor: #tools !
removeSources
	" Remove all development sources of the receiver. "

	self classes do: [:each | each removeSources ]! !

! Smalltalk methodsFor: #json !
readJSON: anObject
	" Return an object read from JSON map object anObject. "

	^self #readJSObject: anObject! !

! Smalltalk methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self! !

! Smalltalk methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self! !

! Smalltalk class methodsFor: #accessing !
at: aGlobalName ifAbsent: aBlock
	" Return the global known as aGlobalName in the receiver or the result of evaluating aBlock. "

	^self current basicAt: aGlobalName ifAbsent: aBlock! !

! Smalltalk class methodsFor: #testing !
includesKey: aGlobalName
	" Return true if the global aGlobalName is defined in the receiver. "

	^self current includesKey: aGlobalName! !

! Smalltalk class methodsFor: #accessing !
at: aGlobalName
	" Return the global known as aGlobalName in the receiver (or nil). "

	^self current basicAt: aGlobalName ifAbsent: [ nil ]! !

! Smalltalk class methodsFor: #accessing !
at: aGlobalName put: anObject
	" Set the global known as aGlobalName in the receiver top anObject. "

	^self current basicAt: aGlobalName put: anObject! !

! Smalltalk class methodsFor: #accessing !
at: aGlobalName ifAbsentPut: aBlock
	" Return the global known as aGlobalName in the receiver or the result of evaluating aBlock.
	Set the global to return value if was absent.
	"

	^self at: aGlobalName ifAbsent: [
		self at: aGlobalName put: aBlock value
	]! !

! Smalltalk class methodsFor: #removing !
removeKey: aKey ifAbsent: aBlock
	" Remove the binding at aKey. "

	^self current removeKey: aKey ifAbsent: aBlock! !

! Smalltalk class methodsFor: #accessing !
@ aGlobalName
	" Return a collection with the objects bound to aGlobalName (or an empty collection if global is not defined).
	It is used in expressions like:
		Smalltalk @ #MyGlobal do: [:global| ... ]
	"

	| binding |
	binding := self at: aGlobalName ifAbsent: [ ^Array new ].
	^Array with: binding! !

! Smalltalk class methodsFor: #accessing !
definePool: poolName with: tuples
	" Define the pool dictionary. "

	^self current definePool: poolName with: tuples! !

! Smalltalk class methodsFor: #accessing !
poolDictionary: aPoolName ifAbsent: aBlock
	" Return the pool dictionary named aPoolName or the result of evaluating aBlock. "

	^self current poolDictionary: aPoolName ifAbsent: aBlock! !

! Smalltalk class methodsFor: #accessing !
objectAt: aName ifAbsent: aBlock
	" Return the (global scoped) smalltalk object at aName or the result of evaluating aBlock (if undefined or not an Object instance). "

	^self current objectAt: aName ifAbsent: aBlock! !

! Smalltalk class methodsFor: #accessing !
jsObjectAt: aName ifAbsent: aBlock
	" OBSOLETE - Return the (global) object at aName or the result of evaluating aBlock (if undefined). "

self note: #OBSOLETE ,' - jsObjectAt:ifAbsent: '.
	^self current nativeObjectAt: aName ifAbsent: aBlock! !

! Smalltalk class methodsFor: #accessing !
jsObjectAt: aName
	" OBSOLETE - Return the (global) object at aName. "

self note: #OBSOLETE ,' - jsObjectAt: '.
	^self nativeObjectAt: aName ifAbsent: [ self error: 'Undefined ',aName ]! !

! Smalltalk class methodsFor: #accessing !
nativeObjectAt: aName ifAbsent: aBlock
	" Return the (global) object at aName or the result of evaluating aBlock (if undefined). "

	^self current nativeObjectAt: aName ifAbsent: aBlock! !

! Smalltalk class methodsFor: #accessing !
nativeObjectAt: aName
	" Return the (global) object at aName. "

	^self nativeObjectAt: aName ifAbsent: [ self error: 'Undefined ',aName ]! !

! Smalltalk class methodsFor: #accessing !
global
	" Return the global context object. "

	^{'' doc'get the global context object'}! !

! Smalltalk class methodsFor: #installing !
install: aBlockClosure as: functionName
	" Install the aBlockClosure as a function in global scope. "

	| old |
	old := self global basicAt: functionName.
	self global basicAt: functionName put: aBlockClosure.
	^old! !

! Smalltalk class methodsFor: #accessing !
classes
	" Return the classes defined in current instance. "

	^self current classes! !

! Smalltalk class methodsFor: #accessing !
rootClasses
	" Return the collection of topLevel classes in the receiver. "

	^self current rootClasses! !

! Smalltalk class methodsFor: #tools !
implementorsOf: aSymbol
	" Return the implementors of aSymbol. "

	^self current implementorsOf: aSymbol! !

! Smalltalk class methodsFor: #environment !
isRunningOn: lang
	" Private - Return true when running hosted by native lang execution environment. "

	^self nativeLanguage = lang! !

! Smalltalk class methodsFor: #tools !
gc
	" Collect garbage. "

	{'' doc'forced full garbage collection'}.
	^false! !

! Smalltalk class methodsFor: #iterating !
allMethodsDo: aBlock
	" Evaluates aBlock for all the methods in the current instance. "

	self classes do: [:each|
		each methodsDo: aBlock.
		each class methodsDo: aBlock.
	]! !

! Smalltalk class methodsFor: #iterating !
allSelectorsDo: aBlock
	" Evaluates aBlock for all the selectors defined in the current instance. "

	| found |
	found := HashedSet new.
	self allMethodsDo: [:mth| | msg |
		msg := mth selector.
		(found includes: msg) ifFalse: [
			aBlock notNil ifTrue: [
				aBlock value: msg
			].
			found add: msg
		]
	].
	^found! !

! Smalltalk class methodsFor: #iterating !
allSelectors
	" Return all the selectors (sorted) defined in the current instance. "

	^(self allSelectorsDo: nil) sorted! !

Object
 subclass: #Behavior
 instanceVariableNames: ''
 category: #Kernel!

! Behavior methodsFor: #instantiation !
new
	" Return an instance of the receiver. "

	^self basicNew initialize! !

! Behavior methodsFor: #instantiation !
new: size
	" Return an instance of the receiver of specified size. "

	^self basicNew: size! !

! Behavior methodsFor: #instantiation !
basicNew: size
	" Return an instance of the receiver of specified size. "

	^self invalidMessage: #basicNew:! !

! Behavior methodsFor: #instantiation !
basicNew
	" Return a new (non-initialized) instance of the receiver. "

	^{'' doc'native instantiation of class'}! !

! Behavior methodsFor: #accessing !
name
	" Return the name of the receiver. "

	^self basicAt: '$className'! !

! Behavior methodsFor: #accessing !
superclass
	" Return the superclass of the receiver. "

	^self basicAt: '$superclass'! !

! Behavior methodsFor: #accessing !
subclasses
	" Return the subclasses of the receiver. "

	^self smalltalk #subclasses: self! !

!Behavior methodsFor: #testing !
allSubclasses
	" Return all the subclasses of the receiver. "

	| result |
	result := Array new.
	self allSubclassesDo: [:each| result add: each ].
	^result! !

!Behavior methodsFor: #testing !
allSubclassesDo: aBlock
	" Evaluates aBlock with all the subclasses of the receiver. "

	self smalltalk #allSubclassesDo: self value: aBlock! !

!Behavior methodsFor: #testing !
withAllSubclassesDo: aBlock
	" Evaluates aBlock with the receiver and all the subclasses. "

	aBlock value: self.
	self allSubclassesDo: aBlock! !

! Behavior methodsFor: #accessing !
withAllSubclasses
	" Return a collection with the receiver and all subclasses. "

	| result |
	result := Array with: self.
	self allSubclassesDo: [:each| result add: each ].
	^result! !

!Behavior methodsFor: #accessing !
allSuperclassesDo: aBlock
	" Evaluate aBlock with the superclasses of the receiver.
	The superclasses are walked in inverse hierarchical order, i.e class Object is last.
	"

	| current |
	current := self superclass.
	[ current notNil ] whileTrue: [
	        aBlock value: current.
		current := current superclass.
	]! !

!Behavior methodsFor: #accessing !
withAllSuperclassesDo: aBlock
	" Evaluate aBlock with the receiver and it's superclasses.
	The superclasses are walked in inverse hierarchical order, i.e class Object is last.
	"

	aBlock value: self.
	^self allSuperclassesDo: aBlock! !

!Behavior methodsFor: #accessing !
withAllSuperclasses
	" Return a collection with the receiver and all superclasses. "

	^(Array with: self) , self allSuperclasses! !

!Behavior methodsFor: #accessing !
allSuperclasses
	" Return a collection of all the superclasses of the receiver.
	The superclasses are in inverse hierarchical order, i.e class Object is last.
	"

	| result |
	result := Array new.
	self allSuperclassesDo: [:each| result add: each ].
	^result! !

! Behavior methodsFor: #testing !
inheritsFrom: aClass
	" Return true if the receiver inherits from aClass. "

	self == aClass ifTrue: [ ^true ].
	self allSuperclassesDo: [:each|
		aClass == each ifTrue: [ ^true ]
	].
	^false! !

! Behavior methodsFor: #accessing !
fn
	" Private - Return the internal function of the receiver (or nil). "

	{'self'}.
	^self basicAt: '$fn' ! !

! Behavior methodsFor: #accessing !
prototype
	" Private - Return the prototype of the receiver (or nil). "

	{'self'}.
	^self fn basicAt: #prototype ! !

! Behavior methodsFor: #accessing !
privateMethods
	" Private, Internal - Return all the methods defined by the receiver. "

	{'self'}.
	^self prototype basicAt: '$methods' ! !

! Behavior methodsFor: #accessing !
constructorName
	" Private - Return the name of the constructor of the receiver (or nil). "

	| result fn |
	fn := self fn.
	result := fn basicAt: #name.
	result notNil ifTrue: [
		result isEmpty ifTrue: [ ^nil ].
		result = '$s8Constructor' ifTrue: [ ^nil ].
		^result
	].
	result := (fn toString upTo: $() asArrayOfSubstrings.
	result size = 2 ifFalse: [ ^nil ].
	^result last! !

! Behavior methodsFor: #accessing !
methodDictionary
	" Return a (pool)dictionary with the methods implemented by the receiver. "

	^PoolDictionary @@ self privateMethods ! !

! Behavior methodsFor: #accessing !
methodsDo: aBlock
	" Evaluates aBlock for all the methods defined by the receiver. "

	self methodDictionary do: [:mth| aBlock value: mth ]! !

! Behavior methodsFor: #accessing !
compiledMethods
	" Return a collection with all the methods defined by the receiver. "

	^self methodDictionary values! !

! Behavior methodsFor: #accessing !
selectorsDo: aBlock
	" Evaluate aBlock for all the selectors implemented by the receiver. "

	^self methodDictionary keys do: aBlock ! !

!Behavior methodsFor: #accessing !
selectors
	" Return the selectors implemented by the receiver. "

	| result |
	result := Array new.
	self selectorsDo: [:each| result add: each ].
	^result! !

!Behavior methodsFor: #accessing !
categorizedSelectors
	" Return the selectors classified by category. "

	| pool |
	pool := PoolDictionary new.
	self methodsDo: [:mth| | category |
		category := mth category.
		category isNil ifTrue: [ category := #undefined ].
		(pool at: category ifAbsentPut: [Array new]) add: mth selector
	].
	^pool keys sorted collect: [:key|
		Array with: key with: (pool at: key) sorted
	]! !

! Behavior methodsFor: #testing !
implements: aSymbol
	" Return true if the receiver implements message aSymbol. "

	^self methodDictionary includesKey: aSymbol ! !

! Behavior methodsFor: #accessing !
compiledMethodAt: aSymbol
	" Return the compiled method at aSymbol implemented by the receiver (or nil). "

	^self methodDictionary at: aSymbol ifAbsent: [] ! !

!Behavior methodsFor: #accessing !
>> aSymbol
	" Return the method at aSymbol (or nil if not implemented by the receiver). "

	^self compiledMethodAt: aSymbol! !

!Behavior methodsFor: #accessing !
compiledMethodFor: aSymbol
	" Return the method at aSymbol searching up in receiver's hierarchy (or nil). "

	| current mth |
	current := self.
	[	mth := current compiledMethodAt: aSymbol.
		current := current superclass.
		mth isNil and: [ current notNil ]
	] whileTrue.
	^mth ! !

!Behavior methodsFor: #accessing !
>>> aSymbol
	" OBSOLETE - Return the method at aSymbol searching up in receiver's hierarchy (or nil if not implemented by the receiver nor its superclasses). "

	^self compiledMethodFor: aSymbol ! !

!Behavior methodsFor: #accessing !
instVarNames
	" Return the instance variable names of the receiver. "

	^self basicAt: '$iVarNames' ifAbsentPut: [#()]! !

!Behavior methodsFor: #accessing !
instVarNamesDo: aBlock
	" Evaluate aBlock with the instance variable names of the receiver. "

	| array |
	array := self basicAt: '$iVarNames'.
	^array notNil ifTrue: [ array do: aBlock ]! !

!Behavior methodsFor: #accessing !
allInstVarNames
	" Return all of the instance variable names defined in the receiver and its superclasses. "

	| result |
	result := self instVarNames.
	self allSuperclassesDo: [ :each | result := each instVarNames , result ].
	^result! !

! Behavior methodsFor: #accessing !
instanceVariableNames
	" Return the instance variable names defined by the receiver. "

	^self instVarNames! !

!Behavior methodsFor: #accessing !
propertyNames
	" Return the property names of the receiver. "

	^self basicAt: '$propertyNames' ifAbsent: [ Array new ]! !

! Behavior methodsFor: #accessing !
propertyNames: anArray
	" Set the property names of the receiver. "

	anArray isNil ifTrue: [ ^self basicDelete: '$propertyNames' ].
	^self basicAt: '$propertyNames' put: anArray! !

! Behavior methodsFor: #accessing !
allPropertyNamesDo: aBlock
	" Evaluates aBlock for the property names defined in the receiver and its superclasses. "

	self propertyNames do: aBlock.
	self allSuperclassesDo: [ :each | each propertyNames do: aBlock ].! !

! Behavior methodsFor: #accessing !
allInstVarNamesDo: aBlock
	" Evaluates aBlock for the instance variable names defined in the receiver and its superclasses. "

	self instVarNamesDo: aBlock.
	self allSuperclassesDo: [ :each | each instVarNamesDo: aBlock ].! !

! Behavior methodsFor: #accessing !
comment
	" Return the comment of the receiver. "

	^self instanceClass basicAt: '$comment' ifAbsent: [ '' ]! !

! Behavior methodsFor: #private !
basicComment: aString
	" Set the comment of the receiver. "

	self instanceClass basicAt: '$comment' put: aString! !

! Behavior methodsFor: #accessing !
comment: aString
	" Set the comment of the receiver. "

	ClassBuilder new
		class: self instanceClass
		comment: aString! !

! Behavior methodsFor: #tools !
removeSources
	" Remove all development sources of the receiver. "

	self instanceClass removeSources! !

! Behavior methodsFor: #accessing !
poolDictionaries
	" Return the poolDictionaries of the receiver. "

	^self instanceClass poolDictionaries! !

! Behavior methodsFor: #accessing !
poolDictionaryNames
	" Return the shared pool names of the receiver. "

	^self instanceClass poolDictionaryNames! !

! Behavior methodsFor: #accessing !
sharedPoolsAndNamesDo: aBlock
	" Execute aBlock for the pool dictionaries known by the receiver. "

	| pool |
	self instanceClass withAllSuperclassesDo: [:each|
		each poolDictionaryNames do: [:aName|
			pool := Smalltalk poolDictionary: aName ifAbsent: [ nil ].
			pool notNil ifTrue: [ aBlock value: aName value: pool ]
		].
	]! !

! Behavior methodsFor: #accessing !
sharedPoolsDo: aBlock
	" Execute aBlock for the pool dictionaries known by the receiver. "

	self sharedPoolsAndNamesDo: [:aName :aPool| aBlock value: aPool ]! !

! Behavior methodsFor: #accessing !
classVariablesDo: aBlock
	" Execute aBlock for the classVariables bindings known by the receiver. "

	| bindings |
	^self instanceClass withAllSuperclassesDo: [:each|
		bindings := each classVariables.
		bindings notNil ifTrue: [
			bindings associationsDo: [:assoc| aBlock value: assoc ]
		]
	]! !

! Behavior methodsFor: #accessing !
classVariableNamesDo: aBlock
	" Execute aBlock for the classVariable names known by the receiver. "

	| bindings |
	^self instanceClass withAllSuperclassesDo: [:each|
		bindings := each classVariables.
		bindings notNil ifTrue: [
			bindings keysDo: [:aName| aBlock value: aName ]
		]
	]! !

! Behavior methodsFor: #accessing !
classVariableAt: aName put: anObject
	" Set the classVariable of the receiver to anObject. "

	^self instanceClass classVariables at: aName put: anObject! !

! Behavior methodsFor: #accessing !
classVariableAt: aName ifAbsent: aBlock
	" Return the value of classVariable named aName. "

	| bindings |
	self instanceClass withAllSuperclassesDo: [:each|
		bindings := each classVariables.
		bindings notNil ifTrue: [
			(bindings includesKey: aName) ifTrue: [
				^bindings at: aName
			]
		]
	].
	^aBlock evaluate! !

! Behavior methodsFor: #accessing !
classVariableAt: aName
	" Return the value of classVariable named aName or generates an error. "

	^self classVariableAt: aName ifAbsent: [
		self error: 'Undefined class variable ',aName ,' in ' ,self name
	]! !

! Behavior methodsFor: #accessing !
pool: poolName variableAt: aName put: anObject
	" Set the pool Variable of the receiver to anObject. "

	| pool |
	pool := Smalltalk
		poolDictionary: poolName
		ifAbsent: [ ^self error: 'Undefined pool variable ',poolName ,' in ' ,self name ].
	^pool at: aName put: anObject! !

! Behavior methodsFor: #loading !
loadRequirementsInto: aCollection for: anEncoder
	" Add the requirements to load the receiver from sources to aCollection.
	Default action do nothing, this method can be refined by the subclasses.
	"

	^self! !

! Behavior methodsFor: #methods !
removeSelectors: aCollection
	" Remove the methods for messages in aCollection, implemented by the receiver. "

	aCollection do: [:each| self removeSelector: each ]! !

! Behavior methodsFor: #private !
removeCompiledMethod: aMethod
	" Remove aMethod from the receiver. "

	self smalltalk #removeMethod: self method: aMethod! !

! Behavior methodsFor: #private !
missingImplementorFor: jsSelector
	" Private - Handle the situation where there is no implementor for jsSelector.
	This can happen when there is no implementation for jsSelector and the message
	 has been dispatched through a superSend.
	"

	^self error: 'Missing implementor for '
		, jsSelector asNativeSelector asLiteral
		,' in ',self name! !

! Behavior methodsFor: #methods !
removeSelector: aSelector
	" Remove the method for message aSelector implemented by the receiver. "

	| method |
	method := self compiledMethodAt: aSelector.
	method notNil ifTrue: [ self removeCompiledMethod: method ].! !

! Behavior methodsFor: #remove !
removeFromSystem
	" Remove the receiver from system. "

	^self instanceClass removeFromSystem ! !

! Behavior methodsFor: #testing !
implementorsOf: aSymbol
	" Return the implementors of aSymbol in the receiver and its subclasses. "

	| result |
	result := #().
	self withAllSubclassesDo: [:each | | method |
		method := each >> aSymbol.
		method notNil ifTrue: [ result add: method ].
		method := each class >> aSymbol.
		method notNil ifTrue: [ result add: method ].
	].
	^result! !

! Behavior methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	^self! !

! Behavior methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self! !

! Behavior methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self! !

! Behavior methodsFor: #converting !
toString
	" Return the receiver as a String. "

	^'{st:',self name,$}! !

Behavior
 subclass: #Class
 instanceVariableNames: ''
 category: #Kernel!

! Class methodsFor: #accessing !
category
	" Return the category of the receiver. "

	^self basicAt: '$category'! !

! Class methodsFor: #accessing !
category: aString
	" Set the category of the receiver. "

	self basicAt: '$category' put: aString! !

! Class methodsFor: #accessing !
classVariables
	" Return the class variables of the receiver. "

	| names |
	names := self basicAt: '$classVariableNames' ifAbsent: [].
	names isNil ifTrue: [ ^self basicAt: '$classVariables' ].
	self basicDelete: '$classVariableNames'.
	self classVariableNames: names.
	^self basicAt: '$classVariables'! !

! Class methodsFor: #accessing !
classVariables: aDictionary
	" Set the class variables of the receiver. "

	(aDictionary isNil or: [ aDictionary isEmpty ]) ifTrue: [
		^self basicDelete: '$classVariables'
	].
	^self basicAt: '$classVariables' put: aDictionary! !

! Class methodsFor: #accessing !
poolDictionaries
	" Return the poolDictionaries of the receiver (anArray or nil). "

	^self basicAt: '$poolDictionaries'! !

! Class methodsFor: #accessing !
poolDictionaries: anArray
	" Set the poolDictionaries of the receiver. "

	anArray notNil ifTrue: [
		anArray isEmpty ifTrue: [
			^self poolDictionaries: nil
		].
		anArray size = anArray asSet size ifFalse: [
			^self error: 'Invalid pool dictionary names ' ,anArray asLiteral
		].
	].
	^self basicAt: '$poolDictionaries' put: anArray! !

! Class methodsFor: #accessing !
classVariableNames
	" Return the class variable names of the receiver. "

	^self classVariables isNil
		ifTrue: [ Array new ]
		ifFalse: [ self classVariables keys ]! !

! Class methodsFor: #accessing !
classVariablesString
	" Return the class variable names of the receiver as aString. "

	^self classVariableNames join: ' '! !

! Class methodsFor: #accessing !
poolDictionaryNames
	" Return the pool dictionary names of the receiver. "

	^self poolDictionaries isNil
		ifTrue: [ Array new ]
		ifFalse: [ self poolDictionaries ]! !

! Class methodsFor: #accessing !
poolDictionariesString
	" Return the poolDictionary names of the receiver as aString. "

	^self poolDictionaryNames join: ' '! !

! Class methodsFor: #accessing !
classVariableNames: classVarNames
	" Set the variable names defined in the receiver. "

	| names dictionary old |
	classVarNames isNil ifTrue: [ ^self classVariables: nil ].
	names := classVarNames isString
		ifTrue: [ classVarNames asArrayOfSubstrings ]
		ifFalse: [ classVarNames ].
	names size = names asSet size ifFalse: [
		^self error: 'Invalid class variable names ' ,names asLiteral
	].
	dictionary := PoolDictionary new.
	old := self classVariables. old isNil ifTrue: [ old := dictionary ].
	names do: [:each|
		dictionary at: each put: (old at: each ifAbsent: [ nil ])
	].
	self classVariables: dictionary.! !

! Class methodsFor: #accessing !
poolDictionaryNames: poolVarNames
	" Set the pool dictionary names known to the receiver. "

	self poolDictionaries: (poolVarNames isString
		ifTrue: [ poolVarNames asArrayOfSubstrings ]
		ifFalse: [ poolVarNames ]).! !

! Class methodsFor: #instantiation !
subclass: aString instanceVariableNames: anotherString
	" Create a subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: anotherString
		category: nil! !

! Class methodsFor: #instantiation !
subclass: aString category: aCategory
	" Create a subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: ''
		category: aCategory! !

! Class methodsFor: #instantiation !
subclass: aString instanceVariableNames: aString2 category: aString3
	" Create a subclass of the receiver. "

	^ClassBuilder new
		superclass: self
		subclass: aString
		instanceVariableNames: aString2
		category: aString3! !

! Class methodsFor: #testing !
isClass
	" Return true if the receiver is a Class. "

	^true! !

! Class methodsFor: #printing !
printString
	" Return a printable representation of the receiver. "

	^self name! !

! Class methodsFor: #accessing !
instanceClass
	" Return the receiver. "

	^self! !

! Class methodsFor: #renaming !
rename: aString
	" Rename the receiver to aString. "

	self smalltalk #renameClass: self to: aString! !

! Class methodsFor: #remove !
removeFromSystem
	" Remove the receiver from system. "

	self subclasses do: [:each| each removeFromSystem ].
	self smalltalk #removeClass: self.
	^true! !

! Class methodsFor: #tools !
removeSources
	" Remove all development sources of the receiver. "

	self class methodsDo: [:each | each removeSources ].
	self methodsDo: [:each | each removeSources ].
	self comment: ''; category: #s8! !

Behavior
 subclass: #Metaclass
 instanceVariableNames: 'instanceClass'
 category: #Kernel!

! Metaclass methodsFor: #accessing !
instanceClass
	" Return the instance class of receiver. "

	^instanceClass! !

! Metaclass methodsFor: #accessing !
instanceVariableNames: aString
	" Set the instance variable names of the receiver. "

	ClassBuilder new
		class: self
		instanceVariableNames: aString.
	^self instanceVariableNames ! !

! Metaclass methodsFor: #testing !
isMetaclass
	" Return true if the receiver is a metaclass. "

	^true! !

! Metaclass methodsFor: #printing !
printString
	" Return a printable representation of the receiver. "

	^self instanceClass name, ' class'! !

! Metaclass methodsFor: #accessing !
classVariables
	" Return the class variables of the receiver. "

	^self instanceClass classVariables! !

Object
 subclass: #CompiledMethod
 instanceVariableNames: ' classField '
 category: #Kernel!

! CompiledMethod methodsFor: #accessing !
source
	" Return the source code of the receiver. "

	^self basicAt: '$source' ifAbsent: ['']! !

! CompiledMethod methodsFor: #accessing !
source: aString
	" Set the source code of the receiver. "

	self basicAt: '$source' put: aString! !

! CompiledMethod methodsFor: #accessing !
category
	" Return the category of the receiver. "

	^self basicAt: '$category' ifAbsent: ['']! !

! CompiledMethod methodsFor: #accessing !
category: aString
	" Set the category of the receiver. "

	self basicAt: '$category' put: aString! !

! CompiledMethod methodsFor: #accessing !
stamp
	" Return the stamp of the receiver. "

	^self basicAt: '$stamp'! !

! CompiledMethod methodsFor: #accessing !
stamp: aString
	" Set the stamp of the receiver. "

	self basicAt: '$stamp' put: aString! !

! CompiledMethod methodsFor: #accessing !
selector
	" Return the selector of the receiver. "

	^self basicAt: '$selector'! !

! CompiledMethod methodsFor: #accessing !
selector: aString
	" Set the selector of the receiver. "

	self basicAt: '$selector' put: aString! !

! CompiledMethod methodsFor: #accessing !
fn
	" Return the native implementation of the receiver (or nil). "

	^self basicAt: '$fn'! !

! CompiledMethod methodsFor: #accessing !
fn: aBlock
	" Set the native implementation of the receiver. "

	self basicAt: '$fn' put: aBlock! !

! CompiledMethod methodsFor: #accessing !
asClosure
	" Private - Return the receiver as a native closure (or nil). "

	{'self'}.
	^self fn! !

! CompiledMethod methodsFor: #accessing !
compiledSource
	" Return the native source code of the receiver (or nil). "

	| closure |
	closure := self asClosure.
	^closure notNil ifTrue: [ closure compiledSourceAt: self ]! !

! CompiledMethod methodsFor: #accessing !
classField
	" Return the implementor of the receiver. "

	^classField! !

! CompiledMethod methodsFor: #accessing !
classField: aClass
	" Set the implementor of the receiver. "

	classField := aClass! !

! CompiledMethod methodsFor: #accessing !
generation
	" Return the generation of the receiver. "

	^self basicAt: '$generation'! !

! CompiledMethod methodsFor: #accessing !
comment
	" Return the comment of the receiver. "

	^(self source stream upTo: $"; upTo: $") trimBlanks! !

! CompiledMethod methodsFor: #printing !
toString
	" Return the receiver as a String. "

	^(classField isNil
		ifTrue: [classField toString]
		ifFalse: [classField name])
		,'>>#' ,self selector toString! !

! CompiledMethod methodsFor: #printing !
printString
	" Return the printable representation of the receiver. "

	{'self'}.
	^self toString! !

! CompiledMethod methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	^self! !

! CompiledMethod methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self! !

! CompiledMethod methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self! !

! CompiledMethod methodsFor: #tools !
removeSources
	" Remove all development sources of the receiver. "

	self source: ''; category: #s8! !

! CompiledMethod methodsFor: #remove !
removeFromSystem
	" Remove the receiver from system. "

	self classField >> self selector = self ifTrue: [
		self classField removeSelector: self selector
	].
	^true! ! 

Object
 subclass: #Boolean
 instanceVariableNames: ''
 category: #Kernel!

! Boolean methodsFor: #comparing !
= anObject
    " Return true if the receiver is equal to argument. "

	{'' doc'boolean test'}! !

! Boolean methodsFor: #control !
ifTrue: aBlock ifFalse: anotherBlock
	" Return the result of evaluating aBlock if the receiver is true, else the result of evaluating anotherBlock. "

	{'' doc'boolean switch'}! !

! Boolean methodsFor: #control !
ifTrue: aBlock
	" Return the result of evaluating aBlock if the receiver is true, else return nil. "

	^self ifTrue: [ aBlock value ] ifFalse: []! !

! Boolean methodsFor: #control !
ifFalse: aBlock
	" Return the result of evaluating aBlock if the receiver is false, else return nil. "

	^self ifTrue: [] ifFalse: [ aBlock value ]! !

! Boolean methodsFor: #control !
ifFalse: aBlock ifTrue: anotherBlock
	" Return the result of evaluating aBlock if the receiver is false, else the result of evaluating anotherBlock. "

	^self ifTrue: anotherBlock ifFalse: [ aBlock value ]! !

! Boolean methodsFor: #control !
and: aBlock
	" Return true if the receiver is true and the result of evaluating aBlock is also true. "

	^self ifTrue: [ aBlock value ] ifFalse: [false]! !

! Boolean methodsFor: #control !
or: aBlock
	" Return true if the receiver is true or the result of evaluating aBlock is true. "

	^self ifTrue: [true] ifFalse: [ aBlock value ]! !

! Boolean methodsFor: #control !
not
	" Return true if the receiver is false. "

	^{'' doc'boolean inversion'}! !

! Boolean methodsFor: #printing !
printString
	" Return a printable representation of the receiver. "

	^self ifTrue: [ #true ] ifFalse: [ #false ]! !

! Boolean methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	{'self'}.
	^self and: [ true ]! !

! Boolean methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self! !

! Boolean methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self! !

! Boolean methodsFor: #testing !
isBoolean
	" Return true if the receiver is a boolean. "

	^true! !

! Object methodsFor: #testing !
isBoolean
	" Return true if the receiver is a boolean. "

	^false! !

Object
 subclass: #Number
 instanceVariableNames: ''
 category: #Kernel!

! Number methodsFor: #comparing !
= anObject
    " Return true if the receiver is equal to argument. "

	{'' doc'number equality test'}! !

! Number methodsFor: #testing !
mustBeNumber
	" Signal an error if the receiver is not a Number.
	WARNING: this check impose a severe loss of perfomance (five times slower).
	If you are sure the operation/compare arguments are ok (a number),
	 comment the implementation of this method making the code empty.
	"

	"(# #isNaN: self) ifTrue: [ super mustBeNumber ]"! !

! Object methodsFor: #testing !
mustBeNumber
	" Signal an error if the receiver is not a Number. "

	^self error: 'Must be a Number'! !

! Number methodsFor: #comparing !
> aNumber
	" Return true if the receiver is greater than argument. "

	^{'' doc'number compare'}! !

! Number methodsFor: #comparing !
< aNumber
	" Return true if the receiver is smaller than argument. "

	^{'' doc'number compare'}! !

! Number methodsFor: #comparing !
>= aNumber
	" Return true if the receiver is greater or equal to argument. "

	^{'' doc'number compare'}! !

! Number methodsFor: #comparing !
<= aNumber
	" Return true if the receiver is smaller or equal to argument. "

	^{'' doc'number compare'}! !

! Number methodsFor: #arithmetic !
+ aNumber
	" Return the result of operation. "

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
- aNumber
	" Return the result of operation. "

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
* aNumber
	" Return the result of operation. "

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
/ aNumber
	" Return the result of operation. "

	^{'' doc'number operation'}! !
		
! Number methodsFor: #signaling !
divideByZero
	" Signal divide by cero attempt. "

	^self error: 'Divide by Zero'! !

! Number methodsFor: #arithmetic !
\\\\ aNumber
	" Return the result of operation. "

	aNumber mustBeNumber.
	aNumber = 0 ifTrue: [ self divideByZero ].
	^self - (self // aNumber * aNumber)! !

! Number methodsFor: #arithmetic !
// aNumber
	" Return the result of operation. "

	aNumber mustBeNumber.
	aNumber = 0 ifTrue: [ self divideByZero ].
	^(self / aNumber) truncated! !

! Number methodsFor: #arithmetic !
sin
	" Return the result of operation"

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
cos
	" Return the result of operation"

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
tan
	" Return the result of operation"

	^{'' doc'number operation'}! !


! Number methodsFor: #arithmetic !
arcCos
	" Return the result of operation"

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
arcSin
	" Return the result of operation"

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
arcTan
	" Return the result of operation"

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
atan2: aNumber
	" Return the result of operation"

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
negated
	" Return the receiver negated"

	^-1 * self ! !

! Number methodsFor: #arithmetic !
squared
	^self * self ! !

! Number methodsFor: #arithmetic !
sqrt
	" Return the result of operation"

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
<< aNumber
	" Return the receiver shifted aNumber of bits to the left. "

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
>> aNumber
	" Return the receiver shifted aNumber of bits to the right. "

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
bitShift: anInteger
	" Return the receiver shifted left anInteger number of bits if anInteger is positive, or shifted right for anInteger negated number of bits if anInteger is negative."

	anInteger < 0 ifTrue: [ ^self >> anInteger negated ].
	^self << anInteger! !

! Number methodsFor: #arithmetic !
raisedTo: aNumber
	" Return the receiver raised to the power of aNumber. "

	aNumber = 0 ifTrue: [ ^1.0 ].
	(self = 0 and: [aNumber > 0]) ifTrue: [ ^0.0 ].
	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
max: aNumber
	" Return the maximum between the receiver and argument. "

	^{'' doc'number operation'}! !

! Number methodsFor: #arithmetic !
min: aNumber
	" Return the minimum between the receiver and argument. "

	^{'' doc'number operation'}! !

! Number methodsFor: #converting !
asNumber
	" Return the receiver as a Number. "

	^self + 0! !

! Number methodsFor: #converting !
abs
	" Return the absolute value of the receiver. "

	self < 0 ifTrue: [ ^self negated ].
	^self! !

! Number methodsFor: #converting !
rounded
	" Return the receiver rounded. "

	^{'' doc'number operation'}! !

! Number methodsFor: #converting !
truncated
	" Return the receiver truncated. "

	^{'' doc'number operation'}! !

! Number methodsFor: #converting !
asCharacter
	" Return the character with the receiver's code. "

	^String fromCharCode: self! !

! Number methodsFor: #iterating !
to: aNumber by: increment value: aBlock
	" Private - Return an iterable to evaluate aBlock on each step of iteration. "

	^Interval from: self to: aNumber by: increment value: aBlock ! !

! Number methodsFor: #iterating !
to: end by: increment
	" Return an iterable from the receiver to end. "

	^self	to: end by: increment
		value: [:index :block| block value: index ]! !

! Number methodsFor: #iterating !
to: end
	" Return an iterable from the receiver to end. "

	^self to: end by: 1 "to iterate like other smalltalks
		(self <= end ifTrue: [1] ifFalse: [-1])"! !

! Number methodsFor: #enumerating !
timesRepeat: aBlock
	" Evaluates aBlock a number of times given by the receiver.
	Return the receiver.
	"

	| fn | fn := self callableOf: aBlock.
	{'' s8'(1 to: self) do: aBlock' doc'iteration'}! !

! Number methodsFor: #iterating !
to: aNumber do: aBlock
	" Evaluate aBlock from the receiver to aNumber.
	Return the receiver.
	"

	| fn | fn := self callableOf: aBlock.
	{''s8' (self to: aNumber) do: aBlock' doc'iteration'}! !

! Number methodsFor: #iterating !
to: aNumber by: increment do: aBlock
	" Evaluate aBlock for the numbers between the receiver and aNumber where each number is the previous number plus increment.
	Return the receiver.
	"

	| fn | fn := self callableOf: aBlock.
	{''s8'(self to: aNumber by: increment) consume: aBlock' doc'iteration'}! !

! Number methodsFor: #converting !
asString
	" Return the receiver as a String. "

	^self printString! !

! Number methodsFor: #converting !
asNative: lang
	" Return the native representation of the receiver. "

	^'(', self printString, ')'! !

! Number methodsFor: #printing !
printString
	" Return the printable representation of the receiver. "

	^{'' doc'native conversion'}! !

! Number methodsFor: #printing !
toString: radix
	" Return the ASCII representation of the receiver in base radix. "

	radix isNumber ifFalse: [
		^#converting toString: radix
	].
	^{'' doc'native conversion'}! !

! Number methodsFor: #printing !
printPaddedWith: aCharacter to: anInteger base: aRadix
	" Return the ASCII representation of the receiver padded on the left with aCharacter to be at least anInteger characters. "

	| result |
	result := self toString: aRadix.
	[ result size < anInteger ] whileTrue: [
		result := aCharacter ,result
	].
	^result! !

! Number methodsFor: #testing !
isNumber
	" Return true if the receiver is a Number. "

	{'' doc'number testing'}! !

! Number methodsFor: #converting !
atRandom
	" Return a random integer in the interval from 1 to the receiver. "

	^(Random new next * self) truncated + 1! !

! Number methodsFor: #converting !
@ aNumber
	" Return a Point with coordinates given by the receiver and aNumber. "

	^Point x: self y: (aNumber isNil ifTrue: [0] ifFalse: [aNumber])! !

! Number methodsFor: #converting !
asPoint
	" Return the receiver as a Point. "

	^Point x: self y: self! !

! Number methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	{'self'}.
	^self + 0
"Note:	we need to normalize result to do not destroy typeof checkings
	Do NOT modify this method if you do not know what you are doing.
	If you remove the + 0 computation, the returned object will return 'object' and not 'number' as result of typeof operation.
"! !

! Number methodsFor: #testing !
between: min and: max
	" Return true if the receiver is between arguments. "

	^min <= self and: [ self <= max ]! !

! Number methodsFor: #testing !
even
	" Return true if the receiver is an even number. "

	^self \\ 2 = 0! !

! Number methodsFor: #testing !
odd
	" Return true if the receiver is an odd number. "

	^self \\ 2 = 1! !

! Number methodsFor: #intervals !
clearInterval

	{'' doc'native timer cancelation'}! !

! Number methodsFor: #intervals !
clearTimeout

	{'' doc'native timer cancelation'}! !

! Number methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self! !

! Number methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self! !

! Number class methodsFor: #instantiation !
pi
	" Return the singular value PI. "

	^{'' doc'native constant'}! !

! Number class methodsFor: #converting !
readFrom: aStream radix: radix
	" Private - Read an integer from <aStream> and return it. "

	| value char |
	value := 0.
	[	char := aStream peek.
		char notNil and: [ char isDigit or: [ char isUpperCase ] ]
	] whileTrue: [
		radix <  char digitValue ifTrue: [ ^value ].
		value := value * radix + char digitValue.
		aStream next
	].
	^value! !

! Number class methodsFor: #converting !
readFrom: aStream
	" Private - Read an integer from <aStream> and return it. "

	| value char |
	value := 0.
	[	char := aStream peek.
		char notNil and: [ char isDigit ]
	] whileTrue: [
		value := value * 10 + char digitValue.
		aStream next
	].
	$r = char ifFalse: [ ^value ].
	aStream next.
	^self readFrom: aStream radix: value! !


! Number class methodsFor: #converting !
readFromOBSOLETE: aStream
	" Private - Read an integer from <aStream> and return it. "

	| value char |
	value := 0.
	[	char := aStream peek.
		char notNil and: [ char isDigit ]
	] whileTrue: [
		value := value * 10 + char digitValue.
		aStream next
	].
	char == "$r"'r' ifFalse: [ ^value ].
	aStream next. "skip the r"
	^self readFrom: aStream radix: value! !

Object
 subclass: #BlockClosure
 instanceVariableNames: ''
 category: #Kernel!

! BlockClosure class methodsFor: #instantiation !
empty
	" Return an empty block. "

	^[]! !

! BlockClosure class methodsFor: #instantiation !
defaultSortBlock
	" Private - Return the block used for default sort. "

	{'self'}.
	^[:a :b| [ a < b ] on: Exception do: [ a toString < b toString ] ]! !

! BlockClosure methodsFor: #accessing !
compiledSourceAt: aCompiledMethod
	" Return the native source code of the receiver. "

	| cached |
	cached := aCompiledMethod notNil ifTrue: [
		aCompiledMethod basicAt: '$nativeSource' ifAbsent: []
	].
	^cached isNil ifTrue: [ self compiledSource ] ifFalse: [ cached ]! !

! BlockClosure methodsFor: #accessing !
compiledSource
	" Return the native source code of the receiver. "

	{'self'}.
	^self toString! !

! BlockClosure methodsFor: #iterating !
whileTrue: aBlock
	" Evaluate aBlock while the result of evaluating the receiver is true. "

	| fn | fn := self callableOf: aBlock.
	{''s8' self value == true ifTrue: [
		aBlock value.
		self whileTrue: aBlock
	]' doc'iteration'}! !

! BlockClosure methodsFor: #iterating !
whileFalse: aBlock
	" Evaluate aBlock while the result of evaluating the receiver is false. "

	| fn | fn := self callableOf: aBlock.
	{''s8'	self value == false ifTrue: [
		aBlock value.
		self whileFalse: aBlock
	]' doc'iteration'}! !

! BlockClosure methodsFor: #iterating !
whileTrue
	" Evaluate the receiver while the result is true. "

	^self whileTrue: []! !

! BlockClosure methodsFor: #iterating !
whileFalse
	" Evaluate the receiver while the result is false. "

	^self whileFalse: []! !

! BlockClosure methodsFor: #evaluating !
value
	" Return the result of evaluating the receiver. "

	^{'self()' doc'native activation'}! !

! BlockClosure methodsFor: #evaluating !
value: anObject
	" Return the result of evaluating the receiver with anObject. "

	^{'self(anObject)' doc'native activation'}! !

! BlockClosure methodsFor: #evaluating !
value: arg1 value: arg2
	" Return the result of evaluating the receiver with arguments. "

	^{'self(arg1,arg2)' doc'native activation'}! !

! BlockClosure methodsFor: #evaluating !
value: arg1 value: arg2 value: arg3
	" Return the result of evaluating the receiver with arguments. "

	^{'self(arg1,arg2,arg3)' doc'native activation'}! !

! BlockClosure methodsFor: #evaluating !
value: arg1 value: arg2 value: arg3 value: arg4
	" Return the result of evaluating the receiver with arguments. "

	^{'self(arg1,arg2,arg3,arg4)' doc'native activation'}! !

! BlockClosure methodsFor: #evaluating !
valueWithPossibleArguments: array
	" Return the result of evaluating the receiver with potential arguments. "

	array isArray ifFalse: [
		self error: 'Invalid arguments in #valueWithPossibleArguments: (must be anArray). receiver=' ,self toString
	].
	^{'' doc'native activation'}! !

! BlockClosure methodsFor: #private !
exception: anException matches: type
	" Private - Return true if anException is of specified type. "

	type isNil ifTrue: [ ^true ].
	(self isObject: type) ifFalse: [ ^type = anException ].
	type isClass ifTrue: [ ^anException isKindOf: type ].
	type = #all ifTrue: [ ^true ].
	type isArray ifTrue: [
		type
			detect: [:one| self exception: anException matches: one ]
			ifNone: [ ^false ].
		^true
	].
	type isString ifTrue: [ | cls |
		cls := self smalltalk at: cls ifAbsent: [ ^false ].
		^self exception: anException matches: cls
	].
	^(type evaluateWith: anException) == true ! !

! BlockClosure methodsFor: #evaluating !
on: type do: aBlock
	" Return the result of evaluating the receiver catching exceptions of specified type. "

	^self try: self catch: [:anException |
		(self exception: anException matches: type)
		ifTrue: [ aBlock evaluateWith: anException ]
		ifFalse: [ anException signal ]
	]! !

! BlockClosure methodsFor: #private !
try: aBlock catch: anotherBlock
	" Private - Try to evaluate aBlock catching exceptions with catchBlock. "

	| fn ab |
	fn := self callableOf: aBlock.
	ab := self callableOf: anotherBlock.
	{'' doc'secured activation'}! !

! BlockClosure methodsFor: #evaluating !
valueDeferred: milliseconds
	" Evaluate the receiver deferred in time.
	Return the receiver.
	"

	self valueWithTimeout: milliseconds! !

! BlockClosure class methodsFor: #deferredEvaluation !
hasDeferredSupport
	" Return true if the platform can defer instance evaluation. "

	[ #{setTimeout} notNil ] on: Exception do: [ ^false ].
	^#{setTimeout} notNil! !

! BlockClosure methodsFor: #evaluating !
valueDeferred
	" Evaluate the receiver deferred in time.
	Return the receiver.
	"

	self valueDeferred: 0! !

! BlockClosure methodsFor: #timeout !
valueWithTimeout: aNumber
	" Evaluate the receiver with time out. "

	^{'' doc'native timer activation'}! !

! BlockClosure methodsFor: #timeout !
valueWithInterval: aNumber
	" Evaluate the receiver with interval. "

	^{'' doc'native timer activation'}! !

! BlockClosure methodsFor: #evaluating !
ensure: aBlock
	" Return the result of evaluating the receiver.
	Execute aBlock after the receiver is executed.
	"

	| fn | fn := self callableOf: aBlock.
	{'' doc'protected activation'}! !

! BlockClosure methodsFor: #evaluating !
ifCurtailed: exceptionBlock
	" Return the result of evaluating the receiver.
	If evaluation is limited by exiting before completion, the exception block is evaluated.
	"

	| curtailed result |
	curtailed := true.
	{'' doc'protected activation'}! !

! BlockClosure methodsFor: #evaluating !
argumentCount
	" Return the number of arguments of the receiver. "

	^{'' doc'activation constants'}! !

! BlockClosure methodsFor: #installing !
installAs: functionName
	" Install the receiver as a function in global scope. "

	^Smalltalk install: self as: functionName! !

! BlockClosure methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	^self! !

! BlockClosure methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self! !

! BlockClosure methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self! !

! BlockClosure methodsFor: #testing !
isValuable
	" Return true if the receiver can be evaluated. "

	^true! !

Object
 subclass: #UndefinedObject
 instanceVariableNames: ''
 category: #Kernel!

! UndefinedObject class methodsFor: #instantiation !
new
	" Return an instance of the receiver. "

	^self error: 'UndefinedObject can''t be instantiated.'! !

! UndefinedObject class methodsFor: #alias !
alias
	" Private - Return an alias of nil.
	Note that is is a hack used to wrap nil in an alias instance to make possible for Lua collections to contain nil
	"

	^self basicAt: '@alias' ifAbsentPut: [ self basicNew ]! !

! UndefinedObject methodsFor: #testing !
ifNil: aBlock
	" Return the receiver if it is not nil, or the result of evaluating aBlock (if it is nil). "

	^aBlock value! !

! UndefinedObject methodsFor: #testing !
ifNotNil: aBlock
	" Return the receiver if it is nil, or the result of evaluating aBlock. "

	^self! !

! UndefinedObject methodsFor: #testing !
ifNil: aBlock ifNotNil: anotherBlock
	" Return the receiver if it is nil, or the result of evaluating aBlock. "

	^aBlock value! !

! UndefinedObject methodsFor: #testing !
ifNotNil: aBlock ifNil: anotherBlock
	" Return the result of evaluating anotherBlock if teh receiver is nil, or the result of evaluating aBlock. "

	^anotherBlock value! !

! UndefinedObject methodsFor: #testing !
isNil
	" Return true if the receiver is nil. "

	^true! !

! UndefinedObject methodsFor: #testing !
notNil
	" Return true if the receiver is not nil. "

	^false! !

! UndefinedObject methodsFor: #printing !
printString
	" Return the text representation of the receiver. "

	^#nil! !

! UndefinedObject methodsFor: #instantiation !
subclass: aString instanceVariableNames: anotherString
	" Create a subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: anotherString
		category: nil! !

! UndefinedObject methodsFor: #instantiation !
subclass: aString instanceVariableNames: aString2 category: aString3
	" Create a subclass of the receiver. "

	^ClassBuilder new
		superclass: self
		subclass: aString
		instanceVariableNames: aString2
		category: aString3! !

! UndefinedObject methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	^{'' doc'native constant'}! !

! UndefinedObject methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self! !

! UndefinedObject methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self! !

! UndefinedObject methodsFor: #evaluating !
value
	" Return the result of evaluating the receiver. "! !

! UndefinedObject methodsFor: #converting !
@ aNumber
	" Return a Point with coordinates given by the receiver and aNumber. "

	^Point x: 0 y: aNumber! !

! UndefinedObject methodsFor: #comparing !
= anObject
	" Return true if the receiver is equal to argument. "

	^anObject isNil! !

! UndefinedObject methodsFor: #comparing !
~= anObject
	" Return true if the receiver is not equal to anObject. "

	^anObject notNil! !

Object
 subclass: #Collection
 instanceVariableNames: ''
 category: #Kernel!

! Collection methodsFor: #alias !
wrap: anObject
	" Private - Return an element holding anObject. "

	{'self'}.
	^anObject! !

! Collection methodsFor: #alias !
unwrap: anObject
	" Private - Return object held by anObject. "

	{'self'}.
	^anObject! !

! Collection methodsFor: #accessing !
size
	" Return the size of the receiver. "

	^self subclassResponsibility: #size! !

! Collection methodsFor: #streams !
readStream
	" Return a read stream on the receiver. "

	^self stream! !

! Collection methodsFor: #streams !
writeStream
	" Return a write stream on the receiver. "

	^self stream! !

! Collection methodsFor: #streams !
stream
	" Return a stream on the receiver. "

	^self streamClass on: self! !

! Collection methodsFor: #streams !
streamClass
	" Private - Return the support for streamming on the receiver. "

	^self class streamClass! !

! Collection methodsFor: #error !
objectNotFound
	" Private - Generates an error.
	An object was not found in the receiver's contents.
	"

	self smalltalk #objectNotFound: nil at: nil in: self.
	^self error: 'Object is not in the collection'! !

! Collection methodsFor: #error !
invalidIndex: anIndex
	" Private - Generates an error.
	An index to the receiver's contents is invalid/wrong.
	"

	^self error: 'Invalid index ' ,anIndex ! !

! Collection methodsFor: #error !
objectNotFound: anObject
	" Private - Generates an error.
	anObject was not found in the receiver's contents.
	"

	self smalltalk #objectNotFound: anObject at: nil in: self.
	^self error: 'Object not found: ' ,(self toString: anObject)! !

! Collection methodsFor: #error !
objectNotFoundAt: indexOrKey
	" Private - Generates an error.
	The object at key/index was not found in the receiver's contents.
	"

	self smalltalk #objectNotFound: nil at: indexOrKey in: self.
	^self error: 'Object not found at '
		,(self toString: indexOrKey)
		,' in ' ,self asLiteral! !

! Collection methodsFor: #adding !
add: anObject
	" Add anObject to the receiver. "

	^self subclassResponsibility: #add:! !

! Collection methodsFor: #adding !
addAll: aCollection
	" Add objects in aCollection to the receiver. "

	aCollection do: [:each | self add: each ].
	^aCollection! !

! Collection methodsFor: #copying !
, aCollection
	" Return a copy of the receiver with all contents of aCollection. "

	^self copy
		addAll: aCollection;
		yourself! !

! Collection methodsFor: #copying !
copyWith: anObject
	" Return a copy of the receiver with anObject. "

	^self copy add: anObject; yourself! !

! Collection methodsFor: #copying !
copyWithAll: aCollection
	" Return a copy of the receiver with all contents of aCollection. "

	^self copy
		addAll: aCollection;
		yourself! !

! Collection methodsFor: #converting !
asArray
	" Return an array with the contents of the receiver. "

	| result |
	result := Array new.
	self do: [:each| result add: each ].
	^result! !

! Collection methodsFor: #enumerating !
do: aBlock
	" Evaluate aBlock with the contents of the receiver. "

	^self subclassResponsibility: #do:! !

! Collection methodsFor: #enumerating !
detect: aBlock ifNone: anotherBlock
	" Return the object in the receiver that return true when evaluated with aBlock or the result of evaluating anotherBlock. "

	^self subclassResponsibility: #detect:ifNone:! !

! Collection methodsFor: #enumerating !
reverseDo: aBlock
	" Evaluate aBlock with the contents of the receiver in reverse order of iteration. "

	^self subclassResponsibility: #reverseDo:! !

! Collection methodsFor: #enumerating !
collect: aBlock
	" Return a collection with the result of evaluating aBlock with the contents of the receiver. "

	| result |
	result := self species new.
	self do: [:each | result add: (aBlock value: each) ].
	^result! !

! Collection methodsFor: #enumerating !
select: aBlock
	" Return the contents of the receiver which evaluates aBlock to true. "

	| result |
	result := self species new.
	self do: [:each|
	    (aBlock value: each) ifTrue: [ result add: each ]
	].
	^result! !

! Collection methodsFor: #enumerating !
detect: aBlock
	" Return the object in the receiver that return true when evaluated with aBlock.
	Generates an error if there is no object that evaluates aBlock to true.
	"

	^self detect: aBlock ifNone: [ self objectNotFound ]! !

! Collection methodsFor: #enumerating !
do: aBlock andBetweenDo: separatorBlock
	" For each element in the receiver, evaluate aBlock with that element as the argument.
	Evaluate the separatorBlock between successive elements of the receiver.
	"

	^self do: aBlock separatedBy: separatorBlock! !

! Collection methodsFor: #enumerating !
do: aBlock separatedBy: anotherBlock
	" Evaluate aBlock for contents of the receiver evaluating anotherBlock inbetween aBlock evaluations. "

	| first |
	first := true.
	self do: [:each |
	    first
		ifTrue: [ first := false ]
		ifFalse: [ anotherBlock value ].
	    aBlock value: each
	]! !

! Collection methodsFor: #enumerating !
inject: anObject into: aBlock
	" Evaluate aBlock for result of previous evaluation and contents of the receiver.
	On first evaluation anObject is used as first argument to evaluate aBlock.
	"

	| result |
	result := anObject.
	self do: [:each | result := aBlock value: result value: each ].
	^result! !

! Collection methodsFor: #enumerating !
reject: aBlock
	" Return the contents of the receiver wich evaluates aBlock to false. "

	^self select: [:each | (aBlock value: each) = false]! !

! Collection methodsFor: #enumerating !
select: selectBlock thenCollect: collectBlock
	" Return a new collection with the result of evaluating collectBlock on the contents of the receiver wich evaluates selectBlock to true. "

	^(self select: selectBlock) collect: collectBlock! !

! Collection methodsFor: #enumerating !
collect: collectBlock thenSelect: selectBlock
	" Return a new collection with the result of evaluating collectBlock on the contents of the receiver wich evaluates selectBlock to true. "

	^(self collect: collectBlock) select: selectBlock! !

! Collection methodsFor: #copying !
reversed
	" Return an array with the contents of the receiver in reverse order. "

	^self subclassResponsibility: #reversed! !

! Collection methodsFor: #testing !
includes: anObject
	" Return true if anObject is in the receiver. "

	^self subclassResponsibility: #includes:! !

! Collection methodsFor: #testing !
notEmpty
	" Return true if the receiver is not empty. "

	^self isEmpty not! !

! Collection methodsFor: #testing !
isEmpty
	" Return true if the receiver is empty. "

	^self size == 0! !

! Collection methodsFor: #removing !
remove: anObject
	" Remove anObject from receiver, return the removed object. "

	^self remove: anObject ifAbsent: [ self objectNotFound: anObject ]! !

! Collection methodsFor: #removing !
remove: anObject ifAbsent: aBlock
	" Remove anObject from receiver.
	Return the removed object or the result of evaluating aBlock (if the object is not found).
	"

	^self subclassResponsibility: #remove:ifAbsent:! !

! Collection methodsFor: #removing !
removeAll: aCollection
	" Remove aCollection from receiver. "

	aCollection do: [:each| self remove: each ].
	^aCollection! !

! Collection methodsFor: #removing !
removeAll
	" Remove all elements in the receiver. "

	self removeAll: self shallowCopy! !

! Collection methodsFor: #counting !
occurrencesOf: anObject
	" Return the occurrences of anObject in the receiver. "

	^self inject: 0 into: [:total :each|
		each = anObject ifTrue: [ total + 1 ] ifFalse: [total]
	]! !

! Collection methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self collect: [:each| each ]! !

! Collection methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self collect: [:each| each deepCopy ]! !

! Collection methodsFor: #iterating !
on: target
	" Return an iterable on the target contents accessed by the receiver. "

	| array |
	array := self asArray.
	^1 to: array size by: 1 value: [:index :block|
		block value: (target at: (array at: index) ifAbsent:[])
	]! !

! Collection class methodsFor: #streams !
streamClass
	" Private - Return the support for streamming on the receiver's instances. "

	^Stream! !

! Collection class methodsFor: #instantiation !
with: anObject
	" Return an instance of the receiver. "

	^self new
		add: anObject;
		yourself! !

! Collection class methodsFor: #instantiation !
with: anObject with: anotherObject
	" Return an instance of the receiver. "

	^self new
		add: anObject;
		add: anotherObject;
		yourself! !

! Collection class methodsFor: #instantiation !
with: firstObject with: secondObject with: thirdObject
	" Return an instance of the receiver. "

	^self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		yourself! !

! Collection class methodsFor: #instantiation !
with: firstObject with: secondObject with: thirdObject with: fourthObject
	" Return an instance of the receiver. "

	^self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		add: fourthObject;
		yourself! !

! Collection class methodsFor: #instantiation !
withAll: aCollection
	" Return an instance of the receiver. "

	^self new
		addAll: aCollection;
		yourself! !

! Collection class methodsFor: #instantiation !
basicNew: size
	" Return an instance of the receiver of specified size. "

	^self subclassResponsibility: #basicNew:! !

Collection
	subclass: #Interval
	instanceVariableNames: 'start end step value'
	category: #Kernel !

! Interval class methodsFor: #instantiation !
from: start to: end by: step value: aBlock
	" Return an instance of the receiver. "

	^self basicNew initialize: start to: end by: step value: aBlock! !

! Interval methodsFor: #initialize !
initialize: from to: to by: increment value: aBlock
	" Private - Initialize the receiver. "

	start := from.
	end := to.
	step := increment.
	value := aBlock! !

! Interval methodsFor: #accessing !
start
	" Return the start of the receiver. "

	^start ! !

! Interval methodsFor: #accessing !
end
	" Return the end of the receiver. "

	^end ! !

! Interval methodsFor: #accessing !
step
	" Return the step of the receiver. "

	^step ! !

! Interval methodsFor: #accessing !
start: from
	" Set the start of the receiver. "

	start := from ! !

! Interval methodsFor: #accessing !
end: to
	" Set the end of the receiver. "

	end := to ! !

! Interval methodsFor: #accessing !
step: increment
	" Set the step of the receiver. "

	step := increment ! !

! Interval methodsFor: #species !
species
	" Return the species to instantiate objects like the receiver. "

	^Array ! !

! Interval methodsFor: #error !
subclassResponsibility: aSelector
	" Private - The concrete implementation on the receiver is not complete.
	Note: Unimplemented methods handle invalid messages sent to the receiver.
	"

	^self invalidMessage: aSelector! !

! Interval methodsFor: #accessing !
size
	" Return the size of the receiver. "

	^0 max: end - start // step + 1! !

! Interval methodsFor: #iterating !
consume: aBlock
	" Private - Evaluates aBlock while the receiver is not empty. "

	[ self isEmpty ] whileFalse: [
		aBlock value: start.
		start := start + step
	] ! !

! Interval methodsFor: #testing !
atEnd
	" Return true if the receiver is at end. "

	^self size = 0 ! !

! Interval methodsFor: #testing !
isEmpty
	" Return true if the receiver is empty. "

	^self size = 0 ! !

! Interval methodsFor: #testing !
notEmpty
	" Return true if the receiver is not empty. "

	^self isEmpty not! !

! Interval methodsFor: #copying !
reversed
	" Return an Interval with the contents of the receiver in reverse order. "

	^self class from: end to: start by: step negated value: value ! !

! Interval methodsFor: #enumerating !
reverseDo: aBlock
	" Evaluate aBlock with the contents of the receiver in reverse order of iteration. "

	^self reversed do: aBlock ! !

! Interval methodsFor: #enumerating !
do: aBlock
	" Evaluate aBlock with the contents of the receiver. "

	start to: end by: step do: [:i|
		value value: i value: aBlock
	]! !

! Interval methodsFor: #arithmetic !
+ aNumber
	" Return the result of operation. "

	^start + aNumber to: end + aNumber by: step value: value! !

! Interval methodsFor: #arithmetic !
- aNumber
	" Return the result of operation. "

	^start - aNumber to: end - aNumber by: step value: value! !

! Interval methodsFor: #printing !
printString
	" Return the string representation of the receiver. "

	^$( ,start ,#to: ,end ,$) ! !

! Interval methodsFor: #converting !
toString
	" Return the receiver as a String. "

	(start isNil or: [end isNil]) ifTrue: [ ^super toString ].
	^self printString ! !

Collection
 subclass: #Set
 instanceVariableNames: 'contents'
 category: #Kernel!

! Collection methodsFor: #converting !
asSet
	" Return the receiver contents in a Set. "

	^Set withAll: self! !

! Collection methodsFor: #converting !
asHashedSet
	" Return the receiver contents in a HashedSet. "

	^HashedSet withAll: self! !

! Set class methodsFor: #private !
newContents: contents
	" Return an instance of the receiver initialized from given contents. "

	^self basicNew initialize: contents! !

! Set class methodsFor: #instantiation !
basicNew: size
	" Return an instance of the receiver of specified size. "

	^self new! !

! Set methodsFor: #initialize !
initialize
	" Private - Initialize the receiver. "

	super initialize.
	contents := #().! !

! Set methodsFor: #initialize !
initialize: initializer
	" Private - Initialize the receiver. "

	contents := initializer.! !

! Set methodsFor: #accessing !
size
	" Return the size of the receiver. "

	^contents size! !

! Set methodsFor: #private !
basicAdd: anObject
	" Private - Add anObject to the receiver's contents. "

	contents add: anObject.
	^anObject! !

! Set methodsFor: #testing !
includes: anObject
	" Return true if anObject is in the receiver. "

	^contents includes: anObject! !

! Set methodsFor: #adding !
add: anObject
	" Add anObject to the receiver. "

	(self includes: anObject) ifTrue: [ ^anObject ].
	^self basicAdd: anObject! !

! Set methodsFor: #removing !
remove: anObject ifAbsent: aBlock
	" Remove anObject from the receiver.
	Return the result of evaluating aBlock if anObject is not in the receiver.
	"

	^contents remove: anObject ifAbsent: aBlock! !

! Set methodsFor: #copying !
copyFrom: anIndex to: anotherIndex
	" Return a copy of the receiver from anIndex to anotherIndex. "

	^self shouldNotImplement: #copyFrom:to:! !

! Set methodsFor: #converting !
asNative: lang
	" Return the native representation of the receiver. "

	| send |
	send := lang = #lua ifTrue: [ $: ] ifFalse: [ $. ].
	^'smalltalk.' ,self class name
		,send ,'newContents_('
		,(contents asNative: lang)
		,')'! !

! Set methodsFor: #comparing !
= aSet
	" Return true if the receiver is equal to argument. "

	self class = aSet class ifFalse: [^false].
	self size = aSet size ifFalse: [^false].
	self do: [:each | (aSet includes: each) ifFalse: [^false] ].
	^true! !

! Set methodsFor: #converting !
asArray
	" Return an array with the contents of the receiver. "

	^contents copy! !

! Set methodsFor: #converting !
toString
	" Convert the receiver to a String. "

	^self class name,' withAll: ',self asArray printString! !

! Set methodsFor: #sorting !
sorted
	" Return the contents of the receiver sorted. "

	^self asArray sort! !

! Set methodsFor: #sorting !
sorted: aBlock
	" Return a sorted copy of the receiver. "

	^self asArray sort: aBlock! !

! Set methodsFor: #enumerating !
do: aBlock
	" Evaluate aBlock with the contents of the receiver. "

	^contents do: aBlock! !

! Set methodsFor: #enumerating !
detect: aBlock ifNone: anotherBlock
	" Return the object in the receiver that return true when evaluated with aBlock or the result of evaluating anotherBlock. "

	^contents detect: aBlock ifNone: anotherBlock! !

! Set methodsFor: #enumerating !
reverseDo: aBlock
	" Evaluate aBlock with the contents of the receiver in reverse order of iteration. "

	^contents reverseDo: aBlock! !

! Set methodsFor: #enumerating !
collect: aBlock
	" Return a collection with the result of evaluating aBlock with the contents of the receiver. "

	^self class withAll: (contents collect: aBlock)! !

! Set methodsFor: #enumerating !
do: aBlock separatedBy: anotherBlock
	" Evaluate aBlock for contents of the receiver evaluating anotherBlock inbetween aBlock evaluations. "

	^contents do: aBlock separatedBy: anotherBlock! !

! Set methodsFor: #enumerating !
inject: anObject into: aBlock
	" Evaluate aBlock for result of previous evaluation and contents of the receiver.
	On first evaluation anObject is used as first argument to evaluate aBlock.
	"

	^contents inject: anObject into: aBlock! !

! Set methodsFor: #enumerating !
select: aBlock
	" Return a copy of the receiver with elements wich evaluates aBlock to true. "

	^self class withAll: (contents select: aBlock)! !

! Set methodsFor: #enumerating !
reject: aBlock
	" Return the contents of the receiver wich evaluates aBlock to false. "

	^self class withAll: (contents reject: aBlock)! !

! Set methodsFor: #copying !
reversed
	" Return an array with the contents of the receiver in reverse order. "

	^contents reversed! !

! Set methodsFor: #testing !
notEmpty
	" Return true if the receiver is not empty. "

	^contents notEmpty! !

! Set methodsFor: #testing !
isEmpty
	" Return true if the receiver is empty. "

	^contents isEmpty! !

! Set methodsFor: #counting !
occurrencesOf: anObject
	" Return the occurrences of anObject in the receiver. "

	^(self includes: anObject) ifTrue: [ 1 ] ifFalse: [ 0 ]! !

! Set methodsFor: #converting !
asSet
	" Return the receiver contents in a Set. "

	^self! !

! Set methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self species newContents: contents copy! !

! Set methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self species newContents: contents deepCopy! !

Set
 subclass: #HashedSet
 instanceVariableNames: 'hash'
 category: #Kernel!

! HashedSet methodsFor: #initialize !
initialize
	" Private - Initialize the receiver. "

	super initialize.
	self initializeHash.! !

! HashedSet methodsFor: #initialize !
initialize: initializer
	" Private - Initialize the receiver. "

	contents := initializer.
	self rehash.! !

! HashedSet methodsFor: #private !
initializeHash
	" Private - Initialize the hash of the receiver. "

	hash := {'' doc'get native sentinel'}.! !

! HashedSet methodsFor: #private !
rehash
	" Private - Rehash the receiver contents. "

	self initializeHash.
	contents do: [:each|
		self privateAdd: (self wrap: each)
	]! !

! HashedSet methodsFor: #converting !
asHashedSet
	" Return the receiver contents in a HashedSet. "

	^self! !

! HashedSet methodsFor: #private !
indexOf: anObject ifAbsent: aBlock
	" Return the position of anObject in the receiver, or the result of evaluating aBlock. "

	(self includes: anObject) ifFalse: [ ^aBlock value ].
	^contents indexOf: anObject ifAbsent: aBlock! !

! HashedSet methodsFor: #private !
withIndexDo: aBlock
	" Evaluate aBlock with all the index positions in the receiver. "

	^contents withIndexDo: aBlock! !

! HashedSet methodsFor: #private !
removeIndex: index
	" Remove the element at index. "

	^self remove: (contents at: index)! !

! HashedSet methodsFor: #private !
privateAdd: cookie
	" Private - Register a cookie onto the receiver's hash. "

	{'self'}.
	hash basicAt: cookie put: cookie! !

! HashedSet methodsFor: #private !
privateRemove: cookie
	" Private - Remove a cookie in the receiver's hash. "

	{'' doc'native remotion'}! !

! HashedSet methodsFor: #testing !
includes: anObject
	" Return true if anObject is in the receiver. "

	| cookie |
	cookie := self wrap: anObject.
	^{'' doc'testing'}! !

! HashedSet methodsFor: #adding !
add: anObject
	" Add anObject to the receiver. "

	(self includes: anObject) ifTrue: [ ^anObject ].
	self privateAdd: (self wrap: anObject).
	^self basicAdd: anObject! !

! HashedSet methodsFor: #removing !
remove: anObject ifAbsent: aBlock
	" Remove anObject from the receiver.
	Return the result of evaluating aBlock if anObject is not in the receiver.
	"

	(self includes: anObject) ifFalse: [ ^aBlock value ].
	contents remove: anObject ifAbsent: [ ].
	self privateRemove: (self wrap: anObject).
	^anObject! !


Collection
 subclass: #SequenceableCollection
 instanceVariableNames: ''
 category: #Kernel!

! SequenceableCollection methodsFor: #accessing !
at: anIndex
	" Return the object in the receiver at anIndex position. "

	^self at: anIndex ifAbsent: [ self objectNotFoundAt: anIndex ]! !

! SequenceableCollection methodsFor: #accessing !
at: anIndex ifAbsent: aBlock
	" Return the object in the receiver at anIndex position or the result of evaluating aBlock. "

	^self subclassResponsibility: #at:ifAbsent:! !

! SequenceableCollection methodsFor: #accessing !
at: anIndex put: anObject
	" Set the object in the receiver at anIndex position. "

	^self subclassResponsibility: #at:put:! !

! SequenceableCollection methodsFor: #accessing !
at: aKey ifAbsentPut: aBlock
	" Return the value at aKey, or add the result of evaluating aBlock to the receiver if it is absent. "

	^self at: aKey ifAbsent: [ self at: aKey put: aBlock value ]! !

! SequenceableCollection methodsFor: #copying !
copyFrom: anIndex to: anotherIndex
	" Return a copy of the receiver. "

	^self subclassResponsibility: #copyFrom:to:! !

! SequenceableCollection methodsFor: #testing !
indexOf: anObject
	" Return the position of anObject in the receiver, or 0. "

	^self indexOf: anObject ifAbsent: [ 0 ]! !

! SequenceableCollection methodsFor: #testing !
indexOf: anObject ifAbsent: aBlock
	" Return the position of anObject in the receiver, or the result of evaluating aBlock. "

	{'' doc'detection'}! !

! SequenceableCollection methodsFor: #enumerating !
do: aBlock
	" Evaluate aBlock with the contents of the receiver. "

	^self implementedBySubclass: #do: ! !

! SequenceableCollection methodsFor: #enumerating !
detect: aBlock ifNone: anotherBlock
	" Return the object in the receiver that return true when evaluated with aBlock or the result of evaluating anotherBlock. "

	| fn | fn := self callableOf: aBlock.
	{''s8'	self do: [:each|
		(aBlock value: each) ifTrue: [ ^each ]
	].
	^anotherBlock value'
	doc'detection'}! !

! SequenceableCollection methodsFor: #copying !
reversed
	" Return an array with the contents of the receiver in reverse order. "

	{'' doc'converting'}! !

! SequenceableCollection methodsFor: #testing !
isEmpty
	" Return true if the receiver is empty.
	Note: ^self size = 0 is slower than inlined impl.
	"

	^{'' doc'testing'}! !

! SequenceableCollection methodsFor: #enumerating !
reverseDo: aBlock
	" Evaluate aBlock with the contents of the receiver in reverse order of iteration. "

	| fn | fn := self callableOf: aBlock.
	{''s8'	self size to: 1 by: -1 do: [:i|
		aBlock value: (self at: i)
	]' doc'(reversed) iteration'}! !

! SequenceableCollection methodsFor: #testing !
includes: anObject
	" Return true if anObject is in the receiver. "

	{'self'}.
	(self isObject: anObject) ifFalse: [ ^(self #indexOf: anObject) > -1 ].
	{'' doc'native detection'}! !

! SequenceableCollection methodsFor: #counting !
occurrencesOf: anObject
	" Return the occurrences of anObject in the receiver. "

	{'' doc'native counting'}! !

! SequenceableCollection methodsFor: #accessing !
first
	" Return the object in the receiver. "

	^self at: 1! !

! SequenceableCollection methodsFor: #accessing !
fourth
	" Return the object in the receiver. "

	^self at: 4! !

! SequenceableCollection methodsFor: #accessing !
last
	" Return the object in the receiver. "

	^self at: self size! !

! SequenceableCollection methodsFor: #accessing !
second
	" Return the object in the receiver. "

	^self at: 2! !

! SequenceableCollection methodsFor: #accessing !
third
	" Return the object in the receiver. "

	^self at: 3! !

! SequenceableCollection methodsFor: #removing !
removeLast
	" Remove the last object in the receiver. "

	^self remove: self last! !

! SequenceableCollection methodsFor: #removing !
removeFirst
	" Remove the first object in the receiver. "

	^self remove: self first! !

! SequenceableCollection methodsFor: #adding !
addLast: anObject
	" Add anObject to the receiver (at end). "

	^self add: anObject! !

! SequenceableCollection methodsFor: #accessing !
atAll: indices put: anObject
	" Return the receiver after replacing those elements, indexed by the indices, with anObject. "

	indices do: [ :index | self at: index put: anObject ]! !

! SequenceableCollection methodsFor: #accessing !
atAllPut: anObject
	" Replace all elements of the receiver with anObject. "

	(1 to: self size) do: [ :index | self at: index put: anObject ]! !

! SequenceableCollection methodsFor: #enumerating !
withIndexDo: aBlock
	" Evaluate aBlock with all the index positions in the receiver. "

	(1 to: self size) do: [ :index |
		aBlock value: (self at: index) value: index
	]! !

! SequenceableCollection methodsFor: #testing !
= aCollection
	" Return true if the receiver is equal to argument. "

	| index |
	self == aCollection ifTrue: [ ^true ].
	self class == aCollection class ifFalse: [ ^false ].
	index := self size.
	index ~= aCollection size ifTrue: [ ^false ].
	[ index <= 0 ] whileFalse: [
		(self at: index) = (aCollection at: index) ifFalse: [ ^false ].
		index := index - 1
	].
	^true! !

! SequenceableCollection methodsFor: #testing !
beginsWith: aCollection
	" Return true if the receiver begins with aCollection. "

	self size < aCollection size ifTrue: [ ^false ].
	self size = aCollection size ifTrue: [ ^aCollection = self ].
	^aCollection = (self copyFrom: 1 to: aCollection size)! !

! SequenceableCollection methodsFor: #testing !
endsWith: aCollection
	" Return true if the receiver ends with aCollection. "

	self size < aCollection size ifTrue: [ ^false ].
	self size = aCollection size ifTrue: [ ^aCollection = self ].
	^aCollection = (self copyFrom: self size - aCollection size + 1 to: self size)! !

! SequenceableCollection methodsFor: #copying !
copyReplaceFrom: start to: stop with: aCollection
	" Return a collection containing a copy of the receiver with the elements at index positions from start through stop replaced with the elements of aCollection. "

	^(self copyFrom: 1 to: start - 1)
		, aCollection
		, (self copyFrom: stop + 1 to: self size)! !

! SequenceableCollection methodsFor: #copying !
copyWithout: anObject
	" Return a copy of the receiver excluding anObject, if any. "

	| index |
	index := self indexOf: anObject ifAbsent: [ ^self copy ].
	^(self copyFrom: 1 to: index - 1)
		,((self copyFrom: index + 1 to: self size)
			copyWithout: anObject)! !

! SequenceableCollection methodsFor: #copying !
copyWithAll: aCollection
	" Return a copy of the receiver with all contents of aCollection. "

	^self , aCollection ! !

! SequenceableCollection methodsFor: #searching !
findFirst: aBlock
	" Return the index of the first element of the receiver that causes aBlock to evaluate to true.
         If no such element is found, report an error.
	"

	^self findFirst: aBlock ifAbsent: [ self errorAbsentObject ]! !

! SequenceableCollection methodsFor: #searching !
findFirst: aBlock ifAbsent: absentBlock
	" Return the index of the first element of the receiver that causes aBlock to evaluate to true.
	If no such element is found, return the result of evaluating the argument block.
	"

	| index size |
	size := self size.
	index := 1.
	[ index <= size ] whileTrue: [
		(aBlock value: (self at: index)) ifTrue: [ ^index ].
		index := index + 1
	].
	^absentBlock value! !

! SequenceableCollection methodsFor: #searching !
findLast: aBlock
	" Return the index of the last element of the receiver that causes aBlock to evaluate to true.
	If no such element is found, report an error.
	"

	^self findLast: aBlock ifAbsent: [ self errorAbsentObject ]! !

! SequenceableCollection methodsFor: #searching !
findLast: aBlock ifAbsent: absentBlock
	" Return the index of the last element of the receiver that causes aBlock to evaluate to true.
	If no such element is found, return the result of evaluating the argument block.
	"

	| index |
	index := self size.
	[ index > 0 ] whileTrue: [
		(aBlock value: (self at: index)) ifTrue: [ ^index ].
		index := index - 1
	].
	^absentBlock value! !

! SequenceableCollection methodsFor: #replace !
replaceFrom: start to: stop with: aCollection
	" Return the receiver.
	Replace the elements of the receiver at index positions start through stop, with the elements of aCollection. "

	stop - start + 1 = aCollection size ifFalse: [
		^self error: 'Replacement has wrong size'
	].
	self
		replaceFrom: start
		to: stop
		with: aCollection
		startingAt: 1! !

! SequenceableCollection methodsFor: #replace !
replaceFrom: start
    to: stop
    with: aCollection
    startingAt: repStart
	" Return the receiver.
	Replace the elements of the receiver at index positions start through stop with consecutive
         elements of aCollection beginning at index position repStart.
	"

	| index1 index2 |
	(self == aCollection and: [ repStart < start ]) ifTrue: [ "do backward move for same object"
		index2 := repStart + stop - start.
		index1 := stop.
		[ start <= index1 ] whileTrue: [
			self at: index1 put: (aCollection at: index2).
			index1 := index1 - 1.
			index2 := index2 - 1
		].
		^self
	].
	index2 := repStart.
	index1 := start.
	[ index1 <= stop ] whileTrue: [
		self at: index1 put: (aCollection at: index2).
		index1 := index1 + 1.
		index2 := index2 + 1
	]! !

! SequenceableCollection methodsFor: #replace !
replaceFrom: start
    to: stop
    with: replacement
    startingWith: replacementStart
	" Replace the elements of the receiver. "

	^self
		replaceFrom: start
		to: stop
		with: replacement
		startingAt: replacementStart! !

! SequenceableCollection methodsFor: #replace !
replaceFrom: start to: stop withObject: anObject
	"Replace each of the elements of the receiver at index positions start through stop with anObject. "

	| index |
	index := start.
	[ index <= stop ] whileTrue: [
		self at: index put: anObject.
		index := index + 1
	].
	^anObject! !

! SequenceableCollection methodsFor: #iterating !
with: aCollection do: aBlock
	" For each pair of elements evaluate aBlock with those elements as the arguments. "

	| first second |
	first	:= self stream.
	second	:= aCollection stream.
	[ first atEnd or: [ second atEnd ] ] whileFalse: [
		aBlock value: first next value: second next
	].! !

! SequenceableCollection methodsFor: #sorting !
sort
	" Sort (inplace) the receiver's contents with default criteria. "

	self sort: nil! !

! SequenceableCollection methodsFor: #sorting !
sort: aBlock
	" Sort (inplace) the receiver's contents with aBlock. "

	self implementedBySubclass: #sort:! !

! SequenceableCollection methodsFor: #iterating !
from: start to: end by: increment
	" Return an iterable on the receiver's contents from start to end. "

	^start to: end by: increment value: [:index :block|
		block value: (self at: index)
	]! !

! SequenceableCollection methodsFor: #iterating !
from: start to: end
	" Return an iterable on the receiver's contents from start to end. "

	^self	from: start to: end
		by: (start <= end ifTrue: [1] ifFalse: [-1]) ! !

SequenceableCollection
 subclass: #String
 instanceVariableNames: ''
 category: #Kernel!

! String class methodsFor: #instantiation !
fromString: aString
	" Return an instance of the receiver. "

	{'' doc'instatiation'}! !

! String class methodsFor: #instantiation !
new
	" Return an instance of the receiver. "

	^{'' doc'instantiation'}! !


! String class methodsFor: #constants !
cr
	" Return the singular instance. "

	^{'' doc'native constant'}! !

! String class methodsFor: #constants !
lf
	" Return the singular instance. "

	^{'' doc'native constant'}! !

! String class methodsFor: #constants !
crlf
	" Return the singular instance. "

	^{'' doc'native constant'}! !

! String class methodsFor: #constants !
space
	" Return the singular instance. "

	^{'' doc'native constant'}! !

! String class methodsFor: #constants !
tab
	" Return the singular instance. "

	^{'' doc'native constant'}! !

! String class methodsFor: #constants !
zero
	" Return the singular instance. "

	^{'' doc'native constant'}! !

! String class methodsFor: #instantiation !
with: aCharacter
	" Return an instance of the receiver. "

	^aCharacter! !

! String class methodsFor: #instantiation !
with: aCharacter with: bCharacter
	" Return an instance of the receiver. "

	^aCharacter , bCharacter! !

! String class methodsFor: #instantiation !
with: aCharacter with: bCharacter with: cCharacter
	" Return an instance of the receiver. "

	^aCharacter , bCharacter , cCharacter! !

! String class methodsFor: #instantiation !
with: aCharacter with: bCharacter
	with: cCharacter with: dCharacter
	" Return an instance of the receiver. "

	^aCharacter , bCharacter , cCharacter , dCharacter! !

! String class methodsFor: #instantiation !
fromCharCode: anArrayOrNumber
	" Return an instance of the receiver. "

	{'' doc'native conversion'}! !

! String class methodsFor: #constants !
doItSelector
	" Return the selector to use for doIt evaluations.
	Warning: this selector will not be dump on image nor change log expressions.
	"

	^#s8DoIt! !

! String class methodsFor: #fileNames !
nativeFileExtension
	" Return the file extension to use for files with native code. "

	^self nativeLanguage toString! !

! String class methodsFor: #fileNames !
sourceFileExtension
	" Return the file extension to use for source files. "

	^#st ! !

! String class methodsFor: #fileNames !
nativeLibraryExtension
	" Return the file extension to use for native library files. "

	^self sourceFileExtension ,$. ,self nativeFileExtension ! !

! String methodsFor: #comparing !
= anObject
    " Return true if the receiver is equal to argument. "

	{'' doc'native testing'}! !

! String methodsFor: #accessing !
size
	" Return the size of the receiver. "

	^{'' doc'native string size'}! !

! String methodsFor: #accessing !
at: anIndex
	" Return the receiver at anIndex. "

	^{'' doc'native indexing'}! !

! String methodsFor: #accessing !
at: anIndex put: anObject
	" Set the receiver at anIndex. "

	^self errorReadOnly: #at:put:! !

! String methodsFor: #accessing !
at: anIndex ifAbsent: aBlock
	" Return the receiver at anIndex position or the result of evaluating aBlock. "

	^(self at: anIndex) ifNil: aBlock! !

! String methodsFor: #enumerating !
do: aBlock
	" Evaluate aBlock with the contents of the receiver. "

	| fn | fn := self callableOf: aBlock.
	{''s8'	1 to: self size do: [:i|
		aBlock value: (self at: i)
	]' doc'native iteration'}! !

! String methodsFor: #enumerating !
detect: aBlock ifNone: anotherBlock
	" Return the object in the receiver that return true when evaluated with aBlock or the result of evaluating anotherBlock. "

	| fn | fn := self callableOf: aBlock.
	{''s8'	self do: [:each|
		(aBlock value: each) ifTrue: [ ^each ]
	].
	^anotherBlock value' doc'native detection'}! !

! String methodsFor: #testing !
includes: anObject
	" Return true if anObject is in the receiver. "

	{'self'}.
	^(self indexOf: anObject) > 0! !

! String methodsFor: #enumerating !
withIndexDo: aBlock
	" Evaluate aBlock with all the index positions in the receiver. "

	| fn | fn := self callableOf: aBlock.
	{''s8'	1 to: self size do: [:i|
		aBlock value: (self at: i) value: i
	]' doc'native iteration'}! !

! String methodsFor: #copying !
copyWithout: anObject
	" Return a copy of the receiver excluding anObject, if any. "

	^self replace: anObject with: ''! !

! String methodsFor: #conversion !
asSymbol
	" Return a symbol with the same characters as the receiver. "

	^self "we do not discriminate singular strings"! !

! String class methodsFor: #constants !
binarySelectorChars
	" Private - Return the characters that can be included in binary selectors. "

	^'|&+/=><,@~!!*-%\?\\' ! !

! String class methodsFor: #private !
isBinarySelector: aString
	" Private - Return true if aString is a binary selector. "

	| chars |
	chars := self binarySelectorChars.
	aString do: [:c| (chars includes: c) ifFalse: [ ^false ] ].
	^true ! !

! String class methodsFor: #private !
isKeywordSelector: aString
	" Private - Return true if aString is a keyword selector. "

	{'' js'self' doc'native testing'}.
	^((RegularExpression fromString: '[^a-zA-Z0-9_:]') exec: aString) isNil ! !

! String class methodsFor: #private !
isMessageSelector: aString
	" Private - Return true if aString is a message selector. "

	^(self isKeywordSelector: aString)
	or: [ self isBinarySelector: aString ] ! !

! String methodsFor: #evaluating !
argumentCount
	" Return the number of arguments of the receiver. "

	self isEmpty ifTrue: [ ^0 ].
	(self class binarySelectorChars includes: self first) ifTrue: [ ^1 ].
	^self occurrencesOf: $:! !

! String methodsFor: #testing !
isLiteral
	" Return true if the receiver can be output using literal notation. "

	{'' js'self' doc'native testing'}.
	^self notEmpty and: [ self class isMessageSelector: self ]! !

! String methodsFor: #converting !
asLiteral
	" Return the receiver in literal notation. "

	^self isLiteral ifTrue: [ $# ,self ] ifFalse: [ self printString ]! !

! Object methodsFor: #converting !
asLiteral
	" Return the receiver in literal notation. "

	^self printString! !

! String methodsFor: #testing !
occurrencesOf: aString
	" Return the times aString can be found in the receiver. "

	| result index |
	{'' doc'native counting'}.
	aString isString ifFalse: [
		result := self #match: aString.
		result isNil ifTrue: [ ^0 ].
		^result#length
	].
	aString size = 1 ifTrue: [
		result := self #match: (RegularExpression
				fromString: '[',aString,']'
				flag: 'g').
		result isNil ifTrue: [ ^0 ].
		^result#length
	].
	result := 0. index := 0.
	[	index := self indexOf: aString startingAt: index + 1.
		index > 1 ] whileTrue: [ result := result + 1 ].
	^result! !

! String methodsFor: #testing !
startsWith: aString
	" Return true if the receiver begins with the characters of aString. "

	{'' doc'native testing'}.
	^self beginsWith: aString! !

! String methodsFor: #converting !
escaped
	" Return the receiver with escape codes. "

	{'' doc'native conversion'}! !

! String methodsFor: #converting !
unescaped
	" Return the receiver from encoded URI escape codes. "

	{'' doc'native conversion'}! !

! String methodsFor: #converting !
escapedCode
	" Return the code to reproduce receiver from code. "

	| escaped |
	escaped := self escaped.
	escaped = self ifTrue: [ ^$" ,escaped ,$" ].
	^'unescape("' ,escaped ,'")'! !

! String methodsFor: #converting !
asEncodedURI
	" Return the receiver encoded as an URI component.
	Encodes special characters and it encodes the following characters: , / ? : @ & = + $ #
	"

	{'' doc'native conversion'}! !

! String methodsFor: #converting !
asDecodedURI
	" Return the receiver decoded as an URI component. "

	{'' doc'native conversion'}! !

! String methodsFor: #adding !
add: anObject
	" Add anObject to the receiver. "

	^self errorReadOnly: #add:! !

! String methodsFor: #converting !
--> anObject
	" Return a far-association of the receiver to anObject. "

	^Association key: (self asArrayOfSubstringsSeparatedBy: $.) value: anObject! !

! String methodsFor: #copying !
, aCollection
	" Return a copy of the receiver with all contents of aCollection. "

	{'' doc'native composition'}! !

! String methodsFor: #copying !
copyWith: anObject
	" Return a copy of the receiver with anObject. "

	^{'' doc'native composition'}! !

! String methodsFor: #copying !
copyFrom: anIndex to: anotherIndex
	" Return the receiver between limits. "

	^{'' doc'native copy'}! !

! String methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self! !

! String methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self! !

! String methodsFor: #copying !
reversed
	" Return a string with the contents of the receiver in reverse order. "

	| result |
	{'self' doc'native conversion'}.
	result := ''.
	self reverseDo: [:c| result := result , c ].
	^result! !

! String methodsFor: #accessing !
indexOf: aString
	" Return the position of first occurrence of aString (or 0 if not present). "

	{'' doc'native detection'}! !

! String methodsFor: #accessing !
indexOf: aString ifAbsent: aBlock
	" Return the position of first occurrence of aString (or the result of evaluating aBlock). "

	| result |
	result := self indexOf: aString.
	^result = 0 ifTrue: aBlock ifFalse: [ result ]! !

! String methodsFor: #accessing !
indexOf: aString startingAt: position
	" Return the position of first occurrence of aString after position (or 0 if not present). "

	{'' doc'native detection'}! !

! String methodsFor: #accessing !
lastIndexOf: aString
	" Return the position of last occurrence of aString (or 0 if not present). "

	{'' doc'native detection'}! !

! String methodsFor: #accessing !
lastIndexOf: aString ifAbsent: aBlock
	" Return the position of last occurrence of aString (or the result of evaluating aBlock). "

	| result |
	result := self lastIndexOf: aString.
	^result = 0 ifTrue: aBlock ifFalse: [ result ]! !

! String methodsFor: #accessing !
lastIndexOf: aString startingAt: position
	" Return the position of last occurrence of aString before position (or 0 if not present). "

	{'' doc'native detection'}! !

! String methodsFor: #accessing !
upTo: aString startingAt: position
	" Return the receiver upto first occurrence of aString after position. "

	| index |
	index := self indexOf: aString startingAt: position.
	index := index = 0 ifTrue: [ self size ] ifFalse: [ index - 1 ].
	^self copyFrom: position to: index! !

! String methodsFor: #accessing !
upTo: aCharacter
	" Return the receiver upto first occurrence of aCharacter. "

	^self upTo: aCharacter startingAt: 1! !

! String methodsFor: #accessing !
upToLast: aString
	" Return the receiver upto last occurrence of aString. "

	| index |
	index := self lastIndexOf: aString ifAbsent: [ ^self ].
	^self copyFrom: 1 to: index - 1! !

! String methodsFor: #accessing !
fromLast: aString
	" Return the receiver contents from last occurrence of aString. "

	| index |
	index := self lastIndexOf: aString ifAbsent: [ ^self ].
	^self copyFrom: index + aString size to: self size! !

! String methodsFor: #js !
asJsName
	" Private - Return the receiver as a valid native name. "

	(String reservedWords includes: self) ifTrue: [
		self = #toString ifFalse: [ ^'$$' ,self ]
	].
	^self! !

! String methodsFor: #js !
asNativeName
	" Return the receiver as a valid native name. "

	^{'' doc'native conversion'}! !

! String class methodsFor: #js !
reservedArray
	" Private - Return the reserved names of native language. "

	^'abstract as boolean break byte case catch char class
	continue const debugger default delete do double else
	enum export extends false final finally float for function
	goto if implements import in instanceof int interface is
	long namespace native new null package private
	protected public return short static super switch
	synchronized this throw throws transient true try typeof
	use var void volatile while with' asArrayOfSubstrings! !

! String class methodsFor: #js !
reservedWords
	" Private - Return the reserved names of native language. "

	^self basicAt: '$reservedSet' ifAbsentPut: [
		HashedSet withAll: self reservedArray
	]! !

! String methodsFor: #converting !
asPrivateSelector
	" Private - Return the receiver as an encoded selector. "

	| result |
	{'' doc'native conversion'}.
	^result ! !

! String methodsFor: #converting !
asSelector
	" Private - Return the receiver as a valid native selector. "

	(self startsWith: $#) ifTrue: [
		^self upTo: $: startingAt: 2
	].
	^self asPrivateSelector asNativeName ! !

! String methodsFor: #converting !
asNativeSelector
	" Return a smalltalk selector from a native function name.
	Note: it must be the reverse operation of #asSelector
	"

	| operations |
	self = '$' ifTrue: [ ^'' ].
	self = #_bk ifTrue: [ ^92 asCharacter ].
	(self startsWith: '$$') ifTrue: [
		^(self copyFrom: 3 to: self size) asNativeSelector
	].
	operations := #(
		#( + _plus )
		#( - _minus )
		#( * _star )
		#( / _slash )
		#( | _bar )
		#( & _amp )
		#( > _gt )
		#( < _lt )
		#( = _eq )
		#( , _comma )
		#( @ _at )
		#( ~ _no )
	).
	operations add: (Array with: 92 asCharacter with: #_bk).
	operations do: [:tuple|
		(self startsWith: tuple last) ifTrue: [
			^operations reversed inject: self into: [:result :each|
				result replace: each last with: each first
			]
		].
	].
	^self replace: $_ with: $:! !

! String methodsFor: #converting !
asStSelector
	" OBSOLETE - Return a smalltalk selector from a native function name.
	Note: use #asNativeSelector
	"

	self note: 'OBSOLETE - #asStSelector (please use #asNativeSelector)'.
	^self asNativeSelector! !

! String methodsFor: #converting !
slashEncoded
	" Private - Return the native representation of the receiver slashed. "

	| encoded bar |
	bar := 92 asCharacter.
	encoded := self.
	encoded := encoded replace: bar with: bar ,bar.
	encoded := encoded replace: String lf with: bar ,$n.
	encoded := encoded replace: String cr with: bar ,$r.
	encoded := encoded replace: String tab with: bar ,$t.
	encoded := encoded replace: $" with: bar ,$".
	^$" ,encoded ,$"! !

! String methodsFor: #converting !
asNative: lang
	" Return the native representation of the receiver. "

	lang = #lua ifTrue: [ ^self slashEncoded ].
	{'' doc'native conversion'}! !

! String methodsFor: #regularExpressions !
replace: aString with: anotherString
	" Return the result of replacement of aString with anotherString in the receiver. "

	{'self' doc'native conversion'}.
	aString == $\ ifTrue: [ ^self replace: '\\' with: anotherString ].
	^self
		replaceRegexp: (RegularExpression fromString: aString flag: 'g')
		with: anotherString! !

! String methodsFor: #regularExpressions !
replaceRegexp: aRegexp with: aString
	" Return the result of replacement of aRegexp with aString in the receiver. "

	^{'' doc'native conversion'}! !

! String methodsFor: #converting !
tokenize: aString
	" Return the split tokenization of the receiver. "

	{'' doc'native conversion'}! !

! String methodsFor: #converting !
asArrayOfSubstrings
	" Return an array of substrings from the receiver.
	The receiver is divided into substrings by one or more occurrences of separators.
	"

	| result start i |
	result := Array new. i := 1.
	self do: [:c|
		start isNil ifTrue: [ "in separator"
			c isSeparator ifFalse: [ "start of substring"
				start := i.
			]
		] ifFalse: [ "in word"
			c isSeparator ifTrue: [ "end of substring"
				result add: (self copyFrom: start to: i - 1).
				start := nil.
			].
		].
		i := i + 1.
	].
	start notNil ifTrue: [
		result add: (self copyFrom: start to: self size)
	].
	^result! !

! String methodsFor: #converting !
asArrayOfSubstringsSeparatedBy: aCharacter
	" Return an array of substrings from the receiver.
	The receiver is divided into substrings by one or more occurrences of aCharacter.
	"

	^self tokenize: aCharacter! !

! String methodsFor: #regularExpressions !
match: aRegexp
	" Return the match index of aRegexp in the receiver. "

	^{'' doc'native testing'}! !

! String methodsFor: #converting !
asString
	" Return the text representation of the receiver. "

	^self! !

! String methodsFor: #testing !
between: min and: max
	" Return true if the receiver is between arguments. "

	^min <= self and: [ self <= max ]! !

! String methodsFor: #testing !
isLetter
	" Return true if the receiver is a one character letter. "

	^{'' doc'native testing'}! !

! String methodsFor: #testing !
isDigit
	" Return true if the receiver is one character numeric digit. "

	^{'' doc'native testing'}! !

! String methodsFor: #testing !
isAlphaNumeric
	" Return true if the receiver is an alphanumeric character. "

	{'self' doc'native testing'}.
	^self isDigit or: [self isLetter]! !

! String methodsFor: #testing !
isUppercase
	" Return true if all the receiver characters are uppercase. "

	^self isUpperCase! !

! String methodsFor: #testing !
isUpperCase
	" Return true if all the receiver characters are uppercase. "

	{'self' doc'native testing'}.
	^(self match: '[^A-Z]') not! !

! String methodsFor: #testing !
isLowerCase
	" Return true if all the receiver characters are lowercase. "

	{'self' doc'native testing'}.
	^(self match: '[^a-z]') not! !

! String methodsFor: #testing !
isLowercase
	" Return true if the receiver is one character lowercase. "

	^self isLowerCase! !

! String methodsFor: #converting !
asUppercase
	" Return a copy of the receiver converted to uppercase. "

	^{'' doc'native testing'}! !

! String methodsFor: #converting !
asUpperCase
	" Return a copy of the receiver converted to uppercase. "

	^self asUppercase! !

! String methodsFor: #converting !
asLowercase
	" Return a copy of the receiver converted to lowercase. "

	^{'' doc'native testing'}! !

! String methodsFor: #converting !
asLowerCase
	" Return a copy of the receiver converted to lowercase. "

	^self asLowercase! !

! String methodsFor: #converting !
asSmalltalkName: upCase
	" Return a copy of the receiver converted to a smalltalk global name. "

	| up result |
	(upCase not and: [
		(self detect: [:c| c >= $a ] ifNone: []) isNil
	]) ifTrue: [
		^self asLowercase asSmalltalkName: upCase
	].
	up := upCase.
	result := self inject: '' into: [:s :each|
		('_:.&%$?' includes: each) ifTrue: [ up := true. s ]
		ifFalse: [ s , (up
			ifTrue: [  up := false. each asUppercase ]
			ifFalse: [ each ])
		]
	].
	result isEmpty ifTrue: [ ^$_ ].
	result first isDigit ifTrue: [ ^$_ ,result ].
	upCase ifTrue: [
		result first isLowercase ifTrue: [
			^result first asUppercase ,(result copyFrom: 2 to: result size)
		]
	] ifFalse: [
		result first isUppercase ifTrue: [
			^result first asLowercase ,(result copyFrom: 2 to: result size)
		]
	].
	^result! !

! String methodsFor: #converting !
asGlobalName
	" Return a copy of the receiver converted to a smalltalk global name. "

	^self asSmalltalkName: true! !

! String methodsFor: #converting !
asLocalName
	" Return a copy of the receiver converted to a smalltalk local name. "

	^self asSmalltalkName: false! !

! String methodsFor: #converting !
digitValue
	" Return the digit value of the receiver (or -1). "

	^('0123456789ABCDEF' indexOf: self) - 1 max: 0 - 1! !

! String methodsFor: #converting !
asNumber
	" Return the receiver as a Number. "

	(self includes: $r) ifTrue: [
		^Number readFrom: self readStream
	].
	^{'' doc'native converting'}! !

! String methodsFor: #error !
errorReadOnly
	" DEPRECATED, Private - The receiver is read-only instance. "

	^self error: 'Strings are inmutable. Attempt to change ' ,self asLiteral! !

! String methodsFor: #error !
errorReadOnly: selector
	" Private - The receiver is read-only instance. "

	^self error: 'Strings are inmutable. '
		,selector asLiteral
		,' sent to ' ,self asLiteral! !

! String methodsFor: #printing !
printString
	" Return the printable representation of the receiver. "

	| body |
	body := (self includes: $') ifTrue: [
		(self asArrayOfSubstringsSeparatedBy: $') join: ''''''
	] ifFalse: [ self ].
	^$' ,body ,$'! !

! String methodsFor: #testing !
isString
	" Return true if the receiver is a String. "

	^true! !

! String methodsFor: #testing !
isBackSlash
	" Return true if the receiver is a back-slash string. "

	^self asciiValue = 92 ! !

! String methodsFor: #comparing !
> aString
	" Return true if the receiver is greater than argument. "

	^{'' doc'native compare'}! !

! String methodsFor: #comparing !
< aString
	" Return true if the receiver is smaller than argument. "

	^{'' doc'native compare'}! !

! String methodsFor: #comparing !
>= aString
	" Return true if the receiver is greater or equal to argument. "

	^{'' doc'native compare'}! !

! String methodsFor: #comparing !
<= aString
	" Return true if the receiver is smaller or equal to argument. "

	^{'' doc'native compare'}! !

! String methodsFor: #removing !
remove: anObject ifAbsent: aBlock
	" Remove anObject from the receiver. "

	^self errorReadOnly: #remove:ifAbsent:! !

! String methodsFor: #removing !
removeAll: aCollection
	" Remove aCollection from receiver. "

	^self errorReadOnly: #removeAll:! !

! String methodsFor: #removing !
removeAll
	" Remove all elements in the receiver. "

	^self errorReadOnly: #removeAll! !

! String methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	{'self'}.
	^self , ''
"Note:	we need to normalize result to do not destroy typeof checkings
	Do NOT modify this method if you do not know what you are doing.
	If you remove the ,'' computation, the returned object will return 'object' and not 'string' as result of typeof operation.
"! !

! Object methodsFor: #testing !
isCharacter
	" Return true if the receiver is a character. "

	^false! !

! String methodsFor: #testing !
isCharacter
	" Implemented to support Character protocoll for one-character Strings. "

	^{'' doc'native testing'}! !

! String methodsFor: #testing !
asciiValue
	" Return the number corresponding to the ASCII encoding of the first character of the receiver. "

	^{'' doc'native converting'}! !

! String methodsFor: #testing !
isSeparator
	" Return true if the receiver is either a space, tab, new-line, form-feed character, or double-byte space.
	Note that new-line has been extended to support NEL character.
	See https://en.wikipedia.org/wiki/Newline
	"

	^{'' doc'native testing'}! !

! String methodsFor: #converting !
asAsciiZ
	" Return the receiver with ascii zero appended. "

	^self , self class zero! !

! String methodsFor: #selecting !
trimFirst: firstBlock last: lastBlock
	" Return a String containing the result of trimming characters matching blocks from both sides of the receiver. "

	| start stop |
	start := 1. stop := self size.
	[	start > stop ifTrue: [ ^String new ].
		firstBlock value: (self at: start)
	] whileTrue: [ start := start + 1 ].
	[	stop < start ifTrue: [ ^String new ].
		lastBlock value: (self at: stop)
	] whileTrue: [ stop := stop - 1 ].
	^self copyFrom: start to: stop! !

! String methodsFor: #selecting !
trim: aBlock
	" Return a String containing the result of trimming characters matching aBlock from both sides of the receiver. "

	^self trimFirst: aBlock last: aBlock! !

! String methodsFor: #selecting !
trimFirst: aBlock
	" Return a String containing the result of trimming characters matching aBlock from start of the receiver. "

	^self trimFirst: aBlock last: [:c| false ]! !

! String methodsFor: #selecting !
trimLast: aBlock
	" Return a String containing the result of trimming characters matching aBlock from end of the receiver. "

	^self trimFirst: [:c| false ] last: aBlock! !

! String methodsFor: #selecting !
trimBlanks
	" Return a String containing the receiver string with leading and trailing blanks removed. "

	^self trim: [:c| c isSeparator ]! !

! String methodsFor: #iterating !
to: end
	" Return an iterable from the receiver to end. "

	^self to: end by: (self <= end ifTrue: [1] ifFalse: [-1])! !

! String methodsFor: #iterating !
to: end by: increment
	" Return an iterable from the receiver to end. "

	^self asciiValue
		to: end asciiValue
		by: increment
		value: [:index :block| block value: index asCharacter ]! !

SequenceableCollection
 subclass: #Array
 instanceVariableNames: ''
 category: #Kernel!

! Array class methodsFor: #instantiation !
new
	" Return an instance of the receiver. "

	^#()! !

! Array class methodsFor: #instantiation !
basicNew: size
	" Return an instance of the receiver of specified size. "

	^{'' doc'native instantiation'}! !

! Array class methodsFor: #instantiation !
new: size
	" Return an instance of the receiver of specified size. "

	^(self basicNew: size) atAllPut: nil! !

! Array class methodsFor: #instantiation !
with: anObject
	" Return an instance of the receiver. "

	{'' doc'native instantiation'}.
	^self new
		add: anObject;
		yourself! !

! Array class methodsFor: #instantiation !
with: firstObject with: secondObject
	" Return an instance of the receiver. "

	{'' doc'native instantiation'}.
	^self new
		add: firstObject;
		add: secondObject;
		yourself! !

! Array class methodsFor: #instantiation !
with: firstObject with: secondObject with: thirdObject
	" Return an instance of the receiver. "

	{'' doc'native instantiation'}.
	^self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		yourself! !

! Array class methodsFor: #instantiation !
with: firstObject with: secondObject with: thirdObject with: fourthObject
	" Return an instance of the receiver. "

	{'' doc'native instantiation'}.
	^self new
		add: firstObject;
		add: secondObject;
		add: thirdObject;
		add: fourthObject;
		yourself! !

! Array class methodsFor: #instantiation !
withAll: aCollection
	" Return an instance of the receiver. "

	{'' doc'native instantiation'}! !

! Array class methodsFor: #util !
slice: anIterable
	" Return an Array instance from contents of anIterable. "

	^{'' doc'native instantiation'}! !

! Array methodsFor: #accessing !
size
	" Return the size of the receiver. "

	^{'' doc'native size'}! !

! Array methodsFor: #accessing !
at: anIndex
	" Return the contents of the receiver at anIndex. "

	{'' doc'native indexing'}! !

! Array methodsFor: #accessing !
at: anIndex ifAbsent: aBlock
	" Return the contents of the receiver at anIndex or the result of evaluating aBlock. "

	{'' doc'native indexing'}! !

! Array methodsFor: #accessing !
at: anIndex put: anObject
	" Set the contents of the receiver at anIndex. "

	{'' doc'native indexing'}! !

! Array methodsFor: #enumerating !
do: aBlock
	" Evaluate aBlock with the contents of the receiver. "

	| fn | fn := self callableOf: aBlock.
	{''s8'	1 to: self size do: [:i|
		aBlock value: (self at: i)
	]' doc'native iteration'}! !

! Array methodsFor: #testing !
isArray
	" Return true if the receiver is an Array. "

	^self class == Array! !

! Array methodsFor: #adding !
addAll: aCollection
	" Add objects in aCollection to the receiver. "

	{'' doc'native composing'}.
	^aCollection! !

! Array methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	| result |
	{'' doc'native copy'}.
	^result! !

! Array methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self collect: [:each | each deepCopy ]! !

! Array methodsFor: #copying !
copyFrom: anIndex to: anotherIndex
	" Return the receiver contents in the interval. "

	| result |
	result := self species new.
	anIndex to: anotherIndex do: [:each | result add: (self at: each) ].
	^result! !

! Array methodsFor: #copying !
copyWithAll: aCollection
	" Return a copy of the receiver with all contents of aCollection. "

	{'' doc'native composition'}! !

! Array methodsFor: #copying !
, aCollection
	" Return a copy of the receiver with all contents of aCollection. "

	{'' doc'native composition'}! !

! Array methodsFor: #composing !
join: aString
	" Return a string with the receiver's contents separated by aString. "

	^{'' doc'native composition'}! !

! Array methodsFor: #composing !
compose: aBlock with: aString
	" Return a string composed with the result of evaluating aBlock into receiver's contents. "

	^(self	collect: aBlock
		thenSelect: [:each| each notNil ])
		join: aString! !

! Array methodsFor: #composing !
compose: aBlock
	" Return a string composed with the result of evaluating aBlock into receiver's contents. "

	^self compose: aBlock with: ''! !

! Array methodsFor: #converting !
asNative: lang
	" Return the native representation of the receiver. "

	| list |
	list := (self collect: [:each | each asNative: lang ]) join: ', '.
	lang = #lua ifTrue: [ ^'smalltalk:newArray' ,$( ,${ ,list ,$} ,$) ].
	^$[ ,list ,$]! !

! Array methodsFor: #sorting !
sort: aBlock
	" Sort (inplace) the receiver's contents with aBlock. "

	| fn | fn := self callableOf: aBlock.
	{'' doc'native sorting'}! !

! Array methodsFor: #converting !
asArray
	" Return an array with the contents of the receiver. "

	^self! !

! Array methodsFor: #removing !
remove: anObject ifAbsent: aBlock
	" Remove anObject from the receiver.
	Return the result of evaluating aBlock if anObject is not in the receiver.
	"

	{'' doc'native remotion'}! !

! Array methodsFor: #adding !
add: anObject
	" Add anObject to the receiver. "

	{'' doc'native addition'}.
	^anObject! !

! Array methodsFor: #adding !
addFirst: anObject
	" Add anObject to the receiver. "

	{'' doc'native addition'}.
	^anObject! !

! Array methodsFor: #adding !
addLast: anObject
	" Add anObject to the receiver (at end). "

	{'' doc'native addition'}.
	^anObject! !

! Array methodsFor: #removing !
removeAll
	" Remove all the objects in the receiver. "

	{'' doc'native remotion'}! !

! Array methodsFor: #removing !
removeFirst
	" Remove the first object in the receiver. "

	| result |
	result := self first.
	{'' doc'native remotion'}.
	^result! !

! Array methodsFor: #removing !
removeIndex: index
	" Remove the element at index. "

	{'' doc'native remotion'}! !

! Array methodsFor: #removing !
removeLast
	" Remove the last object in the receiver. "

	^{'' doc'native remotion'}! !

! Array methodsFor: #sorting !
sorted
	" Return a sorted copy of the receiver. "

	^self copy sort: nil! !

! Array methodsFor: #sorting !
sorted: aBlock
	" Return a sorted copy of the receiver. "

	^self copy sort: aBlock! !

! Array methodsFor: #accessing !
atAllPut: anObject
	" Replace all elements of the receiver with anObject. "

	{'' doc'native operation'}! !

! Array methodsFor: #accessing !
first
	" Return the object in the receiver. "

	{'' doc'native indexing'}.
	^self at: 1! !

! Array methodsFor: #accessing !
second
	" Return the object in the receiver. "

	{'' doc'native indexing'}.
	^self at: 2! !

! Array methodsFor: #accessing !
third
	" Return the object in the receiver. "

	{'' doc'native indexing'}.
	^self at: 3! !

! Array methodsFor: #accessing !
fourth
	" Return the object in the receiver. "

	{'' doc'native indexing'}.
	^self at: 4! !

! Array methodsFor: #accessing !
fifth
	" Return the fifth object in the receiver. "

	{'' doc'native indexing'}.
	^self at: 5! !

! Array methodsFor: #accessing !
sixth
	" Return the sixth object in the receiver. "

	{'' doc'native indexing'}.
	^self at: 6! !

! Array methodsFor: #enumerating !
withIndexDo: aBlock
	" Evaluate aBlock with all the index positions in the receiver. "

	| index |
	index := 1.
	self do: [:each|
		aBlock value: each value: index.
		index := index + 1
	]! !

! Array methodsFor: #printing !
printString
	" DEBUG - Temporary printing method (must check for recursion). "

	| result |
	self isEmpty ifTrue: [ ^'#()' ].
	result := '#( '.
	self do: [:each|
		result size > 10000 ifTrue: [ ^result,'...)' ].
		result := result , (
			(self isObject: each)
			ifTrue: [ each printString ]
			ifFalse: [ each ]) ,' '
	].
	^result,')'! !

! Array methodsFor: #converting !
asLiteral
	" Return the receiver in literal notation. "

	| literals |
	literals := self collect: [:each|
		(self isObject: each)
		ifTrue: [ each asLiteral ]
		ifFalse: [ each toString ]
	].
	^'#(' ,(literals join: $ ) ,$)! !

! Array methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	| result |
	{'self' doc'native conversion'}.
	result := self collect: [:each| 
		(self isObject: each)
		ifTrue: [ each asJSONObject ]
		ifFalse: [ each ]
	].
	{'self' doc'native closing'}.
	^result! !

! Array methodsFor: #converting !
json
	" Return the receiver as a hash object. "

	| result |
	{'self' doc'native conversion'}.
	result := self collect: [:each| 
		(self isObject: each)
		ifTrue: [ each json ]
		ifFalse: [ each ]
	].
	{'self' doc'native closing'}.
	^result! !

Object
 subclass: #RegularExpression
 instanceVariableNames: ''
 category: #Kernel!

! RegularExpression methodsFor: #evaluating !
compile: aString
	" Return the result of compilation of aString. "

	^self #compile: aString! !

! RegularExpression methodsFor: #evaluating !
exec: aString
	" Return the result of excecution of aString. "

	^self #exec: aString! !

! RegularExpression methodsFor: #evaluating !
test: aString
	" Return the result of testing of aString. "

	^self #test: aString! !

! RegularExpression class methodsFor: #instantiation !
fromString: aString flag: anotherString
	" Return an instance of the receiver. "

	^{'' doc'native instantiation'}! !

! RegularExpression class methodsFor: #instantiation !
fromString: aString
	" Return an instance of the receiver. "

	    ^self fromString: aString flag: ''! !

! RegularExpression methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self! !

! RegularExpression methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	^self! !

! RegularExpression methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	^self! !

! Error class methodsFor: #instantiation !
signal: aString
	" Signal an instance of the receiver. "

	^self new
		messageText: aString;
		captureStack;
		signal! !

! Exception methodsFor: #accessing !
messageText
	" Return the messageText of the receiver. "

	^messageText! !

! Exception methodsFor: #accessing !
description
	" Return the description of the receiver. "

	| txt ex |
	txt := self messageText.
	txt isNil ifTrue: [ txt := '' ].
	txt := self toString: txt.
	ex := self nativeException.
	ex notNil ifTrue: [
		txt := txt
			,' [' ,(self typeOf: ex)
			,'] ' ,(self toString: ex)
	].
	^self class name ,' - ' ,txt ! !

! Exception methodsFor: #accessing !
messageText: aString
	" Set the messageText of the receiver. "

	messageText := aString! !

! Exception methodsFor: #signaling !
signal
	" Signal/throw the receiver. "

	self note: self description.
	{'' doc'jump into native exception'}! !

SequenceableCollection
 subclass: #ArrayedContents
 instanceVariableNames: ' contents '
 category: #Kernel!

! ArrayedContents methodsFor: #private !
contents
	" Private - Return the contents of the receiver. "

	contents isNil ifTrue: [ contents := Array new ].
	^contents! !

! ArrayedContents methodsFor: #adding !
add: anObject
	" Add anObject to the receiver. "

	^self contents add: anObject! !

! ArrayedContents methodsFor: #accessing !
at: anIndex ifAbsent: aBlock
	" Return the contents of the receiver at anIndex or the result of evaluating aBlock. "

	^self contents at: anIndex ifAbsent: aBlock! !

! ArrayedContents methodsFor: #accessing !
at: anIndex put: anObject
	" Set the contents of the receiver at anIndex. "

	^self contents at: anIndex put: anObject! !

! ArrayedContents methodsFor: #accessing !
atAllPut: anObject
	" Replace all elements of the receiver with anObject. "

	^self contents atAllPut: anObject! !

! ArrayedContents methodsFor: #copying !
copyFrom: anIndex to: anotherIndex
	" Return the receiver contents in the interval. "

	| result |
	result := self species new.
	anIndex to: anotherIndex do: [:each | result add: (self at: each) ].
	^result! !

! ArrayedContents methodsFor: #testing !
indexOf: anObject ifAbsent: aBlock
	" Return the position of anObject in the receiver, or the result of evaluating aBlock. "

	^self contents indexOf: anObject ifAbsent: aBlock! !

! ArrayedContents methodsFor: #enumerating !
detect: aBlock ifNone: anotherBlock
	" Return the object in the receiver that return true when evaluated with aBlock or the result of evaluating anotherBlock. "

	^self contents detect: aBlock ifNone: anotherBlock! !

! ArrayedContents methodsFor: #enumerating !
do: aBlock
	" Evaluate aBlock with the contents of the receiver. "

	^self contents do: aBlock! !

! ArrayedContents methodsFor: #testing !
includes: anObject
	" Return true if anObject is in the receiver. "

	^self contents includes: anObject! !

! ArrayedContents methodsFor: #testing !
isEmpty
	" Return true if the receiver is empty. "

	^self contents isEmpty! !

! ArrayedContents methodsFor: #counting !
occurrencesOf: anObject
	" Return the occurrences of anObject in the receiver. "

	^self contents occurrencesOf: anObject! !

! ArrayedContents methodsFor: #enumerating !
reverseDo: aBlock
	" Evaluate aBlock with the contents of the receiver in reverse order of iteration. "

	^self contents reverseDo: aBlock! !

! ArrayedContents methodsFor: #copying !
reversed
	" Return an array with the contents of the receiver in reverse order. "

	^self class withAll: self contents reversed! !

! ArrayedContents methodsFor: #enumerating !
detect: aBlock ifNone: anotherBlock
	" Return the object in the receiver that return true when evaluated with aBlock or the result of evaluating anotherBlock. "

	^self contents detect: aBlock ifNone: anotherBlock! !

! ArrayedContents methodsFor: #removing !
remove: anObject ifAbsent: aBlock
	" Remove anObject from the receiver.
	Return the result of evaluating aBlock if anObject is not in the receiver.
	"

	^self contents remove: anObject ifAbsent: aBlock! !

! ArrayedContents methodsFor: #removing !
removeAll
	" Remove all the objects in the receiver. "

	contents := nil! !

! ArrayedContents methodsFor: #removing !
removeFirst
	" Remove the first object in the receiver. "

	^self contents removeFirst! !

! ArrayedContents methodsFor: #removing !
removeLast
	" Remove the last object in the receiver. "

	^self contents removeLast! !

! ArrayedContents methodsFor: #removing !
removeIndex: index
	" Remove the element at index. "

	^self contents removeIndex: index! !

! ArrayedContents methodsFor: #accesing !
size
	" Return the size of the receiver. "

	^self contents size! !

! ArrayedContents methodsFor: #enumerating !
withIndexDo: aBlock
	" Evaluate aBlock with all the index positions in the receiver. "

	^self contents withIndexDo: aBlock! !

ArrayedContents
 subclass: #ComposedAssociations
 instanceVariableNames: ' '
 category: #Kernel!

! ComposedAssociations class methodsFor: #composing !
, anObject
	" Return an instance of the receiver with the argument. "

	^self new ,anObject! !

! ComposedAssociations methodsFor: #composing !
, anObject
	" Return the receiver composed with the argument. "

	anObject isAssociation ifTrue: [
		^self add: anObject; yourself
	].
	^anObject isNil ifTrue: [ self ] ifFalse: [ super , anObject ]! !

! ComposedAssociations methodsFor: #private !
hashObject
	" Private - Return an empty hash object. "

	^{'' doc'native instantiation'}! !

! ComposedAssociations methodsFor: #private !
install: key value: value into: jsObject
	" Private - Install a key/value pair into jsObject. "

	((self isObject: key) and: [ key isArray ]) ifTrue: [
		key size = 1 ifTrue: [ ^self install: key first value: value into: jsObject ].
		key size > 1 ifTrue: [ | target |
			target := jsObject basicAt: key first ifAbsent: [
				jsObject basicAt: (self toString: key first) put: self hashObject
			].
			^self install: (key copyFrom: 2 to: key size) value: value into: target
		].
	].
	^jsObject
		basicAt: (self toString: key) "it is IMPORTANT to send #toString here (v8{4.7.0} bug will expose {object} instead of {string} Ale-14/oct/2015)"
		put: value! !

! ComposedAssociations methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	| result aValue |
	result := self hashObject.
	self do: [:assoc |
		aValue := (self isObject: assoc value)
			ifTrue: [ assoc value json ]
			ifFalse: [ assoc value ].
		self install: assoc key value: aValue into: result
	].
	^result! !

! ComposedAssociations methodsFor: #converting !
handle
	" Return the receiver as a native object. "

	^self json! !

! ComposedAssociations methodsFor: #accessing !
associationAt: aKey ifAbsent: aBlock
	" Return the association at aKey (or the result of evaluating aBlock).
	Note: the search will introspect in case of composite values.
	"

	self reverseDo: [:assoc| | value |
		assoc key = aKey ifTrue: [ ^assoc ].
		value := assoc value.
		((self isObject: value) and: [value isKindOf: self class]) ifTrue: [
			value := value associationAt: aKey ifAbsent: [].
			value notNil ifTrue: [ ^value ].
		]
	].
	^aBlock evaluate! !

! ComposedAssociations methodsFor: #accessing !
@ aKey
	" Return the value at aKey (or nil).
	Note: the search will introspect in case of composite values.
	"

	| assoc |
	assoc := self associationAt: aKey ifAbsent: [ ^nil ].
	^assoc value! !

Object
 subclass: #Association
 instanceVariableNames: 'key value'
 category: #Kernel!

! Association class methodsFor: #instantiation !
key: aKey
	" Return an instance of the receiver. "

	^self new key: aKey! !

! Association class methodsFor: #instantiation !
key: aKey value: aValue
	" Return an instance of the receiver. "

	^(self key: aKey) value: aValue! !

! Association methodsFor: #comparing !
= anAssociation
	" Return true if the receiver is equal to argument. "

	^self class = anAssociation class and: [
	    self key = anAssociation key and: [
		self value = anAssociation value]]! !

! Association methodsFor: #accessing !
key: aKey
	" Set the key of the receiver. "

	key := aKey! !

! Association methodsFor: #accessing !
key
	" Return the key of the receiver. "

	^key! !

! Association methodsFor: #accessing !
value: aValue
	" Set the value of the receiver. "

	value := aValue! !

! Association methodsFor: #accessing !
value
	" Return the value of the receiver. "

	^value! !

! Association methodsFor: #composing !
, anAssociation
	" Return the receiver composed with the argument. "

	^ComposedAssociations with: self with: anAssociation! !

! Association methodsFor: #converting !
asJSONObject
	" Return the receiver as a hash object. "

	^(ComposedAssociations with: self) json! !

! Association methodsFor: #testing !
isAssociation
	" Return true if the receiver is anAssociation. "

	^true! !

! Object methodsFor: #testing !
isAssociation
	" Return true if the receiver is anAssociation. "

	^false! !

! Association methodsFor: #printing !
printString
	" Return a printable representation of the receiver. "

	^self toString ! !

! Association methodsFor: #converting !
toString
	" Return the receiver as a String. "

	| string |
	string := [:x| [
		x isNil ifTrue: [ #nil ] ifFalse: [
			(self isObject: x) ifTrue: [ x asLiteral ]
			ifFalse: [ self toString: x ] ]
		] on: Exception do: [:ex| '??? ' ,ex description ] ].
	^$( ,(string value: key) ,' -> ' ,(string value: value) ,$)! !

Collection
 subclass: #KeyedCollection
 instanceVariableNames: ''
 category: #Kernel!

KeyedCollection
 subclass: #Dictionary
 instanceVariableNames: 'keys values'
 category: #Kernel!

KeyedCollection
	subclass: #PoolDictionary
	instanceVariableNames: ' hash size '
	category: #Kernel !

! KeyedCollection class methodsFor: #instantiation !
basicNew: size
	" Return an instance of the receiver of specified size. "

	^self new! !

! KeyedCollection class methodsFor: #instantiation !
@ jsObject
	" Return an instance of the receiver with the properties owned by jsObject (or nil). "

	| result parts |
	jsObject isNil ifTrue: [ ^nil ].
	result := self new.
	{'' doc'native accessing'}.
	parts do: [:key|
		result at: key put: (jsObject basicAt: key)
	].
	^result! !

! KeyedCollection class methodsFor: #instantiation !
fromTuples: tuples
	" Return an instance of the receiver. "

	| result |
	result := self new.
	tuples do: [:tuple| result at: tuple first put: tuple last ].
	^result! !

! KeyedCollection methodsFor: #copying !
, aCollection
	" Return a copy of the receiver with all contents of aCollection. "

	^self shouldNotImplement: #,! !

! KeyedCollection methodsFor: #comparing !
= anObject
	" Return true if the receiver is equal to argument. "

	self class = anObject class ifFalse: [^false].
	self size = anObject size ifFalse: [^false].
	self keysAndValuesDo: [:key :value |
		(anObject at: key ifAbsent: [^false])
			= value ifFalse: [^false]
	].
	^true! !

! KeyedCollection methodsFor: #adding !
add: anAssociation
	" Bind anAssociation to the receiver. "

	self at: anAssociation key put: anAssociation value.
	^anAssociation! !

! KeyedCollection methodsFor: #adding !
addAll: aKeyedCollection
	" Bind all associations in aKeyedCollection to the receiver. "

	super addAll: aKeyedCollection associations.
	^aKeyedCollection! !

! KeyedCollection methodsFor: #private !
newHash
	" Private - Return a new hash object. "

	^{'' doc'native instantiation'}! !

! KeyedCollection methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	| result value |
	result := self newHash.
	self keysAndValuesDo: [:key :aValue |
		value := (self isObject: aValue)
			ifTrue: [ aValue asJSONObject ]
			ifFalse: [ aValue ].
		result basicAt: key put: value
	].
	^result! !

! KeyedCollection methodsFor: #accessing !
associations
	" Return the associations of the receiver. "

	| result |
	result := Array new.
	self associationsDo: [:each| result add: each ].
	^result! !

! KeyedCollection methodsFor: #enumerating !
associationsDo: aBlock
	" Evaluate aBlock for all the receiver's associations. "

	self keysAndValuesDo: [:key :value|
		aBlock value: (Association key: key value: value)
	].! !

! KeyedCollection methodsFor: #accessing !
at: aKey
	" Return the value at aKey.
	Signal an error if the key is not present.
	"

	^self at: aKey ifAbsent: [ self objectNotFoundAt: aKey ]! !

! KeyedCollection methodsFor: #accessing !
at: aKey ifAbsent: aBlock
	" Return the value at aKey or the result of evaluating aBlock. "

	^self implementedBySubclass: #at:ifAbsent:! !

! KeyedCollection methodsFor: #accessing !
values
	" Return the values of the receiver. "

	^self implementedBySubclass: #values! !

! KeyedCollection methodsFor: #accessing !
at: aKey ifAbsentPut: aBlock
	" Return the value at aKey, or add the result of evaluating aBlock to the receiver if it is absent. "

	^self at: aKey ifAbsent: [ self at: aKey put: aBlock value ]! !

! KeyedCollection methodsFor: #accessing !
at: aKey put: aValue
	" Set the value at aKey of the receiver. "

	^self implementedBySubclass: #at:put:! !

! KeyedCollection methodsFor: #enumerating !
detect: aBlock ifNone: anotherBlock
	" Detect the value who matches aBlock.
	Return the result of evaluating aBlock if no value matches detection block.
	"

	^self implementedBySubclass: #detect:ifNone:! !

! KeyedCollection methodsFor: #enumerating !
do: aBlock
	" Evaluates aBlock for all values in the receiver. "

	^self implementedBySubclass: #do:! !

! KeyedCollection methodsFor: #testing !
includes: anObject
	" Return true if anObject is in the receiver. "

	^self implementedBySubclass: #includes:! !

! KeyedCollection methodsFor: #accessing !
keyForValue: anObject ifAbsent: aBlock
	" Return the key in the receiver for value anObject, or the result of evaluating aBlock. "

	self keysAndValuesDo: [:key :value| anObject = value ifTrue: [ ^key ] ].
	^aBlock value! !

! KeyedCollection methodsFor: #accessing !
keys
	" Return the keys of the receiver. "

	^self implementedBySubclass: #keys! !

! KeyedCollection methodsFor: #enumerating !
keysAndValuesDo: aBlock
	" Evaluate aBlock for key/value pairs of the receiver. "

	^self implementedBySubclass: #keysAndValuesDo:! !

! KeyedCollection methodsFor: #enumerating !
keysDo: aBlock
	" Evaluate aBlock with the keys of the receiver. "

	^self implementedBySubclass: #keysDo:! !

! KeyedCollection methodsFor: #testing !
includesKey: aKey
	" Return true if the receiver's keys include aKey. "

	^self implementedBySubclass: #includesKey:! !

! KeyedCollection methodsFor: #removing !
removeKey: aKey ifAbsent: aBlock
	" Remove the binding at aKey. "

	^self implementedBySubclass: #removeKey:ifAbsent:! !

! KeyedCollection methodsFor: #removing !
removeKeys: aCollection
	" Remove aCollection of keys in the receiver. "

	aCollection do: [:each| self removeKey: each ifAbsent: [] ]! !

! KeyedCollection methodsFor: #removing !
removeAll
	" Remove all elements in the receiver. "

	self initialize! !

! KeyedCollection methodsFor: #testing !
isEmpty
	" Return true if the receiver is empty. "

	^self size = 0! !

! KeyedCollection methodsFor: #accessing !
at: aKey ifPresent: aBlock
	" Return the result of evaluating aBlock if there is a value at aKey (or nil). "

	| aValue |
	aValue := self at: aKey ifAbsent: [ ^nil ].
	^aBlock value: aValue! !

! KeyedCollection methodsFor: #accessing !
at: aKey ifPresent: aBlock ifAbsent: anotherBlock
	" Return the result of evaluating aBlock if there is a value at aKey (or the result of evaluating anotherBlock). "

	| aValue |
	aValue := self at: aKey ifAbsent: [ ^anotherBlock value ].
	^aBlock value: aValue! !

! KeyedCollection methodsFor: #enumerating !
collect: aBlock
	" Return aKeyedCollection with keys and result of evaluating aBlock with value. "

	| result |
	result := self species new.
	self keysAndValuesDo: [:key :value |
		result at: key put: (aBlock value: value)
	].
	^result! !

! KeyedCollection methodsFor: #enumerating !
select: aBlock
	" Return the contents of the receiver wich evaluates aBlock to true. "

	| result |
	result := self species new.
	self keysAndValuesDo: [:key :value |
		(aBlock value: value) ifTrue: [ result at: key put: value ]
	].
	^result! !

! KeyedCollection methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	| result |
	result := self species new.
	self keysAndValuesDo: [:key :value | result at: key put: value ].
	^result! !

! KeyedCollection methodsFor: #accessing !
tuples
	" Return an array of tuples built from the receiver's contents. "

	| result |
	result := Array new.
	self keysAndValuesDo: [:key :value|
		result add: (Array with: key with: value)
	].
	^result! !

! KeyedCollection methodsFor: #copying !
copyFrom: anIndex to: anotherIndex
	" Return a copy of the receiver from anIndex to anotherIndex. "

	^self shouldNotImplement: #copyFrom:to:! !

! KeyedCollection methodsFor: #copying !
deepCopy
	" Return a deep copy of the receiver. "

	| result |
	result := self species new.
	self keysAndValuesDo: [:key :value |
		result at: key put: value deepCopy
	].
	^result! !

! KeyedCollection methodsFor: #removing !
remove: aKey
	" Remove the binding at aKey. "

	self removeKey: aKey! !

! KeyedCollection methodsFor: #removing !
removeKey: aKey
	" Remove the binding at aKey.
	Signal an error if the key is not present.
	"

	^self removeKey: aKey ifAbsent: [ self objectNotFoundAt: aKey ]! !

! KeyedCollection methodsFor: #converting !
printString
	" Return a printable representation of the receiver. "

	^self class name ,' fromTuples: ' ,self tuples asLiteral! !

! KeyedCollection methodsFor: #testing !
isDictionary
	" Return true if the receiver is a dictionary. "

	^true! !

! Object methodsFor: #testing !
isDictionary
	" Return true if the receiver is a dictionary. "

	^false! !

! Dictionary class methodsFor: #instantiation !
basicNew
	" Return an instance of the receiver. "

	{'self' doc'native instantiation'}.
	^super basicNew! !

! Dictionary class methodsFor: #instantiation !
fromKeys: keys and: values
	" Return an instance of the receiver. "

	^self basicNew initializeKeys: keys values: values! !

! Dictionary methodsFor: #initialize !
initialize
	" Private - Initialize the receiver. "

	super initialize.
	keys := HashedSet new.
	values := #().! !

! Dictionary methodsFor: #initialize !
initializeKeys: privateKeys values: privateValues
	" Private - Initialize the receiver. "

	super initialize.
	keys := HashedSet newContents: privateKeys.
	values := privateValues asArray.! !

! Dictionary methodsFor: #accessing !
size
	" Return the size of the receiver. "

	^keys size! !

! Dictionary methodsFor: #testing !
isEmpty
	" Return true if the receiver is empty. "

	^keys isEmpty! !

! Dictionary methodsFor: #private !
lookUpKey: aKey
	" Private - Return the index where aKey is found (or nil). "

	^keys indexOf: aKey ifAbsent: [ nil ]! !

! Dictionary methodsFor: #enumerating !
keysAndValuesDo: aBlock
	" Evaluate aBlock for key/value pairs of the receiver. "

	keys withIndexDo: [:aKey :index | aBlock value: aKey value: (values at: index) ]! !

! Dictionary methodsFor: #accessing !
keys
	" Return the keys of the receiver. "

	^keys copy! !

! Dictionary methodsFor: #enumerating !
keysDo: aBlock
	" Evaluate aBlock with the keys of the receiver. "

	keys do: aBlock! !

! Dictionary methodsFor: #accessing !
values
	" Return the values of the receiver. "

	^values copy! !

! Dictionary methodsFor: #accessing !
at: aKey put: aValue
	" Set the value at aKey of the receiver. "

	| index |
	index := self lookUpKey: aKey.
	index isNil
		ifTrue: [ keys add: aKey. values add: aValue.]
		ifFalse: [ values at: index put: aValue ].
	^aValue! !

! Dictionary methodsFor: #accessing !
at: aKey ifAbsent: aBlock
	" Return the value at aKey or the result of evaluating aBlock. "

	| index |
	index := self lookUpKey: aKey.
	index isNil ifTrue: [ ^aBlock value ].
	^values at: index! !

! Dictionary methodsFor: #accessing !
at: aKey ifAbsentPut: aBlock
	" Return the value at aKey, or add the result of evaluating aBlock to the receiver if it is absent. "

	| index |
	index := self lookUpKey: aKey.
	index isNil ifTrue: [ ^self at: aKey put: aBlock value ].
	^values at: index! !

! Dictionary methodsFor: #enumerating !
do: aBlock
	" Evaluates aBlock for all values in the receiver. "

	^values do: aBlock! !

! Dictionary methodsFor: #enumerating !
detect: aBlock ifNone: anotherBlock
	" Detect the value who matches aBlock.
	Return the result of evaluating aBlock if no value matches detection block.
	"

	^values detect: aBlock ifNone: anotherBlock! !

! Dictionary methodsFor: #testing !
includes: anObject
	" Return true if anObject is in the receiver. "

	^values includes: anObject! !

! Dictionary methodsFor: #enumerating !
includesKey: aKey
	" Return true if the receiver's keys include aKey. "

	^keys includes: aKey! !

! Dictionary methodsFor: #removing !
removeKey: aKey ifAbsent: aBlock
	" Remove the binding at aKey. "

	| index |
	index := self lookUpKey: aKey.
	index isNil ifTrue: [ ^aBlock value ].
	keys removeIndex: index.
	values removeIndex: index.! !

! Dictionary methodsFor: #accessing !
keyForValue: anObject ifAbsent: aBlock
	" Return the key in the receiver for value anObject, or the result of evaluating aBlock. "

	self keysAndValuesDo: [:key :aValue| anObject == aValue ifTrue: [ ^key ] ].
	^aBlock value! !

! Dictionary methodsFor: #converting !
asNative: lang
	" Return the native representation of the receiver. "

	| send |
	send := lang = #lua ifTrue: [$:] ifFalse: [$.].
	^'smalltalk.' ,self class name
		,send ,'fromKeys_and_('
			,(keys asArray asNative: lang)
			,',' ,(values asNative: lang)
		,')'! !

! PoolDictionary class methodsFor: #instantiation !
@@ aHash
	" Return an instance of the receiver. "
	" The instance will hack aHash violating encapsulation of aHash and instance.
	 Both will share internal state. "

	^self basicNew
		instVarAt: #hash put: aHash;
		yourself! !

! PoolDictionary methodsFor: #initialize !
initialize
	" Private - Initialize the receiver. "

	super initialize.
	hash := self newHash.! !

! PoolDictionary methodsFor: #initialize !
validKey: aKey
	" Private - Returns a valid key to store a value in the receiver. "

	{'self'}.
	"(self typeOf: aKey) = #string ifFalse: [ ^self error: 'Invalid key (must be aString).' ]."
	^aKey ! !

! PoolDictionary methodsFor: #accessing !
at: aKey put: aValue
	" Set the value at aKey of the receiver. "

	size := nil.
	{'self'}.
	hash	basicAt: (self validKey: aKey)
		put: (self wrap: aValue).
	^aValue! !

! PoolDictionary methodsFor: #removing !
removeKey: aKey ifAbsent: aBlock
	" Remove the binding at aKey. "

	(self includesKey: aKey) ifFalse: [ ^aBlock value ].
	size := nil.
	{'' doc'native remotion'}! !

! PoolDictionary methodsFor: #accessing !
size
	" Return the size of the receiver. "

	size isNil ifTrue: [ size := self keys size ].
	^size! !

! PoolDictionary methodsFor: #accessing !
at: aKey ifAbsent: aBlock
	" Return the value at aKey or the result of evaluating aBlock. "

	{'' doc'native indexing'}! !

! PoolDictionary methodsFor: #testing !
includesKey: aKey
	" Return true if the receiver's keys include aKey. "

	^{'' doc'native testing'}! !

! PoolDictionary methodsFor: #accessing !
keys
	" Return the keys of the receiver. "

	{'' doc'native accessing'}! !

! PoolDictionary methodsFor: #enumerating !
keysAndValuesDo: aBlock
	" Evaluate aBlock for key/value pairs of the receiver. "

	self keysDo: [:key |
		aBlock value: key value: (self at: key)
	]! !

! PoolDictionary methodsFor: #accessing !
values
	" Return the values of the receiver. "

	^self keys collect: [:key| self at: key ]! !

! PoolDictionary methodsFor: #converting !
asNative: lang
	" Return the native representation of the receiver. "

	| send |
	send := lang = #lua ifTrue: [$:] ifFalse: [$.].
	^'smalltalk.' ,self class name
		,send ,'fromTuples_('
		,(self tuples asNative: lang)
		,')'! !

! PoolDictionary methodsFor: #enumerating !
detect: aBlock ifNone: anotherBlock
	" Detect the value who matches aBlock.
	Return the result of evaluating aBlock if no value matches detection block.
	"

	self keysAndValuesDo: [:key :value |
		(aBlock value: value) ifTrue: [ ^value ]
	].
	^anotherBlock value! !

! PoolDictionary methodsFor: #enumerating !
do: aBlock
	" Evaluates aBlock for all values in the receiver. "

	self keysAndValuesDo: [:key :value | aBlock value: value ]! !

! PoolDictionary methodsFor: #testing !
includes: anObject
	" Return true if anObject is in the receiver. "

	self keysAndValuesDo: [:key :value |
		anObject = value ifTrue: [ ^true ].
	].
	^false! !

! PoolDictionary methodsFor: #enumerating !
keysDo: aBlock
	" Evaluate aBlock with the keys of the receiver. "

	self keys do: aBlock! !

! PoolDictionary methodsFor: #accessing !
@ keyOrPath
	" Return the result of accessing the receiver at key or path (or nil). "

	| result |
	keyOrPath isArray ifFalse: [ ^self at: keyOrPath ifAbsent: [] ].
	result := keyOrPath inject: self into: [:pool :key | | value |
		(value := pool @ key) isNil ifTrue: [ ^nil ].
		value
	].
	^result! !

Object
 subclass: #ClassBuilder
 instanceVariableNames: ''
 category: #Kernel!

! ClassBuilder methodsFor: #instantiation !
superclass: aClass subclass: aString
	" Create/Register a new class in the system. "

	^self
		superclass: aClass
		subclass: aString
		instanceVariableNames: ''
		category: nil! !

!ClassBuilder methodsFor: #instantiation!
superclass: aClass
	subclass: aString
	instanceVariableNames: aString2
	classVariableNames: classVarNames
	poolDictionaries: poolVarNames
	category: aString3
	" Create/Register a new class in the system. "

	| result aCategory |
	aCategory := aString3 isNil ifTrue: [
			aClass isNil ifTrue: [ #unclassified ]
			ifFalse: [ aClass category ]
		] ifFalse: [ aString3 ].
	result := self
		addSubclassOf: aClass named: aString
		instanceVariableNames: (self instanceVariableNamesFor: aString2).
	self setupClass: result.
	result 
		classVariableNames: classVarNames;
		poolDictionaryNames: poolVarNames;
		category: aCategory.
	(aClass notNil and: [ Smalltalk includesKey: #Exporter ]) ifTrue: [
		self print: (Exporter new exportNew: result)
	].
	^result! !

!ClassBuilder methodsFor: #instantiation!
superclass: aClass
	subclass: aString
	instanceVariableNames: aString2
	category: aString3
	" Create/Register a new class in the system. "

	^self
		superclass: aClass
		subclass: aString
		instanceVariableNames: aString2
		classVariableNames: ''
		poolDictionaries: ''
		category: aString3! !

!ClassBuilder methodsFor: #edition!
class: aClass comment: aString
	" Set/Register the class comment in the system. "

	(aClass isNil or: [ aClass comment = aString ]) ifTrue: [ ^self ].
	aClass basicComment: aString.
	(Smalltalk includesKey: #Exporter) ifTrue: [
		self print: (Exporter new exportCommentOf: aClass)
	].! !

!ClassBuilder methodsFor: #private !
basicClass: aMetaClass instanceVariableNames: anArray
	" Private - Register instance variables onto aMetaClass. "

	^self smalltalk #metaVarNames: aMetaClass array: anArray! !

!ClassBuilder methodsFor: #metaclass !
class: aMetaClass instanceVariableNames: aString
	" Register instance variable names onto aMetaClass. "

	| varNames |
	aMetaClass isMetaclass ifFalse: [
		^self error: aMetaClass name, ' is not a metaclass'
	].
	varNames := self instanceVariableNamesFor: aString.
	varNames = aMetaClass instanceVariableNames ifFalse: [
		(self basicClass: aMetaClass instanceVariableNames: varNames) ifFalse: [ ^false ].
	].
	self setupClass: aMetaClass.
	(Smalltalk includesKey: #Exporter) ifTrue: [
		self print: (Exporter new exportMetaDefinitionOf: aMetaClass instanceClass)
	].
	^true ! !

! ClassBuilder methodsFor: #private !
instanceVariableNamesFor: aString
	" Private - Return the instance variable names string. "

	| result |
	result := aString asArrayOfSubstrings.
	result size = result asSet size ifFalse: [
		^self error: 'Invalid instance variable names ' ,result asLiteral
	].
	^result! !

! ClassBuilder methodsFor: #private !
addSubclassOf: aClass named: aString instanceVariableNames: aCollection
	" Private - Add the subclass to system. "

	{'' doc'native instantiation'}! !

! ClassBuilder methodsFor: #private !
setupClass: aClass
	" Private - Bind/initialize the class delegation chain. "! !

Object
 subclass: #Stream
 instanceVariableNames: 'collection position readLimit'
 category: #Kernel!

! Stream methodsFor: #accessing !
collection
	" Private - Return the collection of the receiver. "

	^collection! !

! Stream methodsFor: #accessing !
collection: aCollection
	" Private - Set the collection of the receiver. "

	collection := aCollection! !

! Stream methodsFor: #accessing !
setCollection: aCollection
	" Private - Set the receiver stream collection to aCollection. "

	self collection: aCollection! !

! Stream methodsFor: #initialize !
initialize
	" Private - Initialize the receiver. "

	super initialize.
	position := 0.! !

! Stream methodsFor: #accessing !
position
	" Return the position of the receiver. "

	^position! !

! Stream methodsFor: #accessing !
position: anInteger
	" Set the position of the receiver. "

	position := anInteger! !

! Stream methodsFor: #accessing !
readLimit
	" Private - Return the read limit of the receiver. "

	readLimit == nil ifTrue: [ readLimit := self collection size ].
	^readLimit! !

! Stream methodsFor: #accessing !
readLimit: anInteger
	" Private - Set the stream size of the receiver. "

	readLimit := anInteger! !

! Stream methodsFor: #accessing !
contents
	" Return the contents of the receiver. "

	| result |
	result := self collection.
	self readLimit = result size ifTrue: [ ^result ].
	^result copyFrom: 1 to: self readLimit! !

! Stream methodsFor: #accessing !
size
	" Return the size of the receiver. "

	^self readLimit! !

! Stream methodsFor: #actions !
reset
	" Reset the receiver. "

	self position: 0! !

! Stream methodsFor: #actions !
close
	" Close the receiver. "! !

! Stream methodsFor: #io !
flush
	" Flush the receiver. "! !

! Stream methodsFor: #actions !
resetContents
	" Reset the receiver contents. "

	self reset; readLimit: 0! !

! Stream methodsFor: #enumerating !
do: aBlock
	" Evaluates aBlock streaming on the receiver. "

	[ self atEnd ] whileFalse: [ aBlock value: self next ]! !

! Stream methodsFor: #positioning !
setToEnd
	" Position the receiver at end. "

	self position: self size! !

! Stream methodsFor: #positioning !
skip: anInteger
	" Skip next anInteger readable elements in the receiver. "

	self position: ((anInteger + self position min: self size) max: 0)! !

! Stream methodsFor: #reading !
next
	" Return the next readable element in the receiver (or nil). "

	self position: self position + 1. 
	^collection at: self position! !

! Stream methodsFor: #reading !
next: anInteger
	" Return the next anInteger elements read from the receiver.
	If the receiver reach its end, the returned collection has less elements.
	"

	| result |
	result := self collection species new.
	anInteger timesRepeat: [
		self atEnd ifTrue: [ ^result ].
		result add: self next.
	].
	^result! !

! Stream methodsFor: #writing !
nextPut: anObject
	" Put anObject onto the receiver. "

	self position: self position + 1.
	self collection at: position put: anObject.
	self readLimit: (self readLimit max: position)! !

! Stream methodsFor: #writing !
nextPutAll: aCollection
	" Put aCollection onto the receiver. "

	aCollection do: [:each | self nextPut: each ]! !

! Stream methodsFor: #writing !
cr
	" Put a line break sequence onto the receiver.
	Note: this implementation MUST be consistent with #nextLine
	"

	self nextPutAll: String crlf! !

! Stream methodsFor: #writing !
space
	" Put a space sequence onto the receiver. "

	self nextPutAll: String space! !

! Stream methodsFor: #writing !
tab
	" Put a tab sequence onto the receiver. "

	self nextPutAll: String tab! !

! Stream methodsFor: #reading !
peek
	" Return the next readable element from the receiver without advancing the receiver (or nil). "

	^self atEnd ifFalse: [
	    self collection at: self position + 1
	]! !

! Stream methodsFor: #reading !
peekFor: anObject
	" Return true if the next readable element from the receiver equals anObject.
	Advance the receiver if the result is true.
	"

	^anObject = self peek and: [
		self atEnd ifFalse: [ self next ].
		true
	]! !

! Stream methodsFor: #reading !
peekForAll: aCollection
	" Return true if the next readable elements from the receiver matches aCollection.
	Advance the receiver if the result is true.
	"

	| back |
	back := self position.
	aCollection do: [:each|
		(self peekFor: each) ifFalse: [
			self position: back.
			^false
		].
	].
	^true! !

! Stream methodsFor: #testing !
atEnd
	" Return true if the receiver is at end. "

	^self position >= self size! !

! Stream methodsFor: #testing !
atStart
	" Return true if the receiver is at start position. "

	^self position = 0! !

! Stream methodsFor: #testing !
isEmpty
	" Return true if the receiver is empty. "

	^self size = 0! !

! Stream methodsFor: #testing !
notEmpty
	" Return true if the receiver is not empty. "

	^self isEmpty not! !

! Stream methodsFor: #reading !
upTo: anObject
	" Return the elements read from the receiver upto anObject (or end). "

	| result |
	result := self collection species new.
	[ self atEnd or: [ self peek = anObject ]] whileFalse: [
		result add: self next
	].
	self atEnd ifFalse: [ self next ].
	^result! !

! Stream methodsFor: #reading !
upToEnd
	" Return the contents of the receiver upto end. "

	^self upTo: Object new! !

! Stream methodsFor: #reading !
nextLine
	" Return the characters read from the receiver up to the line delimiter. "

	| result start |
	start := self position.
	result := self upTo: String cr.
	(result includes: String lf) ifTrue: [
		^self position: start; upTo: String lf
	].
	(self peek = String lf) ifTrue: [ self next ].
	^result! !

! Stream methodsFor: #reading !
skipTo: anObject
	" Advance the receiver upto anObject (or end). "

	[ self atEnd ] whileFalse: [
		self next = anObject ifTrue: [ ^true ].
	].
	^false! !

! Stream methodsFor: #reading !
skipSeparators
	" Advance the receiver upto a non-separator (or end).
	Return true  if a non-separator contents is found,
	return false if run upto end of stream.
	"

	[ self atEnd ifTrue: [ ^false ].
	  self peek isSeparator
	] whileTrue: [ self next ].
	^true! !

! Stream class methodsFor: #instantiation !
on: aCollection
	" Return an instance of the receiver. "

	^self new 
		collection: aCollection;
		readLimit: aCollection size;
		yourself! !

! Stream methodsFor: #reading !
upToEnd
	" Return the contents of the receiver upto end. "

	| result |
	result := self collection copyFrom: self position + 1 to: self readLimit.
	self setToEnd.
	^result! !

Stream
 subclass: #StringStream
 instanceVariableNames: ''
 category: #Kernel!

! StringStream methodsFor: #reading !
next: anInteger
	" Return the next anInteger elements read from the receiver.
	If the receiver reach its end, the returned collection has less elements.
	"

	| start result stop |
	start := self position.
	stop := start + anInteger min: self readLimit.
	result := self collection copyFrom: start + 1 to: stop.
	self position: stop.
	^result! !

! StringStream methodsFor: #writing !
nextPut: aString
	" Put aString onto the receiver. "

	self nextPutAll: aString! !

! StringStream methodsFor: #writing !
nextPutAll: aString
	" Put aString onto the receiver at current position; expanding internal contents. "

	| result |
	self position = self collection size
	ifTrue: [ result := self collection , aString ]
	ifFalse: [ | substring |
		substring := self collection copyFrom: 1 to: self position.
		result := substring isEmpty ifTrue: [ aString ] ifFalse: [ substring , aString ].
		substring := self collection
			copyFrom: self position + 1 + aString size
			to: self collection size.
		result := substring isEmpty ifTrue: [ result ] ifFalse: [ result , substring ].
	].
	self collection: result.
	self position: self position + aString size.
	self readLimit: (self readLimit max: position)! !

! StringStream methodsFor: #reading !
skipTo: anObject
	" Advance the receiver upto anObject (or end). "

	| index |
	anObject isString ifFalse: [ self setToEnd. ^false ].
	index := self collection indexOf: anObject startingAt: self position + 1.
	(index between: self position + 1 and: self readLimit) ifTrue: [
		self position: index + anObject size - 1.
		^true
	].
	self setToEnd.
	^false! !

! StringStream methodsFor: #reading !
upTo: anObject
	" Return the elements read from the receiver upto anObject (or end). "

	| start found |
	anObject isString ifFalse: [ ^self upToEnd ].
	start := self position + 1.
	found := (self skipTo: anObject) ifTrue: [ anObject size ] ifFalse: [ 0 ].
	^self collection copyFrom: start to: self position - found! !

! StringStream methodsFor: #reading !
peek
	" Return the next readable element from the receiver without advancing (or nil). "

	{'' doc'native accessing'}! !

! StringStream methodsFor: #reading !
next
	" Return the next readable element in the receiver (or nil). "

	| result |
	result := self peek.
	result notNil ifTrue: [ position := position + 1 ].
	^result! !

! StringStream methodsFor: #reading !
peekForAll: aCollection
	" Return true if the next readable elements from the receiver matches aCollection.
	Advance the receiver if the result is true.
	"

	aCollection isString ifFalse: [ ^super peekForAll: aCollection ].
	{'' doc'native accessing'}.
	position := position + aCollection size.
	^true! !

! StringStream methodsFor: #reading !
upToSeparator
	" Return the elements read from the receiver upto a separator (or end). "

	| start |
	start := self position + 1.
	[self atEnd not and: [self peek isSeparator not]] whileTrue: [
		self next
	].
	^self collection copyFrom: start to: self position! !

! StringStream methodsFor: #tools !
nextMessagePattern
	" Return a message pattern read from the receiver (or nil).
	Note: it is assumed that there is a correct message header in the receiver.
	"

	| part selector args |
	args := #().
	self skipSeparators.
	selector := self upToSeparator.
	selector isEmpty ifTrue: [ ^nil ].
	selector argumentCount > 0 ifTrue: [
		[ self skipSeparators.
		  args add: self upToSeparator.
		  self skipSeparators.
		  part := self peek = $" ifTrue: [ '' ]
			ifFalse: [ self upToSeparator ].
		  part notEmpty and: [ part last = $: ]
		] whileTrue: [
		  selector := selector , part
		].
	].
	part := Message selector: selector arguments: args.
	self skipSeparators.
	self peek = $" ifTrue: [
		part comment: (self next; upTo: $") trimBlanks
	].
	^part! !

! String class methodsFor: #streams !
streamClass
	" Return the streamming support for the receiver's instances. "

	^StringStream! !

Stream
	subclass: #StringWriteStream
	instanceVariableNames: ''
	category: #Kernel !

! StringWriteStream methodsFor: #accessing !
collection
	" Private - Return the collection of the receiver. "

	^self invalidMessage: #collection! !

! StringWriteStream methodsFor: #accessing !
collection: aCollection
	" Private - Set the collection of the receiver. "

	aCollection isString ifFalse: [ self error: 'Collection must be a String' ].
	collection := Array with: aCollection! !

! StringWriteStream methodsFor: #accessing !
contents
	" Return the contents of the receiver. "

	^collection join: ''! !

! StringWriteStream methodsFor: #accessing !
position
	" Return the position of the receiver. "

	^self readLimit! !

! StringWriteStream methodsFor: #accessing !
position: anInteger
	" Set the position of the receiver. "

	^self invalidMessage: #position:! !

! StringWriteStream methodsFor: #accessing !
readLimit
	" Private - Return the read limit of the receiver. "

	^readLimit! !

! StringWriteStream methodsFor: #accessing !
readLimit: anInteger
	" Private - Set the stream size of the receiver. "

	readLimit := anInteger! !

! StringWriteStream methodsFor: #accessing !
setCollection: aCollection
	" Private - Set the receiver stream collection to aCollection. "

	^self invalidMessage: #setCollection:! !

! StringWriteStream methodsFor: #accessing !
size
	" Return the size of the receiver. "

	^self readLimit! !

! StringWriteStream methodsFor: #actions !
close
	" Close the receiver. "! !

! StringWriteStream methodsFor: #actions !
reset
	" Reset the receiver. "

	self initialize! !

! StringWriteStream methodsFor: #enumerating !
do: aBlock
	" Evaluates aBlock streaming on the receiver. "

	^self invalidMessage: #do:! !

! StringWriteStream methodsFor: #initialize !
initialize
	" Private - Initialize the receiver. "

	super initialize.
	collection := Array new.
	readLimit := position := 0.! !

! StringWriteStream methodsFor: #positioning !
setToEnd
	" Position the receiver at end. "! !

! StringWriteStream methodsFor: #positioning !
skip: anInteger
	" Skip next anInteger readable elements in the receiver. "

	^self invalidMessage: #skip:! !

! StringWriteStream methodsFor: #reading !
next
	" Return the next readable element in the receiver (or nil). "

	^self invalidMessage: #next! !

! StringWriteStream methodsFor: #reading !
next: anInteger
	" Return the next anInteger elements read from the receiver. "

	^self invalidMessage: #next! !

! StringWriteStream methodsFor: #reading !
nextLine
	" Return the characters read from the receiver up to the line delimiter. "

	^self invalidMessage: #nextLine! !

! StringWriteStream methodsFor: #reading !
peek
	" Return the next readable element from the receiver without advancing the receiver (or nil). "

	^self invalidMessage: #peek! !

! StringWriteStream methodsFor: #reading !
upTo: anObject
	" Return the elements read from the receiver upto anObject (or end). "

	^self invalidMessage: #upTo:! !

! StringWriteStream methodsFor: #reading !
upToEnd
	" Return the contents of the receiver upto end. "

	^self invalidMessage: #upToEnd! !

! StringWriteStream methodsFor: #testing !
atEnd
	" Return true if the receiver is at end. "

	^true! !

! StringWriteStream methodsFor: #writing !
nextPut: anObject
	" Put anObject onto the receiver. "

	| string |
	string :=  anObject asString.
	collection add: string.
	readLimit := readLimit + string size! !

! StringWriteStream methodsFor: #writing !
nextPutAll: aCollection
	" Put aCollection onto the receiver. "

	self nextPut: aCollection! !

! StringWriteStream methodsFor: #back !
removeLastInput
	" Remove last input recorded in the receiver. "

	readLimit := readLimit - collection removeLast size! !

! String methodsFor: #streams !
writeStream
	" Return a write stream on the receiver. "

	^StringWriteStream on: self! !

Object
 subclass: #Random
 instanceVariableNames: ''
 category: #Kernel!

! Random methodsFor: #accessing !
next
	" Return the next random number. "

	{'' doc'native accessing'}! !

! Random methodsFor: #accessing !
next: anInteger
	" Return the next anInteger random numbers. "

	^(1 to: anInteger) collect: [:each | self next ]! !

Object
 subclass: #Point
 instanceVariableNames: 'x y'
 category: #Kernel!

! Point class methodsFor: #instantiation !
x: x y: y
	" Return an instance of the receiver. "

	^self basicNew x: x; y: y! !

! Point class methodsFor: #instantiation !
zero
	" Return an instance of the receiver. "

	^self basicNew x: 0 y: 0! !

! Point methodsFor: #accessing !
x
	" Return the x of the receiver. "

	^x! !

! Point methodsFor: #accessing !
y
	" Return the y of the receiver. "

	^y! !

! Point methodsFor: #accessing !
x: xValue y: yValue
	" Set the receiver's coordinates. "

	x := xValue.
	y := yValue! !

! Point methodsFor: #accessing !
y: aNumber
	" Set the y of the receiver. "

	y := aNumber isNil ifTrue: [ 0 ] ifFalse: [ aNumber ]! !

! Point methodsFor: #accessing !
x: aNumber
	" Set the x of the receiver. "

	x := aNumber isNil ifTrue: [ 0 ] ifFalse: [ aNumber ]! !

! Point methodsFor: #arithmetic !
* aPoint
	" Return the result of operation. "

	(self isNumber: aPoint) ifTrue: [ ^x * aPoint @ (y * aPoint) ].
	^Point	x: x * aPoint asPoint x
		y: y * aPoint asPoint y! !

! Point methodsFor: #arithmetic !
+ aPoint
	" Return the result of operation. "

	(self isNumber: aPoint) ifTrue: [ ^x + aPoint @ (y + aPoint) ].
	^Point	x: x + aPoint asPoint x
		y: y + aPoint asPoint y! !

! Point methodsFor: #arithmetic !
- aPoint
	" Return the result of operation. "

	(self isNumber: aPoint) ifTrue: [ ^x - aPoint @ (y - aPoint) ].
	^Point	x: x - aPoint asPoint x
		y: y - aPoint asPoint y! !

! Point methodsFor: #arithmetic !
/ aPoint
	" Return the result of operation. "

	(self isNumber: aPoint) ifTrue: [ ^x / aPoint @ (y / aPoint) ].
	^Point	x: x / aPoint asPoint x
		y: y / aPoint asPoint y! !

! Point methodsFor: #arithmetic !
// aPoint
	" Return the result of operation. "

	(self isNumber: aPoint) ifTrue: [ ^x // aPoint @ (y // aPoint) ].
	^Point	x: x // aPoint asPoint x
		y: y // aPoint asPoint y! !

! Point methodsFor: #arithmetic !
\\\\ aPoint
	" Return the result of operation. "

	(self isNumber: aPoint) ifTrue: [ ^x \\ aPoint @ (y \\ aPoint) ].
	^Point	x: x \\ aPoint asPoint x
		y: y \\ aPoint asPoint y! !

! Point methodsFor: #comparing !
< aPoint
	" Return true if the receiver is smaller than argument. "

	(self isNumber: aPoint) ifTrue: [ ^self < aPoint asPoint ].
	^x < aPoint x and: [ y < aPoint y ]! !

! Point methodsFor: #comparing !
<= aPoint
	" Return true if the receiver is smaller or equal to argument. "

	(self isNumber: aPoint) ifTrue: [ ^self <= aPoint asPoint ].
	^x <= aPoint x and: [ y <= aPoint y ]! !

! Point methodsFor: #comparing !
> aPoint
	" Return true if the receiver is greater than argument. "

	(self isNumber: aPoint) ifTrue: [ ^self > aPoint asPoint ].
	^x > aPoint x and: [ y > aPoint y ]! !

! Point methodsFor: #comparing !
>= aPoint
	" Return true if the receiver is greater or equal to argument. "

	(self isNumber: aPoint) ifTrue: [ ^self >= aPoint asPoint ].
	^x >= aPoint x and: [ y >= aPoint y ]! !

! Point methodsFor: #converting !
rounded
	" Return the receiver rounded. "

	^x rounded @ y rounded! !

! Point methodsFor: #converting !
truncated
	" Return the receiver truncated. "

	^x truncated @ y truncated! !

! Point methodsFor: #converting !
asPoint
	" Return the receiver as a Point. "

	^self! !

! Point methodsFor: #printing !
printString
	" Return the printable representation of the receiver. "

	^x toString ,' @ ' ,y toString! !

! Point methodsFor: #printing !
toString
	" Return the printable representation of the receiver. "

	^x toString ,$@ ,y toString! !

! Point methodsFor: #converting !
asLiteral
	" Return the receiver in literal notation. "

	| literal |
	literal := [:v| | string |
		string := (self isObject: v) ifTrue: [v asLiteral] ifFalse: [v toString].
		(string isEmpty or: [string first isDigit or: ['#''-+' includes: string first]])
		ifTrue: [string] ifFalse: [ $( ,string ,$) ]
	].
	^(literal value: x) ,$@ ,(literal value: y)! !

! Point methodsFor: #copying !
shallowCopy
	" Return a shallow copy of the receiver. "

	^self class basicNew
		x: self x y: self y! !

! Point methodsFor: #testing !
= aPoint
	" Return true if the receiver is equal to argument. "

	aPoint isNil ifTrue: [ ^false ].
	(aPoint isMemberOf: self class) ifFalse: [
		^(aPoint respondsTo: #asPoint) and: [
			aPoint asPoint = self
		]
	].
	^x = aPoint x and: [ y = aPoint y ]! !

! Point methodsFor: #magnitude !
min: aPoint
	" Return the minimum of the receiver and aPoint. "

	^(x min: aPoint x) @ (y min: aPoint y)! !

! Point methodsFor: #magnitude !
max: aPoint
	" Return the maximum of the receiver and aPoint. "

	^(x max: aPoint x) @ (y max: aPoint y)! !

Object
 subclass: #Rectangle
 instanceVariableNames: 'leftTop rightBottom'
 category: #Kernel!

! Object methodsFor: #testing !
isRectangle
	" Return true if receiver is a Rectangle. "

	^false! !

! Point methodsFor: #converting !
extent: extent
	" Return a Rectangle from the receiver with specified extent. "

	^Rectangle leftTop: self rightBottom: self + extent! !

! Point methodsFor: #converting !
rightBottom: rightBottom
	" Return a Rectangle from the receiver to rightBottom. "

	^Rectangle leftTop: self rightBottom: rightBottom! !

! Point methodsFor: #converting !
to: rightBottom
	" Return a Rectangle from the receiver to rightBottom. "

	^Rectangle leftTop: self rightBottom: rightBottom! !

! Rectangle class methodsFor: #instantiation !
leftTop: leftTop extent: extent
	" Return an instance of the receiver. "

	^self new leftTop: leftTop rightBottom: leftTop + extent! !

! Rectangle class methodsFor: #instantiation !
leftTop: leftTop rightBottom: rightBottom
	" Return an instance of the receiver. "

	^self new leftTop: leftTop rightBottom: rightBottom! !

! Rectangle methodsFor: #testing !
= aRectangle
	" Return true if the receiver is equal to argument. "

	^aRectangle isRectangle
	and: [ leftTop = aRectangle leftTop
	and: [ rightBottom = aRectangle rightBottom ]]! !

! Rectangle methodsFor: #accessing !
area
	" Return the area of the receiver."

	^self height * self width! !

! Rectangle methodsFor: #accessing !
areasOutside: aRectangle
	"Return a collection of rectangles where the receiver do not intersect aRectangle. "

	(self intersects: aRectangle) ifFalse: [ ^Array with: self ].
	((aRectangle containsPoint: self topLeft)
	and: [aRectangle containsPoint: self bottomRight])
        ifTrue: [ ^#() ].
	^(Array
	with: (aRectangle top > self top ifTrue: [ self topLeft rightBottom: self right @ aRectangle top])
	with: (aRectangle bottom < self bottom ifTrue: [ self left @ aRectangle bottom rightBottom: self bottomRight])
	with: (aRectangle left > self left ifTrue: [ self left @ (self top max: aRectangle top) rightBottom: aRectangle left @ (self bottom min: aRectangle bottom) ])
	with: (aRectangle right < self right ifTrue: [ aRectangle right @ (self top max: aRectangle top) rightBottom: self right @ (self bottom min: aRectangle bottom) ])
	) select: [:each| each notNil ]! !

! Rectangle methodsFor: #accessing !
bottom
	" Return the bottom of the receiver."

	^rightBottom y! !

! Rectangle methodsFor: #accessing !
bottomCenter
	" Return the center of the bottom edge of the receiver. "

	^self center x @ self bottom! !

! Rectangle methodsFor: #accessing !
bottomLeft
	" Return the bottom-left corner of the receiver. "

	^self left @ self bottom! !

! Rectangle methodsFor: #accessing !
bottomRight
	" Return the bottom-right corner of the receiver. "

	^rightBottom! !

! Rectangle methodsFor: #accessing !
center
	" Return the center of the receiver. "

	^leftTop + rightBottom // 2! !

! Rectangle methodsFor: #accessing !
centerTop
	" Return the center top of the receiver. "

	^self right + self left // 2 @ self top! !

! Rectangle methodsFor: #testing !
contains: aRectangle
	" Return true if aRectangle lies completely within or on the edge of the receiver. "

	^self left <= aRectangle left
	and: [ aRectangle right <= self right
	and: [ self bottom >= aRectangle bottom
	and: [ aRectangle top >= self top ]]]! !

! Rectangle methodsFor: #testing !
containsPoint: aPoint
	" Return true if aPoint is contained within the receiver. "

	^(self left <= aPoint x)
	and: [ aPoint x < self right
	and: [ self bottom > aPoint y
	and: [ aPoint y >= self top ]]]! !

! Rectangle methodsFor: #accessing !
expandBy: delta
	" Return a Rectangle obtained from the receiver expanded by delta. "

	delta isRectangle ifTrue: [
            ^leftTop - delta leftTop rightBottom: rightBottom + delta rightBottom
	].
	^leftTop - delta rightBottom: rightBottom + delta! !

! Rectangle methodsFor: #accessing !
extent
	" Return receiver's width and height."

	^self width @ self height! !

! Rectangle methodsFor: #accessing !
extent: aPoint
	" Set the extent of receiver to aPoint. "

	self width: aPoint x.
	self height: aPoint y! !

! Rectangle methodsFor: #accessing !
height
	" Return the height of the receiver. "

	^(leftTop y - rightBottom y) abs! !

! Rectangle methodsFor: #accessing !
height: aNumber
	" Change the height of the receiver. "

	rightBottom y: leftTop y + aNumber! !

! Rectangle methodsFor: #accessing !
insetBy: delta
	" Return the receiver inset by delta. "

	delta isRectangle ifTrue: [
            ^leftTop + delta leftTop rightBottom: rightBottom - delta rightBottom
	].
	^leftTop + delta rightBottom: rightBottom - delta! !

! Rectangle methodsFor: #accessing !
intersect: aRectangle
	" Return the rectangle in which the receiver and aRectangle overlap. "

	aRectangle isNil ifTrue: [ ^self ].
	^(leftTop max: aRectangle leftTop)
        rightBottom: (rightBottom min: aRectangle rightBottom)! !

! Rectangle methodsFor: #accessing !
intersects: aRectangle
	" Return true if the receiver and aRectangle have any area in common. "

	^leftTop < aRectangle rightBottom
	and: [ aRectangle leftTop < rightBottom ]! !

! Rectangle methodsFor: #testing !
isRectangle
	" Return true if receiver is a Rectangle. "

	^true! !

! Rectangle methodsFor: #accessing !
left
	" Return the left of the receiver. "

	^leftTop x! !

! Rectangle methodsFor: #accessing !
leftBottom
	" Return the left bottom corner of the receiver. "

	^self left @ self bottom! !

! Rectangle methodsFor: #accessing !
leftCenter
	" Return the center of the left edge of the receiver. "

	^self left @ self center y! !

! Rectangle methodsFor: #accessing !
leftTop
	" Return the left top of the receiver. "

	^leftTop! !

! Rectangle methodsFor: #accessing !
leftTop: aPoint
	" Set the leftTop of the receiver."

	leftTop := aPoint! !

! Rectangle methodsFor: #accessing !
leftTop: leftTopPoint rightBottom: rightBottomPoint
	"Set the receiver's coordinates. "

	leftTop := leftTopPoint.
	rightBottom := rightBottomPoint.! !

! Rectangle methodsFor: #accessing !
merge: aRectangle
	" Return a Rectangle representing the area in which the receiver and aRectangle merged together. "

	aRectangle isNil ifTrue: [ ^self ].
	^(leftTop min: aRectangle leftTop) rightBottom: (rightBottom max: aRectangle rightBottom)! !

! Rectangle methodsFor: #accessing !
moveBy: aPoint
	" Move the receiver by aPoint. "

	leftTop := leftTop + aPoint.
	rightBottom := rightBottom + aPoint! !

! Rectangle methodsFor: #accessing !
moveTo: aPoint
	" Move the receiver to aPoint. "

	self moveBy: aPoint - leftTop! !

! Rectangle methodsFor: #accessing !
moveToLeftTop: aPoint
	" Move the receiver to leftTop. "

	self moveBy: aPoint - self leftTop! !

! Rectangle methodsFor: #accessing !
nonIntersections: aRectangle
	" Return the rectangles describing areas of the receiver outside aRectangle. "

	| areas |
	(self intersects: aRectangle) ifFalse: [ ^Array with: self ].
	((aRectangle containsPoint: leftTop)
		and: [ aRectangle containsPoint: rightBottom ])
		ifTrue: [ ^#() ].
	^(Array
	with: (aRectangle top > self top ifTrue: [ leftTop rightBottom: rightBottom x @ aRectangle top ])
	with: (aRectangle bottom < self bottom ifTrue: [ leftTop x @ aRectangle bottom rightBottom: rightBottom ])
	with: (aRectangle left > self left ifTrue: [ leftTop x @ (leftTop y max: aRectangle top)
                rightBottom: aRectangle left @ (rightBottom y min: aRectangle bottom) ])
	with: (aRectangle right < self right ifTrue: [ aRectangle right @ (leftTop y max: aRectangle top)
                rightBottom: rightBottom x @ (rightBottom y min: aRectangle bottom) ])
	) select: [:each| each notNil ]! !

! Rectangle methodsFor: #printing !
toString
	" Return the printable representation of the receiver. "

	^leftTop toString ,' to: ' ,rightBottom toString ! !

! Rectangle methodsFor: #printing !
printString
	" Return the printable representation of the receiver. "

	{'self'}.
	^self toString! !

! Rectangle methodsFor: #accessing !
relativeTo: aRectangle
	" Return a Rectangle proportionally relative to aRectangle. "

	| extent |
	extent := aRectangle extent.
	(aRectangle extent x = 0 or: [ aRectangle extent y = 0 ]) ifTrue: [ ^self ].
	^leftTop - aRectangle leftTop / extent extent: self extent / extent! !

! Rectangle methodsFor: #accessing !
right
	" Return the receiver's right side. "

	^rightBottom x! !

! Rectangle methodsFor: #accessing !
rightBottom
	" Return the right bottom corner. "

	^rightBottom! !

! Rectangle methodsFor: #accessing !
rightBottom: aPoint
	" Set the right bottom corner of the receiver. "

	rightBottom := aPoint! !

! Rectangle methodsFor: #accessing !
rightCenter
	" Return the center of the right edge of the receiver. "

	^self right @ self center y! !

! Rectangle methodsFor: #accessing !
rightTop
        " Return the right top of the receiver. "

	^self right @ self top! !

! Rectangle methodsFor: #accessing !
rounded
	" Set the receiver with coordinates rounded to integers. "

	leftTop := leftTop rounded.
	rightBottom := rightBottom rounded.! !

! Rectangle methodsFor: #accessing !
scaleBy: delta
	" Return the receiver multiplied by delta. "

	^leftTop * delta rightBottom: rightBottom * delta! !

! Rectangle methodsFor: #accessing !
scaleTo: aRectangle
	" Return a Rectangle whose size is proportional to the receiver with ratios specified by aRectangle. "

	^(aRectangle scaleBy: self extent) truncated moveBy: self leftTop! !

! Rectangle methodsFor: #accessing !
top
	" Return the top side of the receiver. "

	^leftTop y! !

! Rectangle methodsFor: #accessing !
topCenter
	" Return the center of the top edge of the receiver. "

	^self center x @ self top! !

! Rectangle methodsFor: #accessing !
topLeft
	" Return the top-left corner of the receiver. "

	^leftTop! !

! Rectangle methodsFor: #accessing !
topLeft: aPoint
	"Set the top left corner of the receiver. "

	leftTop := aPoint! !

! Rectangle methodsFor: #accessing !
topRight
	" Return the top-right corner of the receiver. "

	^self right @ self top! !

! Rectangle methodsFor: #accessing !
translateBy: delta
	" Return the receiver with position incremented by delta. "

	^leftTop + delta rightBottom: rightBottom + delta! !

! Rectangle methodsFor: #accessing !
truncated
	" Set the receiver with the coordinates truncated to integers. "

	leftTop := leftTop truncated.
	rightBottom := rightBottom truncated.! !

! Rectangle methodsFor: #accessing !
width
	" Return the width of the receiver. "

	^(rightBottom x - leftTop x) abs! !

! Rectangle methodsFor: #accessing !
width: anInteger
	" Set the width of the receiver. "

	rightBottom x: leftTop x + anInteger! !

Object
 subclass: #EvaluableAction
 instanceVariableNames: ''
 category: #Kernel!

EvaluableAction comment: '
	Action model, evaluable and composable actions are implemented to support evaluations.
'!

EvaluableAction
 subclass: #Message
 instanceVariableNames: 'selector arguments receiver'
 category: #Kernel!

! Message class methodsFor: #instantiation !
selector: aString arguments: anArray
	" Return an instance of the receiver. "

	^self new
		selector: aString;
		arguments: anArray;
		yourself! !

! Message class methodsFor: #instantiation !
receiver: anObject selector: aString arguments: anArray
	" Return an instance of the receiver. "

	^self new
		receiver: anObject;
		selector: aString;
		arguments: anArray;
		yourself! !

! Message methodsFor: #accessing !
selector
	" Return the selector of the receiver. "

	^selector! !

! Message methodsFor: #accessing !
selector: aString
	" Set the selector of the receiver. "

	selector := aString! !

! Message methodsFor: #accessing !
arguments: anArray
	" Set the arguments of the receiver. "

	arguments := anArray! !

! Message methodsFor: #accessing !
arguments
	" Return the arguments of the receiver. "

	^arguments! !

! Message methodsFor: #accessing !
receiver
	" Return the receiver of the receiver. "

	^receiver! !

! Message methodsFor: #accessing !
receiver: anObject
	" Set the receiver of the receiver. "

	receiver := anObject! !

! Message methodsFor: #accessing !
comment
	" Return the comment of the receiver (or nil). "

	^self basicAt: '@comment'! !

! Message methodsFor: #accessing !
comment: aString
	" Set the comment of the receiver. "

	aString isNil ifTrue: [ ^self basicDelete: '@comment' ].
	self basicAt: '@comment' put: aString! !

! Message methodsFor: #private !
perform
	" Return the result of sending the receiver. "

	^self receiver
		perform: self selector
		withArguments: self arguments! !

! Message methodsFor: #evaluating !
performDeferred
	" Evaluate the receiver deferred in time.
	Return the receiver.
	"

	[ self perform ] valueDeferred.! !

! Message methodsFor: #evaluating !
evaluate
	" Return the result of evaluating the receiver. "

	^self perform! !

! Message methodsFor: #evaluating !
evaluateFor: anObject
	" Return the result of evaluating the receiver with anObject. "

	^self evaluateWithArguments: (Array with: anObject)! !

! Message methodsFor: #evaluating !
evaluateWithArguments: anArray
	" Return the result of evaluating the receiver. "

	^self receiver
		perform: self selector
		withArguments: (self collectArguments: anArray)! !

! Message methodsFor: #private !
initializeArguments
	" Private - Initialize the arguments of the receiver. "

	arguments := Array new: self selector argumentCount! !

! Message methodsFor: #private !
argumentCount
	" Return the number of arguments of the receiver. "

	arguments isNil ifTrue: [ self initializeArguments ].
	^arguments size! !

! Message methodsFor: #private !
collectArguments: evaluationArguments
	" Private - Return the arguments to evaluate the receiver. "

	arguments size = evaluationArguments size ifTrue: [ ^evaluationArguments ].
	^evaluationArguments copyFrom: 1 to: (self argumentCount min: evaluationArguments size)! !

! Message methodsFor: #resolution !
resolve: what with: aPoolOrBlock
	" Private - Return the receiver to be evaluated with mappings given by aPoolOrBlock. "

	(aPoolOrBlock isMemberOf: PoolDictionary) ifFalse: [
		^aPoolOrBlock evaluateWith: what
	].
	what isNil ifTrue: [ ^what ].
	(self isObject: what) ifFalse: [ ^what ].
	what isString ifTrue: [ ^aPoolOrBlock at: what ifAbsent: [ what ] ].
	^what! !

! Message methodsFor: #resolution !
resolveWith: aPoolOrBlock
	" Return a message set to be evaluated with mappings given by aPoolOrBlock. "

	| args |
	args := arguments notNil ifTrue: [
		arguments collect: [:each|
			self resolve: each with: aPoolOrBlock
		]
	].
	^self class
		receiver: (self resolve: receiver with: aPoolOrBlock)
		selector: self selector
		arguments: args! !

! Message methodsFor: #testing !
isMessage
	" Return true if the receiver is a message. "

	^true! !

! Object methodsFor: #testing !
isMessage
	" Return true if the receiver is a message. "

	^false! !

Error
 subclass: #MessageNotUnderstood
 instanceVariableNames: 'message receiver'
 category: #Kernel!

! MessageNotUnderstood methodsFor: #accessing !
message
	" Return the message of the receiver. "

	^message! !

! MessageNotUnderstood methodsFor: #accessing !
message: aMessage
	" Set the message of the receiver. "

	message := aMessage! !

! MessageNotUnderstood methodsFor: #accessing !
receiver
	" Return the receiver of the receiver. "

	^receiver! !

! MessageNotUnderstood methodsFor: #accessing !
receiver: anObject
	" Set the receiver of the receiver. "

	receiver := anObject! !

! MessageNotUnderstood methodsFor: #accessing !
messageText
	" Return the message text of the receiver. "

	| type |
	message isNil ifTrue: [ ^#unknown ].
	type := (self isObject: receiver)
		ifTrue: [ 'instance of ' ,receiver class name ]
		ifFalse: [ 'native object' ].
	^(message isString
		ifTrue: [ message asLiteral ]
		ifFalse: [ message selector ])
		,' not understood by ' ,type! !

! BlockClosure methodsFor: #accessing !
ms
	" Return the milliseconds taken to evaluate the receiver. "

	^DateTime millisecondsToRun: self! !

! Class class methodsFor: #export !
categories
	" Return a collection with all defined categories. "

	^(Smalltalk classes collect: [:each | each category ]) asHashedSet! !

! Class class methodsFor: #export !
uncategorized
	" Return a collection with classes without category. "

	^Smalltalk classes select: [:each | each category = #unclassified ]! !

! Class methodsFor: #instantiation !
variableByteSubclass: aString
	classVariableNames: classVarNames
	poolDictionaries: poolVarNames
	" Create a new subclass of the receiver. "

"	self note: 'WARNING: ',aString,' defined as variableByteSubclass:...'.
"	^self 
		subclass: aString
		instanceVariableNames: ''
		classVariableNames: classVarNames
		poolDictionaries: poolVarNames! !

! Class methodsFor: #instantiation !
variableSubclass: aString
	instanceVariableNames: aString2
	classVariableNames: classVarNames
	poolDictionaries: poolVarNames
	" Create a new subclass of the receiver. "

	self note: 'WARNING: ',aString,' defined as variableSubclass:...'.
	^self 
		subclass: aString
		instanceVariableNames: aString2
		classVariableNames: classVarNames
		poolDictionaries: poolVarNames! !

! Class methodsFor: #instantiation !
subclass: aString
	instanceVariableNames: aString2
	classVariableNames: classVarNames
	poolDictionaries: poolVarNames
	category: aCategory
	" Create a new subclass of the receiver. "

	(self smalltalk globalAt: aString) notNil ifTrue: [
		self note: 'WARNING: class REdefinition for ',aString.
		"^self smalltalk globalAt: aString"
	].

"	self note: 'Creating ',aString."
	^ClassBuilder new
		superclass: self
		subclass: aString
		instanceVariableNames: aString2
		classVariableNames: classVarNames
		poolDictionaries: poolVarNames
		category: aCategory! !

! Class methodsFor: #instantiation !
subclass: aString
	instanceVariableNames: aString2
	classVariableNames: classVarNames
	category: aCategory
	" Create a new subclass of the receiver. "

	^self subclass: aString
		instanceVariableNames: aString2
		classVariableNames: classVarNames
		poolDictionaries: ''
		category: aCategory! !

! Class methodsFor: #instantiation !
subclass: aString
	instanceVariableNames: aString2
	classVariableNames: classVarNames
	poolDictionaries: poolVarNames
	" Create a new subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: aString2
		classVariableNames: classVarNames
		poolDictionaries: poolVarNames
		category: nil! !

! Class methodsFor: #instantiation !
subclass: aString
	instanceVariableNames: aString2
	poolDictionaries: poolVarNames
	category: aCategory
	" Create a new subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: aString2
		classVariableNames: ''
		poolDictionaries: poolVarNames
		category: aCategory! !

! Class methodsFor: #instantiation !
variableByteSubclass: aString classVariableNames: classVarNames
	" Create a new subclass of the receiver. "

	^self
		variableByteSubclass: aString
		classVariableNames: classVarNames
		poolDictionaries: ''! !

! Class methodsFor: #instantiation !
variableByteSubclass: aString poolDictionaries: aString2
	" Create a new subclass of the receiver. "

	^self
		variableByteSubclass: aString
		classVariableNames: ''
		poolDictionaries: aString2!

variableByteSubclass: aString
	" Create a new subclass of the receiver. "

	^self
		variableByteSubclass: aString
		classVariableNames: ''! !

! Class methodsFor: #instantiation !
variableSubclass: aString
	instanceVariableNames: aString2
	classVariableNames: classVarNames
	" Create a new subclass of the receiver. "

	^self 
		variableSubclass: aString
		instanceVariableNames: aString2
		classVariableNames: classVarNames
		poolDictionaries: ''! !

! Class methodsFor: #instantiation !
variableSubclass: aString instanceVariableNames: aString2
	" Create a new subclass of the receiver. "

	^self 
		variableSubclass: aString
		instanceVariableNames: aString2
		classVariableNames: ''! !

! Class methodsFor: #instantiation !
variableSubclass: aString classVariableNames: aString2
	" Create a new subclass of the receiver. "

	^self 
		variableSubclass: aString
		instanceVariableNames: ''
		classVariableNames: aString2! !

! Class methodsFor: #instantiation !
variableSubclass: aString poolDictionaries: aString2
	" Create a new subclass of the receiver. "

	^self 
		variableSubclass: aString
		instanceVariableNames: ''
		classVariableNames: ''
		poolDictionaries: aString2! !

! Class methodsFor: #instantiation !
variableSubclass: aString
	" Create a new subclass of the receiver. "

	^self 
		variableSubclass: aString
		instanceVariableNames: ''! !

! Class methodsFor: #instantiation !
subclass: aString
	instanceVariableNames: aString2
	classVariableNames: classVarNames
	" Create a new subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: aString2
		classVariableNames: classVarNames
		poolDictionaries: ''! !

! Class methodsFor: #instantiation !
subclass: aString
	instanceVariableNames: aString2
	poolDictionaries: aString3
	" Create a new subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: aString2
		classVariableNames: ''
		poolDictionaries: aString3! !

! Class methodsFor: #instantiation !
subclass: aString instanceVariableNames: aString2
	" Create a new subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: aString2
		classVariableNames: ''! !

! Class methodsFor: #instantiation !
subclass: aString classVariableNames: aString2
	" Create a new subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: ''
		classVariableNames: aString2
		poolDictionaries: ''! !

! Class methodsFor: #instantiation !
subclass: aString poolDictionaries: aString2
	" Create a new subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: ''
		classVariableNames: ''
		poolDictionaries: aString2! !

! Class methodsFor: #instantiation !
subclass: aString
	" Create a new subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: ''! !

! UndefinedObject methodsFor: #instantiation!
category: ignored
	" BUG - throw an error in ClassBuilder. " ! !

! UndefinedObject methodsFor: #instantiation!
subclass: aString
	instanceVariableNames: aString2
	classVariableNames: classVarNames
	poolDictionaries: poolVarNames
	category: aCategory
	" Create a new subclass of the receiver. "

	(self smalltalk globalAt: aString) notNil ifTrue: [
		self note: 'WARNING: class REdefinition for ',aString.
		"^self smalltalk globalAt: aString"
	].

"	self note: 'Creating ROOT class ',aString."
	^ClassBuilder new
		superclass: self
		subclass: aString
		instanceVariableNames: aString2
		classVariableNames: classVarNames
		poolDictionaries: poolVarNames
		category: aCategory! !

! UndefinedObject methodsFor: #instantiation!
subclass: aString
	instanceVariableNames: aString2
	classVariableNames: classVarNames
	poolDictionaries: poolVarNames
	" Create a new subclass of the receiver. "

	^self
		subclass: aString
		instanceVariableNames: aString2
		classVariableNames: classVarNames
		poolDictionaries: poolVarNames
		category: nil! !

! Object methodsFor: #evaluating !
argumentCount
	" Return the number of arguments of the receiver. "

	^0! !

! Object methodsFor: #evaluating !
numberOfArguments
	" Return the number of arguments of the receiver. "

	^self argumentCount! !

! Object methodsFor: #evaluating !
evaluate
	" Return the result of evaluating the receiver. "

	^self! !

! Object methodsFor: #evaluating !
evaluateDeferred
	" Evaluating the receiver deferred in time.
	Return the receiver. "

	[ self evaluate ] valueDeferred.! !

! Object methodsFor: #evaluating !
evaluateWithArguments: anArray
	" Return the result of evaluating the receiver. "

	^self! !

! Object methodsFor: #evaluating !
evaluateWith: anObject
	" Return the result of evaluating the receiver. "

	^self evaluateWithArguments: (Array with: anObject)! !

! Object methodsFor: #evaluating !
logEvaluate: source
	" Return the result of evaluating the source.
	This method should be refined/replaced to log the source expression as reference of evaluation.
	"

	^self evaluate! !

! String methodsFor: #evaluating !
evaluateFor: anObject
	" Return the result of evaluating the receiver with anObject. "

	^anObject perform: self! !

! BlockClosure methodsFor: #evaluating !
evaluateFor: anObject
	" Return the result of evaluating the receiver with anObject. "

	^# #self: anObject! !

! BlockClosure methodsFor: #evaluating !
evaluate
	" Return the result of evaluating the receiver. "

	^# #self! !

! BlockClosure methodsFor: #evaluating !
evaluateWithArguments: anArray
	" Return the result of evaluating the receiver. "

	^self valueWithPossibleArguments: anArray! !

! BlockClosure methodsFor: #evaluating !
valueWithArguments: anArray
	" Return the result of evaluating the receiver with arguments. "

	^self valueWithPossibleArguments: anArray! !

Object
 subclass: #ActionSequence
 instanceVariableNames: 'contents'
 category: #Kernel!

! EvaluableAction methodsFor: #composing !
, anAction
	" Return the receiver composed with the argument. "

	^ActionSequence with: self with: anAction! !

! Object methodsFor: #converting !
asActionSequence
	" Return an ActionSequence from the receiver. "

	^ActionSequence with: self! !

! SequenceableCollection methodsFor: #converting !
asActionSequence
	" Return an ActionSequence from the receiver. "

	^ActionSequence withAll: self! !

! ActionSequence class methodsFor: #instantiation !
withAll: aCollection
	" Return an instance of the receiver. "

	^self new
		contents: aCollection;
		yourself! !

! ActionSequence class methodsFor: #instantiation !
with: anObject
	" Return an instance of the receiver. "

	^self withAll: (Array with: anObject)! !

! ActionSequence class methodsFor: #instantiation !
with: anObject with: bObject
	" Return an instance of the receiver. "

	^self withAll: (Array with: anObject with: bObject)! !

! ActionSequence class methodsFor: #instantiation !
with: anObject with: bObject with: cObject
	" Return an instance of the receiver. "

	^self withAll: (Array with: anObject with: bObject with: cObject)! !

! ActionSequence methodsFor: #private !
contents: aCollection
	" Private - Set the contents of the receiver. "

	contents := aCollection! !

! ActionSequence methodsFor: #private !
contents
	" Private - Return the contents of the receiver. "

	contents isNil ifTrue: [ contents := Array new ].
	^contents! !

! ActionSequence methodsFor: #testing !
isEmpty
	" Return true if the receiver is empty. "

	^contents isNil or: [ contents isEmpty ]! !

! ActionSequence methodsFor: #testing !
notEmpty
	" Return true if the receiver is not empty. "

	^contents notNil and: [ contents notEmpty ]! !

! ActionSequence methodsFor: #converting !
asArray
	" Return an array with the contents of the receiver. "

	self isEmpty ifTrue: [ ^#() ].
	^contents asArray! !

! ActionSequence methodsFor: #converting !
asActionSequence
	" Return an ActionSequence from the receiver. "

	^self! !

! ActionSequence methodsFor: #converting !
, anAction
	" Return an action composed with the receiver and anAction. "

	self isEmpty ifTrue: [ ^anAction asActionSequence ].
	^self class withAll: self contents , anAction asActionSequence contents! !

! ActionSequence methodsFor: #converting !
copyWith: anObject
	" Return an action composed with the receiver and anObject. "

	^self class withAll: (self contents copyWith: anObject)! !

! ActionSequence methodsFor: #converting !
do: aBlock
	" Evaluate aBlock for each object in the receiver's contents. "

	^self notEmpty ifTrue: [ self contents do: aBlock ]! !

! ActionSequence methodsFor: #testing !
includes: anObject
	" Return true if anObject is in the receiver. "

	^self notEmpty and: [ self contents includes: anObject ]! !

! ActionSequence methodsFor: #enumerating !
select: aBlock
	" Return the a copy of the receiver with elements wich evaluates aBlock to true. "

	^self class withAll: (self asArray select: aBlock)! !

! ActionSequence methodsFor: #enumerating !
reject: aBlock
	" Return the a copy of the receiver with elements wich evaluates aBlock to false. "

	^self class withAll: (self asArray reject: aBlock)! !

! ActionSequence methodsFor: #evaluating !
evaluate
	" Return the result of evaluating the receiver. "

	| result |
	self do: [:each| result := each evaluate ].
	^result! !

! ActionSequence methodsFor: #evaluating !
evaluateWithArguments: anArray
	" Return the result of evaluating the receiver. "

	| result |
	self do: [:each| result := each evaluateWithArguments: anArray ].
	^result! !

! ActionSequence methodsFor: #evaluating !
evaluateFor: anObject
	" Return the result of evaluating the receiver with anObject. "

	| result |
	self do: [:each| result := each evaluateFor: anObject ].
	^result! !

! ActionSequence methodsFor: #events !
reduced
	" Return the receiver reduced to minimal evaluable expression. "

	^self notEmpty ifTrue: [
		contents size = 1
		ifTrue: [ contents first ]
		ifFalse: [ self ]
	]! !

! ActionSequence  methodsFor: #changes !
evaluateWithSelector: aSelector withArguments: array
	" Private - Return the result of evaluating the receiver with aSelector and arguments array. "

	| result |
	self do: [:each| result := each receiver perform: aSelector withArguments: array ].
	^result! !

! ActionSequence methodsFor: #resolution !
resolveWith: aPoolOrBlock
	" Return an instance set to be evaluated with mappings given by aPoolOrBlock. "

	^self class withAll: (self asArray collect: [:each|
		(each respondsTo: #resolveWith:)
		ifTrue: [ each resolveWith: aPoolOrBlock ]
		ifFalse: [ each ]
	])! !

Object
	subclass: #EventManager
	instanceVariableNames:  ' handlers '
	category: #Kernel!
EventManager class instanceVariableNames: ' eventsTriggered '!
EventManager comment: '
	Event system, properties and change/update/release mechanism.
	We hook $properties and $handlers internal properties to manage properties and events at Object level.
	EventManager instances implement events handler as pure instance variable.
'!

! EventManager class methodsFor: #events !
eventsTriggered
	" Return all the events managed by instances of the receiver. "

	eventsTriggered isNil ifTrue: [ self initializeEventsTriggered ].
	^eventsTriggered! !

! EventManager class methodsFor: #events !
eventsTriggered: aCollection
	" Set all the events managed by instances of the receiver. "

	eventsTriggered := aCollection! !

! EventManager class methodsFor: #events !
initializeEventsTriggered
	" Private - Initialize all the events managed by instances of the receiver. "

	self eventsTriggered: self constructEventsTriggered! !

! EventManager methodsFor: #events !
eventTable
	" Private - Return the mapping event names to actions of the receiver. "

	^handlers isNil
		ifTrue: [ self eventTableCreate ]
		ifFalse: [ handlers ]! !

! EventManager methodsFor: #events !
eventTableForEdit
	" Private - Return the mapping event names to actions of the receiver.
	Set the internal map if not present.
	"

	handlers isNil ifTrue: [ handlers := self eventTableCreate ].
	^handlers! !

! EventManager methodsFor: #events !
releaseEventTable
	" Private - Release the event table. "

	handlers := nil! !

! Object methodsFor: #events !
eventTable
	" Private - Return the mapping event names to actions of the receiver. "

	^self basicAt: '@handlers' ifAbsent: [ self eventTableCreate ]! !

! Object methodsFor: #events !
eventTableForEdit
	" Private - Return the mapping event names to actions of the receiver.
	Set the internal map if not present.
	"

	^self basicAt: '@handlers' ifAbsent: [
		self basicAt: '@handlers' put: self eventTableCreate
	]! !

! Object methodsFor: #events !
releaseEventTable
	" Private - Release the event table. "

	self basicDelete: '@handlers'! !

! Object methodsFor: #release !
release
	" Release the receiver. "

	self releaseEventTable ! !

! Object methodsFor: #events !
eventTableCreate
	" Private - Return a new mapping for the receiver. "

	^PoolDictionary new! !

! Object class methodsFor: #events !
eventsTriggered
	" Return all the events managed by instances of the receiver. "

	^self constructEventsTriggered! !

! Object class methodsFor: #events !
constructEventsTriggered
	" Private - Return the events managed by instances of the receiver.
	The default implementation return support for MVC change/update mechanism.
	This method can be refined by the subclasses to extend event support.
	"

	^#( #changed )! !

! Object class methodsFor: #events !
canTriggerEvent: selector
	" Return true if instances of the receiver can trigger event."

	^self eventsTriggered includes: selector ! !

! Object methodsFor: #events !
canTriggerEvent: selector
	" Return true if the receiver can trigger event."

	^self class canTriggerEvent: selector ! !

! Object methodsFor: #events !
invalidEvent: selector
	" Private - Notify the situation where an unknown event is registered.
	Default implementation ignore error during unknown event registration.
	"

	^self
"DEBUG-	^self note: self class name,' invalidEvent: ' ,selector asLiteral "! !

! Object methodsFor: #events !
when: selector evaluate: anAction
	" Register anAction to the be evaluated when the receiver triggers the event. "

	| actions |
	(self canTriggerEvent: selector) ifFalse: [ self invalidEvent: selector ].
	actions := self actionListForEvent: selector.
	(actions includes: anAction) ifTrue: [ ^self ].
	self
		setActionList: (actions copyWith: anAction)
		forEvent: selector! !

! Object methodsFor: #events !
actionListForEvent: selector
	" Private - Return a collection of actions to evaluate when the event is triggered. "

	| result |
	result := self eventTable at: selector ifAbsent: [ ^ActionSequence new ].
	^result asActionSequence! !

! Object methodsFor: #events !
setActionList: actionSequence forEvent: selector
	" Private - Set the list of actions to evaluate when the event is triggered by the receiver. "

	| action |
	action := actionSequence asActionSequence reduced.
	action isNil ifTrue: [ ^self removeActionsForEvent: selector ].
	self eventTableForEdit at: selector put: action ! !

! Object methodsFor: #events !
removeActionsSatisfying: aBlock forEvent: selector
	" Private - Remove all actions for the event that evaluates aBlock to true."

	self
		setActionList: ((self actionListForEvent: selector) asArray reject: aBlock)
		forEvent: selector ! !

! Object methodsFor: #events !
removeActionsWithReceiver: anObject forEvent: selector
	" Remove all actions for the event in the receiver's event table which have anObject as receiver. "

	self
		removeActionsSatisfying: [:each| (each respondsTo: #receiver) and: [ each receiver == anObject ] ]
		forEvent: selector ! !

! Object methodsFor: #events !
removeAction: anAction forEvent: selector
	" Remove anAction for the event in the receiver's event table which have anObject as receiver. "

	self
		removeActionsSatisfying: [:each| each = anAction ]
		forEvent: selector ! !

! Object methodsFor: #events !
when: selector do: aBlock
	" Evaluate aBlock when the receiver triggers the event. "

	self when: selector evaluate: aBlock! !

Message
 subclass: #LinkMessage
 instanceVariableNames: 'resultAction'
 category: #Kernel!
LinkMessage comment: '
	Implement evaluation cascading argument evaluations.'!

! LinkMessage methodsFor: #accessing !
resultAction
	" Return the resultAction of the receiver. "

	^resultAction! !

! LinkMessage methodsFor: #accessing !
resultAction: anObject
	" Set the resultAction of the receiver. "

	resultAction := anObject! !

! LinkMessage methodsFor: #evaluating !
evaluate
	" Return the result of evaluating the receiver. "

	^self evaluateWithArguments: #()! !

! LinkMessage methodsFor: #evaluating !
collectArguments: evaluationArguments
	" Private - Return the arguments to evaluate the receiver. "

	^self arguments collect: [:each| each evaluateWithArguments: evaluationArguments ]! !

! Object methodsFor: #events !
when: selector send: message to: anObject withArguments: array
	" Send the message when the receiver triggers the event. "

	self
		when: selector
		evaluate: (LinkMessage
			receiver: anObject
			selector: message
			arguments: array)! !

! Object methodsFor: #events !
when: selector send: message to: anObject
	" Send the message when the receiver triggers the event. "

	self
		when: selector
		evaluate: (Message
			receiver: anObject selector: message
			arguments: (Array new: message argumentCount))! !

! Object methodsFor: #events !
triggerEvent: selector ifNotHandled: aBlock
	" Trigger the event.
	If the event is not handled, return the value of evaluating aBlock,
        or the value returned by the most recently defined event handler.
	"

	^(self eventTable at: selector ifAbsent: [ aBlock ]) evaluate! !

! Object methodsFor: #events !
actionForEvent: selector
	" Return the action to evaluate when the event is triggered by the receiver (or nil). "

	^self eventTable at: selector ifAbsent: [ nil ]! !

! Object methodsFor: #events !
triggerEvent: selector
	" Return the result of triggering the event. "

	^(self actionForEvent: selector) evaluate! !

! Object methodsFor: #events !
triggerEvent: selector withArguments: array
	" Return the result of triggering the event. "

	^(self actionForEvent: selector) evaluateWithArguments: array! !

! Object methodsFor: #events !
triggerEvent: selector with: anObject
	" Return the result of triggering the event. "

	^self triggerEvent: selector withArguments: (Array with: anObject)! !

! Object methodsFor: #events !
triggerEvent: selector with: anObject with: anotherObject
	" Return the result of triggering the event. "

	^self triggerEvent: selector withArguments: (Array with: anObject with: anotherObject)! !

! Object methodsFor: #events !
triggerEvent: selector withArguments: array ifNotHandled: aBlock
	" Return the result of triggering the event with the arguments.
	If the event is not handled, return the value of evaluating aBlock,
	or the value returned by the most recently defined event handler action.
	"

	^(self eventTable at: selector ifAbsent: [ aBlock ]) evaluateWithArguments: array! !

! Object methodsFor: #events !
triggerEvent: selector with: anObject ifNotHandled: aBlock
	" Return the result of triggering the event with anObject. "

	^self triggerEvent: selector withArguments: (Array with: anObject) ifNotHandled: aBlock! !

! Object methodsFor: #events !
hasActionForEvent: selector
	" Return true if the receiver has an action registered for the event. "

	^self eventTable includesKey: selector ! !

! Object methodsFor: #events !
eventsHandled
	" Return the events for which there are actions registered in the receiver. "

	^self eventTable keys! !

! Object methodsFor: #events !
removeAllActionsWithReceiver: anObject
	" Remove all actions for all events in the receiver's event table with anObject as receiver. "

	self eventTable keys do: [:each|self removeActionsWithReceiver: anObject forEvent: each ]! !

! Object methodsFor: #events !
removeActionsForEvent: selector
	" Remove all actions for the event. "

	| actions |
	actions := self eventTable.
	actions removeKey: selector ifAbsent: [ ^self ].
	actions isEmpty ifTrue: [ self releaseEventTable ].! !

! Object methodsFor: #changes !
addDependent: anObject
	" Make anObject dependent of the receiver. "

	self when: #changed send: #update to: anObject! !

! Object methodsFor: #changes !
changed
	" Update the dependents of the receiver. "

	self triggerEvent: #changed! !

! Object methodsFor: #changes !
updateEvent: selector selector: message arguments: array
	" Private - Trigger the event with the message selector and the arguments. "

	| action |
	action := self eventTable at: selector ifAbsent: [ ^nil ].
	action asActionSequence
		evaluateWithSelector: message
		withArguments: array ! !

! Object methodsFor: #changes !
changed: aspect
	" Update the dependents on aspect of the receiver. "

	self	updateEvent: #changed selector: #update:
		arguments: (Array with: aspect)! !

! Object methodsFor: #changes !
changed: aspect with: anObject
	" Update the dependents on aspect of the receiver. "

	self	updateEvent: #changed selector: #update:with:
		arguments: (Array with: aspect with: anObject)! !

! Object methodsFor: #changes !
changed: aspect with: anObject with: anotherObject
	" Update the dependents on aspect of the receiver. "

	self	updateEvent: #changed selector: #update:with:with:
		arguments: (Array with: aspect with: anObject with: anotherObject)! !

! Object methodsFor: #changes !
dependents
	" Return the dependents of the receiver. "

	^(self actionListForEvent: #changed) collect: [:each | each receiver ]! !

! Object methodsFor: #changes !
update: anObject with: anotherObject
	" Update the receiver.
	Default implementation do nothing.
	This method must be refined by subclasses to support change/update protocoll.
	"! !

! Object methodsFor: #changes !
update: anObject
	" Update the receiver.
	Default implementation do nothing.
	This method must be refined by subclasses to support change/update protocoll.
	"! !

! Object methodsFor: #printing !
printingWith: printing do: aBlock
	" Return the result of evaluating aBlock, redirecting output to printing block.
	Note: native print() will be overwritten with printing block and restored to original after aBlock is evaluated.
	"

	| old result |
	{'self' doc'native printing'}.
	[	old := Smalltalk install: printing as: #print.
		result := aBlock value.
	] ensure: [ Smalltalk install: old as: #print ].
	^result! !

! Behavior methodsFor: #private !
aboutToAdd: aMethod
	" Private - Prepare aMethod to be added onto the receiver. " ! !

! Behavior methodsFor: #private !
canAdd: isDoIt method: aMethod
	" Private - Returns true if aMethod be added onto the receiver. "

	(OverwriteMethods notNil and: [ OverwriteMethods ]) ifFalse: [
		(self implements: aMethod selector) ifTrue: [
			isDoIt ifFalse: [
				(DumpOverwritenMethods notNil
				and: [ DumpOverwritenMethods
				and: [ Exporter notNil ]]) ifTrue: [
					self note: '? ',self name,'>>#',aMethod selector.
					self print: (Exporter new export: self method: aMethod).
				] ifFalse: [
					self note: '-Overwrite Conflict- ',self name,'>>#',aMethod selector.
				].
				^false
			].
		].
	].
	self aboutToAdd: aMethod.
	^true ! !

! Behavior methodsFor: #private !
basicAddCompiledMethod: aMethod
	" Private - Register aMethod onto the receiver. "

	self smalltalk
		#addMethod: aMethod selector asSelector
		method: aMethod to: self! !

! Behavior methodsFor: #methods !
addCompiledMethod: aMethod
	" Add the compiled method, if appropiate, and register the method in recent methods. "

	| isDoIt |
	isDoIt := aMethod selector = String doItSelector.
	(self canAdd: isDoIt method: aMethod) ifFalse: [ ^nil ].
	(Exporter notNil and: [ isDoIt not ]) ifTrue: [
" DEBUG -	self note: self name,'>>#',aMethod selector."
		self print: (Exporter new export: self method: aMethod).
	].
	^self basicAddCompiledMethod: aMethod! !

ArrayedContents
	subclass: #ComposedObjects
	instanceVariableNames: 'species'
	category: #Kernel !

! ComposedObjects methodsFor: #composing !
, anObject
	" Return the receiver composed with the argument. "

	anObject isNil ifTrue: [ ^self ].
	(anObject isArray or: [ anObject respondsTo: #asArray ]) ifTrue: [
		anObject asArray do: [:each| self , each ].
		^self
	].

	(species notNil and: [ (anObject isKindOf: species) not ]) ifTrue: [
		self error: 'Invalid object to compose'
	].

	self add: anObject.! !

! ComposedObjects methodsFor: #accessing !
species: aBehavior
	" Set the species to compose in the receiver. "

	species := aBehavior! !

! Behavior methodsFor: #composing !
collection
	" Return a collection to compose instances of the receiver. "

	^ComposedObjects new
		species: self;
		yourself! !

Object
 subclass: #DateTime
 instanceVariableNames: ''
 category: #Kernel!

! DateTime class methodsFor: #instantiation !
new: anObject
	" Return an instance of the receiver. "

	{'' doc'native instantiation'}! !

! DateTime class methodsFor: #instantiation !
fromString: aString
	" Return an instance of the receiver.
	DateTime fromString: '1962/12/15 01:40:00'
	"

	^self new: aString! !

! DateTime class methodsFor: #instantiation !
fromSeconds: aNumber
	" Return an instance of the receiver. "

	^self fromMilliseconds: aNumber * 1000! !

! DateTime class methodsFor: #instantiation !
fromMilliseconds: aNumber
	" Return an instance of the receiver. "

	^self new: aNumber! !

! DateTime class methodsFor: #instantiation !
today
	" Return an instance of the receiver. "

	{'' doc'native instantiation'}.
	^self new! !

! DateTime class methodsFor: #instantiation !
now
	" Return an instance of the receiver. "

	{'' doc'native instantiation'}.
	^self today! !

! DateTime class methodsFor: #instantiation !
msNow
	" Private - Return a magnitude to measure milliseconds. "

	{'' doc'native instantiation'}.
	^self new! !

! DateTime class methodsFor: #instantiation !
millisecondsToRun: aBlock
	" Return the time elapsed during evaluation of aBlock. "

	| start |
	start := self msNow.
	aBlock value.
	^self msNow - start! !

! DateTime methodsFor: #accessing !
year
	" Return the year of the receiver. "

	^{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
month
	" Return the month of the receiver. "

	^{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
month: aNumber
	" Set the month of the receiver. "

	{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
day
	" Return the day of the receiver. "

	^{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
dayOfWeek
	" Return the day of week of the receiver. "

	self invalidMessage: #dayOfWeek.! !

! DateTime methodsFor: #accessing !
dayOfWeek: aNumber
	" Set the day of week of the receiver. "

	self invalidMessage: #dayOfWeek:.! !

! DateTime methodsFor: #accessing !
day: aNumber
	" Set the day of the receiver. "

	{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
year: aNumber
	" Set the year of the receiver. "

	{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
dayOfMonth
	" Return the day of month of the receiver. "

	^{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
dayOfMonth: aNumber
	" Set the day of month of the receiver. "

	{'' doc'native accessing'}! !

! DateTime methodsFor: #converting !
asString
	" Return the text representation of the receiver. "

	{'self' doc'native conversion'}.
	^self toString! !

! DateTime methodsFor: #printing !
printString
	" Return the printable representation of the receiver. "

	^self asString! !

! DateTime methodsFor: #converting !
asMilliseconds
	" Return the receiver as milliseconds. "

	^self time! !

! DateTime methodsFor: #accessing !
time
	" Return the time of the receiver. "

	^{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
time: aNumber
	" Set the time of the receiver. "

	{'' doc'native accessing'}! !

! DateTime methodsFor: #converting !
asDateString
	" Return the date string of the receiver. "

	^{'' doc'native converting'}! !

! DateTime methodsFor: #converting !
asTimeString
	" Return the time string of the receiver. "

	^{'' doc'native converting'}! !

! DateTime methodsFor: #converting !
asLocaleString
	" Return the locale string of the receiver. "

	^{'' doc'native converting'}! !

! DateTime methodsFor: #converting !
asNumber
	" Return the receiver as a Number. "

	^self asMilliseconds! !

! DateTime methodsFor: #accessing !
hours: aNumber
	" Set the hours of the receiver. "

	{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
minutes: aNumber
	" Set the minutes of the receiver. "

	{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
seconds: aNumber
	" Set the seconds of the receiver. "

	{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
milliseconds: aNumber
	" Set the milliseconds of the receiver. "

	{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
hours
	" Return the hours of the receiver. "

	^{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
minutes
	" Return the minutes of the receiver. "

	^{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
seconds
	" Return the seconds of the receiver. "

	^{'' doc'native accessing'}! !

! DateTime methodsFor: #accessing !
milliseconds
	" Return the milliseconds of the receiver. "

	^{'' doc'native accessing'}! !

! DateTime methodsFor: #comparing !
< aDateTime
	" Return true if the receiver is smaller than argument. "

	^{'' doc'native compare'}! !

! DateTime methodsFor: #comparing !
> aDateTime
	" Return true if the receiver is greater than argument. "

	^{'' doc'native compare'}! !

! DateTime methodsFor: #comparing !
<= aDateTime
	" Return true if the receiver is smaller or equal to argument. "

	^{'' doc'native compare'}! !

! DateTime methodsFor: #comparing !
>= aDateTime
	" Return true if the receiver is greater or equal to argument. "

	^{'' doc'native compare'}! !

! DateTime methodsFor: #arithmetic !
- aDateTime
	" Return the result of operation. "

	^{'' doc'native operation'}! !

! DateTime methodsFor: #arithmetic !
+ aDateTime
	" Return the result of operation. "

	^{'' doc'native operation'}! !

! DateTime methodsFor: #testing !
isValid
	" Return true if the receiver is a valid instance. "

	^{'' doc'native testing'}! !

! DateTime methodsFor: #json !
asJSONObject
	" Return the receiver as a hash object. "

	{'self' doc'native converting'}.
	^self! !